proc oval_pos {oval args} {
    unpack [.canvas coords $oval] x1 y1 x2 y2
    set xr [expr ($x2-$x1)/2]
    set yr [expr ($y2-$y1)/2]
    set x [expr $x1+$xr]
    set y [expr $y1+$yr]
    if {$args != ""} {
	unpack $args x y xr yr
	set x1 [expr $x-$xr]
	set x2 [expr $x+$xr]
	set y1 [expr $y-$yr]
	set y2 [expr $y+$yr]
	.canvas coords $oval $x1 $y1 $x2 $y2
    }
    return [list $x $y $xr $yr]
}

proc image_zoom {img up args} {
    set down $args
    image create photo image_zoom_temp
    image_zoom_temp copy $img -zoom $up
    if {$down != ""} {
	$img configure -width [expr [image width image_zoom_temp]/$down]
	$img configure -height [expr [image height image_zoom_temp]/$down]
	$img copy image_zoom_temp -subsample $down -shrink
    } else {
	$img configure -width [image width image_zoom_temp]
	$img configure -height [image height image_zoom_temp]
	$img copy image_zoom_temp
    }
    image delete image_zoom_temp
}

proc matrix_zoom {m image zoom} {
    if {$zoom == 1} {
	matrix image $m $image
    } else {
	image create photo matrix_zoom_temp
	matrix image $m matrix_zoom_temp
	#puts stderr "->image copy"
	$image copy matrix_zoom_temp -zoom $zoom
	#puts stderr "<-image copy"
	image delete matrix_zoom_temp
    }
}

proc matrix_row {m row args} {
    if {![llength $args]} {
	set result {}
	for {set i 0} {$i < [matrix cols $m]} {incr i} {
	    lappend result [matrix at $m $row $i]
	}
	return $result
    } else {
	set i 0
	foreach v $args {
	    matrix at $m $row $i $v
	    incr i
	}
	return $args
    }
}

proc matrix_append {m args} {
    if {[llength $args] != [matrix cols $m]} {
	error "[llength $args != [matrix cols $m]"
    }
    set new_m [matrix create [expr 1+[matrix rows $m]] [matrix cols $m]]
    matrix move $new_m $m
    eval matrix_row $new_m [matrix rows $m] $args
    return $new_m
}

proc shuffle lst {
    upvar $lst a
    set dists ""
    foreach m $a {
	lappend dists [ran]
    }
    sort -real dists a
    set a $a
}

# copy the values of array srcVar into array destVar
proc arrayCopy {destVar srcVar} {
    upvar $srcVar src $destVar dest
    foreach n [array names src] {
	set dest($n) $src($n)
    }
}

proc max {a b} {
    if {$a < $b} {return $b} else {return $a}
}
proc min {a b} {
    if {$a < $b} {return $a} else {return $b}
}
# Greatest positive common factor. E.g. gcd(12, -18) == 6.
proc gcd {a b} {
    if {$b == 0} {set g $a} else {set g [gcd $b [expr $a % $b]]}
    return [expr abs($g)]
}

# repeat a value k times
# e.g.  repeat hello 3
proc repeat {v k} {
    set result ""
    for {set i 0} {$i < $k} {incr i} {
	append result $v
    }
    return $result
}
# same thing but returns a list
proc lrepeat {v k} {
    set result ""
    for {set i 0} {$i < $k} {incr i} {
	lappend result $v
    }
    return $result
}

# Return a list containing the integers from start to end
proc range {start end} {
    set lst ""
    for {set i $start} {$i <= $end} {incr i} {
	lappend lst $i
    }
    return $lst
}

# unpack a list into variables
# e.g.  unpack {3 4 5} a b c
# slow; should be part of the language
proc unpack {list args} {
    # args is a list of vars in caller's scope
    set len [expr [min [llength $list] [llength $args]] - 1]
    for {set i 0} {$i < $len} {incr i} {
	upvar [lindex $args $i] x
	set x [lindex $list $i]
    }
    # last variable eats the rest of the list
    upvar [lindex $args $len] x
    set x [lrange $list $len end]
}

# returns a fresh, obscure variable name in the global scope
proc tmpnam {} {
    for {set i 0} {[uplevel #0 info exists tmp!nam$i]} {incr i} {}
    return tmp!nam$i
}

# evaluates code in a local scope
# must use an obscure argument name to avoid clash
proc block {co!de} {
    eval ${co!de}
}

# utility functions for switching modes
# mutates a list of events into a list of event/binding pairs
proc get_bindings {widget var} {
  upvar $var v
  foreach event $v {
    lappend v2 [list $event [bind $widget <$event>]]
  }
  uplevel set $var [list $v2]
}
# uses a list of event/bindings pairs to set bindings en masse
proc set_bindings {widget var} {
  upvar $var v
  foreach p $v {
    bind $widget <[lindex $p 0]> [lindex $p 1]
  }
}

# Create an image viewer widget, with zooming, in the style of rleshow.
# Based on code from Kris Popat.
# Examples:
#   imgview .v -file foo.ppm
#   imgview .v -palette 256 -format {raw 256 256 1} -file foo.raw
#   imgview .v -image my_image
#   pack .v
proc imgview {name args} {
    upvar #0 $name v
    set config ""
    while {[llength $args]} {
	set option [lindex $args 0]
	set value [lindex $args 1]
	set args [lrange $args 2 end]
	switch -- $option {
	    -file {set v(image) [eval image create photo $config -file $value]}
	    -image {set v(image) $value}
	    default {lappend config $option $value}
	}
    }
    set v(buffer) [eval image create photo $config]
    $v(buffer) copy $v(image)
    eval label $name -image $v(buffer)
    # try to get image updates
    global tk_version
    if {$tk_version >= 4.1} {
	catch {image_notify $v(image) $name "imgview_update $name"}
    }
    set v(zoom) 1
    set v(xpos) 0
    set v(ypos) 0
    set v(width) [image width $v(image)]
    set v(height) [image height $v(image)]
    set v(x1) 0
    set v(y1) 0
    set v(x2) $v(width)
    set v(y2) $v(height)
    bind $name <Button-1> "imgview_inczoom $name 1 %x %y"
    bind $name <Button-2> "imgview_inczoom $name 0 %x %y"
    bind $name <Button-3> "imgview_inczoom $name -1 %x %y"
}
# callbacks used by imgview
proc imgview_update {name x y width height} {
    upvar #0 $name v
    # if the image has not changed size, copy only the updated area
    # otherwise recompute zoom params and refresh
    set img_width [image width $v(image)]
    set img_height [image height $v(image)]
    if {$img_width == $v(width) && $img_height == $v(height)} {
	# copy only the updated area
	set xw [expr $x + $width]
	set yh [expr $y + $height]
	# clip to viewing area
	set x1 [max $v(x1) [min $v(x2) $x]]
	set y1 [max $v(y1) [min $v(y2) $y]]
	set x2 [max $v(x1) [min $v(x2) $xw]]
	set y2 [max $v(y1) [min $v(y2) $yh]]
	# destination
	set dx [expr ($x1 - $v(x1)) * $v(zoom)]
	set dy [expr ($y1 - $v(y1)) * $v(zoom)]
	$v(buffer) copy $v(image) \
		-from $x1 $y1 $x2 $y2 -to $dx $dy -zoom $v(zoom)
    } else {
	set v(width) $img_width
	set v(height) $img_height
	imgview_inczoom $name 0 $v(xpos) $v(ypos)
    }
}
proc imgview_inczoom {name increment xpos ypos} {
    upvar #0 $name v

    $name config -cursor watch

    ## First we figure out the x and y position of the cursor, relative
    ## to the original (unzoomed) image.  Note that $xpos and $ypos
    ## give the cursor position in the displayed (possibly zoomed) image.
    set xposeff [expr int($xpos / $v(zoom) + $v(x1))]
    set yposeff [expr int($ypos / $v(zoom) + $v(y1))]

    ## Now we update the zoom factor, making sure it's never less than 1.
    set v(zoom) [max 1 [expr $v(zoom) + $increment]]

    set img_width [image width $v(image)]
    set img_height [image height $v(image)]
    set width [expr int($img_width / $v(zoom))]
    set height [expr int($img_height / $v(zoom))]

    set v(x1) [expr $xposeff - int(0.5 * $width)]
    set v(x1) [max 0 [min $v(x1) $img_width]]
    set v(x2) [expr $v(x1) + $width]
    if {$v(x2) > $img_width} {
	set v(x2) $img_width
	set v(x1) [expr $v(x2) - $width]
    }

    set v(y1) [expr $yposeff - int(0.5 * $height)]
    set v(y1) [max 0 [min $v(y1) $img_height]]
    set v(y2) [expr $v(y1) + $height]
    if {$v(y2) > $img_height} {
	set v(y2) $img_height
	set v(y1) [expr $v(y2) - $height]
    }
    # save the current viewing position, in case the image size changes
    set v(xpos) [expr int(0.5 * $width) * $v(zoom)]
    set v(ypos) [expr int(0.5 * $height) * $v(zoom)]
    $v(buffer) copy $v(image) -from $v(x1) $v(y1) $v(x2) $v(y2) -zoom $v(zoom)

    $name config -cursor crosshair
}

# create a canvas with bindings for placing and moving points
# example:
#   pointcanvas .c -background black -foreground blue -radius 5
#   pack .c -fill both -expand 1
proc pointcanvas {name args} {
    upvar #0 $name v
    set v(radius) 5
    set v(foreground) blue
    set v(highlight) red
    set v(outline) white
    set v(add_cb) list
    set v(move_cb) list
    set v(delete_cb) list

    set config ""
    while {[llength $args]} {
	set option [lindex $args 0]
	set value [lindex $args 1]
	set args [lrange $args 2 end]
	switch -- $option {
	    -radius {set v(radius) $value}
	    -foreground {set v(foreground) $value}
	    -highlight {set v(highlight) $value}
	    -add_cb {set v(add_cb) $value}
	    -move_cb {set v(move_cb) $value}
	    -delete_cb {set v(delete_cb) $value}
	    default {lappend config $option $value}
	}
    }

    eval canvas $name $config

    # make the points light up as we pass over them
    $name bind point <Any-Enter> "$name itemconfig current -fill $v(highlight)"
    $name bind point <Any-Leave> "$name itemconfig current -fill $v(foreground)"

    # add points with button 1
    bind $name <1> {
	set name %W
	set x [$name canvasx %x]
	set y [$name canvasy %y]
	set id [pointcanvas_add $name $x $y]
    }

    # allow points to be grabbed and moved with button 2
    $name bind point <2> {
	set name %W
	upvar #0 $name v
	$name addtag selected withtag current
	$name raise current
	set v(lastX) %x
	set v(lastY) %y
    }

    $name bind point <ButtonRelease-2> {
	%W dtag selected
    }
    bind $name <B2-Motion> {
	set name %W
	upvar #0 $name v
	set x %x
	set y %y
	set id [$name find withtag selected]
	set tags [$name gettags $id]
	# only move points
	if {[lsearch $tags point] != -1} {
	    $name move selected [expr $x-$v(lastX)] [expr $y-$v(lastY)]
	    $v(move_cb) $id [$name canvasx $x] [$name canvasy $y]
	}
	set v(lastX) $x
	set v(lastY) $y
    }

    # remove points with button 3
    $name bind point <3> {
	set name %W
	upvar #0 $name v
	set id [$name find withtag current]
	set tags [$name gettags $id]
	# only delete points
	if {[lsearch $tags point] != -1} {
	    $name delete current
	    $v(delete_cb) $id
	}
    }
}

# Add a point centered at (x,y) to the pointcanvas
proc pointcanvas_add {name x y} {
    upvar #0 $name v
    set id [$name create oval \
	    [expr $x-$v(radius)] [expr $y-$v(radius)] \
	    [expr $x+$v(radius)] [expr $y+$v(radius)] \
	    -width 1 -outline $v(outline) -fill $v(foreground) -tags point]
    $v(add_cb) $id $x $y
    return $id
}

proc bbox_center {bbox} {
    unpack $bbox x1 y1 x2 y2
    return "[expr ($x2+$x1)/2] [expr ($y2+$y1)/2]"
}

# Return a list of point centers on the pointcanvas
proc pointcanvas_points {name} {
    set output ""
    foreach pt [$name find withtag point] {
	lappend output [bbox_center [$name coords $pt]]
    }
    return $output
}

# load a library by searching through user's LD_LIBRARY_PATH
proc load-lib lib {
    global env
    if [info exists env(LD_LIBRARY_PATH)] {
	set dirs [split $env(LD_LIBRARY_PATH) :]
	foreach d $dirs {
	    set fname $d/lib$lib[info sharedlibextension]
	    if [file exists $fname] {
		#puts stderr $fname
		set code [catch {load $fname} msg]
                # this catches "couldn't find procedure %_Init"
		if {[lindex $msg 1] == "find"} return
		return -code $code $msg
	    }
	}
    }
    set code [catch {load lib$lib[info sharedlibextension] "%"} msg]
    if {[lindex $msg 1] == "find"} return
    return -code $code $msg
}
# call load-lib on all of the arguments
proc load-libs args {
    foreach lib $args {load-lib $lib}
}

proc get_build {} {
    global build env
    set build ""
    if {[info exists env(BUILD)]} {
	if {$env(BUILD) == "debug"} {set build "-mc"}
    }
    #puts stderr "using build = $build"
    if {![info exists env(PKG)]} {
	# is there a "pkg" in the current directory?
	if {[file exists "pkg"]} {
	    set env(PKG) "pkg"
	} else {
	    set env(PKG) ~tpminka/pkg
	}
    }
}

# pkg is the name of a directory under $env(PKG), e.g. tpm, boa, lapack.
# It can also be a prefix, e.g. jpeg for jpeg-6a.
# In any case, libpkg.so must be the library name.
proc require pkg {
    global build env required
    if {[info exists required($pkg)]} {return}
    set required($pkg) 1
    #puts stderr "require $pkg"
    if {![info exists build]} {get_build}
    set dirnames [glob $env(PKG)/$pkg*]
    foreach dir $dirnames {
	set env(LD_LIBRARY_PATH) $dir/lib/$env(MACHTYPE):$env(LD_LIBRARY_PATH)
	#puts $env(LD_LIBRARY_PATH)

	# try to source an autoload script
	set code [catch {set f [open $dir/src/tcl/tclIndex]}]
	if $code {
	    # special cases
	    switch $pkg {
		lapack {load-libs f2c blas}
	    }
	} else {
	    global auto_index
	    eval [read $f]
	    close $f
	}

	# try to load the library
	set code [catch {load-lib $pkg$build} msg]
	if {$code == 0} return

	# try to load the library without $build
	set code [catch {load-lib $pkg} msg]
	if {$code == 0} {return}
    }
    return -code $code $msg
}

proc load_ph {} {
    global build env
    require learn
    load $env(PHOTOBOOK_ROOT)/lib/$env(MACHTYPE)/libph$build[info sharedlibextension]
}

proc load-formats {} {
    global env
    set no_jpeg [expr [info exists env(NO_JPEG)] || [catch {require jpeg}]]
    set no_tiff [expr [info exists env(NO_TIFF)] || [catch {require tiff}]]
    set no_sgi [expr [info exists env(NO_SGI)] || [catch {require gl_image}]]
    # are we running Tk?
    global tk_version
    if {![info exists tk_version]} {return}
    if {[catch {formatRaw}]} {
	global auto_path
	auto_reset
	lappend auto_path $env(PKG)/tpmtcl/lib
	#lappend auto_path $env(PKG)/tpmtcl/lib/common
        formatRaw
    }
    if {! $no_jpeg} {formatJpeg}
    if {! $no_tiff} {formatTiff}
    if {! $no_sgi} {formatSgi}
}

if {[info exists env(MACHTYPE)] && $env(MACHTYPE) == "sun4-solaris"} {
    catch {load-lib gcc}
}
