#!/usr/local/bin/wish
# Clustering Tutorial
# operations:
#
# add/move/erase points
# load/save a point file
# load and display a tree file
# generate a tree

# todo:
# fast neighbors
# multiple cluster algs - each has own config window

set point(radius) 5

# Canvas bindings #############################################################
canvas .canvas
pack .canvas -fill both -expand 1

# make the points light up as we pass over them
.canvas bind point <Any-Enter> {.canvas itemconfig current -fill red}
.canvas bind point <Any-Leave> {.canvas itemconfig current -fill blue}

proc addpt {x y} {
  global point
  .canvas create oval \
          [expr $x-$point(radius)] [expr $y-$point(radius)] \
          [expr $x+$point(radius)] [expr $y+$point(radius)] \
          -width 1 -outline black -fill blue -tags point
}

# add points with button 1
bind .canvas <1> {addpt %x %y}

# allow points to be grabbed and moved with button 2
.canvas bind point <2> {
  global point
  .canvas addtag selected withtag current
  .canvas raise current
  set point(lastX) %x
  set point(lastY) %y
}
.canvas bind point <ButtonRelease-2> {
  .canvas dtag selected
}
bind .canvas <B2-Motion> {
  global point
  set x %x
  set y %y
  .canvas move selected [expr $x-$point(lastX)] [expr $y-$point(lastY)]
  set point(lastX) $x
  set point(lastY) $y
}

# remove points with button 3
.canvas bind point <3> {
  .canvas delete current
}

# Load/save points ############################################################
frame .iof
pack .iof -fill x -expand 1
button .loadpts -text "Load Points" -command {
  set fp [open [.ptfile get]]
  while {[gets $fp line] >= 0} {
    eval addpt $line
  }
  close $fp
}
proc savepts {} {
  global point
  set fp [open [.ptfile get] w]
  foreach pt [.canvas find withtag point] {
    set c [.canvas coords $pt]
    puts $fp "[expr [lindex $c 0]+$point(radius)] \
              [expr [lindex $c 1]+$point(radius)]"
  }
  close $fp
}
button .savepts -text "Save Points" -command savepts
entry .ptfile
.ptfile insert 0 "cluster.pts"
pack .loadpts .savepts .ptfile -side left -in .iof -fill x -expand 1

# Load tree ###################################################################

# compute a bounding box for one point
proc pt_bbox {coords} {
  global point
  return "[expr [lindex $coords 0]-$point(radius)] \
          [expr [lindex $coords 1]-$point(radius)] \
          [expr [lindex $coords 0]+$point(radius)] \
          [expr [lindex $coords 1]+$point(radius)]"
}

# compute a bounding box for the given rectangles
proc merge_bbox {boxes} {
  set first [lindex $boxes 0]
  set min_x [lindex $first 0]
  set min_y [lindex $first 1]
  set max_x [lindex $first 2]
  set max_y [lindex $first 3]
  foreach box $boxes {
    if {[lindex $box 0] < $min_x} {
      set min_x [lindex $box 0]
    }
    if {[lindex $box 1] < $min_y} {
      set min_y [lindex $box 1]
    }
    if {[lindex $box 2] > $max_x} {
      set max_x [lindex $box 2]
    }
    if {[lindex $box 3] > $max_y} {
      set max_y [lindex $box 3]
    }
  }
  return "$min_x $min_y $max_x $max_y"
}

frame .treef
pack .treef -fill x -expand 1
button .loadtree -text "Load Tree" -command {loadtree [.treefile get]}
proc loadtree {treefile} {
  .canvas delete tree
  set pts [.canvas find withtag point]
  set fp [open $treefile]
  set ptstack ""
  while {[gets $fp line] >= 0} {
    # is this a parent node?
    if {[lindex $line 0] == ":"} {
      # merge top items on stack
      set pos [expr [lindex $line 1]-1]
      set bbox [merge_bbox [lrange $ptstack 0 $pos]]
      eval .canvas create rect $bbox -outline green -width 2 -tags tree
      set ptstack [lreplace $ptstack 0 $pos $bbox]
    } else {
      # add to stack
      set pt [.canvas coords [lindex $pts $line]]
      set ptstack [linsert $ptstack 0 $pt]
    }
  }
  close $fp  
}
entry .treefile
.treefile insert 0 "cluster.tree"
pack .loadtree .treefile -side left -in .treef -fill x -expand 1

# Clustering functions ########################################################

proc clus_jp {} {
    set ptfile [.ptfile get]
    set treefile [.treefile get]
    set shared [.shared get]

    set npoints [llength [.canvas find withtag point]]
    puts "$npoints points"
    puts "Computing neighbors"
    #  exec awk -f neighbors.awk < $ptfile > cluster.nn
    exec neighbors $ptfile $npoints 2 $npoints > neighbors.out
    
    puts "Clustering"
    catch {exec nnhierarchy neighbors.out $npoints $shared [expr $shared+1] [expr $npoints-1] | simplify > $treefile} msg
    puts $msg
    
    loadtree $treefile
}

proc clus_nnmatrix {mflag tflag} {
    set ptfile [.ptfile get]
    set treefile [.treefile get]

    set npoints [llength [.canvas find withtag point]]
    puts "$npoints points"
    puts "Computing neighbors"
    exec neighbors $ptfile $npoints 2 $npoints > neighbors.out
    
    puts "Computing matrix"
    exec nnmatrix $mflag neighbors.out > cluster.matrix
    puts "Compressing matrix" 
    catch {exec matrix2tree $tflag cluster.matrix > $treefile} msg
    puts $msg

    loadtree $treefile
}

proc clus_nncluster {mflag tflag} {
    set ptfile [.ptfile get]
    set treefile [.treefile get]

    set npoints [llength [.canvas find withtag point]]
    puts "$npoints points"
    puts "Computing neighbors"
    exec neighbors $ptfile $npoints 2 $npoints > neighbors.out

    puts "Computing and compressing matrix"
    catch {exec nncluster $tflag $mflag < neighbors.out > $treefile} msg
    puts $msg

    loadtree $treefile
}

proc clus_slink {} {
    set ptfile [.ptfile get]
    set treefile [.treefile get]

    puts "Computing neighbors and clustering"
    catch {exec matrix_slink < $ptfile > $treefile} msg
    puts $msg
    
    loadtree $treefile
}

proc clus_distr {} {
    set ptfile [.ptfile get]
    set treefile [.treefile get]
    set temp [expr [.temp get].0/100]

    catch {exec neighbor_distr < $ptfile > neighbors.out} msg
    puts $msg

    set treefile neighbors.out
    #catch {exec dist_clus 3 $temp < neighbors.out > $treefile} msg
    #puts $msg

    # show the first cluster via circles
    .canvas delete tree
    set pts [.canvas find withtag point]
    set fp [open $treefile]
    gets $fp line
    close $fp
    puts $line
    for {set i 0} {$i < [llength $line]} {incr i} {
	set d [expr sqrt(-2000*log([lindex $line $i]))]
	set pt [.canvas coords [lindex $pts $i]]
	set cx [expr ([lindex $pt 0] + [lindex $pt 2])/2]
	set cy [expr ([lindex $pt 1] + [lindex $pt 3])/2]
	puts "$d $pt"
	.canvas create oval [expr $cx-$d] [expr $cy-$d] \
		            [expr $cx+$d] [expr $cy+$d] \
			    -width 1 -outline green -tags tree
    }
}

frame .actions
pack .actions -fill x -expand 1
set clus_types {{max s} {max c} {sum s} {sum c}}
foreach y $clus_types {
    button .actions.clus_$y -text $y -command "savepts; clus_nnmatrix [lindex $y 0] [lindex $y 1]"
    pack .actions.clus_$y -fill x -expand 1 -side left
}

frame .actions2
pack .actions2 -fill x -expand 1
set clus_types {{max min} {max max} {max avg} {sum min} {sum max} {sum avg}}
foreach y $clus_types {
    button .actions2.clus_$y -text $y -command "savepts; clus_nncluster [lindex $y 0] [lindex $y 1]"
    pack .actions2.clus_$y -fill x -expand 1 -side left
}

frame .jp
pack .jp -fill x -expand 1
button .clus_jp -text jp -command "savepts; clus_jp"
pack .clus_jp -side left -in .jp
scale .shared -from 0 -to 20 -orient horiz
pack .shared -fill x -expand 1 -side left -in .jp

button .clus_slink -text SLINK -command "savepts; clus_slink"
pack .clus_slink -expand 1 -fill x

frame .distr
pack .distr -fill x -expand 1
button .clus_distr -text distr -command "savepts; clus_distr"
pack .clus_distr -side left -in .distr
scale .temp -from 1 -to 100 -orient horiz
pack .temp -fill x -expand 1 -side left -in .distr

# Other operations ############################################################
frame .controlf
pack .controlf -fill x -expand 1
button .clear -text "Clear All" -command {.canvas delete all}
button .quit -text "Quit" -command exit
pack .clear .quit -side left -in .controlf -fill x -expand 1


