/* Distributional clustering 
 * Pereira, Tishby, and Lee
 * http://www.das.harvard.edu/users/students/Lillian_Lee/poster.html
 */
#include <tpm/tpm.h>
#include <tpm/matrix.h>

#define CHI 1
#define VERBOSE 0

static Matrix InitialCenters(int ncenters, Matrix sample);
double Clus_Distribution(Matrix sample, Matrix center, Matrix joint,
			 double temperature, 
			 char *center_file, char *joint_file);

void main(int argc, char *argv[])
{
  int nclusters;
  Matrix sample; /* sample distributions */
  Matrix center; /* centers */
  Matrix joint;  /* prob of center given sample */
  double temperature;
  FILE *fp;

  if(argc < 4) {
    fprintf(stderr, "Usage:\n%s k center_output joint_output < sample_matrix\n", argv[0]);
    fprintf(stderr, "To generate k clusters of the sample distributions.\n");
    exit(1);
  }
  nclusters = atoi(argv[1]);

  sample = MatrixScan(stdin);
  if(!sample) {
    fprintf(stderr, "Error reading sample distribution matrix\n");
    exit(1);
  }

  center = InitialCenters(nclusters, sample);
  joint = MatrixCreate(sample->rows, center->rows);
  for(temperature = 1; temperature >= 1e-2; temperature = temperature / 10) {
    fprintf(stderr, "temperature = %g\n", temperature);
    Clus_Distribution(sample, center, joint, temperature, argv[2], argv[3]);
  }

  fp = fopen(argv[2], "w");
  MatrixPrint(fp, center);
  fclose(fp);
  fp = fopen(argv[3], "w");
  MatrixPrint(fp, joint);
  fclose(fp);
  
  MatrixFree(joint);
  MatrixFree(center);
  MatrixFree(sample);
  MEM_BLOCKS();
  MEM_STATUS();
  exit(0);
}

/* Initialize centers at perturbed mean of samples. */
static Matrix InitialCenters(int ncenters, Matrix sample)
{
  int i,j;
  Matrix center;
  Vector mean = MatrixColumnSum(sample);
  VectorScale(mean, 1.0/sample->rows);
  center = MatrixCreate(ncenters, sample->cols);
  for(i=0;i<center->rows;i++) {
    for(j=0;j<center->cols;j++) {
      center->d2[i][j] = mean->d[j] + Rand() / 100;
    }
  }
  VectorFree(mean);
  return center;
}

double RelativeEntropy(double *p, double *q, int n)
{
  double d = 0;
  int i;
  for(i=0;i<n;i++) {
    if(*p > 0) {
      /* doesn't seem to matter whether we include p's entropy */
#if CHI
      /* Chi-squared approximation */
      double f = *p - *q;
      d += f * f / *q;
#else
#if 1
      d += *p * log(*p / *q);
#else
      d -= *p * log(*q);
#endif
#endif
    }
    p++; q++;
  }
#if CHI
  d *= 0.5;
#endif
  return d;
}

/* Modifies the centers to better approximate the samples.
 * I.e. reduce the average KL distortion between them.
 * Returns the previous distortion.
 */
double Clus_DistributionOneStep(Matrix sample, Matrix center, Matrix joint,
				double temperature)
{
  int s, c, i;
  double distortion = 0;

  /* compute distortion */
  for(s=0; s < sample->rows; s++) {
    double center_sum = 0, distortion_subtotal = 0;
#if 0
    if(s % 10 == 0) fprintf(stderr, "computing distortion for %d\n", s);
#endif
    for(c=0; c < center->rows; c++) {
      double d = RelativeEntropy(sample->d2[s], center->d2[c], sample->cols);
      if(d+1 == d) { /* check for infinity */
	joint->d2[s][c] = 0;
      }
      else {
	center_sum += joint->d2[s][c] = exp(-d/temperature);
	distortion_subtotal += joint->d2[s][c] * d;
      }
    }
    if(center_sum > 0.0) {
      distortion += distortion_subtotal / center_sum;
      for(c=0; c < center->rows; c++) {
	joint->d2[s][c] /= center_sum;
      }
    }
  }
#if 0
  fprintf(stderr, "distortion: %g\n", distortion);
#endif

  /* loop clusters and compute new mean */
  VectorSet((Vector)center, 0);
  for(c=0; c < center->rows; c++) {
    double sample_sum = 0;
#if 0
    fprintf(stderr, "updating center %d\n", c);
#endif
    for(s=0; s < sample->rows; s++) {
      sample_sum += joint->d2[s][c];
      for(i=0; i < center->cols; i++) {
	center->d2[c][i] += joint->d2[s][c] * sample->d2[s][i];
      }
    }
    if(sample_sum > 0.0) {
      for(i=0; i < center->cols; i++) {
	center->d2[c][i] /= sample_sum;
      }
    }
  }

  return distortion;
}

/* Modifies: center, joint */
double Clus_Distribution(Matrix sample, Matrix center, Matrix joint, 
			 double temperature, 
			 char *center_file, char *joint_file)
{
  FILE *fp;
  double distortion = 1e12, old_distortion;
  do {
    old_distortion = distortion;
#if VERBOSE
    fprintf(stderr, "centers:\n");
    MatrixPrint(stderr, center);
#endif
    distortion = Clus_DistributionOneStep(sample, center, joint, temperature);
    fp = fopen(center_file, "w");
    MatrixPrint(fp, center);
    fclose(fp);
    fp = fopen(joint_file, "w");
    MatrixPrint(fp, joint);
    fclose(fp);
    /* have to use fabs since distortion may momentarily increase */
  } while(fabs(old_distortion - distortion) > 1e-1*temperature);
  return distortion;
}
