/* Performs Jarvis+Patrick nearest-neighbor clustering.
 * This is a single-link method: the clusters of two points are merged if
 * those points are k-nearest neighbors of each other and have kt of
 * their k-nearest neighbors in common. (k is the distance threshold;
 * kt is the shared sample threshold.)
 */

#include <tpm/stream.h>
#include <tpm/partition.h>

Partition NNPartition(int **samples, int num_samples, 
		      int dist_thres, int shared_thres);
int CmpNN(int *a, int *b, int dist_thres, int shared_thres);

int main(int argc, char **argv)
{
  char *nn_file;
  FileHandle fp;
  int **samples, **s1;
  int num_samples;
  int dist_thres, shared_thres;
  int i,j;
  Partition p;

  if(argc < 5) {
    fprintf(stderr, 
	    "Usage:\n%s <nn_file> <num_samples> <dist_thres> <shared_thres>\n",
	    argv[0]);
    fprintf(stderr, 
	    "To cluster <num_samples> lines of neighbor file <nn_file>.\n");
    fprintf(stderr, 
	    "dist_thres = distance threshold (k-nearest)\n");
    fprintf(stderr, 
	    "shared_thres = shared sample threshold (kt-nearest in common)\n");
    fprintf(stderr, "shared_thres < dist_thres <= neighbors in nn_file\n");
    fprintf(stderr, "Partition goes to stdout.\n");
    exit(1);
  }

  nn_file = argv[1];
  num_samples = atoi(argv[2]);
  dist_thres = atoi(argv[3]);
  shared_thres = atoi(argv[4])-1;
  if(shared_thres > dist_thres) {
    fprintf(stderr, "Error: shared_thres (%d) > dist_thres (%d)\n",
	    shared_thres, dist_thres);
    exit(1);
  }

  /* Read in neighbor file */
  fp = FileOpen(nn_file, "r");
  samples = Allocate(num_samples, int*);
  for(s1=samples,i=0;i<num_samples;s1++,i++) {
#if 0
    if(i%1000 == 0) fprintf(stderr, "%d\n", i);
#endif
    *s1 = Allocate(dist_thres+1, int);
    for(j=0;j<=dist_thres;j++) {
      fscanf(fp, "%d", &(*s1)[j]);
    }
    while(fgetc(fp) != '\n') if(feof(fp)) break;
    /* sort for better comparision performance */
    qsort(&(*s1)[1], dist_thres, sizeof(int), (int(*)())IntCmp);
  }
  FileClose(fp);

  /* Form clusters */
  p = NNPartition(samples, num_samples, dist_thres, shared_thres);
  WritePartition(stdout, p);
  return 0;
}

Partition NNPartition(int **samples, int num_samples, 
		      int dist_thres, int shared_thres)
{
  Partition p = CreatePartition(num_samples);
  int i,j,k;
  
  for(i=0;i<num_samples;i++) {
#if 0
    if(i%1000 == 0) fprintf(stderr, "%d\n", i);
#endif
    /* loop the neighbors of samples[i] */
    for(j=1;j<=dist_thres;j++) {
      k = samples[i][j];
      if((k < i) || (k >= num_samples)) continue;
      if(!CmpNN(samples[i], samples[k], 
		dist_thres, shared_thres)) {
	MergeElements(p, i, k);
      }
    }
  }
  NormalizePartition(p);
  return p;
}

/* Given: *b is in a 
 */
int CmpNN(int *a, int *b, int dist_thres, int shared_thres)
{
  int i,j, count;
  int found_a, found_b;

  /* Check if *a is in b */
  for(i=1;i<=dist_thres;i++) 
    if(b[i] == *a) goto found_a;
    else if(b[i] > *a) break;
  return 1;
found_a:

  /* Check that a and b have shared_thres elements in common */
  count = 0;
  for(i=1,j=1;(i<=dist_thres)&&(j<=dist_thres);) {
    if(a[i] == b[j]) {
      count++;
      i++; j++;
    }
    else if(a[i] > b[j]) j++;
    else if(a[i] < b[j]) i++;
    if(count > shared_thres) break;
  }
  return !(count > shared_thres);
}
