/* Performs hierarchical clustering on a dissimilarity matrix.
 * Algorithms taken from Jain and Dubes: single/complete link, Ward's.
 */
#include <tpm/gtree.h>
#include <tpm/matrixI.h>

#define VERBOSE 1

typedef void HierFunc(MatrixI mat, int *sizes, int r, int s, int k);
HierFunc SingleLink, CompleteLink, Wards;

void FindMin(MatrixI mat, int *r, int *s);

int main(int argc, char *argv[])
{
  FILE *fp;
  char *fname;
  MatrixI mat;
  int i, r, s, last_r;
  int *sizes;
  Tree *top, newtree;
  HierFunc *hierFunc;
  double last_val;

  if(argc < 3) {
    fprintf(stderr, "Usage: %s [c|s|w] matrix > output\n", argv[0]);
    fprintf(stderr, "To form a hierarchical clustering from a dissimilarity matrix,\n");
    fprintf(stderr, "e.g. the output of nnmatrix.\n");
    fprintf(stderr, "c = complete link\n");
    fprintf(stderr, "s = single link\n");
    fprintf(stderr, "w = Ward's method\n");
    exit(1);
  }

  switch(argv[1][0]) {
  case 'c':
    hierFunc = CompleteLink;
    break;
  case 's':
    hierFunc = SingleLink;
    break;
  case 'w':
    hierFunc = Wards;
    break;
  default:
    fprintf(stderr, "bad algorithm type; must be c, s, or w.\n");
    exit(1);
  }

  /* read matrix file */
  fname = argv[2];
  fp = fopen(fname, "r");
  if(!fp) {
    fprintf(stderr, "Cannot open `%s'\n", fname);
    exit(1);
  }
  mat = MatrixIScanBinary(fp);
  fclose(fp);

  /* initialize */
  top = Allocate(mat->rows, Tree);
  sizes = Allocate(mat->rows, int);
  for(i=0;i<mat->rows;i++) {
    top[i] = TreeCreate(RealCreate(i));
    sizes[i] = 1;
  }

  /* main loop: compress the matrix */
  last_val = 0;
  last_r = -1;
  while(mat->rows > 1) {
#if 0
    if(mat->rows % 1 == 0) fprintf(stderr, "%d\n", mat->rows);
#endif
    FindMin(mat, &r, &s);

    /* make these rows siblings */
#if VERBOSE
    fprintf(stderr, "(%d,%d) %d\n", r, s, mat->d2[r][s]);
#endif
    if((last_val == mat->d2[r][s]) &&
       ((last_r == r) || (last_r == s))) {
      if(last_r == r) {
	TreeAddChild(top[r], top[s]);
      }
      else {
	TreeAddChild(top[s], top[r]);
	top[r] = top[s];
      }
    }
    else {
      newtree = TreeCreate(RealCreate(mat->d2[r][s]));
      TreeAddChild(newtree, top[r]);
      TreeAddChild(newtree, top[s]);
      top[r] = newtree;
    }
    last_val = mat->d2[r][s];
    last_r = r;
    
    /* change the matrix entries */
    for(i=0;i<mat->rows;i++) {
      hierFunc(mat, sizes, r, s, i);
    }
    sizes[r] += sizes[s];

    /* remove the s row */
    mat->rows--;
    for(i=s;i<mat->rows;i++) {
      mat->d2[i] = mat->d2[i+1];
      sizes[i] = sizes[i+1];
      top[i] = top[i+1];
    }

    /* remove the s column */
    mat->cols--;
    for(r=0;r<mat->rows;r++) {
      for(i=s;i<mat->cols;i++)
	mat->d2[r][i] = mat->d2[r][i+1];
    }
  }

  /* unify the trees */
  if(mat->rows > 1) {
    newtree = TreeCreate(RealCreate(0));
    for(i=0;i<mat->rows;i++)
      TreeAddChild(newtree, top[i]);
    top[0] = newtree;
  }
  TreeWrite(stdout, top[0], RealWrite);
  exit(0);
}

void SingleLink(MatrixI mat, int *sizes, int r, int s, int k)
{
  mat->d2[r][k] = min(mat->d2[r][k], mat->d2[s][k]);
  mat->d2[k][r] = min(mat->d2[k][r], mat->d2[k][s]);
}

void CompleteLink(MatrixI mat, int *sizes, int r, int s, int k)
{
  mat->d2[r][k] = max(mat->d2[r][k], mat->d2[s][k]);
  mat->d2[k][r] = max(mat->d2[k][r], mat->d2[k][s]);
}

/* this code is currently broken */
void Wards(MatrixI mat, int *sizes, int r, int s, int k)
{
  double a1,a2,a3, d1,d2,d3;

  a3 = sizes[r]+sizes[s]+sizes[k];
  a1 = (sizes[r]+sizes[k])/a3;
  a2 = (sizes[s]+sizes[k])/a3;
  a3 = -sizes[k]/a3;

  d1 = mat->d2[r][k];
  d2 = mat->d2[s][k];
  d3 = mat->d2[r][s];
  mat->d2[r][k] = a1*d2*d3 + a2*d1*d3 + a3*d1*d2;
  if(mat->d2[r][k]) mat->d2[r][k] = d1*d2*d3 / mat->d2[r][k];

  d1 = mat->d2[k][r];
  d2 = mat->d2[k][s];
  d3 = mat->d2[s][r];
  mat->d2[k][r] = a1*d2*d3 + a2*d1*d3 + a3*d1*d2;
  if(mat->d2[k][r]) mat->d2[k][r] = d1*d2*d3 / mat->d2[k][r];
}

void FindMin(MatrixI mat, int *r, int *s)
{
  int i,j;
  int v, best;

  best = mat->d2[*r=0][*s=1];
  for(i=0;i<mat->rows;i++) {
    for(j=0;j<mat->cols;j++) {
      if(i == j) continue;
      v = mat->d2[i][j];
      if(v < best) {
	*r = i; *s = j; 
	best = v;
      }
    }
  }
}
