/* Generates a nearest-neighbor file, given a list of feature vectors.
 * Uses Euclidean distance.
 */
/*
cc -O -I/mas/vision/local/include -o neighbors neighbors.c
*/
#include <tpm/tpm.h>

/* if NN is 0, will write out distances instead. */
#define NN 1

typedef double Real;
typedef struct SampleStruct {
  int index;
  Real distance;
  Real *vector;
} *Sample;

Real EuclidSqrDist(Real *a, Real *b, int len);
int SampleCompare(Sample a, Sample b);

void main(int argc, char *argv[])
{
  char *feature_file;
  int nsamples, nfeatures, dist_thres;
  Sample features;
  int i,j,k;
  FILE *fp;
  float *dist;

  if(argc < 5) {
    fprintf(stderr, 
	    "Usage: %s features nsamples nfeatures threshold > output\n", 
	    argv[0]);
    fprintf(stderr, "To generate a nearest-neighbor file, given feature vectors.\n");
    fprintf(stderr, "Threshold is the number of neighbors to output per sample.\n");
    exit(1);
  }
  feature_file = argv[1];
  nsamples = atoi(argv[2]);
  nfeatures = atoi(argv[3]);
  dist_thres = atoi(argv[4]);
  
  fp = fopen(feature_file, "r");
  if(!fp) {
    fprintf(stderr, "Cannot open `%s'\n", feature_file);
    exit(1);
  }
  
  /* Read the feature file */
  features = Allocate(nsamples, struct SampleStruct);
  for(i=0;i<nsamples;i++) {
    features[i].index = i;
    features[i].vector = Allocate(nfeatures, Real);
    for(j=0;j<nfeatures;j++) {
      fscanf(fp, "%lg", &features[i].vector[j]);
      if(feof(fp)) { fprintf(stderr, "Error: premature EOF\n"); exit(1); }
    }
    /* skip the rest of the line */
    while((j = fgetc(fp)) != EOF) if(j == '\n') break;
  }
  fclose(fp);
  
  if(nfeatures == 1) {
    /* on a line, nearest-neighbor is easy */
    for(i=0;i<nsamples;i++) features[i].distance = features[i].vector[0];
    qsort((void*)features, nsamples, sizeof(struct SampleStruct),
	  (int(*)())SampleCompare);
    for(i=0;i<nsamples;i++) {
      int a,b;
#if 0
      if(i % 1000 == 0) printf("%d\n", i);
#endif
      /* Find the sample with index i */
      for(k=0;k<nsamples;k++) {   
	if(features[k].index == i) break;
      }
      printf("%d ", i);
      a = k-1;
      b = k+1;
      for(j=1;j<dist_thres;j++) {
	if((a > 0) && ((b >= nsamples) ||
	   (features[k].distance - features[a].distance <
	    features[b].distance - features[a].distance))) {
	  printf("%d ", features[a].index);
	  a--;
	}
	else {
	  printf("%d ", features[b].index);
	  b++;
	}
      }
      printf("\n");
    }
    exit(0);
  }


#if !NN
  dist = Allocate(nsamples, float);
#endif

  /* Loop all samples */
  for(i=0;i<nsamples;i++) {
#if 0
    if(i % 100 == 0) fprintf(stderr, "%d\n", i);
#endif

    /* Find the sample with index i */
    for(k=0;k<nsamples;k++) {   
      if(features[k].index == i) break;
    }
    
    /* Compute distances to sample k */
    for(j=0;j<nsamples;j++) {
      features[j].distance = 
	EuclidSqrDist(features[j].vector, 
		      features[k].vector, 
		      nfeatures);
    }
    
    /* Sort the samples by distance */
    qsort((void*)features, nsamples, sizeof(struct SampleStruct),
	  (int(*)())SampleCompare);

#if NN
    /* Write out the nearest */
    /* Make sure that i is the first number on the line */
    printf("%d ", i);
    for(j=0;j<dist_thres;j++) {
      if(j == nsamples) break; 
      if(features[j].index == i) continue;
      printf("%d ", features[j].index);
    }
    printf("\n");
#else
    /* Write out distances */
    for(j=0;j<nsamples;j++) {
      dist[features[j].index] = features[j].distance;
    }
    for(j=0;j<nsamples;j++) {
      printf("%g ", dist[j]);
    }
    printf("\n");
#endif
  }
  exit(0);
}

int SampleCompare(Sample a, Sample b)
{
  if(a->distance > b->distance) return 1;
  else return -1;
}

Real EuclidSqrDist(Real *a, Real *b, int len)
{
  Real d, sum = 0.0;
  while(len--) {
    d = *a++ - *b++;
    sum += d*d;
  }
  return sum; /* no sqrt */
}
