/* Generates a hierarchical clustering from nearest-neighbor
 * data.  Space overhead is 18 bytes per entry in the NN matrix.
 * The "sum" combination rule only makes sense with a square NN matrix.
 */
/* 
cc -O -I/mas/vision/local/include nncluster.c -o nncluster -L/mas/vision/local/lib/$MACHTYPE -larray -ltpm

neighbors data 369 30 368 | nncluster avg > tree
*/
#include <tpm/tpm.h>
#include <tpm/matrixI.h>
#include <tpm/matrixSparseP.h>
#include <tpm/gtree.h>
#include <tpm/heap.h>

#define VERBOSE 0
#define WEIGHT 10000

typedef struct Similarity {
  ushort l, h; /* l < h */
  ushort d;    /* dissimilarity */
  int heap_index;
} *Similarity;

Similarity SimilarityCreate(ushort i, ushort j, ushort d)
{
  Similarity s = Allocate(1, struct Similarity);
  if(i < j) { s->l = i; s->h = j; }
  else { s->l = j; s->h = i; }
  s->d = d;
  return s;
}

static int SimilarityCompare(Similarity a, Similarity b)
{
  if(a->d < b->d) return -1;
  else if(a->d > b->d) return 1;
  else return 0;
}

static void SimilarityIndex(Similarity s, int index)
{
  s->heap_index = index;
}

typedef enum { COMBINE_MAX, COMBINE_SUM } SimilarityCombine;
typedef enum { CLUSTER_MIN, CLUSTER_MAX, CLUSTER_AVG, CLUSTER_WARD, CLUSTER_WAVG } ClusterMethod;

void SimilarityMatrix(MatrixI nbrs, SimilarityCombine comb,
		      MatrixSparseP *matrix_return,
		      Similarity **similarity_return, int *count_return);
Tree MatrixHierarchy(MatrixSparseP matrix, Similarity *similarity, int count,
		     ClusterMethod method);

int main(int argc, char **argv)
{
  MatrixI nbrs;
  SimilarityCombine comb;
  ClusterMethod method;
  MatrixSparseP matrix;
  Similarity *similarity;
  Tree tree;
  int count;

  if(argc < 2) {
    fprintf(stderr, "Usage:\n%s [min|max|avg|ward] [max|sum] < neighbor_matrix\n",
	   argv[0]);
    fprintf(stderr, "To generate a hierarchical clustering from nearest-neighbor data,\n");
    fprintf(stderr, "such as the output of `neighbors'.\n");
    fprintf(stderr, "min|max|avg is the rule for combining clusters.\n");
    fprintf(stderr, "max|sum is the rule for combining similarities (default max).\n");
    exit(1);
  }

  if(!strcmp(argv[1], "min")) method = CLUSTER_MIN;
  else if(!strcmp(argv[1], "max")) method = CLUSTER_MAX;
  else if(!strcmp(argv[1], "avg")) method = CLUSTER_AVG;
  else if(!strcmp(argv[1], "ward")) method = CLUSTER_WARD;
  else if(!strcmp(argv[1], "wavg")) method = CLUSTER_WAVG;
  else {
    fprintf(stderr, "Unknown clustering method `%s'\n", argv[1]);
    exit(1);
  }
  if(argc > 2) {
    if(!strcmp(argv[2], "max")) comb = COMBINE_MAX;
    else if(!strcmp(argv[2], "sum")) comb = COMBINE_SUM;
    else {
      fprintf(stderr, "Unknown combination rule `%s'\n", argv[2]);
      exit(1);
    }
  }
  else comb = COMBINE_MAX;

  fprintf(stderr, "Reading NN matrix...");
  nbrs = MatrixIScan(stdin);
  fprintf(stderr, "Done\n");
  if(!nbrs) {
    fprintf(stderr, "Error reading neighbor matrix\n");
    exit(1);
  }

  SimilarityMatrix(nbrs, comb, &matrix, &similarity, &count);
  tree = MatrixHierarchy(matrix, similarity, count, method);
  TreeWrite(stdout, tree, IntWrite);
  TreeFree(tree, GenericFree);

  MEM_BLOCKS();
  MEM_STATUS();
  exit(0);
}

void SimilarityPrint(FILE *fp, Similarity s)
{
  fprintf(fp, "(%2d, %2d, %2d) ", s->l, s->h, s->d);
}

void MatrixSparsePPrint(FILE *fp, MatrixSparseP m)
{
  int i, j;
  fprintf(fp, "matrix:\n");
  for(i=0;i<m->rows;i++) {
    for(j=0;j<m->cols;j++) {
      Similarity s = (Similarity)MatrixSparsePAt(m, i, j);
      if(!s) fprintf(fp, "             ");
      else SimilarityPrint(fp, s);
    }
    fprintf(fp, "\n");
  }
  fprintf(fp, "\n");
}

/* Computes a sparse similarity matrix from nearest-neighbor data.
 * Since only half of the entries are unique, the matrix just contains
 * pointers to Similarity structures, stored in an array.
 * Frees nbrs. 
 */
void SimilarityMatrix(MatrixI nbrs, SimilarityCombine comb,
		      MatrixSparseP *matrix_return,
		      Similarity **similarity_return, int *count_return)
{
  MatrixSparseP matrix;
  Similarity *similarity, s;
  int i, j, ctr;

  /* Create a sparse similarity matrix with padding for future insertions */
  matrix = MatrixSparsePCreate(nbrs->rows, 2*(nbrs->cols-1));
  matrix->cols = matrix->rows; /* for printing purposes */

  /* Create an array of Similarities. 
   * Will probably have less than this number of entries, 
   * but we don't know for sure.
   */
  similarity = Allocate(nbrs->rows*(nbrs->cols-1)*2, Similarity);

  /* Compute Similarities */
  ctr = 0;
  for(i=0;i<nbrs->rows;i++) {
    if(VERBOSE || (i % 1000 == 0)) fprintf(stderr, "%d\n", i);
    /* loop neighborhood of i */
    for(j=1;j<nbrs->cols;j++) {
      int n = nbrs->d2[i][j];
      int p;
      /* allocate an entry in the matrix */
      /* does this Similarity already exist? */
      if((i < n) || ((p = MatrixSparsePIndex(matrix, n, i)) == matrix->len)) {
	/* no; make a new one */
	s = SimilarityCreate(i, n, j);
	similarity[ctr++] = s;
	MatrixSparsePInsert(matrix, i, n, s);
	/* make the matrix symmetric */
	MatrixSparsePInsert(matrix, n, i, s);
      }
      else {
	Similarity s = (Similarity)matrix->d[p];
	/* yes; share it */
	if(comb == COMBINE_MAX) {
	  /* Compute max of the nbr position and the entry in the matrix */
	  if(s->d < j) s->d = j;
	}
	else if(comb == COMBINE_SUM) {
	  /* Compute sum of nbr position and entry in matrix */
	  s->d += j;
	}
      }
    }
  }
  MatrixIFree(nbrs);

  *count_return = ctr;
  *matrix_return = matrix;
  *similarity_return = similarity;

#if VERBOSE
  MatrixSparsePPrint(stderr, matrix);
#endif
}

void HeapPrint(FILE *fp, Heap heap)
{
  int i;
  fprintf(fp, "heap:\n");
  for(i=0;i<heap->length;i++) SimilarityPrint(fp, (Similarity)heap->data[i]);
  fprintf(fp, "\n\n");
}

/* Computes a tree from a similarity matrix.
 * Frees matrix and similarity. 
 */
Tree MatrixHierarchy(MatrixSparseP matrix, Similarity *similarity, int count,
		     ClusterMethod method)
{
  Tree tree;
  Heap heap;
  Tree *top; /* [row] trees */
  ushort *d; /* [row] d number of each tree */
  int *size; /* [row] size of each tree */
  int i, l, h, pl, ph;
  Similarity s;

  top = Allocate(matrix->rows, Tree);
  for(i=0;i<matrix->rows;i++) top[i] = TreeCreate(IntCreate(i));
  d = AllocateC(matrix->rows, ushort);
  size = Allocate(matrix->rows, int);
  for(i=0;i<matrix->rows;i++) size[i] = 1;

  /* Make a heap out of the similarities.  This allows us to rapidly extract
   * the pair of clusters with the smallest distance.
   */
  heap = HeapFromArray((Any*)similarity, count, (CmpFunc*)SimilarityCompare, 
		       GenericFree, (HeapIndexFunc*)SimilarityIndex);
#if VERBOSE
  HeapPrint(stderr, heap);
#endif
  i = 0;
  while(s = HeapExtractMin(heap)) {
    if(i++ % 100 == 0) 
      fprintf(stderr, "dist: %d  pos: (%d, %d)  size: (%d, %d)\n", 
	      s->d, s->l, s->h, size[s->l], size[s->h]);
    l = s->l; h = s->h;
    assert(l != h);
    /* unify rows l and h */
    if(d[l] == s->d) {
      if(d[h] == s->d) {
	/* steal top[h]'s children */
	TreeAddChild(top[l], top[h]->child);
	top[h]->child = NULL;
	TreeFree(top[h], GenericFree);
      }
      else {
	TreeAddChild(top[l], top[h]);
      }
    }
    else if(d[h] == s->d) {
      TreeAddChild(top[h], top[l]);
      top[l] = top[h];
    }
    else {
      tree = TreeCreate(IntCreate(s->d));
      TreeAddChild(tree, top[l]);
      TreeAddChild(tree, top[h]);
      top[l] = tree;
    }
    top[h] = NULL;
    d[l] = s->d;

    /* change the matrix entries */
    free(s);
    MatrixSparsePDelete(matrix, l, h);
    MatrixSparsePDelete(matrix, h, l);
    /* loop l's row */
    for(pl=matrix->start[l]; pl < matrix->end[l]; pl++) {
      /* loop h's row, looking for a match */
      for(ph=matrix->start[h]; ph < matrix->end[h]; ph++) {
	if(matrix->col[ph] == matrix->col[pl]) {
	  Similarity sl = (Similarity)matrix->d[pl];
	  Similarity sh = (Similarity)matrix->d[ph];
	  int t, k = (sh->h == h) ? sh->l : sh->h;
	  /* this is only part of the code that depends on the 
             clustering method. */
	  if(method == CLUSTER_AVG) {
	    sl->d = (size[l] * sl->d + size[h] * sh->d) / (size[l] + size[h]);
	    HeapChangedKey(heap, sl->heap_index);
	  }
	  else if(method == CLUSTER_WARD) {
	    /* cast to float first to avoid truncation errors */
	    sl->d = (ushort)((float)((size[l] + size[k]) * sl->d +
				     (size[h] + size[k]) * sh->d -
				     size[k] * d[l]) / 
			     (size[l] + size[h] + size[k]) + 0.5);
	    HeapChangedKey(heap, sl->heap_index);
	  }
	  else if(method == CLUSTER_WAVG) {
#if 0
	    sl->d = (sl->d + sh->d) / 2;
#else
	    sl->d = (ushort)((float)(size[l] * sl->d + size[h] * sh->d + 
			     WEIGHT * (sl->d < sh->d ? sl->d : sh->d)) / 
			     (size[l] + size[h] + WEIGHT) + 0.5);
#endif
	    HeapChangedKey(heap, sl->heap_index);
	  }
	  else if(((method == CLUSTER_MIN) && (sh->d < sl->d)) ||
		  ((method == CLUSTER_MAX) && (sh->d > sl->d))) {
	    sl->d = sh->d;
	    HeapChangedKey(heap, sl->heap_index);
	  }
	  /* clobber sh */
	  HeapDelete(heap, sh->heap_index);
	  t = matrix->start[l];
	  MatrixSparsePDelete(matrix, h, k);
	  MatrixSparsePDelete(matrix, k, h);
	  if(t != matrix->start[l]) {
	    pl = matrix->start[l] - 1; /* must restart if changed */
	  }
	  break;
	}
      }
    }
    /* loop h's row; anything that's left must be moved into l's row */
    while((ph=matrix->start[h]) < matrix->end[h]) {
      int c = matrix->col[ph];
      s = (Similarity)matrix->d[ph];
      if(s->h == h) if(s->l < l) s->h = l; else { s->h = s->l; s->l = l; }
      else if(s->h > l) s->l = l; else { s->l = s->h; s->h = l; }
      MatrixSparsePDelete(matrix, h, c);
      MatrixSparsePInsert(matrix, l, c, s);
      c = MatrixSparsePIndex(matrix, c, h);
      matrix->col[c] = l;
    }
    size[l] += size[h];
#if VERBOSE
    MatrixSparsePPrint(stderr, matrix);
    HeapPrint(stderr, heap);
#endif
  }

  
  /* unify the remaining trees */
  tree = top[0];
  for(i=1;i<matrix->rows;i++) {
    if(top[i]) {
      tree = TreeCreate(RealCreate(0));
      TreeAddChild(tree, top[0]);
      TreeAddChild(tree, top[i]);
      for(i++;i<matrix->rows;i++) {
	if(top[i]) TreeAddChild(tree, top[i]);
      }
      break;
    }
  }
  free(top);

  HeapFree(heap);
  MatrixSparsePFree(matrix);
  free(d);
  free(size);
  return tree;
}
