/* Generates a NN or MNV dissimilarity matrix from a neighbor file,
 * for use by matrix2tree. Higher numbers are less similar.
 */
#include <tpm/tpm.h>
#include <tpm/matrixI.h>

int main(int argc, char **argv)
{
  char *file;
  FILE *fp;
  int i,j;
  MatrixI matrix, nbrs;
  enum { MAX, SUM } method;
  int share = 0;

  if(argc < 3) {
    fprintf(stderr, "Usage:\n%s [max|sum] neighbor_matrix\n",
	   argv[0]);
    fprintf(stderr, "To generate a binary NN dissimilarity matrix on stdout\n");
    fprintf(stderr, "from an ascii nearest-neighbor file.\n");
    exit(1);
  }

  method = !strcmp(argv[1], "sum");

  file = argv[2];
  fp = fopen(file, "r");
  if(!fp) {
    fprintf(stderr, "Cannot open `%s'\n", file);
    exit(1);
  }
  nbrs = MatrixIScan(fp);
  fclose(fp);
  if(!nbrs) {
    fprintf(stderr, "Error reading neighbor file\n");
    exit(1);
  }

  /* Create full N by N similarity matrix */
  matrix = MatrixICreate(nbrs->rows, nbrs->rows);
  VectorISet((VectorI)matrix, 0);

  if(share) {
    /* add in the sqrt of absolute distances between neighbor lists */
    /* (looks a lot like Euclidean dist) */
    for(i=0;i<nbrs->rows;i++) {
      /* loop neighborhood of i */
      for(j=1;j<nbrs->cols;j++) {
	int n = nbrs->d2[i][j];
	int k, l, dist = 0;
	for(k=0;k<nbrs->cols;k++) {
	  for(l=0;l<nbrs->cols;l++) {
	    if(nbrs->d2[i][k] == nbrs->d2[n][l]) {
	      if(k > l) dist += k-l;
	      else dist += l-k;
	      break;
	    }
	  }
	}
	matrix->d2[i][n] += isqrt(dist);
      }
    }
  }

  for(i=0;i<nbrs->rows;i++) {
    /* loop neighborhood of i */
    for(j=1;j<nbrs->cols;j++) {
      int n = nbrs->d2[i][j];
      if(method == MAX) {
	/* Compute max of the nbr position and the entry in the matrix */
	if(matrix->d2[i][n] < j) matrix->d2[i][n] = j;
	if(matrix->d2[n][i] < j) matrix->d2[n][i] = j;
      }
      else if(method == SUM) {
	/* Compute sum of nbr position and entry in matrix */
	matrix->d2[i][n] += j;
	matrix->d2[n][i] += j;
      }
    }
  }

  /* Set all remaining zeros to an arbitrary high value */
  for(i=0;i<matrix->len;i++) if(!matrix->d[i]) matrix->d[i] = 1000000;

  /* write out the matrix */
  MatrixIPrintBinary(stdout, matrix);
  exit(0);
}
