/* Rotates s and c by n*n matrix v before computing distance. */
static double Mahalanobis(double *s, double *c, double *v, int n, double *r)
{
  double sum = 0;
  int i,j;
  Vector re = VectorCreate(n);

  /* Loop eigenvectors */
  for(i=0;i<n;i++) {
    double d = 0;
    double m = 0;
    /* take dot product of s and c with the current eigenvector (v) 
     * and subtract the results.
     */
    for(j=0;j<n;j++) {
      d += (s[j] - c[j]) * v[j];
      m += v[j]*v[j];
    }
    re->d[i] = d / sqrt(m);
    r[i] = 0;
    sum += d*d;
    /* advance to next eigenvector */
    v += n;
  }
  /* project r back to real space */
  v -= n*n;
  for(i=0;i<n;i++) {
    for(j=0;j<n;j++) {
      r[j] += re->d[i] * v[j];
    }
    /* advance to next eigenvector */
    v += n;
  }
  VectorFree(re);
  return sum;
}

/*
 * Modifies: sample, cov, *nsample_p.
 */
void SurfaceRepel2(Matrix sample, Matrix cov, Matrix joint, 
		   double temperature, double timestep, int *nsample_p)
{
  int nsample = *nsample_p;
  int i, j, k;
  Matrix sample_d = MatrixCreate(sample->rows, sample->cols);
  /* temporary storage for inner loop */
  Matrix r = MatrixCreate(sample->rows, sample->cols);
  Vector dist = VectorCreate(sample->rows);

  VectorSet((Vector)joint, 0);
  VectorSet((Vector)sample_d, 0);

  /* compute membership */
  for(i=0;i<sample->rows;i++) {
    double joint_sum = 0, mag;
    for(j=0;j<sample->rows;j++) {
      double d;
      if(i == j) continue;
      d = Mahalanobis(sample->d2[i], sample->d2[j], cov->d2[j], sample->cols, 
		      r->d2[j]);
      dist->d[j] = d; /* save d for later */
      if(d+1 != d) { /* check for infinity */
	joint_sum += joint->d2[i][j] = exp(-d/temperature);
      }
    }
    if(joint_sum > 0.0) {
      for(j=0; j < sample->rows; j++) {
	joint->d2[i][j] /= joint_sum;
	for(k=0;k<sample->cols;k++) {
	  sample_d->d2[i][k] += joint->d2[i][j] * r->d2[j][k];
	}
      }
    }
    else {
      /* fall in here if temperature is too low */
      int closest = 0;
      /* find closest point */
      for(j=1; j < sample->rows; j++) {
	if(dist->d[j] < dist->d[closest]) closest = j;
      }
      for(j=0; j < sample->rows; j++) {
	if(j == closest) joint->d2[i][j] = 1;
	else joint->d2[i][j] = 0;
      }
      for(k=0;k<sample->cols;k++) {
	sample_d->d2[i][k] = r->d2[closest][k];
      }
    }

    /* randomly perturb the derivative */
    for(k=0;k<sample->cols;k++) {
      fprintf(stderr, "%g ", sample_d->d2[i][k]);
      sample_d->d2[i][k] += (RandReal()-0.5)/timestep;
    }
    fprintf(stderr, "\n");

    /* project the derivative */
    SurfaceProject(sample_d->d2[i], sample->d2[i], sample->cols, 0.5/timestep);

    /* at equilibrium? */
    mag = 0;
    for(k=0;k<sample->cols;k++) {
      mag += sample_d->d2[i][k] * sample_d->d2[i][k] * timestep * timestep;
    }
    fprintf(stderr, "mag: %g\n", sqrt(mag));
  }

  /* apply the derivatives */
  VectorScale((Vector)sample_d, timestep);
  VectorAdd((Vector)sample, (Vector)sample_d);

  VectorFree(dist);
  MatrixFree(r);
  MatrixFree(sample_d);
  *nsample_p = nsample;
}
