#define TIMESTEP 0.03
#define AMPLITUDE 6
#define FEEDBACK 15
#define DESIRED_E (.8 * AMPLITUDE)
#define DESIRED_R 40
#define EQUILIBRIUM 4
#define RADIUS_MAX (1.5 * DESIRED_R)
#define FISSION .2
#define DEATH .7

/* Modifies d so that s+d lies on the surface. */
static void SurfaceProject(double *d, double *s, int n)
{
  /* project onto a circle */
  double dx = s[0] - 200;
  double dy = s[1] - 200;
  /* gradient is [2dx 2dy] */
  double r = dx*dx + dy*dy;
  double f = r - 100*100;
  double dot = 2 * (d[0] * dx + d[1] * dy);
  d[0] -= (dot + FEEDBACK * f) * dx / 2 / r;
  d[1] -= (dot + FEEDBACK * f) * dy / 2 / r;
}

/*
 * Modifies: sample, radius, *nsample_p.
 */
void SurfaceRepel(Matrix sample, Vector radius, int *nsample_p)
{
  int nsample = *nsample_p;
  int i, j, k;
  Vector r = VectorCreate(sample->cols);
  Vector energy = VectorCreate(sample->rows);
  Vector radius_d = VectorCreate(radius->len);
  Matrix sample_d = MatrixCreate(sample->rows, sample->cols);
  VectorSet((Vector)sample_d, 0);
  VectorSet((Vector)energy, 0);

  for(i=0;i<nsample;i++) {
    double radius_gradient = 0;
    double ri2 = radius->d[i] * radius->d[i];

    for(j=0;j<nsample;j++) {
      double rmag = 0, eij, eji;
      double rj2 = radius->d[j] * radius->d[j];
      if(i == j) continue;

      /* compute distance betw s and s2 */
      for(k=0;k<sample->cols;k++) {
	r->d[k] = sample->d2[i][k] - sample->d2[j][k];
	rmag += r->d[k] * r->d[k];
      }
      eij = AMPLITUDE * exp(-rmag / 2 / ri2);
      eji = AMPLITUDE * exp(-rmag / 2 / rj2);
      energy->d[i] += eij;
      radius_gradient += rmag * eij;
	
      if(rmag > 0) {
	/* move samples apart */
	for(k=0;k<sample->cols;k++) {
	  sample_d->d2[i][k] += r->d[k] * (eij + eji * ri2 / rj2);
	}
      }
      else {
	/* randomly perturb samples */
	for(k=0;k<sample->cols;k++) {
	  sample_d->d2[i][k] += (RandReal()-0.5)*10;
	}
      }
    }

    /* project the derivative */
    SurfaceProject(sample_d->d2[i], sample->d2[i], sample->cols);

    /* compute new radius */
    fprintf(stderr, "energy: %g r_grad: %g\n", energy->d[i], radius_gradient);
    radius_gradient /= ri2 * radius->d[i];
    radius_d->d[i] = FEEDBACK * (DESIRED_E - energy->d[i]) / (radius_gradient+10);
  }

  /* apply the derivatives */
  VectorScale((Vector)sample_d, TIMESTEP);
  VectorAdd((Vector)sample, (Vector)sample_d);
  VectorScale(radius_d, TIMESTEP);
  VectorAdd(radius, radius_d);

  for(i=0;i<nsample;i++) {
    /* at equilibrium? */
    double dmag = 0;
    for(k=0;k<sample->cols;k++) {
      dmag += sample_d->d2[i][k] * sample_d->d2[i][k];
    }
    if(dmag < EQUILIBRIUM * radius->d[i]) {
      /* fission? */
      if((nsample < sample->rows) &&
	 ((radius->d[i] > RADIUS_MAX) || 
	  ((energy->d[i] > FISSION * DESIRED_E) && 
	   (radius->d[i] > DESIRED_R)))) {
	fprintf(stderr, "fission %d\n", i);
	/* make a new sample at a random offset */
	for(k=0;k<sample->cols;k++) {
	  sample->d2[nsample][k] = sample->d2[i][k] + 
	    (RandReal()-0.5)*radius->d[i];
	}
	radius->d[nsample] = radius->d[i] /= sqrt(2);
	/* the new sample is not at equilibrium */
	sample_d->d2[nsample][0] = 1000;
	nsample++;
      }
      /* kill? */
      else if(nsample && (radius->d[i] < RandReal()*DEATH*DESIRED_R)) {
	fprintf(stderr, "kill %d\n", i);
	nsample--;
	if(nsample > i) {
	  memcpy(sample->d2[i], sample->d2[nsample], 
		 sample->cols * sizeof(double));
	  radius->d[i] = radius->d[nsample];
	}
      }
    }
  }

  VectorFree(r);
  VectorFree(energy);
  MatrixFree(sample_d);
  VectorFree(radius_d);
  *nsample_p = nsample;
}
