#include <boa/boa.h>
#include <tpm/hash.h>

#define VERBOSE 0

/* Globals *******************************************************************/
#if GC
static int Cleaning = 0;      /* if true, don't remove objs from Used */
static HashTable Used = NULL; /* stores all allocated objs */
#endif

/* Prototypes ****************************************************************/

Obj Boa_CreateFromData(Type type, Any value);
Obj Boa_Create(Type type);
void Boa_Free(Obj t);
void Boa_FreeWithType(Obj t);
Obj BoaHold(Obj t);
Obj BoaCopyShallow(Obj t);
Obj BoaCopy(Obj t);
int BoaHash(Obj t);

int BoaFreeTest10(Obj t);
int BoaFreeTest01(Obj t);
void BoaFreeOR10(Obj t);
void BoaFreeOR01(Obj t);
void BoaFreeAND01(Obj t);
void BoaFreeAND10(Obj t);
static BoaFreeFunc Free11, Free10, Free01, Free00;

void GcStartup(void);
void GcShutdown(void);
void GcClean(int count, ...);

/* Functions *****************************************************************/

Obj Boa_CreateFromData(Type type, Any value)
{
  HashEntry *e;
  Obj t = Allocate(1, struct Obj);
  t->ref_count = 1;
  if(value) t->freeFunc = Free11;
  else t->freeFunc = Free00;
  t->value = value;
  t->type = type;
#if GC
  /* remember this object, for GC */
  e = HashTableAddEntry(Used, (char*)t, NULL); HashEntryValue(e) = t;
#endif
  return t;
}

/* Instantiates a Type. */
Obj Boa_Create(Type type)
{
  Obj t;
  if(!type) return NULL;
#if VERBOSE
  fprintf(stderr, "creating %d bytes\n", TypeSize(type));
#endif
  t = Boa_CreateFromData(type, AllocateC(TypeSize(type), char));
  Boa_Init(t);
  return t;
}

/* Functional form of Boa_Release. */
void Boa_Free(Obj t)
{
  Boa_Release(t);
}

/* Functional form of Boa_Hold. */
Obj BoaHold(Obj t)
{
  Boa_Hold(t);
  return t;
}

/* Releases t and its type. */
void Boa_FreeWithType(Obj t)
{
  Type type = t->type;
  Boa_Release(t);
  Boa_Release(type);
}

Obj BoaCopyShallow(Obj t)
{
  if(!t) return NULL;
  t = Boa_CreateFromData(t->type, 
			 AllocCopy(TypeSize(t->type), char, t->value));
  t->freeFunc = Free01; /* does not own the data */
  return t;
}

Obj BoaCopy(Obj t)
{
  void *data;
  if(!t) return NULL;
  data = AllocateC(TypeSize(t->type), char);
  if(Boa_Copy(t, data) != TPM_OK) {
    free(data);
    return NULL;
  }
  return Boa_CreateFromData(t->type, data);
}

/* There are four free functions for an Obj, corresponding to whether
 * the Obj owns the "value" memory block or the data in that block.
 * They are named using a two-bit mask, e.g.
 *   Free10 means the Obj owns the data but not the block.
 * Freeing an Obj never frees its type; Types are assumed to live forever.
 */
int BoaFreeTest10(Obj t)
{
  return ((t)->freeFunc == Free11 || (t)->freeFunc == Free10);
}
int BoaFreeTest01(Obj t)
{
  return ((t)->freeFunc == Free11 || (t)->freeFunc == Free01);
}

/* Modifies t->freeFunc so that it owns the value. */
void BoaFreeOR10(Obj t)
{
  if(t->freeFunc == Free01)
    t->freeFunc = Free11;
  else if(t->freeFunc == Free00)
    t->freeFunc = Free10;
}

/* Modifies t->freeFunc so that it owns the value memory. */
void BoaFreeOR01(Obj t)
{
  if(t->freeFunc == Free10)
    t->freeFunc = Free11;
  else if(t->freeFunc == Free00)
    t->freeFunc = Free01;
}

/* Modifies t->freeFunc so that it doesn't destroy the value. */
void BoaFreeAND01(Obj t)
{
  if(t->freeFunc == Free11)
    t->freeFunc = Free01;       
  else if(t->freeFunc == Free10)  
    t->freeFunc = Free00; 
}

/* Modifies t->freeFunc so that it doesn't free the value. */
void BoaFreeAND10(Obj t)
{
  if(t->freeFunc == Free11)
    t->freeFunc = Free10;
  else if(t->freeFunc == Free01)
    t->freeFunc = Free00;
}

/* BoaFreeFuncs */

static void Free11(Obj t)
{
  Boa_Final(t);
  free(t->value);
  Free00(t);
}

static void Free10(Obj t)
{
  Boa_Final(t);
  Free00(t);
}

static void Free01(Obj t)
{
  free(t->value);
  Free00(t);
}

static void Free00(Obj t)
{
#if GC
  if(Cleaning) free(t);
  else HashTableRemoveEntry(Used, HashTableFindEntry(Used, (char*)t));
#else
  free(t);
#endif
}

/* GC ************************************************************************/

#if GC

void GcStartup(void)
{
  Used = HashTableCreate(HASH_WORD, GenericFree);
}

static void ObjDestroy(Obj t)
{
  t->freeFunc(t);
}

/* Destroy all Objs */
void GcShutdown(void)
{
  /* GcClean(0); */
}

#define IsDead(t) ((t)->ref_count & 0x8000)
#define MakeDead(t) (t)->ref_count |= 0x8000
#define IsLive(t) !IsDead(t)
#define MakeLive(t) (t)->ref_count &= ~0x8000
#define IsType(t) ((t)->ref_count & 0x4000)
#define MakeType(t) (t)->ref_count |= 0x4000
#define UnType(t) (t)->ref_count &= ~0x4000
#define Reset(t) (t)->ref_count &= ~0xC000

static void MarkReachable(Obj t, Any data)
{
  if(IsLive(t)) return;
  MakeLive(t);
  Boa_Clean(t, MarkReachable, NULL);
}

static int ZeroCount(Obj t, Any dummy)
{
  return t->ref_count > 0;
}

/* XXX doesn't work */
/* Release all Objs not reachable from the given Obj arrays.
 * Usage: GcClean(3, array, 5, array, 0)
 */
void GcClean(int count, ...)
{
  Obj *ts, t;
  va_list args;

  /* Mark all Objs as Dead */
  {HashIter(b, e, t, Used) {
    MakeDead(t);
  }}

  va_start(args, count);
  while(count) {
    ts = va_arg(args, Obj*);
    /* Mark all reachable Objs as Live */
    while(count--) {
      MarkReachable(*ts, NULL);
    }
    count = va_arg(args, int);
  }
  va_end(args);
  /* count is now 0 */
  
  for(;;) {
    /* Mark unreachableBoa_Types */
    {HashIter(b, e, t, Used) {
      if(IsDead(t->type)) { MakeLive(t->type); MakeType(t->type); }
    }}
  
    /* Release all dead Objs */
    {HashIter(b, e, t, Used) {
      if(IsType(t)) MakeDead(t);
    }}
    Used->ff = (FreeFunc*)ObjDestroy;
    Cleaning = 1;
    count = HashTableRemoveValueAll(Used, NULL, (CmpFunc*)ZeroCount);
    Cleaning = 0;
    Used->ff = GenericFree;
    if(!count) break;
    fprintf(stderr, "GcClean: cleaned %d objects\n", count);
  }
}

#endif /* GC */
