/* A stream is just an Obj which responds to certain kinds of requests,
 * namely read/write/seek/tell.  This allows arbitrary data structures
 * to be used as streams.
 * This file contains convenience routines for making these requests on 
 * streams. 
 */
#include <boa/boa.h>

/* Prototypes ****************************************************************/
int Boa_Tell(Obj stream);
int Boa_Seek(Obj stream, int pos, int whence);
char *Boa_TellString(Obj stream);
int Boa_SeekString(Obj stream, char *pos);

int Boa_Getc(Obj stream);
char *Boa_GetLine(Obj stream);
int Boa_Scan(Obj stream, Obj value);

int Boa_Putc(Obj stream, char c);
int Boa_Puts(Obj stream, char *s);

/* Functions *****************************************************************/

/* fgetc */
int Boa_Getc(Obj stream)
{
  Obj t;
  char c;
  if(Boa_IntGet(stream, 1, 0, &T_Char, &t) != TPM_OK) return EOF;
  c = Boa_Char_Value(t);
  Boa_Release(t);
  return (int)c;
}

/* Eats value from stream. */
int Boa_Scan(Obj stream, Obj value)
{
  char *p;
  if(value->type == T_Char) {
    if(Boa_Getc(stream) != Boa_Char_Value(value)) {
      ErrorSet("Boa_Scan: failed to match `%c'", Boa_Char_Value(value));
      /* reset the stream */
      Boa_Seek(stream, -1, SEEK_CUR);
      return TPM_BOTTOM;
    }
    return TPM_OK;
  }
  if(value->type != T_String) return TPM_BOTTOM;
  p = Boa_String_Value(value);
  if(*p) {
    int pos = Boa_Tell(stream);
    do {
      if(Boa_Getc(stream) != *p) {
	ErrorSet("Boa_Scan: failed to match `%s'", Boa_String_Value(value));
	/* reset the stream */
	Boa_Seek(stream, pos, 0);
	return TPM_BOTTOM;
      }
    } while(*++p);
  }
  return TPM_OK;
}

/* getline */
char *Boa_GetLine(Obj stream)
{
  char *s;
  Obj line;
  if(Boa_StringGet(stream, "line", 0, &T_String, &line) != TPM_OK) return NULL;
  s = strdup(Boa_String_Value(line));
  Boa_Release(line);
  return s;
}

/* fputc */
int Boa_Putc(Obj stream, char c)
{
  Obj t;
  int status;
  t = Boa_CreateFromData(T_Char, NULL);
  t->value = &c;
  status = Boa_IntSet(stream, 1, 0, t);
  Boa_Release(t);
  return status;
}

/* fputs */
int Boa_Puts(Obj stream, char *s)
{
  Obj t;
  int status;
  t = Boa_CreateFromData(T_String, NULL);
  t->value = &s;
  status = Boa_IntSet(stream, 1, 0, t);
  Boa_Release(t);
  return status;
}

/* ftell */
int Boa_Tell(Obj stream)
{
  int i;
  Obj pos;
  if(Boa_StringGet(stream, "tell", 0, &T_Int, &pos) != TPM_OK) return -1;
  i = Boa_Int_Value(pos);
  Boa_Release(pos);
  return i;
}

/* fseek, without SEEK_END.
 * returns an error code.
 */
int Boa_Seek(Obj stream, int pos, int whence)
{
  if(whence == 0) {
    Obj seek;
    int status = Boa_StringGet(stream, "seek", 0, NULL, &seek);
    if(status == TPM_OK) {
      status = Boa_IntGet(seek, pos, 0, NULL, NULL);
      Boa_Release(seek);
    }
    if(status != TPM_OK) fprintf(stderr, "seek error\n");
    return status;
  }
  else return Boa_IntGet(stream, pos, 0, NULL, NULL);
}

/* string version of tell; returns name, if any, of current position.
 * returns NULL on failure, otherwise
 * caller must free the string.
 */
char *Boa_TellString(Obj stream)
{
  char *s;
  Obj pos;
  if(Boa_StringGet(stream, "tell", 0, &T_String, &pos) != TPM_OK) return NULL;
  s = strdup(Boa_String_Value(pos));
  Boa_Release(pos);
  return s;
}

/* seek to a string position */
int Boa_SeekString(Obj stream, char *pos)
{
  Obj seek;
  int status = Boa_StringGet(stream, "seek", 0, NULL, &seek);
  if(status == TPM_OK) {
    status = Boa_StringGet(seek, pos, 0, NULL, NULL);
    Boa_Release(seek);
  }
  return status;
}
