/* Type is just an Obj which describes Objs.
 * For efficiency reasons, all Types must store:
 *   1. a method table 
 *   2. the size of an instance
 *   3. the required alignment of an instance
 * Like all Objs, Types have a meta-Obj which describes them. These
 * are also Types, and provide accessors like "size" for the Type.  
 * The regress ends with Type SimpleType, which has itself as a meta-Obj.
 */
/* Types must supply methods for the Boa API:
 * initializer/finalizer
 * human-readable output method
 * copy/compare/hash methods
 * dictionary access methods (get/set with various kinds of keys)
 */
#include <tpm/tpm.h>       /* for FILE, MakeAligned, TPM_OK */
#include <tpm/strstream.h> /* for SPrint */
#include <boa/obj.h>

/* Basic operations. Return TPM_ERROR on failure. */
/* Initializes t->value, a block which has already been allocated and
 * cleared. 
 */
typedef void Boa_InitFunc(Obj t);
/* Called when t->ref_count hits zero.  Cleans up memory besides that
 * used for t and t->value.
 */
typedef void Boa_FinalFunc(Obj t);
/* Calls func(token, data) on all tokens held by object. */
typedef void Boa_CleanFunc(Obj t, ObjFunc *func, Any data);
/* human-readable description */
typedef int Boa_SPrintFunc(Obj t, StrStream ss);
#if 0
/* t releases nargs arguments and replaces them with nrets return values.
 * In case of error, return values are omitted but args are still released.
 * args is a vector of max(nargs, nrets) Objs.
 */
typedef int Boa_CallFunc(Obj t, int nargs, int nrets, Obj *args);
#endif
/* Writes a deep copy of object into the block at *value_return. */
typedef int Boa_CopyFunc(Obj t, Ptr value_return);
/* Returns 0 iff a and b have equal values. */
typedef int Boa_CompareFunc(Obj a, Obj b);
/* If equal(a, b), then hash(a) == hash(b). */
typedef int Boa_HashFunc(Obj t);

/* Context operations. 
 * Return TPM_OK on success, TPM_ERROR on bad request, 
 * and TPM_BOTTOM on missing item. Type mismatch should be handled
 * as a missing item, to allow multi-methods.
 */
/* Flags:
 *   BOA_REPLACE  Replace the old value if item is present (Set only)
 *   BOA_INSERT   Insert the value if item is not present
 *   BOA_DELETE   Remove the item, if present (Get only)
 *   BOA_SPLIT    Name may have components separated by BOA_SEP(flags)
 */
#define BOA_INSERT (1<<9)
#define BOA_DELETE (1<<10)
#define BOA_REPLACE BOA_DELETE
#define BOA_SPLIT 0x2e
#define BOA_SEP(f) (char)((f) & 0xff)

/* Queries the value of an item.
 * If value_return == NULL, checks for existence of the item only.
 * If DELETE is set, returns the old value if value_return is non-NULL.
 * If type_in_out == NULL, no type check is performed and no type is returned.
 * If *type_in_out == NULL, no type check, type of the item is returned.
 * If *type_in_out != NULL, only queries items matching that type.
 * If successful, *value_return gives hold to caller.
 */
typedef int Boa_StringGetFunc(Obj self, char *name, int flags,
			      Type *type_in_out, Obj *value_return);
typedef int Boa_IntGetFunc(Obj self, int index, int flags,
			   Type *type_in_out, Obj *value_return);
typedef int Boa_ObjGetFunc(Obj self, Obj index, int flags,
			   Type *type_in_out, Obj *value_return);

/* Modifies/creates an item.
 * If the item exists and REPLACE is clear, an error is returned.
 * If item exists and REPLACE is set, the item is released and assigned
 * to the given value, which must be non-NULL.
 * If the item doesn't exist and INSERT is set, a new item is created
 * whose default value is given (may be NULL); otherwise BOTTOM is returned.
 * If successful, caller relinquishes hold on value.
 */
typedef int Boa_StringSetFunc(Obj self, char *name, int flags, Obj value);
typedef int Boa_IntSetFunc(Obj self, int index, int flags, Obj value);
typedef int Boa_ObjSetFunc(Obj self, Obj index, int flags, Obj value);

/* method table */
typedef struct Boa_Methods {
  Boa_InitFunc *init;
  Boa_FinalFunc *final;
  Boa_CleanFunc *clean;
  Boa_SPrintFunc *sprint;
  Boa_CopyFunc *copy;
  Boa_CompareFunc *compare;
  Boa_HashFunc *hash;

  Boa_StringGetFunc *stringGet;
  Boa_StringSetFunc *stringSet;
  Boa_IntGetFunc *intGet;
  Boa_IntSetFunc *intSet;
  Boa_ObjGetFunc *objGet;
  Boa_ObjSetFunc *objSet;
} *Boa_Methods;

/* Must be first data members in all structures claiming to be Types */
#define TYPE_HEAD  \
  Boa_Methods methods; \
  int size, align; \
  char *name;

typedef struct Boa_SimpleType {
  TYPE_HEAD
} *Boa_SimpleType;

#define TypeMethods(t) ((Boa_SimpleType)(t)->value)->methods
#define TypeSize(t) ((Boa_SimpleType)(t)->value)->size
#define TypeAlign(t) ((Boa_SimpleType)(t)->value)->align
#define TypeName(t) ((Boa_SimpleType)(t)->value)->name

#define Boa_Methods(t) TypeMethods((t)->type)
#define Boa_Size(t) TypeSize((t)->type)
#define Boa_Align(t) TypeAlign((t)->type)

/* Convenience macros for calling methods */
#define Boa_Init(t) Boa_Methods(t)->init(t)
#define Boa_Final(t) Boa_Methods(t)->final(t)
#define Boa_Clean(t, f, d) Boa_Methods(t)->clean(t, f, d)
#define Boa_SPrint(t, ss) Boa_Methods(t)->sprint(t, ss)
#define Boa_Copy(t, d) Boa_Methods(t)->copy(t, d)
#define Boa_Compare(a, b) ((a != b) && BoaCompare(a, b))
#define Boa_Hash(t) Boa_Methods(t)->hash(t)
#define Boa_StringGet(t, n, f, type, v) Boa_Methods(t)->stringGet(t, n, f, type, v)
#define Boa_StringSet(t, n, f, v) Boa_Methods(t)->stringSet(t, n, f, v)
#define Boa_IntGet(t, n, f, type, v) Boa_Methods(t)->intGet(t, n, f, type, v)
#define Boa_IntSet(t, n, f, v) Boa_Methods(t)->intSet(t, n, f, v)
#define Boa_ObjGet(t, n, f, type, v) Boa_Methods(t)->objGet(t, n, f, type, v)
#define Boa_ObjSet(t, n, f, v) Boa_Methods(t)->objSet(t, n, f, v)

/* methods.c *****************************************************************/
Boa_InitFunc Empty_Init;
Boa_FinalFunc Empty_Final;
Boa_CleanFunc Empty_Clean;

Boa_FinalFunc Generic_Final;
Boa_SPrintFunc Generic_SPrint;
Boa_CopyFunc Generic_Copy;
Boa_CompareFunc Generic_Compare;
Boa_CompareFunc Identity_Compare;
Boa_HashFunc Generic_Hash;
Boa_FinalFunc Sub_Final;
Boa_CleanFunc Sub_Clean;
Boa_IntGetFunc Sub_IntGet;
Boa_IntSetFunc Sub_IntSet;
Boa_StringGetFunc Sub_StringGet;
Boa_StringSetFunc Sub_StringSet;
Boa_StringGetFunc Generic_StringGet;
Boa_StringSetFunc Generic_StringSet;
Boa_ObjGetFunc Generic_ObjGet;
Boa_ObjSetFunc Generic_ObjSet;
Boa_StringGetFunc Generic_SplitGet;
Boa_StringSetFunc Generic_SplitSet;

/* Always return TPM_ERROR */
Boa_CopyFunc Err_Copy;
Boa_SPrintFunc Err_SPrint;
Boa_StringGetFunc Err_StringGet;
Boa_StringSetFunc Err_StringSet;
Boa_IntGetFunc Err_IntGet;
Boa_IntSetFunc Err_IntSet;
Boa_ObjGetFunc Err_ObjGet;
Boa_ObjSetFunc Err_ObjSet;
