#include "parse.h"
#include <tpm/list.h>
#include <ctype.h>

ST ST_Parse(StrStream ss);
void ST_Free(ST st);
void ST_Print(FILE *fp, ST st);

/* Functions *****************************************************************/

/* Reads up to len bytes (including terminator) into s,
 * stopping at c, discarding that character.
 * Returns TPM_ERROR if no data in ss or len <= 0.
 */
int StrStreamGetUntil(StrStream ss, char c, char *s, int len)
{
  if((ss->rpos == ss->wpos) || (len <= 0)) {
    ErrorString = TPM_EEMPTY;
    return TPM_ERROR;
  }
  for(len--;len && (ss->rpos < ss->wpos);len--) {
    *s = ss->data[ss->rpos++];
    if(*s == c) break;
    else s++;
  }
  *s = 0;
  return TPM_OK;
}

ST ST_Parse(StrStream ss)
{
  char c, s[256];

  do {
    /* read a character */
    if(StrStreamGetc(ss, &c) != TPM_OK) return NULL;
  } while(isspace(c));
  
  /* close? */
  if(c == ')') return NULL;

  /* application? */
  if(c == '(') {
    ST st, child;
    List children = ListCreate(NULL);
    st = Allocate(1, union ST);
    st->m.node_type = ST_APP;

    for(;;) {
      child = ST_Parse(ss);
      if(!child) break;
      ListAddRear(children, child);
    }

    st->m.child = ListToPtrArray(children, &st->m.children);
    ListFree(children);
    return st;
  }

  /* int? */
  if(isdigit(c)) {
    ST st;
    char *p = s;
    *p++ = c;
    for(;;) {
      if(StrStreamGetc(ss, &c) != TPM_OK) break;
      if(!isdigit(c)) { StrStreamUnget(ss, 1); break; }
      *p++ = c;
    }
    *p = '\0';
    st = Allocate(1, union ST);
    st->v.node_type = ST_INT;
    st->v.value = (Any)atoi(s);
    return st;
  }

  /* string? */
  if(c == '"') {
    ST st;
    StrStreamGetUntil(ss, '"', s, 256);
    st = Allocate(1, union ST);
    st->v.node_type = ST_STRING;
    st->v.value = (Any)strdup(s);
    return st;
  }

  /* name */
  {
    ST st;
    char *p = s;
    *p++ = c;
    for(;;) {
      if(StrStreamGetc(ss, &c) != TPM_OK) break;
      if(isspace(c)) break;
      if(c == '(' || c == ')') { StrStreamUnget(ss, 1); break; }
      *p++ = c;
    }
    *p = '\0';
    st = Allocate(1, union ST);
    st->v.node_type = ST_NAME;
    st->v.value = (Any)strdup(s);
    return st;
  }
}

void ST_Free(ST st)
{
  int i;
  switch(st->any.node_type) {
  case ST_APP:
    for(i=0;i<st->m.children;i++) ST_Free(st->m.child[i]);
    free(st->m.child);
    break;
  case ST_STRING:
  case ST_NAME:
    free(st->v.value);
    break;
  }
  free(st);
}

void ST_Print(FILE *fp, ST st)
{
  int i;
  switch(st->any.node_type) {
  case ST_INT:
    fprintf(fp, "%d ", (int)st->v.value);
    break;
  case ST_STRING:
    fprintf(fp, "\"%s\" ", (char*)st->v.value);
    break;
  case ST_NAME:
    fprintf(fp, "%s ", (char*)st->v.value);
    break;
  case ST_APP:
    fprintf(fp, "(");
    for(i=0;i<st->m.children;i++) ST_Print(fp, st->m.child[i]);
    fprintf(fp, ") ");
    break;
  }
}
