/* Evaluator object.  StringGet with an s-expression will return a value.
 * Attributes:
 *   context    Where to resolve names
 * Requires: Int, String, SimpleType.
 */

#include <boa/boa.h>
#include "parse.h"
#include <ctype.h>

typedef struct Eval {
  Obj context;
} *Eval;

/* Globals *******************************************************************/
Type T_Eval;

/* Functions *****************************************************************/

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Eval e = (Eval)self->value;
  func(e->context, data);
}

static int SPrint(Obj self, StrStream ss)
{
  return StrStreamPuts(ss, "anEval");
}

static Obj Evaluate(ST st, Obj context)
{
  Obj t;
  int i, status;
  switch(st->any.node_type) {
  case ST_INT:
    return Boa_Int_Create((int)st->v.value);
  case ST_STRING:
    return Boa_String_Create(strdup(st->v.value));
  case ST_NAME:
    if(Boa_StringGet(context, st->v.value, 0, NULL, &t) != TPM_OK) {
      ErrorPrepend("unbound name `%s'", st->v.value);
      return NULL;
    }
    return t;
  case ST_APP:
    if(st->m.children < 1) return NULL;
    t = Evaluate(st->m.child[0], context);
    if(!t) return NULL;
    for(i=1;i<st->m.children;i++) {
      Obj arg = Evaluate(st->m.child[i], context);
      Obj nt;
      if(!arg) { /* abort */
	Boa_Release(t);
	return NULL; 
      }
      status = Boa_ObjGet(t, arg, 0, NULL, &nt);
      Boa_Release(arg);
      if(status != TPM_OK) break;
      Boa_Release(t);
      t = nt;
    }
    if(status == TPM_OK) return t;
    else return NULL;
  }
  fprintf(stderr, "Internal error: bad node type (%d) in Evaluate\n", 
	  st->any.node_type);
  return NULL;
}

static int StringGet(Obj self, char *name, int flags,
		     Type *type_in_out, Obj *value_return)
{
  Eval e = (Eval)self->value;
  ST st;
  StrStream ss;
  Obj t;

  /* delete not supported */
  if(flags & BOA_DELETE) return TPM_ERROR;

  /* parse the request into a syntax tree */
  ss = StrStreamFromString(name, strlen(name), 0);
  st = ST_Parse(ss);
  StrStreamFree(ss);
  if(!st) return TPM_BOTTOM;

  /* evaluate the syntax tree */
  t = Evaluate(st, e->context);
  ST_Free(st);
  if(!t) return TPM_BOTTOM;
  BOA_TYPECHECK_GET(t->type);
  if(value_return) *value_return = t;
  else Boa_Release(t);
  return TPM_OK;
}

/* Read an expression into a StrStream and call StringGet */
static int ObjGet(Obj self, Obj stream, int flags, 
		  Type *type_in_out, Obj *value_return)
{
  StrStream ss = StrStreamCreate(16, STRSTREAM_UNLIMITED);
  char depth = 0, gotchar = 0, quote = 0;
  int status, pos = Boa_Tell(stream);
  for(;;) {
    int c = Boa_Getc(stream);
    if(c == EOF) { if(!depth && !quote) break; else goto error; }
    if(!depth && !quote && isspace(c)) { if(gotchar) break; else continue; }
    gotchar = 1;
    StrStreamPutc(ss, (char)c);
    if(c == '"') quote = !quote;
    else if(!quote) {
      if(c == '(') depth++;
      else if(c == ')') { 
	if(--depth == 0) {
	  /* eat one character of whitespace */
	  Boa_Getc(stream); break; 
	}
      }
    }
  }
  status = StringGet(self, StrStreamAsString(ss), flags, 
		     type_in_out, value_return);
  StrStreamFree(ss);
  return status;

 error:
  StrStreamFree(ss);
  Boa_Seek(stream, pos, 0);
  return TPM_BOTTOM;
}

/* Write a representation of an object to the stream, which can be read
 * back in with the same Eval.
 * For now, just hope that SPrint will do this.
 * Could also scan through the context for a matching object.
 */
static int ObjSet(Obj self, Obj stream, int flags, Obj value)
{
  int status;
  StrStream ss = StrStreamCreate(16, STRSTREAM_UNLIMITED);
  status = Boa_SPrint(value, ss);
  if(status != TPM_OK) return TPM_BOTTOM;
  status = Boa_Puts(stream, StrStreamAsString(ss));
  StrStreamFree(ss);
  if(status != TPM_OK) return TPM_BOTTOM;
  return status;
}

static struct Boa_Methods methods = {
  Empty_Init, Generic_Final, Clean,
  SPrint, Generic_Copy, Generic_Compare, Generic_Hash,
  StringGet, Err_StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_Eval_Define(void)
{
  T_Eval = Boa_SimpleType_Create("eval", &methods, 
				 sizeof(struct Eval),
				 sizeof(Ptr));
  return T_Eval;
}
