/* The Boa package.  Add to Tcl with Boatcl_Init(interp).
 * Commands:
 *   create type
 *   hold obj
 *   free obj
 *   parse formatter type string
 *   unparse ?formatter? obj
 *   print ?file? obj
 *   get obj item	  (StringGet)
 *   getu obj item	  (StringGet)
 *   intGet obj item
 *   objGet obj item
 *   probe obj item
 *   delete obj item ?var?
 *   set obj item value	  (StringSet)
 *   intSet obj item value
 *   objSet obj item value
 *   copy obj
 *   typeof obj
 *   settype obj type
 *   refs obj
 *   cmd name obj	Define a Tcl command <name> which calls <obj>
 *   tk_image obj img   Copy a TpmImage into an existing TkImage (img)
 *   tcl_var obj var    Unparse an Obj into a Tcl variable, releasing the obj
 *   string value
 *   int value
 *   float value
 *   double value
 *   type name
 *   formatter name
 *   global name
 *   pkg name           Load a package by calling Boa_Name_Pkg()
 */
#include <tcl.h>
#include "tcl_macros.h"
#include <tk.h>  /* for Tk_PhotoHandle */

#include <boa/boa.h>
#include <boa/tcl.h>
#include <boa/image.h>
#include <tpm/dyn.h>  /* for pkg */

/* this is defined in type.h */
#undef Boa_Init

/* Prototypes ****************************************************************/
int Boa_Init(Tcl_Interp *interp);
int Boa_Cmd(ClientData clientData, Tcl_Interp *interp,
	       int argc, char *argv[]);

/* Private */
static void CreateCommand(Tcl_Interp *interp, char *cmd, Obj closure);

/* Functions *****************************************************************/

static int SetObjResult(Tcl_Interp *interp, Obj obj)
{
  char *s = Boa_AsString(obj);
  Err(!s, "sprint failed");
  Tcl_SetResult(interp, s, TCL_DYNAMIC);
  return TCL_OK;
}

int Boa_Cmd(ClientData clientData, Tcl_Interp *interp,
	    int argc, char *argv[])
{
  Obj obj;
  int status;
  Err(argc < 2, "missing directive: must be one of: get, set, delete, create, free, parse, unparse, print, scan, type, refs, tk_image");
  if(!strcmp(argv[1], "get")) {
    Obj value;
    Err(argc != 4, "usage: get obj item");
    GetPtr(argv[2], obj);
    Err(Boa_StringGet(obj, argv[3], '/', NULL, &value), 
	ErrorString);
    ReturnPtr(value);
  }
  else if(!strcmp(argv[1], "getu")) {
    Obj value;
    Err(argc != 4, "usage: getu obj item");
    GetPtr(argv[2], obj);
    Err(Boa_StringGet(obj, argv[3], 0, NULL, &value), 
	ErrorString);
    ReturnPtr(value);
  }
  else if(!strcmp(argv[1], "intGet")) {
    int index;
    Obj value;
    Err(argc != 4, "usage: intGet obj index");
    GetPtr(argv[2], obj);
    GetInt(argv[3], index);
    Err(Boa_IntGet(obj, index, '/', NULL, &value), 
	ErrorString);
    ReturnPtr(value);
  }
  else if(!strcmp(argv[1], "objGet")) {
    Obj index, value;
    Err(argc != 4, "usage: objGet obj item");
    GetPtr(argv[2], obj);
    GetPtr(argv[3], index);
    Err(Boa_ObjGet(obj, index, '/', NULL, &value), 
	ErrorString);
    ReturnPtr(value);
  }
  else if(!strcmp(argv[1], "keyGet")) {
    int index;
    Obj key;
    Err(argc != 4, "usage: keyGet obj index");
    GetPtr(argv[2], obj);
    GetInt(argv[3], index);
    if(obj->type == T_Ord) {
      Err((index < 0) || (index >= Boa_Ord_Length(obj)), "index out of range");
      key = Boa_Ord_Name(obj, index);
      Boa_Hold(key);
    }
    else {
      Err(1, "don't know how to keyGet from that type");
    }
    ReturnPtr(key);
  }
  else if(!strcmp(argv[1], "string")) {
    Err(argc != 3, "usage: string value");
    obj = Boa_String_Create(strdup(argv[2]));
    ReturnPtr(obj);
  }
  else if(!strcmp(argv[1], "type")) {
    Err(argc != 3, "usage: type name");
    if(Boa_StringGet(Boa_Types, argv[2], 0, NULL, &obj) != TPM_OK) {
      Tcl_AppendResult(interp, "type \"", argv[2], "\" doesn't exist", NULL);
      return TCL_ERROR;
    }
    ReturnPtr(obj);
  }
  else if(!strcmp(argv[1], "formatter")) {
    Err(argc != 3, "usage: formatter name");
    if(Boa_StringGet(Boa_Formatters, argv[2], 0, NULL, &obj) != TPM_OK) {
      Tcl_AppendResult(interp, "formatter \"", argv[2], "\" doesn't exist", NULL);
      return TCL_ERROR;
    }
    ReturnPtr(obj);
  }
  else if(!strcmp(argv[1], "global")) {
    Err(argc != 3, "usage: global name");
    if(Boa_StringGet(Boa_Globals, argv[2], '/', NULL, &obj) != TPM_OK) {
      Tcl_AppendResult(interp, "global \"", argv[2], "\" doesn't exist", NULL);
      return TCL_ERROR;
    }
    ReturnPtr(obj);
  }
  else if(!strcmp(argv[1], "int")) {
    int i;
    Err(argc != 3, "usage: int value");
    GetInt(argv[2], i);
    obj = Boa_Int_Create(i);
    ReturnPtr(obj);
  }
  else if(!strcmp(argv[1], "float")) {
    double d;
    Err(argc != 3, "usage: float value");
    GetDouble(argv[2], d);
    obj = Boa_Float_Create(d);
    ReturnPtr(obj);
  }
  else if(!strcmp(argv[1], "double")) {
    double d;
    Err(argc != 3, "usage: double value");
    GetDouble(argv[2], d);
    obj = Boa_Double_Create(d);
    ReturnPtr(obj);
  }
  else if(!strcmp(argv[1], "tcl_var")) {
    char *s, *p;
    Err(argc != 4, "usage: tcl_var obj var");
    GetPtr(argv[2], obj);
    s = Boa_AsString(obj);
    Boa_Release(obj);
    Err(!s, ErrorString);
    /* remove outermost braces and quotes from s */
    if(s[0] == '{' || s[0] == '"') {
      p = s+1;
      p[strlen(p)-1] = 0;
    }
    else p = s;
    status = Tcl_SetVar(interp, argv[3], p, TCL_LEAVE_ERR_MSG)
             ? TCL_OK : TCL_ERROR;
    free(s);
    if(status != TCL_OK) return status;
  }
  else if(!strcmp(argv[1], "probe")) {
    Type type;
    Err(argc != 4, "usage: probe obj item");
    GetPtr(argv[2], obj);
    type = NULL;
    status = Boa_StringGet(obj, argv[3], 0, &type, NULL);
    /* release type? */
    return SetObjResult(interp, type);
  }
#ifdef _TK
  else if(!strcmp(argv[1], "tk_image")) {
    Tk_PhotoHandle photo;
    Err(argc != 4, "usage: tk_image obj img");
    GetPtr(argv[2], obj);
    /* is this a valid image name? */
    if((photo = Tk_FindPhoto(argv[3])) == NULL) {
      Tcl_AppendResult(interp, "image \"", argv[3],
		       "\" doesn't exist", (char *) NULL);
      return TCL_ERROR;
    }
    TpmImageFillTk(photo, Boa_Image_Value(obj));
    Boa_Release(obj);
  }
#endif
  else if(!strcmp(argv[1], "set")) {
    Obj value;
    Err(argc != 5, "usage: set obj item value");
    GetPtr(argv[2], obj);
    GetPtr(argv[4], value);
    Err(Boa_StringSet(obj, argv[3], BOA_INSERT|BOA_REPLACE|'/', value),
	ErrorString);
  }
  else if(!strcmp(argv[1], "intSet")) {
    int index;
    Obj value;
    Err(argc != 5, "usage: intSet obj item value");
    GetPtr(argv[2], obj);
    GetInt(argv[3], index);
    GetPtr(argv[4], value);
    Err(Boa_IntSet(obj, index, BOA_INSERT|BOA_REPLACE|'/', value),
	ErrorString);
  }
  else if(!strcmp(argv[1], "objSet")) {
    Obj index, value;
    Err(argc != 5, "usage: objSet obj item value");
    GetPtr(argv[2], obj);
    GetPtr(argv[3], index);
    GetPtr(argv[4], value);
    Err(Boa_ObjSet(obj, index, BOA_INSERT|BOA_REPLACE|'/', value),
	ErrorString);
  }
  else if(!strcmp(argv[1], "copy")) {
    Obj value;
    Err(argc != 3, "usage: copy obj");
    GetPtr(argv[2], obj);
    value = BoaCopy(obj);
    ReturnPtr(value);
  }
  else if(!strcmp(argv[1], "delete")) {
    Obj value, *value_return;
    Err((argc < 4) || (argc > 5), "usage: delete obj item ?var?");
    GetPtr(argv[2], obj);
    if(argc == 5) value_return = &value;
    else value_return = NULL;
    Err(Boa_StringGet(obj, argv[3], BOA_DELETE, NULL, value_return), 
	ErrorString);
    if(value_return) {
      char str[10];
      sprintf(str, "%ld", (long)*value_return);
      Tcl_SetVar(interp, argv[4], str, 0);
    }
  }
  else if(!strcmp(argv[1], "parse")) {
    /* to define a type, use
     *   parse type 0 int
     */
    Obj formatter, stream;
    Type type;
    Err(argc != 5, "usage: parse formatter type string");
    GetPtr(argv[3], type);
    formatter = Boa_GetFormatter(argv[2]);
    if(!formatter) {
      Tcl_AppendResult(interp, "no such formatter `", argv[2], "'", NULL);
      return TCL_ERROR;
    }
    stream = Boa_StrStream_FromString(argv[4]);
    status = Boa_ObjGet(formatter, stream, 0, &type, &obj);
    Boa_Release(stream);
    Boa_Release(formatter);
    Err(status, ErrorString);
    ReturnPtr(obj);
  }
  else if(!strcmp(argv[1], "unparse")) {
    Err((argc < 3) || (argc > 4), "usage: unparse ?formatter? obj");
    if(argc == 3) {
      GetPtr(argv[2], obj);
      return SetObjResult(interp, obj);
    }
    else {
      Obj formatter, stream;
      formatter = Boa_GetFormatter(argv[2]);
      if(!formatter) {
	Tcl_AppendResult(interp, "no such formatter `", argv[2], "'", NULL);
	return TCL_ERROR;
      }
      GetPtr(argv[3], obj);
      stream = Boa_Create(T_StrStream);
      Boa_Hold(obj);
      status = Boa_ObjSet(formatter, stream, 0, obj);
      Tcl_SetResult(interp, StrStreamAsString(Boa_StrStream_Value(stream)), 
		    TCL_VOLATILE);
      Boa_Release(stream);
      Boa_Release(formatter);
      Err(status, ErrorString);
    }
  }
  else if(!strcmp(argv[1], "create")) {
    Type type;
    Err(argc < 3, "usage: create type");
    GetPtr(argv[2], type);
    obj = Boa_Create(type);
    ReturnPtr(obj);
  }
  else if(!strcmp(argv[1], "hold")) {
    Err(argc != 3, "usage: hold obj");
    GetPtr(argv[2], obj);
    Boa_Hold(obj);
    ReturnPtr(obj);
  }
  else if(!strcmp(argv[1], "free")) {
    Err(argc != 3, "usage: free obj");
    GetPtr(argv[2], obj);
    Boa_Release(obj);
  }
  else if(!strcmp(argv[1], "print")) {
    FILE *fp;
    int i;
    Err((argc < 3) || (argc > 4), "usage: print ?file? obj");
    if(argc == 3) { fp = stdout; i = 2; }
    else { fp = stderr; i = 3; }
    GetPtr(argv[i], obj);
    Err(Boa_Print(fp, obj), ErrorString);
    putc('\n', fp);
  }
  else if(!strcmp(argv[1], "value")) {
    Err(argc != 3, "usage: value obj");
    GetPtr(argv[2], obj);
    ReturnPtr(obj->value);
  }
  else if(!strcmp(argv[1], "value0")) {
    Err(argc < 3, "usage: value0 obj ?ptr?");
    GetPtr(argv[2], obj);
    if(argc > 3) {
      GetPtr(argv[3], ((Obj**)obj->value)[0]);
    }
    ReturnPtr(((Obj**)obj->value)[0]);
  }
  else if(!strcmp(argv[1], "typeof")) {
    Err(argc != 3, "usage: typeof obj");
    GetPtr(argv[2], obj);
    ReturnPtr(obj->type);
  }
  else if(!strcmp(argv[1], "settype")) {
    Err(argc != 4, "usage: settype obj type");
    GetPtr(argv[2], obj);
    GetPtr(argv[3], obj->type);
  }
  else if(!strcmp(argv[1], "refs")) {
    Err(argc != 3, "usage: refs obj");
    GetPtr(argv[2], obj);
    sprintf(interp->result, "%ld", obj->ref_count);
  }
  else if(!strcmp(argv[1], "pkg")) {
    char name[256];
    Boa_PkgFunc *f;
    Err(argc != 3, "usage: pkg name");
    sprintf(name, "Boa_%s_Pkg", argv[2]);
    /* Capitalize the package name */
    name[4] += 'A' - 'a';
    f = (Boa_PkgFunc*)DynSymbol(NULL, name);
    if(!f) {
      Tcl_AppendResult(interp, "symbol \"", name, "\" doesn't exist", NULL);
      return TCL_ERROR;
    }
    f();
  }
#if 0
  else if(!strcmp(argv[1], "cmd")) {
    Err(argc != 4, "usage: cmd name closure");
    GetPtr(argv[3], obj);
    Boa_Hold(obj);
    CreateCommand(interp, argv[2], obj);
  }
#endif
  else if(!strcmp(argv[1], "eval")) {
    Tcl_Interp *sub_interp;
    int status;
    Err(argc != 4, "usage: eval tcl_obj cmd");
    GetPtr(argv[2], obj);
    sub_interp = Boa_Tcl_Interp(obj);
    status = Tcl_Eval(sub_interp, argv[3]);
    Tcl_SetResult(interp, sub_interp->result, TCL_VOLATILE);
    return status;
  }
  else {
    Err(1, "unknown directive");
  }
  return TCL_OK;
}

/* This is executed only when Boa is loaded into Tcl 
 * (not the other way around).
 */
int Boa_Init(Tcl_Interp *interp)
{
  Boa_Startup();
  Boa_Tcl_Pkg();
  Boa_Image_Pkg();
  Tcl_CreateCommand(interp, "boa", Boa_Cmd, NULL, 
		    (Tcl_CmdDeleteProc*)Boa_Shutdown);
  return Tcl_PkgProvide(interp, "boa", "1");
}

#if 0
/* This is the Tcl command created by "boa cmd".
 * It calls the closure with the string arguments from Tcl.
 */
static int ProcCmd(Obj closure, Tcl_Interp *interp,
		   int argc, char *argv[])
{
  Obj name_tok, t;
  char **names;
  int i;

  TRY(BoaCall(closure, 0, 0, NULL));
  /* clear anything the closure might have done */
  Tcl_ResetResult(interp);
  return TCL_OK;
}

static void ProcDelete(Obj closure)
{
  Boa_Release(closure);
}

static void CreateCommand(Tcl_Interp *interp, char *cmd, Obj closure)
{
  Obj mux, cvt, tcl;
  
  /* create a Tcl context for our interp */
  tcl = ObjNew(T_Tcl);
  BoaTcl_Interp(tcl->value) = interp;

  /* create a TclCvt for communicating with Tcl */
  cvt = ObjNew(T_TclCvt);
  ObjHold(tcl);
  BoaSetObj(cvt, "_tcl", tcl);
  mux = ObjNew(T_Mux);
  /* use a dict to hold the arguments */
  Mux_Set(mux->value, 1, ObjNew(T_Dict));
  /* let Tcl handle the rest */
  Mux_Set(mux->value, 2, tcl);
  BoaSetObj(cvt, "_context", mux);

  /* make a mux between Tcl and the closure */
  mux = ObjNew(T_Mux);
  Mux_Set(mux->value, 1, cvt);
  Mux_Set(mux->value, 2, closure);

  Tcl_CreateCommand(interp, cmd, (Tcl_CmdProc*)ProcCmd, mux,
		    (Tcl_CmdDeleteProc*)ProcDelete);
}
#endif
