/* Definition of the vm command created by "boa vm cmd".
 * Commands:
 *   vlen		Size of the vstack
 *   v index		Token on the vstack (top is index 0, second is index 1)
 *   vpush token	
 *   vpop
 *
 *   clen		Size of the cstack
 *   c index		Token on the cstack
 *   cpush token      
 *   cpop
 *
 *   llen		Size of the loop stack, i.e. nesting level
 *
 *   get name		Get name in C
 *   set name token	Set name in C to token
 *   start token	Start an iteration (same as "start" VM instruction)
 *   stop		Stop an iteration ("break")
 *   nextvn		Get next {name token} or {} if done
 *   exec cmds		Execute instructions in cmds
 */
 
#include <boa/boa.h>
#include <boa/vm.h>
#include <boa/asm.h>
#include <tcl.h>
#include <tcl_macros.h>

/* Prototypes ****************************************************************/
int BoaVMCmd(BoaVM vm, Tcl_Interp *interp,
	     int argc, char *argv[]);

/* Private */
static int AppendTok(Tcl_Interp *interp, Token token);
static int Exec(BoaVM vm, char *s);

/* Functions *****************************************************************/
int BoaVMCmd(BoaVM vm, Tcl_Interp *interp,
	     int argc, char *argv[])
{
  Token token;
  BoaCode code;
  BoaVMBlock b = vm->b-1;
  Err(argc < 2, "missing directive: must be one of: v, vlen, c, clen, exec, ...");
  if(!strcmp(argv[1], "v")) {
    int i;
    Err(argc != 3, "usage: v index");
    GetInt(argv[2], i);
    i++;
    Err((i <= 0) || (b->v-i < (b-1)->v), "index out of bounds");
    token = b->v[-i];
    TokenHold(token);
    ReturnPtr(token);
  }
  else if(!strcmp(argv[1], "vlen")) {
    ReturnInt(b->v - (b-1)->v);
  }
  else if(!strcmp(argv[1], "vpush")) {
    Err(argc != 3, "usage: vpush token");
    GetPtr(argv[2], token);
    *b->v++ = token;
  }
  else if(!strcmp(argv[1], "vpop")) {
    b->v--;
    TokenRelease(*b->v);
  }
  else if(!strcmp(argv[1], "c")) {
    int i;
    Err(argc != 3, "usage: c index");
    GetInt(argv[2], i);
    i++;
    Err((i <= 0) || (b->c-i < (b-1)->c), "index out of bounds");
    token = b->c[-i];
    TokenHold(token);
    ReturnPtr(token);
  }
  else if(!strcmp(argv[1], "clen")) {
    ReturnInt(b->c - (b-1)->c);
  }
  else if(!strcmp(argv[1], "cpush")) {
    Err(argc != 3, "usage: cpush token");
    GetPtr(argv[2], token);
    *b->c++ = token;
  }
  else if(!strcmp(argv[1], "cpop")) {
    b->c--;
    TokenRelease(*b->c);
  }
  else if(!strcmp(argv[1], "llen")) {
    ReturnInt(b->l - (b-1)->l);
  }
  else if(!strcmp(argv[1], "get")) {
    char cmd[128];
    Err(argc != 3, "usage: get name");
    sprintf(cmd, "get %s", argv[2]);
    Err(Exec(vm, cmd), TpmError);
    b->v--;
    ReturnPtr(*b->v);
  }
  else if(!strcmp(argv[1], "set")) {
    char cmd[128];
    Err(argc != 4, "usage: set name token");
    GetPtr(argv[3], token);
    *b->v++ = token;
    sprintf(cmd, "set %s", argv[2]);
    Err(Exec(vm, cmd), TpmError);
    b->v--;
    ReturnPtr(*b->v);
  }
  else if(!strcmp(argv[1], "start")) {
    Err(argc != 3, "usage: start token");
    GetPtr(argv[2], token);
    *b->v++ = token;
    Err(Exec(vm, "start"), TpmError);
  }
  else if(!strcmp(argv[1], "stop")) {
    Err(Exec(vm, "stop"), TpmError);
  }
  else if(!strcmp(argv[1], "nextvn")) {
    Token *top = b->v;
    Err(Exec(vm, "nextvn fail :fail"), TpmError);
    /* did anything get pushed on the stack? */
    if(b->v > top) {
      char str[20];
      b->v--; token = *b->v;
      Tcl_AppendElement(interp, *(char**)token->value);
      TokenRelease(token);
      b->v--; token = *b->v;
      sprintf(str, "%ld", (long)token);
      Tcl_AppendElement(interp, str);
    }
  }
  else if(!strcmp(argv[1], "exec")) {
    Err(argc != 3, "usage: exec cmds");
    Err(Exec(vm, argv[2]), TpmError);
  }
  else {
    Err(1, "unknown directive");
  }
  return TCL_OK;
}

static int AppendTok(Tcl_Interp *interp, Token token)
{
  StrStream ss = StrStreamCreate(16, STRSTREAM_UNLIMITED);
  if(BoaSPrint(token, ss) == TPM_ERROR) {
    StrStreamFree(ss); TpmPrependError("sprint failed"); Err(1, TpmError);
  }
  /* give the data in the stream to Tcl */
  Tcl_AppendElement(interp, StrStreamAsString(ss));
  StrStreamFree(ss);
  return TPM_OK;
}

static int Exec(BoaVM vm, char *s)
{
  int status;
  BoaCode code;
  BoaAsm asm = BoaAsmCreate(NULL);
  StrStream ss = StrStreamFromString(s, strlen(s), 0);
  BoaAssemble(asm, ss);
  StrStreamFree(ss);
  code = BoaAssembleDone(asm);
  status = BoaVMExec(vm, code->code);
  BoaCodeFree(code);
  return status;
}
