/* Array formatter.
 * ObjGet with a Stream will return an Obj.
 * ObjSet with a Stream will write the Obj.
 * Forwards to a formatter for the elements.
 * Attributes
 *   formatter     Element formatter
 *   separator     Obj to place between items.  Defaults to "".
 */
#include <boa/boa.h>

typedef struct Formatter {
  Obj formatter, separator;
} *Formatter;

/* Globals *******************************************************************/
Type T_FormatArray;
Obj Boa_FormatArray_DefaultFormatter;

/* Functions *****************************************************************/

static void Init(Obj self)
{
  Formatter f = (Formatter)self->value;
  Boa_Hold(Boa_FormatArray_DefaultFormatter);
  f->formatter = Boa_FormatArray_DefaultFormatter;
  f->separator = Boa_String_Create(strdup(""));
}

static void Final(Obj self)
{
  Formatter f = (Formatter)self->value;
  Boa_Release(f->formatter);
  Boa_Release(f->separator);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Formatter f = (Formatter)self->value;
  func(f->formatter, data);
  func(f->separator, data);
}

static int SPrint(Obj self, StrStream ss)
{
  return StrStreamPuts(ss, "aFormatArray");
}

static int StringSet(Obj self, char *name, int flags, Obj value)
{
  Formatter f = (Formatter)self->value;
  if(name[0] == '_') {
    if(!strcmp(name+1, "formatter")) {
      Boa_Release(f->formatter);
      f->formatter = value;
      return TPM_OK;
    }
    else if(!strcmp(name+1, "separator")) {
      Boa_Release(f->separator);
      f->separator = value;
      return TPM_OK;
    }
  }
  return TPM_BOTTOM;
}

static int ObjGet(Obj self, Obj stream, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Formatter f = (Formatter)self->value;
  int pos, i, length;
  Type subtype;
  char *data;
  Obj k, t;
  
  /* can't guess the length of the array */
  if(!type_in_out || !*type_in_out) return TPM_BOTTOM;
  if((*type_in_out)->type != T_Array)  return TPM_BOTTOM;
  
  pos = Boa_Tell(stream);
  subtype = Boa_Array_Subtype(*type_in_out);
  length = Boa_Array_Length(*type_in_out);
  /* would like to create t without initialization, but then if an error
   * occurred during reading, the array would be inconsistent.
   */
  t = Boa_Create(*type_in_out);
  data = (char*)t->value;
  k = Boa_CreateFromData(subtype, NULL);
  for(i=0;i<length;i++) {
    /* call ourselves for each element */
    /* has to copy the object each time: slow! */
    Obj elt;
    k->value = data;
    Boa_Final(k);  /* finalize the old value */
    if(Boa_ObjGet(f->formatter, stream, flags, &subtype, &elt) != TPM_OK)
      goto error;
    /* copy in the new value */
    memcpy(data, elt->value, TypeSize(subtype));
    BoaFreeAND01(elt);
    Boa_Release(elt);
    /* read the separator */
    if(Boa_Scan(stream, f->separator) != TPM_OK) goto error;
    data += TypeSize(subtype);
  }
  Boa_Release(k);
  if(value_return) *value_return = t;
  else Boa_Release(t);  /* should back up, too */
  return TPM_OK;

 error:
  Boa_Release(k);
  Boa_Release(t);
  /* reset the stream */
  Boa_Seek(stream, pos, 0);
  return TPM_BOTTOM;
}

static int ObjSet(Obj self, Obj stream, int flags, Obj value)
{
  Formatter f = (Formatter)self->value;
  int pos;
  Type subtype;
  int i, length;
  char *data;

  if(value->type->type != T_Array) return TPM_BOTTOM;

  pos = Boa_Tell(stream);
  subtype = Boa_Array_Subtype(value->type);
  length = Boa_Array_Length(value->type);
  data = (char*)value->value;
  for(i=0;i<length;i++) {
    /* call ourselves for each element */
    Obj t = Boa_CreateFromData(subtype, NULL);
    t->value = data;
    if(Boa_ObjSet(f->formatter, stream, flags, t) != TPM_OK) {
      /* reset the stream */
      Boa_Seek(stream, pos, 0);
      return TPM_BOTTOM;
    }
    assert(Boa_IntSet(stream, 1, 0, Boa_Hold(f->separator)) == TPM_OK);
    data += TypeSize(subtype);
  }
  Boa_Release(value);
  return TPM_OK;
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Err_StringGet, StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_FormatArray_Define(void)
{
  Boa_FormatArray_DefaultFormatter = Boa_Create(T_Merge);
  T_FormatArray = Boa_SimpleType_Create("formatArray", &methods,
					sizeof(struct Formatter),
					sizeof(Obj));
  return T_FormatArray;
}
