/* Machine-independent binary formatter.
 * Words are always stored little-endian.
 * ObjGet with a Stream will return an Obj.
 * ObjSet with a Stream will write the Obj.
 * Dispatches on the return type to specific read/write routines.
 * Behavior object; has no data.
 * Requires: Array, Int, Double, Char, String, SimpleType.
 */
/* todo: Float */
#include <boa/boa.h>

/* Globals *******************************************************************/
Type T_FormatBinary;

/* Functions *****************************************************************/

static int SPrint(Obj self, StrStream ss)
{
  return StrStreamPuts(ss, "aFormatBinary");
}

static int ObjGet(Obj self, Obj stream, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Obj t;
  /* can't guess the type */
  if(!type_in_out || !*type_in_out) return TPM_BOTTOM;
  if(*type_in_out == T_Char) {
    return Boa_IntGet(stream, 1, 0, type_in_out, value_return);
  }
  else if(*type_in_out == T_Int) {
    unsigned char *p;
    int i;
    /* read four characters */
    if(Boa_IntGet(stream, sizeof(int), 0, &T_String, &t) != TPM_OK) 
      return TPM_BOTTOM;
    p = (unsigned char *)Boa_String_Value(t);
#ifdef TPM_SWAP
    /* this always works, but is slower than a typecast */
    i  = (int)*p++ << 0;
    i += (int)*p++ << 8;
    i += (int)*p++ << 16;
    i += (int)*p++ << 24;
#else
    i = *(int*)p;
#endif
    Boa_Release(t);
    if(value_return) *value_return = Boa_Int_Create(i);
    /* back up otherwise */
    return TPM_OK;
  }
  else if(*type_in_out == T_Double) {
    char *p, *q;
    double d;
    /* read eight characters */
    if(Boa_IntGet(stream, sizeof(double), 0, &T_String, &t) != TPM_OK)
      return TPM_BOTTOM;
    p = Boa_String_Value(t);
#ifdef TPM_SWAP
    q = (char *)&d;
    q[0] = p[7];
    q[1] = p[6];
    q[2] = p[5];
    q[3] = p[4];
    q[4] = p[3];
    q[5] = p[2];
    q[6] = p[1];
    q[7] = p[0];
#else
    d = *(double*)p;
#endif
    Boa_Release(t);
    if(value_return) *value_return = Boa_Double_Create(d);
    /* back up otherwise */
    return TPM_OK;
  }
  /* special case array read */
  else if((*type_in_out)->type == T_Array) {
    Type subtype = Boa_Array_Subtype(*type_in_out);
    int length = Boa_Array_Length(*type_in_out);
    Obj k;
    if((subtype == T_Char) || (subtype == T_Int) || (subtype == T_Double)) {
      if(Boa_IntGet(stream, TypeSize(*type_in_out), 0, &T_String, &k) 
	 != TPM_OK) return TPM_BOTTOM;
      /* swap bytes, if necessary */
      AdjustBuffer(Boa_String_Value(k), length, TypeSize(subtype));
      /* steal the value from k */
      BoaFreeAND01(k);
      t = Boa_CreateFromData(*type_in_out, Boa_String_Value(k));
      Boa_Release(k);
      if(value_return) *value_return = t;
      else Boa_Release(t);  /* should back up, too */
      return TPM_OK;
    }
  }
  return TPM_BOTTOM;   /* don't know how to read it */
}

static int ObjSet(Obj self, Obj stream, int flags, Obj value)
{
  Obj t;
  if((value->type == T_Char) ||
     (value->type == T_String)) {
    /* all streams should support these */
    return Boa_IntSet(stream, 1, flags, value);
  }
  else if(value->type == T_Int) {
    char *data = "0123";
    /* make a four-character string */
    t = Boa_CreateFromData(T_Ptr, NULL);
    t->value = &data;
#ifdef TPM_SWAP
    /* this always works, but is slower than a typecast */
    data[0] = (char)(Boa_Int_Value(value) >> 0);
    data[1] = (char)(Boa_Int_Value(value) >> 8);
    data[2] = (char)(Boa_Int_Value(value) >> 16);
    data[3] = (char)(Boa_Int_Value(value) >> 24);
#else
    *(int*)data = Boa_Int_Value(value);
#endif
    Boa_Release(value);
    return Boa_IntSet(stream, sizeof(int), 0, t);
  }
  else if(value->type == T_Double) {
    char *p, *q = "01234567";
    /* make an eight-character string */
    t = Boa_CreateFromData(T_Ptr, NULL);
    t->value = &q;
#ifdef TPM_SWAP
    p = (char *)value->value;
    q[0] = p[7];
    q[1] = p[6];
    q[2] = p[5];
    q[3] = p[4];
    q[4] = p[3];
    q[5] = p[2];
    q[6] = p[1];
    q[7] = p[0];
#else
    *(double*)q = Boa_Double_Value(value);
#endif
    Boa_Release(value);
    return Boa_IntSet(stream, sizeof(double), 0, t);
  }
  return TPM_BOTTOM;   /* don't know how to write it */
}

static struct Boa_Methods methods = {
  Empty_Init, Empty_Final, Empty_Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Err_StringGet, Err_StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_FormatBinary_Define(void)
{
  T_FormatBinary = Boa_SimpleType_Create("formatBinary", &methods, 1, 1);
  return T_FormatBinary;
}
