/* The Boa stream formatter.  
 * ObjGet with a Stream will return an Obj.
 * ObjSet with a Stream will write the Obj.
 * File format is
 *   BOA formatter
 *   type
 *   data
 * where the formatter names an object in r->formatters.  The formatter
 * is given the same Stream, positioned at the start of the data.
 * The formatter might itself be FormatBoa, etc.
 * Stores the format to be used for certain objects in "objects".
 * Attributes:
 *   format        dictionary of object/formatter
 *   path          location of current object
 * Requires: IdDict, String, SimpleType.
 */
#include <boa/boa.h>

#define REMEMBER 0
#define VERBOSE 0

typedef struct Formatter {
  Obj formatters;  /* dictionary of formatters (for reading) */
  Obj objects;     /* object/formatter-name pairs (for writing) */
  Obj path;        /* location of the object */
} *Formatter;

/* Globals *******************************************************************/
Type T_FormatBoa;
Obj Boa_FormatBoa_DefaultFormatters;

/* Functions *****************************************************************/

static void Init(Obj self)
{
  Formatter r = (Formatter)self->value;
  Boa_Hold(Boa_Formatters);
  r->formatters = Boa_Formatters;
  r->objects = Boa_Create(T_IdDict);
}

static void Final(Obj self)
{
  Formatter r = (Formatter)self->value;
  Boa_Release(r->formatters);
  if(r->objects) Boa_Release(r->objects);
  if(r->path) Boa_Release(r->path);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Formatter r = (Formatter)self->value;
  func(r->formatters, data);
  if(r->objects) func(r->objects, data);
  if(r->path) func(r->path, data);
}

static int SPrint(Obj self, StrStream ss)
{
  return StrStreamPuts(ss, "aFormatBoa");
}

static int StringGet(Obj self, char *name, int flags,
		     Type *type_in_out, Obj *value_return)
{
  Formatter r = (Formatter)self->value;
  if(name[0] == '_') {
    if(!strcmp(name+1, "format")) {
      if(!r->objects) return TPM_BOTTOM;
      BOA_TYPECHECK_GET(r->objects->type);
      Boa_Hold(r->objects);
      *value_return = r->objects;
      return TPM_OK;
    }
  }
  return TPM_BOTTOM;
}

static int StringSet(Obj self, char *name, int flags, Obj value)
{
  Formatter r = (Formatter)self->value;
  if(name[0] == '_') {
    if(!strcmp(name+1, "path")) {
      if(value->type != T_String) return TPM_BOTTOM;
      if(r->path) Boa_Release(r->path);
      r->path = value;
      return TPM_OK;
    }
  }
  return TPM_BOTTOM;
}

static int ObjGet(Obj self, Obj stream, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Formatter r = (Formatter)self->value;
  Obj type_formatter = NULL, data_formatter = NULL;
  Type type;
  int pos;
  char *formatter_name = NULL;

  /* remember our position */
  pos = Boa_Tell(stream);

  /* read "BOA " */
  if(Boa_Getc(stream) != 'B' ||
     Boa_Getc(stream) != 'O' ||
     Boa_Getc(stream) != 'A' ||
     Boa_Getc(stream) != ' ') goto error;

  /* read the name of the formatter */
  formatter_name = Boa_GetLine(stream);
  if(!formatter_name) goto error;
  if(Boa_StringGet(r->formatters, formatter_name, 0, 
		  NULL, &data_formatter) != TPM_OK) {
    ErrorPrepend("no data formatter");
    goto error;
  }

  /* read the type */
  if(Boa_StringGet(r->formatters, "type", 0, NULL, &type_formatter) != TPM_OK) {
    ErrorPrepend("no type formatter");
    goto error;
  }
  if(Boa_ObjGet(type_formatter, stream, 0, NULL, &type) != TPM_OK) {
    ErrorPrepend("no such type");
    goto error;
  }
  Boa_Release(type_formatter); type_formatter = NULL;
  /* type mismatch? */
  if(type_in_out) {
    if(!*type_in_out) *type_in_out = type;
    else if(Boa_Compare(*type_in_out, type)) goto error;
  }

  if(r->path) {
    /* tell the data formatter the path */
#if VERBOSE
    fprintf(stderr, "FormatBoa: telling path `%s'\n", Boa_String_Value(r->path));
#endif
    if(Boa_StringSet(data_formatter, "_path", 0, r->path) != TPM_OK) {
      Boa_Release(r->path);
    }
    r->path = NULL;
  }
  /* ask the data formatter to read the rest */
  if(Boa_ObjGet(data_formatter, stream, flags, &type, value_return) != TPM_OK) {
    /* error in data */
    goto error;
  }

  /* remember the formatter we used for this object */
  if(REMEMBER && r->objects) {
    Boa_ObjSet(r->objects, *value_return, BOA_INSERT|BOA_REPLACE, 
	       Boa_String_Create(formatter_name));
  }
  else {
    free(formatter_name);
    Boa_Release(data_formatter);
  }
  return TPM_OK;

 error:
  if(formatter_name) free(formatter_name);
  if(type_formatter) Boa_Release(type_formatter);
  if(data_formatter) Boa_Release(data_formatter);
#if VERBOSE
  fprintf(stderr, "FormatBoa: seeking back to %d\n", pos);
#endif
  Boa_Seek(stream, pos, 0);
  return TPM_BOTTOM;
}

static int ObjSet(Obj self, Obj stream, int flags, Obj value)
{
  Formatter r = (Formatter)self->value;
  Obj formatter_name;
  int pos, status, i;
  char header[] = "BOA ";
  Obj type_formatter, data_formatter;

  if(Boa_StringGet(r->formatters, "type", 0, NULL, &type_formatter) != TPM_OK) {
    /* no type formatter */
    return TPM_BOTTOM;
  }

  /* remember our position */
  pos = Boa_Tell(stream);

  i = -1;
  /* have we seen this object before? */
  if(!r->objects ||
     (Boa_ObjGet(r->objects, value, 0, &T_String, &formatter_name) != TPM_OK)) {
    /* no */
    i++;
    Boa_IntGet(Boa_FormatBoa_DefaultFormatters, i, 0, 
	       &T_String, &formatter_name);
  }

  /* try each of the default formatters */
  for(;;) {
    /* write the header */
    Boa_Puts(stream, header);

    /* write the name of the formatter */
    Boa_Puts(stream, Boa_String_Value(formatter_name));

    /* write a newline */
    Boa_Putc(stream, '\n');

    /* write the type */
    Boa_Hold(value->type);  /* type_formatter will do a release */
    status = Boa_ObjSet(type_formatter, stream, 0, value->type);
    if(status != TPM_OK) { Boa_Release(value->type); goto error; }

    /* write a newline */
    Boa_Putc(stream, '\n');

    /* write the value */
    if(Boa_StringGet(r->formatters, Boa_String_Value(formatter_name), 0, 
		     NULL, &data_formatter) != TPM_OK) {
      /* no data formatter */
      ErrorSet("No such formatter `%s'\n", Boa_String_Value(formatter_name));
      goto error;
    }
    status = Boa_ObjSet(data_formatter, stream, flags, value);
    Boa_Release(data_formatter);
    if(status == TPM_OK) break;

    i++;
    if(i == Boa_Ord_Length(Boa_FormatBoa_DefaultFormatters)) goto error;
    Boa_Release(formatter_name);
    Boa_IntGet(Boa_FormatBoa_DefaultFormatters, i, 0, 
	       &T_String, &formatter_name);
    Boa_Seek(stream, pos, 0);
  }

  /* write a newline */
  if(Boa_Putc(stream, '\n') != TPM_OK) goto error;

  Boa_Release(formatter_name);
  Boa_Release(type_formatter);
  return TPM_OK;

 error:
  Boa_Release(formatter_name);
  Boa_Release(type_formatter);
  Boa_Seek(stream, pos, 0);
  return TPM_BOTTOM;
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  StringGet, StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_FormatBoa_Define(void)
{
  Boa_FormatBoa_DefaultFormatters = Boa_Create(T_Ord);
  T_FormatBoa = Boa_SimpleType_Create("formatBoa", &methods,
				      sizeof(struct Formatter),
				      sizeof(Ptr));
  return T_FormatBoa;
}
