/* Dictionary formatter.  Uses StringSet to build up a context.
 * ObjGet with a Stream will return an Obj.
 * ObjSet with a Stream will write the Obj.
 * Requires: Eval, FormatAscii, Dict, SimpleType.
 * Format is one binding per line:
 *   name type value
 * The "type" formatter is used for the type, and "ascii" for the data.
 */
#include <boa/boa.h>

typedef struct Formatter {
  Obj formatters;  /* dictionary of formatters */
} *Formatter;

/* Globals *******************************************************************/
Type T_FormatDict;

/* Functions *****************************************************************/

static void Init(Obj self)
{
  Formatter r = (Formatter)self->value;
  r->formatters = Boa_Formatters;
}

static int SPrint(Obj self, StrStream ss)
{
  return StrStreamPuts(ss, "aFormatDict");
}

static int ObjGet(Obj self, Obj stream, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Formatter r = (Formatter)self->value;
  Obj dict, name=NULL, type, value, newline;
  Obj type_formatter, data_formatter;
  int pos = Boa_Tell(stream);
  type_formatter = Boa_GetFormatter("type");
  data_formatter = Boa_GetFormatter("ascii");
  newline = Boa_Char_Create('\n');
  if(!type_in_out) dict = Boa_Create(T_Dict);
  else {
    if(!*type_in_out) *type_in_out = T_Dict;
    dict = Boa_Create(*type_in_out);
  }
  /* read left brace */
  if(Boa_StringGet(stream, "line", 0, NULL, &name) != TPM_OK) goto error;
  /* open-coded strcmp(name, "{") */
  if(Boa_String_Value(name)[0] != '{' || Boa_String_Value(name)[1]) {
    ErrorString = "missing open brace";
    goto error;
  }
  for(;;) {
    Boa_Release(name);
    /* read a line */
    if(Boa_StringGet(stream, "line", 0, NULL, &name) != TPM_OK) {
      ErrorString = "missing close brace";
      goto error;
    }
    /* is it a right brace? */
    if(Boa_String_Value(name)[0] == '}' && !Boa_String_Value(name)[1]) break;
    /* read the type */
    if(Boa_ObjGet(type_formatter, stream, 0, NULL, &type) != TPM_OK) goto error;
    /* type formatter ate the space */
    /* read the value */
    if(Boa_ObjGet(data_formatter, stream, 0, &type, &value) != TPM_OK)
      goto error;
    /* enter it into the dictionary */
    if(Boa_StringSet(dict, Boa_String_Value(name), BOA_INSERT, value) != TPM_OK) {
      Boa_Release(value);
      goto error;
    }
    /* XXX release the type? */
    /* eat the newline */
    if(Boa_Scan(stream, newline) != TPM_OK) goto error;
  }
  Boa_Release(newline);
  Boa_Release(name);
  *value_return = dict;
  return TPM_OK;

 error:
  Boa_Release(newline);
  if(name) Boa_Release(name);
  Boa_Seek(stream, pos, 0);
  return TPM_BOTTOM;
}

static int ObjSet(Obj self, Obj stream, int flags, Obj value)
{
  Formatter r = (Formatter)self->value;
  Obj iter, t;
  Obj type_formatter, data_formatter;
  char *name;
  int pos = Boa_Tell(stream);
  /* Use an iterator to get the bindings */
  if(Boa_StringGet(value, "_iter", 0, NULL, &iter) != TPM_OK) {
    /* no iterator */
    return TPM_BOTTOM;
  }
  if(Boa_StringGet(r->formatters, "type", 0, NULL, &type_formatter) != TPM_OK) {
    /* no type formatter */
    return TPM_BOTTOM;
  }
  if(Boa_StringGet(r->formatters, "ascii", 0, NULL, &data_formatter) 
     != TPM_OK) {
    /* no data formatter */
    return TPM_BOTTOM;
  }
  Boa_Puts(stream, "{\n");
  while(name = Boa_TellString(iter)) {
    /* name */
    t = Boa_String_Create(name);
    Boa_IntSet(stream, 1, 0, t);  /* frees name for us */
    Boa_Putc(stream, '\n');

    if(Boa_IntGet(iter, 1, 0, NULL, &t) != TPM_OK) break;
    /* type */
    Boa_Hold(t->type);  /* type_formatter will do a release */
    if(Boa_ObjSet(type_formatter, stream, 0, t->type) != TPM_OK) {
      Boa_Release(t->type);
      Boa_Release(t);
      goto error;
    }
    /* space */
    Boa_Putc(stream, ' ');
    /* value */
    if(Boa_ObjSet(data_formatter, stream, 0, t) != TPM_OK) {
      Boa_Release(t);
      goto error;
    }
    /* newline */
    Boa_Putc(stream, '\n');
  }
  Boa_Puts(stream, "}\n");
  Boa_Release(iter);
  Boa_Release(type_formatter);
  Boa_Release(data_formatter);
  return TPM_OK;

 error:
  Boa_Release(iter);
  Boa_Release(type_formatter);
  Boa_Release(data_formatter);
  Boa_Seek(stream, pos, 0);
  return TPM_BOTTOM;
}

static struct Boa_Methods methods = {
  Init, Sub_Final, Sub_Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Err_StringGet, Err_StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_FormatDict_Define(void)
{
  T_FormatDict = Boa_SimpleType_Create("formatDict", &methods,
				       sizeof(struct Formatter),
				       sizeof(Ptr));
  return T_FormatDict;
}
