/* Formatter for raw images, i.e. continguous pixels without a header.
 * ObjGet with a Stream will return an Image.
 * ObjSet with a Stream will write the Image.
 * Attributes:
 *   width
 *   height
 *   channels
 *   format     N, RGB, BGR
 * Requires: Image, SimpleType.
 */
#include <boa/boa.h>
#include <boa/image.h>

typedef struct Formatter {
  int width, height, channels;
  char *format;
} *Formatter;

/* Globals *******************************************************************/
Type T_FormatRaw;

/* Functions *****************************************************************/

static void Init(Obj self)
{
  Formatter f = (Formatter)self->value;
  f->format = strdup("N");
}

static int SPrint(Obj self, StrStream ss)
{
  Formatter f = (Formatter)self->value;
  StrStreamPuts(ss, "formatRaw(");
  StrStreamPutInt(ss, f->width);
  StrStreamPutc(ss, ' ');
  StrStreamPutInt(ss, f->height);
  StrStreamPutc(ss, ' ');
  StrStreamPutInt(ss, f->channels);
  StrStreamPutc(ss, ' ');
  StrStreamPuts(ss, f->format);
  return StrStreamPutc(ss, ')');
}

static int StringGet(Obj self, char *name, int flags, 
		     Type *type_in_out, Obj *value_return)
{
  Formatter f = (Formatter)self->value;
  if(name[0] == '_') {
    if(!strcmp(name+1, "width")) {
      BOA_TYPECHECK_GET(T_Int);
      if(value_return) *value_return = Boa_Int_Create(f->width);
      return TPM_OK;
    }
    else if(!strcmp(name+1, "height")) {
      BOA_TYPECHECK_GET(T_Int);
      if(value_return) *value_return = Boa_Int_Create(f->height);
      return TPM_OK;
    }
    else if(!strcmp(name+1, "channels")) {
      BOA_TYPECHECK_GET(T_Int);
      if(value_return) *value_return = Boa_Int_Create(f->channels);
      return TPM_OK;
    }
    else if(!strcmp(name+1, "format")) {
      BOA_TYPECHECK_GET(T_String);
      if(value_return) *value_return = Boa_String_Create(f->format);
      return TPM_OK;
    }
  }
  return TPM_BOTTOM;
}

static int StringSet(Obj self, char *name, int flags, Obj value)
{
  Formatter f = (Formatter)self->value;
  if(name[0] == '_') {
    if(!strcmp(name+1, "width")) {
      if(value->type != T_Int) return TPM_BOTTOM;
      f->width = Boa_Int_Value(value);
      Boa_Release(value);
      return TPM_OK;
    }
    else if(!strcmp(name+1, "height")) {
      if(value->type != T_Int) return TPM_BOTTOM;
      f->height = Boa_Int_Value(value);
      Boa_Release(value);
      return TPM_OK;
    }
    else if(!strcmp(name+1, "channels")) {
      if(value->type != T_Int) return TPM_BOTTOM;
      f->channels = Boa_Int_Value(value);
      Boa_Release(value);
      return TPM_OK;
    }
    else if(!strcmp(name+1, "format")) {
      if(value->type != T_String) return TPM_BOTTOM;
      free(f->format);
      f->format = strdup(Boa_String_Value(value));
      Boa_Release(value);
      return TPM_OK;
    }
  }
  return TPM_BOTTOM;
}

static int ObjGet(Obj self, Obj stream, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Formatter f = (Formatter)self->value;
  int len = f->width*f->height*f->channels;
  Obj t;
  TpmImage image;

  BOA_TYPECHECK_GET(T_Image);
  if(!value_return) return Boa_IntGet(stream, len, 0, &T_String, NULL);
  if(Boa_IntGet(stream, len, 0, &T_String, &t) != TPM_OK) {
    return TPM_BOTTOM;
  }
  image = TpmImageCreate(f->width, f->height, f->channels);
  if((f->channels == 3) && strcmp(f->format, "N")) {
    unsigned char *r, *g, *b, *p;
    int i, size = f->width * f->height;
    p = (unsigned char *)Boa_String_Value(t);  /* image data */
    r = *image->d3[0]; g = *image->d3[1]; b = *image->d3[2];
    if(!strcmp(f->format, "RGB")) {
      /* remember, more compact code runs faster... */
      for(i=0;i<size;i++) { *r++ = *p++; *g++ = *p++; *b++ = *p++; }
    }
    else if(!strcmp(f->format, "BGR")) {
      for(i=0;i<size;i++) { *b++ = *p++; *g++ = *p++; *r++ = *p++; }
    }
    else {
      ErrorSet("unknown format `%s'", f->format);
      Boa_Release(t);
      TpmImageFree(image);
      return TPM_BOTTOM;
    }
  }
  else {
    memcpy(image->d, Boa_String_Value(t), len);
  }
  Boa_Release(t);
  t = Boa_Create(T_Image);
  Boa_Image_Value(t) = image;
  *value_return = t;
  return TPM_OK;
}

static int ObjSet(Obj self, Obj stream, int flags, Obj value)
{
  Formatter f = (Formatter)self->value;
  int len = f->width*f->height*f->channels;
  TpmImage image;
  Obj t = Boa_Create(T_Ptr);
  if(value->type != T_Image) return TPM_BOTTOM;
  image = Boa_Image_Value(value);
  if((f->channels == 3) && strcmp(f->format, "N")) {
    unsigned char *r, *g, *b, *p;
    unsigned char *data = Allocate(len, unsigned char);
    int i, size = f->width * f->height;
    r = *image->d3[0]; g = *image->d3[1]; b = *image->d3[2];
    p = data;
    if(!strcmp(f->format, "RGB")) {
      for(i=0;i<size;i++) { *p++ = *r++; *p++ = *g++; *p++ = *b++; }
    }
    else if(!strcmp(f->format, "BGR")) {
      for(i=0;i<size;i++) { *p++ = *b++; *p++ = *g++; *p++ = *r++; }
    }
    else {
      ErrorSet("unknown format `%s'", f->format);
      Boa_Release(t);
      TpmImageFree(image);
      free(data);
      return TPM_BOTTOM;
    }
    Boa_Ptr_Value(t) = data;
    Boa_IntSet(stream, len, 0, t);
    Boa_Release(t);
    free(data);
  }
  else {
    Boa_Ptr_Value(t) = image->d;
    Boa_IntSet(stream, len, 0, t);
    Boa_Release(t);
  }
  Boa_Release(value);
  return TPM_OK;
}

static struct Boa_Methods methods = {
  Init, Empty_Final, Empty_Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  StringGet, StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_FormatRaw_Define(void)
{
  T_FormatRaw = Boa_SimpleType_Create("formatRaw", &methods, 
				      sizeof(struct Formatter), 
				      sizeof(Ptr));
  return T_FormatRaw;
}
