/* Tcl formatter.
 * ObjGet with a Stream will return an Obj.
 * ObjSet with a Stream will write the Obj.
 * Dispatches on the return type to specific read/write routines.
 * Uses a Tcl_Interp for help.
 * Requires: Array, Int, Double, Char, String, SimpleType.
 */
#include <boa/boa.h>
#include <boa/tcl.h>

typedef struct Formatter {
  Obj tcl;   /* T_Tcl */
} *Formatter;

/* Globals *******************************************************************/
Type T_FormatTcl;

/* Functions *****************************************************************/

static int SPrint(Obj self, StrStream ss)
{
  return StrStreamPuts(ss, "aFormatTcl");
}

/* Sets *value_return to an Obj of type "type", read from string.
 * Does not free string.
 * Requires: value_return != NULL 
 */
static int ConvertTcl(Tcl_Interp *interp, char *string,
		      Type type, Obj *value_return)
{
  if(type == T_String) {
    *value_return = Boa_String_Create(strdup(string));
  }
  else if(type == T_Int) {
    int i;
    /* value really must be an integer; will not truncate, 
     * even if fractional part is 0.
     */
    if(Tcl_GetInt(interp, string, &i) != TCL_OK) return TPM_BOTTOM;
    *value_return = Boa_Int_Create(i);
  }
  else if(type == T_Double) {
    double d;
    if(Tcl_GetDouble(interp, string, &d) != TCL_OK) return TPM_BOTTOM;
    *value_return = Boa_Double_Create(d);
  }
  else if(type->type == T_Array) {
    Type subtype = Boa_Array_Subtype(type);
    int i, length = Boa_Array_Length(type);
    Obj t, k;
    char *data;
    int argc;
    char **argv;
    Tcl_SplitList(interp, string, &argc, &argv);
    MEM_ALLOC_NOTIFY(argv);
    if(argc != length) {
      free(argv);
      fprintf(stderr, "FormatTcl: array length (%d) not the expected %d\n", 
	      argc, length);
      fprintf(stderr, "`%s'\n", string);
      return TPM_BOTTOM;
    }
    t = Boa_Create(type);
    data = (char*)t->value;
    k = Boa_CreateFromData(subtype, NULL);
    for(i=0;i<length;i++) {
      /* call ourselves for each element */
      Obj elt;
      k->value = data;
      Boa_Final(k);  /* finalize the old value */
      if(ConvertTcl(interp, argv[i], subtype, &elt) != TPM_OK) {
	Boa_Release(k);
	Boa_Release(t);
	free(argv);
	return TPM_BOTTOM;
      }
      /* copy in the new value */
      memcpy(data, elt->value, TypeSize(subtype));
      BoaFreeAND01(elt);
      Boa_Release(elt);
      data += TypeSize(subtype);
    }
    Boa_Release(k);
    free(argv);
    *value_return = t;
  }
  else return TPM_BOTTOM;   /* don't know how to read it */
  return TPM_OK;
}

static int ObjGet(Obj self, Obj stream, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Formatter f = (Formatter)self->value;
  StrStream ss = StrStreamCreate(1024, STRSTREAM_UNLIMITED);
  int pos = Boa_Tell(stream);
  Obj t;
  int status;
  Tcl_Interp *interp = Boa_Tcl_Interp(f->tcl);
  /* read the whole stream into a big buffer, and call Tcl on it. */
  /* this way tcl commands, including those that access or modify global 
   * state, can be embedded in the expression */
  while(Boa_StringGet(stream, "line", 0, &T_String, &t) == TPM_OK) {
    StrStreamPuts(ss, Boa_String_Value(t));
    Boa_Release(t);
    StrStreamPutc(ss, '\n');
  }
  status = Tcl_ExprString(interp, StrStreamAsString(ss));
  StrStreamFree(ss); 
  if(status != TCL_OK) { 
    fprintf(stderr, "FormatTcl: %s\n", interp->result);
    Boa_Seek(stream, pos, 0); return TPM_BOTTOM; 
  }
  if(type_in_out && !*type_in_out) *type_in_out = T_String;
  if(!value_return) {
    Boa_Seek(stream, pos, 0);
    return TPM_OK;
  }
  return ConvertTcl(interp, interp->result,
		    type_in_out?*type_in_out:T_String, value_return);
}

static int ObjSet(Obj self, Obj stream, int flags, Obj value)
{
  if((value->type == T_Int) ||
     (value->type == T_Double)) {
    /* all streams should support these */
    return Boa_IntSet(stream, 1, flags, value);
  }
  else if((value->type == T_Char) ||
	  (value->type == T_String)) {
    /* use Tcl_Merge to quote things correctly */
    char *s, data[2];
    if(value->type == T_Char) {
      /* make into a string */
      s = data;
      s[0] = Boa_Char_Value(value);
      s[1] = 0;
    }
    else s = Boa_String_Value(value);
    Boa_Release(value);
    s = Tcl_Merge(1, &s);
    MEM_ALLOC_NOTIFY(s);
    Boa_IntSet(stream, 1, flags, Boa_String_Create(s));
  }
  /* same as FormatAscii, except for surrounding braces */
  else if(value->type->type == T_Array) {
    Type subtype = Boa_Array_Subtype(value->type);
    int i, length = Boa_Array_Length(value->type);
    char *data = (char*)value->value;
    int pos = Boa_Tell(stream);
    /* write a Tcl list */
    Boa_Putc(stream, '{');
    for(i=0;i<length;i++) {
      /* call ourselves for each element */
      Obj t = Boa_CreateFromData(subtype, NULL);
      t->value = data;
      if(ObjSet(self, stream, flags, t) != TPM_OK) {
	/* reset the stream */
	Boa_Seek(stream, pos, 0);
	return TPM_BOTTOM;
      }
      Boa_Putc(stream, ' ');
      data += TypeSize(subtype);
    }
    Boa_Release(value);
    Boa_Putc(stream, '}');
  }
  /* dictionaries could be done as name/value pairs */
  else return TPM_BOTTOM;   /* don't know how to write it */
  return TPM_OK;
}

static struct Boa_Methods methods = {
  Empty_Init, Sub_Final, Sub_Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Err_StringGet, Err_StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_FormatTcl_Define(void)
{
  T_FormatTcl = Boa_SimpleType_Create("formatTcl", &methods, 
				      sizeof(struct Formatter), sizeof(Ptr));
  return T_FormatTcl;
}
