/* Tcl_Eval formatter.
 * ObjGet with a Stream will return an Obj.
 * ObjSet with a Stream will write the Obj.
 * Attributes:
 *   path	location of current object
 *   interp     Tcl interpreter
 */
#include <boa/boa.h>
#include <boa/tcl.h>
#include <tpm/list.h>

typedef struct Formatter {
  Obj tcl;    /* T_Tcl */
  List paths; /* List of Objs */
} *Formatter;

/* Globals *******************************************************************/
Type T_FormatTclEval;

/* Functions *****************************************************************/

static void Init(Obj self)
{
  Formatter f = (Formatter)self->value;
  /* what about f->tcl? */
  f->paths = ListCreate((FreeFunc*)Boa_Free);
}

static void Final(Obj self)
{
  Formatter f = (Formatter)self->value;
  Boa_Release(f->tcl);
  ListFree(f->paths);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Formatter f = (Formatter)self->value;
  Obj t;
  func(f->tcl, data);
  {ListIter(p, t, f->paths) {
    if(t) func(t, data);
  }}
}

static int SPrint(Obj self, StrStream ss)
{
  return StrStreamPuts(ss, "aFormatTclEval");
}

static int ObjGet(Obj self, Obj stream, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Formatter f = (Formatter)self->value;
  StrStream ss = StrStreamCreate(1024, STRSTREAM_UNLIMITED);
  int pos = Boa_Tell(stream);
  Obj t, dir;
  int status;
  Tcl_Interp *interp = Boa_Tcl_Interp(f->tcl);
  /* read the whole stream into a big buffer, and call Tcl on it. */
  /* this way tcl commands, including those that access or modify global 
   * state, can be embedded in the expression */
  /* however, this must occur inside of a local scope,
   * to catch temporary variables. "block" is defined in util.tcl. */
  /* should convert boa globals into tcl globals in a more general way. */
  StrStreamPuts(ss, "block {global dir boaDir dlopen dlsym atoi;\n");
  while(Boa_StringGet(stream, "line", 0, &T_String, &t) == TPM_OK) {
    StrStreamPuts(ss, Boa_String_Value(t));
    Boa_Release(t);
    StrStreamPutc(ss, '\n');
  }
  StrStreamPuts(ss, "}"); /* close the block */
  MEM_VERIFY_ALL();
  status = Tcl_Eval(interp, StrStreamAsString(ss));
  StrStreamFree(ss); 

  /* restore the old _path of boaDir and the old value of dir */
  dir = BoaTclStringGet(f->tcl, "boaDir");
  assert(dir);
  ListRemoveFront(f->paths, (void**)&t);
  if(t) {
    assert(Boa_StringSet(f->tcl, "dir", BOA_INSERT|BOA_REPLACE, 
			 Boa_Hold(t)) == TPM_OK);
    assert(Boa_StringSet(dir, "_path", 0, t) == TPM_OK);
  }
  
  if(status != TCL_OK) { 
    ErrorSet("%s", interp->result); 
    ErrorPrint(stderr);
    Boa_Seek(stream, pos, 0); 
    return TPM_ERROR;
  }
  if(!value_return) {
    Boa_Seek(stream, pos, 0);
    return TPM_OK;
  }
  t = BoaTclToObj(interp, interp->result);
  if(!t) return TPM_ERROR;
  BOA_TYPECHECK_GET(t->type);
  *value_return = t;
  return TPM_OK;
}

static int ObjSet(Obj self, Obj stream, int flags, Obj value)
{
  Obj type_formatter, tcl_formatter;
  /* forward to the regular Tcl formatter */
  TRY_RETURN(Boa_StringGet(Boa_Formatters, "type", 0, NULL, &type_formatter));
  TRY_RETURN(Boa_StringGet(Boa_Formatters, "tcl", 0, NULL, &tcl_formatter));
  Boa_Puts(stream, "boa parse tcl [boa parse type 0 {");
  Boa_Hold(value->type);
  /* the type better have matched braces */
  TRY_RETURN(Boa_ObjSet(type_formatter, stream, 0, value->type));
  Boa_Release(type_formatter);
  Boa_Puts(stream, "}] {");
  TRY_RETURN(Boa_ObjSet(tcl_formatter, stream, 0, value));
  Boa_Release(tcl_formatter);
  Boa_Putc(stream, '}');
  return TPM_OK;
}

static int StringGet(Obj self, char *name, int flags, 
		     Type *type_in_out, Obj *value_return)
{
  Formatter f = (Formatter)self->value;
  if(name[0] == '_') {
    if(!strcmp(name+1, "interp")) {
      BOA_TYPECHECK_GET(T_Tcl);
      if(value_return) {
	Boa_Hold(f->tcl);
	*value_return = f->tcl;
      }
      return TPM_OK;
    }
  }
  return TPM_BOTTOM;
}

static int StringSet(Obj self, char *name, int flags, Obj value)
{
  Formatter f = (Formatter)self->value;
  if(name[0] == '_') {
    if(!strcmp(name+1, "path")) {
      /* _path is the name of the file we're reading */
      Obj s,t;
      int status;
      char *last;
      /* move the "boaDir" objDir in the Tcl interpreter */
      t = BoaTclStringGet(f->tcl, "boaDir");
      if(!t) {
	/* doesn't exist; create it */
	t = Boa_Create(T_ObjDir);
	Boa_StringSet(f->tcl, "boaDir", BOA_INSERT, BoaObjToTcl(t));
      }
      /* extract the directory component of the path */
      last = strrchr(Boa_String_Value(value), '/');
      assert(last);
      /* XXX unsafe mutation */
      *last = 0;
#if 0
      fprintf(stderr, "setting path to %s\n", Boa_String_Value(value));
#endif
      Boa_StringSet(f->tcl, "dir", BOA_INSERT|BOA_REPLACE, Boa_Hold(value));
      /* save the current _path */
      assert(Boa_StringGet(t, "_path", 0, &T_String, &s) == TPM_OK);
      ListAddFront(f->paths, s);
      assert(Boa_StringSet(t, "_path", BOA_REPLACE, value) == TPM_OK);
      return TPM_OK;
    }
  }
  return TPM_BOTTOM;
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  StringGet, StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_FormatTclEval_Define(void)
{
  T_FormatTclEval = Boa_SimpleType_Create("formatTclEval", &methods, 
					  sizeof(struct Formatter), 
					  sizeof(Ptr));
  return T_FormatTclEval;
}
