/* Formatter for vector/matrix/volume.
 * ObjGet with a Stream will return an Obj.
 * ObjSet with a Stream will write the Obj.
 * Requires: SimpleType.
 */
#include <boa/boa.h>
#include <boa/vector.h>

/* Globals *******************************************************************/
Type T_FormatVector;

/* Functions *****************************************************************/

static int SPrint(Obj self, StrStream ss)
{
  return StrStreamPuts(ss, "aFormatVector");
}

/* Vector is ascii len, whitespace, binary data.
 * Matrix is ascii rows, whitespace, cols, whitespace, binary data.
 * Volume is zdim, whitespace, ydim, whitespace, xdim, whitespace, binary data.
 */
static int ObjGet(Obj self, Obj stream, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Obj t;
  int pos = Boa_Tell(stream);
  /* if don't know the type, assume it's a vector */
  if(!type_in_out || !*type_in_out || (*type_in_out == T_Vector)) {
    Vector result;
    int len;
    if(Boa_IntGetInt(stream, 1, &len) != TPM_OK) return TPM_BOTTOM;
    Boa_Getc(stream); /* discard whitespace */
    if(Boa_IntGet(stream, len*sizeof(double), 0, &T_String, &t) != TPM_OK) 
      goto error;
    BOA_TYPECHECK_GET(T_Vector);
    if(value_return) {
      /* steal the data from the string */
      result = VectorFromData(len, (double*)Boa_String_Value(t));
      BoaFreeAND01(t);
      Boa_Release(t);
      AdjustBuffer(result->d, result->len, sizeof(double));
      *value_return = Boa_Vector_Create(result);
    }
    return TPM_OK;
  }
  else if(*type_in_out == T_Matrix) {
    Matrix result;
    int rows, cols;
    if(Boa_IntGetInt(stream, 1, &rows) != TPM_OK) return TPM_BOTTOM;
    if(Boa_IntGetInt(stream, 1, &cols) != TPM_OK) goto error;
    Boa_Getc(stream); /* discard whitespace */
    if(Boa_IntGet(stream, rows*cols*sizeof(double), 0, &T_String, &t) != TPM_OK)
      goto error;
    if(value_return) {
      /* steal the data from the string */
      result = MatrixFromData(rows, cols, (double*)Boa_String_Value(t));
      BoaFreeAND01(t);
      Boa_Release(t);
      AdjustBuffer(result->d, result->len, sizeof(double));
      *value_return = Boa_Matrix_Create(result);
    }
    return TPM_OK;
  }
  else if(*type_in_out == T_Volume) {
    Volume result;
    int zdim, ydim, xdim;
    if(Boa_IntGetInt(stream, 1, &zdim) != TPM_OK) return TPM_BOTTOM;
    if(Boa_IntGetInt(stream, 1, &ydim) != TPM_OK) goto error;
    if(Boa_IntGetInt(stream, 1, &xdim) != TPM_OK) goto error;
    Boa_Getc(stream); /* discard whitespace */
    if(Boa_IntGet(stream, zdim*ydim*xdim*sizeof(double), 0, &T_String, &t) != TPM_OK)
      goto error;
    if(value_return) {
      /* steal the data from the string */
      result = VolumeFromData(zdim, ydim, xdim, (double*)Boa_String_Value(t));
      BoaFreeAND01(t);
      Boa_Release(t);
      AdjustBuffer(result->d, result->len, sizeof(double));
      *value_return = Boa_Volume_Create(result);
    }
    return TPM_OK;
  }
  return TPM_BOTTOM;
 error:
  Boa_Seek(stream, pos, 0);
  return TPM_BOTTOM;
}

static int ObjSet(Obj self, Obj stream, int flags, Obj value)
{
  int status;
  Vector v = Boa_Vector_Value(value);
  if(value->type == T_Vector) {
    Boa_IntSet(stream, 1, 0, Boa_Int_Create(v->len));
    Boa_Putc(stream, '\n');
  }
  else if(value->type == T_Matrix) {
    Matrix m = Boa_Matrix_Value(value);
    Boa_IntSet(stream, 1, 0, Boa_Int_Create(m->rows));
    Boa_Putc(stream, ' ');
    Boa_IntSet(stream, 1, 0, Boa_Int_Create(m->cols));
    Boa_Putc(stream, '\n');
  }
  else if(value->type == T_Volume) {
    Volume vol = Boa_Volume_Value(value);
    Boa_IntSet(stream, 1, 0, Boa_Int_Create(vol->zdim));
    Boa_Putc(stream, ' ');
    Boa_IntSet(stream, 1, 0, Boa_Int_Create(vol->ydim));
    Boa_Putc(stream, ' ');
    Boa_IntSet(stream, 1, 0, Boa_Int_Create(vol->xdim));
    Boa_Putc(stream, '\n');
  }
  else return TPM_BOTTOM;
  AdjustBuffer(v->d, v->len, sizeof(double));
  status = Boa_IntSet(stream, v->len*sizeof(double), 0, Boa_Ptr_Create(v->d));
  AdjustBuffer(v->d, v->len, sizeof(double));
  if(status != TPM_OK) return TPM_BOTTOM;
  Boa_Release(value);
  return TPM_OK;
}

static struct Boa_Methods methods = {
  Empty_Init, Empty_Final, Empty_Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Err_StringGet, Err_StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_FormatVector_Define(void)
{
  T_FormatVector = Boa_SimpleType_Create("formatVector", &methods, 1, 1);
  return T_FormatVector;
}
