/* Formatter for vector/matrix/volume.
 * ObjGet with a Stream will return an Obj.
 * ObjSet with a Stream will write the Obj.
 * Requires: SimpleType.
 */
#include <boa/boa.h>
#include <boa/vector.h>

/* Globals *******************************************************************/
Type T_FormatVectorAscii;

/* Functions *****************************************************************/

static int SPrint(Obj self, StrStream ss)
{
  return StrStreamPuts(ss, "aFormatVectorAscii");
}

static int ReadVector(Vector v, Obj stream)
{
  int i;
  for(i=0;i<v->len;i++) {
    Obj t;
    if(Boa_IntGet(stream, 1, 0, &T_Double, &t) != TPM_OK) return TPM_ERROR;
    v->d[i] = Boa_Double_Value(t);
    Boa_Release(t);
  }
  return TPM_OK;
}

/* Vector is ascii len, whitespace, ascii data.
 * Matrix is ascii rows, whitespace, cols, whitespace, ascii data.
 * Volume is zdim, whitespace, ydim, whitespace, xdim, whitespace, ascii data.
 */
static int ObjGet(Obj self, Obj stream, int flags,
		  Type *type_in_out, Obj *value_return)
{
  int pos = Boa_Tell(stream);
  /* if don't know the type, assume it's a vector */
  if(!type_in_out || !*type_in_out || (*type_in_out == T_Vector)) {
    Vector result;
    int len;
    if(Boa_IntGetInt(stream, 1, &len) != TPM_OK) return TPM_BOTTOM;
    Boa_Getc(stream); /* discard whitespace */
    result = VectorCreate(len);
    if(ReadVector(result, stream) != TPM_OK) goto error;
    BOA_TYPECHECK_GET(T_Vector);
    if(value_return) {
      *value_return = Boa_Vector_Create(result);
    }
    else VectorFree(result);
    return TPM_OK;
  }
  else if(*type_in_out == T_Matrix) {
    Matrix result;
    int rows, cols;
    if(Boa_IntGetInt(stream, 1, &rows) != TPM_OK) return TPM_BOTTOM;
    if(Boa_IntGetInt(stream, 1, &cols) != TPM_OK) goto error;
    Boa_Getc(stream); /* discard whitespace */
    result = MatrixCreate(rows, cols);
    if(ReadVector((Vector)result, stream) != TPM_OK) goto error;
    if(value_return) {
      *value_return = Boa_Matrix_Create(result);
    }
    else MatrixFree(result);
    return TPM_OK;
  }
  else if(*type_in_out == T_Volume) {
    Volume result;
    int zdim, ydim, xdim;
    if(Boa_IntGetInt(stream, 1, &zdim) != TPM_OK) return TPM_BOTTOM;
    if(Boa_IntGetInt(stream, 1, &ydim) != TPM_OK) goto error;
    if(Boa_IntGetInt(stream, 1, &xdim) != TPM_OK) goto error;
    Boa_Getc(stream); /* discard whitespace */
    result = VolumeCreate(zdim, ydim, xdim);
    if(ReadVector((Vector)result, stream) != TPM_OK) goto error;
    if(value_return) {
      *value_return = Boa_Volume_Create(result);
    }
    else VolumeFree(result);
    return TPM_OK;
  }
  return TPM_BOTTOM;
 error:
  Boa_Seek(stream, pos, 0);
  return TPM_BOTTOM;
}

static int WriteVector(Vector v, Obj stream)
{
  int i;
  Obj t = Boa_Create(T_Double);
  for(i=0;i<v->len;i++) {
    Boa_Double_Value(t) = v->d[i];
    Boa_Hold(t);
    if(Boa_IntSet(stream, 1, 0, t) != TPM_OK) return TPM_ERROR;
    Boa_Putc(stream, ' ');
  }
  Boa_Release(t);
  return TPM_OK;
}

static int ObjSet(Obj self, Obj stream, int flags, Obj value)
{
  Vector v = Boa_Vector_Value(value);
  if(value->type == T_Vector) {
    Boa_IntSet(stream, 1, 0, Boa_Int_Create(v->len));
    Boa_Putc(stream, '\n');
  }
  else if(value->type == T_Matrix) {
    Matrix m = Boa_Matrix_Value(value);
    Boa_IntSet(stream, 1, 0, Boa_Int_Create(m->rows));
    Boa_Putc(stream, ' ');
    Boa_IntSet(stream, 1, 0, Boa_Int_Create(m->cols));
    Boa_Putc(stream, '\n');
  }
  else if(value->type == T_Volume) {
    Volume vol = Boa_Volume_Value(value);
    Boa_IntSet(stream, 1, 0, Boa_Int_Create(vol->zdim));
    Boa_Putc(stream, ' ');
    Boa_IntSet(stream, 1, 0, Boa_Int_Create(vol->ydim));
    Boa_Putc(stream, ' ');
    Boa_IntSet(stream, 1, 0, Boa_Int_Create(vol->xdim));
    Boa_Putc(stream, '\n');
  }
  else return TPM_BOTTOM;
  if(WriteVector(v, stream) != TPM_OK) return TPM_BOTTOM;
  Boa_Release(value);
  return TPM_OK;
}

static struct Boa_Methods methods = {
  Empty_Init, Empty_Final, Empty_Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Err_StringGet, Err_StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_FormatVectorAscii_Define(void)
{
  T_FormatVectorAscii = Boa_SimpleType_Create("formatVectorAscii", &methods, 1, 1);
  return T_FormatVectorAscii;
}
