/* Formatter for vector/matrix/volume without a header.
 * Attributes:
 *   xdim
 *   ydim     If specified, will read matrices.  Defaults to 1.
 *   zdim     If specified, will read volumes.  Defaults to 1.
 */
#include <boa/boa.h>
#include <boa/vector.h>

typedef struct Formatter {
  int xdim, ydim, zdim;
} *Formatter;

/* Globals *******************************************************************/
Type T_FormatVectorRaw;

/* Functions *****************************************************************/

static int SPrint(Obj self, StrStream ss)
{
  Formatter f = (Formatter)self->value;
  StrStreamPuts(ss, "formatVectorRaw(");
  StrStreamPutInt(ss, f->xdim);
  StrStreamPutc(ss, ' ');
  StrStreamPutInt(ss, f->ydim);
  StrStreamPutc(ss, ' ');
  StrStreamPutInt(ss, f->zdim);
  return StrStreamPutc(ss, ')');
}

static int StringSet(Obj self, char *name, int flags, Obj value)
{
  Formatter f = (Formatter)self->value;
  if(name[0] == '_') {
    if(!strcmp(name+1, "xdim")) {
      if(value->type != T_Int) return TPM_BOTTOM;
      f->xdim = Boa_Int_Value(value);
      Boa_Release(value);
      return TPM_OK;
    }
    else if(!strcmp(name+1, "ydim")) {
      if(value->type != T_Int) return TPM_BOTTOM;
      f->ydim = Boa_Int_Value(value);
      Boa_Release(value);
      return TPM_OK;
    }
    else if(!strcmp(name+1, "zdim")) {
      if(value->type != T_Int) return TPM_BOTTOM;
      f->zdim = Boa_Int_Value(value);
      Boa_Release(value);
      return TPM_OK;
    }
  }
  return TPM_BOTTOM;
}

static int ObjGet(Obj self, Obj stream, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Formatter f = (Formatter)self->value;
  int len;
  Type type;
  double *data;
  Obj t;

  /* check type */
  if(!type_in_out || !*type_in_out) {
    type = f->ydim ? f->zdim ? T_Volume : T_Matrix : T_Vector;
    if(type_in_out) *type_in_out = type;
  }
  else {
    type = *type_in_out;
    if((type != T_Vector) && (type != T_Matrix) && (type != T_Volume))
      return TPM_BOTTOM;
  }

  /* how many bytes to read? */
  len = f->xdim;
  if((type == T_Matrix) && f->ydim) len *= f->ydim;
  else if((type == T_Volume) && f->zdim) len *= f->zdim;

  if(!value_return) return Boa_IntGet(stream, len*sizeof(double), 0, &T_String, NULL);
  /* read the data into a string */
  if(Boa_IntGet(stream, len*sizeof(double), 0, &T_String, &t) != TPM_OK) {
    return TPM_BOTTOM;
  }
  data = (double*)Boa_String_Value(t);
  /* assume we are the only holder of t */
  assert(t->ref_count == 1);
  BoaFreeAND01(t);
  Boa_Release(t);
  AdjustBuffer(data, len, sizeof(double));
  if(type == T_Vector) {
    *value_return = Boa_Vector_Create(VectorFromData(f->xdim, data));
  }
  else if(type == T_Matrix) {
    *value_return = Boa_Matrix_Create(MatrixFromData(f->ydim?f->ydim:1, f->xdim, data));
  }
  else {
    *value_return = Boa_Volume_Create(VolumeFromData(f->zdim?f->zdim:1, f->ydim?f->ydim:1, f->xdim, data));
  }
  return TPM_OK;
}

static int ObjSet(Obj self, Obj stream, int flags, Obj value)
{
  int status;
  Vector v = Boa_Vector_Value(value);
  AdjustBuffer(v->d, v->len, sizeof(double));
  status = Boa_IntSet(stream, v->len*sizeof(double), 0, Boa_Ptr_Create(v->d));
  AdjustBuffer(v->d, v->len, sizeof(double));
  return status;
}

static struct Boa_Methods methods = {
  Empty_Init, Empty_Final, Empty_Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Err_StringGet, StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_FormatVectorRaw_Define(void)
{
  T_FormatVectorRaw = Boa_SimpleType_Create("formatVectorRaw", &methods, 
					    sizeof(struct Formatter), 
					    sizeof(Ptr));
  return T_FormatVectorRaw;
}
