/* array(n, t) returns an array type which holds exactly n elements of type t.
 * Attributes
 *   length
 */
#include <boa/boa.h>

/* Globals *******************************************************************/
Type T_Array;

/* Functions *****************************************************************/

#define length Boa_Int_Value(Boa_PType_Item(self->type, 0))
#define subtype Boa_PType_Item(self->type, 1)

static void Init(Obj self)
{
  char *data = self->value;
  int i;
  Obj t;
  if(TypeMethods(subtype)->init == Empty_Init) return;
  /* t is a fake object which does not own its data. 
   * we will use this trick repeatedly.
   */   
  t = Boa_CreateFromData(subtype, NULL);
  for(i=0;i<length;i++) {
    t->value = data;
    Boa_Init(t);
    data += TypeSize(subtype);
  }
  Boa_Release(t);
}

static void Final(Obj self)
{
  char *data = self->value;
  int i;
  Obj t;
  if(TypeMethods(subtype)->final == Empty_Final) return;
  t = Boa_CreateFromData(subtype, NULL);
  for(i=0;i<length;i++) {
    t->value = data;
    Boa_Final(t);
    data += TypeSize(subtype);
  }
  Boa_Release(t);
}

static void Clean(Obj self, ObjFunc *func, Any d)
{
  char *data = self->value;
  int i;
  Obj t;
  if(TypeMethods(subtype)->clean == Empty_Clean) return;
  t = Boa_CreateFromData(subtype, NULL);
  for(i=0;i<length;i++) {
    t->value = data;
    func(t, d);
    data += TypeSize(subtype);
  }
  Boa_Release(t);
}

static int Copy(Obj self, Ptr value_return)
{
  char *data = self->value;
  int i;
  Obj t;
  if(TypeMethods(subtype)->copy == Generic_Copy) {
    return Generic_Copy(self, value_return);
  }
  t = Boa_CreateFromData(subtype, NULL);
  for(i=0;i<length;i++) {
    t->value = data;
    /* XXX leak */
    TRY_RETURN(Boa_Copy(t, value_return));
    data += TypeSize(subtype);
    value_return = (char*)value_return + TypeSize(subtype);
  }
  Boa_Release(t);
  return TPM_OK;
}

static int Compare(Obj self, Obj b)
{
  char *ap = self->value, *bp = b->value;
  int i, result = 0;
  Obj at, bt;
  if(TypeMethods(subtype)->compare == Generic_Compare) {
    return Generic_Compare(self, b);
  }
  at = Boa_CreateFromData(subtype, NULL);
  bt = Boa_CreateFromData(subtype, NULL);
  for(i=0;i<length;i++) {
    at->value = ap;
    bt->value = bp;
    if(result |= Boa_Compare(at, bt)) break;
    ap += TypeSize(subtype);
    bp += TypeSize(subtype);
  }
  Boa_Release(at);
  Boa_Release(bt);
  return result;
}

static int Hash(Obj self)
{
  char *data = self->value;
  int i, result = 1;
  Obj t;
  if(TypeMethods(subtype)->hash == Generic_Hash) {
    return Generic_Hash(self);
  }
  t = Boa_CreateFromData(subtype, NULL);
  for(i=0;i<length;i++) {
    t->value = data;
    result += Boa_Hash(t);
  }
  Boa_Release(t);
  return result * 1103515245;
}

/* prints like a Tcl list */
static int SPrint(Obj self, StrStream ss)
{
  char *data = self->value;
  int i;
  Obj t = Boa_CreateFromData(subtype, NULL);
  StrStreamPutc(ss, '{');
  for(i=0;i<length;i++) {
    t->value = data;
    Boa_SPrint(t, ss);
    if(i < length-1) StrStreamPutc(ss, ' ');
    data += TypeSize(subtype);
  }
  Boa_Release(t);
  StrStreamPutc(ss, '}');
  return TPM_OK;
}

static int IntGet(Obj self, int index, int flags, 
		  Type *type_in_out, Obj *value_return)
{
  char *data = self->value;
  if((index < 0) || (index >= length)) return TPM_BOTTOM;
  /* length is immutable */
  if(flags & BOA_DELETE) { ErrorString = BOA_EFLAGS; return TPM_ERROR; }
  BOA_TYPECHECK_GET(subtype);
  if(value_return) {
    Obj t = Boa_CreateFromData(subtype, NULL);
    t->value = data + TypeSize(subtype)*index;
    /* Return a copy, so that changes to the array do not show thru. */
    *value_return = BoaCopy(t);
    Boa_Release(t);
    if(!*value_return) return TPM_BOTTOM;
  }
  return TPM_OK;
}
/* converts name to index and forwards to IntGet */
static int StringGet(Obj self, char *name, int flags, 
		     Type *type_in_out, Obj *value_return)
{
  int index;
  if(name[0] == '_') {
    if(!strcmp(name+1, "length")) {
      return Boa_IntGet(self->type, 0, flags, type_in_out, value_return);
    }
  }
  index = atoi(name);
  if(!index && (name[0] != '0')) return TPM_BOTTOM;
  return IntGet(self, index, flags, type_in_out, value_return);
}

static int IntSet(Obj self, int index, int flags, Obj value)
{
  char *data = self->value;
  int status;
  if((index < 0) || (index >= length)) return TPM_BOTTOM;
  /* must do a replace */
  if(!(flags & BOA_REPLACE)) { ErrorString = BOA_EFLAGS; return TPM_ERROR; }
  if(Boa_Compare(subtype, value->type)) return TPM_BOTTOM;
  data += TypeSize(subtype)*index;
  status = Boa_Copy(value, data);
  Boa_Release(value);
  return status;
}
/* converts name to index and forwards to IntSet */
static int StringSet(Obj self, char *name, int flags, Obj value)
{
  int index = atoi(name);
  if(!index && (name[0] != '0')) return TPM_BOTTOM;
  return IntSet(self, index, flags, value);
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  SPrint, Copy, Compare, Hash,
  StringGet, StringSet,
  IntGet, IntSet,
  Generic_ObjGet, Generic_ObjSet
};

/* assumes num_args and args have already been filled in */
static void Type_Init(Obj self)
{
  Boa_ParamTypeInstance data = (Boa_ParamTypeInstance)self->value;
  data->methods = &methods;
  data->size = Boa_Int_Value(data->args[0]) * TypeSize(data->args[1]);
  data->align = TypeAlign(data->args[1]);
}

Obj Boa_Array_Define(void)
{
  T_Array = Boa_ParamType_Create("array", Type_Init, 2, T_Int, NULL);
  return T_Array;
}
