/* Caches a context. Cache size can be limited to a certain number of entries.
 * Attributes:
 *   context    context being cached
 *   cache      cache buffer (must be ordered)
 *   itemlimit
 *   depth      number of leading components to cache
 *   nitems     current number of cached items (read-only)
 * Requires: Int, SimpleType.
 */
/* todo: richer cache context, learning access patterns */
#include <boa/boa.h>
#include <tpm/dlist.h>

#define VERBOSE 0
#define SHOW_CACHE 0

typedef struct Cache {
  Obj context, cache;
  DList queue;   /* holds (type, index) pairs */
  unsigned item_limit, total_items, depth;
} *Cache;

/* Globals *******************************************************************/
Type T_Cache;

/* Cache-specific Functions **************************************************/

/* overloading would help here */
static int AnyGet(int kind, Obj self, Any index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  switch(kind) {
  case 0:
    return Boa_StringGet(self, (char*)index, flags, 
			 type_in_out, value_return);
  case 1:
    return Boa_IntGet(self, (int)index, flags, 
		      type_in_out, value_return);
  default:
    return Boa_ObjGet(self, (Obj)index, flags, 
		      type_in_out, value_return);
  }
}

static int AnySet(int kind, Obj self, Any index, int flags, Obj value)
{
  switch(kind) {
  case 0:
    return Boa_StringSet(self, (char*)index, flags, value);
  case 1:
    return Boa_IntSet(self, (int)index, flags, value);
  default:
    return Boa_ObjSet(self, (Obj)index, flags, value);
  }
}

typedef struct QueueData {
  int kind;
  Any index;
} *QueueData;

static QueueData QueueDataCreate(int kind, Any index)
{
  QueueData data = Allocate(1, struct QueueData);
  data->kind = kind;
  switch(kind) {
  case 0:
    data->index = strdup((char*)index);
    break;
  case 1:
    data->index = index;
    break;
  case 2:
    Boa_Hold((Obj)index);
    data->index = index;
  }
  return data;
}

static void QueueDataFree(QueueData data)
{
  switch(data->kind) {
  case 0:
    free((char*)data->index);
    break;
  case 2:
    Boa_Release((Obj)data->index);
    break;
  }
  free(data);
}

static void IndexPrint(FILE *fp, int kind, Any index)
{
  switch(kind) {
  case 0:
    fprintf(fp, "%s", (char*)index);
    break;
  case 1:
    fprintf(fp, "%d", (int)index);
    break;
  case 2:
    Boa_Print(fp, (Obj)index);
  }
}

/* Removes the least recently used item in the cache.
 * This should be the last node in the queue.
 */
static void Uncache(Cache cache)
{
  QueueData data;
  if(DListSize(cache->queue) == 0) return;
  DListRemove(cache->queue, DListFront(cache->queue), (void**)&data);
  if(AnyGet(data->kind, cache->cache, data->index, BOA_DELETE, NULL, NULL)
     != TPM_OK) {
    /* shout and hope someone listens */
    fprintf(stderr, "Uncaching error on ");
    IndexPrint(stderr, data->kind, data->index);
    fprintf(stderr, "\n");
  }
  cache->total_items--;
}

/* Removes items from the cache so that it meets its limits */
static void Shrink(Cache cache)
{
  while(cache->total_items > cache->item_limit) Uncache(cache);
}

/* Remove all entries from the cache */
static void Flush(Cache cache)
{
  int limit = cache->item_limit;
  cache->item_limit = 0;
  Shrink(cache);
  cache->item_limit = limit;
}

static void Insert(int kind, Cache cache, Any index, Obj value)
{
  /* make room */
  cache->total_items++;
  Shrink(cache);
  if(AnySet(kind, cache->cache, index, BOA_INSERT|BOA_REPLACE, value) 
     != TPM_OK) {
#if VERBOSE
    fprintf(stderr, "Caching error on ");
    IndexPrint(stderr, kind, index);
    fprintf(stderr, "\n");
#endif
    cache->total_items--;
    Boa_Release(value);
  }
  /* queue it */
  DListAddFront(cache->queue, QueueDataCreate(kind, index));
}

/* Functions *****************************************************************/

/* don't initialize cache->context since many callers will set it with
 * Boa_Cache_Context.
 */
static void Init(Obj self)
{
  Cache cache = (Cache)self->value;
  cache->cache = Boa_Create(T_Dict);
  cache->queue = DListCreate((FreeFunc*)QueueDataFree);
  cache->item_limit = (unsigned)-1;
}

/* these functions cannot assume cache->context exists */

static void Final(Obj self)
{
  Cache cache = (Cache)self->value;
  Boa_Release(cache->cache);
  DListFree(cache->queue);
  if(cache->context) Boa_Release(cache->context);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Cache cache = (Cache)self->value;
  func(cache->cache, data);
  if(cache->context) func(cache->context, data);
}

static int SPrint(Obj self, StrStream ss)
{
  Cache cache = (Cache)self->value;
#if SHOW_CACHE
  StrStreamPuts(ss, "cache(");
#endif
  if(cache->context) { TRY_RETURN(Boa_SPrint(cache->context, ss)); }
  else StrStreamPuts(ss, "NULL");
#if SHOW_CACHE
  StrStreamPuts(ss, ")");
#endif
  return TPM_OK;
}

/* kind is 0 for string, 1 for int, 2 for obj */
static int DoGet(int kind, Cache cache, Any index, int flags,
		 Type *type_in_out, Obj *value_return)
{
  int status;
  /* do we have it in the cache? */
  status = AnyGet(kind, cache->cache, index, flags, type_in_out, value_return);
  if(status == TPM_OK) {
#if VERBOSE
    fprintf(stderr, "Cache hit on ");
    IndexPrint(stderr, kind, index);
    fprintf(stderr, "\n");
#endif
    if(flags & BOA_DELETE) {
      /* must remove from context as well */
      AnyGet(kind, cache->context, index, flags, type_in_out, NULL);
    }
    else {
      /* move to the front of the queue */
    }
    return TPM_OK;
  }

#if VERBOSE
  fprintf(stderr, "Cache miss on ");
  IndexPrint(stderr, kind, index);
  fprintf(stderr, "\n");
#endif
  /* get from the context */
  if(!cache->context) return TPM_BOTTOM;
  if(flags & BOA_DELETE) {
    return AnyGet(kind, cache->context, index, flags, 
		  type_in_out, value_return);
  }
  else {
    Obj t;
    status = AnyGet(kind, cache->context, index, flags, type_in_out, &t);
    if(status != TPM_OK) return status;
    /* save the result */
    if(cache->depth > 1) {
      /* make a cache of that item */
      Obj c = Boa_Create(T_Cache);
      Cache new = (Cache)c->value;
      new->item_limit = cache->item_limit;
      new->depth = cache->depth-1;
      new->context = t;
      t = c;
    }
    Boa_Hold(t);
    Insert(kind, cache, index, t);
    if(value_return) *value_return = t;
    else Boa_Release(t);
    return TPM_OK;
  }
}

static int StringGet(Obj self, char *name, int flags,
		     Type *type_in_out, Obj *value_return)
{
  Cache cache = (Cache)self->value;
  if(cache->depth && (flags & BOA_SPLIT))
    return Generic_SplitGet(self, name, flags, type_in_out, value_return);
  if(*name == '_') {  /* attribute */
    if(!strcmp(name+1, "itemlimit")) {
      BOA_TYPECHECK_GET(T_Int);
      if(value_return) *value_return = Boa_Int_Create(cache->item_limit);
      return TPM_OK;
    }
    else if(!strcmp(name+1, "nitems")) {
      BOA_TYPECHECK_GET(T_Int);
      if(value_return) *value_return = Boa_Int_Create(cache->total_items);
      return TPM_OK;
    }
    else if(!strcmp(name+1, "depth")) {
      BOA_TYPECHECK_GET(T_Int);
      if(value_return) *value_return = Boa_Int_Create(cache->depth);
      return TPM_OK;
    }
    else if(!strcmp(name+1, "context")) {
      if(!cache->context) return TPM_BOTTOM;
      BOA_TYPECHECK_GET(cache->context->type);
      if(value_return) {
	Boa_Hold(cache->context);
	*value_return = cache->context;
      }
      return TPM_OK;
    }
    else if(!strcmp(name+1, "cache")) {
      BOA_TYPECHECK_GET(cache->cache->type);
      if(value_return) {
	Boa_Hold(cache->cache);
	*value_return = cache->cache;
      }
      return TPM_OK;
    }
    /* fall through if no matching attribute */
  }
  return DoGet(0, cache, name, flags, type_in_out, value_return);
}
static int IntGet(Obj self, int index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Cache cache = (Cache)self->value;
  return DoGet(1, cache, (Any)index, flags, type_in_out, value_return);
}
static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Cache cache = (Cache)self->value;
  return DoGet(2, cache, (Any)index, flags, type_in_out, value_return);
}

/* Write-through to the original context, retaining a reference.
 */
static int DoSet(int kind, Cache cache, Any index, int flags, Obj value)
{
  int status;
  if(!cache->context) return TPM_BOTTOM;
  Boa_Hold(value);  /* hold in case we can save it */
  status = AnySet(kind, cache->context, index, flags, value);
  if(status == TPM_OK) Insert(kind, cache, index, value);
  else Boa_Release(value);
  return status;
}

static int StringSet(Obj self, char *name, int flags, Obj value)
{
  Cache cache = (Cache)self->value;

  if(flags & BOA_SPLIT)
    return Generic_SplitSet(self, name, flags, value);

  /* is it an attribute? */
  if(*name == '_') {
    if(!strcmp(name+1, "itemlimit")) {
      if(value->type != T_Int) return TPM_BOTTOM;
      cache->item_limit = Boa_Int_Value(value);
      Boa_Release(value);
      /* clean up to meet the new limit */
      Shrink(cache);
      return TPM_OK;
    }
    else if(!strcmp(name+1, "depth")) {
      if(value->type != T_Int) return TPM_BOTTOM;
      cache->depth = Boa_Int_Value(value);
      Boa_Release(value);
      return TPM_OK;
    }
    else if(!strcmp(name+1, "context")) {
      if(cache->context) { Flush(cache); Boa_Release(cache->context); }
      cache->context = value;
      return TPM_OK;
    }
    else if(!strcmp(name+1, "cache")) {
      Boa_Release(cache->cache);
      cache->cache = value;
      return TPM_OK;
    }
    /* fall through */
  }
  return DoSet(0, cache, name, flags, value);
}
static int IntSet(Obj self, int index, int flags, Obj value)
{
  Cache cache = (Cache)self->value;
  return DoSet(1, cache, (Any)index, flags, value);
}
static int ObjSet(Obj self, Obj index, int flags, Obj value)
{
  Cache cache = (Cache)self->value;
  return DoSet(2, cache, (Any)index, flags, value);
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  StringGet, StringSet,
  IntGet, IntSet,
  ObjGet, ObjSet
};

Obj Boa_Cache_Define(void)
{
  T_Cache = Boa_SimpleType_Create("cache", &methods,
				  sizeof(struct Cache),
				  sizeof(Ptr));
  return T_Cache;
}

Obj Boa_Cache_Create(Obj context)
{
  Obj t = Boa_Create(T_Cache);
  Boa_Cache_Context(t) = context;
  return t;
}
