/* Holds a C function which mutates a context argument.
 * Attributes:
 *   func	C function (ContextFunc).
 *   context	Context holding argument values.
 *   args	Index of positional argument names.
 * Requires: SimpleType.
 */
/* todo: extract currying into its own context */
#include <boa/boa.h>

typedef struct ContextProc {
  Obj context;
  ContextFunc *func;
  Obj args;
} *ContextProc;

/* Globals *******************************************************************/
Type T_ContextProc;

/* Functions *****************************************************************/

static int DefaultFunc(Obj context, Type *type_in_out, Obj *value_return)
{
  return TPM_BOTTOM;
}

static void Init(Obj self)
{
  ContextProc cp = (ContextProc)self->value;
  cp->context = Boa_Create(T_Dict);
  cp->func = DefaultFunc;
  cp->args = Boa_Create(T_Index);
}

static void Final(Obj self)
{
  ContextProc cp = (ContextProc)self->value;
  Boa_Release(cp->context);
  Boa_Release(cp->args);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  ContextProc cp = (ContextProc)self->value;
  func(cp->context, data);
  func(cp->args, data);
}

static int SPrint(Obj self, StrStream ss)
{
  ContextProc cp = (ContextProc)self->value;
  StrStreamPuts(ss, "contextProc(");
  TRY_RETURN(Boa_SPrint(cp->args, ss));
  StrStreamPutc(ss, ' ');
  TRY_RETURN(Boa_SPrint(cp->context, ss));
  return StrStreamPutc(ss, ')');
}

static int Copy(Obj self, Ptr value_return)
{
  ContextProc cp = (ContextProc)self->value;
  ContextProc new = (ContextProc)value_return;
  new->context = Boa_Hold(cp->context);
  new->func = cp->func;
  new->args = Boa_Hold(cp->args);
  return TPM_OK;
}

#if 0
static int Call(Obj self, int nargs, int nrets, Obj *args)
{
  ContextProc cp = (ContextProc)self->value;
  Obj t;
  /* must have no args or >1 rets, for now */
  if(nargs || (nrets > 1)) return TPM_BOTTOM;
  return cp->func(cp->context, NULL, args);
}
#endif

static int Bind(Obj self, Obj value, int flags,
		Type *type_in_out, Obj *value_return)
{
  ContextProc cp = (ContextProc)self->value;
  int i, arg, unbound = 0;
  Obj new_context;
  /* find an unbound argument */
  for(i=0;i<Boa_Index_Length(cp->args);i++) {
    if(Boa_ObjGet(cp->context, Boa_Index_Item(cp->args, i), 0, NULL, NULL) 
       != TPM_OK) {
      if(!unbound) {
	arg = i;
#if 0
	fprintf(stderr, "binding ");
	Boa_Print(stderr, Boa_Index_Item(cp->args, arg));
	fprintf(stderr, " to ");
	Boa_PrintNL(stderr, value);
#endif
	new_context = BoaCopy(cp->context);
	if(!new_context ||
	   (Boa_ObjSet(new_context, Boa_Index_Item(cp->args, arg), 
		       BOA_INSERT, value) != TPM_OK)) {
	  char *s = Boa_AsString(Boa_Index_Item(cp->args, arg));
	  ErrorPrepend("cannot bind argument %s", s);
	  free(s);
	  Boa_Release(value);
	  return TPM_BOTTOM;
	}
      }
      unbound++;
    }
  }
  if(!unbound) { 
    /* kludge: return the item */
    if(value->type == T_String) {
      int status = Boa_StringGet(cp->context, Boa_String_Value(value), 
				 flags, type_in_out, value_return);
      Boa_Release(value);
      return status;
    }
    else {
      Boa_Release(value); 
      return TPM_BOTTOM; 
    }
  }
  if(unbound == 1) {
    Obj t;
    /* call the function */
    int status = cp->func(new_context, type_in_out, &t);
    Boa_Release(new_context);
    if(status != TPM_OK) return status;
    if(value_return) *value_return = t;
    return TPM_OK;
  }
  /* return a curried copy of ourself */
  BOA_TYPECHECK_GET(T_ContextProc);
  if(value_return) {
    ContextProc new = (ContextProc)Allocate(TypeSize(T_ContextProc), char);
    new->context = new_context;
    new->func = cp->func;
    new->args = Boa_Hold(cp->args);
    *value_return = Boa_CreateFromData(T_ContextProc, new);
  }
  return TPM_OK;
}

static int StringGet(Obj self, char *name, int flags,
		     Type *type_in_out, Obj *value_return)
{
  ContextProc cp = (ContextProc)self->value;
  if(flags & BOA_SPLIT) 
    return Generic_SplitGet(self, name, flags, type_in_out, value_return);
  if(name[0] == '_') {
    if(!strcmp(name+1, "func")) {
      BOA_TYPECHECK_GET(T_Ptr);
      if(value_return) *value_return = Boa_Ptr_Create((Ptr)cp->func);
      return TPM_OK;
    }
    else if(!strcmp(name+1, "context")) {
      BOA_TYPECHECK_GET(cp->context->type);
      if(value_return) *value_return = Boa_Hold(cp->context);
      return TPM_OK;
    }
    else if(!strcmp(name+1, "args")) {
      BOA_TYPECHECK_GET(cp->args->type);
      if(value_return) *value_return = Boa_Hold(cp->args);
      return TPM_OK;
    }
  }
  /* bind a positional argument to a string */
  return Bind(self, Boa_String_Create(strdup(name)), flags, 
	      type_in_out, value_return);
}
/* sets forward to the subcontext (call must be explicit) */
static int StringSet(Obj self, char *name, int flags, Obj value)
{
  ContextProc cp = (ContextProc)self->value;
  if(flags & BOA_SPLIT)
    return Generic_SplitSet(self, name, flags, value);
  if(name[0] == '_') {
    if(!strcmp(name+1, "func")) {
      if(value->type != T_Ptr) return TPM_BOTTOM;
      cp->func = (ContextFunc*)Boa_Ptr_Value(value);
      Boa_Release(value);
      return TPM_OK;
    }
    else if(!strcmp(name+1, "context")) {
      if(cp->context) Boa_Release(cp->context);
      cp->context = value;
      return TPM_OK;
    }
    else if(!strcmp(name+1, "args")) {
      if(cp->args) Boa_Release(cp->args);
      cp->args = value;
      return TPM_OK;
    }
  }
  return Boa_StringSet(cp->context, name, flags, value);
}
static int IntGet(Obj self, int index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  return Bind(self, Boa_Int_Create(index), flags, type_in_out, value_return);
}
static int IntSet(Obj self, int index, int flags, Obj value)
{
  ContextProc cp = (ContextProc)self->value;
  return Boa_IntSet(cp->context, index, flags, value);
}
static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  return Bind(self, Boa_Hold(index), flags, type_in_out, value_return);
}
static int ObjSet(Obj self, Obj index, int flags, Obj value)
{
  ContextProc cp = (ContextProc)self->value;
  return Boa_ObjSet(cp->context, index, flags, value);
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  SPrint, Copy, Generic_Compare, Generic_Hash,
  StringGet, StringSet,
  IntGet, IntSet,
  ObjGet, ObjSet
};

Obj Boa_ContextProc_Define(void)
{
  T_ContextProc = Boa_SimpleType_Create("contextProc", &methods,
					sizeof(struct ContextProc),
					sizeof(Ptr));
  return T_ContextProc;
}
