/* Simple context with no inheritance. Implemented as a GHashTable of Objs.
 * Requires: String, SimpleType.
 */

#include <boa/boa.h>
#include <tpm/hash.h>

typedef HashTable Dict;

/* Globals *******************************************************************/
Type T_Dict;

/* Functions *****************************************************************/

static void Init(Obj self)
{
  Dict dict = (Dict)self->value;
  HashTable new = HashTableCreate((HashFunc*)BoaHash, (CmpFunc*)BoaCompare,
				  (FreeFunc*)Boa_Free, (FreeFunc*)Boa_Free);
  *dict = *new;
  free(new);
}

static int Copy(Obj self, Ptr value_return)
{
  Dict dict = (Dict)self->value;
  /* must add holds to the objects */
  HashTable new = HashTableCopy(dict, 
				(CopyFunc*)BoaHold, (CopyFunc*)BoaHold);
  *(Dict)value_return = *new;
  free(new);
  return TPM_OK;
}

static int Compare(Obj self, Obj b)
{
  return HashTableCompare((HashTable)self->value, (HashTable)b->value,
			   (CmpFunc*)BoaCompare);
}

static int Hash(Obj self)
{
  Dict dict = (Dict)self->value;
  int result=1;
  Obj t;
  HashIter(b, e, t, dict) {
    result += Boa_Hash(t);
  }
  return result * 1103515245;
}

static void Final(Obj self)
{
  Dict dict = (Dict)self->value;
  /* taken from HashTableFree */
  HashTableFlush(dict);
  free(dict->buckets);
  free(dict->occupancy);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Obj t;
  HashIter(b, e, t, (Dict)self->value) {
    func(t, data);
  }
}

static int SPrint(Obj self, StrStream ss)
{
  Dict dict = (Dict)self->value;
  Obj t;
  int space = 0;
  StrStreamPutc(ss, '{');
  {HashIter(b, e, t, dict) {
    if(space) StrStreamPuts(ss, ", ");
    TRY_RETURN(Boa_SPrint((Obj)e->key, ss));
    StrStreamPuts(ss, ": ");
    TRY_RETURN(Boa_SPrint(t, ss));
    space = 1;
  }}
  return StrStreamPutc(ss, '}');
}

static int IndexObj(Obj self, Obj key)
{
  Dict dict = (Dict)self->value;
  HashEntry *e = HashTableFindEntry(dict, key);
  if(!e) return NAI;
  return HashTableIndexEntry(dict, e);
}
static Obj IndexInt(Obj self, int index)
{
  Dict dict = (Dict)self->value;
  HashEntry *e = HashTableEntryAtIndex(dict, index);
  if(!e) return NULL;
  return Boa_Hold((Obj)e->key);
}

static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Dict dict = (Dict)self->value;
  Obj t;
  HashEntry *e;

  if(index->type == T_String) {
    char *name = Boa_String_Value(index);
    if(name[0] == '_') {
      if(!strcmp(name+1, "length")) {
	BOA_TYPECHECK_GET(T_Int);
	if(value_return) *value_return = Boa_Int_Create(dict->num_entries);
	return TPM_OK;
      }
#if 0
      else if(!strcmp(name+1, "iter")) {
	if(value_return) {
	  Obj iter = Boa_Create(T_Iter);
	  Obj indexer = Boa_Create(T_Indexer);
	  Boa_Indexer_Data(indexer) = Boa_Hold(self);
	  Boa_Indexer_ObjFunc(indexer) = IndexObj;
	  Boa_Indexer_IntFunc(indexer) = IndexInt;
	  Boa_Iter_Data(iter) = Boa_Hold(self);
	  Boa_Iter_Indexer(iter) = indexer;
	  *value_return = iter;
	}
	return TPM_OK;
      }
#endif
      /* should also have _index */
    }
  }

  e = HashTableFindEntry(dict, index);
  if(!e) return TPM_BOTTOM;

  t = e->value;
  BOA_TYPECHECK_GET(t->type);
  if(flags & BOA_DELETE) {
    if(value_return) {
      /* remove without freeing */
      HashTableRemoveNoFree(dict, e);
      *value_return = t;
    }
    else HashTableRemoveEntry(dict, e);
  }
  else if(value_return) *value_return = Boa_Hold(t);
  return TPM_OK;
}

static int ObjSet(Obj self, Obj index, int flags, Obj value)
{
  Dict dict = (Dict)self->value;
  HashEntry *e;
  int found;

  if(flags & BOA_INSERT) {
    e = HashTableAddEntry(dict, index, &found);
    if(!found) assert(e->key = Boa_Stabilize(index));
  }
  else {
    found = !!(e = HashTableFindEntry(dict, index));
    if(!found) return TPM_BOTTOM;
  }
  if(found) {
    if(!(flags & BOA_REPLACE)) { ErrorString = BOA_EFLAGS; return TPM_ERROR; }
    Boa_Release((Obj)e->value);
  }
  e->value = value;
  return TPM_OK;
}

/* really slow */
static int IntGet(Obj self, int index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  int status;
  Obj t = IndexInt(self, index);
  if(!t) return TPM_BOTTOM;
  status = ObjGet(self, t, flags, type_in_out, value_return);
  Boa_Release(t);
  return status;
}
static int IntSet(Obj self, int index, int flags, Obj value)
{
  Dict dict = (Dict)self->value;
  int status;
  Obj t;
  if(index == dict->num_entries) {
    /* act like an index and insert it */
    t = Boa_Int_Create(index);
    status = Boa_ObjSet(self, t, flags, value);
    Boa_Release(t);
    return status;
  }
  t = IndexInt(self, index);
  if(!t) return TPM_BOTTOM;
  status = ObjSet(self, t, flags, value);
  Boa_Release(t);
  return status;
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  SPrint, Copy, Compare, Hash,
  Generic_StringGet, Generic_StringSet,
  IntGet, IntSet,
  ObjGet, ObjSet
};

/* Exported functions ********************************************************/
/* This function will be called when this type is loaded. */
Obj Boa_Dict_Define(void)
{
  T_Dict = Boa_SimpleType_Create("dict", &methods,
				 sizeof(struct HashTable), /* size */
				 sizeof(Ptr));	       /* align */
  return T_Dict;
}
