/* Like an Array, but stored in a stream.
 * Currently read-only.
 * Uses the binary formatter.
 * Attributes:
 *   file        file name of the stream, if any
 *   formatter   defaults to FormatBinary
 *   itemsize    size of an element on disk (each must be same size)
 *               defaults to TypeSize(subtype)
 * Requires: FormatBinary, Int, ParamType.
 */
#include <boa/boa.h>

typedef struct DiskArray {
  Obj stream, /* could be NULL if a file open failed */
      formatter;
  int item_size; /* size of an element on disk */
  int pos;    /* cached here to avoid fseeks */
} *DiskArray;

/* Globals *******************************************************************/
Type T_DiskArray;
Obj Boa_DiskArray_DefaultFormatter;

/* Functions *****************************************************************/

#define length Boa_Int_Value(Boa_PType_Item(self->type, 0))
#define subtype Boa_PType_Item(self->type, 1)

static void Init(Obj self)
{
  DiskArray da = (DiskArray)self->value;
  Boa_Hold(Boa_DiskArray_DefaultFormatter);
  da->formatter = Boa_DiskArray_DefaultFormatter;
  da->item_size = TypeSize(subtype);
}

static void Final(Obj self)
{
  DiskArray da = (DiskArray)self->value;
  if(da->stream) Boa_Release(da->stream);
  Boa_Release(da->formatter);
}

static void Clean(Obj self, ObjFunc *func, Any d)
{
  DiskArray da = (DiskArray)self->value;
  if(da->stream) func(da->stream, d);
  func(da->formatter, d);
}

static int SPrint(Obj self, StrStream ss)
{
  DiskArray da = (DiskArray)self->value;
  StrStreamPuts(ss, "diskArray{");
  if(da->stream) { TRY_RETURN(Boa_SPrint(da->stream, ss)); }
  else StrStreamPuts(ss, "NULL");
  StrStreamPutc(ss, ' ');
  Boa_SPrint(da->formatter, ss);
  return StrStreamPutc(ss, '}');
}

static int Copy(Obj self, Ptr value_return)
{
  DiskArray da = (DiskArray)self->value;
  DiskArray new = (DiskArray)value_return;
  if(da->stream) Boa_Hold(da->stream);
  new->stream = da->stream;
  Boa_Hold(da->formatter);
  new->formatter = da->formatter;
  new->item_size = da->item_size;
  new->pos = da->pos;
  return TPM_OK;
}

static int IntGet(Obj self, int index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  DiskArray da = (DiskArray)self->value;
  int pos;

  if(!da->stream) return TPM_BOTTOM;
  if((index < 0) || (index >= length)) return TPM_BOTTOM;
  BOA_TYPECHECK_GET(subtype);

  /* read data from the file, avoiding fseeks. */
  pos = index * da->item_size;
  if(da->pos != pos) {
    if(Boa_Seek(da->stream, pos - da->pos, 1) != TPM_OK) {
      ErrorSet("DiskArray: bad file");
      return TPM_ERROR;
    }
  }
  if(Boa_ObjGet(da->formatter, da->stream, 0, &subtype, value_return) != TPM_OK)
    return TPM_BOTTOM;
  da->pos = pos + da->item_size;
  return TPM_OK;
}

static int IntSet(Obj self, int index, int flags, Obj value)
{
  DiskArray da = (DiskArray)self->value;
  int pos;

  if(!da->stream) { ErrorSet("DiskArray: bad file"); return TPM_BOTTOM; }
  if((index < 0) || (index > length)) { 
    ErrorSet("DiskArray: index out of bounds"); return TPM_BOTTOM; 
  }
  if(index == length) {
    if(!(flags & BOA_INSERT)) { ErrorString = BOA_EFLAGS; return TPM_BOTTOM; }
  }
  else if(!(flags & BOA_REPLACE)) { ErrorString = BOA_EFLAGS; return TPM_BOTTOM; }
  if(Boa_Compare(value->type, subtype)) {
    ErrorSet("DiskArray: type mismatch"); return TPM_BOTTOM;
  }

  /* write data to the file, avoiding fseeks. */
  pos = index * da->item_size;
  if(da->pos != pos) {
    if(Boa_Seek(da->stream, pos - da->pos, 1) != TPM_OK) {
      ErrorSet("DiskArray: bad file");
      return TPM_ERROR;
    }
  }
  if(Boa_ObjSet(da->formatter, da->stream, 0, value) != TPM_OK)
    return TPM_BOTTOM;
  da->pos = pos + da->item_size;
  return TPM_OK;
}

/* converts name to index and forwards to IntGet */
static int StringGet(Obj self, char *name, int flags, 
		     Type *type_in_out, Obj *value_return)
{
  DiskArray da = (DiskArray)self->value;
  int index;
  if(name[0] == '_') {
    if(!strcmp(name+1, "file")) {
      return TPM_BOTTOM;  /* only know the stream, not the name */
    }
    else if(!strcmp(name+1, "formatter")) {
      if(value_return) { 
	Boa_Hold(da->formatter); *value_return = da->formatter;
      }
      return TPM_OK;
    }
    else if(!strcmp(name+1, "itemsize")) {
      BOA_TYPECHECK_GET(T_Int);
      if(value_return) *value_return = Boa_Int_Create(da->item_size);
      return TPM_OK;
    }
  }

  index = atoi(name);
  if(!index && (name[0] != '0')) return TPM_BOTTOM;
  return IntGet(self, index, flags, type_in_out, value_return);
}
static int StringSet(Obj self, char *name, int flags, Obj value)
{
  DiskArray da = (DiskArray)self->value;
  int index;
  if(name[0] == '_') {
    /* attribute */
    if(!strcmp(name+1, "file")) {
      if(value->type != T_String) return TPM_BOTTOM;
      if(da->stream) Boa_Release(da->stream);
      da->stream = FileHandle_Open(Boa_String_Value(value), "r+");
      if(!da->stream) /* try read-only */
	da->stream = FileHandle_Open(Boa_String_Value(value), "r");
      if(!da->stream) /* try create */
	da->stream = FileHandle_Open(Boa_String_Value(value), "w+");
      Boa_Release(value);
      if(!da->stream) return TPM_ERROR;
      da->pos = 0;
      return TPM_OK;
    }
    else if(!strcmp(name+1, "formatter")) {
      Boa_Release(da->formatter);
      da->formatter = value;
      return TPM_OK;
    }
    else if(!strcmp(name+1, "itemsize")) {
      if(value->type != T_Int) return TPM_BOTTOM;
      da->item_size = Boa_Int_Value(value);
      return TPM_OK;
    }
  }

  index = atoi(name);
  if(!index && (name[0] != '0')) return TPM_BOTTOM;
  return IntSet(self, index, flags, value);
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  SPrint, Copy, Generic_Compare, Generic_Hash,
  StringGet, StringSet,
  IntGet, IntSet,
  Generic_ObjGet, Generic_ObjSet
};

/* assumes num_args and args have already been filled in */
static void Type_Init(Obj self)
{
  Boa_ParamTypeInstance data = (Boa_ParamTypeInstance)self->value;
  data->methods = &methods;
  data->size = sizeof(struct DiskArray);
  data->align = sizeof(Ptr);
}

Obj Boa_DiskArray_Define(void)
{
  Boa_DiskArray_DefaultFormatter = Boa_Create(T_Merge);
  T_DiskArray = Boa_ParamType_Create("diskArray", Type_Init, 2, T_Int, NULL);
  return T_DiskArray;
}
