/* Dynamic linking contexts.
 * DLOpen loads a library, given an abbreviated name.
 * DLSym gets a symbol from the current program.
 * If the name has a leading component, it is assumed to be the 
 * abbreviated name of a library to load if the symbol isn't found.
 * E.g. "RGB2Ohta" fails if the symbol doesn't exist, but
 * "imgops/RGB2Ohta" will load libimgops.sl/so and try again.
 * Requires: Ptr, SimpleType.
 */
#include <boa/boa.h>
#include <tpm/dyn.h>

/* no data */

/* Globals *******************************************************************/
Type T_DLOpen, T_DLSym;

/* DLOpen ********************************************************************/

static int Open_SPrint(Obj self, StrStream ss)
{
  return StrStreamPuts(ss, "aDLOpen");
}

static int Open_StringGet(Obj self, char *name, int flags,
			  Type *type_in_out, Obj *value_return)
{
  void *handle;
  char lib[128];
  BOA_TYPECHECK_GET(T_Ptr);
#ifdef hpux
  sprintf(lib, "lib%s.sl", name);
#else
  sprintf(lib, "lib%s.so", name);
#endif
  /* load library */
  handle = DynLoad(lib);
  if(!handle) {
    ErrorSet("Cannot load library `%s'", lib);
    ErrorAppend(DynError());
    ErrorPrint(stderr);
    return TPM_BOTTOM;
  }
  if(value_return) *value_return = Boa_Ptr_Create(handle);
  return TPM_OK;
}

static struct Boa_Methods open_methods = {
  Empty_Init, Empty_Final, Empty_Clean,
  Open_SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Open_StringGet, Err_StringSet,
  Err_IntGet, Err_IntSet,
  Generic_ObjGet, Generic_ObjSet
};

Obj Boa_DLOpen_Define(void)
{
  T_DLOpen = Boa_SimpleType_Create("dlopen", &open_methods, 1, 1);
  return T_DLOpen;
}

/* DLSym *********************************************************************/

static int Sym_SPrint(Obj self, StrStream ss)
{
  return StrStreamPuts(ss, "aDLSym");
}

/* modifies name temporarily */
static int Sym_StringGet(Obj self, char *name, int flags,
			 Type *type_in_out, Obj *value_return)
{
  Ptr ptr;
  /* does the name have a leading component? */
  if(flags & BOA_SPLIT) {
    char *p;
    for(p=name;*p;p++) {
      if(*p == BOA_SEP(flags)) {
	ptr = DynSymbol(NULL, p+1);
	if(!ptr) {
	  char lib[128];
	  int status;
	  *p = '\0';
#if 1
#ifdef hpux
	  sprintf(lib, "lib%s.sl", name);
#else
	  sprintf(lib, "lib%s.so", name);
#endif
	  ptr = DynSymbol(lib, p+1);
	  *p = BOA_SEP(flags);
#else
	  status = Open_StringGet(NULL, name, 0, NULL, NULL);
	  *p = BOA_SEP(flags);
	  if(status == TPM_OK) {
	    /* try again */
	    ptr = DynSymbol(NULL, p+1);
	  }
#endif
	}
	if(ptr) {
	  BOA_TYPECHECK_GET(T_Ptr);
	  if(value_return) *value_return = Boa_Ptr_Create(ptr);
	  return TPM_OK;
	}
	else {
	  ErrorSet("Cannot find symbol `%s'", p+1);
	  ErrorAppend(DynError());
	  ErrorPrint(stderr);
	  return TPM_BOTTOM;
	}
      }
    }
  }
  ptr = DynSymbol(NULL, name);
  if(!ptr) {
    ErrorSet("Cannot find symbol `%s'", name);
    ErrorAppend(DynError());
    ErrorPrint(stderr);
    return TPM_BOTTOM;
  }
  BOA_TYPECHECK_GET(T_Ptr);
  if(value_return) *value_return = Boa_Ptr_Create(ptr);
  return TPM_OK;
}

static struct Boa_Methods sym_methods = {
  Empty_Init, Empty_Final, Empty_Clean,
  Sym_SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Sym_StringGet, Err_StringSet,
  Err_IntGet, Err_IntSet,
  Generic_ObjGet, Generic_ObjSet
};

Obj Boa_DLSym_Define(void)
{
  T_DLSym = Boa_SimpleType_Create("dlsym", &sym_methods, 1, 1);
  return T_DLSym;
}
