/* Adapts a FILE* to the Boa Stream protocol.
 * Requires: Ptr, Float, Double, Int, String, Char, SimpleType.
 */
#include <boa/boa.h>
#include <ctype.h>

typedef struct File {
  FILE *handle;
} *File;

typedef struct Seek {
  Obj file;
} *Seek;

/* Globals *******************************************************************/
Type T_FileHandle;
static Type T_FileSeek;

/* Protptypes ****************************************************************/
Obj FileHandle_Open(char *name, char *mode);

/* Functions *****************************************************************/

Obj FileHandle_Open(char *name, char *mode)
{
  Obj t;
  FILE *fp;
  /* check for stdin/stdout */
  if(name[0] == '-' && !name[1]) {
    if(mode[0] == 'r') fp = stdin;
    else fp = stdout;
  }
  else {
    fp = fopen(name, mode);
    if(!fp) { ErrorSetErrno(); return NULL; }
  }
  t = Boa_Create(T_FileHandle);
  Boa_FileHandle_Value(t) = fp;
  return t;
}

static void Final(Obj self)
{
  File f = (File)self->value;
  if(f->handle != stdout) fclose(f->handle);
}

static int SPrint(Obj self, StrStream ss)
{
  return StrStreamPuts(ss, "aFileHandle");
}

static int FileSeek(FILE *fp, int pos, int whence)
{
  if(fseek(fp, pos, whence)) {
    ErrorSetErrno(); return TPM_ERROR;
  }
  return TPM_OK;
}

/* If type is NULL, returns a character.
 * Otherwise can return:
 *   T_String, where index determines the length, up to EOF.
 *   T_Int, T_Float, T_Double, where index 0 means probe. 
 *                             index assumed 1 otherwise.
 */
static int IntGet(Obj self, int index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  File f = (File)self->value;
  if(type_in_out) {
    if(*type_in_out == T_String) {
      size_t pos = ftell(f->handle);
      if(value_return) {
	char *data = Allocate(index+1, char);
	if(fread(data, 1, index, f->handle) != index) {
	  free(data); fseek(f->handle, pos, 0); return TPM_BOTTOM;
	}
	data[index] = 0;
	*value_return = Boa_String_Create(data);
      }
      else if((pos + index >= FileSize(f->handle)) ||
	      FileSeek(f->handle, index, SEEK_CUR)) {
	return TPM_BOTTOM;
      }
      return TPM_OK;
    }
    else if(*type_in_out == T_Int) {
      int i;
      int pos;
      if(index == 0) pos = ftell(f->handle);
      if(fscanf(f->handle, "%d", &i) != 1) return TPM_BOTTOM;
      if(index == 0) fseek(f->handle, pos, 0);
      if(value_return) *value_return = Boa_Int_Create(i);
      return TPM_OK;
    }
    else if(*type_in_out == T_Float) {
      float d;
      int pos;
      if(index == 0) pos = ftell(f->handle);
      if(fscanf(f->handle, "%g", &d) != 1) return TPM_BOTTOM;
      if(index == 0) fseek(f->handle, pos, 0);
      if(value_return) *value_return = Boa_Float_Create(d);
      return TPM_OK;
    }
    else if(*type_in_out == T_Double) {
      double d;
      int pos;
      if(index == 0) pos = ftell(f->handle);
      if(fscanf(f->handle, "%lg", &d) != 1) return TPM_BOTTOM;
      if(index == 0) fseek(f->handle, pos, 0);
      if(value_return) *value_return = Boa_Double_Create(d);
      return TPM_OK;
    }
    else if(*type_in_out != T_Char) return TPM_BOTTOM;  /* can't read that */
  }
  /* reading T_Char */
  BOA_TYPECHECK_GET(T_Char);
  if(value_return) {
    char c = getc(f->handle);
    if(index != 1) {
      if(FileSeek(f->handle, index-1, SEEK_CUR)) return TPM_BOTTOM;
    }
    *value_return = Boa_Char_Create(c);
  }
  else if(index) {
    if(FileSeek(f->handle, index, SEEK_CUR)) return TPM_BOTTOM;
  }
  else {
    /* test if the file is ready to read */
  }
  return TPM_OK;
}

/* If type is Ptr, outputs "index" characters.
 * Otherwise assumes index is 1.
 * Can output: ptr, char, string, int, double, float
 */
static int IntSet(Obj self, int index, int flags, Obj value)
{
  File f = (File)self->value;
  if(value->type == T_Char) {
    if(putc(Boa_Char_Value(value), f->handle) == EOF) {
      ErrorSetErrno(); return TPM_BOTTOM;
    }
    Boa_Release(value);
  }
  else if(value->type == T_String) {
    if(fputs(Boa_String_Value(value), f->handle) == EOF) {
      ErrorSetErrno(); return TPM_BOTTOM;
    }
    Boa_Release(value);
  }
  else if(value->type == T_Int) {
    if(fprintf(f->handle, "%d", Boa_Int_Value(value)) < 0) {
      ErrorSetErrno(); return TPM_BOTTOM;
    }
    Boa_Release(value);
  }
  else if(value->type == T_Float) {
    if(fprintf(f->handle, "%g", Boa_Float_Value(value)) < 0) {
      ErrorSetErrno(); return TPM_BOTTOM;
    }
    Boa_Release(value);
  }
  else if(value->type == T_Double) {
    if(fprintf(f->handle, "%lg", Boa_Double_Value(value)) < 0) {
      ErrorSetErrno(); return TPM_BOTTOM;
    }
    Boa_Release(value);
  }
  else if(value->type == T_Ptr) {
    if(fwrite(Boa_Ptr_Value(value), 1, index, f->handle) != index) {
      ErrorSetErrno(); return TPM_BOTTOM;
    }
    Boa_Release(value);
  }
  else return TPM_BOTTOM;  /* don't know how to write it */
  return TPM_OK;
}

static int StringGet(Obj self, char *name, int flags,
		     Type *type_in_out, Obj *value_return)
{
  File f = (File)self->value;
  if(!strcmp(name, "seek")) {
    if(value_return) {
      Obj t = Boa_Create(T_FileSeek);
      Boa_Hold(self);
      ((Seek)t->value)->file = self;
      *value_return = t;
    }
    return TPM_OK;
  }
  else if(!strcmp(name, "tell")) {
    BOA_TYPECHECK_GET(T_Int);
    if(value_return) *value_return = Boa_Int_Create(ftell(f->handle));
    return TPM_OK;
  }
  else if(!strcmp(name, "line")) {
    BOA_TYPECHECK_GET(T_String);
    if(value_return) {
#if 1
      StrStream ss = StrStreamCreate(128, STRSTREAM_UNLIMITED);
      if(StrStreamScanLine(f->handle, ss) != TPM_OK) {
	StrStreamFree(ss); return TPM_BOTTOM;
      }
      *value_return = Boa_String_Create(strdup(StrStreamAsString(ss)));
      StrStreamFree(ss);
#else
      /* fast but unsafe version */
      char *line = Allocate(256, char);
      if(!getline(line, 256, f->handle)) return TPM_BOTTOM;
      *value_return = Boa_String_Create(line);
#endif
    }
    else {
      if(feof(f->handle)) { ErrorString = TPM_EEMPTY; return TPM_BOTTOM; }
      for(;;) {
	int c = getc(f->handle);
	if(c == EOF) break;
	if(c == '\n') break;
      }
    }
    return TPM_OK;
  }
  else if(!strcmp(name, "word")) {
    BOA_TYPECHECK_GET(T_String);
    if(value_return) {
      StrStream ss = StrStreamCreate(128, STRSTREAM_UNLIMITED);
      StrStreamScanWord(f->handle, ss);
      *value_return = Boa_String_Create(strdup(StrStreamAsString(ss)));
      StrStreamFree(ss);
    }
    else {
      int c, gotchar = 0;
      for(;;) {
	c = getc(f->handle);
	if(c == EOF) break;
	if(isspace(c)) { if(gotchar) break; else continue; }
	gotchar = 1;
      }
    }
    return TPM_OK;
  }
  else return TPM_BOTTOM;
}

static struct Boa_Methods methods = {
  Empty_Init, Final, Empty_Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  StringGet, Err_StringSet,
  IntGet, IntSet,
  Generic_ObjGet, Generic_ObjSet
};

/* Seek object ***************************************************************/

static int SeekSPrint(Obj self, StrStream ss)
{
  return StrStreamPuts(ss, "seek{aFileHandle}");
}

static int SeekIntGet(Obj self, int index, int flags,
		      Type *type_in_out, Obj *value_return)
{
  Seek seek = (Seek)self->value;
  File f = (File)seek->file->value;
  if(fseek(f->handle, index, 0) || (ftell(f->handle) != index)) { 
    ErrorSetErrno(); return TPM_BOTTOM; 
  }
  if(type_in_out || value_return)
    return IntGet(seek->file, 0, flags, type_in_out, value_return);
  else
    return TPM_OK;
}
static int SeekIntSet(Obj self, int index, int flags, Obj value)
{
  Seek seek = (Seek)self->value;
  File f = (File)seek->file->value;
  if(fseek(f->handle, index, 0) || (ftell(f->handle) != index)) { 
    ErrorSetErrno(); return TPM_BOTTOM; 
  }
  return IntSet(seek->file, 0, flags, value);
}

static struct Boa_Methods seek_methods = {
  Empty_Init, Sub_Final, Sub_Clean,
  SeekSPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Err_StringGet, Err_StringSet,
  SeekIntGet, SeekIntSet,
  Generic_ObjGet, Generic_ObjSet
};

Obj Boa_FileHandle_Define(void)
{
  T_FileSeek = Boa_SimpleType_Create("fileSeek", &seek_methods,
				     sizeof(struct Seek),
				     sizeof(Ptr));
  T_FileHandle = Boa_SimpleType_Create("fileHandle", &methods,
				       sizeof(struct File),
				       sizeof(Ptr));
  return T_FileHandle;
}
