/* Decorator for modifying requests to a context.
 * Does not filter attribute references.
 * Uses three separate maps for string, int and other Objs.
 * The output type and value determines the request to forward.
 * Attributes:
 *   context     Obj to forward to
 *   stringMap
 *   intMap
 *   objMap
 *   valueMap    given index/value, returns a new value
 */
#include <boa/boa.h>

#define DISPATCH_INT 1

/* must be kept synchronized with macros in std.h */
typedef struct Decorator {
  Obj context, stringMap, intMap, objMap, valueMap;
} *Decorator;

/* Globals *******************************************************************/
Type T_Filter;

/* Functions *****************************************************************/

static void Final(Obj self)
{
  Decorator dec = (Decorator)self->value;
  if(dec->context) Boa_Release(dec->context);
  if(dec->stringMap) Boa_Release(dec->stringMap);
  if(dec->intMap) Boa_Release(dec->intMap);
  if(dec->objMap) Boa_Release(dec->objMap);
  if(dec->valueMap) Boa_Release(dec->valueMap);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Decorator dec = (Decorator)self->value;
  if(dec->context) func(dec->context, data);
  if(dec->stringMap) func(dec->stringMap, data);
  if(dec->intMap) func(dec->intMap, data);
  if(dec->objMap) func(dec->objMap, data);
  if(dec->valueMap) func(dec->valueMap, data);
}

static int SPrint(Obj self, StrStream ss)
{
  Decorator dec = (Decorator)self->value;
  StrStreamPuts(ss, "filter(");
  if(dec->context) TRY_RETURN(Boa_SPrint(dec->context, ss));
  else StrStreamPuts(ss, "NULL");
  StrStreamPutc(ss, ' ');
  if(dec->stringMap) TRY_RETURN(Boa_SPrint(dec->stringMap, ss));
  else StrStreamPuts(ss, "NULL");
  StrStreamPutc(ss, ' ');
  if(dec->intMap) TRY_RETURN(Boa_SPrint(dec->intMap, ss));
  else StrStreamPuts(ss, "NULL");
  StrStreamPutc(ss, ' ');
  if(dec->objMap) TRY_RETURN(Boa_SPrint(dec->objMap, ss));
  else StrStreamPuts(ss, "NULL");
  StrStreamPutc(ss, ' ');
  if(dec->valueMap) TRY_RETURN(Boa_SPrint(dec->valueMap, ss));
  else StrStreamPuts(ss, "NULL");
  return StrStreamPutc(ss, ')');
}

static int Copy(Obj self, Ptr value_return)
{
  Decorator dec = (Decorator)self->value;
  Decorator new = (Decorator)value_return;
  if(dec->context) {
    new->context = BoaCopy(dec->context);
    if(!new->context) return TPM_ERROR;
  }
  if(dec->stringMap) new->stringMap = Boa_Hold(dec->stringMap);
  if(dec->intMap) new->intMap = Boa_Hold(dec->intMap);
  if(dec->objMap) new->objMap = Boa_Hold(dec->objMap);
  if(dec->valueMap) new->valueMap = Boa_Hold(dec->valueMap);
  return TPM_OK;
}

static int DispatchGet(Obj context, Obj index, int flags, 
		       Type *type_in_out, Obj *value_return)
{
  int status;
  if(index->type == T_String) {
    status = Boa_StringGet(context, Boa_String_Value(index), flags,
			 type_in_out, value_return);
  }
#if DISPATCH_INT
  else if(index->type == T_Int) {
    status = Boa_IntGet(context, Boa_Int_Value(index), flags,
			type_in_out, value_return);
  }
#endif
  else {
    status = Boa_ObjGet(context, index, flags,
			type_in_out, value_return);
  }
  Boa_Release(index);
  return status;
}
static int DispatchSet(Obj context, Obj index, int flags, Obj value)
{
  int status;
  if(index->type == T_String) {
    status = Boa_StringSet(context, Boa_String_Value(index), flags, value);
  }
#if DISPATCH_INT
  else if(index->type == T_Int) {
    status = Boa_IntSet(context, Boa_Int_Value(index), flags, value);
  }
#endif
  else {
    status = Boa_ObjSet(context, index, flags, value);
  }
  Boa_Release(index);
  return status;
}

static int StringGet(Obj self, char *name, int flags,
		     Type *type_in_out, Obj *value_return)
{
  Decorator dec = (Decorator)self->value;
  Obj t;
  if(name[0] == '_') {
    if(!strcmp(name+1, "context")) {
      BOA_TYPECHECK_GET(dec->context);
      if(value_return) *value_return = Boa_Hold(dec->context);
      return TPM_OK;
    }
    else if(!strcmp(name+1, "stringMap")) {
      BOA_TYPECHECK_GET(dec->stringMap);
      if(value_return) *value_return = Boa_Hold(dec->stringMap);
      return TPM_OK;
    }
    else if(!strcmp(name+1, "intMap")) {
      BOA_TYPECHECK_GET(dec->intMap);
      if(value_return) *value_return = Boa_Hold(dec->intMap);
      return TPM_OK;
    }
    else if(!strcmp(name+1, "objMap")) {
      BOA_TYPECHECK_GET(dec->objMap);
      if(value_return) *value_return = Boa_Hold(dec->objMap);
      return TPM_OK;
    }
    else if(!strcmp(name+1, "valueMap")) {
      BOA_TYPECHECK_GET(dec->valueMap);
      if(value_return) *value_return = Boa_Hold(dec->valueMap);
      return TPM_OK;
    }
  }
  if(!dec->context) return TPM_BOTTOM;
  if(!dec->stringMap || (name[0] == '_')) 
    return Boa_StringGet(dec->context, name, flags, type_in_out, value_return);
  else if(Boa_StringGet(dec->stringMap, name, flags, NULL, &t) != TPM_OK)
    return TPM_BOTTOM;
  return DispatchGet(dec->context, t, flags, type_in_out, value_return);
}
static int IntGet(Obj self, int index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Decorator dec = (Decorator)self->value;
  Obj t;
  if(!dec->context) return TPM_BOTTOM;
  if(!dec->intMap) 
    return Boa_IntGet(dec->context, index, flags, type_in_out, value_return);
  else if(Boa_IntGet(dec->intMap, index, flags, NULL, &t) != TPM_OK)
    return TPM_BOTTOM;
  return DispatchGet(dec->context, t, flags, type_in_out, value_return);
}
static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Decorator dec = (Decorator)self->value;
  Obj t;
  if(!dec->context) return TPM_BOTTOM;
  if(!dec->objMap) 
    return Boa_ObjGet(dec->context, index, flags, type_in_out, value_return);
  else if(Boa_ObjGet(dec->objMap, index, flags, NULL, &t) != TPM_OK)
    return TPM_BOTTOM;
  return DispatchGet(dec->context, t, flags, type_in_out, value_return);
}

static int StringSet(Obj self, char *name, int flags, Obj value)
{
  Decorator dec = (Decorator)self->value;
  Obj t;
  if(name[0] == '_') {
    if(!strcmp(name+1, "context")) {
      if(dec->context) Boa_Release(dec->context);
      dec->context = value;
      return TPM_OK;
    }
    else if(!strcmp(name+1, "stringMap")) {
      if(dec->stringMap) Boa_Release(dec->stringMap);
      dec->stringMap = value;
      return TPM_OK;
    }
    else if(!strcmp(name+1, "intMap")) {
      if(dec->intMap) Boa_Release(dec->intMap);
      dec->intMap = value;
      return TPM_OK;
    }
    else if(!strcmp(name+1, "objMap")) {
      if(dec->objMap) Boa_Release(dec->objMap);
      dec->objMap = value;
      return TPM_OK;
    }
    else if(!strcmp(name+1, "valueMap")) {
      if(dec->valueMap) Boa_Release(dec->valueMap);
      dec->valueMap = value;
      return TPM_OK;
    }
    /* fall through (could set attr of the indexer) */
  }
  if(!dec->context) return TPM_BOTTOM;
  if(!dec->valueMap) {
    if(!dec->stringMap || (name[0] == '_'))
      return Boa_StringSet(dec->context, name, flags, value);
    else if(Boa_StringGet(dec->stringMap, name, flags&~BOA_REPLACE, NULL, &t) != TPM_OK) 
      return TPM_BOTTOM;
    return DispatchSet(dec->context, t, flags, value);
  }
  else {
    Obj map;
    /* ignore the stringMap */
    /* use the index to get a map */
    if(Boa_StringGet(dec->valueMap, name, 0, NULL, &map) != TPM_OK) {
      /* no map for this index; forward unchanged */
      return Boa_StringSet(dec->context, name, flags, value);
    }
    /* take the value through the map */
    if(DispatchGet(map, value, 0, NULL, &t) != TPM_OK) {
      Boa_Release(map);
      ErrorPrepend("map failed");
      return TPM_BOTTOM;
    }
    Boa_Release(map);
    /* forward the modified request */
    return Boa_StringSet(dec->context, name, flags, t);
  }
}
static int IntSet(Obj self, int index, int flags, Obj value)
{
  Decorator dec = (Decorator)self->value;
  Obj t;
  if(!dec->context) return TPM_BOTTOM;
  if(!dec->valueMap) {
    if(!dec->intMap)
      return Boa_IntSet(dec->context, index, flags, value);
    else if(Boa_IntGet(dec->intMap, index, flags&~BOA_REPLACE, NULL, &t) 
	    != TPM_OK) 
      return TPM_BOTTOM;
    return DispatchSet(dec->context, t, flags, value);
  }
  else {
    Obj map;
    /* ignore the intMap */
    /* use the index to get a map */
    if(Boa_IntGet(dec->valueMap, index, 0, NULL, &map) != TPM_OK) {
      /* no map for this index; forward unchanged */
      return Boa_IntSet(dec->context, index, flags, value);
    }
    /* take the value through the map */
    if(DispatchGet(map, value, 0, NULL, &t) != TPM_OK) {
      Boa_Release(map);
      ErrorPrepend("map failed");
      return TPM_BOTTOM;
    }
    Boa_Release(map);
    /* forward the modified request */
    return Boa_IntSet(dec->context, index, flags, t);
  }
}
static int ObjSet(Obj self, Obj index, int flags, Obj value)
{
  Decorator dec = (Decorator)self->value;
  Obj t;
  if(!dec->context) return TPM_BOTTOM;
  if(!dec->valueMap) {
    if(!dec->objMap)
      return Boa_ObjSet(dec->context, index, flags, value);
    else if(Boa_ObjGet(dec->objMap, index, flags&~BOA_REPLACE, NULL, &t) 
	    != TPM_OK) 
      return TPM_BOTTOM;
    return DispatchSet(dec->context, t, flags, value);
  }
  else {
    Obj map;
    /* ignore the objMap */
    /* use the index to get a map */
    if(Boa_ObjGet(dec->valueMap, index, 0, NULL, &map) != TPM_OK) {
      /* no map for this index; forward unchanged */
      return Boa_ObjSet(dec->context, index, flags, value);
    }
    /* take the value through the map */
    if(DispatchGet(map, value, 0, NULL, &t) != TPM_OK) {
      Boa_Release(map);
      ErrorPrepend("map failed");
      return TPM_BOTTOM;
    }
    Boa_Release(map);
    /* forward the modified request */
    return Boa_ObjSet(dec->context, index, flags, t);
  }
}

static struct Boa_Methods methods = {
  Empty_Init, Final, Clean,
  SPrint, Copy, Generic_Compare, Generic_Hash,
  StringGet, StringSet,
  IntGet, IntSet,
  ObjGet, ObjSet
};

Obj Boa_Filter_Define(void)
{
  T_Filter = Boa_SimpleType_Create("filter", &methods,
				   sizeof(struct Decorator),
				   sizeof(Ptr));
  return T_Filter;
}
