/* A HashTable of Objs, keyed on their address and contents.
 * Used to check if an Obj is exactly the same one seen before.
 * Assumes pointers are unique and have infinite lifetimes (not!).
 * Requires: SimpleType.
 * ** doesn't match on contents yet **
 */

#include <boa/boa.h>
#include <tpm/hash.h>

typedef HashTable Dict;

/* Globals *******************************************************************/
Type T_IdDict;

/* Functions *****************************************************************/

static void Init(Obj self)
{
  Dict dict = (Dict)self->value;
  /* This is only difference with t_dict.c */
  HashTable new = HashTableCreate(HashPtr, PtrCmp, (FreeFunc*)Boa_Free, 
				  (FreeFunc*)Boa_Free);
  *dict = *new;
  free(new);
}

static void Final(Obj self)
{
  Dict dict = (Dict)self->value;
  /* taken from HashTableFree */
  HashTableFlush(dict);
  free(dict->buckets);
  free(dict->occupancy);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Obj t;
  HashIter(b, e, t, (Dict)self->value) {
    func(t, data);
  }
}

/* verbatim from t_dict.c */
static int SPrint(Obj self, StrStream ss)
{
  Dict dict = (Dict)self->value;
  Obj t;
  int space = 0;
  StrStreamPutc(ss, '{');
  {HashIter(b, e, t, dict) {
    if(space) StrStreamPuts(ss, ", ");
    TRY_RETURN(Boa_SPrint((Obj)e->key, ss));
    StrStreamPuts(ss, " : ");
    TRY_RETURN(Boa_SPrint(t, ss));
    space = 1;
  }}
  return StrStreamPutc(ss, '}');
}

/* verbatim from t_dict.c */
static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Dict dict = (Dict)self->value;
  Obj t;
  HashEntry *e;

  e = HashTableFindEntry(dict, index);
  if(!e) return TPM_BOTTOM;

  t = e->value;
  BOA_TYPECHECK_GET(t->type);
  if(flags & BOA_DELETE) {
    if(value_return) {
      /* remove without freeing */
      HashTableRemoveNoFree(dict, e);
      *value_return = t;
    }
    else HashTableRemoveEntry(dict, e);
  }
  else if(value_return) *value_return = Boa_Hold(t);
  return TPM_OK;
}

/* verbatim from t_dict.c */
static int ObjSet(Obj self, Obj index, int flags, Obj value)
{
  Dict dict = (Dict)self->value;
  HashEntry *e;
  int found;

  if(flags & BOA_INSERT) {
    e = HashTableAddEntry(dict, index, &found);
    if(!found) assert(e->key = Boa_Stabilize(index));
  }
  else {
    found = !!(e = HashTableFindEntry(dict, index));
    if(!found) return TPM_BOTTOM;
  }
  if(found) {
    if(!(flags & BOA_REPLACE)) { ErrorString = BOA_EFLAGS; return TPM_ERROR; }
    Boa_Release((Obj)e->value);
  }
  e->value = value;
  return TPM_OK;
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Err_StringGet, Err_StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_IdDict_Define(void)
{
  T_IdDict = Boa_SimpleType_Create("idDict", &methods,
				   sizeof(struct HashTable), /* size */
				   sizeof(Ptr));             /* align */
  return T_IdDict;
}
