/* Index context.  Maintains an obj->int mapping from 0..N-1.
 * Insertions get index N++.
 * Deletions shift all numbers down, i.e. there are no holes.
 */
#include <boa/boa.h>

/* both items and dict hold the Objs */
typedef struct Index {
  Obj *items;   /* maps int -> Obj */
  int length, bufsize;
  Obj dict;     /* reverse mapping: Obj -> int */
  int missing;  /* first index which is not in the dict */
} *Index;

/* Globals *******************************************************************/
Type T_Index;

/* Functions *****************************************************************/
static void Init(Obj self)
{
  Index ind = (Index)self->value;
  ind->bufsize = 1024;
  ind->items = Allocate(ind->bufsize, Obj);
  ind->dict = Boa_Create(T_Dict);
  ind->missing = 0;
}

static void Final(Obj self)
{
  Index ind = (Index)self->value;
  int i;
  for(i=0;i<ind->length;i++) {
    Boa_Release(ind->items[i]);
  }
  free(ind->items);
  Boa_Release(ind->dict);
}

/* prints just like an array */
static int SPrint(Obj self, StrStream ss)
{
  Index ind = (Index)self->value;
  int i;
  StrStreamPutc(ss, '{');
  for(i=0;i<ind->length;i++) {
    TRY_RETURN(Boa_SPrint(ind->items[i], ss));
    StrStreamPutc(ss, ' ');
  }
  return StrStreamPutc(ss, '}');
}

static int StringGet(Obj self, char *name, int flags,
		     Type *type_in_out, Obj *value_return)
{
  return Generic_StringGet(self, name, flags & ~BOA_SPLIT, 
			   type_in_out, value_return);
}
static int StringSet(Obj self, char *name, int flags, Obj value)
{
  return Generic_StringSet(self, name, flags & ~BOA_SPLIT, value);
}

static void Insert(Index ind, Obj index)
{
  ind->length++;
  if(ind->length > ind->bufsize) {
    ind->bufsize *= 2;
    Reallocate(ind->items, ind->bufsize, Obj);
  }
  ind->items[ind->length-1] = index;
}

static int IntGet(Obj self, int index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Index ind = (Index)self->value;
  if((index < 0) || (index >= ind->length)) return TPM_BOTTOM;
  BOA_TYPECHECK_GET(ind->items[index]->type);
  if(value_return) *value_return = Boa_Hold(ind->items[index]);
  return TPM_OK;
}

/* Set can only add to the end. */
static int IntSet(Obj self, int index, int flags, Obj value)
{
  Index ind = (Index)self->value;
  if(index != ind->length) {
    ErrorSet("can only add to the end of an index");
    return TPM_BOTTOM;
  }
  if(!(flags & BOA_INSERT)) { ErrorString = BOA_EFLAGS; return TPM_ERROR; }
  value = Boa_Stabilize(value);
  if(!value) return TPM_BOTTOM;
  Insert(ind, value);
  return TPM_OK;
}

/* Forwards to IntSet(value, index) */
static int ObjSet(Obj self, Obj index, int flags, Obj value)
{
  int i;
  if(value->type != T_Int) return TPM_BOTTOM;
  i = Boa_Int_Value(value);
  Boa_Release(value);
  return IntSet(self, i, flags, index);
}

static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Index ind = (Index)self->value;
  Obj t;
  int i, status;
  if(index->type == T_String) {
    char *name = Boa_String_Value(index);
    if(name[0] == '_') {
      if(!strcmp(name+1, "length")) {
	BOA_TYPECHECK_GET(T_Int);
	if(value_return) *value_return = Boa_Int_Create(ind->length);
	return TPM_OK;
      }
    /* fall through */
    }
  }

  status = Boa_ObjGet(ind->dict, index, flags, &T_Int, &t);
  if(status != TPM_OK) {
    if(ind->missing < ind->length) {
      /* insert any missing items */
      for(;ind->missing < ind->length;ind->missing++) {
	/* ignore failure */
	Boa_ObjSet(ind->dict, ind->items[ind->missing], BOA_INSERT, 
		   Boa_Int_Create(ind->missing));
      }
      /* and try again */
      status = Boa_ObjGet(ind->dict, index, flags, &T_Int, &t);
    }
    if(status != TPM_OK) {
      if(!(flags & BOA_INSERT)) return TPM_BOTTOM;
      t = Boa_Int_Create(ind->length);
      if(value_return) *value_return = Boa_Hold(t);
      return ObjSet(self, index, flags, t);
    }
  }
  BOA_TYPECHECK_GET(T_Int);
  i = Boa_Int_Value(t);
  if(value_return) *value_return = t;
  else Boa_Release(t);

  if(flags & BOA_DELETE) {
    Boa_Release(ind->items[i]);
    ind->length--;
    if(ind->length < ind->bufsize/4) {
      ind->bufsize /= 2;
      Reallocate(ind->items, ind->bufsize, Obj);
    }
    if(ind->missing > i) ind->missing--;
    /* shift down */
    memmove(&ind->items[i], &ind->items[i+1], 
	    (ind->length-i)*sizeof(Obj));
    for(;i<ind->length;i++) {
      assert(Boa_ObjSet(ind->dict, ind->items[i], 
			BOA_REPLACE, Boa_Int_Create(i)) == TPM_OK);
    }
  }
  return TPM_OK;
}

static struct Boa_Methods methods = {
  Init, Final, Empty_Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  StringGet, StringSet,
  IntGet, IntSet,
  ObjGet, ObjSet
};

Obj Boa_Index_Define(void)
{
  T_Index = Boa_SimpleType_Create("index", &methods,
				  sizeof(struct Index), sizeof(Ptr));
  return T_Index;
}
