/* Iterator for contexts with integer indexing. */
/* An iter must obey the Stream API:
 * Get/IntSet will reposition the head after the operation.
 * Get/Set to "seek" will access position determined by indexer object,
 * but not move it afterwards.
 * "tell" will yield the current integer position.
 */
#include <boa/boa.h>

/* these must be synchronized with macros in std.h */
typedef struct Iter {
  Obj data;
  Obj seek;
  int pos;
} *Iter;

/* this could be simplified by using an AddStrings decorator */
typedef struct Seek {
  Obj indexer, iter;
} *Seek;

/* Globals *******************************************************************/
Type T_Iter;
static Type T_IterSeek;

/* Functions *****************************************************************/

/* caller must set data and seek->indexer */
static void Init(Obj self)
{
  Iter iter = (Iter)self->value;
  iter->seek = Boa_Create(T_IterSeek);
  Boa_Hold(self);
  ((Seek)iter->seek->value)->iter = self;
}

/* these should all check that data != NULL */
static void Final(Obj self)
{
  Iter iter = (Iter)self->value;
  Boa_Release(iter->data);
  Boa_Release(iter->seek);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Iter iter = (Iter)self->value;
  func(iter->data, data);
  func(iter->seek, data);
}

static int SPrint(Obj self, StrStream ss)
{
  Iter iter = (Iter)self->value;
  TRY_RETURN(StrStreamPuts(ss, "iter{"));
  TRY_RETURN(Boa_SPrint(iter->data, ss));
  return StrStreamPutc(ss, '}');
}

/* These two just forward the operation and move the head. */
static int IntGet(Obj self, int index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Iter iter = (Iter)self->value;
  int status = Boa_IntGet(iter->data, iter->pos, flags, 
			  type_in_out, value_return);
  iter->pos += index;
  return status;
}
static int IntSet(Obj self, int index, int flags, Obj value)
{
  Iter iter = (Iter)self->value;
  int status = Boa_IntSet(iter->data, iter->pos, flags, value);
  iter->pos += index;
  return status;
}

static int StringGet(Obj self, char *name, int flags,
		     Type *type_in_out, Obj *value_return)
{
  Iter iter = (Iter)self->value;
  if(!strcmp(name, "seek")) {
    if(value_return) {
      Boa_Hold(iter->seek);
      *value_return = iter->seek;
    }
    return TPM_OK;
  }
  else if(!strcmp(name, "tell")) {
    if(!type_in_out || (!*type_in_out && (*type_in_out = T_Int)) ||
       (*type_in_out == T_Int)) {
      if(value_return) {
	Obj t = Boa_Create(T_Int);
	Boa_Int_Value(t) = iter->pos;
	*value_return = t;
      }
      return TPM_OK;
    }
    else if(*type_in_out == T_String) {
      /* ask the seek indexer */
      /* could do this for Int, too */
      return Boa_IntGet(((Seek)iter->seek->value)->indexer, iter->pos, 0, 
		       type_in_out, value_return);
    }
    return TPM_BOTTOM;  /* wanted neither int nor string */
  }
  else return TPM_BOTTOM;
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  StringGet, Err_StringSet,
  IntGet, IntSet,
  Generic_ObjGet, Generic_ObjSet
};

/* Seek object ***************************************************************/

/* these should all check that indexer != NULL */
static void SeekFinal(Obj self)
{
  Seek seek = (Seek)self->value;
  if(seek->indexer) Boa_Release(seek->indexer);
  Boa_Release(seek->iter);
}

static void SeekClean(Obj self, ObjFunc *func, Any data)
{
  Seek seek = (Seek)self->value;
  if(seek->indexer) func(seek->indexer, data);
  func(seek->iter, data);
}

static int SeekSPrint(Obj self, StrStream ss)
{
  Seek seek = (Seek)self->value;
  StrStreamPuts(ss, "iter_seek{");
  if(seek->indexer) Boa_SPrint(seek->indexer, ss);
  else StrStreamPuts(ss, "NULL");
  StrStreamPuts(ss, ", ");
  Boa_SPrint(seek->iter, ss);
  return StrStreamPutc(ss, '}');
}

static int SeekStringGet(Obj self, char *name, int flags,
			 Type *type_in_out, Obj *value_return)
{
  Seek seek = (Seek)self->value;
  if(!seek->indexer) return TPM_BOTTOM;
  /* ask the seek indexer for the position */
  Boa_StringGetInt(seek->indexer, name, &((Iter)seek->iter->value)->pos);
  /* forward the operation to the iter */
  return IntGet(seek->iter, 0, flags, type_in_out, value_return);
}

static int SeekIntGet(Obj self, int index, int flags,
		      Type *type_in_out, Obj *value_return)
{
  Seek seek = (Seek)self->value;
  if(!seek->indexer) ((Iter)seek->iter->value)->pos = index;
  else {
    /* ask the seek indexer for the position */
    Boa_IntGetInt(seek->indexer, index, &((Iter)seek->iter->value)->pos);
  }
  /* forward the operation to the iter */
  return IntGet(seek->iter, 0, flags, type_in_out, value_return);
}

static int SeekStringSet(Obj self, char *name, int flags, Obj value)
{
  Seek seek = (Seek)self->value;
  if(!seek->indexer) return TPM_BOTTOM;
  /* ask the seek indexer for the position */
  Boa_StringGetInt(seek->indexer, name, &((Iter)seek->iter->value)->pos);
  /* forward the operation to the iter */
  return IntSet(seek->iter, 0, flags, value);
}

static int SeekIntSet(Obj self, int index, int flags, Obj value)
{
  Seek seek = (Seek)self->value;
  if(!seek->indexer) ((Iter)seek->iter->value)->pos = index;
  else {
    /* ask the seek indexer for the position */
    Boa_IntGetInt(seek->indexer, index, &((Iter)seek->iter->value)->pos);
  }
  /* forward the operation to the iter */
  return IntSet(seek->iter, 0, flags, value);
}

static struct Boa_Methods seek_methods = {
  Empty_Init, SeekFinal, SeekClean,
  SeekSPrint, Err_Copy, Generic_Compare, Generic_Hash,
  SeekStringGet, SeekStringSet,
  SeekIntGet, SeekIntSet,
  Generic_ObjGet, Generic_ObjSet
};

Obj Boa_Iter_Define(void)
{
  T_IterSeek = Boa_SimpleType_Create("iter_seek", &seek_methods,
				     sizeof(struct Seek),
				     sizeof(Ptr));
  T_Iter = Boa_SimpleType_Create("iter", &methods,
				 sizeof(struct Iter),
				 sizeof(Ptr));
  return T_Iter;
}
