/* If BOA_SPLIT is set, takes a collection of requests to a context 
 * and returns a collection of results.
 * Otherwise, just forwards to the context.
 * Named to sound like Lisp's "mapcar".
 * Attributes:
 *   context
 */
#include <tpm/list.h>
#include <tpm/hash.h>
#include <boa/boa.h>

typedef struct Map {
  Obj context;
} *Map;

/* Globals *******************************************************************/
Type T_MapCon;

/* Functions *****************************************************************/

Obj Boa_MapCon_Create(Obj context)
{
  Obj self = Boa_Create(T_MapCon);
  Map m = (Map)self->value;
  Boa_Release(m->context);
  m->context = context;
  return self;
}

static void Init(Obj self)
{
  Map m = (Map)self->value;
  m->context = Boa_Create(T_Merge);   /* dummy context */
}

static void Final(Obj self)
{
  Map m = (Map)self->value;
  Boa_Release(m->context);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Map m = (Map)self->value;
  func(m->context, data);
}

static int SPrint(Obj self, StrStream ss)
{
  Map m = (Map)self->value;
  StrStreamPuts(ss, "mapCon");
  return Boa_SPrint(m->context, ss);
}

static int StringGet(Obj self, char *name, int flags,
		     Type *type_in_out, Obj *value_return)
{
  Map m = (Map)self->value;
  if(name[0] == '_') {
    if(!strcmp(name+1, "context")) {
      BOA_TYPECHECK_GET(m->context->type);
      if(value_return) *value_return = Boa_Hold(m->context);
      return TPM_OK;
    }
  }
  return TPM_BOTTOM;
}
static int StringSet(Obj self, char *name, int flags, Obj value)
{
  Map m = (Map)self->value;
  if(name[0] == '_') {
    if(!strcmp(name+1, "context")) {
      Boa_Release(m->context);
      m->context = value;
      return TPM_OK;
    }
  }
  return TPM_BOTTOM;
}

/* If BOA_SPLIT is set, gets all of the values in index, 
 * returning all of the results in a compound structure. 
 * Otherwise forwards to the context.
 * If type_in_out == NULL, uses the same structure as the input.
 * XXX uses same flags for each value; is this right?
 */
static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Map m = (Map)self->value;
  Obj in, out, result;
#if 0
  if((index->type == T_String) || !(flags & BOA_SPLIT)) {
    /* forward to the context */
    return Boa_ObjGet(m->context, index, flags, type_in_out, value_return);
  }
#endif
  if(index->type == T_List) {
    /* Output is a positional collection */
    int i = 0;
    if(!type_in_out || (!*type_in_out && (*type_in_out = index->type)))
      result = Boa_Create(index->type);
    else
      result = Boa_Create(*type_in_out);
    {ListIter(p, in, Boa_List_Value(index)) {
      int status = Boa_ObjGet(m->context, in, flags, NULL, &out);
      if(status != TPM_OK) {
	Boa_Release(result);
	return status;
      }
      status = Boa_IntSet(result, i, BOA_INSERT|BOA_REPLACE, out);
      if(status != TPM_OK) {
	ErrorPrint(stderr);
	Boa_Release(out);
	Boa_Release(result);
	return status;
      }
      i++;
    }}
    BOA_TYPECHECK_GET(result->type);
    *value_return = result;
    return TPM_OK;
  }
  else if(index->type == T_Dict) {
    /* Output is an associative collection */
    if(!type_in_out || (!*type_in_out && (*type_in_out = index->type)))
      result = Boa_Create(index->type);
    else
      result = Boa_Create(*type_in_out);
    {HashIter(b, e, in, (HashTable)index->value) {
      int status = Boa_ObjGet(m->context, in, flags, NULL, &out);
      if(status != TPM_OK) {
	Boa_Release(result);
	return status;
      }
      assert(Boa_ObjSet(result, in, BOA_INSERT|BOA_REPLACE, out) == TPM_OK);
    }}
    BOA_TYPECHECK_GET(result->type);
    *value_return = result;
    return TPM_OK;
  }
  else if(index->type == T_Index) {
    /* Output is a positional collection */
    int i;
    if(!type_in_out || (!*type_in_out && (*type_in_out = index->type)))
      result = Boa_Create(index->type);
    else
      result = Boa_Create(*type_in_out);
    for(i=0;i<Boa_Index_Length(index);i++) {
      int status;
      in = Boa_Index_Item(index, i);
      /* this is lame.  I wish I could unify stringGet and objGet. */
      if(in->type == T_String)
	status = Boa_StringGet(m->context, Boa_String_Value(in), flags, 
			       NULL, &out);
      else 
	status = Boa_ObjGet(m->context, in, flags, NULL, &out);
      if(status != TPM_OK) {
	Boa_Release(result);
	return status;
      }
      status = Boa_IntSet(result, i, BOA_INSERT|BOA_REPLACE, out);
      if(status != TPM_OK) {
	ErrorPrint(stderr);
	Boa_Release(out);
	Boa_Release(result);
	return status;
      }
    }
    BOA_TYPECHECK_GET(result->type);
    *value_return = result;
    return TPM_OK;
  }
  else if(index->type->type == T_Array) {
    /* Output is a positional collection */
    int i;
    if(!type_in_out || (!*type_in_out && (*type_in_out = index->type)))
      result = Boa_Create(index->type);
    else
      result = Boa_Create(*type_in_out);
    if(Boa_Array_Subtype(index->type) == T_String) {
      /* do StringGets */
      char **data = (char**)index->value;
      for(i=0;i<Boa_Array_Length(index->type);i++) {
	int status = Boa_StringGet(m->context, data[i], flags, NULL, &out);
	if(status != TPM_OK) {
	  Boa_Release(result);
	  return status;
	}
	status = Boa_IntSet(result, i, BOA_INSERT|BOA_REPLACE, out);
	if(status != TPM_OK) {
	  ErrorPrint(stderr);
	  Boa_Release(out);
	  Boa_Release(result);
	  return status;
	}
      }
    }
    else {
      /* do ObjGets */
      Obj t = Boa_CreateFromData(Boa_Array_Subtype(index->type), NULL);
      for(i=0;i<Boa_Array_Length(index->type);i++) {
	int status;
	t->value = (char*)index->value + i*TypeSize(t->type);
	status = Boa_ObjGet(m->context, t, flags, NULL, &out);
	if(status != TPM_OK) {
	  Boa_Release(t);
	  Boa_Release(result);
	  return status;
	}
	status = Boa_IntSet(result, i, BOA_INSERT|BOA_REPLACE, out);
	if(status != TPM_OK) {
	  ErrorPrint(stderr);
	  Boa_Release(out);
	  Boa_Release(result);
	  return status;
	}
      }
      Boa_Release(t);
    }
    if(value_return) *value_return = result;
    else Boa_Release(result);
    return TPM_OK;
  }
  return TPM_BOTTOM;
}

/* Sets all of the values in index to value. */
static int ObjSet(Obj self, Obj index, int flags, Obj value)
{
  Map m = (Map)self->value;
  if((index->type == T_String) || !(flags & BOA_SPLIT)) {
    return Boa_ObjSet(m->context, index, flags, value);
  }
  return TPM_BOTTOM;
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  StringGet, StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_MapCon_Define(void)
{
  T_MapCon = Boa_SimpleType_Create("mapCon", &methods,
				   sizeof(struct Map), sizeof(Ptr));
  return T_MapCon;
}
