/* An ordered merge of inherited contexts.
 * Boa_Methods are dispatched to the inherited objects, in turn, until
 * succeeding on one of them.
 * Attributes:
 *   _     Ord of inherited objects
 */

#include <boa/boa.h>

typedef struct Merge {
  Obj inherit; /* Ord */
} *Merge;

/* Globals *******************************************************************/
Type T_Merge;

/* Functions *****************************************************************/

static void Init(Obj self)
{
  Merge merge = (Merge)self->value;
  merge->inherit = Boa_Create(T_Ord);
}

static void Final(Obj self)
{
  Merge merge = (Merge)self->value;
  Boa_Release(merge->inherit);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Merge merge = (Merge)self->value;
  func(merge->inherit, data);
}

static int SPrint(Obj self, StrStream ss)
{
  Merge merge = (Merge)self->value;
  StrStreamPuts(ss, "merge");
  return Boa_SPrint(merge->inherit, ss);
}

static int StringGet(Obj self, char *name, int flags,
		     Type *type_in_out, Obj *value_return)
{
  Merge merge = (Merge)self->value;
  int i;

  if(name[0] == '_') {
    if(!name[1]) {
      /* cannot delete the Ord */
      if(flags & BOA_DELETE) return TPM_ERROR;
      BOA_TYPECHECK_GET(merge->inherit->type);
      if(value_return) {
	Boa_Hold(merge->inherit);
	*value_return = merge->inherit;
      }
      return TPM_OK;
    }
    else if((flags & BOA_SPLIT) && (name[1] == BOA_SEP(flags))) {
      return Boa_StringGet(merge->inherit, name+2, flags, 
			  type_in_out, value_return);
    }
  }

  /* dispatch to the inherited objects in turn, until one succeeds */
  for(i=0;i < Boa_Ord_Length(merge->inherit);i++) {
    Obj t = Boa_Ord_Item(merge->inherit, i);
    int status = Boa_StringGet(t, name, flags,
			      type_in_out, value_return);
    if(status != TPM_BOTTOM) return status;
  }
  return TPM_BOTTOM;
}

static int IntGet(Obj self, int index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Merge merge = (Merge)self->value;
  int i;
  /* dispatch to the inherited objects in turn, until one succeeds */
  for(i=0;i < Boa_Ord_Length(merge->inherit);i++) {
    Obj t = Boa_Ord_Item(merge->inherit, i);
    int status = Boa_IntGet(t, index, flags,
			   type_in_out, value_return);
    if(status != TPM_BOTTOM) return status;
  }
  return TPM_BOTTOM;
}

static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Merge merge = (Merge)self->value;
  int i;
  /* dispatch to the inherited objects in turn, until one succeeds */
  for(i=0;i < Boa_Ord_Length(merge->inherit);i++) {
    Obj t = Boa_Ord_Item(merge->inherit, i);
    int status = Boa_ObjGet(t, index, flags,
			   type_in_out, value_return);
    if(status != TPM_BOTTOM) return status;
  }
  return TPM_BOTTOM;
}

static int StringSet(Obj self, char *name, int flags, Obj value)
{
  Merge merge = (Merge)self->value;
  int i;

  if(name[0] == '_') {
    if(!name[1]) {
      Boa_Release(merge->inherit);
      merge->inherit = value;
      return TPM_OK;
    }
    else if((flags & BOA_SPLIT) && (name[1] == BOA_SEP(flags))) {
      return Boa_StringSet(merge->inherit, name+2, flags, value);
    }
  }

  /* dispatch to the inherited objects in turn, until one succeeds */
  for(i=0;i < Boa_Ord_Length(merge->inherit);i++) {
    Obj t = Boa_Ord_Item(merge->inherit, i);
    int status = Boa_StringSet(t, name, flags, value);
    if(status != TPM_BOTTOM) return status;
  }
  return TPM_BOTTOM;
}

static int IntSet(Obj self, int index, int flags, Obj value)
{
  Merge merge = (Merge)self->value;
  int i;
  /* dispatch to the inherited objects in turn, until one succeeds */
  for(i=0;i < Boa_Ord_Length(merge->inherit);i++) {
    Obj t = Boa_Ord_Item(merge->inherit, i);
    int status = Boa_IntSet(t, index, flags, value);
    if(status != TPM_BOTTOM) return status;
  }
  return TPM_BOTTOM;
}

static int ObjSet(Obj self, Obj index, int flags, Obj value)
{
  Merge merge = (Merge)self->value;
  int i;
  /* dispatch to the inherited objects in turn, until one succeeds */
  for(i=0;i < Boa_Ord_Length(merge->inherit);i++) {
    Obj t = Boa_Ord_Item(merge->inherit, i);
    int status = Boa_ObjSet(t, index, flags, value);
    if(status != TPM_BOTTOM) return status;
  }
  return TPM_BOTTOM;
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  StringGet, StringSet,
  IntGet, IntSet,
  ObjGet, ObjSet
};

Obj Boa_Merge_Define(void)
{
  T_Merge = Boa_SimpleType_Create("merge", &methods,
				  sizeof(struct Merge),
				  sizeof(Ptr));
  return T_Merge;
}
