/* Obj struct.  An array of objs whose values live in a 
 * contiguous C struct.  Insertion adds to the end of the struct;
 * deletion shifts downward.  Both make sure to preserve alignment.
 */
#include <boa/boa.h>

typedef struct ObStruct {
  char *data;
  int length; /* size of the data in bytes */
  int nitems; /* total number of items */
  Obj *items;
} *ObStruct;

/* Globals *******************************************************************/
Type T_ObStruct;

/* Functions *****************************************************************/

/* pull t's value out of the struct */
#define Extract(t) \
 {(t)->value = AllocCopy(TypeSize((t)->type), char, (t)->value); \
  BoaFreeOR01(t);}

static void ItemFree(Obj t)
{
  if(--t->ref_count == 0) t->freeFunc(t);
  else Extract(t);
}

static void Init(Obj self)
{
  ObStruct ob = (ObStruct)self->value;
  ob->items = Allocate(1, Obj);
  ob->data = Allocate(1, char);
}

static void Final(Obj self)
{
  ObStruct ob = (ObStruct)self->value;
  int i;
  for(i=0;i<ob->nitems;i++) ItemFree(ob->items[i]);
  free(ob->items);
  free(ob->data);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  ObStruct ob = (ObStruct)self->value;
  int i;
  for(i=0;i<ob->nitems;i++) func(ob->items[i], data);
}

/* prints like a Tcl list */
static int SPrint(Obj self, StrStream ss)
{
  ObStruct ob = (ObStruct)self->value;
  int i;
  StrStreamPutc(ss, '{');
  for(i=0;i<ob->nitems;i++) {
    TRY_RETURN(Boa_SPrint(ob->items[i], ss));
    StrStreamPutc(ss, ' ');
  }
  StrStreamPutc(ss, '}');
  return TPM_OK;
}

static int StringGet(Obj self, char *name, int flags,
		     Type *type_in_out, Obj *value_return)
{
  ObStruct ob = (ObStruct)self->value;
  if(name[0] == '_') {
    if(!strcmp(name+1, "length")) {
      BOA_TYPECHECK_GET(T_Int);
      if(value_return) *value_return = Boa_Int_Create(ob->nitems);
      return TPM_OK;
    }
    else if(!strcmp(name+1, "iter")) {
      Obj iter = Boa_Create(T_Iter);
      Boa_Iter_Data(iter) = self;
      Boa_StringGet(Boa_Globals, "atoi", 0, NULL, &Boa_Iter_Indexer(iter));
      *value_return = iter;
      return TPM_OK;
    }
  }
  return TPM_BOTTOM;
}

static int IntGet(Obj self, int index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  ObStruct ob = (ObStruct)self->value;
  Obj t;

  if((index < 0) || (index >= ob->nitems)) return TPM_BOTTOM;
  t = ob->items[index];
  BOA_TYPECHECK_GET(t->type);

  if(flags & BOA_DELETE) {
    char *dest, *src;
    if(value_return) {
      Extract(t);
      *value_return = t;
    }
    /* remove from the Obj array; don't realloc */
    ob->nitems--;
    memmove(&ob->items[index], &ob->items[index+1], 
	   (ob->nitems-index)*sizeof(Obj));

    /* shrink the data block, but don't realloc; 
     * that will happen on the next insert. */
    /* position dest immediately after the previous item */
    if(index == 0) dest = ob->data;
    else dest = (char*)ob->items[index-1]->value + 
      TypeSize(ob->items[index-1]->type);
    for(;;index++) {
      if(index == ob->nitems) {
	ob->length = (long)dest - (long)ob->data;
	break;
      }
      t = ob->items[index];
      src = t->value;
      /* align dest to that item */
      MakeAligned(dest, TypeAlign(t->type));
      if(dest == src) break; /* no change; we're done */
      /* move src to dest (may overlap) */
      memmove(dest, src, TypeSize(t->type));
      t->value = dest;
      dest += TypeSize(t->type);
    }
  }
  else if(value_return) {
    Boa_Hold(t);
    *value_return = t;
  }
  return TPM_OK;
}

/* Allocates a new space at the end for an item of the given type.
 * Repositions items if the data ptr was moved.
 * Returns the address of the new item.
 */
static char *NewItem(ObStruct ob, Type type)
{
  char *old_data = ob->data;
  ob->nitems++;
  Reallocate(ob->items, ob->nitems, Obj);
  MakeAlignedInt(ob->length, TypeAlign(type));
  ob->length += TypeSize(type);
  Reallocate(ob->data, ob->length, char);
  if(ob->data != old_data) {
    /* reposition entries */
    Obj t;
    int i;
    for(i=0;i<ob->nitems-1;i++) {
      t = ob->items[i];
      t->value = (char*)t->value + (long)ob->data - (long)old_data;
    }
  }
  return ob->data + ob->length - TypeSize(type);
}

static int IntSet(Obj self, int index, int flags, Obj value)
{
  ObStruct ob = (ObStruct)self->value;
  Obj t;
  char *p;

  if(index < 0) return TPM_BOTTOM;
  if((flags & BOA_INSERT) && (index == ob->nitems)) {
    /* add a space on the end */
    p = NewItem(ob, value->type);
  }
  else {
    if(index >= ob->nitems) return TPM_BOTTOM;
    if(!(flags & BOA_REPLACE)) { ErrorString = BOA_EFLAGS; return TPM_ERROR; }
    t = ob->items[index];
    /* replacement must have the same type */
    if(Boa_Compare(t->type, value->type)) return TPM_BOTTOM;
    p = t->value;
    ItemFree(t);
    /* fall through */
  }

  /* place the new value in p */
  /* can we free the Obj's value block? */
  if(BoaFreeTest01(value)) {
    /* grab that Obj and seize the value */
    ob->items[index] = value;
    memcpy(p, value->value, TypeSize(value->type));
    /* free the Obj's value block */
    free(value->value);
    BoaFreeAND10(value);
    value->value = p;
  }
  else {
    /* make a copy of the value */
    if(Boa_Copy(value, p) != TPM_OK) {
      ErrorString = "copy failure in PkDict";
      fprintf(stderr, "%s\n", ErrorString);
      return TPM_ERROR;
    }
    /* make a new Obj */
    t = Boa_CreateFromData(value->type, p);
    BoaFreeAND10(t);
    ob->items[index] = t;
    Boa_Release(value);
  }
  return TPM_OK;
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  StringGet, Err_StringSet,
  IntGet, IntSet,
  Generic_ObjGet, Generic_ObjSet
};

Obj Boa_ObStruct_Define(void)
{
  T_ObStruct = Boa_SimpleType_Create("obStruct", &methods,
				     sizeof(struct ObStruct), sizeof(Ptr));
  return T_ObStruct;
}
