/* Allows objects to be read from a UNIX directory. 
 * A formatter strategy does the stream/object conversion.
 * Attributes:
 *   path
 *   formatter
 * Requires: FileHandle, SimpleType.
 */
/* can the path lookup be optimized? e.g. chdir?
 * no - would affect the whole process
 */
#include <boa/boa.h>
#include <sys/stat.h>
#include <dirent.h>

#define VERBOSE 0

typedef struct ObjDir {
  char *path;
  Obj formatter;
} *ObjDir;

/* Globals *******************************************************************/
Type T_ObjDir;
Obj Boa_ObjDir_DefaultFormatter;

/* Functions *****************************************************************/

static void Init(Obj self)
{
  ObjDir od = (ObjDir)self->value;
  od->path = strdup(".");
  Boa_Hold(Boa_ObjDir_DefaultFormatter);
  od->formatter = Boa_ObjDir_DefaultFormatter;
}

static void Final(Obj self)
{
  ObjDir od = (ObjDir)self->value;
  free(od->path);
  Boa_Release(od->formatter);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  ObjDir od = (ObjDir)self->value;
  func(od->formatter, data);
}

static int SPrint(Obj self, StrStream ss)
{
  ObjDir od = (ObjDir)self->value;
  StrStreamPuts(ss, "objdir(");
  StrStreamPuts(ss, od->path);
  StrStreamPutc(ss, ' ');
  TRY_RETURN(Boa_SPrint(od->formatter, ss));
  return StrStreamPutc(ss, ')');
}

static int ReadFile(Obj formatter, char *name, 
		    Type *type_in_out, Obj *value_return)
{
  int status;
  Obj t;
  Obj file = FileHandle_Open(name, "r");
  if(!file) { ErrorSet("Cannot read `%s'", name); ErrorAppendErrno(); return TPM_BOTTOM; }
#if 1
  fprintf(stderr, "objDir: reading `%s' with ", name);
  Boa_Print(stderr, formatter);
  putc('\n', stderr);
#endif
  /* tell the formatter the filename */
  t = Boa_String_Create(strdup(name));
  /* if this fails, we have to release t ourselves */
  if(Boa_StringSet(formatter, "_path", 0, t) != TPM_OK) Boa_Release(t);
  status = Boa_ObjGet(formatter, file, 0, type_in_out, value_return);
  Boa_Release(file);  /* closes fp */
  return status;
}

static char *TranslatePath(char *path)
{
  if(path[0] == '~') {
    /* replace with an absolute path */
    char *p, *str;
    for(p=path+1;*p && *p != '/';p++);
    if(*p) {
      *p = 0;
      str = Path_Home(path+1);
      *p = '/';
      if(!str) return NULL;
      path = Allocate(strlen(str)+strlen(p)+1, char);
      sprintf(path, "%s%s", str, p);
      free(str);
    }
    else {
      path = Path_Home(path+1);
      if(!path) return NULL;
    }
  }
  else path = strdup(path);
  if(path[0] == '/') {
    /* construct a string of "../"'s equivalent to "/" and work on that */
    char *p, *str = Path_Ups(".");
    p = path+1;
    if(!str) { free(path); return NULL; }
    if(*p) {
      char *t = path;
      path = Allocate(strlen(str)+strlen(p)+2, char);
      sprintf(path, "%s/%s", str, p);
      free(str);
      free(t);
    }
    else { free(path); path = str; }
  }
  return path;
}

static int StringGet(Obj self, char *name, int flags,
		     Type *type_in_out, Obj *value_return)
{
  ObjDir od = (ObjDir)self->value;
  char *subname;
  int status;
  struct stat statinfo;

  if(name[0] == '_') {
    if(!strcmp(name+1, "path")) {
      BOA_TYPECHECK_GET(T_String);
      if(value_return) *value_return = Boa_String_Create(strdup(od->path));
      return TPM_OK;
    }
    else if(!strcmp(name+1, "formatter")) {
      if(value_return) {
	Boa_Hold(od->formatter);
	*value_return = od->formatter;
      }
      return TPM_OK;
    }
    else if(!strcmp(name+1, "index")) {
      Obj index;
      DIR *dir = opendir(od->path);
      struct dirent *ent;
      int i = 0;
      if(dir == NULL) return TPM_ERROR;
      index = Boa_Create(T_Index);
      while(ent = readdir(dir)) {
	if(!strcmp(ent->d_name, ".") || !strcmp(ent->d_name, "..")) continue;
	assert(Boa_IntSet(index, i, BOA_INSERT, 
			  Boa_String_Create(strdup(ent->d_name))) == TPM_OK);
	i++;
      }
      closedir(dir);
      *value_return = index;
      return TPM_OK;
    }
    /* fall through */
  }
  /* stdin? */
  else if(name[0] == '-' && !name[1]) {
    /* ignore delete */
    if(value_return) {
      return ReadFile(od->formatter, name, type_in_out, value_return);
    }
    return TPM_OK;
  }
  name = TranslatePath(name);

  if(flags & BOA_SPLIT) {
    status = Generic_SplitGet(self, name, flags, type_in_out, value_return);
    free(name);
    return status;
  }

  /* construct new name */
  subname = Allocate(strlen(od->path)+strlen(name)+2, char);
  sprintf(subname, "%s/%s", od->path, name);
  free(name);

  /* what type of file is it? */
  if(stat(subname, &statinfo)) { 
    ErrorSet("Cannot stat `%s'", subname); ErrorAppendErrno(); 
    goto error; 
  }
  if((statinfo.st_mode & S_IFMT) == S_IFDIR) {
    /* directory; create a new objdir */
#if VERBOSE
    fprintf(stderr, "objDir: reading directory `%s'\n", subname);
#endif
    /* BOA_TYPECHECK_GET(type) with goto error */
    if(type_in_out) {
      if(!*type_in_out) *type_in_out = T_ObjDir;
      else if(*type_in_out != T_ObjDir) goto error;
    }
    if(value_return) {
      ObjDir d;
      Obj t = Boa_Create(T_ObjDir);
      d = (ObjDir)t->value;
      free(d->path);
      d->path = subname;
      Boa_Hold(od->formatter);
      d->formatter = od->formatter;
      *value_return = t;
    }
#if 0
    if(flags & BOA_DELETE) {
      /* doesn't work if non-empty */
      if(rmdir(subname) == -1) perror("rmdir");
    }
#endif
    return TPM_OK;
  }

  /* otherwise assume it's a regular file */
  /* don't read if we're only probing */
  if(type_in_out || value_return)
    status = ReadFile(od->formatter, subname, type_in_out, value_return);
  else
    status = TPM_OK;
#if 0
  if(flags & BOA_DELETE) {
    if(unlink(subname) == -1) perror("unlink");
  }
#endif
  free(subname);
  return status;

 error:
  free(subname);
  return TPM_BOTTOM;
}

static int StringSet(Obj self, char *name, int flags, Obj value)
{
  ObjDir od = (ObjDir)self->value;
  char *subname;
  int found, status;
  struct stat statinfo;
  Obj file;

  /* check for stdout */
  if(name[0] == '-' && !name[1]) subname = strdup(name);
  else {
    if(name[0] == '_') {
      if(!strcmp(name+1, "path")) {
	if(value->type != T_String) return TPM_BOTTOM;
	free(od->path);
	od->path = strdup(Boa_String_Value(value));
	Boa_Release(value);
	return TPM_OK;
      }
      else if(!strcmp(name+1, "formatter")) {
	Boa_Release(od->formatter);
	od->formatter = value;
	return TPM_OK;
      }
      /* fall through */
    }
    name = TranslatePath(name);

    if(flags & BOA_SPLIT) {
      status = Generic_SplitSet(self, name, flags, value);
      free(name);
      return status;
    }

    /* construct new name */
    subname = Allocate(strlen(od->path) + strlen(name) + 2, char);
    sprintf(subname, "%s/%s", od->path, name);
    free(name);

    found = !stat(subname, &statinfo);
    if(found && !(flags & BOA_REPLACE)) { ErrorString = BOA_EFLAGS; goto error; }
    if(!found && !(flags & BOA_INSERT)) { ErrorString = BOA_EFLAGS; goto error; }
  }
  /* write the Obj */
  file = FileHandle_Open(subname, "w");
  free(subname);
  if(!file) return TPM_BOTTOM;
  status = Boa_ObjSet(od->formatter, file, 0, value);
  Boa_Release(file);  /* closes fp */
  return status;

 error:
  free(subname);
  return TPM_BOTTOM;
}

static struct Boa_Methods methods = {
  Init, Final, Empty_Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  StringGet, StringSet,
  Err_IntGet, Err_IntSet,
  Generic_ObjGet, Generic_ObjSet
};

Obj Boa_ObjDir_Define(void)
{
  Boa_ObjDir_DefaultFormatter = Boa_Create(T_Merge);
  T_ObjDir = Boa_SimpleType_Create("objDir", &methods, 
				   sizeof(struct ObjDir),
				   sizeof(Ptr));
  return T_ObjDir;
}
