/* A small ordered list of named Objs.
 * Insertion adds to the end of the list, deletion shifts downward.
 * Items can be accessed via Boa_Ord_Item(ord, index);
 * the length via Boa_Ord_Length(ord).
 * Requires: Indexer, Iter, SimpleType.
 */
#include <boa/boa.h>

/* must keep these synchronized with the macros in std.h */
typedef struct {
  Obj obj;
  Obj name;
} Item;

typedef struct Ord {
  Item *items;
  int length;
} *Ord;

/* Globals *******************************************************************/
Type T_Ord;

/* Functions *****************************************************************/

static void Init(Obj self)
{
  Ord ord = (Ord)self->value;
  ord->items = Allocate(1,Item);
}

static void Final(Obj self)
{
  Ord ord = (Ord)self->value;
  int i;
  for(i=0;i<ord->length;i++) {
    Boa_Release(ord->items[i].obj);
    free(ord->items[i].name);
  }
  free(ord->items);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Ord ord = (Ord)self->value;
  int i;
  for(i=0;i<ord->length;i++) {
    func(ord->items[i].obj, data);
  }
}

#if 0
static int SPrint(Obj self, StrStream ss)
{
  Ord ord = (Ord)self->value;
  int i, space = 0;
  StrStreamPutc(ss, '{');
  for(i=0;i<ord->length;i++) {
    if(space) StrStreamPuts(ss, ", ");
    TRY_RETURN(Boa_SPrint(ord->items[i].name, ss));
    StrStreamPuts(ss, ": ");
    TRY_RETURN(Boa_SPrint(ord->items[i].obj, ss));
    space = 1;
  }
  return StrStreamPutc(ss, '}');
}
#else
/* Prints like a Tcl association list */
static int SPrint(Obj self, StrStream ss)
{
  Ord ord = (Ord)self->value;
  int i, space = 0;
  StrStreamPutc(ss, '{');
  for(i=0;i<ord->length;i++) {
    if(space) StrStreamPutc(ss, ' ');
    StrStreamPutc(ss, '{');
    TRY_RETURN(Boa_SPrint(ord->items[i].name, ss));
    StrStreamPutc(ss, ' ');
    TRY_RETURN(Boa_SPrint(ord->items[i].obj, ss));
    StrStreamPutc(ss, '}');
    space = 1;
  }
  return StrStreamPutc(ss, '}');
}
#endif

static int IntGet(Obj self, int index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Ord ord = (Ord)self->value;
  Obj t;
  if((index < 0) || (index >= ord->length)) return TPM_BOTTOM;
  t = ord->items[index].obj;
  BOA_TYPECHECK_GET(t->type);
  if(flags & BOA_DELETE) {
    Boa_Release(ord->items[index].name);
    if(value_return) *value_return = t;
    else Boa_Release(t);
    /* shift items down */
    ord->length--;
    for(;index < ord->length;index++)
      ord->items[index] = ord->items[index+1];
  }
  else if(value_return) {
    Boa_Hold(t);
    *value_return = t;
  }
  return TPM_OK;
}

static int IndexObj(Obj self, Obj key)
{
  Ord ord = (Ord)self->value;
  int i;
  for(i=0;i<ord->length;i++) {
    if(!Boa_Compare(key, ord->items[i].name)) break;
  }
  if(i == ord->length) return NAI;
  return i;
}
static Obj IndexInt(Obj self, int index)
{
  Ord ord = (Ord)self->value;
  if((index < 0) || (index >= ord->length)) return NULL;
  return Boa_Hold(ord->items[index].name);
}

static int IntSet(Obj self, int index, int flags, Obj value)
{
  Ord ord = (Ord)self->value;
  if((index < 0) || (index >= ord->length)) return TPM_BOTTOM;
  if(!(flags & BOA_REPLACE)) { ErrorString = BOA_EFLAGS; return TPM_ERROR; }
  Boa_Release(ord->items[index].obj);
  ord->items[index].obj = value;
  return TPM_OK;
}

static int ObjGet(Obj self, Obj index, int flags,
		     Type *type_in_out, Obj *value_return)
{
  Ord ord = (Ord)self->value;
  int i;

  if(index->type == T_String) {
    char *name = Boa_String_Value(index);
    if(name[0] == '_') {
      if(!strcmp(name+1, "length")) {
	BOA_TYPECHECK_GET(T_Int);
	if(value_return) *value_return = Boa_Int_Create(ord->length);
	return TPM_OK;
      }
#if 1
      else if(!strcmp(name+1, "iter")) {
	if(value_return) {
	  Obj iter = Boa_Create(T_Iter);
	  Obj indexer = Boa_Create(T_Indexer);
	  Boa_Indexer_Data(indexer) = Boa_Hold(self);
	  Boa_Indexer_ObjFunc(indexer) = IndexObj;
	  Boa_Indexer_IntFunc(indexer) = IndexInt;
	  Boa_Iter_Data(iter) = Boa_Hold(self);
	  Boa_Iter_Indexer(iter) = indexer;
	  *value_return = iter;
	}
	return TPM_OK;
      }
#endif
      /* should also have _index */
    }
    /* fall through */
  }

  for(i=0;i<ord->length;i++) {
    if(!Boa_Compare(index, ord->items[i].name)) {
      return IntGet(self, i, flags, type_in_out, value_return);
    }
  }
  return TPM_BOTTOM;
}

static int ObjSet(Obj self, Obj index, int flags, Obj value)
{
  Ord ord = (Ord)self->value;
  int i;
  for(i=0;i<ord->length;i++) {
    if(!Boa_Compare(index, ord->items[i].name)) break;
  }

  if(i == ord->length) {
    /* not found */
    if(!(flags & BOA_INSERT)) { ErrorString = BOA_EFLAGS; return TPM_BOTTOM; }
    /* create a new item at the end */
    index = Boa_Stabilize(index);
    if(!index) return TPM_ERROR;
    ord->length++;
    Reallocate(ord->items, ord->length, Item);
    ord->items[i].name = index;
    ord->items[i].obj = value;
    return TPM_OK;
  }
  else return IntSet(self, i, flags, value);
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Generic_StringGet, Generic_StringSet,
  IntGet, IntSet,
  ObjGet, ObjSet
};

Obj Boa_Ord_Define(void)
{
  T_Ord = Boa_SimpleType_Create("ord", &methods,
				sizeof(struct Ord),
				sizeof(Ptr));
  return T_Ord;
}
