/* The Parameterized Type type.
 * An Obj of type T_ParamType is a type-function, which when called yields
 * a new type.
 * Requires: Int, SimpleType.
 */
#include <boa/boa.h>

/* Globals *******************************************************************/
Type T_ParamType;

/* Convenience routines for instances ****************************************/

static void Instance_Final(Obj self)
{
  Boa_ParamTypeInstance data = (Boa_ParamTypeInstance)self->value;
  int i;
  for(i=0;i<data->num_args;i++) {
    Boa_Release(data->args[i]);
  }
  free(data->args);
}

static void Instance_Clean(Obj self, ObjFunc *func, Any d)
{
  Boa_ParamTypeInstance data = (Boa_ParamTypeInstance)self->value;
  int i;
  for(i=0;i<data->num_args;i++) {
    func(data->args[i], d);
  }
}

/* output a form that can be parsed back into this type */
static int Instance_SPrint(Obj self, StrStream ss)
{
  int i;
  Boa_ParamTypeInstance data = (Boa_ParamTypeInstance)self->value;
  Boa_ParamType tdata = (Boa_ParamType)self->type->value;

  StrStreamPutc(ss, '(');
  StrStreamPuts(ss, tdata->name);
  /* loop arguments */
  for(i=0;i<data->num_args;i++) {
    StrStreamPutc(ss, ' ');
    TRY_RETURN(Boa_SPrint(data->args[i], ss));
  }
  return StrStreamPutc(ss, ')');
}

static int Instance_Compare(Obj self, Obj b)
{
  Boa_ParamTypeInstance data = (Boa_ParamTypeInstance)self->value;
  Boa_ParamTypeInstance data2 = (Boa_ParamTypeInstance)b->value;
  int i, result = 0;
  for(i=0;i<data->num_args;i++) {
    if(result |= Boa_Compare(data->args[i], data2->args[i])) break;
  }
  return result;
}

static int Instance_Hash(Obj self)
{
  Boa_ParamTypeInstance data = (Boa_ParamTypeInstance)self->value;
  int i, result = 0;
  for(i=0;i<data->num_args;i++) {
    result += Boa_Hash(data->args[i]);
  }
  return result * 1103515245;
}

/* returns args[index] */
static int Instance_IntGet(Obj self, int index, int flags,
			   Type *type_in_out, Obj *value_return)
{
  Boa_ParamTypeInstance data = (Boa_ParamTypeInstance)self->value;
  if((index < 0) || (index >= data->num_args)) return TPM_BOTTOM;
  if(flags & BOA_DELETE) { ErrorString = BOA_EFLAGS; return TPM_ERROR; }
  BOA_TYPECHECK_GET(data->args[index]->type);
  if(value_return) *value_return = Boa_Hold(data->args[index]);
  return TPM_OK;
}

static struct Boa_Methods instance_methods = {
  Empty_Init, Instance_Final, Instance_Clean,
  Instance_SPrint, Generic_Copy, Instance_Compare, Instance_Hash,
  Err_StringGet, Err_StringSet,
  Instance_IntGet, Err_IntSet,
  Generic_ObjGet, Generic_ObjSet
};

Type Boa_ParamType_Create(char *name, Boa_InitFunc *init,
			  int num_args, ...)
{
  int i;
  va_list args;
  Boa_ParamType data = Allocate(1, struct Boa_ParamType);
  data->name = name;
  data->methods = AllocCopy(1, struct Boa_Methods, &instance_methods);
  data->methods->init = init;
  data->size = sizeof(struct Boa_ParamTypeInstance);
  data->align = sizeof(Ptr);
  data->num_args = num_args;
  data->known = AllocateC(num_args, Obj);
  data->unknown = Allocate(num_args, Type);
  va_start(args, num_args);
  for(i=0;i<num_args;i++) data->unknown[i] = va_arg(args, Type);
  va_end(args);
  data->parent = Boa_CreateFromData(T_ParamType, data);
  return data->parent;
}

/* Functions *****************************************************************/

static int SPrint(Obj self, StrStream ss)
{
  int i;
  Boa_ParamType data = (Boa_ParamType)self->value;
  StrStreamPuts(ss, "(ptype [");
  for(i=0;i<data->num_args;i++) {
    if(!data->known[i]) break;
    Boa_SPrint(data->known[i], ss);
    StrStreamPutc(ss, ' ');
  }
  StrStreamPuts(ss, "] ");
  for(;i<data->num_args;i++) {
    if(data->unknown[i]) Boa_SPrint(data->unknown[i], ss);
    else StrStreamPuts(ss, "T");
    StrStreamPutc(ss, ' ');
  }
  return StrStreamPutc(ss, ')');
}

static void Final(Obj self)
{
  Boa_ParamType data = (Boa_ParamType)self->value;
  int i;
  for(i=0;i<data->num_args;i++) {
    if(data->known[i]) Boa_Release(data->known[i]);
    if(data->unknown[i]) Boa_Release(data->unknown[i]);
  }
  free(data->known);
  free(data->unknown);
}

static void Clean(Obj self, ObjFunc *func, Any d)
{
  Boa_ParamType data = (Boa_ParamType)self->value;
  int i;
  for(i=0;i<data->num_args;i++) {
    if(data->known[i]) func(data->known[i], d);
    if(data->unknown[i]) func(data->unknown[i], d);
  }
}

#if 0
/* Expects to get data->num_args parameters of the types given by data->args.
 */
static int Call(Obj self, int nargs, int nrets, Obj *args)
{
  int i;
  Boa_ParamType data = (Boa_ParamType)self->value;
  Boa_ParamTypeInstance new_data;
  Obj t;
  if((nargs != data->num_args) || (nrets != 1)) goto error;
  for(i=0;i<nargs;i++) {
    if(data->args[i] && Boa_Compare(data->args[i], args[i]->type))
      goto error;
  }
  /* make a new instance, which takes over ownership of the args */
  new_data = Allocate(1, struct Boa_ParamTypeInstance);
  new_data->num_args = nargs;
  new_data->args = AllocCopy(nargs, Obj, args);
  t = Boa_CreateFromData(self, new_data);
  Boa_Init(t); /* this will set methods, size, and align */
  args[0] = t;
  return TPM_OK;

 error:
  for(i=0;i<nargs;i++) Boa_Release(args[i]);
  return TPM_ERROR;
}
#endif

/* Fill in an unknown.  If all unknowns have been filled, return a new 
 * ParamTypeInstance.  Otherwise return a new partially-instantiated ParamType.
 */
static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Boa_ParamType data = (Boa_ParamType)self->value, new_data;
  int i,j;
  Obj t;

  /* set i to the position we want to fill */
  for(i=0;i<data->num_args;i++) if(!data->known[i]) break;
  assert(i < data->num_args);
  /* check type */
  if(data->unknown[i] && Boa_Compare(data->unknown[i], index->type))
    return TPM_BOTTOM;

  /* last argument? */
  if(i == data->num_args - 1) {
    Boa_ParamTypeInstance instance_data;
    instance_data = Allocate(1, struct Boa_ParamTypeInstance);
    instance_data->num_args = data->num_args;
    instance_data->args = AllocCopy(data->num_args, Obj, data->known);
    instance_data->args[i] = index;
    for(j=0;j<data->num_args;j++) Boa_Hold(instance_data->args[j]);
    t = Boa_CreateFromData(data->parent, instance_data);
    Boa_Init(t); /* this will set methods, size, and align */
    if(value_return) *value_return = t;
    else Boa_Release(t);
    return TPM_OK;
  }

  new_data = Allocate(1, struct Boa_ParamType);
  *new_data = *data;
  new_data->unknown = AllocCopy(data->num_args, Type, data->unknown);
  new_data->known = AllocCopy(data->num_args, Obj, data->known);
  new_data->known[i] = index;
  for(j=0;j<data->num_args;j++) {
    if(new_data->unknown[j]) Boa_Hold(new_data->unknown[j]);
    if(new_data->known[j]) Boa_Hold(new_data->known[j]);
  }
  new_data->parent = data->parent;
  t = Boa_CreateFromData(self->type, new_data);
  if(value_return) *value_return = t;
  else Boa_Release(t);
  return TPM_OK;
}

static struct Boa_Methods methods = {
  Empty_Init, Final, Clean,
  SPrint, Err_Copy, Identity_Compare, Generic_Hash,
  Err_StringGet, Err_StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, Generic_ObjSet
};

Obj Boa_ParamType_Define(void)
{
  T_ParamType = Boa_SimpleType_Create("ptype", &methods, 
				      sizeof(struct Boa_ParamType),
				      sizeof(Ptr));
  return T_ParamType;
}
