/* Boa adapter for the SpanTree data type.
 * A SpanTree is a compact representation of a spanning tree with float edge
 * costs.
 * Requires: Int, ParamType.
 */
#include <boa/boa.h>
#include <tpm/spanTree.h>

#define length Boa_Int_Value(Boa_PType_Item(self->type, 0))

/* Globals *******************************************************************/
Type T_SpanTree = NULL;

/* Functions *****************************************************************/

static void Init(Obj self)
{
  SpanTree st = (SpanTree)self->value;
  /* taken from SpanTreeCreate */
  st->num_nodes = length;
  st->node = Allocate(st->num_nodes, struct SpanTreeNode);
}

static void Final(Obj self)
{
  SpanTree st = (SpanTree)self->value;
  /* taken from SpanTreeFree */
  free(st->node);
}

static int SPrint(Obj self, StrStream ss)
{
  return Boa_SPrint(self->type, ss);
}

static int IntGet(Obj self, int index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  SpanTree st = (SpanTree)self->value;
  if((index < 0) || (index >= length)) return TPM_ERROR;
  /* length is immutable */
  if(flags & BOA_DELETE) { ErrorString = BOA_EFLAGS; return TPM_ERROR; }
  BOA_TYPECHECK_GET(T_Float);
  if(value_return) {
    *value_return = Boa_Float_Create(st->node[index].edge);
  }
  return TPM_OK;
}

static int IndexObj(Obj self, Obj key)
{
  SpanTree st = (SpanTree)self->value;
  int i;
  Obj t;
  /* if types don't match, we're done */
  Boa_PrintNL(stderr, key);
  if(Boa_Compare(T_Float, key->type)) return NAI;
  t = Boa_CreateFromData(T_Float, NULL);
  for(i=0;i<st->num_nodes;i++) {
    t->value = &st->node[i].edge;
    if(!Boa_Compare(t, key)) {
      Boa_Release(t);
      return i;
    }
  }
  Boa_Release(t);
  return NAI;
}
static Obj IndexInt(Obj self, int index)
{
  SpanTree st = (SpanTree)self->value;
  Obj t;
  if((index < 0) || (index >= length)) return NULL;
  return Boa_Float_Create(st->node[index].edge);
}

/* convert name to integer and forward to IntGet */
static int StringGet(Obj self, char *name, int flags,
		     Type *type_in_out, Obj *value_return)
{
  int index;
  if(name[0] == '_') {
    if(!strcmp(name+1, "length")) {
      return Boa_IntGet(self->type, 0, flags, type_in_out, value_return);
    }
    else if(!strcmp(name+1, "index")) {
      Obj t = Boa_Create(T_Indexer);
      Boa_Indexer_Data(t) = Boa_Hold(self);
      Boa_Indexer_ObjFunc(t) = IndexObj;
      Boa_Indexer_IntFunc(t) = IndexInt;
      *value_return = t;
      return TPM_OK;
    }
  }
  index = atoi(name);
  if(!index && (name[0] != '0')) return TPM_BOTTOM;
  return IntGet(self, index, flags, type_in_out, value_return);
}

static struct Boa_Methods methods = {
  Init, Final, Empty_Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  StringGet, Err_StringSet,
  IntGet, Err_IntSet,
  Generic_ObjGet, Generic_ObjSet
};

/* assumes num_args and args have already been filled in */
static void Type_Init(Obj self)
{
  Boa_ParamTypeInstance data = (Boa_ParamTypeInstance)self->value;
  data->methods = &methods;
  data->size = sizeof(struct SpanTree);
  data->align = sizeof(Ptr);
}

Obj Boa_SpanTree_Define(void)
{
  T_SpanTree = Boa_ParamType_Create("spanTree", Type_Init, 1, T_Int);
  return T_SpanTree;
}
