/* Adapts a StrStream to the Boa Stream protocol.
 * Requires: Ptr, Float, Double, Int, String, Char, SimpleType.
 */
#include <boa/boa.h>

typedef struct Stream {
  StrStream ss;
} *Stream;

typedef struct Seek {
  Obj stream;
} *Seek;

/* Globals *******************************************************************/
Type T_StrStream;
static Type T_StrSeek;

/* Prototypes ****************************************************************/

Obj Boa_StrStream_FromString(char *string);

/* Functions *****************************************************************/

Obj Boa_StrStream_FromString(char *string)
{
  Obj t = Boa_Create(T_StrStream);
  StrStreamFree(Boa_StrStream_Value(t));
  Boa_StrStream_Value(t) = StrStreamFromString(string, strlen(string)+1, 0);
  return t;
}

static void Init(Obj self)
{
  Stream s = (Stream)self->value;
  s->ss = StrStreamCreate(16, STRSTREAM_UNLIMITED);
}

static void Final(Obj self)
{
  Stream s = (Stream)self->value;
  StrStreamFree(s->ss);
}

static int SPrint(Obj self, StrStream ss)
{
  return StrStreamPuts(ss, "aStrStream");
}

/* mimic of FileHandle::IntGet */
static int IntGet(Obj self, int index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Stream s = (Stream)self->value;
  if(type_in_out) {
    if(*type_in_out == T_String) {
      if(value_return) {
	char *data = Allocate(index+1, char);
	if(StrStreamGet(s->ss, data, index) != TPM_OK) {
	  free(data); return TPM_BOTTOM;
	}
	data[index] = 0;
	*value_return = Boa_String_Create(data);
      }
      else if(StrStreamSeekR(s->ss, index, SEEK_CUR) != TPM_OK) {
	return TPM_BOTTOM;
      }
      return TPM_OK;
    }
    else if(*type_in_out == T_Int) {
      int i;
      if(StrStreamGetInt(s->ss, &i) != TPM_OK) return TPM_BOTTOM;
      if(value_return) *value_return = Boa_Int_Create(i);
      return TPM_OK;
    }
    else if(*type_in_out == T_Float) {
      double d;
      if(StrStreamGetDouble(s->ss, &d) != TPM_OK) return TPM_BOTTOM;
      if(value_return) *value_return = Boa_Float_Create((float)d);
      return TPM_OK;
    }
    else if(*type_in_out == T_Double) {
      double d;
      if(StrStreamGetDouble(s->ss, &d) != TPM_OK) return TPM_BOTTOM;
      if(value_return) *value_return = Boa_Double_Create(d);
      return TPM_OK;
    }
    else if(*type_in_out != T_Char) return TPM_BOTTOM;
  }
  /* reading T_Char */
  BOA_TYPECHECK_GET(T_Char);
  if(value_return) {
    char c;
    if(StrStreamGetc(s->ss, &c) != TPM_OK) return TPM_BOTTOM;
    if(index != 1) {
      if(StrStreamSeekR(s->ss, index-1, SEEK_CUR) != TPM_OK) return TPM_BOTTOM;
    }
    *value_return = Boa_Char_Create(c);
  }
  else if(index) {
    if(StrStreamSeekR(s->ss, index, SEEK_CUR) != TPM_OK) return TPM_BOTTOM;
  }
  else if(StrStreamEmpty(s->ss)) return TPM_BOTTOM;
  return TPM_OK;
}

static int IntSet(Obj self, int index, int flags, Obj value)
{
  Stream s = (Stream)self->value;
  if(value->type == T_Char) {
    if(StrStreamPutc(s->ss, Boa_Char_Value(value)) != TPM_OK) return TPM_BOTTOM;
    Boa_Release(value);
  }
  else if(value->type == T_String) {
    if(StrStreamPuts(s->ss, Boa_String_Value(value)) != TPM_OK) return TPM_BOTTOM;
    Boa_Release(value);
  }
  else if(value->type == T_Int) {
    if(StrStreamPutInt(s->ss, Boa_Int_Value(value)) != TPM_OK) return TPM_BOTTOM;
    Boa_Release(value);
  }
  else if(value->type == T_Float) {
    if(StrStreamPutDouble(s->ss, (double)Boa_Float_Value(value)) != TPM_OK) return TPM_BOTTOM;
    Boa_Release(value);
  }
  else if(value->type == T_Double) {
    if(StrStreamPutDouble(s->ss, Boa_Double_Value(value)) != TPM_OK) return TPM_BOTTOM;
    Boa_Release(value);
  }
  else if(value->type == T_Ptr) {
    if(StrStreamPut(s->ss, Boa_Ptr_Value(value), index) != TPM_OK) return TPM_BOTTOM;
    Boa_Release(value);
  }
  else return TPM_BOTTOM;
  return TPM_OK;
}

static int StringGet(Obj self, char *name, int flags,
		     Type *type_in_out, Obj *value_return)
{
  Stream s = (Stream)self->value;
  if(!strcmp(name, "seek")) {
    if(value_return) {
      Obj t = Boa_Create(T_StrSeek);
      Boa_Hold(self);
      ((Seek)t->value)->stream = self;
      *value_return = t;
    }
    return TPM_OK;
  }
  else if(!strcmp(name, "tell")) {
    BOA_TYPECHECK_GET(T_Int);
    if(value_return) *value_return = Boa_Int_Create(s->ss->rpos);
    return TPM_OK;
  }
  else if(!strcmp(name, "line")) {
    char *data;
    BOA_TYPECHECK_GET(T_String);
    data = StrStreamGetLine(s->ss);
    if(!data) return TPM_BOTTOM;
    if(value_return) *value_return = Boa_String_Create(data);
    return TPM_OK;
  }
  else if(!strcmp(name, "word")) {
    char *data;
    BOA_TYPECHECK_GET(T_String);
    data = StrStreamGetWord(s->ss);
    if(!data) return TPM_BOTTOM;
    if(value_return) *value_return = Boa_String_Create(data);
    return TPM_OK;
  }
  else return TPM_BOTTOM;
}

static struct Boa_Methods methods = {
  Init, Final, Empty_Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  StringGet, Err_StringSet,
  IntGet, IntSet,
  Generic_ObjGet, Generic_ObjSet
};

/* Seek object ***************************************************************/

static int SeekSPrint(Obj self, StrStream ss)
{
  return StrStreamPuts(ss, "seek(aStrStream)");
}

static int SeekIntGet(Obj self, int index, int flags,
		      Type *type_in_out, Obj *value_return)
{
  Seek seek = (Seek)self->value;
  Stream s = (Stream)seek->stream->value;
  if(StrStreamSeekR(s->ss, index, 0)) return TPM_BOTTOM;
  return IntGet(seek->stream, 0, flags, type_in_out, value_return);
}
static int SeekIntSet(Obj self, int index, int flags, Obj value)
{
  Seek seek = (Seek)self->value;
  Stream s = (Stream)seek->stream->value;
  if(StrStreamSeekR(s->ss, index, 0)) return TPM_BOTTOM;
  return IntSet(seek->stream, 0, flags, value);
}

static struct Boa_Methods seek_methods = {
  Empty_Init, Sub_Final, Sub_Clean,
  SeekSPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Err_StringGet, Err_StringSet,
  SeekIntGet, SeekIntSet,
  Generic_ObjGet, Generic_ObjSet
};

Obj Boa_StrStream_Define(void)
{
  T_StrSeek = Boa_SimpleType_Create("strSeek", &seek_methods,
				    sizeof(struct Seek),
				    sizeof(Ptr));
  T_StrStream = Boa_SimpleType_Create("strStream", &methods,
				      sizeof(struct Stream),
				      sizeof(Ptr));
  return T_StrStream;
}
