/* Tcl context. Interfaces to a Tcl_Interp.
 * Get reads the (string) value of a Tcl variable.
 * Set writes a string value to a Tcl variable.
 * Attributes:
 *   interp     Tcl_Interp
 *   command	Tcl command to execute
 *   call	Executes _command in Tcl
 *   status     Tcl status code of last command
 *   result     Tcl result of last command
 */
#include <boa/boa.h>
#include <boa/tcl.h>

typedef struct Tcl {
  Tcl_Interp *interp;
  char *command;
  int status;
} *Tcl;

/* Globals *******************************************************************/
Type T_Tcl = NULL;
static char *initCmd = "";

/* Functions *****************************************************************/

static void Init(Obj self)
{
  Tcl tcl = (Tcl)self->value;
  tcl->interp = Tcl_CreateInterp();
  if(Tcl_Init(tcl->interp) == TCL_ERROR) {
    fprintf(stderr, "Tcl_Init failed\n");
  }
  Tcl_CreateCommand(tcl->interp, "boa", Boa_Cmd, NULL, NULL);
  Tcl_PkgProvide(tcl->interp, "boa", "1");
  if(Tcl_Eval(tcl->interp, initCmd) != TCL_OK) {
    fprintf(stderr, "Tcl interpreter initialization `%s' failed\n", initCmd);
  }
  tcl->command = strdup("");
  tcl->status = TCL_OK;
}

static void Final(Obj self)
{
  Tcl tcl = (Tcl)self->value;
  Tcl_DeleteInterp(tcl->interp);
  free(tcl->command);
}

static int SPrint(Obj self, StrStream ss)
{
  return StrStreamPuts(ss, "aTcl");
}

#if 0
static int Eval(Type type, Tcl tcl)
{
  StrStream ss = StrStreamCreate(16, STRSTREAM_UNLIMITED);
  if(!tcl->interp) return TPM_ERROR;
  /* evaluate _command at the global level */
  StrStreamPuts(ss, "uplevel #0 {");
  StrStreamPuts(ss, tcl->command);
  StrStreamPuts(ss, "}");
  tcl->status = Tcl_Eval(tcl->interp, StrStreamAsString(ss));
  StrStreamFree(ss);
  return TPM_OK;
}
#endif

static int StringGet(Obj self, char *name, int flags,
		     Type *type_in_out, Obj *value_return)
{
  Tcl tcl = (Tcl)self->value;
  char *str;

  /* attribute */
  if(name[0] == '_') {
    if(!strcmp(name+1, "result")) {
      /* return the result of the last command */
      BOA_TYPECHECK_GET(T_String);
      if(!tcl->interp) { ErrorString = BOA_EINIT; return TPM_BOTTOM; }
      if(value_return) {
	*value_return = Boa_String_Create(strdup(tcl->interp->result));
      }
      return TPM_OK;
    }
#if 0
    else if(!strcmp(name+1, "call")) {
      /* return the command evaluation procedure */
      CallFunc *proc = (CallFunc*)Eval;
      BOA_TYPECHECK_GET(T_Proc);
      BOA_ITEM_GET(T_Proc, &proc);
      return TPM_OK;
    }
#endif
    else if(!strcmp(name+1, "status")) {
      /* return the return status of the last command */
      BOA_TYPECHECK_GET(T_Int);
      if(value_return) *value_return = Boa_Int_Create(tcl->status);
      return TPM_OK;
    }
    else if(!strcmp(name+1, "command")) {
      /* return the last command */
      BOA_TYPECHECK_GET(T_String);
      if(value_return) {
	*value_return = Boa_String_Create(strdup(tcl->command));
      }
      return TPM_OK;
    }
    else if(!strcmp(name+1, "interp")) {
      BOA_TYPECHECK_GET(T_Ptr);
      if(value_return) *value_return = Boa_Ptr_Create(tcl->interp);
      return TPM_OK;
    }
    /* todo: _iter */
  }
  
  if(!tcl->interp) { ErrorString = BOA_EINIT; return TPM_BOTTOM; }
  BOA_TYPECHECK_GET(T_String);
  /* read the Tcl variable with the given name */
  str = Tcl_GetVar(tcl->interp, name, TCL_GLOBAL_ONLY);
  if(!str) return TPM_BOTTOM;
  /* return a copy so that Tcl mutations will not be seen */
  if(value_return) *value_return = Boa_String_Create(strdup(str));
  return TPM_OK;
}

static int StringSet(Obj self, char *name, int flags, Obj value)
{
  Tcl tcl = (Tcl)self->value;
  int status;

  /* attribute */
  if(name[0] == '_') {
    if(!strcmp(name+1, "command")) {
      if(value->type != T_String) return TPM_BOTTOM;
      free(tcl->command);
      tcl->command = strdup(Boa_String_Value(value));
      Boa_Release(value);
      return TPM_OK;
    }
    else if(!strcmp(name+1, "interp")) {
      if(value->type != T_Ptr) return TPM_BOTTOM;
      /* does not delete old interp */
      tcl->interp = Boa_Ptr_Value(value);
      Boa_Release(value);
      return TPM_OK;
    }
  }
  
  if(!tcl->interp) return TPM_ERROR;
  /* write the Tcl variable with the given name */
  if(value->type != T_String) return TPM_BOTTOM;
  status = !!Tcl_SetVar(tcl->interp, name, Boa_String_Value(value), 0);
  Boa_Release(value);
  return status ? TPM_OK : TPM_BOTTOM;
}

static struct Boa_Methods methods = {
  Init, Final, Empty_Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  StringGet, StringSet,
  Err_IntGet, Err_IntSet,
  Generic_ObjGet, Generic_ObjSet
};

Obj Boa_Tcl_Define(void)
{
  T_Tcl = Boa_SimpleType_Create("tcl", &methods,
				sizeof(struct Tcl),
				sizeof(Ptr));
  return T_Tcl;
}
