/* EM/K-means clustering
 * Adapted from Kris Popat.
 * Can use diagonal or full covariance matrix.
 */
#include <tpm/cluster.h>

#define VERBOSE 0
/* eigenvalues of covariance matrix cannot go below LAMBDA_FLOOR */
#define LAMBDA_FLOOR 1e-10

/* Prototypes ****************************************************************/

double Clus_EM(Matrix sample, Matrix center, Matrix cov,
	       Matrix mbr, double temperature,
	       char *center_file, char *icov_file, char *mbr_file);
double Clus_EMStepE(Matrix sample, Matrix center, Matrix cov,
		    Matrix mbr, double temperature);
double Clus_EMStepEP(Matrix sample, Matrix center, Matrix cov, Vector prior,
		    Matrix mbr, double temperature);
void Clus_EMStepM(Matrix sample, Matrix center, Matrix mbr);
void Clus_EMStepV(Matrix sample, Matrix center, Matrix cov, Matrix mbr);
void Clus_EMStepP(Vector prior, Matrix mbr);
void VectorPerturb(Vector v, double noise_amp);
void Clus_EMInitialize(Matrix cov);
void Clus_EMClassify(Matrix center, Matrix cov, Vector prior,
		     double temperature,
		     int freq, int scale, Matrix output);
void Clus_EMDistance(Matrix center, Matrix cov, Vector prior,
		     double temperature,
		     Vector point, int scale, Matrix output);

/* Functions *****************************************************************/

void Clus_EMInitialize(Matrix cov)
{
  int i,j,n;
  VectorSet((Vector)cov, 0);
  n = sqrt(cov->cols);
  for(i=0;i<cov->rows;i++) {
    for(j=0;j<n;j++) {
      cov->d2[i][j*n+j] = 1;
    }
  }
}

/* Adds uniformly distributed noise from [-noise_amp..noise_amp] to all 
 * elements of v.
 * Modifies: v.
 */
void VectorPerturb(Vector v, double noise_amp)
{
  int i;
  for(i=0;i<v->len;i++) v->d[i] += 2 * (Rand() - 0.5) * noise_amp;
}

/* Clustering via EM.
 * sample is a matrix of row vectors.
 * center is a matrix of rows of the same dimensionality as sample.
 *
 * Each row of cov is the covariance for a component.
 * The covariance is stored as the concatenation of the eigenvectors,
 * weighted by the inverse root of the eigenvalues.
 * It should be initialized with Clus_EMInitialize().
 *
 * temperature is the annealing parameter.  
 *   1 corresponds to conventional EM
 *   0 corresponds to K-means
 *
 * center_file, if non-NULL, stores the centers after every iteration.
 * cov_file stores the eigenvectors and
 * mbr_file stores the assignment matrix.
 *
 * Modifies: center, cov, mbr 
 */
double Clus_EM(Matrix sample, Matrix center, Matrix cov,
	       Matrix mbr, double temperature,
	       char *center_file, char *cov_file, char *mbr_file)
{
  FILE *fp;
  double distortion = 1e12, old_distortion;
  do {
    /* compute centers */
    Clus_EMStepM(sample, center, mbr);
    /* perturb centers */
    VectorPerturb((Vector)center, 0.01);
#if 0
    /* compute covariances */
    Clus_EMStepV(sample, center, cov, mbr);
#endif

    /* compute membership */
    old_distortion = distortion;
    distortion = Clus_EMStepE(sample, center, cov, mbr, temperature);

    if(center_file && (fp = fopen(center_file, "w"))) {
      MatrixPrint(fp, center);
      fclose(fp);
    }
    if(cov_file && (fp = fopen(cov_file, "w"))) {
      MatrixPrint(fp, cov);
      fclose(fp);
    }
    if(mbr_file && (fp = fopen(mbr_file, "w"))) {
      MatrixPrint(fp, mbr);
      fclose(fp);
    }
    /* have to use fabs since distortion may momentarily increase */
  } while(fabs(old_distortion - distortion) > 1e-1*temperature);
  return distortion;
}

/* An alternative to Mahalanobis */
static double Radial(double *s, double *c, double *v, int n)
{
  double r = 0;
  int i;
  /* compute the radius of s and subtract the radius of c */
  for(i=0;i<n;i++) {
    r += (s[i] - 200) * (s[i] - 200);
    r -= (c[i] - 200) * (c[i] - 200);
  }
  return fabs(r);
}

#define LOGTWOPI   1.8378770664093454

/* Rotates s and c by n*n matrix v before computing distance. */
static double Mahalanobis(double *s, double *c, double *v, int n)
{
  double sum = 0;
  int i,j;

  /* Loop eigenvectors */
  for(i=0;i<n;i++) {
    double d = 0;
    double mag = 0;
    /* take dot product of s and c with the current eigenvector (v) 
     * and subtract the results.
     */
    for(j=0;j<n;j++) {
      d += (s[j] - c[j]) * v[j];
      mag += v[j]*v[j];
    }
    sum += d*d;
    /* mag is now the inverse eigenvalue, i.e. the inverse variance.
     * subtract the log to take care of the inversion.
     */
    sum += LOGTWOPI - log(mag);
    /* advance to next eigenvector */
    v += n;
  }
  return sum/2;
}

/* An alternative to Mahalanobis */
static double WeightedEuclidean(double *s, double *c, double *v, int n)
{
  double sum = 0;
  int i;
  for(i=0;i<n;i++) {
    double d = s[i] - c[i];
    sum += d*d/v[i];
    sum += LOGTWOPI + 2 * log(v[i]);
  }
  return sum/2;
}

static double Euclidean(double *s, double *c, int n)
{
  double sum = 0;
  int i;
  for(i=0;i<n;i++) {
    double d = s[i] - c[i];
    sum += d*d;
  }
  return sum;
}

/* Calls Clus_EMStepEP with uniform mixing parameters. */
double Clus_EMStepE(Matrix sample, Matrix center, Matrix cov,
		    Matrix mbr, double temperature)
{
  double distortion;
  Vector prior = VectorCreate(center->rows);
  VectorSet(prior, 1.0/prior->len);
  distortion = Clus_EMStepEP(sample, center, cov, prior, mbr, temperature);
  VectorFree(prior);
  return distortion;
}

/* Computes the membership of each sample in each center, storing it in mbr.
 * The rows of mbr sum to 1.
 * Returns the total weighted distance from samples to centers, 
 * i.e. the distortion of the clustering.
 * Modifies: mbr.
 */
double Clus_EMStepEP(Matrix sample, Matrix center, Matrix cov, Vector prior,
		     Matrix mbr, double temperature)
{
  int s, c;
  double distortion = 0;
  double likelihood = 0;

  /* compute distortion */
  for(s=0; s < sample->rows; s++) {
    double center_sum = 0, distortion_subtotal = 0;
    double likelihood_subtotal = 0;
    double closest_d = 1.0/0;
    int closest = 0;
#if VERBOSE
    if(s % 100 == 0) fprintf(stderr, "computing distortion for %d\n", s);
#endif
    for(c=0; c < center->rows; c++) {
      double d = Mahalanobis(sample->d2[s], center->d2[c], cov->d2[c], 
			     sample->cols);
#if 0
      /* flip a coin and set d to infinity */
      if(Rand() > 0.99) d = 1.0/0;
#endif
      if(!finite(d)) {
	mbr->d2[s][c] = 0;
      }
      else {
	mbr->d2[s][c] = exp(-d/temperature) * prior->d[c];
	center_sum += mbr->d2[s][c];
	distortion_subtotal += mbr->d2[s][c] * d;
	likelihood_subtotal += exp(-d) * prior->d[c];
	/* prepare the backup plan */
	if(d < closest_d) { closest = c; closest_d = d; }
      }
    }
    if(center_sum > 0.0) {
      distortion += distortion_subtotal / center_sum;
      for(c=0; c < center->rows; c++) {
	mbr->d2[s][c] /= center_sum;
      }
      if(likelihood_subtotal > 0) likelihood -= log(likelihood_subtotal);
      else likelihood = -1.0/0;
    }
    else {
      /* use the backup plan: closest cluster only */
      for(c=0; c < center->rows; c++) {
	if(c == closest) mbr->d2[s][c] = 1;
	else mbr->d2[s][c] = 0;
      }
      distortion += closest_d;
      likelihood += closest_d - log(prior->d[closest]);
    }
  }
#if VERBOSE
  fprintf(stderr, "distortion: %lg\n", distortion);
#endif
#if 1
  return likelihood;
#else
  return distortion;
#endif
}

/* Sets center[c] to be the weighted mean of the samples,
 * with weights given by mbr[:][c].
 */
void Clus_EMStepM(Matrix sample, Matrix center, Matrix mbr)
{
  int s, c, i;
  /* loop clusters and compute new mean */
  VectorSet((Vector)center, 0);
  for(c=0; c < center->rows; c++) {
    double sample_sum = 0;
#if VERBOSE
    if(c % 10 == 0) fprintf(stderr, "updating center %d mean\n", c);
#endif
    for(s=0; s < sample->rows; s++) {
      sample_sum += mbr->d2[s][c];
      for(i=0; i < center->cols; i++) {
	center->d2[c][i] += mbr->d2[s][c] * sample->d2[s][i];
      }
    }
    if(sample_sum > 0.0) {
      for(i=0; i < center->cols; i++) {
	center->d2[c][i] /= sample_sum;
      }
    }
  }
}

/* Modify cov (size rank by rank) to be its eigenvectors scaled 
 * by the inverse root eigenvalues.
 */
static void Cov2EV(double *cov, int rank)
{
  Matrix m = MatrixFromData(rank, rank, cov); /* eigenvectors */
  double *v = Allocate(rank, double); /* eigenvalues */
  int i, j;

  /* extract the eigenvectors of the covariance matrix */
  MatrixEigenvectorsSym(m, v);
  /* take the root of the eigenvalues */
  for(i=0; i < rank; i++) {
    /* this flooring is related to Baback's PhD */
    if(v[i] < LAMBDA_FLOOR) v[i] = LAMBDA_FLOOR;
    else v[i] = sqrt(v[i]);
  }
  /* scale the eigenvectors by the inverse eigenvalues */
  for(i=0; i < rank; i++) {
    for(j=0; j < rank; j++) {
      m->d2[i][j] /= v[i];
    }
  }

  free(v);
  free(m->d2);
  free(m);
}

/* Modify cov (size rank by rank) to be a regular inverse
 * covariance matrix. 
 */
static void EV2Cov(double *cov, int rank)
{
  Matrix m = MatrixFromData(rank, rank, cov); /* eigenvectors */
  MatrixTransposeTimesSelf(m, m);
  free(m->d2);
  free(m);
}

/* Sets cov[c][d] to be the weighted variance of the samples
 * along dimension d, with weights given by mbr[:][c].
 */
void Clus_EMStepV(Matrix sample, Matrix center, Matrix cov, Matrix mbr)
{
  int s, c, i, j;
  /* loop clusters and compute new variance */
  VectorSet((Vector)cov, 0);
  for(c=0; c < center->rows; c++) {
    double sample_sum = 0;
#if VERBOSE
    if(c % 10 == 0) fprintf(stderr, "updating center %d variance\n", c);
#endif
    for(s=0; s < sample->rows; s++) {
      sample_sum += mbr->d2[s][c];
      for(i=0; i < center->cols; i++) {
	double di = sample->d2[s][i] - center->d2[c][i];
	for(j=i; j < center->cols; j++) {
	  double dj = sample->d2[s][j] - center->d2[c][j];
	  cov->d2[c][i*center->cols+j] += mbr->d2[s][c] * di * dj;
	  cov->d2[c][j*center->cols+i] = cov->d2[c][i*center->cols+j];
	}
      }
    }
    if(sample_sum > 0.0) {
      for(i=0; i < cov->cols; i++) {
	cov->d2[c][i] /= sample_sum;
      }
    }
    Cov2EV(cov->d2[c], center->cols);
  }
}

/* Sets prior to be the new mixing proportions given assignments mbr. */
void Clus_EMStepP(Vector prior, Matrix mbr)
{
  int s, c;
  double sum = 0;
  for(c=0; c < mbr->cols; c++) {
    double sample_sum = 0;
    for(s=0; s < mbr->rows; s++) {
      sample_sum += mbr->d2[s][c];
    }
    sum += sample_sum;
    prior->d[c] = sample_sum;
  }
  VectorScale(prior, 1.0/sum);
}

/* Compute the soft assignment of sample to the centers.
 * Modifies: mbr 
 */
static void Classify(Matrix center, Matrix cov, Vector prior,
		     double temperature, Vector sample, double *mbr)
{
  double center_sum = 0;
  double closest_d = 1.0/0;
  int closest = 0;
  int c;
  for(c=0;c<center->rows;c++) {
    double d = Mahalanobis(sample->d, center->d2[c], cov->d2[c], sample->len);
    if(!finite(d)) {
      mbr[c] = 0;
    }
    else {
      mbr[c] = exp(-d/temperature) * prior->d[c];
      center_sum += mbr[c];
      /* prepare the backup plan */
      if(d < closest_d) { closest = c; closest_d = d; }
    }
  }
  if(center_sum > 0.0) {
    for(c=0;c<center->rows;c++) {
      mbr[c] /= center_sum;
    }
  }
  else {
    /* use the backup plan: closest cluster only */
    for(c=0; c < center->rows; c++) {
      if(c == closest) mbr[c] = 1;
      else mbr[c] = 0;
    }
  }
}

/* Returns the soft classification of all points 
 * ( (0.5 .. (n-0.5))*scale, (0.5 .. (n-0.5))*scale )
 * This is useful for visualization.
 */
void Clus_EMClassify(Matrix center, Matrix cov, Vector prior,
		     double temperature,
		     int freq, int scale, Matrix output)
{
  int i,j,c;
  Vector s = VectorCreate(2);
  Vector mbr = VectorCreate(center->rows);

  for(i=0;i<output->rows;i++) {
    s->d[1] = i * scale + scale/2;
    for(j=0;j<output->cols;j++) {
      s->d[0] = j * scale + scale/2;

      Classify(center, cov, prior, temperature, s, mbr->d);

      output->d2[i][j] = 0;
      for(c=0;c<center->rows;c++) {
	if((c / freq) % 2) {
	  output->d2[i][j] += mbr->d[c];
	}
      }
    }
  }
  VectorFree(mbr);
  VectorFree(s);
}

#define CDM 0

/* Returns the mixture-distance from point to all points
 * ( (0.5 .. (n-0.5))*scale, (0.5 .. (n-0.5))*scale )
 * This is useful for visualization.
 */
void Clus_EMDistance(Matrix center, Matrix cov, Vector prior,
		     double temperature,
		     Vector point, int scale, Matrix output)
{
  int i, j, c;
  Vector s = VectorCreate(2);
  Vector mbr = VectorCreate(center->rows);
#if CDM
  Vector mbr2 = VectorCreate(center->rows);
#endif

  Classify(center, cov, prior, temperature, point, mbr->d);

  for(i=0;i<output->rows;i++) {
    s->d[1] = i * scale + scale/2;
    for(j=0;j<output->cols;j++) {
      s->d[0] = j * scale + scale/2;

#if CDM
      Classify(center, cov, prior, temperature, s, mbr2->d);

#if 0
      output->d2[i][j] = 0;
      for(c=0;c<center->rows;c++) {
	output->d2[i][j] += mbr->d[c] * mbr2->d[c];
      }
#else
      output->d2[i][j] = -VectorRelEntropy(mbr, mbr2);
      if(!finite(output->d2[i][j])) output->d2[i][j] = -1e3;
#endif
#else
      output->d2[i][j] = 0;
      for(c=0;c<center->rows;c++) {
	double d = Mahalanobis(point->d, s->d, cov->d2[c], s->len);
	output->d2[i][j] += mbr->d[c] * exp(-d);
      }
#endif
    }
  }

#if CDM
  VectorFree(mbr2);
#endif
  VectorFree(mbr);
  VectorFree(s);
}
