/* Gaussian mixture model data structure. */
#include <tpm/tpm.h>
#include <tpm/gmm.h>
#include <tpm/logdouble.h>

#define VERBOSE 0
/* ROBUST mode leaves some samples unassigned; i.e. mbr will not sum to 1. */
/* this does not interact well with covariance estimation */
#define ROBUST 0
/* ROBUST_T is the probability threshold to determine an outlier. */
#define ROBUST_T 1e-50

/* Prototypes ****************************************************************/

GaussianMixture GaussianMixtureCreate(int ncenters, int ndims);
void GaussianMixtureFree(GaussianMixture gm);
void GaussianMixtureSample(GaussianMixture gm, Vector sample);
double GaussianMixtureLogProb(GaussianMixture gm, Vector point);
double GaussianMixtureClassify(GaussianMixture gm, Vector point, double *mbr);
double GaussianMixtureDistance(GaussianMixture gm, Vector v1, Vector v2);

void GaussianMixtureIdentityCov(GaussianMixture gm);
void GaussianMixtureMeanCenter(GaussianMixture gm, Matrix points);
double GaussianMixtureTrainEM(GaussianMixture gm, Matrix points);
double GaussianMixtureTrainStepE(GaussianMixture gm, Matrix points, Matrix mbr);
void GaussianMixtureTrainStepM(GaussianMixture gm, Matrix points, Matrix mbr);
void GaussianMixtureTrainStepV(GaussianMixture gm, Matrix points, Matrix mbr);
void GaussianMixtureTrainStepP(GaussianMixture gm, Matrix mbr);

void GaussianMixtureTrainIncremental(GaussianMixture gm, Vector point,
				     double rate);
void GaussianMixtureTrainFast(GaussianMixture gm, Matrix points, int iters);

/* Functions *****************************************************************/

/* Creates a new mixture with arbitrary parameters. */
GaussianMixture GaussianMixtureCreate(int ncenters, int ndims)
{
  GaussianMixture gm = Allocate(1, struct GaussianMixture);
  gm->center = MatrixCreate(ncenters, ndims);
  gm->cov = MatrixCreate(ncenters, ndims * ndims);
  gm->cov_type = COV_FULL;
  GaussianMixtureIdentityCov(gm);
  gm->prior = VectorCreate(ncenters);
  VectorSet(gm->prior, 1.0/ncenters);
  gm->temperature = 1.0;
  return gm;
}

void GaussianMixtureFree(GaussianMixture gm)
{
  MatrixFree(gm->center);
  MatrixFree(gm->cov);
  VectorFree(gm->prior);
  free(gm);
}

/* Modifies sample to be a sample from the density. */
void GaussianMixtureSample(GaussianMixture gm, Vector sample)
{
  int c, d;
  Vector t = VectorCreate(sample->len);

  assert(sample->len == gm->center->cols);

  /* Randomly pick a center */
  c = VectorSample(gm->prior);

  /* Sample center->cols normal rv's */
  for(d=0;d<gm->center->cols;d++) {
    t->d[d] = RandNormal();
  }

  if(gm->cov_type == COV_FULL) {
    /* Rotate t according to the covariance */
    Matrix m = MatrixFromData(gm->center->cols, gm->center->cols, 
			      gm->cov->d2[c]);
    MatrixEvInvert(m);
    MatrixVectorMultiply(sample, m, t);
    MatrixEvInvert(m);
    free(m->d2);
    free(m);
  }
  else {
    VectorMove(sample, t);
  }
  VectorFree(t);
  VectorScale(sample, gm->temperature);

  /* Shift according to the mean */
  for(d=0;d<gm->center->cols;d++) {
    sample->d[d] += gm->center->d2[c][d];
  }
}

#define LOGTWOPI   1.8378770664093454

/* Returns the logarithm of the probability of point according to gm. */
double GaussianMixtureLogProb(GaussianMixture gm, Vector point)
{
  double sum = -1.0/0;
  int c;
  /* t is a dummy vector which points to the current mean */
  Vector t = VectorFromData(gm->center->cols, gm->center->d2[0]);
  assert(point->len == gm->center->cols);
  for(c=0;c<gm->center->rows;c++) {
    double d;
    if(gm->prior->d[c] == 0) continue;
    t->d = gm->center->d2[c];
    if(gm->cov_type == COV_FULL) {
      /* m is a dummy matrix which points to the current covariance */
      Matrix m = MatrixFromData(gm->center->cols, gm->center->cols, 
				gm->cov->d2[c]);
      d = VectorMahalanobisEv(point, t, m);
      free(m->d2);
      free(m);
    }
    else {
      d = VectorDistance(point, t);
      d *= d;
      d += point->len * LOGTWOPI;
      d /= 2;
    }
    if(finite(d)) {
#if 0
      sum += exp(-d/gm->temperature) * gm->prior->d[c];
#else
      /* this works even if gm->temperature == 0. */
      double x = -d/gm->temperature + log(gm->prior->d[c]);
      sum = LogDoubleAdd(sum, x);
#endif
    }
  }
  free(t);
  return sum;
}

/* using a logarithmic representation is more accurate. */
#define USE_LOG 1

/* Modifies mbr to be the soft assignment of point to the mixture
 * components.  Returns the log-likelihood of point.
 */
double GaussianMixtureClassify(GaussianMixture gm, Vector point, double *mbr)
{
#if USE_LOG
  double likelihood = -1.0/0;
  int ties = 0;
#else
  double likelihood = 0;
  double closest_d = 1.0/0;
  int closest = 0;
#endif
  int c;
  /* t is a dummy vector which points to the current mean */
  Vector t = VectorFromData(gm->center->cols, gm->center->d2[0]);
  assert(point->len == gm->center->cols);
  for(c=0;c<gm->center->rows;c++) {
    double d;
    t->d = gm->center->d2[c];
    if(gm->cov_type == COV_FULL) {
      /* m is a dummy matrix which points to the current covariance */
      Matrix m = MatrixFromData(gm->center->cols, gm->center->cols, 
				gm->cov->d2[c]);
      d = VectorMahalanobisEv(point, t, m);
      free(m->d2);
      free(m);
    }
    else {
      d = VectorDistance(point, t);
      d *= d;
      d += point->len * LOGTWOPI;
      d /= 2;
    }
    if(!finite(d)) {
      mbr[c] = 0;
    }
    else {
#if USE_LOG
      if(gm->prior->d[c] > 0) {
	if(gm->temperature == 0) {
	  /* store the distance in mbr and let likelihood be the minimum. */
	  mbr[c] = -d;
	  if(mbr[c] > likelihood) {
	    likelihood = mbr[c];
	    ties = 1;
	  }
	  else if(mbr[c] == likelihood) ties++;
	}
	else {
	  mbr[c] = -d/gm->temperature + log(gm->prior->d[c]);
	  likelihood = LogDoubleAdd(likelihood, mbr[c]);
	}
      }
      /* should mbr be set to a dummy value here? */
#else
      mbr[c] = exp(-d/gm->temperature) * gm->prior->d[c];
      likelihood += mbr[c];
      /* prepare the backup plan */
      if(d < closest_d) { closest = c; closest_d = d; }
#endif
    }
  }
  free(t);
#if USE_LOG
  if(gm->temperature == 0) {
    /* mbr[c] holds -d */
    for(c=0;c<gm->center->rows;c++) {
      if(mbr[c] == likelihood) {
	mbr[c] = 1.0/ties;
      }
      else mbr[c] = 0;
    }
  }
  else {
    for(c=0;c<gm->center->rows;c++) {
      mbr[c] = exp(mbr[c] - likelihood);
    }
  }
#else
  if(likelihood > 0.0) {
    if(ROBUST) likelihood += ROBUST_T;
    for(c=0;c<gm->center->rows;c++) {
      mbr[c] /= likelihood;
    }
  }
  else if(!ROBUST) {
    /* use the backup plan: closest cluster only */
    for(c=0; c < gm->center->rows; c++) {
      if(c == closest) mbr[c] = 1;
      else mbr[c] = 0;
    }
    likelihood = -closest_d;
  }
  else {
    /* this point is an outlier */
    for(c=0; c < gm->center->rows; c++) mbr[c] = 0;
  }
#endif
  if(0) {
    /* check */
    assert(likelihood == GaussianMixtureLogProb(gm, point));
  }
  return likelihood;
}

/* Returns the mixture-distance as defined by Cox and Yianilos. 
 * It is asymmetric and technically a probability, not a distance.
 * Uses temperature for the classification but not for the distances.
 */
double GaussianMixtureDistance(GaussianMixture gm, Vector v1, Vector v2)
{
  double distance = 0;
  int c;
  Vector mbr = VectorCreate(gm->center->rows);
  GaussianMixtureClassify(gm, v1, mbr->d);
  for(c=0;c<gm->center->rows;c++) {
    double d;
    if(gm->cov_type == COV_FULL) {
      /* m is a dummy matrix which points to the current covariance */
      Matrix m = MatrixFromData(gm->center->cols, gm->center->cols, 
				gm->cov->d2[c]);
      d = VectorMahalanobisEv(v1, v2, m);
      free(m->d2);
      free(m);
    }
    else {
      d = VectorDistance(v1, v2);
      d *= d;
    }
    distance += exp(-d) * mbr->d[c];
  }
  VectorFree(mbr);
  return distance;
}

/* EM Training ***************************************************************/

/* Set all of the covariance matrices to the identity. */
void GaussianMixtureIdentityCov(GaussianMixture gm)
{
  int i,j,n;
  VectorSet((Vector)gm->cov, 0);
  n = gm->center->cols;
  for(i=0;i<gm->cov->rows;i++) {
    for(j=0;j<n;j++) {
      gm->cov->d2[i][j*n+j] = 1;
    }
  }
  gm->cov_type = COV_IDENTITY;
}

/* Set all of the centers to the mean of points. */
void GaussianMixtureMeanCenter(GaussianMixture gm, Matrix points)
{
  int i,j;
  Vector mean = VectorCreate(points->cols);
  assert(gm->center->cols == points->cols);
  /* VectorRowSum(points, mean); */
  VectorSet(mean, 0);
  for(i=0;i<points->rows;i++) {
    for(j=0;j<points->cols;j++) {
      mean->d[j] += points->d2[i][j];
    }
  }
  VectorScale(mean, 1.0/points->cols);
  for(i=0;i<gm->center->rows;i++) {
    for(j=0;j<gm->center->cols;j++) {
      gm->center->d2[i][j] = mean->d[j];
    }
  }
  VectorFree(mean);
}

/* Train the mixture at a particular temperature until convergence. */
double GaussianMixtureTrainEM(GaussianMixture gm, Matrix points)
{
  int iters;
  double likelihood = 1e12, old_likelihood;
  Matrix mbr = MatrixCreate(points->rows, gm->center->rows);
  assert(gm->center->cols == points->cols);
  VectorSet(gm->prior, 1.0/gm->prior->len);
  for(iters=0;iters<100;iters++) {
    /* compute membership */
    old_likelihood = likelihood;
    likelihood = GaussianMixtureTrainStepE(gm, points, mbr);
    /* compute centers */
    GaussianMixtureTrainStepM(gm, points, mbr);
    if(iters < 2) VectorPerturb((Vector)gm->center, 1e-4);
#if 0
    /* compute covariances */
    GaussianMixtureTrainStepV(gm, points, mbr);
#endif
#if 0
    /* compute weights */
    GaussianMixtureTrainStepP(gm, mbr);
#endif
    /* have to use fabs since likelihood may momentarily increase */
    /* if temperature is very small, this rule will never take effect. */
    if((iters > 2) && 
       (fabs(old_likelihood - likelihood) < 1e-1*gm->temperature))
      break;
    if(points->rows > 1000) fprintf(stderr, "EM iteration %d\n", iters);
  }
#if 0
  fprintf(stderr, "trained for %d iters\n", iters);
#endif
  MatrixFree(mbr);
  return likelihood;
}

/* Modifies mbr to be the membership probabilities for all points.
 * Returns the average log-likelihood.
 */
double GaussianMixtureTrainStepE(GaussianMixture gm, Matrix points, Matrix mbr)
{
  double likelihood = 0;
  int p;
  Vector point = VectorFromData(points->cols, points->d2[0]);
  assert(points->cols == gm->center->cols);
  /* Loop points */
  for(p = 0; p < points->rows; p++) {
    point->d = points->d2[p];
    likelihood += GaussianMixtureClassify(gm, point, mbr->d2[p]);
  }
  free(point);
  return likelihood/points->rows;
}

/* Sets center[c] to be the weighted mean of the points,
 * with weights given by mbr[:][c].
 */
void GaussianMixtureTrainStepM(GaussianMixture gm, Matrix points, Matrix mbr)
{
  int s, c, i;
  /* loop clusters and compute new mean */
  VectorSet((Vector)gm->center, 0);
  for(c=0; c < gm->center->rows; c++) {
    double point_sum = 0;
#if VERBOSE
    if(c % 10 == 0) fprintf(stderr, "updating center %d mean\n", c);
#endif
    for(s=0; s < points->rows; s++) {
      point_sum += mbr->d2[s][c];
      for(i=0; i < gm->center->cols; i++) {
	gm->center->d2[c][i] += mbr->d2[s][c] * points->d2[s][i];
      }
    }
    if(point_sum > 0.0) {
      for(i=0; i < gm->center->cols; i++) {
	gm->center->d2[c][i] /= point_sum;
      }
    }
  }
}

/* eigenvalues of covariance matrix cannot go below LAMBDA_FLOOR */
#define LAMBDA_FLOOR 1e-10

/* Modify cov (size rank by rank) to be its eigenvectors scaled 
 * by the inverse root eigenvalues.
 */
static void Cov2EV(double *cov, int rank)
{
  Matrix m = MatrixFromData(rank, rank, cov); /* eigenvectors */
  double *v = Allocate(rank, double); /* eigenvalues */
  int i, j;

  /* extract the eigenvectors of the covariance matrix */
  MatrixEigenvectorsSym(m, v);
  /* take the root of the eigenvalues */
  for(i=0; i < rank; i++) {
#if 0
    /* this flooring is related to Baback's PhD */
    if(v[i] < LAMBDA_FLOOR) v[i] = LAMBDA_FLOOR;
    else 
#endif
    v[i] = sqrt(v[i]);
  }
  /* scale the eigenvectors by the inverse eigenvalues */
  for(i=0; i < rank; i++) {
    for(j=0; j < rank; j++) {
      m->d2[i][j] /= v[i];
    }
  }

  free(v);
  free(m->d2);
  free(m);
}

/* Modify cov (size rank by rank) to be a regular non-inverted
 * covariance matrix. 
 */
static void EV2Cov(double *cov, int rank)
{
  Matrix m = MatrixFromData(rank, rank, cov); /* eigenvectors */
  int i,j;

  /* this is MatrixEvInvert */
  for(i=0; i < rank; i++) {
    double sum = 0;
    for(j=0; j < rank; j++) {
      sum += m->d2[i][j] * m->d2[i][j];
    }
    for(j=0; j < rank; j++) {
      m->d2[i][j] /= sum;
    }
  }

  MatrixTransposeTimesSelf(m, m);
  free(m->d2);
  free(m);
}

/* Sets cov[c][d] to be the weighted variance of the points
 * along dimension d, with weights given by mbr[:][c].
 */
void GaussianMixtureTrainStepV(GaussianMixture gm, Matrix points, Matrix mbr)
{
  int s, c, i, j;
  /* loop clusters and compute new variance */
  VectorSet((Vector)gm->cov, 0);
  for(c=0; c < gm->center->rows; c++) {
    double point_sum = 0;
    Matrix m = MatrixFromData(gm->center->cols, gm->center->cols, 
			      gm->cov->d2[c]);
#if VERBOSE
    if(c % 10 == 0) fprintf(stderr, "updating center %d variance\n", c);
#endif
    for(s=0; s < points->rows; s++) {
      point_sum += mbr->d2[s][c];
      for(i=0; i < gm->center->cols; i++) {
	double di = points->d2[s][i] - gm->center->d2[c][i];
	for(j=i; j < gm->center->cols; j++) {
	  double dj = points->d2[s][j] - gm->center->d2[c][j];
	  m->d2[i][j] += mbr->d2[s][c] * di * dj;
	  m->d2[j][i] = m->d2[i][j];
	}
      }
    }
#if 1
    /* add a multiple of the identity */
    for(i=0; i < gm->center->cols; i++) {
      m->d2[i][i] += LAMBDA_FLOOR;
    }
    VectorScale((Vector)m, 1.0/(point_sum + LAMBDA_FLOOR));
#else
    if(point_sum > 0.0) {
      VectorScale((Vector)m, 1.0/point_sum);
    }
#endif
    Cov2EV(m->d, gm->center->cols);
    free(m->d2);
    free(m);
  }
  gm->cov_type = COV_FULL;
}

/* Sets prior to be the new mixing proportions given assignments mbr. */
void GaussianMixtureTrainStepP(GaussianMixture gm, Matrix mbr)
{
  int s, c;
  double sum = 0;
  for(c=0; c < mbr->cols; c++) {
    double point_sum = 0;
    for(s=0; s < mbr->rows; s++) {
      point_sum += mbr->d2[s][c];
    }
    sum += point_sum;
    gm->prior->d[c] = point_sum;
  }
  VectorScale(gm->prior, 1.0/sum);
}

/* Incremental Training ******************************************************/

/* Modifies gm to make point more likely. 
 */
void GaussianMixtureTrainIncremental(GaussianMixture gm, Vector point,
				     double rate)
{
  int c, i, j;
  double *mbr = Allocate(gm->center->rows, double);
  GaussianMixtureClassify(gm, point, mbr);
  for(c=0;c<gm->center->rows;c++) {
    Matrix m;
    double a = mbr[c]*rate;

#if 0
    m = MatrixFromData(gm->center->cols, gm->center->cols, 
		       gm->cov->d2[c]);
    EV2Cov(m->d, gm->center->cols);
    /* update covariance */
    for(i=0;i<gm->center->cols;i++) {
      double di = point->d[i] - gm->center->d2[c][i];
      for(j=0;j<gm->center->cols;j++) {
	double dj = point->d[j] - gm->center->d2[c][j];
	m->d2[i][j] = (1-a/10)*m->d2[i][j] + a/10 * di * dj;
	fprintf(stderr, "%g %g %g\n", di*dj, m->d2[i][j], a);
	m->d2[j][i] = m->d2[i][j];
      }
    }
    Cov2EV(m->d, gm->center->cols);
    free(m->d2);
    free(m);
    gm->cov_type = COV_FULL;
#endif

    /* update center */
    for(i=0;i<gm->center->cols;i++) {
      gm->center->d2[c][i] = 
	(1-a)*gm->center->d2[c][i] +
	a*point->d[i];
    }

#if 0
    /* update prior */
    gm->prior->d[c] = (1-rate/10)*gm->prior->d[c] + rate/10*mbr[c];
#endif
  }
  free(mbr);
}

/* Calls TrainIncremental iters times using a random row from points. */
void GaussianMixtureTrainFast(GaussianMixture gm, Matrix points, int iters)
{
  Vector v = VectorFromData(points->cols, NULL);
  int i;
  /* this rule is valid over a wide range of iters, at least for test_gmm */
  double rate = 3.0/gm->center->cols/sqrt(iters);

  /* Initialization */
  VectorSet(gm->prior, 1.0/gm->prior->len);
  for(i=0;i<gm->center->rows;i++) {
    v->d = points->d2[ RandInt(0, points->rows-1) ];
    memcpy(gm->center->d2[i], v->d, v->len*sizeof(double));
  }

  /* Refinement */
  for(i=0;i<iters;i++) {
    if((iters % 1000) == 0) 
      fprintf(stderr, "Incremental iteration %d\n", iters);
    v->d = points->d2[ RandInt(0, points->rows-1) ];
    GaussianMixtureTrainIncremental(gm, v, rate);
    if(iters < 2) VectorPerturb((Vector)gm->center, 1e-4);
  }
  free(v);
}
