/* Multinomial mixture model data structure. 
 * This code is almost identical to gmm.c
 */
#include <tpm/tpm.h>
#include <tpm/mnm.h>

#define VERBOSE 0
/* ROBUST mode leaves some samples unassigned; i.e. mbr will not sum to 1. */
#define ROBUST 1
/* ROBUST_T is the probability threshold to determine an outlier. */
#define ROBUST_T 1e-50

/* Prototypes ****************************************************************/

MultinomialMixture MultinomialMixtureCreate(int ncenters, int ndims);
void MultinomialMixtureFree(MultinomialMixture dm);
double MultinomialMixtureClassify(MultinomialMixture dm, Vector point, 
				double *mbr);

double MultinomialMixtureTrainEM(MultinomialMixture dm, Matrix points);
double MultinomialMixtureTrainStepE(MultinomialMixture dm, Matrix points, 
				  Matrix mbr);
void MultinomialMixtureTrainStepM(MultinomialMixture dm, Matrix points, 
				Matrix mbr);
void MultinomialMixtureTrainStepP(MultinomialMixture dm, Matrix mbr);

/* Functions *****************************************************************/

/* Creates a new mixture with arbitrary parameters. */
MultinomialMixture MultinomialMixtureCreate(int ncenters, int ndims)
{
  MultinomialMixture dm = Allocate(1, struct MultinomialMixture);
  dm->center = MatrixCreate(ncenters, ndims);
  dm->prior = VectorCreate(ncenters);
  VectorSet(dm->prior, 1.0/ncenters);
  dm->temperature = 1.0;
  return dm;
}

void MultinomialMixtureFree(MultinomialMixture dm)
{
  MatrixFree(dm->center);
  VectorFree(dm->prior);
  free(dm);
}

/* Modifies mbr to be the soft assignment of sample to the mixture
 * components.  Returns the average distortion.
 */
double MultinomialMixtureClassify(MultinomialMixture dm, Vector point, 
				  double *mbr)
{
  double center_sum = 0;
  double closest_d = 1.0/0;
  int closest = 0;
  int c;
  double distortion = 0;
  /* t is a dummy vector which points to the current mean */
  Vector t = VectorFromData(dm->center->cols, dm->center->d2[0]);
  for(c=0;c<dm->center->rows;c++) {
    double d;
    t->d = dm->center->d2[c];
#if 1
    d = VectorRelEntropy(point, t);
#else
    d = VectorChi(point, t);
#endif
    if(!finite(d)) {
      mbr[c] = 0;
    }
    else {
      mbr[c] = exp(-d/dm->temperature) * dm->prior->d[c];
      center_sum += mbr[c];
      distortion += mbr[c] * d;
      /* prepare the backup plan */
      if(d < closest_d) { closest = c; closest_d = d; }
    }
  }
  free(t);
  if(center_sum > 0.0) {
    if(ROBUST) center_sum += ROBUST_T;
    for(c=0;c<dm->center->rows;c++) {
      mbr[c] /= center_sum;
    }
    distortion /= center_sum;
  }
  else if(!ROBUST) {
    /* use the backup plan: closest cluster only */
    for(c=0; c < dm->center->rows; c++) {
      if(c == closest) mbr[c] = 1;
      else mbr[c] = 0;
    }
    distortion = closest_d;
  }
  else {
    /* this point is an outlier */
    for(c=0; c < dm->center->rows; c++) mbr[c] = 0;
  }
  return distortion;
}

/* EM Training ***************************************************************/

/* Train the mixture at a particular temperature until convergence. */
double MultinomialMixtureTrainEM(MultinomialMixture dm, Matrix points)
{
  int iters;
  double distortion = 1e12, old_distortion;
  Matrix mbr = MatrixCreate(points->rows, dm->center->rows);
  VectorSet((Vector)mbr, 1.0/mbr->cols);
  VectorSet(dm->prior, 1.0/dm->prior->len);
  for(iters=0;iters<100;iters++) {
    /* compute centers */
    MultinomialMixtureTrainStepM(dm, points, mbr);
    if(iters < 2) VectorPerturb((Vector)dm->center, 1e-5);
    /* compute membership */
    old_distortion = distortion;
    distortion = MultinomialMixtureTrainStepE(dm, points, mbr);
#if 0
    /* compute weights */
    MultinomialMixtureTrainStepP(dm, mbr);
#endif
    /* have to use fabs since distortion may momentarily increase */
    if((iters > 2) && 
       (fabs(old_distortion - distortion) < 1e-1*dm->temperature))
      break;
    if(points->rows > 1000) fprintf(stderr, "EM iteration %d\n", iters);
  }
  MatrixFree(mbr);
  return distortion;
}

/* Modifies mbr to be the membership probabilities for all points.
 * Returns the average total distortion.
 */
double MultinomialMixtureTrainStepE(MultinomialMixture dm, Matrix points, 
				    Matrix mbr)
{
  double distortion = 0;
  int p;
  Vector point = VectorFromData(points->cols, points->d2[0]);
  assert(points->cols == dm->center->cols);
  /* Loop points */
  for(p = 0; p < points->rows; p++) {
    point->d = points->d2[p];
    distortion += MultinomialMixtureClassify(dm, point, mbr->d2[p]);
  }
  free(point);
  return distortion;
}

/* Sets center[c] to be the weighted mean of the points,
 * with weights given by mbr[:][c].
 */
void MultinomialMixtureTrainStepM(MultinomialMixture dm, Matrix points, 
				  Matrix mbr)
{
  int s, c, i;
  /* loop clusters and compute new mean */
  VectorSet((Vector)dm->center, 0);
  for(c=0; c < dm->center->rows; c++) {
    double point_sum = 0;
#if VERBOSE
    if(c % 10 == 0) fprintf(stderr, "updating center %d mean\n", c);
#endif
    for(s=0; s < points->rows; s++) {
      point_sum += mbr->d2[s][c];
      for(i=0; i < dm->center->cols; i++) {
	dm->center->d2[c][i] += mbr->d2[s][c] * points->d2[s][i];
      }
    }
    if(point_sum > 0.0) {
      for(i=0; i < dm->center->cols; i++) {
	dm->center->d2[c][i] /= point_sum;
      }
    }
  }
}

/* Sets prior to be the new mixing proportions given assignments mbr. */
void MultinomialMixtureTrainStepP(MultinomialMixture dm, Matrix mbr)
{
  int s, c;
  double sum = 0;
  for(c=0; c < mbr->cols; c++) {
    double point_sum = 0;
    for(s=0; s < mbr->rows; s++) {
      point_sum += mbr->d2[s][c];
    }
    sum += point_sum;
    dm->prior->d[c] = point_sum;
  }
  VectorScale(dm->prior, 1.0/sum);
}
