/* Minimal spanning tree. */
#include <tpm/cluster.h>
#include <tpm/heap.h>

SpanTree Clus_MST(int nsamples, Clus_RowFunc *dist, Any data);

/* Functions *****************************************************************/

typedef struct HeapNode {
  double distance;
  int sample, closest;
} *HeapNode;

static int HeapNodeCompare(HeapNode a, HeapNode b)
{
  if(a->distance < b->distance) return -1;
  else return 1;
}

/* Returns a spanning tree over 0..nsamples-1 such that the sum of the
 * edge costs (given by dist) is minimized.
 */
SpanTree Clus_MST(int nsamples, Clus_RowFunc *dist, Any data)
{
  double *distance = Allocate(nsamples, double);
  int i,j;
  HeapNode *nodes = Allocate(nsamples, HeapNode);
  Heap heap;
  SpanTree st = SpanTreeCreate(nsamples);

  /* set distances to inf */
  for(i=0;i<nsamples;i++) {
    nodes[i] = Allocate(1, struct HeapNode);
    nodes[i]->sample = i;
    nodes[i]->distance = 1.0/0;
  }
  /* except for the first */
  nodes[0]->distance = 0;
  nodes[0]->closest = -1;
  /* put them in a heap for fast extractMin */
  heap = HeapFromArray((Any)nodes, nsamples, (CmpFunc*)HeapNodeCompare, 
		       GenericFree, NULL);

  /* loop nsamples times */
  for(i=0;i<nsamples;i++) {
    /* extract the closest */
    HeapNode n = (HeapNode)HeapExtractMin(heap);
    j = n->sample;
    /* link to the closest in the tree */
    st->node[j].parent = n->closest;
    st->node[j].edge = n->distance;
    free(n);
    /* update distances */
    if(dist(j, nsamples, distance, data) != TPM_OK) {
      fprintf(stderr, "error computing distances to %d\n", j);
      free(distance);
      HeapFree(heap);
      SpanTreeFree(st);
      return NULL;
    }
    {HeapIter(k, n, heap) {
      /* check if j is now the closest. */
      if(distance[n->sample] < n->distance) {
	n->distance = distance[n->sample];
	n->closest = j;
	/* this is a safe mutation during iteration since the
	 * node can only move earlier in the heap.
	 */
	HeapChangedKey(heap, k);
      }
    }}
  }
  free(distance);
  HeapFree(heap);
  return st;
}
