#include <tpm/cluster.h>

#define FISSION 1e-2

/* Rotates s and c by n*n matrix v before computing distance. */
static double Euclidean(double *s, double *c, int n, double *r)
{
  double sum = 0;
  int i;

  for(i=0;i<n;i++) {
    r[i] = s[i] - c[i];
    sum += r[i]*r[i];
  }
  return sum;
}

/* What if we explicitly maximized the total distance between points? */
/* Simplified particle repulsion.
 * Each particle adds the difference vector with its nearest neighbor
 * to its current velocity.
 * Modifies: sample, vel.
 */
double SurfaceRepel(Matrix sample, Matrix vel, double noise,
		    int *nsample_p,
		    ProjectFunc *project, Any data)
{
  double total = 0;
  int nsample = *nsample_p;
  int i, j, k;
  Vector r = VectorCreate(sample->cols);
  Vector closest_r = VectorCreate(sample->cols);

  for(i=0;i<nsample;i++) {
    double min_d, dot;
    int closest;

    /* find closest point and set closest_r to the difference vector */
    min_d = 1.0/0;
    for(j=0;j<nsample;j++) {
      double d;
      if(i == j) continue;
      d = Euclidean(sample->d2[i], sample->d2[j], sample->cols, r->d);
      if(d < min_d) {
	min_d = d;
	memcpy(closest_r->d, r->d, sample->cols*sizeof(double));
	closest = j;
      }
    }
#if 0
    fprintf(stderr, "closest to %d is %d\n", i, closest);
#endif

    total += min_d;

    /* project the difference vector */
    project(closest_r->d, sample->d2[i], sample->cols, 0.0, i, data);
    min_d = 0;
    for(k=0;k<sample->cols;k++) {
      min_d += closest_r->d[k] * closest_r->d[k];
    }

    if(min_d > 0) {
      /* normalize the derivative */
      min_d = sqrt(min_d);
      for(k=0;k<sample->cols;k++) {
	closest_r->d[k] /= min_d;
      }
    }
#if 0
    else {
      /* randomly perturb the derivative */
      for(k=0;k<sample->cols;k++) {
	closest_r->d[k] = Rand()-0.5;
      }
    }
#endif

#if 0
    dot = 0;
    for(k=0;k<sample->cols;k++) {
      dot += vel->d2[i][k] * vel->d2[i][k];
    }
    if(i == 10) fprintf(stderr, "vel mag: %g\n", dot);
#endif

    /* project the old velocity */
    project(vel->d2[i], sample->d2[i], sample->cols, 0.0, i, data);

    /* compute dot product with velocity */
    dot = 0;
    for(k=0;k<sample->cols;k++) {
      dot += closest_r->d[k] * vel->d2[i][k];
    }

    /* add to the velocity */
    if(dot < 0) { dot *= -1.1; /* must be in (1,2) */ }
    else { dot *= 0.5; /* must be in (0,1) */ }
    if(i == 20) fprintf(stderr, "dot: %g\n", dot);
    for(k=0;k<sample->cols;k++) {
      vel->d2[i][k] += dot * closest_r->d[k];
      vel->d2[i][k] += 2*(Rand()-0.5)*noise;
    }

    /* project the new velocity */
    project(vel->d2[i], sample->d2[i], sample->cols, 0.5, i, data);

    /* fission? */
    if((nsample < sample->rows) && (min_d > 0) && (dot/min_d < FISSION)) {
      fprintf(stderr, "fission %d\n", i);
      memcpy(sample->d2[nsample], sample->d2[i], sample->cols*sizeof(double));
      memcpy(vel->d2[nsample], vel->d2[i], vel->cols*sizeof(double));

      dot = 0;
      for(k=0;k<sample->cols;k++) {
	dot += vel->d2[nsample][k] * vel->d2[nsample][k];
      }
      fprintf(stderr, "vel mag: %g\n", dot);

      nsample++;
    }
  }

  /* apply the derivative (does extra work) */
  VectorAdd((Vector)sample, (Vector)vel);

  VectorFree(r);
  VectorFree(closest_r);
  *nsample_p = nsample;
  return total/nsample;
}
