/* Single-link clustering with little memory usage.
 * Uses Sibson's SLINK.
 */
#include <tpm/cluster.h>

#define VERBOSE 0

Tree Clus_SingleLink(int nsamples, Clus_RowFunc *dist, Any data);

/* Functions *****************************************************************/

/* Sibson's SLINK.
 * Mutates the pi[nsamples] and lambda[nsamples] arrays.
 * As given in Info Retr: Data Structures + Algs, by
 * Frakes and Baeza-Yates, ed.
 * Original reference: 
 * Sibson, R. 1973. "SLINK: an Optimally Efficient Algorithm for the 
 * Single-Link Cluster Method." Computer Journal, 16, 30--34.
 */
static int SLINK(int nsamples, Clus_RowFunc *dist, Any data,
		 int *pi, double *lambda)
{
  double *distance = Allocate(nsamples, double);
  int i,j;

  pi[0] = 0;
  lambda[0] = 1.0/0;
  for(i=1;i<nsamples;i++) {
    if(VERBOSE && (i % 100 == 0)) fprintf(stderr, "%d\n", i);
    pi[i] = i;
    lambda[i] = 1.0/0;
    if(dist(i, i, distance, data) != TPM_OK) {
      fprintf(stderr, "error computing distances to %d\n", i);
      free(distance);
      return TPM_ERROR;
    }
    /* not j < i-1 as in book */
    for(j=0;j<i;j++) {
      int next = pi[j];
      if(lambda[j] < distance[j]) {
	if(distance[next] > distance[j]) distance[next] = distance[j];
      }
      else {
	if(distance[next] > lambda[j]) distance[next] = lambda[j];
	pi[j] = i;
	lambda[j] = distance[j];
      }
    }
    for(j=0;j<i-1;j++) {
      int next = pi[j];
      if(lambda[next] < lambda[j]) pi[j] = i;
    }
  }
  free(distance);
  return TPM_OK;
}

/* Convert the pointer representation from SLINK into a Tree. 
 * Not the most efficient way to do this, but not a bottleneck, either.
 */
static Tree SLINK2Tree(int nsamples, int *pi, double *lambda)
{
  Tree result, *top = Allocate(nsamples, Tree);
  int i;
  /* create leaves */
  for(i=0;i<nsamples;i++) top[i] = TreeCreate(IntCreate(i));
  /* create interior nodes */
  for(i=0;i<nsamples-1;i++) {
    int j, smallest_pos;
    double smallest_lambda;
    Tree new_tree;
    if(VERBOSE && (i % 100 == 0)) fprintf(stderr, "%d\n", i);
    /* find the smallest distance */
    smallest_pos = 0;
    smallest_lambda = lambda[0];
    for(j=1;j<nsamples;j++) {
      if(lambda[j] < smallest_lambda) {
	smallest_lambda = lambda[j];
	smallest_pos = j;
      }
    }
    /* merge smallest_pos with next */
    j = pi[smallest_pos];
    lambda[smallest_pos] = 1.0/0; /* so it doesn't get selected again */
    new_tree = TreeCreate(IntCreate((int)smallest_lambda));
    TreeAddChild(new_tree, top[j]);
    TreeAddChild(new_tree, top[smallest_pos]);
    top[j] = new_tree;
    top[smallest_pos] = NULL;
  }
  result = NULL;
  for(i=0;i<nsamples;i++) {
    if(top[i]) {
      assert(!result);  /* make sure there is only one */
      result = top[i];
    }
  }
  free(top);
  return result;
}

/* Returns a single-link clustering according to the given distance function.
 * Returns a Tree whose leaves are 0-based sample numbers. 
 */
Tree Clus_SingleLink(int nsamples, Clus_RowFunc *dist, Any data)
{
  Tree result;
  int *pi = Allocate(nsamples, int);
  double *lambda = Allocate(nsamples, double);
  if(SLINK(nsamples, dist, data, pi, lambda) != TPM_OK) return NULL;
  result = SLINK2Tree(nsamples, pi, lambda);
  free(pi);
  free(lambda);
  return result;
}
