/* Definition of the cluster package. Add it to Tcl with
 *   Clustertcl_Init(interp);
 * Commands:
 *   em_init sample center cov
 *   em_step sample center cov mbr temperature
 */
#include <tcl.h>
#include "tcl_macros.h"
#include <tpm/cluster.h>
#include <tpm/gmm.h>

/* Prototypes ****************************************************************/
int Clustertcl_Init(Tcl_Interp *interp);

/* Functions *****************************************************************/

/* Modifies output to be the soft classification of all points 
 * ( (0.5 .. (n-0.5))*scale, (0.5 .. (n-0.5))*scale ).
 * This is useful for visualization.
 */
static void ClassifyRegion(GaussianMixture gm, 
			   int freq, int scale, Matrix output)
{
  int i,j,c;
  Vector s = VectorCreate(2);
  Vector mbr = VectorCreate(gm->center->rows);

  for(i=0;i<output->rows;i++) {
    s->d[1] = i * scale + scale/2;
    for(j=0;j<output->cols;j++) {
      s->d[0] = j * scale + scale/2;

      GaussianMixtureClassify(gm, s, mbr->d);

      output->d2[i][j] = 0;
      for(c=0;c<gm->center->rows;c++) {
	if((c / freq) % 2) {
	  output->d2[i][j] += mbr->d[c];
	}
      }
    }
  }
  VectorFree(mbr);
  VectorFree(s);
}

/* Modifies output to be the mixture-distance from point to all points
 * ( (0.5 .. (n-0.5))*scale, (0.5 .. (n-0.5))*scale ).
 * This is useful for visualization.
 */
static void DistanceRegion(GaussianMixture gm, Vector point,
			   int scale, Matrix output)
{
  int i,j,c;
  Vector s = VectorCreate(2);

#if 0
  /* slow version */
  for(i=0;i<output->rows;i++) {
    s->d[1] = i * scale + scale/2;
    for(j=0;j<output->cols;j++) {
      s->d[0] = j * scale + scale/2;

      output->d2[i][j] = GaussianMixtureDistance(gm, point, s);
    }
  }
#else
  /* fast version */
  Vector mbr = VectorCreate(gm->center->rows);
  GaussianMixtureClassify(gm, point, mbr->d);
  VectorSet((Vector)output, 0);
  /* Loop centers */
  for(c=0;c<gm->center->rows;c++) {
    double d;
    /* m is a dummy matrix which points to the current covariance */
    Matrix m = MatrixFromData(gm->center->cols, gm->center->cols, 
			      gm->cov->d2[c]);

    for(i=0;i<output->rows;i++) {
      s->d[1] = i * scale + scale/2;
      for(j=0;j<output->cols;j++) {
	s->d[0] = j * scale + scale/2;

	d = VectorMahalanobisEv(point, s, m);
	output->d2[i][j] += exp(-d) * mbr->d[c];
      }
    }

    free(m->d2);
    free(m);
  }
  VectorFree(mbr);
#endif
  VectorFree(s);
}

/* Modifies output to be the divergence in class membership
 * from point to all points
 * ( (0.5 .. (n-0.5))*scale, (0.5 .. (n-0.5))*scale ).
 * This is useful for visualization.
 */
static void DivergenceRegion(GaussianMixture gm, Vector point, 
			     int scale, Matrix output)
{
  int i, j, c;
  Vector s = VectorCreate(2);
  Vector mbr = VectorCreate(gm->center->rows);
  Vector mbr2 = VectorCreate(gm->center->rows);

  GaussianMixtureClassify(gm, point, mbr->d);

  for(i=0;i<output->rows;i++) {
    s->d[1] = i * scale + scale/2;
    for(j=0;j<output->cols;j++) {
      s->d[0] = j * scale + scale/2;

      GaussianMixtureClassify(gm, s, mbr2->d);

      output->d2[i][j] = -VectorRelEntropy(mbr, mbr2);
      if(!finite(output->d2[i][j])) output->d2[i][j] = -1e3;
    }
  }

  VectorFree(mbr2);
  VectorFree(mbr);
  VectorFree(s);
}

static void CorrelationRegion(GaussianMixture gm, Vector point,
			      int scale, Matrix output)
{
  int k, i, j, c;
  Vector s = VectorCreate(2);
  Vector a = VectorCreate(2);
  Vector b = VectorCreate(2);
  int p_class, s_class;

  VectorSet((Vector)output, 0);
  
  /* Loop pairs of samples from the density */
  for(k=0;k<100;k++) {
    GaussianMixtureSample(gm, a);
    GaussianMixtureSample(gm, b);

    /* is point closer to a or b? */
    p_class = (VectorDistance(point, a) < VectorDistance(point, b));

    for(i=0;i<output->rows;i++) {
      s->d[1] = i * scale + scale/2;
      for(j=0;j<output->cols;j++) {
	s->d[0] = j * scale + scale/2;

	/* is s closer to a or b? */
	s_class = (VectorDistance(s, a) < VectorDistance(s, b));

	output->d2[i][j] += (p_class == s_class);
      }
    }
  }

  VectorFree(b);
  VectorFree(a);
  VectorFree(s);
}

typedef struct SurfaceLines {
  Matrix base, offset, mbr;
} *SurfaceLines;

/* Modifies d so that s+d lies on the surface. */
/* n is number of dimensions. */
static void SurfaceProjectLines(double *d, double *s, int n, double feedback,
				int i, SurfaceLines surf)
{
  int c, k;
  double dot, f;
  Matrix base = surf->base, offset = surf->offset;
  double *mbr = surf->mbr->d2[i];
  double *grad = AllocateC(n, double);
#if 0
  /* compute dot product of d with surface */
  for(c=0;c<base->rows;c++) {
    dot = 0;
    for(k=0;k<n;k++) {
      dot += d[k] * offset->d2[c][k];
    }
    dot *= mbr[c];
    for(k=0;k<n;k++) {
      grad[k] += dot * offset->d2[c][k];
    }

    dot = d[0] * offset->d2[c][1] + d[1] * offset->d2[c][0];
    dot *= mbr[c]*0.5;
    grad[0] += dot * offset->d2[c][1];
    grad[1] += dot * offset->d2[c][0];
  }  
  for(k=0;k<n;k++) {
    d[k] = grad[k];
  }
  memset(grad, 0, n*sizeof(double));
#endif
  /* compute gradient vector */
  for(c=0;c<base->rows;c++) {
    /* compute (sample - base) dot offset */
    dot = 0;
    for(k=0;k<n;k++) {
      dot += (s[k] - base->d2[c][k]) * offset->d2[c][k];
    }
    for(k=0;k<n;k++) {
      grad[k] += mbr[c] * (s[k] - base->d2[c][k] - dot * offset->d2[c][k]);
    }
  }
  /* compute magnitude (distance to surface) and dot product with d */
  dot = f = 0;
  for(k=0;k<n;k++) {
    f += grad[k] * grad[k];
    dot += grad[k] * d[k];
  }
#if 0
  dot = 0;
#endif
  if(f > 0) {
    /* this is a simplification of 
       f = sqrt(f);
       (dot / f + feedback * f) * grad / f 
     */
    dot /= f;
    dot += feedback;
    for(k=0;k<n;k++) {
      d[k] -= dot * grad[k];
    }
  }
  free(grad);
}

/* Modifies d so that s+d lies on the surface. */
static void SurfaceProjectSquare(double *d, double *s, int n, double feedback,
				 int i, Any data)
{
  /* project onto a square (100,100)-(300,300) */
  double f = 0, dot = 0;
  double *dx = Allocate(n, double);
  int j;
  for(j=0;j<n;j++) {
    if(s[j] < 100) {
      /* left corner */
      dx[j] = s[j] - 100;
    }
    else if(s[j] > 300) {
      /* right corner */
      dx[j] = s[j] - 300;
    }
    else {
      /* horizontal edge */
      dx[j] = 0;
    }
    f += dx[j] * dx[j];
    dot += d[j] * dx[j];
  }
  if(f > 0) {
    dot /= f;
    dot += feedback;
    for(j=0;j<n;j++) {
      d[j] -= dot * dx[j];
    }
  }
  free(dx);
}

/* Modifies d so that s+d lies on the surface. */
static void SurfaceProjectCircle(double *d, double *s, int n, double feedback,
				 int i, Any data)
{
  /* project onto a circle */
  double dx = s[0] - 200;
  double dy = s[1] - 200;
  double r = sqrt(dx*dx + dy*dy);
  double f = r - 100;
  double dot;
  /* gradient is [dx/r dy/r] */
  dx /= r;
  dy /= r;
  dot = d[0] * dx + d[1] * dy;
  d[0] -= (dot + feedback * f) * dx;
  d[1] -= (dot + feedback * f) * dy;
}

static int ClusterCmd(ClientData clientData, Tcl_Interp *interp,
		      int argc, char *argv[])
{
  Err(argc < 2, "missing directive; try one of: em_E, em_M, em_V");
  if(!strcmp(argv[1], "perturb")) {
    Vector v;
    double noise;
    Err(argc != 4, "usage: cluster perturb vector noise");
    GetPtr(argv[2], v);
    GetDouble(argv[3], noise);
    VectorPerturb(v, noise);
  }
  else if(!strcmp(argv[1], "gmm")) {
    GaussianMixture gm;
    Err(argc < 3, "missing directive; try one of: create, free, stepE, stepM");
    if(!strcmp(argv[2], "create")) {
      int ncenters, ndims;
      Err(argc != 5, "usage: cluster gmm create ncenters ndims");
      GetInt(argv[3], ncenters);
      GetInt(argv[4], ndims);
      gm = GaussianMixtureCreate(ncenters, ndims);
      ReturnPtr(gm);
    }
    else if(!strcmp(argv[2], "free")) {
      Err(argc != 4, "usage: cluster gmm free gm");
      GetPtr(argv[3], gm);
      GaussianMixtureFree(gm);
    }
    else if(!strcmp(argv[2], "center")) {
      Err(argc != 4, "usage: cluster gmm center gm");
      GetPtr(argv[3], gm);
      ReturnPtr(gm->center);
    }
    else if(!strcmp(argv[2], "cov")) {
      Err(argc != 4, "usage: cluster gmm cov gm");
      GetPtr(argv[3], gm);
      ReturnPtr(gm->cov);
    }
    else if(!strcmp(argv[2], "prior")) {
      Err(argc != 4, "usage: cluster gmm prior gm");
      GetPtr(argv[3], gm);
      ReturnPtr(gm->prior);
    }
    else if(!strcmp(argv[2], "temperature")) {
      Err(argc < 4, "usage: cluster gmm temperature gm ?value?");
      GetPtr(argv[3], gm);
      if(argc > 4) {
	GetDouble(argv[4], gm->temperature);
      }
      ReturnDouble(gm->temperature);
    }
    else if(!strcmp(argv[2], "sample")) {
      Vector v;
      Err(argc != 5, "usage: cluster gmm sample gm v");
      GetPtr(argv[3], gm);
      GetPtr(argv[4], v);
      GaussianMixtureSample(gm, v);
    }    
    else if(!strcmp(argv[2], "identity_cov")) {
      Err(argc != 4, "usage: cluster gmm identity_cov gm");
      GetPtr(argv[3], gm);
      GaussianMixtureIdentityCov(gm);
    }    
    else if(!strcmp(argv[2], "trainE")) {
      Matrix points, mbr;
      Err(argc != 6, "usage: cluster gmm trainE gm points mbr");
      GetPtr(argv[3], gm);
      GetPtr(argv[4], points);
      GetPtr(argv[5], mbr);
      ReturnDouble(GaussianMixtureTrainStepE(gm, points, mbr));
    }    
    else if(!strcmp(argv[2], "trainM")) {
      Matrix points, mbr;
      Err(argc != 6, "usage: cluster gmm trainM gm points mbr");
      GetPtr(argv[3], gm);
      GetPtr(argv[4], points);
      GetPtr(argv[5], mbr);
      GaussianMixtureTrainStepM(gm, points, mbr);
    }    
    else if(!strcmp(argv[2], "trainV")) {
      Matrix points, mbr;
      Err(argc != 6, "usage: cluster gmm trainV gm points mbr");
      GetPtr(argv[3], gm);
      GetPtr(argv[4], points);
      GetPtr(argv[5], mbr);
      GaussianMixtureTrainStepV(gm, points, mbr);
    }    
    else if(!strcmp(argv[2], "trainP")) {
      Matrix mbr;
      Err(argc != 5, "usage: cluster gmm trainP gm mbr");
      GetPtr(argv[3], gm);
      GetPtr(argv[4], mbr);
      GaussianMixtureTrainStepP(gm, mbr);
    }    
    else if(!strcmp(argv[2], "classify_region")) {
      int freq, scale;
      Matrix output;
      Err(argc != 7, "usage: cluster gmm classify_region gm freq scale output_matrix");
      GetPtr(argv[3], gm);
      GetInt(argv[4], freq);
      GetInt(argv[5], scale);
      GetPtr(argv[6], output);
      ClassifyRegion(gm, freq, scale, output);
    }
    else if(!strcmp(argv[2], "distance_region")) {
      Vector point;
      int scale;
      Matrix output;
      Err(argc != 7, "usage: cluster gmm distance_region gm point scale output_matrix");
      GetPtr(argv[3], gm);
      GetPtr(argv[4], point);
      GetInt(argv[5], scale);
      GetPtr(argv[6], output);
      DistanceRegion(gm, point, scale, output);
    }
    else if(!strcmp(argv[2], "divergence_region")) {
      Vector point;
      int scale;
      Matrix output;
      Err(argc != 7, "usage: cluster gmm divergence_region gm point scale output_matrix");
      GetPtr(argv[3], gm);
      GetPtr(argv[4], point);
      GetInt(argv[5], scale);
      GetPtr(argv[6], output);
      DivergenceRegion(gm, point, scale, output);
    }
    else if(!strcmp(argv[2], "correlation_region")) {
      Vector point;
      int scale;
      Matrix output;
      Err(argc != 7, "usage: cluster gmm correlation_region gm point scale output_matrix");
      GetPtr(argv[3], gm);
      GetPtr(argv[4], point);
      GetInt(argv[5], scale);
      GetPtr(argv[6], output);
      CorrelationRegion(gm, point, scale, output);
    }
    else Err(1, "gmm: unknown directive");
  }
#if 0
  else if(!strcmp(argv[1], "surface_repel_wh")) {
    Matrix sample;
    Vector radius;
    int nsample;
    Err(argc != 5, "usage: cluster surface_repel sample_matrix radius_vector nsample");
    GetPtr(argv[2], sample);
    GetPtr(argv[3], radius);
    GetInt(argv[4], nsample);
    SurfaceRepel(sample, radius, &nsample);
    ReturnInt(nsample);
  }
#endif
  else if(!strcmp(argv[1], "surface_repel")) {
    Matrix sample, vel;
    double noise, distance;
    int nsample;
    struct SurfaceLines surf;
    Err(argc < 9, "usage: cluster surface_repel sample_matrix vel_matrix noise nsample base offset mbr");
    GetPtr(argv[2], sample);
    GetPtr(argv[3], vel);
    GetDouble(argv[4], noise);
    GetInt(argv[5], nsample);
    if(argc > 6) {
      GetPtr(argv[6], surf.base);
      GetPtr(argv[7], surf.offset);
      GetPtr(argv[8], surf.mbr);
    }
    distance = SurfaceRepel(sample, vel, noise, &nsample, 
			 (ProjectFunc*)SurfaceProjectSquare, &surf);
    /* ReturnInt(nsample); */
    ReturnDouble(distance);
  }
  else {
    Err(1, "unknown directive");
  }
  return TCL_OK;
}


int Clustertcl_Init(Tcl_Interp *interp)
{
  Tcl_CreateCommand(interp, "cluster", ClusterCmd, NULL, NULL);
  return Tcl_PkgProvide(interp, "clusterTcl", "1");
}


