/* Definition of the cluster package. Add it to Tcl with
 *   Clustertcl_Init(interp);
 * Commands:
 *   em_init sample center cov
 *   em_step sample center cov mbr temperature
 */
#include <tcl.h>
#include "tcl_macros.h"
#include <tpm/cluster.h>

/* Prototypes ****************************************************************/
int Clustertcl_Init(Tcl_Interp *interp);

/* Functions *****************************************************************/

typedef struct SurfaceLines {
  Matrix base, offset, mbr;
} *SurfaceLines;

/* Modifies d so that s+d lies on the surface. */
/* n is number of dimensions. */
static void SurfaceProjectLines(double *d, double *s, int n, double feedback,
				int i, SurfaceLines surf)
{
  int c, k;
  double dot, f;
  Matrix base = surf->base, offset = surf->offset;
  double *mbr = surf->mbr->d2[i];
  double *grad = AllocateC(n, double);
#if 0
  /* compute dot product of d with surface */
  for(c=0;c<base->rows;c++) {
    dot = 0;
    for(k=0;k<n;k++) {
      dot += d[k] * offset->d2[c][k];
    }
    dot *= mbr[c];
    for(k=0;k<n;k++) {
      grad[k] += dot * offset->d2[c][k];
    }

    dot = d[0] * offset->d2[c][1] + d[1] * offset->d2[c][0];
    dot *= mbr[c]*0.5;
    grad[0] += dot * offset->d2[c][1];
    grad[1] += dot * offset->d2[c][0];
  }  
  for(k=0;k<n;k++) {
    d[k] = grad[k];
  }
  memset(grad, 0, n*sizeof(double));
#endif
  /* compute gradient vector */
  for(c=0;c<base->rows;c++) {
    /* compute (sample - base) dot offset */
    dot = 0;
    for(k=0;k<n;k++) {
      dot += (s[k] - base->d2[c][k]) * offset->d2[c][k];
    }
    for(k=0;k<n;k++) {
      grad[k] += mbr[c] * (s[k] - base->d2[c][k] - dot * offset->d2[c][k]);
    }
  }
  /* compute magnitude (distance to surface) and dot product with d */
  dot = f = 0;
  for(k=0;k<n;k++) {
    f += grad[k] * grad[k];
    dot += grad[k] * d[k];
  }
#if 0
  dot = 0;
#endif
  if(f > 0) {
    /* this is a simplification of 
       f = sqrt(f);
       (dot / f + feedback * f) * grad / f 
     */
    dot /= f;
    dot += feedback;
    for(k=0;k<n;k++) {
      d[k] -= dot * grad[k];
    }
  }
  free(grad);
}

/* Modifies d so that s+d lies on the surface. */
static void SurfaceProjectSquare(double *d, double *s, int n, double feedback,
				 int i, Any data)
{
  /* project onto a square (100,100)-(300,300) */
  double f = 0, dot = 0;
  double *dx = Allocate(n, double);
  int j;
  for(j=0;j<n;j++) {
    if(s[j] < 100) {
      /* left corner */
      dx[j] = s[j] - 100;
    }
    else if(s[j] > 300) {
      /* right corner */
      dx[j] = s[j] - 300;
    }
    else {
      /* horizontal edge */
      dx[j] = 0;
    }
    f += dx[j] * dx[j];
    dot += d[j] * dx[j];
  }
  if(f > 0) {
    dot /= f;
    dot += feedback;
    for(j=0;j<n;j++) {
      d[j] -= dot * dx[j];
    }
  }
  free(dx);
}

/* Modifies d so that s+d lies on the surface. */
static void SurfaceProjectCircle(double *d, double *s, int n, double feedback,
				 int i, Any data)
{
  /* project onto a circle */
  double dx = s[0] - 200;
  double dy = s[1] - 200;
  double r = sqrt(dx*dx + dy*dy);
  double f = r - 100;
  double dot;
  /* gradient is [dx/r dy/r] */
  dx /= r;
  dy /= r;
  dot = d[0] * dx + d[1] * dy;
  d[0] -= (dot + feedback * f) * dx;
  d[1] -= (dot + feedback * f) * dy;
}

static int ClusterCmd(ClientData clientData, Tcl_Interp *interp,
		      int argc, char *argv[])
{
  Err(argc < 2, "missing directive; try one of: em_E, em_M, em_V");
  if(!strcmp(argv[1], "em_E")) {
    Matrix sample, center, cov, mbr;
    double temperature;
    Err(argc != 7, "usage: cluster em_E sample_matrix center_matrix covariance_matrix mbr_matrix temperature");
    GetPtr(argv[2], sample);
    GetPtr(argv[3], center);
    GetPtr(argv[4], cov);
    GetPtr(argv[5], mbr);
    GetDouble(argv[6], temperature);
    ReturnDouble(Clus_EMStepE(sample, center, cov, mbr, temperature));
  }
  else if(!strcmp(argv[1], "em_EP")) {
    Matrix sample, center, cov, mbr;
    Vector prior;
    double temperature;
    Err(argc != 8, "usage: cluster em_EP sample_matrix center_matrix covariance_matrix prior_vector mbr_matrix temperature");
    GetPtr(argv[2], sample);
    GetPtr(argv[3], center);
    GetPtr(argv[4], cov);
    GetPtr(argv[5], prior);
    GetPtr(argv[6], mbr);
    GetDouble(argv[7], temperature);
    ReturnDouble(Clus_EMStepEP(sample, center, cov, prior, mbr, temperature));
  }
  else if(!strcmp(argv[1], "em_M")) {
    Matrix sample, center, mbr;
    double noise;
    Err(argc < 5, "usage: cluster em_M sample_matrix center_matrix mbr_matrix ?noise?");
    GetPtr(argv[2], sample);
    GetPtr(argv[3], center);
    GetPtr(argv[4], mbr);
    Clus_EMStepM(sample, center, mbr);
    if(argc >= 6) GetDouble(argv[5], noise);
    else noise = 0.01;
    VectorPerturb((Vector)center, noise);
  }
  else if(!strcmp(argv[1], "perturb")) {
    Vector v;
    double noise;
    Err(argc != 4, "usage: cluster perturb vector noise");
    GetPtr(argv[2], v);
    GetDouble(argv[3], noise);
    VectorPerturb(v, noise);
  }
  else if(!strcmp(argv[1], "em_V")) {
    Matrix sample, center, cov, mbr;
    Err(argc != 6, "usage: cluster em_V sample_matrix center_matrix covariance_matrix mbr_matrix");
    GetPtr(argv[2], sample);
    GetPtr(argv[3], center);
    GetPtr(argv[4], cov);
    GetPtr(argv[5], mbr);
    Clus_EMStepV(sample, center, cov, mbr);
  }
  else if(!strcmp(argv[1], "em_P")) {
    Vector prior;
    Matrix mbr;
    Err(argc != 4, "usage: cluster em_P prior_vector mbr_matrix");
    GetPtr(argv[2], prior);
    GetPtr(argv[3], mbr);
    Clus_EMStepP(prior, mbr);
  }
  else if(!strcmp(argv[1], "em_identity")) {
    Matrix cov;
    int i,j,n;
    Err(argc != 3, "usage: cluster em_identity cov_matrix");
    GetPtr(argv[2], cov);
    VectorSet((Vector)cov, 0);
    n = sqrt(cov->cols);
    for(i=0;i<cov->rows;i++) {
      for(j=0;j<n;j++) {
	cov->d2[i][j*n+j] = 1;
      }
    }
  }
  else if(!strcmp(argv[1], "em_classify")) {
    Matrix center, cov, output;
    Vector prior;
    double temperature;
    int freq, scale;
    Err(argc != 9, "usage: cluster em_classify center_matrix cov_matrix prior_vector temperature freq scale output_matrix");
    GetPtr(argv[2], center);
    GetPtr(argv[3], cov);
    GetPtr(argv[4], prior);
    GetDouble(argv[5], temperature);
    GetInt(argv[6], freq);
    GetInt(argv[7], scale);
    GetPtr(argv[8], output);
    Clus_EMClassify(center, cov, prior, temperature, freq, scale, output);
  }
  else if(!strcmp(argv[1], "em_distance")) {
    Matrix center, cov, output;
    Vector prior, point;
    double temperature;
    int scale;
    Err(argc != 9, "usage: cluster em_distance center_matrix cov_matrix prior_vector temperature point scale output_matrix");
    GetPtr(argv[2], center);
    GetPtr(argv[3], cov);
    GetPtr(argv[4], prior);
    GetDouble(argv[5], temperature);
    GetPtr(argv[6], point);
    GetInt(argv[7], scale);
    GetPtr(argv[8], output);
    Clus_EMDistance(center, cov, prior, temperature, point, scale, output);
  }
#if 0
  else if(!strcmp(argv[1], "surface_repel_wh")) {
    Matrix sample;
    Vector radius;
    int nsample;
    Err(argc != 5, "usage: cluster surface_repel sample_matrix radius_vector nsample");
    GetPtr(argv[2], sample);
    GetPtr(argv[3], radius);
    GetInt(argv[4], nsample);
    SurfaceRepel(sample, radius, &nsample);
    ReturnInt(nsample);
  }
#endif
  else if(!strcmp(argv[1], "surface_repel")) {
    Matrix sample, vel;
    double noise, distance;
    int nsample;
    struct SurfaceLines surf;
    Err(argc < 9, "usage: cluster surface_repel sample_matrix vel_matrix noise nsample base offset mbr");
    GetPtr(argv[2], sample);
    GetPtr(argv[3], vel);
    GetDouble(argv[4], noise);
    GetInt(argv[5], nsample);
    if(argc > 6) {
      GetPtr(argv[6], surf.base);
      GetPtr(argv[7], surf.offset);
      GetPtr(argv[8], surf.mbr);
    }
    distance = SurfaceRepel(sample, vel, noise, &nsample, 
			 (ProjectFunc*)SurfaceProjectSquare, &surf);
    /* ReturnInt(nsample); */
    ReturnDouble(distance);
  }
  else {
    Err(1, "unknown directive");
  }
  return TCL_OK;
}


int Clustertcl_Init(Tcl_Interp *interp)
{
  Tcl_CreateCommand(interp, "cluster", ClusterCmd, NULL, NULL);
  return Tcl_PkgProvide(interp, "clusterTcl", "1");
}


