set ncenters 3
unpack $argv ncenters

require cluster

set helpstring {\
---------------------- Interactive EM clustering ---------------------------

LMB places a point      MMB drags a point        RMB removes a point

Lines show one standard deviation along the major and minor axes of
the covariance matrix.

"Use covariance" lets covariance participate in E-step 
  (otherwise identity covariance, as in K-means).

"Use weighting" estimates priors for the Gaussians
  (otherwise uniform weights, as in K-means).

"Color membership" colors points according to their soft membership.
"Class image" does the same for all pixels.

When temperature is 0, membership (class posterior) is binary.
When temperature is 1, membership is determined as in EM.
Larger temperatures force more smoothing, causing clusters to merge.

"Step" is useful for kick-starting the algorithm after an error.
"Reset" clears the colors and initializes the cluster parameters.

Source code is in ~tpminka/src/cluster/em.c.
A paper about the temperature parameter is in ~tpminka/papers/dist-clus.ps.gz

I have a data set which causes severe convergence problems for the
algorithm, in cluster.pts (just hit "load points").  The smallest
distortion I've been able to get with six clusters is 36700 (67000
with four).  See if you can beat it.

Another interesting data set is a cross.  Can you guess what the
global minimum is for two clusters using covariance?

The Brodatz texture database in two dimensions is in "data/tev.pts".
EM gives fairly different answers depending on whether covariance
and/or weighting is on.  It takes a while to load.  You probably want to 
disable coloring.

Thomas Minka 9/20/96
}

# Globals #####################################################################
set center_radius 3
set center_width 1
set center_color green

# list of point ids
set points {}

# matrix of points
set sample [matrix create 0 2]

# mixture model
set gm [cluster gmm create $ncenters 2]

set bg_size 50
set class(1) [matrix create $bg_size $bg_size]
set class(2) [matrix create $bg_size $bg_size]
set class(4) [matrix create $bg_size $bg_size]

# colors
set red   {0 1 0 1 0 1 0 1}
set green {0 0 1 1 0 0 1 1}
set blue  {0 0 0 0 1 1 1 1}

proc number2color n {
    global red green blue
    return [format #%04x%04x%04x \
	    [expr [lindex $red $n]  *65535] \
	    [expr [lindex $green $n]*65535] \
	    [expr [lindex $blue $n] *65535]]
}
proc next_color {} {
    global color color_number red
    incr color_number
    if {$color_number > [llength $red]} {set color_number 0}
    set color [number2color $color_number]
}
set color_number -1
next_color

# Canvas bindings #############################################################
proc add_cb {id x y} {
    # add the id to the point list
    global points
    lappend points $id

    # add the coordinates to the end of the sample matrix
    global sample
    set new_sample [matrix_append $sample $x $y]
    matrix free $sample
    set sample $new_sample
    #puts [matrix get $sample]

    # extend the mbr matrix
    global mbr ncenters
    if {[info exists mbr]} {matrix free $mbr}
    set mbr [matrix create [matrix rows $sample] $ncenters]
}

proc move_cb {id x y} {
    global points
    set index [lsearch -exact $points $id]

    # update the matrix
    global sample
    matrix_row $sample $index $x $y
}

proc delete_cb {id} {
    # which point is this?
    global points
    set index [lsearch -exact $points $id]

    # remove from point list
    set points [lreplace $points $index $index]
    # remove from samples
    global sample
    set new_sample [matrix delrow $sample $index]
    matrix free $sample
    set sample $new_sample
    # remove from mbr
    global mbr
    set new_mbr [matrix delrow $mbr $index]
    matrix free $mbr
    set mbr $new_mbr
}

pointcanvas .canvas -background black -width 300 -height 300 \
  -scrollregion {-100 0 400 400} -radius 4 \
  -add_cb add_cb -move_cb move_cb -delete_cb delete_cb
pack .canvas -fill both -expand 1

label .info -text "Left button places a point   Middle button drags a point   Right button removes a point"
pack .info -fill x 
#-expand 1

proc dot {x y radius args} {
    set dot [eval .canvas create oval 0 0 0 0 -width 1 -outline blue $args]
    oval_pos $dot $x $y $radius $radius
    return $dot
}

# create centers
for {set i 0} {$i < $ncenters} {incr i} {
    dot 0 0 $center_radius -fill $center_color -tags center$i
    if 0 {
	.canvas create oval 0 0 0 0 -width $center_width -outline $center_color -tags oval$i
    } else {
	.canvas create line 0 0 0 0 -width $center_width -fill $center_color -tags minor$i
	.canvas create line 0 0 0 0 -width $center_width -fill $center_color -tags major$i
    }
}

# background image
set bg_image [image create photo]
.canvas create image 0 0 -anchor nw -image $bg_image

# mouse position
set mouse [vector create 2]
bind .canvas <Motion> {
    vector set $mouse [list [.canvas canvasx %x] [.canvas canvasy %y]]
}

# Controls ####################################################################
frame .controlf
pack .controlf -fill x -side bottom

# Cluster controls ############################################################
frame .checkf
pack .checkf -in .controlf -fill x -expand 1

# if set on by default, get an error on the first point
checkbutton .use_cov -text "Use covariance" -variable use_cov
pack .use_cov -fill x -expand 1 -in .checkf -side left

checkbutton .use_prior -text "Use weighting" -variable use_prior
pack .use_prior -fill x -expand 1 -in .checkf -side left

checkbutton .use_color -text "Color membership" -variable use_color
pack .use_color -fill x -expand 1 -in .checkf -side left
set use_color 0

checkbutton .use_class -text "Class image" -variable use_class
pack .use_class -fill x -expand 1 -in .checkf -side left

frame .temp
pack .temp -fill x -expand 1 -in .controlf
label .temp.label -text Temperature
label .temp.value -textvariable temperature -font fixed
proc set_temp t {
    global temperature gm
    set temperature [format %9.2f [expr exp($t*1.05)-1]]
    cluster gmm temperature $gm $temperature
}
scale .temp.scale -from 0 -to 10 -resolution 0.01 -orient horiz -command set_temp -showvalue no
pack .temp.scale -side left -fill x -expand 1
pack .temp.label .temp.value -side left

frame .like
pack .like -fill x -expand 1 -in .temp
label .like.label -text Likelihood
label .like.value -textvariable likelihood -font fixed
pack .like.label .like.value -side left

frame .noise
pack .noise -fill x -expand 1 -in .controlf
label .noise.label -text "Parameter noise (means only)"
label .noise.value -textvariable noise
proc set_noise t {
    global temperature
    set temperature [expr exp($t*1.05)]
}
scale .noise.scale -from 0 -to 100 -resolution 1 -orient horiz -variable noise -showvalue no
pack .noise.scale -side left -fill x -expand 1
pack .noise.label .noise.value -side left

frame .buttonf
pack .buttonf -in .controlf -fill x -expand 1
button .step -text Step -command clus_one_step
pack .step -in .buttonf -side left -fill x -expand 1

set show_distance 0
button .dist -text Distance -command {
    set show_distance [expr ! $show_distance]
    if {$show_distance} {
	update_distance
    }
}
pack .dist -in .buttonf -side left -fill x -expand 1

button .init -text Reset -command {
    vector fill $mbr [expr 1.0/[matrix cols $mbr]]
    vector fill [cluster gmm prior $gm] [expr 1.0/$ncenters]
    cluster gmm trainM $gm $sample $mbr
    cluster gmm identity_cov $gm
    cluster perturb [cluster gmm center $gm] $noise
    $bg_image blank
    if 0 {
	foreach id $points {
	    .canvas itemconfigure $id -fill black
	}
    }
}
pack .init -in .buttonf -side left -fill x -expand 1

button .help -text "Help" -command {
    toplevel .h
    text .h.text -height 45
    .h.text insert end $helpstring
    .h.text configure -state disabled
    pack .h.text -fill both -expand 1
    button .h.dismiss -text Dismiss -command {destroy .h}
    pack .h.dismiss -fill x -expand 1
}
pack .help -in .buttonf -side left -fill x -expand 1

button .quit -text "Quit" -command exit
pack .quit -in .buttonf -side left -fill x -expand 1

# Load points #################################################################
proc remove_all {} {
    global points sample
    .canvas delete withtag point
    set points {}
    set sample [matrix create 0 2]
}

frame .iof
pack .iof -fill x -expand 1 -in .controlf
button .clearpts -text "Clear" -command {remove_all}
button .loadpts -text "Load" -command {loadpts [.ptfile get]}

proc loadpts fname {
    if 1 {
	remove_all
    } else {
	next_color
    }

    set fp [open $fname]
    while {[gets $fp line] >= 0} {
	eval pointcanvas_add .canvas $line
    }
    close $fp
}
proc savepts fname {
  puts "Save Points is disabled for this demo"
  return
  global point
  set fp [open $fname w]
  foreach pt [.canvas find withtag point] {
    unpack [.canvas coords $pt] x1 y1 x2 y2
    puts $fp "[expr ($x2+$x1)/2] [expr ($y2+$y1)/2]"
  }
  close $fp
}
button .savepts -text "Save" -command {savepts [.ptfile get]}
entry .ptfile -font fixed
.ptfile insert 0 "cluster.pts"

if 0 {
    pack .clearpts .loadpts .savepts .ptfile -side left -in .iof -fill x -expand 1
} else {
    listbox .filelist
    set files [glob *.pts]
    foreach f $files {
	.filelist insert end $f
    }
    bind .filelist <Double-1> {
	loadpts [.filelist get [.filelist curselection]]
    }
    pack .filelist -in .iof -fill x -expand 1
}

# Clustering ##################################################################

proc clus_one_step {} {
    global sample ncenters gm mbr noise use_cov use_prior likelihood
    if {[info exists mbr]} {
	if {$use_cov == 0} {
	    # clear covariances
	    cluster gmm identity_cov $gm
	}
	if {$use_prior == 0} {
	    vector fill [cluster gmm prior $gm] [expr 1.0/$ncenters]
	}
	set likelihood [format %9.2f [cluster gmm trainE $gm $sample $mbr]]
	cluster gmm trainM $gm $sample $mbr 
	cluster perturb [cluster gmm center $gm] $noise
	cluster gmm trainV $gm $sample $mbr
	cluster gmm trainP $gm $mbr

	# draw the new centers
	for {set c 0} {$c < $ncenters} {incr c} {
	    unpack [matrix_row [cluster gmm center $gm] $c] x y
	    oval_pos center$c $x $y
	    .canvas raise center$c
	    if 0 {
		# show diagonal covariance as an ellipse
		unpack [matrix_row $cov $i] xradius yradius
		set xradius [expr sqrt($xradius)]
		set yradius [expr sqrt($yradius)]
		.canvas coords oval$i \
			[expr $x-$xradius] [expr $y-$yradius] \
			[expr $x+$xradius] [expr $y+$yradius]
	    } else {
		# show rotated covariance as major and minor axis
		unpack [matrix_row [cluster gmm cov $gm] $c] v00 v01 v10 v11
		set e0 [vector create 2]
		set e1 [vector create 2]
		vector set $e0 "$v00 $v01"
		vector set $e1 "$v10 $v11"
		set lambda0 [expr 1.0/[vector mag $e0]]
		set lambda1 [expr 1.0/[vector mag $e1]]
		vector scale $e0 [expr $lambda0*$lambda0]
		vector scale $e1 [expr $lambda1*$lambda1]
		.canvas coords minor$c \
			[expr $x-[vector at $e0 0]] \
			[expr $y-[vector at $e0 1]] \
			[expr $x+[vector at $e0 0]] \
			[expr $y+[vector at $e0 1]]
		.canvas raise minor$c
		.canvas coords major$c \
			[expr $x-[vector at $e1 0]] \
			[expr $y-[vector at $e1 1]] \
			[expr $x+[vector at $e1 0]] \
			[expr $y+[vector at $e1 1]]
		.canvas raise major$c
		vector free $e0
		vector free $e1
	    }
	}

	global use_color red ncenters
	if {$use_color && ($ncenters <= [llength $red])} {
	    # color the points according to their membership
	    global points red green blue
	    for {set i 0} {$i < [matrix rows $mbr]} {incr i} {
		set id [lindex $points $i]
		set r 0
		set g 0
		set b 0
		for {set c 0} {$c < [matrix cols $mbr]} {incr c} {
		    set m [matrix at $mbr $i $c]
		    set r [expr $r+[lindex $red $c]*$m]
		    set g [expr $g+[lindex $green $c]*$m]
		    set b [expr $b+[lindex $blue $c]*$m]
		}
		set r [expr int($r*65535)]
		set g [expr int($g*65535)]
		set b [expr int($b*65535)]
		.canvas itemconfigure $id -fill [format #%04x%04x%04x $r $g $b]
	    }
	}

	global use_class class bg_image
	if {$use_class} {
	    set h [lindex [.canvas configure -height] 4]
	    set zoom [expr $h/[matrix rows $class(1)]]
	    foreach freq {1 2 4} {
		cluster gmm classify_region $gm $freq $zoom $class($freq)
		vector scale $class($freq) 255
	    }
	    set temp_image [image create photo]
	    if {$ncenters == 2} {
		matrix image $class(1) $temp_image
	    } else {
		matrix rgb $class(1) $class(2) $class(4) $temp_image
	    }
	    $bg_image copy $temp_image -zoom $zoom
	    image delete $temp_image
	}
    }
    after 100 clus_one_step
}

proc update_distance {} {
    global gm
    global class bg_image mouse

    set h [lindex [.canvas configure -height] 4]
    set zoom [expr $h/[matrix rows $class(1)]]

    cluster gmm distance_region $gm $mouse $zoom $class(1)

    #vector scale $class(1) 255
    vector rescale $class(1) 0 255
    matrix_zoom $class(1) $bg_image $zoom
    
    global show_distance
    if {$show_distance} {
	after 100 update_distance
    }
}

set temperature 1.0
set likelihood 0.0
set noise 2

clus_one_step

puts "LMB places a point"
puts "MMB drags a point"
puts "RMB removes a point"

