#ifndef TPMIMAGE_H_INCLUDED
#define TPMIMAGE_H_INCLUDED

#include <stdio.h>

/* designed to be byte-compatible with a MatrixC */
typedef struct TpmImage {
  unsigned char *d;
  int len;
  int width;
  unsigned char **d2;
  int hc, height;
  unsigned char ***d3;
  int channels;
} *TpmImage;

/* i1 gets the contents of i2. i1 must be already allocated and have exactly
 * the same dimensions as i2.
 */
#define TpmImageMove(i1, i2) memcpy((i1)->d, (i2)->d, (i2)->len)

TpmImage TpmImageCreate(int width, int height, int channels);
TpmImage TpmImageFromData(int width, int height, int channels, 
			  unsigned char *data);
void TpmImageFree(TpmImage image);
TpmImage TpmImageCopy(TpmImage image);
void TpmImageMoveTo(TpmImage dest, TpmImage src, int toX, int toY);
void TpmImageMinMax(TpmImage image, int *min_return, int *max_return);
void TpmImageRescale(TpmImage image, 
		     unsigned char new_min, unsigned char new_max);

void TpmImagePutPixel(TpmImage image, int x, int y, 
		      unsigned char red, unsigned char green, 
		      unsigned char blue);
void TpmImagePutRow(TpmImage image, int x, int y, int width,
		    unsigned char red, unsigned char green, 
		    unsigned char blue);
void TpmImagePutBlock(TpmImage image, int x, int y, int width, int height,
		      unsigned char red, unsigned char green, 
		      unsigned char blue);
TpmImage TpmImageZoom(TpmImage image, int zfact);
TpmImage TpmImageGray(TpmImage image);
int *TpmImageHistogram(TpmImage image);

/* colorspace ****************************************************************/

void RGB2Ohta(unsigned char r, unsigned char g, unsigned char b,
	      unsigned char *i1, unsigned char *i2, unsigned char *i3);
void Ohta2RGB(unsigned char i1, unsigned char i2, unsigned char i3,
	      unsigned char *r, unsigned char *g, unsigned char *b);
void TpmImageRGB2Ohta(TpmImage image);
void TpmImageOhta2RGB(TpmImage image);
void RGB2HSV(unsigned char r, unsigned char g, unsigned char b,
	     unsigned char *h, unsigned char *s, unsigned char *v);
void HSV2RGB(unsigned char h, unsigned char s, unsigned char v,
	     unsigned char *r, unsigned char *g, unsigned char *b);
void TpmImageRGB2HSV(TpmImage image);
void TpmImageHSV2RGB(TpmImage image);
void RGB2Cone(unsigned char r, unsigned char g, unsigned char b,
	      unsigned char *c, unsigned char *s, unsigned char *v);
void Cone2RGB(unsigned char c, unsigned char s, unsigned char v,
	      unsigned char *r, unsigned char *g, unsigned char *b);
void TpmImageRGB2Cone(TpmImage image);

/* feature *******************************************************************/

typedef void TpmImageFunc(void *data, TpmImage image);

void TpmImageWindows(TpmImage image, int xsize, int ysize, int xinc, int yinc,
		     TpmImageFunc *f, void *data);
TpmImage TpmImageQuantizeChannels(TpmImage image, int nbins);
TpmImage TpmImageQuantizeCube(TpmImage image, int rbins, int gbins, int bbins);
TpmImage TpmImageUnQuantizeCube(TpmImage image, int rbins, int gbins, int bbins);
#ifdef MATRIX_H_INCLUDED
#ifdef TPM_REGION_H_INCLUDED
void TpmImageRegionHistogram(TpmImage image, Tpm_Region *region, Vector result);
#endif
void TpmImageHist(TpmImage image, Vector result);
void TpmImageHistWindows(TpmImage image, int xsize, int ysize, 
			 int xinc, int yinc, Matrix result);
void TpmImageAutoCorrelogram(TpmImage image, int spacing[], Matrix result);
void TpmImageAutoCorrelogramWindows(TpmImage image, int spacing[], 
				    int num_spacings,
				    int xsize, int ysize, 
				    int xinc, int yinc, Matrix result);

void TpmImageSynthHist(TpmImage image, Vector hist, int nbins);
void TpmImageSynthHistClus(TpmImage image, Vector center_vec, int nbins);
void TpmImageSynthHistHSV(TpmImage image, Vector hist, int nbins);
void TpmImageSynthBars(TpmImage image, Vector v, 
		       double minimum, double maximum);
#endif

/* Matrix ********************************************************************/

#ifdef MATRIX_H_INCLUDED
TpmImage TpmImageFromMatrix(Matrix m);
Matrix MatrixFromImage(TpmImage image);
void CoocMatrixFromImage(TpmImage image, int dx, int dy, Matrix cooc);
#endif
#ifdef VOLUME_H_INCLUDED
TpmImage TpmImageFromVolume(Volume v);
Volume VolumeFromImage(TpmImage image);
#endif

/* X *************************************************************************/

#ifdef XTPM_APP_H
Pixmap TpmImageToPixmap(XtpmAppData appData, TpmImage image);
void TpmImageFillXImage(XImage *ximage, TpmImage image);
#endif

/* TIFF **********************************************************************/

TpmImage TpmReadTiff(char *fname);
TpmImage TpmReadTiffStream(FILE *fp);
#ifdef _TIFFIO_
TpmImage TpmImageFromTiff(TIFF *tif);
#endif
int TpmImageWriteTiff(FILE *fp, TpmImage image);

/* PNM ***********************************************************************/

typedef struct {
  int cols, rows, maxval, format;
  FILE *fp;
} Pnm;

TpmImage TpmReadPnm(char *fname);
TpmImage TpmReadPnmStream(FILE *fp);
int PnmReadInit(Pnm *pnm);
TpmImage TpmImageFromPnm(Pnm *pnm);
#define PnmClose(pnm) if((pnm)->fp) fclose((pnm)->fp); else
int TpmImageWritePnm(FILE *fp, TpmImage image);

/* JPEG **********************************************************************/

typedef struct {
  int width, height, channels;
  FILE *fp;
#ifdef JPEGLIB_H
  struct jpeg_decompress_struct *cinfo;
#else
  void *cinfo;
#endif
} Jpeg;

TpmImage TpmReadJpeg(char *fname);
TpmImage TpmReadJpegStream(FILE *fp);
void JpegClose(Jpeg *jpeg);
int JpegReadInit(Jpeg *jpeg);
TpmImage TpmImageFromJpeg(Jpeg *jpeg);
int TpmImageWriteJpeg(FILE *fp, TpmImage image);

/* TK ************************************************************************/

#ifdef TK_VERSION
void TpmImageFillTk(Tk_PhotoHandle photo, TpmImage image);
TpmImage TpmImageFromTk(Tk_PhotoHandle photo);
#endif

/* SGI ***********************************************************************/

TpmImage TpmReadSgi(char *fname);
#ifdef __GL_IMAGE_H__
TpmImage TpmImageFromSgi(IMAGE *img);
#endif
int TpmImageWriteSgi(FILE *fp, TpmImage image);

#endif
