#include <tpm/tpm.h>
#include "image.h"

/* Prototypes ****************************************************************/

void RGB2Ohta(unsigned char r, unsigned char g, unsigned char b,
	      unsigned char *i1, unsigned char *i2, unsigned char *i3);
void Ohta2RGB(unsigned char i1, unsigned char i2, unsigned char i3,
	      unsigned char *r, unsigned char *g, unsigned char *b);
void TpmImageRGB2Ohta(TpmImage image);
void TpmImageOhta2RGB(TpmImage image);
void RGB2HSV(unsigned char r, unsigned char g, unsigned char b,
	     unsigned char *h, unsigned char *s, unsigned char *v);
void HSV2RGB(unsigned char h, unsigned char s, unsigned char v,
	     unsigned char *r, unsigned char *g, unsigned char *b);
void TpmImageRGB2HSV(TpmImage image);
void TpmImageHSV2RGB(TpmImage image);
void RGB2Cone(unsigned char r, unsigned char g, unsigned char b,
	      unsigned char *c, unsigned char *s, unsigned char *v);
void Cone2RGB(unsigned char c, unsigned char s, unsigned char v,
	      unsigned char *r, unsigned char *g, unsigned char *b);
void TpmImageRGB2Cone(TpmImage image);

/* Functions *****************************************************************/

void RGB2Ohta(unsigned char r, unsigned char g, unsigned char b,
	      unsigned char *i1, unsigned char *i2, unsigned char *i3)
{
  *i1 = ((float)r + g + b)/3 + 0.5;
  *i2 = ((float)r - b + 256)/2 + 0.5;
  *i3 = ((float)r - 2*g + b + 512)/4 + 0.5;
}

void Ohta2RGB(unsigned char i1, unsigned char i2, unsigned char i3,
	      unsigned char *r, unsigned char *g, unsigned char *b)
{
  *r = max(0, min(255, (float)i1 + i2 + ((float)i3*2 - 256)/3 - 128 + 0.5));
#if 0
  *g = (int)i1 - ((int)i3*4 + 512)/3;
  *b = (int)i1 - i2 + ((int)i3*2 - 256)/3 + 128;
#else
  *g = max(0, min(255, (float)*r - i2 - i3*2 + 384));
  *b = max(0, min(255, (float)*r - 2*i2 + 256));
#endif
}

void TpmImageRGB2Ohta(TpmImage image)
{
  int y, x;
  assert(image->channels == 3);
  for(y=0;y<image->height;y++) {
    for(x=0;x<image->width;x++) {
      RGB2Ohta(image->d3[0][y][x], image->d3[1][y][x], image->d3[2][y][x],
	       &image->d3[0][y][x], &image->d3[1][y][x], &image->d3[2][y][x]);
    }
  }
}

void TpmImageOhta2RGB(TpmImage image)
{
  int y, x;
  assert(image->channels == 3);
  for(y=0;y<image->height;y++) {
    for(x=0;x<image->width;x++) {
      Ohta2RGB(image->d3[0][y][x], image->d3[1][y][x], image->d3[2][y][x],
	       &image->d3[0][y][x], &image->d3[1][y][x], &image->d3[2][y][x]);
    }
  }
}

#if 1
void RGB2HSV(unsigned char r, unsigned char g, unsigned char b,
	     unsigned char *h, unsigned char *s, unsigned char *v)
{
  float maximum, minimum, range, hue;
  if(r > g) { maximum = r; minimum = g; }
  else { maximum = g; minimum = r; }
  if(b > maximum) { maximum = b; }
  else if(b < minimum) { minimum = b; }

  *v = maximum;
  range = maximum - minimum;
  if(maximum == 0) *s = 0;
  else *s = range*255/maximum+0.5;
  if(*s == 0) { *h = 0; return; }

  if(r == maximum) hue = (g - b)/range/6;
  else if(g == maximum) hue = (2 + (b - r)/range)/6;
  else hue = (4 + (r - g)/range)/6;
  if(hue < 0) hue += 1;
  *h = hue*255+0.5;
}
#else
/* Taken from Matt Miller */
static void mlm_RGB2HSV( double R, double G, double B,
			 double *H, double *S, double *V )
{
#ifdef min
#undef min
#endif
#ifdef max
#undef max
#endif
  double min, mid, max;
  enum { RGB, GRB, GBR, BGR, BRG, RBG } order;
  int baseH;
  double fractH;

  #define SET_MMMORDER( v0, v1, v2, h ) \
          { \
            max = v0; \
            mid = v1; \
            min = v2; \
            order = v0##v1##v2; \
            baseH = h; \
          }

  if( R > G )
    if( G > B )
      SET_MMMORDER( R, G, B, 0 )
    else if( R > B )
      SET_MMMORDER( R, B, G, 5 )
    else
      SET_MMMORDER( B, R, G, 4 )
  else
    if( R > B )
      SET_MMMORDER( G, R, B, 1 )
    else if( G > B )
      SET_MMMORDER( G, B, R, 2 )
    else
      SET_MMMORDER( B, G, R, 3 )

  #undef SET_MMMORDER

  *H = 180.;
  *S = 1.;

  *V = max;
  if( *V == 0. ) return;

  *S = 1. - min / max;
  if( *S == 0. ) return;

  if( baseH & 1 )
    fractH = 1. - (mid - min) / (max - min);
  else
    fractH = (mid - min) / (max - min);
  *H = (baseH + fractH) / 6.;
}

void RGB2HSV(unsigned char r, unsigned char g, unsigned char b,
	     unsigned char *h, unsigned char *s, unsigned char *v)
{
  double H, S, V;
  mlm_RGB2HSV(r, g, b, &H, &S, &V);
  *h = H*256+0.5;
  *s = S*256+0.5;
  *v = V;
}
#endif

void HSV2RGB(unsigned char h, unsigned char s, unsigned char v,
	     unsigned char *r, unsigned char *g, unsigned char *b)
{
  int baseH;
  unsigned char p, q, t;
  float S = (float)s/255;
  float fractH;
  if(s == 0) {
    *r = *g = *b = v;
    return;
  }
  fractH = (float)h/255*6;
  baseH = (int)fractH;
  fractH -= baseH;

  p = v*(1-S)+0.5;
  q = v*(1-S*fractH)+0.5;
  t = v*(1-S*(1-fractH))+0.5;
  switch(baseH) {
  case 0: *r = v; *g = t; *b = p; break;
  case 1: *r = q; *g = v; *b = p; break;
  case 2: *r = p; *g = v; *b = t; break;
  case 3: *r = p; *g = q; *b = v; break;
  case 4: *r = t; *g = p; *b = v; break;
  case 5: *r = v; *g = p; *b = q; break;
  default: fprintf(stderr, "hsv = (%d, %d, %d)\n", h, s, v); assert(0);
  }
}

void TpmImageRGB2HSV(TpmImage image)
{
  int y, x;
  assert(image->channels == 3);
  for(y=0;y<image->height;y++) {
    for(x=0;x<image->width;x++) {
      RGB2HSV(image->d3[0][y][x], image->d3[1][y][x], image->d3[2][y][x],
	      &image->d3[0][y][x], &image->d3[1][y][x], &image->d3[2][y][x]);
    }
  }
}

void TpmImageHSV2RGB(TpmImage image)
{
  int y, x;
  for(y=0;y<image->height;y++) {
    for(x=0;x<image->width;x++) {
      HSV2RGB(image->d3[0][y][x], image->d3[1][y][x], image->d3[2][y][x],
	      &image->d3[0][y][x], &image->d3[1][y][x], &image->d3[2][y][x]);
    }
  }
}

void RGB2Cone(unsigned char r, unsigned char g, unsigned char b,
	      unsigned char *c, unsigned char *s, unsigned char *v)
{
  unsigned char H, S, V;
  float f;
  RGB2HSV(r, g, b, &H, &S, &V);
  f = (float)S/255;
  *c = f*V*cos((double)H/255*2*M_PI)/2 + 128 + 0.5;
  *s = f*V*sin((double)H/255*2*M_PI)/2 + 128 + 0.5;
  *v = V;
}

void Cone2RGB(unsigned char c, unsigned char s, unsigned char v,
	      unsigned char *r, unsigned char *g, unsigned char *b)
{
  unsigned char S;
  float H = atan2(2*((double)s-128), 2*((double)c-128));
  if(H == 0) {
    S = min(1, 2*((double)c-128) / v) * 255 + 0.5;
  }
  else {
    S = min(1, 2*((double)s-128) / sin(H) / v) * 255 + 0.5;
  }
  if(H < 0) H += 2*M_PI;
  HSV2RGB(H/2/M_PI*255+0.5, S, v, r, g, b);
}

void TpmImageRGB2Cone(TpmImage image)
{
  int y, x;
  assert(image->channels == 3);
  for(y=0;y<image->height;y++) {
    for(x=0;x<image->width;x++) {
      RGB2Cone(image->d3[0][y][x], image->d3[1][y][x], image->d3[2][y][x],
	       &image->d3[0][y][x], &image->d3[1][y][x], &image->d3[2][y][x]);
    }
  }
}
