#include <tpm/tpm.h>
#include <tpm/matrix.h>
#include <tpm/region.h>
#include "image.h"

/* Prototypes ****************************************************************/

void TpmImageWindows(TpmImage image, int xsize, int ysize, int xinc, int yinc,
		     TpmImageFunc *f, Any data);
TpmImage TpmImageQuantizeChannels(TpmImage image, int nbins);
TpmImage TpmImageQuantizeCube(TpmImage image, int rbins, int gbins, int bbins);
TpmImage TpmImageUnQuantizeCube(TpmImage image, int rbins, int gbins, int bbins);
void TpmImageHist(TpmImage image, Vector result);
void TpmImageRegionHistogram(TpmImage image, Tpm_Region *region, Vector result);
void TpmImageHistWindows(TpmImage image, int xsize, int ysize, 
			 int xinc, int yinc, Matrix result);
void TpmImageAutoCorrelogram(TpmImage image, int spacing[], Matrix result);
void TpmImageAutoCorrelogramWindows(TpmImage image, int spacing[], 
				    int num_spacings,
				    int xsize, int ysize, 
				    int xinc, int yinc, Matrix result);

void TpmImageSynthHist(TpmImage image, Vector hist, int nbins);
void TpmImageSynthHistClus(TpmImage image, Vector center_vec, int nbins);
void TpmImageSynthHistHSV(TpmImage image, Vector hist, int nbins);
void TpmImageSynthBars(TpmImage image, Vector v, 
		       double minimum, double maximum);

/* Functions *****************************************************************/

/* Calls f(data, window) on subwindows of image. */
void TpmImageWindows(TpmImage image, int xsize, int ysize, int xinc, int yinc,
		     TpmImageFunc *f, Any data)
{
  TpmImage window = TpmImageCreate(xsize, ysize, image->channels);
  int x, y, c, i;
  int xhi, yhi;

  xhi = image->width - xsize;
  yhi = image->height - ysize;
  assert(xhi % xinc == 0);
  assert(yhi % yinc == 0);
  for(y=0;y<=yhi;y+=yinc) {
    for(x=0;x<=xhi;x+=xinc) {
      /* copy the rectangle whose upper-left corner is (x,y) into window. */
      for(c=0;c<image->channels;c++) {
	for(i=0;i<ysize;i++) {
	  memcpy(window->d3[c][i], &image->d3[c][y+i][x], xsize);
	}
      }
      f(data, window);
    }
  }
  TpmImageFree(window);
}

TpmImage TpmImageQuantizeChannels(TpmImage image, int nbins)
{
  TpmImage result;
  int c, y, x;
  result = TpmImageCreate(image->width, image->height, image->channels);
  for(c=0;c<image->channels;c++) {
    for(y=0;y<image->height;y++) {
      for(x=0;x<image->width;x++) {
	result->d3[c][y][x] = image->d3[c][y][x]*nbins/256;
      }
    }
  }
  return result;
}

static void QuantizeCube(int rbins, int gbins, int bbins,
			 unsigned char r, unsigned char g, unsigned char b,
			 unsigned char *q)
{
  *q = r*rbins/256;
  *q += g*gbins/256*rbins;
  *q += b*bbins/256*gbins*rbins;
}

static void UnQuantizeCube(int rbins, int gbins, int bbins,
			   unsigned char q,
			   unsigned char *r, unsigned char *g, unsigned char *b)
{
  *b = (double)q / rbins / gbins * 256 / bbins;
  q = q % (rbins * gbins);
  *g = (double)q / rbins * 256 / gbins;
  q = q % rbins;
  *r = (double)q * 256 / rbins;
}

/* Returns a single-channel image with nbins^channels 
 * different pixel values. 
 */
TpmImage TpmImageQuantizeCube(TpmImage image, int rbins, int gbins, int bbins)
{
  TpmImage result;
  int y, x;
  result = TpmImageCreate(image->width, image->height, 1);
  for(y=0;y<image->height;y++) {
    for(x=0;x<image->width;x++) {
#if 1
      QuantizeCube(rbins, gbins, bbins,
		   image->d3[0][y][x], image->d3[1][y][x], image->d3[2][y][x],
		   &result->d2[y][x]);
#else
      int nbins = rbins;
      result->d3[0][y][x] = image->d3[0][y][x]*nbins/256;
      if(image->channels > 1) {
	result->d3[0][y][x] +=
	  image->d3[1][y][x]*nbins/256*nbins +
	  image->d3[2][y][x]*nbins/256*nbins*nbins;
      }
#endif
    }
  }
  return result;
}

TpmImage TpmImageUnQuantizeCube(TpmImage image, int rbins, int gbins, int bbins)
{
  TpmImage result;
  int y, x;
  result = TpmImageCreate(image->width, image->height, 3);
  for(y=0;y<image->height;y++) {
    for(x=0;x<image->width;x++) {
      UnQuantizeCube(rbins, gbins, bbins, image->d2[y][x],
	     &result->d3[0][y][x], &result->d3[1][y][x], &result->d3[2][y][x]);
    }
  }
  return result;
}

/* Each row of result, interpreted as a Matrix, is filled with the 
 * histogram for the corresponding channel.
 */
void TpmImageHist(TpmImage image, Vector result)
{
  int x, y, c;
  int cols = result->len / image->channels;
  double *p = result->d;
  VectorSet((Vector)result, 0);
  for(c=0;c<image->channels;c++) {
    for(y=0;y<image->height;y++) {
      for(x=0;x<image->width;x++) {
	if(result->len <= image->d3[c][y][x]) {
	  fprintf(stderr, "TpmImageHist: image value (%d) exceeds result->len (%d)\n",
		  image->d3[c][y][x], result->len);
	  continue;
	}
	p[ image->d3[c][y][x] ]++;
      }
    }
    p += cols;
  }
}

/* image must be single-channel. */
void TpmImageRegionHistogram(TpmImage image, Tpm_Region *region, Vector result)
{
  int x, y;
  assert(image->channels == 1);
  VectorSet(result, 0);
  Tpm_RegionIter(region, x, y) {
    result->d[ image->d2[y][x] ]++;
  }
}

/* hist should be a PMF. */
void TpmImageSynthHist(TpmImage image, Vector hist, int nbins)
{
  int x, y, count = 0, bin = 0;
  double size = image->height*image->width;
  int limit = hist->d[0] * size;
  for(y=0;y<image->height;y++) {
    for(x=0;x<image->width;x++) {
      while(count >= limit) {
	bin++;
	limit += hist->d[bin] * size;
      }
      UnQuantizeCube(nbins, nbins, nbins, bin,
	       &image->d3[0][y][x], &image->d3[1][y][x], &image->d3[2][y][x]);
      count++;
    }
  }
  if(0) {
    /* check */
    Vector v = VectorCreate(hist->len);
    TpmImage q_image = TpmImageQuantizeCube(image, nbins, nbins, nbins);
    TpmImageHist(q_image, v);
    VectorScale(v, 1.0/q_image->len);
    TpmImageFree(q_image);
    fprintf(stderr, "histogram (%g): ", VectorSum(hist));
    VectorPrint(stderr, hist);
    fprintf(stderr, "synth     (%g): ", VectorSum(v));
    VectorPrint(stderr, v);
    VectorFree(v);
  }
}

/* hist should be a PMF. */
void TpmImageSynthHistHSV(TpmImage image, Vector hist, int nbins)
{
  int x, y, count = 0, pos = 0;
  double size = image->height*image->width;
  int limit = hist->d[0] * size;
  for(y=0;y<image->height;y++) {
    for(x=0;x<image->width;x++) {
      int bin;
#if 0
      bin = VectorSample(hist);
#else
      while(count >= limit) {
	pos++;
	limit += hist->d[pos] * size;
      }
      bin = pos;
#endif
      UnQuantizeCube(nbins, nbins, nbins, bin,
	       &image->d3[0][y][x], &image->d3[1][y][x], &image->d3[2][y][x]);
      HSV2RGB(image->d3[0][y][x], image->d3[1][y][x], image->d3[2][y][x],
	      &image->d3[0][y][x], &image->d3[1][y][x], &image->d3[2][y][x]);
      count++;
    }
  }
  if(0) {
    /* check */
    Vector v = VectorCreate(hist->len);
    TpmImage q_image;
    TpmImage hsv_image = TpmImageCopy(image);
    TpmImageRGB2HSV(hsv_image);
    q_image = TpmImageQuantizeCube(hsv_image, nbins, nbins, nbins);
    TpmImageFree(hsv_image);
    TpmImageHist(q_image, v);
    VectorScale(v, 1.0/q_image->len);
    TpmImageFree(q_image);
    fprintf(stderr, "histogram (%g): ", VectorSum(hist));
    VectorPrint(stderr, hist);
    fprintf(stderr, "synth     (%g): ", VectorSum(v));
    VectorPrint(stderr, v);
    VectorFree(v);
  }
}

/* stacks the synthesized images for the clusters on top of eachother. */
void TpmImageSynthHistClus(TpmImage image, Vector center_vec, int nbins)
{
  int hist_len = nbins*nbins*nbins;
  Matrix center = MatrixFromData(center_vec->len/hist_len, hist_len,
				 center_vec->d);
  Vector hist = VectorFromData(hist_len, NULL);
  int c;
  TpmImage subimage = TpmImageCreate(image->width, image->height/center->rows, 
				     image->channels);
  for(c=0;c<center->rows;c++) {
    hist->d = center->d2[c];
    TpmImageSynthHist(subimage, hist, nbins);
    TpmImageMoveTo(image, subimage, 0, c*image->height/center->rows);
  }

  TpmImageFree(subimage);
  free(hist);
  free(center->d2);
  free(center);
}

void TpmImageSynthBars(TpmImage image, Vector v, 
		       double minimum, double maximum)
{
  int i, x, inc, bar_width, bar_space;
  int spacing = 1;
  double scale = image->height / (maximum - minimum);

  /* determine the bar_width and spacing */
  x = 0;
  inc = 1;
  bar_width = image->width / v->len;
  if(bar_width == 0) {
    /* image is shorter than v */
    bar_width = 1;
    bar_space = 0;
    inc = (int)ceil((double)v->len / image->width);
  }
  else {
    /* image is longer than v */
    bar_width /= spacing + 1;
    if(bar_width == 0) {
      /* spacing is too large; try spacing = 1 */
      bar_width = bar_space = image->width / v->len / 2;
      if(bar_width == 0) {
	/* still too large; no spacing */
	bar_width = 1;
	bar_space = 0;
	/* center the bars */
	x = (image->width - v->len) / 2;
      }
    }
    else { 
      bar_space = bar_width * spacing;
    }
  }

  /* clear to black */
  memset(image->d, 0, image->len);

  /* draw bars of width bar_width spaced apart by bar_space */
  for(i=0;i<v->len;i+=inc,x+=bar_width+bar_space) {
    int j, height;
    double sum = 0.0;
    /* average together inc elements */
    for(j=0;j<inc;j++) sum += v->d[i+j];
    sum /= inc;

    height = (int)((sum - minimum) * scale + 0.5);
    if(height <= 0) continue;
    if(height > image->height) height = image->height;
    TpmImagePutBlock(image, x, image->height-height, bar_width, height,
		     255, 255, 255);
  }
}

/* WindowHist ****************************************************************/

struct WindowHistInfo {
  Matrix result;
  int row;
};

static Vector v = NULL;

static void WindowHist(struct WindowHistInfo *info, TpmImage image)
{
  if(!v) v = VectorFromData(info->result->cols, NULL);
  v->len = info->result->cols;
  v->d = info->result->d2[info->row];
  TpmImageHist(image, v);
  info->row++;
}

void TpmImageHistWindows(TpmImage image, int xsize, int ysize, 
			 int xinc, int yinc, Matrix result)
{
  struct WindowHistInfo info;
  info.result = result;
  info.row = 0;
  TpmImageWindows(image, xsize, ysize, xinc, yinc, 
		  (TpmImageFunc*)WindowHist, &info);
}

/* WindowAutoCorrelogram *****************************************************/

struct WindowAutoCorrelogramInfo {
  Matrix result;
  int *spacing;
  int num_spacings;
  int row;
};

static void WindowAutoCorrelogram(struct WindowAutoCorrelogramInfo *info, 
				  TpmImage image)
{
  Matrix m = MatrixFromData(info->result->cols / info->num_spacings, 
			    info->num_spacings,
			    info->result->d2[info->row]);
  TpmImageAutoCorrelogram(image, info->spacing, m);
  info->row++;
  free(m->d2);
  free(m);
}

void TpmImageAutoCorrelogramWindows(TpmImage image, int spacing[], 
				    int num_spacings,
				    int xsize, int ysize, 
				    int xinc, int yinc, Matrix result)
{
  struct WindowAutoCorrelogramInfo info;
  info.result = result;
  info.row = 0;
  info.spacing = spacing;
  info.num_spacings = num_spacings;
  TpmImageWindows(image, xsize, ysize, xinc, yinc, 
		  (TpmImageFunc*)WindowAutoCorrelogram, &info);
}

/* AutoCorrelogram ***********************************************************/

struct AutoCorrelogramInfo {
  Matrix result;
  Vector hist;
  int *spacing;
};

static void AutoCorrelogramWindow(struct AutoCorrelogramInfo *info, 
				  TpmImage image)
{
  int c, s;
  int x = (image->width-1)/2;
  int y = (image->height-1)/2;
  Matrix result = info->result;
  int *spacing = info->spacing;
  for(s=0;s<result->cols;s++) {
#if 0
    int x_w = max(0, x-spacing[s]);
    int x_e = min(image->width-1, x+spacing[s]);
    int y_n = max(0, y-spacing[s]);
    int y_s = min(image->height-1, y+spacing[s]);
#else
    int x_w = x-spacing[s];
    int x_e = x+spacing[s];
    int y_n = y-spacing[s];
    int y_s = y+spacing[s];
#endif
    for(c=0;c<result->rows;c++) {
      int x2, y2;
      if(image->d2[y][x] != c) continue;
      info->hist->d[c]++;
      /* top and bottom lines */
      for(x2 = x_w; x2 <= x_e; x2++) {
	result->d2[c][s] += (image->d2[y_n][x2] == c);
	result->d2[c][s] += (image->d2[y_s][x2] == c);
      }
      /* left and right sides */
      for(y2 = y_n+1; y2 < y_s; y2++) {
	result->d2[c][s] += (image->d2[y2][x_w] == c);
	result->d2[c][s] += (image->d2[y2][x_e] == c);
      }
    }
  }
}

/* Sets result to the autocorrelogram of image.
 * Ignores pixels close to the boundary.
 * Each row corresponds to one color, each column a spacing.
 * image should be single-channel, e.g. quantized with TpmImageQuantizeCube. 
 */
void TpmImageAutoCorrelogram(TpmImage image, int spacing[], Matrix result)
{
  int num_spacings = result->cols, max_spacing = 0;
  int *hist, i,j;

  assert(image->channels == 1);
  /* Determine the maximum correlation distance */
  for(i=0;i<num_spacings;i++) {
    if(spacing[i] > max_spacing) max_spacing = spacing[i];
  }

  if(1) {
    /* naive method */
    Vector hist = VectorCreate(result->rows);
    struct AutoCorrelogramInfo info;
    info.result = result;
    info.hist = hist;
    info.spacing = spacing;
    VectorSet((Vector)result, 0);
    VectorSet(hist, 0);
    TpmImageWindows(image, 2*max_spacing+1, 2*max_spacing+1, 1, 1,
		    (TpmImageFunc*)AutoCorrelogramWindow, &info);
    /* each pixel was counted num_spacings times */
    VectorScale(hist, 1.0/num_spacings);
    for(i=0;i<result->rows;i++) {
      for(j=0;j<result->cols;j++) {
	if(hist->d[i]) {
	  result->d2[i][j] /= hist->d[i] * 8 * spacing[j];
	}
      }
    }
    VectorFree(hist);
    return;
  }

  if(0) {
    /* Zabih's method.  Doesn't work and takes too much memory. */
    int s, x, y, c;
    int ncolors = result->rows;
    unsigned *h_count, *v_count, *p;
    /* compute horizontal and vertical counts */
    p = h_count = Allocate(image->len*(max_spacing+1)*ncolors, unsigned);
    for(i=0;i<image->len;i++) {
      unsigned *last_p;
      memset(p, 0, ncolors);
      p[ image->d[i] ] = 1;
      for(x=1;x <= max_spacing;x++) {
	last_p = p;
	p += ncolors;
	memcpy(p, last_p, ncolors);
	/* step by two each time */
	p[ image->d[i+2*x-1] ]++;
	p[ image->d[i+2*x] ]++;
      }
      p += ncolors;
    }
    p = v_count = Allocate(image->len*(max_spacing+1)*ncolors, unsigned);
    for(i=0;i<image->len;i++) {
      unsigned *last_p;
      memset(p, 0, ncolors);
      p[ image->d[i] ] = 1;
      for(y=1;y <= max_spacing;y++) {
	last_p = p;
	p += ncolors;
	memcpy(p, last_p, ncolors);
	/* step by two each time */
	p[ image->d[i+(2*y-1)*image->width] ]++;
	p[ image->d[i+2*y*image->width] ]++;
      }
      p += ncolors;
    }

    /* compute cooccurrences */
    VectorSet((Vector)result, 0);
    for(y=max_spacing;y<image->height-max_spacing;y++) {
      for(x=max_spacing;x<image->width-max_spacing;x++) {
	for(s=0;s<num_spacings;s++) {
#if 0
	  int x_w = max(0, x-spacing[s]);
	  int x_e = min(image->width-1, x+spacing[s]);
	  int y_n = max(0, y-spacing[s]);
	  int y_s = min(image->height-1, y+spacing[s]);
#else
	  int x_w = x-spacing[s];
	  int x_e = x+spacing[s];
	  int y_n = y-spacing[s];
	  int y_s = y+spacing[s];
#endif
	  int i_nw = y_n*image->width + x_w;
	  int i_nw2 = i_nw + image->width;
	  int i_sw = y_s*image->width + x_w;
	  int i_ne = i_nw + x_e - x_w + image->width;
	  c = image->d2[y][x];
	  result->d2[c][s] += 
	    /* nw to ne */
	    h_count[(i_nw*(max_spacing+1) + s)*ncolors + c] +
	    /* nw to sw */
	    v_count[(i_nw2*(max_spacing+1) + s-1)*ncolors + c] +
	    /* sw to se */
	    h_count[(i_sw*(max_spacing+1) + s)*ncolors + c] +
	    /* ne to se */
	    v_count[(i_ne*(max_spacing+1) + s-1)*ncolors + c];
	}
      }
    }
    free(h_count);
    free(v_count);

    hist = TpmImageHistogram(image);
    for(i=0;i<result->rows;i++) {
      for(j=0;j<result->cols;j++) {
	if(hist[i]) {
	  result->d2[i][j] /= hist[i] * 8 * spacing[j];
	}
      }
    }
    free(hist);
  }
}
