#include <tpm/tpm.h>
#include <gl/image.h>
#include <tpm/matrix.h>
#include <tpm/image.h>

#include <tpm/gmm.h>
#include <tpm/mnm.h>

#define HSV 1
#define OHTA 0
#define CUBE 4
#define CLUSTERS 4
#define GAUSSIAN 0
#define IMAGE 1

static void GaussianClassImage(TpmImage image, GaussianMixture gm, 
			       Matrix points)
{
  Vector mbr = VectorCreate(gm->center->rows);
  Vector point = VectorFromData(points->cols, NULL);
  int i;
  for(i=0;i<points->rows;i++) {
    point->d = points->d2[i];
    GaussianMixtureClassify(gm, point, mbr->d);
    image->d[i] = VectorMaximum(mbr);
  }
  free(point);
  VectorFree(mbr);
}

static void MultinomialClassImage(TpmImage image, MultinomialMixture dm, 
				Matrix points)
{
  Vector mbr = VectorCreate(dm->center->rows);
  Vector point = VectorFromData(points->cols, NULL);
  int i;
  for(i=0;i<points->rows;i++) {
    point->d = points->d2[i];
    MultinomialMixtureClassify(dm, point, mbr->d);
    image->d[i] = VectorMaximum(mbr);
  }
  free(point);
  VectorFree(mbr);
}

/* Reads image names on stdin and outputs histograms on stdout. */
void main(int argc, char *argv[])
{
  char fname[1024];
  TpmImage rgb_image, image;
  int wsize = 16, inc = 1;

  if(argc > 1) wsize = atoi(argv[1]);
  if(argc > 2) inc = atoi(argv[2]);
  for(;;) {
    if(gets(fname) == NULL) break;
    rgb_image = TpmReadSgi(fname);
    if(!rgb_image) {
      fprintf(stderr, "Cannot open `%s': \n", fname);
      perror("");
      continue;
    }
#if HSV
    TpmImageRGB2HSV(rgb_image);
#endif
    image = TpmImageQuantizeCube(rgb_image, CUBE, CUBE, CUBE);
    TpmImageFree(rgb_image);

    if((image->height - wsize) % inc) {
      fprintf(stderr, "cannot use wsize = %d, inc = %d with height %d\n",
	      wsize, inc, image->height);
    }
    else if((image->width - wsize) % inc) {
      fprintf(stderr, "cannot use wsize = %d, inc = %d with width %d\n",
	      wsize, inc, image->width);
    }
    else {
      int xcount = ((image->width - wsize)/inc+1);
      int ycount = ((image->height - wsize)/inc+1);
      Matrix result = MatrixCreate(ycount*xcount, CUBE*CUBE*CUBE);
      fprintf(stderr, "%d windows of dimension %d\n", 
	      result->rows, result->cols);
      TpmImageHistWindows(image, wsize, wsize, inc, inc, result);
      VectorScale((Vector)result, 1.0/wsize/wsize);

      printf("%s ", fname);
      if(0) {
	MatrixPrint(stdout, result);
      }
      else {
	/* Almost all of the time is spent here. */
#if GAUSSIAN
	GaussianMixture gm = GaussianMixtureCreate(CLUSTERS, result->cols);
	gm->temperature = 1e-2;
	GaussianMixtureMeanCenter(gm, result);
	GaussianMixtureTrainEM(gm, result);
	VectorPrint(stdout, (Vector)gm->center);
	if(IMAGE) {
	  char str[1024];
	  FILE *fp;
	  TpmImage class_image = TpmImageCreate(xcount, ycount, 1);
	  GaussianClassImage(class_image, gm, result);
	  TpmImageRescale(class_image, 0, 255);
	  sprintf(str, "class/%s", fname);
	  fp = fopen(str, "w");
	  if(fp) {
	    TpmImageWritePnm(fp, class_image);
	    fclose(fp);
	  }
	  TpmImageFree(class_image);
	}
	GaussianMixtureFree(gm);
#else
	MultinomialMixture dm = MultinomialMixtureCreate(CLUSTERS, result->cols);
	/* 0.3 is a good value for 100382 and 100868 with 4 clusters */
	dm->temperature = 0.2;
	MultinomialMixtureTrainEM(dm, result);
	VectorPrint(stdout, (Vector)dm->center);
	if(IMAGE) {
	  char str[1024];
	  FILE *fp;
	  TpmImage class_image = TpmImageCreate(xcount, ycount, 1);
	  MultinomialClassImage(class_image, dm, result);
	  TpmImageRescale(class_image, 0, 255);
	  sprintf(str, "class/%s", fname);
	  fp = fopen(str, "w");
	  if(fp) {
	    TpmImageWritePnm(fp, class_image);
	    fclose(fp);
	  }
	  TpmImageFree(class_image);
	}
	MultinomialMixtureFree(dm);
#endif
      }
      printf("\n");
      MatrixFree(result);
    }
    TpmImageFree(image);
  }
  MEM_BLOCKS();
  MEM_STATUS();
  exit(0);
}
