#include <tpm/tpm.h>
#include "image.h"

/* OSF already has uchar in sys/types.h */
#ifndef _OSF_SOURCE
typedef unsigned char uchar;
#endif

/* Prototypes ****************************************************************/

TpmImage TpmImageCreate(int width, int height, int channels);
TpmImage TpmImageFromData(int width, int height, int channels, 
			  unsigned char *data);
void TpmImageFree(TpmImage image);
TpmImage TpmImageCopy(TpmImage image);
void TpmImageMoveTo(TpmImage dest, TpmImage src, int toX, int toY);
void TpmImageMinMax(TpmImage image, int *min_return, int *max_return);
void TpmImageRescale(TpmImage image, uchar new_min, uchar new_max);

void TpmImagePutPixel(TpmImage image, int x, int y, 
		      unsigned char red, unsigned char green, 
		      unsigned char blue);
void TpmImagePutRow(TpmImage image, int x, int y, int width,
		    unsigned char red, unsigned char green, 
		    unsigned char blue);
void TpmImagePutBlock(TpmImage image, int x, int y, int width, int height,
		      unsigned char red, unsigned char green, 
		      unsigned char blue);
TpmImage TpmImageZoom(TpmImage image, int zfact);
TpmImage TpmImageGray(TpmImage image);
int *TpmImageHistogram(TpmImage image);

/* Functions *****************************************************************/

/* Returns a TpmImage whose data is given. The data becomes property 
 * of the image and will be freed when the image is. The data must be
 * width*height*channels bytes long, and is laid out as channel, line, pixel.
 */
TpmImage TpmImageFromData(int width, int height, int channels, 
			  unsigned char *data)
{
  TpmImage image;
  int i;

  image = Allocate(1, struct TpmImage);
  image->width = width;
  image->height = height;
  image->channels = channels;
  image->hc = channels * height;
  image->len = image->hc * width;

  image->d = data;
  image->d2 = Allocate(image->hc, uchar*);
  for(i=0;i<image->hc;i++) image->d2[i] = &image->d[i*width];
  image->d3 = Allocate(height, uchar**);
  for(i=0;i<height;i++) image->d3[i] = &image->d2[i*height];
  return image;
}

/* Returns a new, uninitialized image whose dimensions are given. */
TpmImage TpmImageCreate(int width, int height, int channels)
{
  uchar *data = Allocate(width * height * channels, uchar);
  if(!data) {
    fprintf(stderr, "Cannot allocate %d by %d by %d image\n", 
	    width, height, channels);
  }
  return TpmImageFromData(width, height, channels, data);
}

/* Frees all storage associated with image. */
void TpmImageFree(TpmImage image)
{
  free(image->d3);
  free(image->d2);
  free(image->d);
  free(image);
}

/* Returns a new image whose data is identical to image. */
TpmImage TpmImageCopy(TpmImage image)
{
  TpmImage result = TpmImageCreate(image->width, image->height, 
				   image->channels);
  TpmImageMove(result, image);
  return result;
}

void TpmImageMoveTo(TpmImage dest, TpmImage src, int toX, int toY)
{
  int x, y, c;
  assert(dest->height >= src->height+toY);
  assert(dest->width >= src->width+toX);
  for(c=0;c<src->channels;c++) {
    for(y=0;y<src->height;y++) {
      for(x=0;x<src->width;x++) {
	dest->d3[c][y+toY][x+toX] = src->d3[c][y][x];
      }
    }
  }
}

void TpmImageMinMax(TpmImage image, int *min_return, int *max_return)
{
  uchar minimum, maximum;
  int min_i, max_i;
  int i;
  min_i = max_i = 0;
  minimum = maximum = image->d[0];
  for(i=1;i<image->len;i++) {
    if(image->d[i] > maximum) {
      maximum = image->d[i];
      max_i = i;
    }
    else if(image->d[i] < minimum) {
      minimum = image->d[i];
      min_i = i;
    }
  }
  if(min_return) *min_return = min_i;
  if(max_return) *max_return = max_i;
}

void TpmImageRescale(TpmImage image, uchar new_min, uchar new_max)
{
  int i;
  int min_i, max_i;
  uchar old_min, old_max;
  double f;
  TpmImageMinMax(image, &min_i, &max_i);
  old_min = image->d[min_i];
  old_max = image->d[max_i];
  if(old_max != old_min) {
    f = (new_max - new_min) / (old_max - old_min);
    for(i=0;i<image->len;i++) {
      image->d[i] = (image->d[i] - old_min) * f + new_min;
    }
  }
  else {
    for(i=0;i<image->len;i++) {
      image->d[i] = image->d[i] - old_min + new_min;
    }
  }
}

/*****************************************************************************/

#define ColorToGray(r,g,b) (unsigned char)(0.30*(r)+0.59*(g)+0.11*(b)+0.5)

void TpmImagePutPixel(TpmImage image, int x, int y, 
		      unsigned char red, unsigned char green, 
		      unsigned char blue)
{
  if(image->height>y && image->width>x && x>=0 && y>=0) {
    if(image->channels==3) {
      image->d3[0][y][x]=red;
      image->d3[1][y][x]=green;
      image->d3[2][y][x]=blue;
    }
    else {
      image->d3[0][y][x]=ColorToGray(red,green,blue);
    }
  }
}

void TpmImagePutRow(TpmImage image, int x, int y, int width,
		    unsigned char red, unsigned char green, 
		    unsigned char blue)
{
  int i;

  /* clip to the image boundaries */
  if(x + width >= image->width) {
    width = image->width - x - 1;
  }

  if(image->channels == 3) {
    for(i=0;i<width;i++) {
      image->d3[0][y][x+i] = red;
      image->d3[1][y][x+i] = green;
      image->d3[2][y][x+i] = blue;
    }
  }
  else {
    uchar gray = ColorToGray(red, green, blue);
    for(i=0;i<width;i++) {
      image->d2[y][x+i] = gray;
    }
  }
}

void TpmImagePutBlock(TpmImage image, int x, int y, int width, int height,
		      unsigned char red, unsigned char green, 
		      unsigned char blue)
{
  int i;
  for(i=0;i<height;i++) {
    TpmImagePutRow(image, x, i+y, width, red, green, blue);
  }
}

/* If zfact = -1,0,1 then copies image.
 * If zfact > 1 then returns image enlarged by zfact (bilinear interpolation)
 * If zfact < -1 then return image reduced by -zfact (averaging)
 */
TpmImage TpmImageZoom(TpmImage image, int zfact)
{
  
  int i,j,k,l,m;
  int p,q;
  int a,b,c,d,aw,bw,cw,dw;
  TpmImage zimage;
  
  if(zfact == 0) zfact = 1;
  if(zfact > 1) {
    zimage = TpmImageCreate((image->width)*zfact,
			    (image->height)*zfact, image->channels);
    /* Puts enlarged copy of image->data into zimage->data */
    for(m=0;m<image->channels;m++) {
      for(i=0;i<image->height;i++)  {
	for(j=0;j<image->width;j++)  {
	  for(k=0;k<zfact;k++) {
	    for(l=0;l<zfact;l++) {
	      p = q = 0;
	      a = image->d3[m][i][j];

	      aw = (zfact-l)*(zfact-k);
	      bw = k*(zfact-l);
	      cw = l*(zfact-k);
	      dw = l*k;
                          
	      p += aw*a;
	      q += aw;
	      
	      if(i < image->height-1) {
		b=image->d3[m][i+1][j];
		p+=bw*b;
		q+=bw;
	      }
	      else {
		b=a;
		p+=aw*a;
		q+=aw;
	      }
	      
	      if(j < image->width-1) {
		c=image->d3[m][i][j+1];
		p+=cw*c;
		q+=cw;
	      }
	      else {
		c=a;
		p+=aw*a;
		q+=aw;
	      }
	      
	      if((j < image->width-1) && (i < image->height-1))	{
		d=image->d3[m][i+1][j+1];
		p+=dw*d;
		q+=dw;
	      }
	      else {
		d=a;
		p+=aw*a;
		q+=aw;
	      }
	      zimage->d3[m][zfact*i+k][zfact*j+l]=p/q;
	    }
	  }
	}
      }
    }
  }
  else {  /* zfact is < -1 */
    zfact = -zfact;
    zimage = TpmImageCreate((image->width)/zfact, 
			    (image->height)/zfact, image->channels);
    /* Put shrunken copy of image->data into zimage->data */
    for(m=0;m<zimage->channels;m++) {
      for(i=0;i<zimage->height;i++) {
	for(j=0;j<zimage->width;j++) {
	  int sum = 0;
	  for(k=0;k<zfact;k++) {
	    for(l=0;l<zfact;l++) {
	      sum += image->d3[m][zfact*i+k][zfact*j+l];
	    }
	  }
	  zimage->d3[m][i][j] = sum / (zfact*zfact);
	}
      }
    }
  }
  return zimage;
}

/* Returns the concatenation of 256 bins per channel. */
int *TpmImageHistogram(TpmImage image)
{
  int c, i, image_size = image->width * image->height;
  int *bins = AllocateC(image->channels*256, int);
  int *count = bins;
  for(c=0;c<image->channels;c++) {
    for(i=0;i<image_size;i++) {
      int v = image->d3[c][0][i];
      count[v]++;
    }
    count += 256;
  }
  return bins;
}

/* Produces the NTSC gray equivalent of an RGB image. */
TpmImage TpmImageGray(TpmImage image)
{
  TpmImage result;
  int i;
  if(image->channels == 1) return TpmImageCopy(image);
  result = TpmImageCreate(image->width, image->height, 1);
  for(i=0;i<result->len;i++) {
    result->d[i] = ColorToGray(image->d3[0][0][i], 
			       image->d3[1][0][i], 
			       image->d3[2][0][i]);
  }
  return result;
}
