#include <tpm/tpm.h>
#include <tpm/stream.h>
#include <jpeglib.h>
#include "image.h"
#include <setjmp.h>

/* Prototypes ****************************************************************/
TpmImage TpmReadJpeg(char *fname);
TpmImage TpmReadJpegStream(FILE *fp);
void JpegClose(Jpeg *jpeg);
int JpegReadInit(Jpeg *jpeg);
TpmImage TpmImageFromJpeg(Jpeg *jpeg);
int TpmImageWriteJpeg(FILE *fp, TpmImage image);

/* Functions *****************************************************************/

TpmImage TpmReadJpeg(char *fname)
{
  TpmImage image;
  FILE *fp;
  if(!strcmp(fname, "-")) fp = stdin;
  else {
    fp = fopen(fname, "rb");
    if(!fp) {
      fprintf(stderr, "TpmReadJpeg: cannot open `%s'", fname);
      perror("");
      return NULL;
    }
  }
  image = TpmReadJpegStream(fp);
  fclose(fp);
  return image;
}

TpmImage TpmReadJpegStream(FILE *fp)
{
  TpmImage image;
  Jpeg jpeg;

  jpeg.fp = fp;
  if(JpegReadInit(&jpeg)) return NULL;
  image = TpmImageFromJpeg(&jpeg);
  jpeg.fp = NULL;
  JpegClose(&jpeg);
  return image;
}

void JpegClose(Jpeg *jpeg)
{
  if(jpeg->fp) fclose(jpeg->fp);
  jpeg_destroy_decompress(jpeg->cinfo);
  free(jpeg->cinfo->err);
  free(jpeg->cinfo);
}

/*
 * ERROR HANDLING:
 *
 * The JPEG library's standard error handler (jerror.c) is divided into
 * several "methods" which you can override individually.  This lets you
 * adjust the behavior without duplicating a lot of code, which you might
 * have to update with each future release.
 *
 * Our example here shows how to override the "error_exit" method so that
 * control is returned to the library's caller when a fatal error occurs,
 * rather than calling exit() as the standard error_exit method does.
 *
 * We use C's setjmp/longjmp facility to return control.  This means that the
 * routine which calls the JPEG library must first execute a setjmp() call to
 * establish the return point.  We want the replacement error_exit to do a
 * longjmp().  But we need to make the setjmp buffer accessible to the
 * error_exit routine.  To do this, we make a private extension of the
 * standard JPEG error handler object.  (If we were using C++, we'd say we
 * were making a subclass of the regular error handler.)
 *
 * Here's the extended error handler struct:
 */

typedef struct my_error_mgr {
  struct jpeg_error_mgr pub;	/* "public" fields */

  jmp_buf setjmp_buffer;	/* for return to caller */
} *my_error_ptr;

/*
 * Here's the routine that will replace the standard error_exit method:
 */
static void my_error_exit (j_common_ptr cinfo)
{
  /* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
  my_error_ptr myerr = (my_error_ptr) cinfo->err;

#if 0
  /* Display the message. */
  (*cinfo->err->output_message) (cinfo);
#endif

  /* Return control to the setjmp point */
  longjmp(myerr->setjmp_buffer, 1);
}

int JpegReadInit(Jpeg *jpeg)
{
  /* This struct contains the JPEG decompression parameters and pointers to
   * working space (which is allocated as needed by the JPEG library).
   */
  struct jpeg_decompress_struct *cinfo;
  /* We use our private extension JPEG error handler.
   * Note that this struct must live as long as the main JPEG parameter
   * struct, to avoid dangling-pointer problems.
   */
  my_error_ptr jerr;

  /* Check the header before calling the library */
  if(FileCheckHeader(jpeg->fp, "\377\330\377\340\000\020JFIF\000\001", 12))
    return TPM_ERROR;

  cinfo = Allocate(1, struct jpeg_decompress_struct);
  jerr = Allocate(1, struct my_error_mgr);

  /* Step 1: allocate and initialize JPEG decompression object */
  /* We set up the normal JPEG error routines, then override error_exit. */
  cinfo->err = jpeg_std_error(&jerr->pub);
  jerr->pub.error_exit = my_error_exit;
  /* Establish the setjmp return context for my_error_exit to use. */
  if (setjmp(jerr->setjmp_buffer)) {
    /* If we get here, the JPEG code has signaled an error.
     * We need to clean up the JPEG object and return.
     */
    return TPM_ERROR;
  }
  /* Now we can initialize the JPEG decompression object. */
  jpeg_create_decompress(cinfo);

  /* Step 2: specify data source (eg, a file) */
  jpeg_stdio_src(cinfo, jpeg->fp);

  /* Step 3: read file parameters with jpeg_read_header() */

  (void) jpeg_read_header(cinfo, TRUE);
  /* We can ignore the return value from jpeg_read_header since
   *   (a) suspension is not possible with the stdio data source, and
   *   (b) we passed TRUE to reject a tables-only JPEG file as an error.
   * See libjpeg.doc for more info.
   */

  jpeg->cinfo = cinfo;
  jpeg->width = jpeg->cinfo->image_width;
  jpeg->height = jpeg->cinfo->image_height;
  jpeg->channels = jpeg->cinfo->num_components;
  return TPM_OK;
}

TpmImage TpmImageFromJpeg(Jpeg *jpeg)
{
  struct jpeg_decompress_struct *cinfo = jpeg->cinfo;
  JSAMPARRAY buffer;		/* Output row buffer */
  int row_stride;		/* physical row width in output buffer */
  /* Create the image */
  TpmImage image = TpmImageCreate(jpeg->width, jpeg->height, jpeg->channels);

  /* Step 4: set parameters for decompression */

  /* In this example, we don't need to change any of the defaults set by
   * jpeg_read_header(), so we do nothing here.
   */

  /* Step 5: Start decompressor */

  (void) jpeg_start_decompress(cinfo);
  /* We can ignore the return value since suspension is not possible
   * with the stdio data source.
   */

  /* We may need to do some setup of our own at this point before reading
   * the data.  After jpeg_start_decompress() we have the correct scaled
   * output image dimensions available, as well as the output colormap
   * if we asked for color quantization.
   * In this example, we need to make an output work buffer of the right size.
   */ 
  /* JSAMPLEs per row in output buffer */
  row_stride = cinfo->output_width * cinfo->output_components;

  /* Step 6: while (scan lines remain to be read) */
  /*           jpeg_read_scanlines(...); */

  /* Here we use the library's state variable cinfo.output_scanline as the
   * loop counter, so that we don't have to keep track ourselves.
   */
  if(image->channels == 1) {
    while (cinfo->output_scanline < cinfo->output_height) {
      (void) jpeg_read_scanlines(cinfo, &image->d2[cinfo->output_scanline], 1);
    }
  }
  else {
    /* assume 3 channels */
    unsigned char *rptr, *gptr, *bptr;
    rptr = *image->d3[0]; gptr = *image->d3[1]; bptr = *image->d3[2];
    /* Make a one-row-high sample array that will go away when done 
       with image */
    buffer = (*cinfo->mem->alloc_sarray)
               ((j_common_ptr) cinfo, JPOOL_IMAGE, row_stride, 1);
    while (cinfo->output_scanline < cinfo->output_height) {
      unsigned char *ptr;
      int i;
      /* jpeg_read_scanlines expects an array of pointers to scanlines.
       * Here the array is only one element long, but you could ask for
       * more than one scanline at a time if that's more convenient.
       */
      (void) jpeg_read_scanlines(cinfo, buffer, 1);

      /* Copy the scanline into the TpmImage */
      ptr = buffer[0];
      for(i=0;i<image->width;i++) {
	*rptr++ = *ptr++;
	*gptr++ = *ptr++;
	*bptr++ = *ptr++;
      }
    }
  }

  /* Step 7: Finish decompression */

  (void) jpeg_finish_decompress(cinfo);
  /* We can ignore the return value since suspension is not possible
   * with the stdio data source.
   */

  return image;
}

int TpmImageWriteJpeg(FILE *fp, TpmImage image)
{
  /* This struct contains the JPEG compression parameters and pointers to
   * working space (which is allocated as needed by the JPEG library).
   * It is possible to have several such structures, representing multiple
   * compression/decompression processes, in existence at once.  We refer
   * to any one struct (and its associated working data) as a "JPEG object".
   */
  struct jpeg_compress_struct cinfo;
  struct jpeg_error_mgr jerr;
  int row_stride;		/* physical row width in image buffer */

  /* Step 1: allocate and initialize JPEG compression object */

  /* We have to set up the error handler first, in case the initialization
   * step fails.  (Unlikely, but it could happen if you are out of memory.)
   * This routine fills in the contents of struct jerr, and returns jerr's
   * address which we place into the link field in cinfo.
   */
  cinfo.err = jpeg_std_error(&jerr);
  /* Now we can initialize the JPEG compression object. */
  jpeg_create_compress(&cinfo);

  /* Step 2: specify data destination (eg, a file) */
  jpeg_stdio_dest(&cinfo, fp);

  /* Step 3: set parameters for compression */

  /* First we supply a description of the input image.
   * Four fields of the cinfo struct must be filled in:
   */
  cinfo.image_width = image->width;
  cinfo.image_height = image->height;
  cinfo.input_components = image->channels;
  cinfo.in_color_space = JCS_RGB; 	/* colorspace of input image */
  /* Now use the library's routine to set default compression parameters.
   * (You must set at least cinfo.in_color_space before calling this,
   * since the defaults depend on the source color space.)
   */
  jpeg_set_defaults(&cinfo);
  /* Default quality factor is 75% */
#if 0
  /* Set quality factor to 75% */
  jpeg_set_quality(&cinfo, 75, TRUE /* limit to baseline-JPEG values */);
#endif

  /* Step 4: Start compressor */

  /* TRUE ensures that we will write a complete interchange-JPEG file.
   * Pass TRUE unless you are very sure of what you're doing.
   */
  jpeg_start_compress(&cinfo, TRUE);

  /* Step 5: while (scan lines remain to be written) */
  /*           jpeg_write_scanlines(...); */

  /* Here we use the library's state variable cinfo.next_scanline as the
   * loop counter, so that we don't have to keep track ourselves.
   * To keep things simple, we pass one scanline per call; you can pass
   * more if you wish, though.
   */
  row_stride = image->width * image->channels;

  if(image->channels == 1) {
    while (cinfo.next_scanline < cinfo.image_height) {
      /* jpeg_write_scanlines expects an array of pointers to scanlines. */
      (void) jpeg_write_scanlines(&cinfo, &image->d2[cinfo.next_scanline],
				  cinfo.image_height - cinfo.next_scanline);
    }
  }
  else {
    /* assume 3 channels */
    unsigned char *rptr, *gptr, *bptr;
    unsigned char *buffer = Allocate(row_stride, unsigned char);
    rptr = *image->d3[0];
    gptr = *image->d3[1];
    bptr = *image->d3[2];
    while (cinfo.next_scanline < cinfo.image_height) {
      unsigned char *ptr = buffer;
      int i;
      for(i=0;i<image->width;i++) {
	*ptr++ = *rptr++;
	*ptr++ = *gptr++;
	*ptr++ = *bptr++;
      }
      (void) jpeg_write_scanlines(&cinfo, &buffer, 1);
    }
    free(buffer);
  }

  /* Step 6: Finish compression */

  jpeg_finish_compress(&cinfo);

  /* Step 7: release JPEG compression object */

  /* This is an important step since it will release a good deal of memory. */
  jpeg_destroy_compress(&cinfo);

  /* And we're done! */
  return TPM_OK;
}
