#include <tpm/tpm.h>
#include "image.h"
#include <ctype.h>

/* Prototypes ****************************************************************/
TpmImage TpmReadPnm(char *fname);
TpmImage TpmReadPnmStream(FILE *fp);
int PnmReadInit(Pnm *pnm);
TpmImage TpmImageFromPnm(Pnm *pnm);
int TpmImageWritePnm(FILE *fp, TpmImage image);

/* Functions *****************************************************************/

TpmImage TpmReadPnm(char *fname)
{
  TpmImage image;
  FILE *fp;
  if(!strcmp(fname, "-")) fp = stdin;
  else {
    fp = fopen(fname, "r");
    if(!fp) {
      fprintf(stderr, "TpmReadPnm: cannot open `%s`: ", fname);
      perror("");
      return NULL;
    }
  }
  image = TpmReadPnmStream(fp);
  fclose(fp);
  return image;
}

TpmImage TpmReadPnmStream(FILE *fp)
{
  TpmImage image;
  Pnm pnm;

  pnm.fp = fp;
  if(PnmReadInit(&pnm)) return NULL;
  image = TpmImageFromPnm(&pnm);
  pnm.fp = NULL;
  PnmClose(&pnm);
  return image;
}

static int GetInt(FILE *fp, int *ip)
{
  int c;
  /* look for a digit, skipping comments */
  for(;;) {
    c = getc(fp);
    if(c == EOF) return TPM_ERROR;
    if(c == '#') {
      /* read until end of line */
      do {
	c = getc(fp);
	if(c == EOF) return TPM_ERROR;
      } while(c != '\n');
      continue;
    }
    if(isspace(c)) continue;
    if(!isdigit(c)) return TPM_ERROR; /* got some garbage */
    break; /* found a digit */
  }
  /* read the digits */
  *ip = c - '0';
  for(;;) {
    c = getc(fp);
    if(!isdigit(c)) break;
    *ip = *ip * 10 + c - '0';
  }
  /* eats one char after the number */
  return TPM_OK;
}

int PnmReadInit(Pnm *pnm)
{
  pnm->format = getc(pnm->fp);
  if(pnm->format != 'P') return TPM_ERROR;
  pnm->format = getc(pnm->fp) - '0';
  if((pnm->format < 4) || (pnm->format > 6)) return TPM_ERROR;
  
  TRY_RETURN(GetInt(pnm->fp, &pnm->cols));
  TRY_RETURN(GetInt(pnm->fp, &pnm->rows));
  TRY_RETURN(GetInt(pnm->fp, &pnm->maxval));
  /* pnm->fp is now at beginning of data */
  return TPM_OK;
}

TpmImage TpmImageFromPnm(Pnm *pnm)
{
  TpmImage image;
  unsigned char *rptr, *gptr, *bptr;
  int i, image_size;

  image = TpmImageCreate(pnm->cols, pnm->rows, 
			 pnm->format == 6 ? 3 : 1);
  image_size = pnm->cols * pnm->rows;
  
  switch(pnm->format) {
  case 6:
    rptr = *image->d3[0];
    gptr = *image->d3[1];
    bptr = *image->d3[2];
    for(i=0;i<image_size;i++) {
      *rptr++ = getc(pnm->fp);
      *gptr++ = getc(pnm->fp);
      *bptr++ = getc(pnm->fp);
    }
    break;

  case 5:
    fread(image->d, 1, image_size, pnm->fp);
    break;

  default:
    fprintf(stderr, "TpmImageFromPnm: unsupported format P%d\n", 
	    pnm->format & 0xff);
    return NULL;
  }
  return image;
}

int TpmImageWritePnm(FILE *fp, TpmImage image)
{
  int i, image_size = image->height * image->width;
  if(image->channels == 1) {
    fprintf(fp, "P5\n");
    fprintf(fp, "%d %d\n255\n", image->width, image->height);
    
    fwrite(image->d, 1, image_size, fp);
  }
  else { /* channels == 3 */
    unsigned char *rptr, *gptr, *bptr;
    fprintf(fp, "P6\n");
    fprintf(fp, "%d %d\n255\n", image->width, image->height);

    rptr = *image->d3[0];
    gptr = *image->d3[1];
    bptr = *image->d3[2];
    for(i=0;i<image_size;i++) {
      putc(*rptr++, fp);
      putc(*gptr++, fp);
      putc(*bptr++, fp);
    }
  }
  return TPM_OK;
}

