#include <tpm/tpm.h>
#include <tpm/stream.h>
#include <tiffio.h>
#include "image.h"

#define DEBUG 0

/* Prototypes ****************************************************************/
TpmImage TpmReadTiff(char *fname);
TpmImage TpmReadTiffStream(FILE *fp);
TpmImage TpmImageFromTiff(TIFF *tif);
int TpmImageWriteTiff(FILE *fp, TpmImage image);

/* Functions *****************************************************************/

#define error(msg) \
  {fprintf(stderr, "%s: %s\n", TIFFFileName(tif), msg); \
     return NULL;}

TpmImage TpmImageFromTiff(TIFF *tif)
{
  int32 height, width;
  unsigned short channels;
  unsigned short photometric;
  unsigned short planarconfig, matteing;
  unsigned short bitspersample;
  int ncolors, ncmap, cmaplen;
  unsigned short *cmap;
  unsigned char *tiffbuf, *ptr;
  int x,y;
  TpmImage image;
  
  /* get the info about the image */
  if (TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &width) == 0)
    error("TIFFGetField TIFFTAG_IMAGEWIDTH failed!");
  if (TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &height) == 0)
    error("TIFFGetField TIFFTAG_IMAGELENGTH failed!");
  if (TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &channels) == 0) 
    error("TIFFGetField TIFFTAG_SAMPLESPERPIXEL failed!");
  if (TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &bitspersample) == 0) 
    error("TIFFGetField TIFFTAG_BITSPERSAMPLE failed!");
  if (TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, &photometric) == 0)
    error("TIFFGetField TIFFTAG_PHOTOMETRIC failed!");
  if (TIFFGetField(tif, TIFFTAG_PLANARCONFIG, &planarconfig ) == 0)
    error("TIFFGetField TIFFTAG_PLANARCONFIG failed!");
  /* Might not be present. */
  matteing = 0;
  (void)TIFFGetField(tif, TIFFTAG_MATTEING, &matteing );

#if DEBUG
  fprintf(stderr, "%d x %d x %d\n", width, height, channels);
  fprintf(stderr, "%d bits\n", bitspersample);
  fprintf(stderr, "photometric: %d\n", photometric);
  fprintf(stderr, "planarconfig: %d\n", planarconfig);
  fprintf(stderr, "matteing: %d\n", matteing);
#endif

  if(bitspersample != 8) 
    error("cannot handle bitspersample != 8, sorry.");

  switch ( photometric ) {
  case PHOTOMETRIC_MINISWHITE:
    {
      int size =  1 << bitspersample;
      if ( planarconfig == PLANARCONFIG_SEPARATE && matteing )
        channels = 1;
      ncolors =  1;
      ncmap = 1;
      cmaplen = bitspersample;
/*
      cmap = (unsigned short *)malloc((unsigned)size*sizeof(short));
      for (i=0; i<size; i++)
        cmap[i] = 65535L - ((long)i * 65535L)/(size - 1);
*/
      break;
    }
  case PHOTOMETRIC_MINISBLACK:
    {
      int size =  1 << bitspersample;
      if ( planarconfig == PLANARCONFIG_SEPARATE && matteing )
        channels = 1;
      ncolors =  1;
      ncmap = 1;
      cmaplen = bitspersample;
/*
      cmap = (unsigned short *)malloc((unsigned)size*sizeof(short));
      for (i=0; i<size; i++)
        cmap[i] = ((long)i * 65535L)/(size - 1);
*/
      break;
    }
  case PHOTOMETRIC_RGB:
    if (planarconfig != PLANARCONFIG_CONTIG)
      error("can't handle separate planar config");
    ncolors =  3;
    ncmap = 0;
    cmaplen = 0;
    cmap = 0;
    break;
  case PHOTOMETRIC_PALETTE:     /* Handle Colormap */
    {
      unsigned short    *red, *green, *blue;
      int               size;
      
      size = 1 << bitspersample;
      if ( planarconfig == PLANARCONFIG_SEPARATE && matteing )
                channels = 1;
      if (TIFFGetField(tif, TIFFTAG_COLORMAP, &red, &green, &blue)) {
        ncolors =  1;
        ncmap = 3;
        cmaplen = bitspersample;
/*
        cmap = (unsigned short *)malloc((unsigned)(3*size*sizeof(short)) );
        for (i=0; i<size; i++) {
          cmap[i] = red[i];
          cmap[i + size] = green[i];
          cmap[i + 2*size] = blue[i];
        }
*/
      } else error("TIFFGetField TIFFTAG_COLORMAP failed!");
      break;
    }
  default:
    error("tiff file PhotoMetric Interpretation not recognized");
  }

  tiffbuf = (unsigned char*)malloc((unsigned)TIFFScanlineSize(tif));
  if (!tiffbuf) error("can't allocate tiff scanline buffer");
  image = TpmImageCreate(width, height, channels);

  /* read the TIFF line by line, and copy into the image. */
  if(channels == 1) {
    for(y=0;y<height;y++) {
      TIFFReadScanline(tif, tiffbuf, y, 0);
      memcpy(image->d2[y], tiffbuf, width);
    }
  }
  else {
    /* assume channels == 3 */
    unsigned char *rptr = *image->d3[0],
                  *gptr = *image->d3[1],
                  *bptr = *image->d3[2];
    for(y=0;y<height;y++) {
      TIFFReadScanline(tif, tiffbuf, y, 0);
      ptr = tiffbuf;
      for(x=0;x<width;x++) {
	*rptr++ = *ptr++;
	*gptr++ = *ptr++;
	*bptr++ = *ptr++;
      }
    }
  }
  free(tiffbuf);

  return image;
}

static void errorHandler(const char* module, const char* fmt, va_list ap)
{
#if DEBUG
  if (module != NULL)
    fprintf(stderr, "%s: ", module);
  vfprintf(stderr, fmt, ap);
  fprintf(stderr, ".\n");
#endif
}

TpmImage TpmReadTiff(char *fname)
{
  TpmImage image;
  TIFF *tif;
  TIFFErrorHandler old_handler;
  if(!strcmp(fname, "-")) return TpmReadTiffStream(stdin);
  old_handler = TIFFSetErrorHandler(errorHandler);
  tif = TIFFOpen(fname, "rb");
  if(!tif) { TIFFSetErrorHandler(old_handler); return NULL; }
  image = TpmImageFromTiff(tif);
  TIFFClose(tif);
  TIFFSetErrorHandler(old_handler);
  return image;
}

TpmImage TpmReadTiffStream(FILE *fp)
{
  TpmImage image;
  TIFFErrorHandler old_handler;
  TIFF *tif;

  /* Check the header before calling the library */
  if(FileCheckHeader(fp, "II\000*", 3) && FileCheckHeader(fp, "MM\000*", 3)) {
#if DEBUG
    fprintf(stderr, "header match failed\n");
#endif
    return NULL;
  }

  old_handler = TIFFSetErrorHandler(errorHandler);
  /* sync the fileno with the FILE* */
  lseek(fileno(fp), ftell(fp), 0);
  tif = TIFFFdOpen(fileno(fp), "<unknown>", "rb");
  if(!tif) { TIFFSetErrorHandler(old_handler); return NULL; }
  image = TpmImageFromTiff(tif);
  TIFFClose(tif);
  TIFFSetErrorHandler(old_handler);
  return image;
}

int TpmImageWriteTiff(FILE *fp, TpmImage image)
{
  TIFFErrorHandler old_handler = TIFFSetErrorHandler(errorHandler);
  TIFF *tif = TIFFFdOpen(dup(fileno(fp)), "<unknown>", "wb");
  if(!tif) { TIFFSetErrorHandler(old_handler); return TPM_ERROR; }

  TIFFSetField(tif, TIFFTAG_IMAGEWIDTH, image->width);
  TIFFSetField(tif, TIFFTAG_IMAGELENGTH, image->height);
  TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, image->channels);
  TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE, 8);
  TIFFSetField(tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
  TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
  /* I don't know why I have to set this, but it doesn't work otherwise. */
  TIFFSetField(tif, TIFFTAG_ROWSPERSTRIP, 12);

  if(image->channels == 1) {
    int y;
    for(y=0;y<image->height;y++) {
      TIFFWriteScanline(tif, image->d2[y], y, 0);
    }
  }
  else {
    int x,y;
    /* assume channels == 3 */
    unsigned char *rptr = *image->d3[0],
                  *gptr = *image->d3[1],
                  *bptr = *image->d3[2];
    unsigned char *tiffbuf = (unsigned char*)malloc(TIFFScanlineSize(tif));
    if (!tiffbuf) error("can't allocate tiff scanline buffer");
    for(y=0;y<image->height;y++) {
      unsigned char *ptr = tiffbuf;
      for(x=0;x<image->width;x++) {
	*ptr++ = *rptr++;
	*ptr++ = *gptr++;
	*ptr++ = *bptr++;
      }
      TIFFWriteScanline(tif, tiffbuf, y, 0);
    }
    free(tiffbuf);
  }

  TIFFClose(tif);
  TIFFSetErrorHandler(old_handler);
  return TPM_OK;
}
