#include <tpm/tpm.h>
#include <xtpm/xtpm.h>
#include <xtpm/app.h>
#include "image.h"

#define FAST_FILL 1
#define ColorToGray(r,g,b) (unsigned char)(0.30*(r)+0.59*(g)+0.11*(b)+0.5)

/* Prototypes ****************************************************************/

Pixmap TpmImageToPixmap(XtpmAppData appData, TpmImage image);
void TpmImageFillXImage(XImage *ximage, TpmImage image);

/* Functions *****************************************************************/

/* Fill an XImage with the contents of image */
/* If ximage->bpp > image->channels, values will be replicated.
 * If ximage->bpp < image->channels, low order bits (or bytes) will be used.
 * Should work regardless of byte ordering (MSBfirst or LSBfirst).
 */
#define LSB3to4_1(r,g,b) \
  ((r) + ((g) << 8) + ((b) << 16))
#define LSB3to4_2(r,g,b) \
  (LSB3to4_1(r,g,b) << 8)
#define MSB3to4_1(r,g,b) \
  (((r) << 16) + ((g) << 8) + (b))
#define MSB3to4_2(r,g,b) \
  (MSB3to4_1(r,g,b) << 8)
void TpmImageFillXImage(XImage *ximage, TpmImage image)
{
  int xi_channels, i, j, image_size;
  unsigned char *xdptr;
  int pad; /* extra bytes at end of each ximage row */

  xdptr = (unsigned char *)ximage->data;
  image_size = ximage->width * ximage->height;
  xi_channels = ximage->bytes_per_line / ximage->width;
  /* some servers like to round bytes_per_line up to a multiple of four */
  pad = ximage->bytes_per_line/xi_channels % ximage->width;

  if(image->channels == 3) {
    unsigned char *rptr = *image->d3[0],
                  *gptr = *image->d3[1],
                  *bptr = *image->d3[2];
    if(xi_channels == 4) {
#if FAST_FILL
      /* faster, but not as robust */
      /* HP700: _rgb (red_mask == 0xff0000)
       * SGI:   _bgr (red_mask == 0x0000ff)
       */
      if((ximage->byte_order == MSBFirst) &&
	 (ximage->red_mask == 0xff0000)) {
	for(i=0;i<image_size;i++) {
	  xdptr++;
	  *xdptr++ = *rptr++;
	  *xdptr++ = *gptr++;
	  *xdptr++ = *bptr++;
	}
      }
      else {
	/* reverse of above */
	for(i=0;i<image_size;i++) {
	  xdptr++;
	  *xdptr++ = *bptr++;
	  *xdptr++ = *gptr++;
	  *xdptr++ = *rptr++;
	}
      }
#else
      /* Use XPutPixel for safety */
      if((ximage->byte_order == MSBFirst) &&
	 (ximage->red_mask == 0xff0000)) {
	for(i=0;i<ximage->height;i++) {
	  for(j=0;j<ximage->width;j++) {
	    XPutPixel(ximage, j, i, MSB3to4_1(*rptr++, *gptr++, *bptr++));
	  }
	}
      }
      else {
	for(i=0;i<ximage->height;i++) {
	  for(j=0;j<ximage->width;j++) {
	    XPutPixel(ximage, j, i, LSB3to4_1(*rptr, *gptr++, *bptr++));
	  }
	}
      }
#endif
    }
    else if(xi_channels == 1) {
      /* Compute NTSC gray = 0.30 red, 0.59 green, 0.11 blue */
      for(i=0;i<ximage->height;i++) {
        for(j=0;j<ximage->width;j++) {
          *xdptr++ = ColorToGray(*rptr++, *gptr++, *bptr++);
        }
        xdptr += pad;
      }
    }
    else {
      fprintf(stderr, "TpmImageFillXImage: unexpected xi_channels = %d\n", 
	      xi_channels);
    }
  }
  else if(image->channels == 1) {
    unsigned char *ptr = image->d;
    if(xi_channels == 4) {
      /* put same byte in every channel */
      for(i=0;i<image_size;i++) {
	*xdptr++ = *ptr;
	*xdptr++ = *ptr;
	*xdptr++ = *ptr;
	*xdptr++ = *ptr++;
      }
    }
    else if(xi_channels == 1) {
      if(pad) {
	/* do one scanline at a time */
	for(i=0;i<ximage->height;i++) {
	  memcpy(xdptr, ptr, ximage->width);
	  xdptr += ximage->width + pad;
	  ptr += ximage->width;
	}
      }
      else {
	/* no pad; a single memcpy will do */
	memcpy(xdptr, ptr, image_size);
      }
    }
    else {
      fprintf(stderr, "TpmImageFillXImage: unexpected xi_channels = %d\n", 
	      xi_channels);
    }
  }
  else {
    fprintf(stderr, "TpmImageFillXImage: unexpected image->channels = %d\n", 
	    image->channels);
  }
}

Pixmap TpmImageToPixmap(XtpmAppData appData, TpmImage image)
{
  XImage *ximage;
  Pixmap pixmap;
  int total_size;

  if(!image) return NULL;

  /* copy the image into an ximage */
  ximage = XCreateImage(appData->display, 
                        appData->visual,
                        appData->depth,
                        ZPixmap, 
                        0, 
                        NULL,
                        image->width,
                        image->height,
                        32,
                        0);
  total_size = ximage->bytes_per_line * image->height;
  ximage->data = (char*)malloc(total_size);
  if(!ximage->data) {
    fprintf(stderr, "TpmImageToPixmap: Could not allocate %d by %d ximage\n",
	    ximage->width, ximage->height);
    return NULL;
  }
  TpmImageFillXImage(ximage, image);
  
  /* Allocate pixmap */
  pixmap = XCreatePixmap(appData->display,
                         appData->root,
                         image->width,
                         image->height,
                         appData->depth);
  if(!pixmap) {
    fprintf(stderr, "TpmImageToPixmap: XCreatePixmap failed\n");
    return NULL;
  }
  
  /* Copy the image to the pixmap */
  XPutImage(appData->display, 
            pixmap,
            appData->gc,
            ximage, 
            0, 0, 
            0, 0,
            image->width,
            image->height);

  free(ximage->data); ximage->data = NULL;
  XDestroyImage(ximage);
  return pixmap;
}
