#include <tpm/boa_learn.h>
#include <tpm/hier.h>

#define DEFAULT_CLASSIFIER T_Classifier_TwoLayer

/* Prototypes ****************************************************************/
void Boa_Learn_Pkg(void);

Obj Boa_Classifier_Create(Type classifier_t, Obj features, CmpFunc *compare,
			  InstanceFunc *instance);
Obj Boa_Classifier_CreateBlock(Obj classifier);
List Boa_Classifier_AllMatching(Obj filter, Any item, CmpFunc *cf);
int Boa_Classifier_Dump(Obj filter, Obj dest);

Obj Boa_MultiClass_Create(Obj classifier);
Obj Boa_MultiClass_CreateBlock(Obj classifiers);
Obj Boa_MultiClass_CreateExclusive(Obj classifiers);
Obj Boa_MultiClass_AddClass(Obj classifiers);

Learn_BenchData Learn_BenchData_Create(void);
void Learn_BenchData_Free(Learn_BenchData data);
Learn_BenchGroup Learn_BenchData_AddGroup(Learn_BenchData data);
Learn_BenchInstance Learn_BenchData_AddInstance(Learn_BenchData data, 
						int label, float weight, 
						Obj instance, 
						Learn_BenchGroup group);
int Learn_Benchmark(Obj classifiers, Learn_BenchData truth, 
		    FILE *output, FILE *log);

/* Functions *****************************************************************/

void Boa_Learn_Pkg(void)
{
  if(T_Classifier_SSC) return;  /* don't init twice */

  Boa_Hier_Pkg();
  Boa_StringSet(Boa_Types, "classifier_ssc", BOA_INSERT, 
		Boa_Classifier_SSC_Define());
  Boa_StringSet(Boa_Types, "classifier_sc", BOA_INSERT, 
		Boa_Classifier_SC_Define());
  Boa_StringSet(Boa_Types, "classifier_cover", BOA_INSERT, 
		Boa_Classifier_Cover_Define());
  Boa_StringSet(Boa_Types, "classifier_tree", BOA_INSERT, 
		Boa_Classifier_Tree_Define());
  Boa_StringSet(Boa_Types, "classifier_stump", BOA_INSERT, 
		Boa_Classifier_Stump_Define());

  Boa_StringSet(Boa_Types, "classifier_nearest", BOA_INSERT, 
		Boa_Classifier_Nearest_Define());
  Boa_StringSet(Boa_Types, "classifier_triangle", BOA_INSERT, 
		Boa_Classifier_Triangle_Define());
  Boa_StringSet(Boa_Types, "classifier_kernel", BOA_INSERT, 
		Boa_Classifier_Kernel_Define());
  Boa_StringSet(Boa_Types, "classifier_select", BOA_INSERT, 
		Boa_Classifier_Select_Define());
  Boa_StringSet(Boa_Types, "classifier_kr", BOA_INSERT, 
		Boa_Classifier_KR_Define());
  Boa_StringSet(Boa_Types, "classifier_cdm", BOA_INSERT, 
		Boa_Classifier_CDM_Define());

  Boa_StringSet(Boa_Types, "classifier_belief", BOA_INSERT, 
		Boa_Classifier_Belief_Define());
  Boa_StringSet(Boa_Types, "classifier_spanTree", BOA_INSERT, 
		Boa_Classifier_SpanTree_Define());

  Boa_StringSet(Boa_Types, "classifier_twoLayer", BOA_INSERT, 
		Boa_Classifier_TwoLayer_Define());
#ifndef NO_LAPACK
  Boa_StringSet(Boa_Types, "classifier_butterfly", BOA_INSERT, 
		Boa_Classifier_Butterfly_Define());
#endif

  Boa_StringSet(Boa_Types, "featureMap", BOA_INSERT, Boa_FeatureMap_Define());
  Boa_StringSet(Boa_Types, "exclusive", BOA_INSERT, Boa_Exclusive_Define());
  Boa_StringSet(Boa_Types, "euclid", BOA_INSERT, Boa_Euclid_Define());
  Boa_StringSet(Boa_Types, "cophenetic", BOA_INSERT, Boa_Cophenetic_Define());
  Boa_StringSet(Boa_Types, "proximity", BOA_INSERT, Boa_Proximity_Define());
}

static List IntInstance(void *data)
{
  List instances = ListCreate((FreeFunc*)Boa_Free);
  Obj t = Boa_CreateFromData(T_Int, NULL);
  t->value = data;
  ListAddFront(instances, t);
  return instances;
}

/* Creates a classifier decorated by a cached featureMap.
 * features should be an Ord of Hiers.
 * If compare != NULL, it specifies a comparison function between instances
 * and tree data.
 * If instance != NULL, it will be used to prime the featureMap cache
 * from the leaf data.
 */
Obj Boa_Classifier_Create(Type classifier_t, Obj features, CmpFunc *compare,
			  InstanceFunc *instance)
{
  Obj featureMap, classifier;
  if(!classifier_t) classifier_t = DEFAULT_CLASSIFIER;
  featureMap = Boa_Create(T_FeatureMap);
  Boa_StringSet(featureMap, "_features", BOA_REPLACE, features);
  if(compare) Boa_StringSet(featureMap, "_compare", BOA_REPLACE, 
			    Boa_Ptr_Create((Ptr)compare));

  /* initialize the cache */
  featureMap = Boa_Cache_Create(featureMap);
  if(!instance && (!compare || (compare == IntCmp))) instance = IntInstance;
  if(instance) {
    Obj cache;
    int i;
    Type pairType = Boa_Type_Create("(tuple 2)");
    assert(Boa_StringGet(featureMap, "_cache", 0, NULL, &cache) == TPM_OK);
    for(i=0;i<Boa_Ord_Length(features);i++) {
      Obj feature = Boa_Ord_Item(features, i);
      if(feature->type->type == T_Hier) {
	Hier hier = Boa_Hier_Value(feature);
	int j;
	for(j=0;j<hier->num_nodes;j++) {
	  HierNode node = HierNodeAt(hier, j);
	  Obj t, pair, pairs;
	  List instances;
	  if(node->children) continue;
	  instances = instance(&node->data);
	  {ListIter(p, t, instances) {
	    pair = Boa_Create(pairType);
	    Boa_Hold(feature);
	    Boa_Tuple_Item(pair, 0) = feature;
	    Boa_Tuple_Item(pair, 1) = Boa_Int_Create(j);
	    if(Boa_ObjGet(cache, t, 0, NULL, &pairs) != TPM_OK) {
	      pairs = Boa_Create(T_List);
	      assert(Boa_ObjSet(cache, t, BOA_INSERT, Boa_Hold(pairs)) 
		     == TPM_OK);
	    }
	    ListAddFront(Boa_List_Value(pairs), pair);
	    Boa_Release(pairs);
	  }}
	  ListFree(instances);
	}
      }
    }
    Boa_Release(cache);
  }

  classifier = Boa_Create(T_Filter);
  Boa_Filter_Context(classifier) = Boa_Create(classifier_t);
  Boa_Filter_ObjMap(classifier) = featureMap;
  return classifier;
}

/* Returns a new classifier decorated by a mapCon, decorated by the
 * same cached featureMap decorated by a cached mapCon.  Follow that?
 * Does not take hold of classifier (caller still owns it).
 */
Obj Boa_Classifier_CreateBlock(Obj classifier)
{
  Obj filter, t;
  filter = Boa_Create(T_Filter);

  t = Boa_Filter_Context(classifier);
  Boa_Hold(t);
  Boa_Filter_Context(filter) = Boa_MapCon_Create(t);

  t = Boa_Filter_ObjMap(classifier);
  Boa_Hold(t);
  Boa_Filter_ObjMap(filter) = Boa_Cache_Create(Boa_MapCon_Create(t));
  return filter;
}

#if 0
List Boa_Classifier_AllMatching(Obj filter, Any item, CmpFunc *cf)
{
  List result;
  Obj featureMap, features, t;
  featureMap = Boa_Cache_Context(Boa_Filter_ObjMap(filter));
  Boa_StringGet(featureMap, "_features", 0, NULL, &features);
  Boa_StringGet(featureMap, "_compare", 0, NULL, &t);
  result = BoaClassifier_AllMatching(Boa_Filter_Context(filter), 
				     features, (CmpFunc*)Boa_Ptr_Value(t),
				     item, cf);
  Boa_Release(t);
  Boa_Release(features);
  return result;
}
#endif

int Boa_Classifier_Dump(Obj filter, Obj dest)
{
  Obj classifier = Boa_Filter_Context(filter);
  Obj data = *(Obj*)classifier->value;
  Obj featureMap = Boa_Cache_Context(Boa_Filter_ObjMap(filter));
  Obj features;
  int i;

  assert(Boa_StringGet(featureMap, "_features", 0, NULL, &features) == TPM_OK);
  for(i=0;i<Boa_Ord_Length(features);i++) {
    Obj feature = Boa_Ord_Item(features, i);
    Obj t;
    if(Boa_ObjGet(data, feature, 0, NULL, &t) != TPM_OK) {
      /* haven't seen this feature before */
      continue;
    }
    if(Boa_StringSet(dest, Boa_String_Value(Boa_Ord_Name(features, i)), 
		     BOA_INSERT|BOA_REPLACE, t) != TPM_OK) {
      return TPM_BOTTOM;
    }
  }
  Boa_Release(features);
  return TPM_OK;
}

/* Returns an Ord of classifiers, each decorated by a cached featureMap
 * that they all share.
 * Initially has one class. 
 */
Obj Boa_MultiClass_Create(Obj classifier)
{
  Obj classifiers = Boa_Create(T_Ord);
  Boa_StringSet(classifiers, "0", BOA_INSERT, classifier);
  return classifiers;
}

/* Returns a new Ord where each classifier is decorated to emit
 * negative examples for other classes.
 * This should be done after CreateBlock if you want blocking.
 */
Obj Boa_MultiClass_CreateExclusive(Obj classifiers)
{
  int i;
  Obj result = Boa_Create(T_Ord);
  Obj posEx;
  char name[16];
  if(!Boa_Ord_Length(classifiers)) return result;
  posEx = Boa_Create(T_List);
  for(i=0;i<Boa_Ord_Length(classifiers);i++) {
    Obj classifier, t;
    classifier = Boa_Ord_Item(classifiers, i);
    t = Boa_MultiClassExclusive_Create(Boa_Hold(classifier),
				       result, Boa_Hold(posEx));

    sprintf(name, "%d", i);
    assert(Boa_StringSet(result, name, BOA_INSERT, t) == TPM_OK);
  }  
  Boa_Release(posEx);
  return result;
}

/* Returns a new Ord of classifiers where each classifier has been blocked,
 * as in Boa_Classifier_CreateBlock().
 * Does not take hold of classifiers (caller still owns it).
 */
Obj Boa_MultiClass_CreateBlock(Obj classifiers)
{
  int i;
  Obj result = Boa_Create(T_Ord);
  Obj objMap;
  char name[16];
  if(!Boa_Ord_Length(classifiers)) return result;
  /* one objMap to be shared among all */
  objMap = Boa_Filter_ObjMap(Boa_Ord_Item(classifiers, 0));
  objMap = Boa_Cache_Create(Boa_MapCon_Create(Boa_Hold(objMap)));
  for(i=0;i<Boa_Ord_Length(classifiers);i++) {
    Obj classifier, filter, t;
    classifier = Boa_Ord_Item(classifiers, i);

    filter = Boa_Create(T_Filter);
    Boa_Filter_ObjMap(filter) = Boa_Hold(objMap);
    t = Boa_Filter_Context(classifier);
    /* no cache */
    Boa_Filter_Context(filter) = Boa_MapCon_Create(Boa_Hold(t));

    sprintf(name, "%d", i);
    assert(Boa_StringSet(result, name, BOA_INSERT, filter) == TPM_OK);
  }
  Boa_Release(objMap);
  return result;
}

/* Add a new classifier to the Ord, using the same features. 
 * If existing classifiers are blocked or exclusive, maintains that property.
 */
Obj Boa_MultiClass_AddClass(Obj classifiers)
{
  Obj objMap, classifier, prototype;
  Type classifierType;
  char name[16];
  int blocked;

  prototype = Boa_Ord_Item(classifiers, 0);
  /* give it the same objMap as everyone else, to share the cache */
  /* use get(_objMap) because this will get through a T_Exclusive */
  assert(Boa_StringGet(prototype, "_objMap", 0, 
		       NULL, &objMap) == TPM_OK);
  /* use the same classifier type */
  assert(Boa_StringGet(prototype, "_context", 0, 
		       NULL, &classifier) == TPM_OK);
  if(classifier->type == T_MapCon) {
    Obj t;
    assert(Boa_StringGet(classifier, "_context", 0, NULL, &t) == TPM_OK);
    Boa_Release(classifier);
    classifier = t;
    blocked = 1;
  }
  else blocked = 0;
  classifierType = classifier->type;
  Boa_Release(classifier);

  classifier = Boa_Create(T_Filter);
  Boa_Filter_Context(classifier) = Boa_Create(classifierType);
  if(blocked) {
    /* wrap with a mapCon */
    Boa_Filter_Context(classifier) = 
      Boa_MapCon_Create(Boa_Filter_Context(classifier));
  }
  Boa_Filter_ObjMap(classifier) = objMap;

  sprintf(name, "%d", Boa_Ord_Length(classifiers));
  /* are these classifiers exclusive? */
  if(prototype->type == T_Exclusive) {
    /* get the posEx list from classifier 0 */
    Obj posEx = Boa_Exclusive_PosEx(prototype);
    Obj t;
    /* make all posEx so far negatives for the new class */
    ListIter(p, t, Boa_List_Value(posEx)) {
      assert(Boa_ObjSet(classifier, t, BOA_INSERT, 
			Boa_Float_Create(-1)) == TPM_OK);
      if(0) fprintf(stderr, "adding negative for class %s\n", name);
    }
    classifier = Boa_MultiClassExclusive_Create(classifier, classifiers,
						posEx);
  }
  assert(Boa_StringSet(classifiers, name, BOA_INSERT, classifier) == TPM_OK);
  return classifier;
}

/* Bench *********************************************************************/

static void Learn_BenchGroup_Free(Learn_BenchGroup data)
{
  int i;
  free(data->count);
  for(i=0;i<MAX_LABELS;i++) {
    if(data->errors[i]) ListFree(data->errors[i]);
  }
  free(data->errors);
  free(data);
}
static void Learn_BenchInstance_Free(Learn_BenchInstance data)
{
  Boa_Release(data->instance);
  free(data->weight);
  free(data);
}
Learn_BenchData Learn_BenchData_Create(void)
{
  Learn_BenchData data = Allocate(1, struct Learn_BenchData);
  data->instances = ListCreate((FreeFunc*)Learn_BenchInstance_Free);
  data->groups = ListCreate((FreeFunc*)Learn_BenchGroup_Free);
  return data;
}
void Learn_BenchData_Free(Learn_BenchData data)
{
  ListFree(data->instances);
  ListFree(data->groups);
  free(data);
}

Learn_BenchGroup Learn_BenchData_AddGroup(Learn_BenchData data)
{
  Learn_BenchGroup group = AllocateC(1, struct Learn_BenchGroup);
  group->count = AllocateC(MAX_LABELS, int);
  group->errors = AllocateC(MAX_LABELS, List);
  ListAddFront(data->groups, group);
  return group;
}
static int InstanceCompare(Learn_BenchInstance ins, Obj t)
{
  return ins->instance != t;
}
Learn_BenchInstance Learn_BenchData_AddInstance(Learn_BenchData data, 
						int label, float weight, 
						Obj instance, 
						Learn_BenchGroup group)
{
  Learn_BenchInstance ins;
  int i;
  /* search for the instance */
  i = ListIndexAtValue(data->instances, instance, (CmpFunc*)InstanceCompare);
  if(i != -1) {
    ins = (Learn_BenchInstance)ListValueAtIndex(data->instances, i);
  }
  else {
    ins = Allocate(1, struct Learn_BenchInstance);
    ins->instance = instance;
    ins->group = group;
    ins->weight = AllocateC(MAX_LABELS, float);
    ins->weight[label] = weight;
    ListAddFront(data->instances, ins);
  }
  if(group && (weight > 0)) group->count[label]++;
  return ins;
}

#define VERBOSE 1
#define RANDOM 1
#define RANDOM_WEIGHT 0

/* Benchmark a multiclassifier.
 * Randomly picks a node from truth and gives it to the classifier.
 * The number of classification errors after each step is written
 * to output.
 * The total number of examples to reach zero error is returned.
 * This function should be called multiple times to account for the
 * random component.
 * Currently assumes exclusive classes.
 */
int Learn_Benchmark(Obj classifiers, Learn_BenchData truth, 
		    FILE *output, FILE *log)
{
  Obj instances, classifier, block_classifiers;
  int i, examples, label, labels;
  Learn_BenchInstance ins;
  Learn_BenchGroup grp;
  int *ex_per_class;

  labels = Boa_Ord_Length(classifiers);
  if(RANDOM) fprintf(stderr, "random teaching\n");
  else fprintf(stderr, "sequential teaching\n");
#if VERBOSE
  fprintf(stderr, "%d labels\n", labels);
#endif
  ex_per_class = AllocateC(labels, int);

  /* make block classifiers for fast queries */
  block_classifiers = Boa_MultiClass_CreateBlock(classifiers);

  /* collect the requests */
  instances = Boa_Create(T_Index);
  i = 0;
  {ListIter(p, ins, truth->instances) {
    assert(Boa_IntSet(instances, i, BOA_INSERT, Boa_Hold(ins->instance)) 
	   == TPM_OK);
    i++;
  }}

  /* benchmark the classifier */
  srand(getpid());
  for(examples = 0;; examples++) {
    List errors = ListCreate(NULL);  /* List of shared BenchData */
    int false_pos = 0;  /* number of false labels assigned */

    /* initialize groups */
    {ListIter(p, grp, truth->groups) {
      for(label=0;label<labels;label++) {
	if(grp->errors[label]) ListFree(grp->errors[label]);
	grp->errors[label] = ListCreate(NULL);
      }
    }}

    /* check the membership of all instances */
    /* loop labels */
    for(label = 0; label < Boa_Ord_Length(classifiers); label++) {
      Obj results;
      ListNode truthNode;
      classifier = Boa_Ord_Item(block_classifiers, label);
      /* make the requests as a unit */
      if(Boa_ObjGet(classifier, instances, 0, &T_Index, &results) != TPM_OK) {
	Obj instance, t;
	results = Boa_Create(T_Index);
	for(i=0;i<Boa_Index_Length(instances);i++) {
	  instance = Boa_Index_Item(instances, i);
	  if(Boa_ObjGet(classifier, instance, 0, &T_Float, &t) != TPM_OK) {
	    ErrorPrepend("Classifier get failed for label %d", label);
	    ErrorPrint(stderr);
	    return NAI;
	  }
	  assert(Boa_IntSet(results, i, BOA_INSERT, t) == TPM_OK);
	}
      }
      for(i=0, 
	  truthNode = ListFront(truth->instances);
	  i<Boa_Index_Length(instances);
	  i++,
	  truthNode = truthNode->next) {
	int old_errs, old_grp_errs;
	float f = Boa_Float_Value(Boa_Index_Item(results, i));
	ins = (Learn_BenchInstance)truthNode->data;
#if VERBOSE > 1
	fprintf(stderr, "label %d for ", label);
	Boa_Print(stderr, ins->instance);
	fprintf(stderr, " is %f --> ", f);
	old_errs = ListSize(errors);
	if(ins->group) old_grp_errs = ListSize(ins->group->errors[label]);
	if(log) {
	  Boa_Print(log, ins->instance);
	  fprintf(log, " %g\n", f);
	}
#endif
	/* bad weight? */
	if(f != f) {
	  ListAddFront(errors, ins);
	}
	else {
	  if(!ins->group) {
	    if((ins->weight[label] <= 0) && (f > 0)) {
	      ListAddFront(errors, ins);
	      false_pos++;
	    }
	    else if((ins->weight[label] < 0) && (f == 0)) {
	      ListAddFront(errors, ins);
	    }
	    else if((ins->weight[label] > 0) && (f <= 0)) {
	      ListAddFront(errors, ins);
	    }
	  }
	  else {
	    if((ins->group->count[label] == 0) && (f > 0)) {
	      ListAddFront(ins->group->errors[label], ins);
	    }
	    else if((ins->weight[label] > 0) && (f <= 0)) {
	      ListAddFront(ins->group->errors[label], ins);
	    }
	  }
	}
#if VERBOSE > 1
	if((ListSize(errors) > old_errs) ||
	   (ins->group && 
	    (ListSize(ins->group->errors[label]) > old_grp_errs)))
	  fprintf(stderr, "ERROR\n");
	else fprintf(stderr, "OK\n");
#endif
      }
      Boa_Release(results);
    }

    /* check groups */
    {ListIter(p, grp, truth->groups) {
      for(label=0;label<labels;label++) {
	/* count/2 makes it vote; count says at least 1 */
	if(ListSize(grp->errors[label]) &&
	   (grp->count[label] <= ListSize(grp->errors[label]))) {
	  /* pick an error at random */
	  i = RandInt(1, ListSize(grp->errors[label]));
	  ins = (Learn_BenchInstance)ListValueAtIndex(grp->errors[label], i-1);
	  ListAddFront(errors, ins);
	  if(grp->count[label] == 0) false_pos++;
	}
	/* otherwise the group is covered */
      }
    }}

    fprintf(stderr, "%d errors (%d false pos) after %d examples\n", ListSize(errors), false_pos, examples);
    fprintf(output, "%d %d\n", ListSize(errors), false_pos);
    fflush(output);
    if(!ListSize(errors)) { ListFree(errors); break; }

#if RANDOM
    /* pick an error at random */
    i = RandInt(1, ListSize(errors));
#else
    /* pick the first error (which is last in the list) */
    i = ListSize(errors);
#endif
    ins = (Learn_BenchInstance)ListValueAtIndex(errors, i-1);
    ListFree(errors);

    /* and correct it */
    for(label=0;label<labels;label++) {
      float weight;
      if(ins->weight[label] > 0) {
	if(log) {
	  Boa_Print(log, ins->instance);
	  fprintf(log, " %g %d\n", ins->weight[label], label);
	}
	ex_per_class[label]++;
	weight = ins->weight[label];
      }
      else weight = -1;
#if RANDOM_WEIGHT
      /* my impression of sampling from a linear distribution */
      weight *= sqrt(RandFloat()) * 2 - 1;
#endif
      classifier = Boa_Ord_Item(classifiers, label);
      if(Boa_ObjSet(classifier, ins->instance, BOA_INSERT,
		    Boa_Float_Create(weight)) != TPM_OK) {
	ErrorPrepend("Classifier set failed for label %d", label);
	ErrorPrint(stderr);
      }
    }
    if(log) fflush(log);
  }
  Boa_Release(instances);
  Boa_Release(block_classifiers);

  fprintf(stderr, "Label\tPosExamples\n");
  for(label=0;label<labels;label++) {
    fprintf(stderr, "%d\t%d\n", label, ex_per_class[label]);
  }
  free(ex_per_class);
  fprintf(stderr, "Total\t%d\n", examples);
  return examples;
}
