/* Hier (hierarchy) data type. Designed for efficient storage and access.
 * Format is an array of equally-sized nodes, where each node is an
 * integer count of its children and an integer index to its parent
 * followed by user-defined data.
 * The node order corresponds to a preorder traversal
 * (as opposed to Trees which are stored postorder).
 */

#include "hier.h"

/* Prototypes ****************************************************************/

Hier HierCreate(int num_nodes, int data_size);
void HierFree(Hier hier);
int HierRead(Hier hier, FILE *h);
Hier HierCreateFromStream(FILE *h);
Hier HierCreateFromFile(char *file);
int HierWrite(Hier hier, FILE *h);
int HierRightLeaf(Hier hier, int node);

/* Functions *****************************************************************/

/* Creates a new, uninitialized hierarchy. */
Hier HierCreate(int num_nodes, int data_size)
{
  Hier hier = Allocate(1, struct Hier);
  hier->num_nodes = num_nodes;
  hier->data_size = data_size;
  /* may have to use padding to make the nodes aligned */
  if(data_size % sizeof(int)) 
    hier->pad = sizeof(int) - (data_size % sizeof(int));
  else 
    hier->pad = 0;
  hier->node = Allocate(num_nodes * HierNodeSize(hier), char);
  return hier;
}

void HierFree(Hier hier)
{
  free(hier->node);
  free(hier);
}

/* Reads node data from file into an existing hier. 
 */
int HierRead(Hier hier, FILE *h)
{
  int status = fread(hier->node, 1, hier->num_nodes*HierNodeSize(hier), h);
  if(status != hier->num_nodes*HierNodeSize(hier)) return TPM_ERROR;
  /* this is a kludge which doesn't really work */
#ifndef TPM_SWAP
  SwapBuffer(hier->node, hier->num_nodes*HierNodeSize(hier)/sizeof(int), sizeof(int));
#endif
  return TPM_OK;
}

Hier HierCreateFromStream(FILE *h)
{
  Hier hier;
  int num_nodes, data_size;

  /* determine parameters from the header */
  if(fscanf(h, "hier %d (%d char)", &num_nodes, &data_size) < 2) {
    return NULL;
  }
  getc(h); /* absorb the newline; not safe if done by fscanf */
  hier = HierCreate(num_nodes, data_size);
  if(HierRead(hier, h)) {
    /* Show a message, but don't do anything */
    fprintf(stderr, "HierCreateFromStream: error reading file\n");
  }
  return hier;
}

Hier HierCreateFromFile(char *file)
{
  FILE *h;
  Hier hier;
  if(file[0] == '-' && !file[1]) h = stdin;
  else h = fopen(file, "r");
  if(!h) return NULL;
  hier = HierCreateFromStream(h);
  fclose(h);
  return hier;
}

int HierWrite(Hier hier, FILE *h)
{
  int status;
  fprintf(h, "hier %d (%d char)\n", hier->num_nodes, hier->data_size);
  status = fwrite(hier->node, 1, hier->num_nodes*HierNodeSize(hier), h);
  if(!status) return TPM_ERROR;
#ifndef TPM_SWAP
  SwapBuffer(hier->node, hier->num_nodes*HierNodeSize(hier)/sizeof(int), sizeof(int));
#endif
  return TPM_OK;
}

/* Returns the index of the rightmost leaf under the given node. */
int HierRightLeaf(Hier hier, int node)
{
  int children;
  HierNode n = HierNodeAt(hier, node);
  for(children = n->children;
      children; children--) {
    n = (HierNode)((char*)n + HierNodeSize(hier));
    node++;
    children += n->children;
  }
  return node;
}
