/* Learning benchmark for metrics instead of trees.
 * metric_bench partition db metric metric ...
 */
#include <tpm/boa_learn.h>
#include <boa/boa.h>

#define PARTITION 1
#if PARTITION
#include <tpm/partition.h>
#endif

int main(int argc, char *argv[])
{
  Obj od, partitionObj, features, classifiers, t;
  Obj db, tests, members;
#if PARTITION
  Partition partition;
#else
  float *partition;
  int p_length;
#endif
  int i;
  FILE *fp, *log;
  char logfile[128];
  Learn_BenchData truth;
  Learn_BenchGroup grp;
  char *str;
  char *partitionFile, classifierName[128], *dbName;

  if(argc < 6) {
    fprintf(stderr, "usage: %s classifier partition db metric metric ... output\n", argv[0]);
    fprintf(stderr, "db is the name of a database\n");
    fprintf(stderr, "and metrics are under db/metric.\n");
    fprintf(stderr, "classifiers are sc, ssc, tree, nearest, belief\n");
    exit(1);
  }
  sprintf(classifierName, "classifier_%s", argv[1]);
  Boa_Startup();
  Boa_Hier_Pkg();
  Boa_Learn_Pkg();
  Boa_Tcl_Pkg();

  od = Boa_Create(T_ObjDir);

  /* read partition */
  partitionFile = argv[2];
#if PARTITION
  partition = PartitionCreateFromFile(partitionFile);
  if(!partition) {
    ErrorSetErrno();
    ErrorPrepend("Cannot load partition file `%s'", partitionFile);
    ErrorPrint(stderr);
    exit(1);
  }
#else
  if(Boa_StringGet(od, partitionFile, 0, NULL, &partitionObj) != TPM_OK) {
    ErrorPrepend("Cannot load partition file `%s'", partitionFile);
    ErrorPrint(stderr);
    exit(1);
  }
  assert(partitionObj->type->type == T_Array);
  partition = (float*)partitionObj->value;
  p_length = Boa_Array_Length(partitionObj->type);
#endif

  /* get the database directory */
  str = getenv("PHOTOBOOK_DB_DIR");
  if(!str) str = "";
  Boa_StringSet(od, "_path", BOA_REPLACE, Boa_String_01(str));
  dbName = argv[3];
  if(Boa_StringGet(od, dbName, 0, NULL, &db) != TPM_OK) {
    ErrorPrepend("Cannot open database `%s'", dbName);
    ErrorPrint(stderr);
    exit(1);
  }
  Boa_Release(od);
  assert(Boa_StringGet(db, "index", 0, NULL, &tests) == TPM_OK);
  if(Boa_StringGet(db, "cache/feature", '/', NULL, &members) != TPM_OK) {
    ErrorPrepend("Cannot open %s/cache/feature", dbName);
    ErrorPrint(stderr);
    exit(1);
  }
  /* get the metric directory */
  assert(Boa_StringGet(db, "metric", 0, NULL, &od) == TPM_OK);
  Boa_Release(db);

  /* read metrics */
  features = Boa_Create(T_Ord);
  for(i=4;i<argc-1;i++) {
    if(Boa_StringGet(od, argv[i], '/', NULL, &t) != TPM_OK) {
      ErrorPrepend("Cannot load metric `%s'", argv[i]);
      ErrorPrint(stderr);
      exit(1);
    }
    /* set the test and feature fields */
    assert(Boa_StringSet(t, "features", BOA_INSERT, Boa_Hold(members)) == TPM_OK);
    assert(Boa_StringSet(t, "test", BOA_INSERT, Boa_Hold(tests)) == TPM_OK);
    assert(Boa_StringSet(features, argv[i], BOA_INSERT, t) == TPM_OK);
  }
  Boa_Release(members);

  /* create the classifier */
  t = Boa_Type_Create(classifierName);
  classifiers = Boa_MultiClass_Create(Boa_Classifier_Create(
		  t, features, NULL, NULL));

  /* collect a ground-truth list */
  truth = Learn_BenchData_Create();
#if PARTITION
  for(i=0;i<partition->num_elements;i++) {
    if(!partition->class[i]) continue;
    while(partition->class[i]-1 >= Boa_Ord_Length(classifiers)) {
      Boa_MultiClass_AddClass(classifiers);
    }
    /* get the object corresponding to this index */
    assert(Boa_IntGet(tests, i, 0, NULL, &t) == TPM_OK);
    Learn_BenchData_AddInstance(truth, partition->class[i]-1, 1, t, NULL);
  }
  PartitionFree(partition);
#else
  grp = NULL;
  for(i=0;i<p_length;i++) {
    /* get the object corresponding to this index */
    assert(Boa_IntGet(tests, i, 0, NULL, &t) == TPM_OK);
    Learn_BenchData_AddInstance(truth, 0, partition[i], t, grp);
  }
#endif
  Boa_Release(tests);
  
  /* benchmark the classifier */
  fp = fopen(argv[argc-1], "w");
  sprintf(logfile, "%s.log", argv[argc-1]);
  log = fopen(logfile, "w");
  Learn_Benchmark(classifiers, truth, fp, log);
  fclose(log);
  fclose(fp);
  Learn_BenchData_Free(truth);
#if 0
  assert(Boa_StringSet(od, "_path", 0, Boa_String_01("/tmp/tpminka")) 
	 == TPM_OK);
  t = Boa_Ord_Item(classifiers, 0);
  if(Boa_Classifier_Dump(t, od) != TPM_OK) {
    ErrorPrint(stderr);
  }
#endif
  Boa_Release(classifiers);

  Boa_Release(od);
  Boa_Shutdown();
  exit(0);
}
