/* This file implements pure disjunctive learning.
 * Classes are described by a union of atomic groupings.
 */

#include "learn.h"
#include "common.h"
#include <math.h>

#define DEBUG 0
#define AQ 1
/* test test1.hier qq.hier
 * 0 0 -1
 * 0 1 1
 * 0 2 1
 * 0 4 1
 * 0 5 1
 */
#define RANDOM 0
/* non-flat bias generally negates effect of randomness */
#define SZB 1
/* size bias parameter */
#define SZ 1
/* use tree weights */
#define TW 0
/* be pessimistic */
#define PESS 0

#define Covers(class) (List)lh->classData[class]

/* Prototypes ****************************************************************/
void LearnUpdateClass(LearnHandle lh, int class);
List LearnWithClassMatching(LearnHandle lh, int class, 
			    void *item, CmpFunc *compare);
float LearnClassWeight(LearnHandle lh, int class, List leaves);

int LearnReadBias(LearnHandle lh, int tree, char *fname);
void LearnSaveBias(LearnHandle lh, int tree, char *fname);
void LearnUpdateBias(LearnHandle lh);

void *LearnClassDataInit(int class);
void LearnClassDataFree(void *covers);
void LearnClassDataPrint(LearnHandle lh, int class, FILE *fp);

float *TreeFreq(LearnHandle lh, int class);

/* Functions *****************************************************************/

float *TreeFreq(LearnHandle lh, int class)
{
  float *freq = AllocateC(lh->ntrees, float);
  int total;
  Cover cover;
  /* Loop covers */
  total = ListSize(Covers(class));
  {ListIter(p, cover, Covers(class)) {
    freq[cover->tree] += 1.0/total;
  }}
  return freq;
}

void LearnClassDataPrint(LearnHandle lh, int class, FILE *fp)
{
  fprintf(fp, "covers for class %d: tree low-high\n", class);
  {Cover cover;ListIter(p, cover, Covers(class)) {
    fprintf(fp, "  %d %d-%d\n", cover->tree, cover->node, cover->high);
  }}
}

void *LearnClassDataInit(int class)
{
  return ListCreate(CoverFree);
}
void LearnClassDataFree(void *covers)
{
  ListFree((List)covers);
}

#define SQR(x) ((float)(x)*(x))
static float NodeBias(LearnHandle lh, int tree, int bias, int item)
{
  TreeData *td = &lh->td[tree];
  float score, t;

#if SZB
  /* size bias: s * (1 - s) ^ k */
  t = HierRightLeaf(lh->trees[tree], item) - item + 1;
  t /= lh->trees[tree]->num_nodes;
  /* change the range of t to [0.1,0.9] instead of [0,1] */
  t = t*0.8 + 0.1;
  score = t * pow(1-t, SZ);
#else
  /* p(1-n) */
  score = (float)(td->bpos[bias][item] + 1) / (td->bptotal[bias] + 2);
  t = 1 - (float)(td->bneg[bias][item] + 1) / (td->bntotal[bias] + 2);
  score *= t;
#if PESS
  /* favor small nodes */
  t = HierRightLeaf(lh->trees[tree], item) - item + 1;
  t /= lh->trees[tree]->num_nodes;
  score -= t/1000;
#endif
#endif

#if 0
#if BIAS == 1
  /* (p+1)/(p+n+2) */
  score = (float)(td->bpos[bias][item] + 1) 
    / (td->bpos[bias][item] + td->bneg[bias][item] + 2);
#elif BIAS == 2
  /* p/n */
#if 1
  score = (float)(td->bpos[bias][item] + 1) / (td->bptotal[bias] + 2);
  score /= (float)(td->bneg[bias][item] + 1) / (td->bntotal[bias] + 2);
#else
  if(!td->bptotal[bias]) score = 0.5;
  else if(!td->bpos[bias][item]) score = 1.0 / td->bptotal[bias];
  else score = (float)td->bpos[bias][item] / td->bptotal[bias];
  if(td->bntotal[bias]) {
    if(!td->bneg[bias][item]) score *= td->bntotal[bias];
    else score /= (float)td->bneg[bias][item] / td->bntotal[bias];
  }
#endif
#elif BIAS == 3
  /* p-n */
  score = (float)(td->bpos[bias][item] + 1) / (td->bptotal[bias] + 2);
  score -= (float)(td->bneg[bias][item] + 1) / (td->bntotal[bias] + 2);
  if(score <= 0) score = 1e-10;
#endif
#endif
  return score;
}

void LearnUpdateBias(LearnHandle lh)
{
  int class, bias, tree, i;
  for(class=0;class<lh->nclasses;class++) {
    if(!ListSize(lh->pos_ex[class])) continue;
    for(tree=0;tree<lh->ntrees;tree++) {
      TreeData *td = &lh->td[tree];
      int best = 0;
      float maximum = -1e10;
      if(!td->enabled) continue;
      if(td->nbias == 1) continue;
      for(bias=0;bias<td->nbias;bias++) {
	float score = 0.0;
	float sum = 0.0, sum2 = 0.0, score2 = 0.0;
	float tsum = 0.0, tsum2 = 0.0;
	for(i=0;i<lh->trees[tree]->num_nodes;i++) {
	  float p, n;
	  float b, t;
#if 0
	  if(HierNodeAt(lh->trees[tree], i)->children) continue;
#endif
	  p = (float)td->npos[class][i]/TotalPos(class);
	  if(td->nneg[class][i])
	    n = (float)td->nneg[class][i]/TotalNeg(class);
	  else
	    n = 0;

#if 0
	  b = NodeBias(lh, tree, bias, i);
#else
	  b = (float)(td->bpos[bias][i] + 1) / (td->bptotal[bias] + 2);
	  b /= (float)(td->bneg[bias][i] + 1) / (td->bntotal[bias] + 2);
#endif
	  sum += b*b;
#if 0
	  t = p*(1-n);
#else
	  if(n >= p) t = 0;
	  else t = p-n;
#endif
	  tsum += t;
	  score += t * b;
#if 0
	  b = 1.0/b;
	  sum2 += b;
	  if(p >= n) t = 0;
	  else t = n-p;
	  tsum2 += t;
	  score2 += t * b;
#endif
	}
	score /= sqrt(sum);
	score /= tsum;
	if(score2) { 
	  score2 /= sum2; score2 /= tsum2; 
#if 0
	  score *= score2; 
#endif
	}
#if 0
	fprintf(stderr, "  class %d tree %d bias %d has score %g, %g\n", 
		class, tree, bias, score, score2);
#endif
	if(score > maximum) { maximum = score; best = bias; }
      }
      if(best != class+1) {
	if(best) fprintf(stderr, "!");
	else fprintf(stderr, ".");
      }
      fprintf(stderr, "class %d tree %d has bias %d with score %g\n", 
	      class, tree, best, maximum);
      td->cbias[class] = best;
    }
  }
}

/* find the best covers for the example.
 * Modifies *best_score and best.
 */
static void BestCovers(LearnHandle lh, int class, Example ex, 
		       float *best_score, List best)
{
  TreeData *td;
  Leaf leaf;
  float score;
  int i;
  /* loop leaves */
  {ListIter(p, leaf, ex->leaves) {
    td = &lh->td[leaf->tree];
    if(!td->enabled) continue;
    /* climb the tree and look for highest score */
    for(i = leaf->node; i >= 0;
	i = HierNodeAt(lh->trees[leaf->tree], i)->parent) {
      if(td->nneg[class][i]) break;
      score = td->npos[class][i] * 
#if TW
	td->weight *
#endif
	  NodeBias(lh, leaf->tree, td->cbias[class], i);
#if DEBUG
      fprintf(stderr, "%d %d sc: %g npos: %d bpos: %d/%d bneg: %d/%d\n", 
	      leaf->tree, i, score,
	      td->npos[class][i],
	      td->bpos[td->cbias[class]][i], td->bptotal[td->cbias[class]],
	      td->bneg[td->cbias[class]][i], td->bntotal[td->cbias[class]]);
#endif
      if(score > *best_score) {
	*best_score = score;
	ListFlush(best);
      }
      if(score == *best_score) {
	Cover c = Allocate(1, struct Cover);
	c->tree = leaf->tree;
	c->node = i;
	ListAddFront(best, c);
      }
    }
  }}
}

void LearnUpdateClass(LearnHandle lh, int class)
{
  Example ex, ex2;
  Cover cover;
  Leaf leaf;
  int i;
  float score, best_score;
  TreeData *td;
  List best;

  MEM_VERIFY_ALL();

  /* flush the old cover list */
  LearnClassDataFree(lh->classData[class]);
  lh->classData[class] = LearnClassDataInit(class);

  /* unmark examples */
  {ListIter(p, ex, lh->pos_ex[class]) {
    ex->mark = 0;
  }}
  
  /* loop positive examples */
  {ListIter(p, ex, lh->pos_ex[class]) {
    if(ex->mark) continue;
  try:
    best_score = 0;
    best = ListCreate(CoverFree);
#if AQ
    /* find the best cover for the example */
    BestCovers(lh, class, ex, &best_score, best);
#else
    /* find the best cover overall */
    {ListIter(p2, ex2, lh->pos_ex[class]) {
      if(ex2->mark) continue;
      BestCovers(lh, class, ex2, &best_score, best);
    }}
#endif
    /* was this example disallowed? */
    if(best_score == 0) {
#if DEBUG
      fprintf(stderr, "class %d was disallowed\n", class);
#endif
      ListFree(best);
      continue;
    }

#if RANDOM
    i = RandInt(0, ListSize(best)-1);
#else
    i = 0;
#endif
    cover = AllocCopy(1, struct Cover, (Cover)ListValueAtIndex(best, i));
    ListFree(best);
    cover->high = HierRightLeaf(lh->trees[cover->tree], cover->node);
#if DEBUG
    fprintf(stderr, "cover: %d %d-%d %g\n", 
	    cover->tree, cover->node, cover->high, best_score);
#endif
    ListAddFront(Covers(class), cover);

    /* mark covered examples */
    td = &lh->td[cover->tree];
    for(i=cover->node;i<=cover->high;i++) {
      if(td->ex[class][i]) {
	ListIter(q, ex2, td->ex[class][i]) {
#if DEBUG
	  fprintf(stderr, "marking %d\n", i);
#endif
	  ex2->mark = 1;
	}
      }
    }
    if(!ex->mark) goto try;
  }}
}

/* Returns a List of the hier data for the items of the given class
 * matching item, using the supplied comparison function.
 * If compare == NULL, all items in the class are returned.
 * The result may contain duplicates.
 */
List LearnWithClassMatching(LearnHandle lh, int class, 
			    void *item, CmpFunc *compare)
{
  List result = ListCreate(NULL);
  Cover cover;
  int index;
  HierNode hierNode;

  if(class >= lh->nclasses) return result;
  LearnUpdate(lh);
  /* Loop covers */
  {ListIter(p, cover, Covers(class)) {
    /* Loops the leaves of the cover's tree */
    for(index = cover->node; index <= cover->high; index++) {
      hierNode = HierNodeAt(lh->trees[cover->tree], index);
      if(hierNode->children) continue;
      if(compare && compare(&hierNode->data, item)) continue;
      ListAddFront(result, &hierNode->data);
    }
  }}
  return result;
}

/* Returns the class membership weight, given per-tree leaf indices.
 */
float LearnClassWeight(LearnHandle lh, int class, List leaves)
{
  Cover cover;
  LearnUpdate(lh);
  /* Loop the covers */
  {ListIter(p, cover, Covers(class)) {
    Leaf leaf;
    ListIter(p2, leaf, leaves) {
      if((leaf->tree == cover->tree) &&
	 (leaf->node >= cover->node) &&
	 (leaf->node <= cover->high)) return 1.0;
    }
  }}
  return 0.0;
}

int LearnReadBias(LearnHandle lh, int tree, char *fname)
{
  TreeData *td = &lh->td[tree];
  int i, j, n, len;
  FILE *fp = fopen(fname, "r");
  if(!fp) return TPM_ERROR;
  if(fscanf(fp, "%d (struct int (%d int))", &n, &len) < 2) return TPM_ERROR;
  getc(fp); /* absorb the newline; not safe if done by fscanf */
  fprintf(stderr, "%s has %d biases of %d nodes\n", fname, n, len);
  i = td->nbias;
  td->nbias += n;
  Reallocate(td->bptotal, td->nbias, int);
  Reallocate(td->bpos, td->nbias, int*);
  Reallocate(td->bntotal, td->nbias, int);
  Reallocate(td->bneg, td->nbias, int*);
  for(;n--;i++) {
    ReadAdjust(&td->bptotal[i], TPM_INT, 1, fp);
    td->bpos[i] = Allocate(lh->trees[tree]->num_nodes, int);
    ReadAdjust(td->bpos[i], TPM_INT, lh->trees[tree]->num_nodes, fp);
    ReadAdjust(&td->bntotal[i], TPM_INT, 1, fp);
    td->bneg[i] = Allocate(lh->trees[tree]->num_nodes, int);
    ReadAdjust(td->bneg[i], TPM_INT, lh->trees[tree]->num_nodes, fp);
  }
  fclose(fp);
  return TPM_OK;
}

void LearnSaveBias(LearnHandle lh, int tree, char *fname)
{
  TreeData *td = &lh->td[tree];
  int i,j,b;
  int new;
  FILE *fp = fopen(fname, "w");
  if(!fp) { perror("LearnSaveBias"); return; }

  /* update all biases used by some class */
  new = 0;
  for(i=0;i<lh->nclasses;i++) {
    b = td->cbias[i];
    if(!b) {
      /* don't change default bias; create new */
      new++;
      continue; 
    }
    /* assumes every tree sees every example */
    td->bptotal[b] += TotalPos(i);
    td->bntotal[b] += TotalNeg(i);
    for(j=0;j<lh->trees[tree]->num_nodes;j++) {
      td->bpos[b][j] += td->npos[i][j];
      td->bneg[b][j] += td->nneg[i][j];
    }
  }

  fprintf(fp, "%d (struct int (%d int))\n", 
	  td->nbias-1+new, 2*lh->trees[tree]->num_nodes);
  /* write biases in order */
  for(b=1;b<td->nbias;b++) {
    WriteAdjust(&td->bptotal[b], TPM_INT, 1, fp);
    WriteAdjust(td->bpos[b], TPM_INT, lh->trees[tree]->num_nodes, fp);
    WriteAdjust(&td->bntotal[b], TPM_INT, 1, fp);
    WriteAdjust(td->bneg[b], TPM_INT, lh->trees[tree]->num_nodes, fp);
  }
  /* append new biases */
  for(i=0;i<lh->nclasses;i++) {
    if(td->cbias[i]) continue;
    j = TotalPos(i);
    WriteAdjust(&j, TPM_INT, 1, fp);
    WriteAdjust(td->npos[i], TPM_INT, lh->trees[tree]->num_nodes, fp);
    j = TotalNeg(i);
    WriteAdjust(&j, TPM_INT, 1, fp);
    WriteAdjust(td->nneg[i], TPM_INT, lh->trees[tree]->num_nodes, fp);
  }

  fclose(fp);
}

#if 0
    /* now cover all nodes in the same tree whose bias is at least as large
       as that cover */
    td = &lh->td[cover->tree];
    if(!td->[td->cbias[class]]) continue;
    best_score = NodeBias(lh, cover->tree, td->cbias[class], cover->node);
    if(best_score == 1) continue;
    for(i=0;i<lh->trees[cover->tree]->num_nodes;i++) {
      if(td->nneg[class][i]) continue;
      if((i >= cover->node) && (i <= cover->high)) continue;
      score = NodeBias(lh, cover->tree, td->cbias[class], i);
      if(score >= best_score) {
	Cover newcover = Allocate(1, struct Cover);
	newcover->tree = cover->tree;
	newcover->node = i;
	newcover->high = HierRightLeaf(lh->trees[newcover->tree], 
				       newcover->node);
	ListAddFront(Covers(class), newcover);
	LearnMarkCovered(lh, class, newcover);
	/* skip all children of this cover */
	i = newcover->high;
#if 0
	fprintf(stderr, "newcover: %d-%d %g\n", 
		newcover->node, newcover->high, score);
#endif
      }
    }
#endif
