/* Learning benchmark.
 * bench partition hier hier ...
 * Tries to learn the partition using concepts over the given hierarchies.
 * At each iteration, outputs the number of errors, chooses one at random,
 * and corrects the learner.
 */
#include "learn.h"
#include <tpm/partition.h>
#include "stats.h"

#define FIXEDBIAS 0
#define INORDER 0
#define DEBUG 0
#define REPEAT 1
/* show tree weights */
#define TW 0

/* taken from aq.c */
#define Covers(class) (List)lh->classData[class]

extern float *TreeFreq(LearnHandle lh, int class);
extern void LearnTreeScore(LearnHandle lh, int class);

void LearnPartition(LearnHandle lh, Partition ideal, FILE *fp);

void main(int argc, char *argv[])
{
  LearnHandle lh;
  int i,j;
  int arg;
  Partition ideal;
  FILE *fp;
  char str[128], *outfile = "bench.out";
  int just_tree = 0;

  MEM_ABORT_ON_ERROR(2);
  arg = 1;
  if(argc > 1) {
    if(!strcmp(argv[arg], "-o")) {
      outfile = argv[arg+1];
      arg+=2;
    }
    if(!strcmp(argv[arg], "-t")) {
      just_tree = 1;
      arg++;
    }
  }
  if((argc-arg) < 2) {
    fprintf(stderr, "usage: bench [-o output] [-t] partition hier hier ...\n");
    exit(1);
  }

  ideal = PartitionCreateFromFile(argv[arg]);
  if(!ideal) {
    fprintf(stderr, "Error reading `%s'\n", argv[arg]);
    exit(1);
  }
  arg++;
  fprintf(stderr, "%d elements, %d classes\n", 
	  ideal->num_elements, ideal->num_classes);

  lh = LearnInit();

  /* read the trees */
  for(i=arg;i < argc;i++) {
    char *fname = argv[i];
    Hier tree = HierCreateFromFile(fname);
    if(!tree) {
      fprintf(stderr, "Error reading `%s'\n", fname);
      exit(1);
    }
    LearnAddTree(lh, tree);
    sprintf(str, "%s.bias", fname);
    /* read bias */
    if(LearnReadBias(lh, i-2, str)) fprintf(stderr, "no bias\n");
    else {
      fprintf(stderr, "using %s\n", str);
    }
  }

  /* define classes */
  while(lh->nclasses < ideal->num_classes) LearnAddClass(lh);
#if FIXEDBIAS
  for(i=0;i<lh->ntrees;i++) {
    TreeData *td = &lh->td[i];
    for(j=0;j<lh->nclasses;j++) {
      if(j+1 >= td->nbias) break;
      td->cbias[j] = j+1;
    }
  }
#endif

  if(just_tree) {
    /* give the answer */
    for(i=0;i<ideal->num_elements;i++) {
      if(ideal->class[i])
	LearnAddEx(lh, ideal->class[i]-1, 1, &i, IntCmp);
      else {
	/* negative for all classes */
	for(j=0;j<ideal->num_classes;j++)
	  LearnAddEx(lh, j, -1, &i, IntCmp);
      }
    }
  }
  else {
    /* learn the partition */
    fp = fopen(outfile, "w");
    LearnPartition(lh, ideal, fp);
    fclose(fp);
    LearnCommit(lh);
  }

  /* save the biases */
#if !FIXEDBIAS
  LearnUpdateBias(lh);
#endif
#if 0
  for(i=0;i<lh->ntrees;i++) {
    sprintf(str, "%s.bias", argv[i+2]);
    LearnSaveBias(lh, i, str);
  }
#endif

#if 0
  /* show the concept */
  for(j=0;j<ideal->num_classes;j++)
    LearnClassDataPrint(lh, j, stderr);
#endif

  /* show tree usage */
  if(lh->ntrees > 1) {
    printf("tree usage:\n");
    for(i=0;i<lh->nclasses;i++) {
      float *freq = TreeFreq(lh, i);
      printf("  class %d: ", i);
      for(j=0;j<lh->ntrees;j++) printf("%g ", freq[j]);
      printf("\n");
      free(freq);
    }
  }

  /* save the trees */
  for(i=0;i<lh->ntrees;i++) {
    if(strstr(argv[i+arg], ".new"))
      strcpy(str, argv[i+arg]);
    else 
      sprintf(str, "%s.new", argv[i+arg]);
    LearnSaveTree(lh, i, str);
  }

  LearnFreeTrees(lh);
  LearnFree(lh);
  PartitionFree(ideal);

  MEM_BLOCKS();
  MEM_STATUS();
}

/* Learns ideal by random selection, sending # of errors after every
 * example to fp.
 */
void LearnPartition(LearnHandle lh, Partition ideal, FILE *fp)
{
  int nex, next_update, rep;
  int c, i,j;
  int num_errs, err_sum;
  Stats err_stats, sum_stats;
  double err_std, sum_std;
  char *errors;

  srand(getpid());
  errors = Allocate(ideal->num_elements, char);
  next_update = 1;
  for(nex=0;;nex++) {
    err_stats = StatsCreate();
    sum_stats = StatsCreate();
    for(rep=0;rep<REPEAT;rep++) {
      /* force all classes to update */
      LearnEnableTree(lh, 0);

      /* compute error array */
      for(i=0;i<ideal->num_elements;i++) errors[i] = !!ideal->class[i];
      for(c=0;c<ideal->num_classes;c++) {
	List list = LearnWithClass(lh, c);
	/* have to be careful here, since list may contain duplicates */
	{int *m; ListIter(p, m, list) {
	  /* LearnWithClass numbers from zero, but Partition numbers from 1 */
	  if(ideal->class[*m] == c+1) errors[*m] = 0;
	  /* incorrect class counts twice */
	  else errors[*m] = 2;
	}}
	ListFree(list);
      }

      /* compute error count */
      num_errs = err_sum = 0;
      for(i=0;i<ideal->num_elements;i++) {
	if(errors[i]) { num_errs++; err_sum += errors[i]; }
      }

      /* update stats */
      StatsAddSample(err_stats, num_errs);
      StatsAddSample(sum_stats, err_sum);
    }
    if(!num_errs) break;
#if REPEAT == 1
    printf("%d %d\n", err_sum, num_errs);
    fprintf(fp, "%d\n", err_sum);
#else
    err_std = StatsStd(err_stats);
    sum_std = StatsStd(sum_stats);
    printf("%d (%g %g %g-%g) %d (%g %g %g-%g)\n", 
	   err_sum, sum_stats->mean, sum_std, sum_stats->min, sum_stats->max,
	   num_errs, err_stats->mean, err_std, err_stats->min, err_stats->max);
    fprintf(fp, "%g %g\n", sum_stats->mean, sum_std);
#endif
#if 1
    fflush(fp);
#endif
    StatsFree(err_stats);
    StatsFree(sum_stats);

#if TW
    if(lh->ntrees > 1) {
      for(c=0;c<ideal->num_classes;c++) {
	float *freq = TreeFreq(lh, c);
	printf("class %d\n", c);
	printf("scores: ");
	LearnTreeScore(lh, c);
	for(j=0;j<lh->ntrees;j++) printf("%g ", lh->td[j].weight);
	printf("\nfreq:   ");
	for(j=0;j<lh->ntrees;j++) printf("%g ", freq[j]);
	printf("\n");
	free(freq);
      }
    }
#endif

#if INORDER
    /* fix classes in order */
    for(c=0;c<ideal->num_classes;c++) {
      PartitionIter(ideal, c+1, e) {
	if(errors[e]) { i=e; goto done; }
      }
    }
  done:
#else
    /* make a random correction */
    j = RandInt(1, num_errs);
    for(i=0;i<ideal->num_elements;i++) if(errors[i] && !--j) break;
#endif
#if DEBUG
    printf("setting %d to class %d\n", i, ideal->class[i]);
#endif
    if(ideal->class[i])
      LearnAddEx(lh, ideal->class[i]-1, 1, &i, IntCmp);
    else {
      /* negative for all classes */
      for(c=0;c<ideal->num_classes;c++)
	LearnAddEx(lh, c, -1, &i, IntCmp);
    }

#if !FIXEDBIAS
    /* update biases */
    if(nex == next_update) {
      LearnUpdateBias(lh);
#if 1
      next_update *= 2;
#else
      next_update += 10;
#endif
    }
#endif
  }
  StatsFree(err_stats);
  StatsFree(sum_stats);
  free(errors);
}
