/* This file implements two-sided recursive learning.
 * Class membership is determined by the leaves of a decision tree which 
 * branches on membership in atomic groupings.
 */

#include "learn.h"
#include "common.h"
#include <tpm/gtree.h>
#include <math.h>   /* for log() */

#define DEBUG 1
/* pick the first good node (otherwise last) */
#define FIRST 1
/* pick equally-good nodes at random (overrides FIRST) */
#define RANDOM 1
/* this forces a decision list instead of tree. */
#define LIST 0
/* use the Gini criterion instead of entropy. */
#define GINI 1
/* use a size bias */
#define SZB 0
/* size parameter */
#define SZ 1
/* use tree weights */
#define TW 0

#define ClassTree(class) (Tree)lh->classData[class]

/* Prototypes ****************************************************************/
void LearnUpdateClass(LearnHandle lh, int class);
List LearnWithClassMatching(LearnHandle lh, int class, 
			    void *item, CmpFunc *compare);
float LearnClassWeight(LearnHandle lh, int class, List leaves);

int LearnReadBias(LearnHandle lh, int tree, char *fname);
void LearnSaveBias(LearnHandle lh, int tree, char *fname);
void LearnUpdateBias(LearnHandle lh);

void *LearnClassDataInit(int class);
void LearnClassDataFree(void *covers);
void LearnClassDataPrint(LearnHandle lh, int class, FILE *fp);

float *TreeFreq(LearnHandle lh, int class);

/* private */
static Tree BuildTree(LearnHandle lh, int class);

/* Functions *****************************************************************/

static void TreeCount(Tree tree, float *freq)
{
  Cover cover = (Cover)tree->data;
  if(TreeLeaf(tree)) return;
  freq[cover->tree]++;
  TreeCount(tree->child, freq);
  TreeCount(tree->child->sibling, freq);
}

float *TreeFreq(LearnHandle lh, int class)
{
  float *freq = AllocateC(lh->ntrees, float);
  int i, total = TreeCountNonLeaves(ClassTree(class));
  TreeCount(ClassTree(class), freq);
  for(i=0;i<lh->ntrees;i++) freq[i] /= total;
  return freq;
}

#define IsInside(c) (c->node == 0)
static Cover AllInside(void)
{
  Cover cover = Allocate(1, struct Cover);
  cover->node = 0;
  return cover;
}
#define IsOutside(c) (c->node == -1)
static Cover AllOutside(void)
{
  Cover cover = Allocate(1, struct Cover);
  cover->node = -1;
  return cover;
}

static void CoverWrite(FILE *fp, Cover cover)
{
  if(!cover->node) fprintf(fp, "0 0");
  else if(cover->node == -1) fprintf(fp, "0 -1");
  else fprintf(fp, "%d %d-%d", cover->tree, cover->node, cover->high);
}

void LearnClassDataPrint(LearnHandle lh, int class, FILE *fp)
{
  fprintf(fp, "tree for class %d:\n", class);
  TreeWrite(fp, ClassTree(class), (WriteFunc*)CoverWrite);
}

void *LearnClassDataInit(int class)
{
  return TreeCreate(AllOutside());
}
void LearnClassDataFree(void *tree)
{
  TreeFree((Tree)tree, CoverFree);
}

/* Calls BuildTree with the given example set.
 * examples is freed.
 */
static Tree BuildSubTree(LearnHandle lh, int class, List examples)
{
  int tree, i;
  Tree t;
  LearnHandle save;

  save = LearnSaveState(lh, class);
  LearnClearState(lh, class);

  /* create a new example list */
  {Example ex;ListIter(p, ex, examples) {
    AddEx(lh, class, ex->leaves, ex->weight);
  }}
  ListFree(examples);

  /* build a tree */
  t = BuildTree(lh, class);

  LearnRestoreState(lh, class, save);
  return t;
}

/* Returns a Tree of Covers. The left child of a node is inside.
 * Leaf data is 1 for in-class, 0 for out-class.
 */
static Tree BuildTree(LearnHandle lh, int class)
{
  Cover cover;
  int tree, node;
  float best_score;
  Tree t;
  List inside, outside;
  TreeData *td;
  int i;
  Example ex;
  List best;

  /* pure set of examples? */
  if(TotalNeg(class) == 0) {
    return TreeCreate(AllInside());
  }
  if(TotalPos(class) == 0) {
    return TreeCreate(AllOutside());
  }

  /* find best split */
  best_score = 0;
  best = ListCreate(CoverFree);
  for(tree=0;tree<lh->ntrees;tree++) {
    td = &lh->td[tree];
    for(node=0;node<lh->trees[tree]->num_nodes;node++) {
      int in_p = td->npos[class][node], in_n = td->nneg[class][node];
      int out_p = TotalPos(class) - in_p, out_n = TotalNeg(class) - in_n;
      int in_size = in_p + in_n, out_size = out_p + out_n;
      float score, v;
      /* if one of the bins is empty, we won't make progress */
      if(!in_size || !out_size) continue;
#if LIST
      /* skip if one of the bins is impure */
      if(in_p && in_n && out_p && out_n) continue;
#endif
      score = 0.0;
      if(in_p && in_n) {
	/* inside entropy */
	v = (float)in_p / in_size;
	/* both of these range from 0 (best) to 1 (worst) */
#if GINI
	/* Dietterich, Kearns, and Mansour recommend a sqrt here */
	score += v * (1-v) * in_size * 4;
#else
	score += (v * log(v) + (1-v) * log(1-v)) * in_size 
	  / 0.69314718055994530942; /* log(2) */
#endif
      }
      if(out_p && out_n) {
	/* outside entropy */
	v = (float)out_p / out_size;
#if GINI
	score += v * (1-v) * out_size * 4;
#else
	score += (v * log(v) + (1-v) * log(1-v)) * out_size 
	  / 0.69314718055994530942; /* log(2) */
#endif
      }
      /* now score ranges from 0 (worst) to 1 (best) */
      score = 1 - score / (in_size + out_size);
#if SZB
      v = HierRightLeaf(lh->trees[tree], node) - node + 1;
      v	/= lh->trees[tree]->num_nodes;
      /* change the range of v to [0.1,0.9] instead of [0,1] */
      v = v*0.8 + 0.1;
      score *= v * pow(1-v, SZ);
#endif
#if TW
      score *= td->weight;
#endif
#if DEBUG
      fprintf(stderr, "cover: %d %d %g\n", 
	      tree, node, score);
#endif
      if(score > best_score) {
	best_score = score;
	ListFlush(best);
      }
      if(score == best_score) {
	Cover c = Allocate(1, struct Cover);
	c->tree = tree;
	c->node = node;
	ListAddFront(best, c);
      }
    }
  }
#if RANDOM
  i = RandInt(0, ListSize(best)-1);
#elif FIRST
  i = ListSize(best)-1;
#else
  i = 0;
#endif
  cover = AllocCopy(1, struct Cover, (Cover)ListValueAtIndex(best, i));
  ListFree(best);
  cover->high = HierRightLeaf(lh->trees[cover->tree], cover->node);
#if DEBUG
  fprintf(stderr, "best cover: %d %d-%d %g\n", 
	  cover->tree, cover->node, cover->high, best_score);
#endif
  t = TreeCreate(cover);

  /* divide examples according to membership in the cover */
  inside = ListCreate(NULL);
  outside = ListCreate(NULL);
  td = &lh->td[cover->tree];
  for(i=cover->node;i<=cover->high;i++) {
    if(td->ex[class][i]) {
      ListIter(q, ex, td->ex[class][i]) {
#if DEBUG
	fprintf(stderr, "inside: %d %g\n", i, ex->weight);
#endif
	ListAddFront(inside, ex);
	ex->mark = 1;
      }
    }
  }
  {ListIter(p, ex, lh->examples[class]) {
    if(!ex->mark) {
#if DEBUG
      fprintf(stderr, "outside: %d %g\n", 
	      ((Leaf)ListFront(ex->leaves)->data)->node, ex->weight);
#endif
      ListAddFront(outside, ex);
    }
    ex->mark = 0;
  }}
  /* now all examples are unmarked again */

  /* recurse on interior */
  t->child = BuildSubTree(lh, class, inside);

  /* recurse on exterior */
  t->child->sibling = BuildSubTree(lh, class, outside);

  return t;
}

void LearnUpdateClass(LearnHandle lh, int class)
{
  FILE *fp;

  /* flush the old tree */
  LearnClassDataFree(lh->classData[class]);
  lh->classData[class] = BuildTree(lh, class);
#if DEBUG
  fp = fopen("qq.tree", "w");
  TreeWrite(fp, ClassTree(class), (WriteFunc*)CoverWrite);
  fclose(fp);
#endif
}

/* Returns the membership weight in a decision tree, 
   given per-tree leaf indices. */
static float Search(Tree t, List leaves)
{
  Cover cover = (Cover)t->data;
  if(TreeLeaf(t)) {
    if(IsInside(cover)) return 1.0;
    else return 0.0;
  }
  {Leaf leaf; ListIter(p, leaf, leaves) {
    if((leaf->tree == cover->tree) &&
       (leaf->node >= cover->node) &&
       (leaf->node <= cover->high)) {
      /* inside */
      return Search(t->child, leaves);
    }
  }}
  /* outside */
  return Search(t->child->sibling, leaves);
}

/* Returns the class membership weight, given per-tree leaf indices. */
float LearnClassWeight(LearnHandle lh, int class, List leaves)
{
  return Search(ClassTree(class), leaves);
}

List LearnWithClassMatching(LearnHandle lh, int class, 
			    void *item, CmpFunc *compare)
{
  return LearnWithClassMatchingGeneric(lh, class, item, compare);
}

int LearnReadBias(LearnHandle lh, int tree, char *fname)
{
  return TPM_ERROR;
}

void LearnSaveBias(LearnHandle lh, int tree, char *fname)
{
}

void LearnUpdateBias(LearnHandle lh)
{
}
