/* This file implements two-sided recursive learning.
 * Class membership is determined by the leaves of a decision tree which 
 * branches on membership in atomic groupings.
 */

#include "learn.h"
#include "common.h"
#include <tpm/gtree.h>
#include <math.h>   /* for log() */
#include "isa.h"

#define DEBUG 1

/* Prototypes ****************************************************************/
void LearnUpdateClass(LearnHandle lh, int class);
List LearnWithClassMatching(LearnHandle lh, int class, 
			    void *item, CmpFunc *compare);
float LearnClassWeight(LearnHandle lh, int class, List leaves);

int LearnReadBias(LearnHandle lh, int tree, char *fname);
void LearnSaveBias(LearnHandle lh, int tree, char *fname);
void LearnUpdateBias(LearnHandle lh);

void *LearnClassDataInit(int class);
void LearnClassDataFree(void *covers);

/* Functions *****************************************************************/

#define IsInside(c) (c->node == 0)
static Cover AllInside(void)
{
  Cover cover = Allocate(1, struct Cover);
  cover->node = 0;
  return cover;
}
#define IsOutside(c) (c->node == -1)
static Cover AllOutside(void)
{
  Cover cover = Allocate(1, struct Cover);
  cover->node = -1;
  return cover;
}

void *LearnClassDataInit(int class)
{
  return TreeCreate(AllOutside());
}
void LearnClassDataFree(void *tree)
{
  TreeFree((Tree)tree, CoverFree);
}

/* Returns a Tree of Covers. The left child of a node is inside.
 * Leaf data is 1 for in-class, 0 for out-class.
 */
static Tree BuildTree(LearnHandle lh, int class, 
		      IntStackArray *pa, IntStackArray *na)
{
  Cover cover;
  int tree, node;
  float best_score;
  int best_size;
  Tree t, t_outside, t_inside;
  List inside, outside, inside_pos, outside_pos;
  List ex_save, pos_save;
  TreeData *td;
  int i;
  Example ex;
  Leaf leaf;

  /* pure set of examples? */
  if(TotalNeg(class) == 0) {
    return TreeCreate(AllInside());
  }
  if(TotalPos(class) == 0) {
    return TreeCreate(AllOutside());
  }

  /* find node with least entropy */
  cover = Allocate(1, struct Cover);
  best_score = 1e10;
  /* loop examples */
  {ListIter(p, ex, lh->examples[class]) {
    ListIter(q, leaf, ex->leaves) {
      td = &lh->td[leaf->tree];
      for(node = leaf->node; node >= 0; 
	  node = HierNodeAt(lh->trees[leaf->tree], node)->parent) {
	int p = td->npos[class][node], n = td->nneg[class][node];
	int size = p + n;
	float score, v;
	/* if one of the bins is empty, we can't make progress */
	if(!size || (size == TotalEx(class))) continue;
	score = 0.0;
	if(p && n) {
	  /* inside entropy */
	  v = (float)p / size;
	  score -= (v * (float)log((double)v) +
		    (1-v) * (float)log((double)(1-v))) * size;
	  if(score > 1) score /= 1.5;
	}
	if((p != TotalPos(class)) && (n != TotalNeg(class))) {
	  /* outside entropy */
	  v = (float)(TotalPos(class) - p) / (TotalEx(class) - size);
	  score -= (v * (float)log((double)v) +
		    (1-v) * (float)log((double)(1-v))) 
	    * (TotalEx(class) - size);
	}
#if 0
	fprintf(stderr, "cover: %d %d %g\n", 
		leaf->tree, node, score);
#endif
	if((score < best_score) ||
	   ((score == best_score) && 
	    ((size > best_size) || (node < cover->node)))) {
	  best_score = score;
	  best_size = size;
	  cover->tree = leaf->tree;
	  cover->node = node;
	}
      }
    }
  }}
  cover->high = HierRightLeaf(lh->trees[cover->tree], cover->node);
#if DEBUG
  fprintf(stderr, "best cover: %d %d-%d %g\n", 
	  cover->tree, cover->node, cover->high, best_score);
#endif
  t = TreeCreate(cover);

  /* divide examples according to membership in the cover */
  inside = ListCreate(NULL);
  inside_pos = ListCreate(NULL);
  outside = ListCreate(NULL);
  outside_pos = ListCreate(NULL);
  td = &lh->td[cover->tree];
  for(i=cover->node;i<=cover->high;i++) {
    if(td->ex[class][i]) {
      ListIter(q, ex, td->ex[class][i]) {
	if(ListIndexAtValue(lh->examples[class], 
 == -1) continue;
#if DEBUG
	fprintf(stderr, "inside: %d %g\n", i, ex->weight);
#endif
	ListAddFront(inside, ex);
	if(ex->weight > 0) ListAddFront(inside_pos, ex);
	ex->mark = 1;
      }
    }
  }
  {ListIter(p, ex, lh->examples[class]) {
    if(!ex->mark) {
#if DEBUG
      fprintf(stderr, "outside: %d %g\n", 
	      ((Leaf)ListFront(ex->leaves)->data)->node, ex->weight);
#endif
      ListAddFront(outside, ex);
      if(ex->weight > 0) ListAddFront(outside_pos, ex);
    }
    ex->mark = 0;
  }}
  /* now all examples are unmarked again */

  ex_save = lh->examples[class];
  pos_save = lh->pos_ex[class];
  /* remove */
  {ListIter(p, ex, inside) {
    ListIter(q, leaf, ex->leaves) {
      td = &lh->td[leaf->tree];
      for(node = leaf->node; node >= 0; 
	  node = HierNodeAt(lh->trees[leaf->tree], node)->parent) {
	IntStackArrayPush(pa[leaf->tree], node);
	IntStackArrayPush(na[leaf->tree], node);
	if(ex->weight > 0) IntStackArrayElt(pa[leaf->tree], node)--;
	else IntStackArrayElt(na[leaf->tree], node)--;
      }
    }
  }}
  lh->examples[class] = outside;
  lh->pos_ex[class] = outside_pos;
  t_outside = BuildTree(lh, class, pa, na);
  ListFree(outside);
  ListFree(outside_pos);

  /* zero */
  {ListIter(p, ex, inside) {
    ListIter(q, leaf, ex->leaves) {
      td = &lh->td[leaf->tree];
      for(node = leaf->node; node >= 0; 
	  node = HierNodeAt(lh->trees[leaf->tree], node)->parent) {
	IntStackArrayElt(pa[leaf->tree], node) = 0;
	IntStackArrayElt(na[leaf->tree], node) = 0;
      }
    }
  }}
  /* fill */
  {ListIter(p, ex, inside) {
    ListIter(q, leaf, ex->leaves) {
      td = &lh->td[leaf->tree];
      for(node = leaf->node; node >= 0; 
	  node = HierNodeAt(lh->trees[leaf->tree], node)->parent) {
	if(ex->weight > 0) IntStackArrayElt(pa[leaf->tree], node)++;
	else IntStackArrayElt(na[leaf->tree], node)++;
      }
    }
  }}
  lh->examples[class] = inside;
  lh->pos_ex[class] = inside_pos;
  t_inside = BuildTree(lh, class, pa, na);
  ListFree(inside);
  ListFree(inside_pos);

  /* restore */
  {ListIter(p, ex, inside) {
    ListIter(q, leaf, ex->leaves) {
      td = &lh->td[leaf->tree];
      for(node = leaf->node; node >= 0; 
	  node = HierNodeAt(lh->trees[leaf->tree], node)->parent) {
	IntStackArrayPop(pa[leaf->tree], node);
	IntStackArrayPop(na[leaf->tree], node);
      }
    }
  }}
  lh->examples[class] = ex_save;
  lh->pos_ex[class] = pos_save;

  t->child = t_inside;  t->child->sibling = t_outside;
  return t;
}

static void CoverWrite(FILE *fp, Cover cover)
{
#if 0
  fprintf(fp, "%d %d", cover->tree, cover->node);
#else
  fprintf(fp, "%d", cover->node);
#endif
}
void LearnUpdateClass(LearnHandle lh, int class)
{
  FILE *fp;
  int tree;
  IntStackArray *pa, *na;

  /* wrap stackArrays around npos, nneg */
  pa = Allocate(lh->ntrees, IntStackArray);
  na = Allocate(lh->ntrees, IntStackArray);
  for(tree=0;tree<lh->ntrees;tree++) {
    pa[tree] = IntStackArrayFromData(lh->td[tree].npos[class], 
				     lh->trees[tree]->num_nodes);
    na[tree] = IntStackArrayFromData(lh->td[tree].nneg[class], 
				     lh->trees[tree]->num_nodes);
  }

  /* flush the old tree */
  LearnClassDataFree(lh->classData[class]);
  lh->classData[class] = BuildTree(lh, class, pa, na);

  /* free stackArrays */
  for(tree=0;tree<lh->ntrees;tree++) {
    IntStackArrayAlmostFree(pa[tree]);
    IntStackArrayAlmostFree(na[tree]);
  }
  free(pa);
  free(na);

#if DEBUG
  fp = fopen("qq.tree", "w");
  TreeWrite(fp, (Tree)lh->classData[class], (WriteFunc*)CoverWrite);
  fclose(fp);
#endif
}

/* Returns the membership weight in a decision tree, 
   given per-tree leaf indices. */
static float Search(Tree t, List leaves)
{
  Cover cover = (Cover)t->data;
  if(TreeLeaf(t)) {
    if(IsInside(cover)) return 1.0;
    else return 0.0;
  }
  {Leaf leaf; ListIter(p, leaf, leaves) {
    if((leaf->tree == cover->tree) &&
       (leaf->node >= cover->node) &&
       (leaf->node <= cover->high)) {
      /* inside */
      return Search(t->child, leaves);
    }
  }}
  /* outside */
  return Search(t->child->sibling, leaves);
}

/* Returns the class membership weight, given per-tree leaf indices. */
float LearnClassWeight(LearnHandle lh, int class, List leaves)
{
  return Search((Tree)lh->classData[class], leaves);
}

/* Returns a List of the hier data for the items of the given class
 * matching item, using the supplied comparison function.
 * If compare == NULL, all items in the class are returned.
 * The result may contain duplicates.
 */
List LearnWithClassMatching(LearnHandle lh, int class, 
			    void *item, CmpFunc *compare)
{
  int tree, node;
  List result = ListCreate(NULL);
  List leaves = ListCreate(NULL);
  struct Leaf leaf;

  if(class >= lh->nclasses) return result;
  LearnUpdate(lh);
  ListAddFront(leaves, &leaf);
  /* loop all leaves in all trees, adding them to result if they are
     included in lh->covers[class] */
  for(tree=0;tree<lh->ntrees;tree++) {
    leaf.tree = tree;
    for(node=0;node<lh->trees[tree]->num_nodes;node++) {
      HierNode hierNode = HierNodeAt(lh->trees[tree], node);
      /* is it a leaf? */
      if(hierNode->children) continue;
      /* does it match item? */
      if(compare && compare(&hierNode->data, item)) continue;
      leaf.node = node;
      if(LearnClassWeight(lh, class, leaves) > 0.0) 
	ListAddFront(result, &hierNode->data);
    }
  }
  ListFree(leaves);
  return result;
}

int LearnReadBias(LearnHandle lh, int tree, char *fname)
{
  return TPM_ERROR;
}

void LearnSaveBias(LearnHandle lh, int tree, char *fname)
{
}

void LearnUpdateBias(LearnHandle lh)
{
}
