/* Learning module.
 * Given a set of hierarchies from which to draw atomic groupings,
 * partitions the instances according to a set of positive and negative 
 * examples for a set of classes.
 *
 * LearnInit() - returns a LearnHandle for internal bookkeeping.
 * LearnFreeTrees(lh) - frees trees used by lh.
 *                      must be called before LearnFree.
 * LearnFree(lh) - frees storage held by lh, except for loaded trees.
 * LearnClear(lh) - erases memory of examples (but not trees).
 * LearnAddTree(lh, tree) - adds a hierarchy for lh to use.
 *                Each node in the hierarchy should have satellite data for
 *                identification with an instance.
 * LearnAddClass(lh) - adds a new class entry to lh, 
 *                     returning the index for that class.
 * LearnAddEx(lh, class, weight, item, compare) - 
 *           adds an example for the given class.
 *           The sign of weight indicates positive or negative;
 *           the magnitude indicates strength or confidence.
 *           All nodes whose data matches item, according to the compare fcn,
 *           will be made into examples.
 * LearnWithClass(lh, class) - returns a list of node data, for nodes which 
 *               have been assigned to the class by the learning algorithm.
 * LearnDisableTree(lh, tree) - temporarily remove tree's groupings from 
 *                              consideration.
 * LearnEnableTree(lh, tree) - undo LearnDisableTree.
 */

#include <tpm/gtree.h> /* for LearnSaveTree */
#include "learn.h"
#include "common.h"
#include <tpm/hash.h>

#define NET 0

/* Globals *******************************************************************/

/* Prototypes ****************************************************************/

LearnHandle LearnInit(void);
void LearnFreeTrees(LearnHandle lh);
void LearnFree(LearnHandle lh);
void LearnClear(LearnHandle lh);
void LearnAddTree(LearnHandle lh, Hier tree);
int LearnAddClass(LearnHandle lh);
void LearnEnableTree(LearnHandle lh, int tree);
void LearnDisableTree(LearnHandle lh, int tree);
void LearnAddEx(LearnHandle lh, int class, float weight, 
		void *item, CmpFunc *compare);
void LearnUpdate(LearnHandle lh);
List LearnWithClass(LearnHandle lh, int class);
void LearnCommit(LearnHandle lh);
void LearnSaveTree(LearnHandle lh, int tree, char *fname);
void LearnTreeScore(LearnHandle lh, int class);

void LearnExampleFree(Example ex);
void AddEx(LearnHandle lh, int class, List leaves, float weight);
LearnHandle LearnSaveState(LearnHandle lh, int class);
void LearnClearState(LearnHandle lh, int class);
void LearnRestoreState(LearnHandle lh, int class, LearnHandle save);

/* Functions *****************************************************************/

/* Create a new LearnHandle */
LearnHandle LearnInit(void)
{
  LearnHandle lh = Allocate(1, struct LearnHandle);

  /* Tree info */
  lh->ntrees = 0;
  lh->trees = Allocate(1, Hier);
  lh->td = Allocate(1, TreeData);

  /* Class info */
  lh->nclasses = 0;
  lh->updated = Allocate(1, int);
  lh->classData = Allocate(1, void*);
  
  /* Example info */
  lh->examples = Allocate(1, List);
  lh->pos_ex = Allocate(1, List);
  lh->leafCache = ListCreate((FreeFunc*)ListFree);

  return lh;
}

void LearnFreeTrees(LearnHandle lh)
{
  int i;
  for(i=0;i<lh->ntrees;i++) HierFree(lh->trees[i]);
}

static void TreeDataFree(LearnHandle lh, int tree)
{
  TreeData *td = &lh->td[tree];
  int i,j;
  for(i=0;i<lh->nclasses;i++) {
    free(td->npos[i]);
    free(td->nneg[i]);
    for(j=0;j<lh->trees[tree]->num_nodes;j++) {
      if(td->ex[i][j]) ListFree(td->ex[i][j]);
    }
    free(td->ex[i]);
  }
  free(td->npos);
  free(td->nneg);
  free(td->ex);
#if NET
  for(j=0;j<lh->trees[tree]->num_nodes;j++) {
    ListFree(td->leaf[j]);
  }
  free(td->leaf);
#endif
  free(td->cbias);
  for(i=0;i<td->nbias;i++) {
    free(td->bpos[i]);
    free(td->bneg[i]);
  }
  free(td->bpos);
  free(td->bneg);
  free(td->bptotal);
  free(td->bntotal);
}

void LearnFree(LearnHandle lh)
{
  int i,j;

  /* Tree info */
  free(lh->trees);
  for(i=0;i<lh->ntrees;i++) TreeDataFree(lh, i);
  free(lh->td);

  /* Class info */
  free(lh->updated);
  for(i=0;i<lh->nclasses;i++) LearnClassDataFree(lh->classData[i]);
  free(lh->classData);

  /* Example info */
  for(i=0;i<lh->nclasses;i++) {
    ListFree(lh->examples[i]);
    ListFree(lh->pos_ex[i]);
  }
  free(lh->examples);
  free(lh->pos_ex);
  ListFree(lh->leafCache);

  free(lh);
}

/* Erase memory of all examples */
void LearnClear(LearnHandle lh)
{
  int i,j,k;
  for(i=0;i<lh->nclasses;i++) {
    lh->updated[i] = 0;
    /* forget examples */
    ListRemoveAll(lh->examples[i], NULL, AllMatchCmp);
    ListRemoveAll(lh->pos_ex[i], NULL, AllMatchCmp);
    /* initialize scores */
    for(j=0;j<lh->ntrees;j++) {
      TreeData *td = &lh->td[j];
      td->cbias[i] = 0;
      for(k=0;k<lh->trees[j]->num_nodes;k++) {
	td->npos[i][k] = td->nneg[i][k] = 0;
	if(td->ex[i][k]) ListFree(td->ex[i][k]);
	td->ex[i][k] = NULL;
      }
    }
  }
}

static void TreeDataInit(LearnHandle lh, int tree)
{
  TreeData *td = &lh->td[tree];
  int i, node, t, node2;

  td->enabled = 1;
  td->weight = 1.0;
  td->npos = Allocate(lh->nclasses, int*);
  td->nneg = Allocate(lh->nclasses, int*);
  td->ex = Allocate(lh->nclasses, List*);
  td->cbias = Allocate(lh->nclasses, int);
  for(i=0;i<lh->nclasses;i++) {
    td->npos[i] = AllocateC(lh->trees[tree]->num_nodes, int);
    td->nneg[i] = AllocateC(lh->trees[tree]->num_nodes, int);
    td->ex[i] = AllocateC(lh->trees[tree]->num_nodes, List);
    td->cbias[i] = 0;
  }
#if NET
  td->leaf = AllocateC(lh->trees[tree]->num_nodes, List);
  for(node=0;node<lh->trees[tree]->num_nodes;node++) {
    /* match this node with nodes in other trees */
    HierNode n1 = HierNodeAt(lh->trees[tree], node);
    td->leaf[node] = ListCreate(LeafFree);
    if(n1->children) continue;
    for(t=0;t<lh->ntrees;t++) {
      for(node2=0;node2<lh->trees[t]->num_nodes;node2++) {
	HierNode n2 = HierNodeAt(lh->trees[t], node2);
	if(n2->children) continue;
	if(!IntCmp(&n2->data, &n1->data)) {
	  Leaf leaf = Allocate(1, struct Leaf);
	  leaf->tree = t;
	  leaf->node = node2;
	  ListAddFront(td->leaf[node], leaf);
	  if(lh->td[t].leaf[node2]) {
	    leaf = Allocate(1, struct Leaf);
	    leaf->tree = tree;
	    leaf->node = node;
	    ListAddFront(lh->td[t].leaf[node2], leaf);
	  }
	}
      }
    }
  }
#endif

  td->nbias = 1;
  td->bptotal = Allocate(1, int);
  td->bntotal = Allocate(1, int);
  td->bpos = Allocate(1, int*);
  td->bneg = Allocate(1, int*);

  /* empty default bias */
  td->bptotal[0] = td->bntotal[0] = 0;
  td->bpos[0] = (int*)calloc(lh->trees[tree]->num_nodes, sizeof(int));
  td->bneg[0] = (int*)calloc(lh->trees[tree]->num_nodes, sizeof(int));
}

/* Add a new tree for the learner to use.
 */
void LearnAddTree(LearnHandle lh, Hier tree)
{
  int t = lh->ntrees++;

  Reallocate(lh->trees, lh->ntrees, Hier);
  lh->trees[t] = tree;

  Reallocate(lh->td, lh->ntrees, TreeData);
  TreeDataInit(lh, t);

  /* example nodes must be recalculated */
  /* score arrays must be reallocated */
  /* for now, require no classes */
  assert(!lh->nclasses);
}

static void AddExToTree(LearnHandle lh, int class, int tree, 
			int item, float weight)
{
  /* climb the tree, changing scores */
  for(;item >= 0;item = HierNodeAt(lh->trees[tree], item)->parent) {
    if(weight > 0) lh->td[tree].npos[class][item]++;
    else lh->td[tree].nneg[class][item]++;
  }
}

/* Calls AddExToTree for all nodes and records the example */
void AddEx(LearnHandle lh, int class, List leaves, float weight)
{
  Example ex;
  Leaf leaf;
  ListIter(p, leaf, leaves) {
    AddExToTree(lh, class, leaf->tree, leaf->node, weight);
  }
  lh->updated[class] = 0;

  /* remember the example */
  ex = Allocate(1, struct Example);
  ex->mark = 0;
  ex->weight = weight;
  ex->leaves = leaves;
  ListAddFront(lh->examples[class], ex);
  if(weight > 0) ListAddFront(lh->pos_ex[class], ex);

  /* make trees point to it */
  {ListIter(p, leaf, leaves) {
    TreeData *td = &lh->td[leaf->tree];
    if(!td->ex[class][leaf->node]) 
      td->ex[class][leaf->node] = ListCreate(NULL);
    ListAddFront(td->ex[class][leaf->node], ex);
  }}
}

/* Add a new class entry. Returns the index of the class.
 */
int LearnAddClass(LearnHandle lh)
{
  int i,j;
  int c = lh->nclasses++;

  Reallocate(lh->updated, lh->nclasses, int);
  lh->updated[c] = 1;

  Reallocate(lh->examples, lh->nclasses, List);
  lh->examples[c] = ListCreate((FreeFunc*)ExampleFree);

  Reallocate(lh->pos_ex, lh->nclasses, List);
  lh->pos_ex[c] = ListCreate(NULL);

  Reallocate(lh->classData, lh->nclasses, void*);
  lh->classData[c] = LearnClassDataInit(c);

  for(i=0;i<lh->ntrees;i++) {
    TreeData *td = &lh->td[i];

    Reallocate(td->cbias, lh->nclasses, int);
    td->cbias[c] = 0; /* default bias */

    Reallocate(td->npos, lh->nclasses, int*);
    Reallocate(td->nneg, lh->nclasses, int*);
    Reallocate(td->ex, lh->nclasses, List*);
    td->npos[c] = AllocateC(lh->trees[i]->num_nodes, int);
    td->nneg[c] = AllocateC(lh->trees[i]->num_nodes, int);
    td->ex[c] = AllocateC(lh->trees[i]->num_nodes, List);
  }
  
  /* add implicit negative examples to this class, for the positive examples
   * of other classes.
   */
  for(i=0;i<c;i++) {
    Example ex;
    ListIter(p, ex, lh->pos_ex[i]) {
      AddEx(lh, c, ex->leaves, -1);
    }
  }
  return c;
}

void LearnEnableTree(LearnHandle lh, int tree)
{
  int i;
  /* all classes need to be updated */
  for(i=0;i<lh->nclasses;i++) lh->updated[i] = 0;
  lh->td[tree].enabled = 1;
}

void LearnDisableTree(LearnHandle lh, int tree)
{
  int i;
  /* all classes need to be updated */
  for(i=0;i<lh->nclasses;i++) lh->updated[i] = 0;
  lh->td[tree].enabled = 0;
}

static List GetLeaves(LearnHandle lh, int class, void *item, CmpFunc *compare)
{
  int tree, i;
  List leaves;

  /* Check the leaf cache? */

  /* Determine node positions in each tree */
  leaves = ListCreate(LeafFree);
  /* Loop trees */
  for(tree=0;tree<lh->ntrees;tree++) {
    /* Loop leaves */
    for(i=0;i < lh->trees[tree]->num_nodes;i++) {
      HierNode node = HierNodeAt(lh->trees[tree], i);
      if(node->children) continue;
#if 0
      /* already an example? */
      if(lh->td[tree].ex[class][i]) continue; /* first come, first served */
#endif
      if(!compare(&node->data, item)) {
	Leaf leaf = Allocate(1, struct Leaf);
	leaf->tree = tree;
	leaf->node = i;
	/* add to rear to keep trees in order */
	ListAddRear(leaves, leaf);
#if 0
	/* assume at most one leaf per tree */
	break;
#endif
      }
    }
  }
  if(ListEmpty(leaves)) { ListFree(leaves); return NULL; }
  ListAddFront(lh->leafCache, leaves);
  return leaves;
}

/* Add an example of class with the given weight.
 * weight < 0 means negative example.
 * compare is used to match item with tree leaves.
 */
void LearnAddEx(LearnHandle lh, int class, float weight, 
		void *item, CmpFunc *compare)
{
  List leaves;
  Example ex;

  leaves = GetLeaves(lh, class, item, compare);
  if(!leaves) return;
  AddEx(lh, class, leaves, weight);

  if(weight > 0) {
    int i;
    /* add implicit negatives for other classes */
    for(i=0;i<lh->nclasses;i++) {
      if(i == class) continue;
      AddEx(lh, i, leaves, -1);
    }
  }
}

void LearnUpdate(LearnHandle lh)
{
  int c;
  for(c=0;c<lh->nclasses;c++) {
    if(lh->updated[c]) continue;
#if DEBUG
    fprintf(stderr, "updating class %d\n", c);
#endif
    LearnUpdateClass(lh, c);
    lh->updated[c] = 1;
  }
}

List LearnWithClass(LearnHandle lh, int class)
{
  return LearnWithClassMatching(lh, class, NULL, NULL);
}

/* Returns a List of the hier data for the items of the given class
 * matching item, using the supplied comparison function.
 * If compare == NULL, all items in the class are returned.
 * The result may contain duplicates.
 * Uses LearnClassWeight to do the work.
 */
List LearnWithClassMatchingGeneric(LearnHandle lh, int class, 
				   void *item, CmpFunc *compare)
{
  int tree, node;
  List result = ListCreate(NULL);
#if !NET
  List leaves = ListCreate(NULL);
  struct Leaf leaf;
#endif

  if(class >= lh->nclasses) return result;
  LearnUpdate(lh);
#if !NET
  ListAddFront(leaves, &leaf);
#endif
  /* loop all leaves in all trees, adding them to result if they are
     included in lh->covers[class] */
  for(tree=0;tree<lh->ntrees;tree++) {
#if !NET
    leaf.tree = tree;
#endif
    for(node=0;node<lh->trees[tree]->num_nodes;node++) {
      HierNode hierNode = HierNodeAt(lh->trees[tree], node);
      /* is it a leaf? */
      if(hierNode->children) continue;
      /* does it match item? */
      if(compare && compare(&hierNode->data, item)) continue;
#if NET
      if(LearnClassWeight(lh, class, lh->td[tree].leaf[node]) > 0.0) 
	ListAddFront(result, &hierNode->data);
#else
      leaf.node = node;
      if(LearnClassWeight(lh, class, leaves) > 0.0) 
	ListAddFront(result, &hierNode->data);
#endif
    }
  }
#if !NET
  ListFree(leaves);
#endif
  return result;
}

/* Makes all covered nodes examples of that class */
static void CommitClass(LearnHandle lh, int class)
{
  List list = LearnWithClass(lh, class);
  void *data;
  ListIter(p, data, list) {
    /* XXX cmpfunc is not general */
    /* let LearnAddEx take care of duplicates */
    LearnAddEx(lh, class, 1, data, IntCmp);
  }
  ListFree(list);
}

/* Make all nodes assigned to a class by the algorithm real examples of 
 * that class.  I.e. pin down the learner's classifications.
 */
void LearnCommit(LearnHandle lh)
{
  int i;
  /* commit all classes */
  for(i=0;i<lh->nclasses;i++) {
    /* prevent updates */
    lh->updated[i] = 1;
    CommitClass(lh, i);
  }
}

/* Returns a LearnHandle which stores the example state of the
 * given class.  The Example structures are shared with lh.
 */
LearnHandle LearnSaveState(LearnHandle lh, int class)
{
  LearnHandle save = Allocate(1, struct LearnHandle);
  int tree;

  save->td = Allocate(lh->ntrees, TreeData);
  for(tree=0;tree<lh->ntrees;tree++) {
    TreeData *std = &save->td[tree];
    TreeData *td = &lh->td[tree];
    std->npos = Allocate(1, int*);
    std->npos[0] = td->npos[class];
    std->nneg = Allocate(1, int*);
    std->nneg[0] = td->nneg[class];
    std->ex = Allocate(1, List*);
    std->ex[0] = td->ex[class];
  }
  save->examples = Allocate(1, List);
  save->examples[0] = lh->examples[class];
  save->pos_ex = Allocate(1, List);
  save->pos_ex[0] = lh->pos_ex[class];
  return save;
}

/* Clears all examples from the given class, without freeing the
 * Example structures.
 */
void LearnClearState(LearnHandle lh, int class)
{
  int tree;
  for(tree=0;tree<lh->ntrees;tree++) {
    TreeData *td = &lh->td[tree];
    td->npos[class] = AllocateC(lh->trees[tree]->num_nodes, int);
    td->nneg[class] = AllocateC(lh->trees[tree]->num_nodes, int);
    td->ex[class] = AllocateC(lh->trees[tree]->num_nodes, List);
  }
  lh->examples[class] = ListCreate(GenericFree);  /* not ExampleFree */
  lh->pos_ex[class] = ListCreate(NULL);
}

/* Restores the example state saved by LearnSaveState. */
void LearnRestoreState(LearnHandle lh, int class, LearnHandle save)
{
  int tree, i;

  /* restore examples */
  ListFree(lh->examples[class]);
  lh->examples[class] = save->examples[0];
  free(save->examples);
  ListFree(lh->pos_ex[class]);
  lh->pos_ex[class] = save->pos_ex[0];
  free(save->pos_ex);

  /* restore score arrays */
  for(tree=0;tree<lh->ntrees;tree++) {
    TreeData *std = &save->td[tree];
    TreeData *td = &lh->td[tree];
    free(td->npos[class]);
    td->npos[class] = std->npos[0];
    free(std->npos);
    free(td->nneg[class]);
    td->nneg[class] = std->nneg[0];
    free(std->nneg);
    for(i=0;i<lh->trees[tree]->num_nodes;i++) 
      if(td->ex[class][i]) ListFree(td->ex[class][i]);
    free(td->ex[class]);
    td->ex[class] = std->ex[0];
    free(std->ex);
  }
  free(save->td);
  free(save);
}

/* Returns the average number of pos in pure covers of tree */
static float TreeScore(LearnHandle lh, int tree, int class)
{
  float score = 0;
  TreeData *td = &lh->td[tree];
  int i;
  for(i=0;i<lh->trees[tree]->num_nodes;i++) {
    if(td->nneg[class][i]) continue;
    score += td->npos[class][i];
  }
  return score /= lh->trees[tree]->num_nodes;
}

void LearnTreeScore(LearnHandle lh, int class)
{
  int i;
  for(i=0;i<lh->ntrees;i++) {
    lh->td[i].weight = TreeScore(lh, i, class);
  }
}

/*****************************************************************************/
typedef struct {
  void *data;
  int index;
} NodeData;

/* Appends the preorder traversal index to each tree node. */
/* Requires that tree.c uses TreeAddChild (not LeftChild). */
static void AugmentTree(Tree tree, int *pos)
{
  NodeData *d;
  if(!tree) return;
  d = Allocate(1, NodeData);
  d->data = tree->data;
  d->index = (*pos)++;
  tree->data = d;
  AugmentTree(tree->child, pos);
  AugmentTree(tree->sibling, pos);
}

/* Restores the tree's original data. */
static void StripTree(Tree tree)
{
  NodeData *d;
  if(!tree) return;
  d = tree->data;
  tree->data = d->data;
  free(d);
  StripTree(tree->child);
  StripTree(tree->sibling);
}

static void *NewParentData(Tree tree)
{
  return IntCreate(-1);
}

/* Unite positive siblings, and promote positive singletons. 
 * Modifies *tree.
 * Increments *count for each node added.
 */
#define MARK -1
static void UpdateTree(TreeData *td, Tree parent, Tree *tree, 
			int class, int *count)
{
  Tree *t, c, new;
  if(!*tree) return;

  /* Loop children, recursing on mixed subtrees. */
  for(t = tree; c = *t; t = &(*t)->sibling) {
    NodeData *d = (*t)->data;
    if(d->index == MARK) continue; /* created by us; must be pure negative */
    if(td->npos[class][d->index] && td->nneg[class][d->index]) {
      UpdateTree(td, c, &c->child, class, count);
    }
  }

  /* Loop siblings, collecting pure positive subtrees. */
  new = TreeCreate(NULL);
  for(t = tree; c = *t; ) {
    NodeData *d = (*t)->data;
    if((d->index == MARK)  /* created by us; must be pure negative */
       || !td->npos[class][d->index] 
       || td->nneg[class][d->index]) t = &(*t)->sibling;
    else {
      *t = c->sibling;
      c->sibling = NULL;
      TreeAddChild(new, c);
    }
  }

  /* Removed them all? */
  if(!*tree) {
    /* Put back */
    /* could flatten here */
    *tree = new->child;
    for(t = &new->child; c = *t; t = &(*t)->sibling) c->parent = parent;
    new->child = NULL;
#if VERBOSE
    fprintf(stderr, "removed all\n");
#endif
  }

  /* Got one? */
  if(new->child) {
    /* Got more than one? */
    if(new->child->sibling) {
      /* Add new as a child. */
      NodeData *d = Allocate(1, NodeData);
      d->data = NewParentData(new);
      d->index = MARK;
      new->data = d;
      new->parent = parent;
      new->sibling = *tree;
      *tree = new;
      (*count)++;
#if VERBOSE
      fprintf(stderr, "unite\n");
#endif
    }
    else {
      /* Promote child */
      c = new->child;
      new->child = NULL;
      TreeFree(new, NULL);
      if(parent) {
#if VERBOSE
	fprintf(stderr, "promote\n");
#endif
	c->parent = parent->parent;
	c->sibling = parent->sibling;
	parent->sibling = c;
      }
      else {
	/* put back */
	c->parent = (*tree)->parent;
	c->sibling = (*tree)->sibling;
	(*tree)->sibling = c; 
#if VERBOSE
	fprintf(stderr, "put back\n");
#endif
      }
    }
  }
  else TreeFree(new, NULL);
}

void LearnSaveTree(LearnHandle lh, int tree, char *fname)
{
  Tree t;
  TreeData *td = &lh->td[tree];
  Hier hier;
  int i, count;
  FILE *fp = fopen(fname, "w");
  if(!fp) { perror("LearnSaveTree"); return; }

  /* update tree */
  t = TreeFromHier(lh->trees[tree]);
  i = 0;
  AugmentTree(t, &i);

  /* can do classes separately since their pos ex are disjoint */
  count = lh->trees[tree]->num_nodes;
  for(i=0;i<lh->nclasses;i++) {
    /* show pos ex */
#if VERBOSE
    printf("class %d: ", i);
    {Example ex;ListIter(p, ex, lh->pos_ex[i]) {
      Leaf leaf = (Leaf)ListFront(ex->leaves)->data;
      printf("%d ", 
	     *(int*)(&HierNodeAt(lh->trees[leaf->tree], leaf->node)->data));
    }}
    printf("\n");
#endif
    UpdateTree(td, NULL, &t, i, &count);
  }

  StripTree(t); 
#if 0
  TreeWrite(stdout, t, IntWrite); putchar('\n');
#endif
  count -= TreeSimplify(t, GenericFree);
#if 0
  TreeWrite(stdout, t, IntWrite); putchar('\n');
#endif

  hier = HierCreate(count, lh->trees[tree]->data_size);
  HierFromTree(hier, t);
  HierWrite(hier, fp);
  fclose(fp);
  HierFree(hier);
  TreeFree(t, GenericFree);
}
