/* Binary classifier.
 * SetObj with a sample and a number in [-1, 1] trains the classifier.
 * GetObj with a sample returns a float in [-1, 1], where zero means 
 * "no opinion".
 *
 * A sample is a List of (feature, key) pairs, where
 * key is an index into the feature object.
 * E.g. if feature is a Hier then key is a node number.
 *
 * Uses an IdDict to store the features.
 * For each feature, there is associated information for performing
 * classification.  For example, pos/neg counts for each node of a Hier.
 *
 * This classifier works like a Bayesian tree.  It uses an upward/downward
 * algorithm to compute probabilities.  Uses ratios of alpha and beta values
 * to avoid numerical underflow.
 */
#include <tpm/list.h>
#include <boa/boa.h>
#include <tpm/hier.h>

#define VERBOSE 0

typedef struct Classifier {
  Obj features;
} *Classifier;

/* Globals *******************************************************************/
Type T_Classifier_Belief;

/* Functions *****************************************************************/

static void Init(Obj self)
{
  Classifier c = (Classifier)self->value;
  c->features = Boa_Create(T_IdDict);
}

static void Final(Obj self)
{
  Classifier c = (Classifier)self->value;
  /* XXX go through and manually free all the child lists */
  Boa_Release(c->features);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Classifier c = (Classifier)self->value;
  func(c->features, data);
}

/* P_T_T is probability of child true given parent true, etc. */
#define ODDS 2
#define P_T_T ((double)ODDS/(1+ODDS))
#define P_F_T (1-P_T_T)
#define P_F_F P_T_T
#define P_T_F (1-P_F_F)

typedef struct TreeData {
  List children;	/* list of node numbers */
  double alpha, beta;	/* ratios F/T */
  /* double-precision is needed for large runs */
} *TreeData;

static void ComputeProbs(Hier hier, TreeData data)
{
  static int ctr = 0;
  int i;
  HierNode node;
  /* compute beta values upward */
  for(i=0;i<hier->num_nodes;i++) {
    if(data[i].children) data[i].beta = 1;
  }
  for(i=hier->num_nodes;i;) {
    i--;
    node = HierNodeAt(hier, i);
    if(node->parent < 0) continue;
    if(data[i].beta + 1 == data[i].beta) /* check for infinity */
      data[node->parent].beta *= P_F_F / P_F_T;
    else 
      data[node->parent].beta *= 
	(P_T_F + P_F_F*data[i].beta) / (P_T_T + P_F_T*data[i].beta);
  }
  /* compute alpha values downward */
  for(i=1;i<hier->num_nodes;i++) {
    double a;
    node = HierNodeAt(hier, i);
    if(node->parent < 0) continue;
    a = data[node->parent].alpha;
    if(a + 1 == a) /* check for infinity */
      data[i].alpha = P_F_F / P_T_F;
    else {
      int j; 
      /* ignore warnings about this line */
      ListIterInt(p, j, data[node->parent].children) {
	if(j == i) continue;
	if(data[j].beta + 1 == data[j].beta) /* check for infinity */
	  a *= P_F_F / P_F_T;
	else
	  a *= (P_T_F + P_F_F*data[j].beta) / (P_T_T + P_F_T*data[j].beta); 
      }
      data[i].alpha = (P_F_F*a + P_F_T) / (P_T_F*a + P_T_T); 
    }
  }
  if(VERBOSE && !(ctr++ % 100)) {
    /* display results */
    for(i=0;i<hier->num_nodes;i++) {
      double p = 1.0/(1+data[i].beta*data[i].alpha);
      fprintf(stderr, "%d: %g %g %g\n", i, data[i].beta, data[i].alpha, p);
    }
  }
}

#if 0
/* Returns the data of all instances in the class, i.e. with weight > 0.
 * classifier is a Classifier.
 * features is an Ord.
 * If compare != NULL, only instance data matching item will be returned.
 */
List BoaClassifier_AllMatching(Obj self, Obj features, 
			       CmpFunc *compare,
			       Any item, CmpFunc *include_compare)
{
  Classifier c = (Classifier)self->value;
  int f, i;
  double not_for, not_against;
  List in = ListCreate(NULL), out = ListCreate(NULL);

  /* loop features */
  for(f=0;f<Boa_Ord_Length(features);f++) {
    Obj feature = Boa_Ord_Item(features, f);
    Obj t;
    if(Boa_ObjGet(c->features, feature, 0, NULL, &t) 
       != TPM_OK) {
      /* haven't seen this feature before */
      continue;
    }

    if(feature->type->type == T_Hier) {
      /* Loop all nodes and collect in/out */
      int i;
      Hier hier = Boa_Hier_Value(feature);
      TreeData data = (TreeData)t->value;
      for(i=0;i<hier->num_nodes;i++) {
	HierNode node = HierNodeAt(hier, i);
	double p;
	if(node->children) continue;
	if(include_compare && include_compare(&node->data, item)) continue;
	not_for = not_against = 1;
	/* compute probs here */
	p = 1.0/(1+data[i].beta*data[i].alpha);
	not_for *= 1-p;
	not_against *= p;
	if(not_against > not_for) ListAddFront(in, &node->data);
	else if(not_for > not_against) ListAddFront(out, &node->data);
      }
    }
    else {
      fprintf(stderr, "Classifier: unknown feature type ");
      Boa_PrintNL(stderr, feature->type);
    }
    Boa_Release(t);
  }
  /* subtract out from in */
  {Any data; ListIter(p, data, out) {
    ListRemoveAll(in, data, compare);
  }}
  ListFree(out);
  return in;
}
#endif

/* index is a List of (feature, key) tuples.
 * returns a float certainty factor.
 */
static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Classifier c = (Classifier)self->value;
  List pairs;
  Obj pair;
  double not_for, not_against;
  if(index->type != T_List) return TPM_BOTTOM;
  pairs = Boa_List_Value(index);

  not_for = not_against = 1;
  {ListIter(p, pair, pairs) {
    Obj feature, t;

    assert(Boa_Tuple_Length(pair) == 2);
    feature = Boa_Tuple_Item(pair, 0);
    if(Boa_ObjGet(c->features, feature, 0, NULL, &t) 
       != TPM_OK) {
      /* haven't seen this feature before */
      continue;
    }

    if(feature->type->type == T_Hier) {
      TreeData data = (TreeData)Boa_Tuple_Item(t, 0)->value;
      int i = Boa_Int_Value(Boa_Tuple_Item(pair, 1));
      double p;
      if(Boa_Int_Value(Boa_Tuple_Item(t, 1)) == 0) {
	/* recompute probabilities */
	ComputeProbs(Boa_Hier_Value(feature), data);
	Boa_Int_Value(Boa_Tuple_Item(t, 1)) = 1;
      }
      p = 1.0/(1+data[i].beta*data[i].alpha);
      not_for *= 1-p;
      not_against *= p;
    }
    else {
      fprintf(stderr, "Classifier: unknown feature type ");
      Boa_PrintNL(stderr, feature->type);
    }
    Boa_Release(t);
  }}

#if 0
  /* force 1 or -1 */
  if(not_against > not_for) { not_against = 1; not_for = 0; }
  else { not_against = 0; not_for = 1; }
#endif

  BOA_TYPECHECK_GET(T_Float);
  if(value_return) *value_return = Boa_Float_Create(not_against - not_for);
  return TPM_OK;
}

/* index is a List of (feature, key) tuples.
 * value is a float certainty factor.
 */
static int ObjSet(Obj self, Obj index, int flags, Obj value)
{
  Classifier c = (Classifier)self->value;
  float certainty;
  List pairs;
  Obj pair;
  if(index->type != T_List) return TPM_BOTTOM;
  pairs = Boa_List_Value(index);
  if(value->type != T_Float) return TPM_BOTTOM;
  certainty = Boa_Float_Value(value);
  /* assumes we won't get an error */
  Boa_Release(value);

  {ListIter(p, pair, pairs) {
    Obj feature, t;
    int i;
    Hier hier;
    TreeData data;

    assert(Boa_Tuple_Length(pair) == 2);
    feature = Boa_Tuple_Item(pair, 0);
    assert(feature->type->type == T_Hier);
    hier = Boa_Hier_Value(feature);
    i = Boa_Int_Value(Boa_Tuple_Item(pair, 1));

    if(Boa_ObjGet(c->features, feature, 0, NULL, &t) 
       != TPM_OK) {
      int j;
      char str[64];
      /* haven't seen this feature before */
      /* create an array of TreeData for this feature */
      sprintf(str, "(array %d (array %d char))", 
	      hier->num_nodes, sizeof(struct TreeData));
      t = Boa_Create(Boa_Type_Create(str));
      data = (TreeData)t->value;
      /* initialize the treeData */
      for(j=0;j<hier->num_nodes;j++) {
	HierNode node = HierNodeAt(hier, j);
	/* add myself as a child of my parent */
	if(node->parent >= 0) {
	  ListAddFront(data[node->parent].children, (Any)j);
	}
	if(node->children) data[j].children = ListCreate(NULL);
	data[j].beta = 1;
	data[j].alpha = 1;
      }
      pair = Boa_Create(Boa_Type_Create("(tuple 2)"));
      Boa_Tuple_Item(pair, 0) = t;
      Boa_Tuple_Item(pair, 1) = Boa_Int_Create(0);  /* updated flag */
      Boa_ObjSet(c->features, feature, BOA_INSERT, pair);
    }
    else {
      data = (TreeData)Boa_Tuple_Item(t, 0)->value;
      Boa_Int_Value(Boa_Tuple_Item(t, 1)) = 0;
    }

    /* update the beta value of this leaf */
    if(certainty <= 0) data[i].beta = 1/(1+certainty);
    else data[i].beta = 1 - certainty;
  }} 
  return TPM_OK;
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  Generic_SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Err_StringGet, Err_StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_Classifier_Belief_Define(void)
{
  T_Classifier_Belief = Boa_SimpleType_Create("classifier_belief", &methods, 
					      sizeof(struct Classifier), 
					      sizeof(Obj));
  return T_Classifier_Belief;
}
