/* Binary classifier.
 * SetObj with a sample and a number in [-1, 1] trains the classifier.
 * GetObj with a sample returns a number in [-1, 1], where zero means 
 * "no opinion".
 *
 * A sample is a List of (feature, key) pairs, where
 * key is an index into the feature object.
 * E.g. if feature is a Hier then key is a node number.
 *
 * Uses an IdDict to store the features.
 * Assumes all features are metrics.
 *
 * This classifier is a generalized two-layer perceptron, where any
 * likelihood function can be used at the hidden nodes.  The likelihood
 * function is derived from the distance d as (1 - 2*d).
 */
#include <tpm/list.h>
#include <tpm/hash.h>
#include <tpm/optim.h>
#include <boa/boa.h>
#include <boa/vector.h>

#define DEBUG 0
#define HIDDEN_MAX 400
#define ZERO_MEAN 1
#define BIAS 1
#define PARAMETRIC 1
#define OFFSET 0
/* the cross term is expensive to compute and usually negligible */
#define CROSS_TERM 0

typedef struct Classifier {
  Obj features;
  float *certainty;
  int ntests;
  int nhidden;
  int *hidden_pos, *hidden_neg;
  Vector hidden_weight;
  int updated;
} *Classifier;

/* Globals *******************************************************************/
Type T_Classifier_Butterfly;

/* Functions *****************************************************************/

static void Init(Obj self)
{
  Classifier c = (Classifier)self->value;
  c->features = Boa_Create(T_IdDict);
  c->nhidden = 1;
  c->hidden_pos = Allocate(c->nhidden, int);
  c->hidden_neg = Allocate(c->nhidden, int);
  c->hidden_weight = VectorCreate(c->nhidden+BIAS);
}

static void Final(Obj self)
{
  Classifier c = (Classifier)self->value;
  Boa_Release(c->features);
  if(c->certainty) free(c->certainty);
  free(c->hidden_pos);
  free(c->hidden_neg);
  VectorFree(c->hidden_weight);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Classifier c = (Classifier)self->value;
  func(c->features, data);
}

static int FeatureKeyIndex(Obj feature, Obj key) 
{
  Obj tests;
  int i;
  /* assume feature is a metric object, and key is a query */
  if(Boa_StringGet(feature, "_context/test", '/', NULL, &tests) == TPM_OK) {
    assert(Boa_ObjGetInt(tests, key, &i) == TPM_OK);
    Boa_Release(tests);
    return i;
  }
  else if(key->type == T_Int) {
    return Boa_Int_Value(key);
  } 
  fprintf(stderr, "FeatureKeyIndex: don't know how to index ");
  Boa_PrintNL(stderr, key);
  return -1;
}

static int FeatureNumKeys(Obj feature)
{
  Obj tests;
  int i;
  /* assume feature is a metric object, and key is a query */
  if(Boa_StringGet(feature, "_context/test", '/', NULL, &tests) == TPM_OK) {
    assert(Boa_StringGetInt(tests, "_length", &i) == TPM_OK);
    Boa_Release(tests);
  }
  else assert(Boa_StringGetInt(feature, "_length", &i) == TPM_OK);
  return i;
}

/* If use dp/(dp+dn), get a sigmoid instead. */
static double Butterfly(double dp, double dn, double dpn)
{
  double dp2 = dp*dp;
  double dn2 = dn*dn;
  double cross = dp*dn*dpn;
  if(dp2 - 2*cross + dn2 < 1e-20) return 0;
  return (dp2*dp - dn2*dn + cross*(dn - dp)) / (dp2 - 2*cross + dn2);
}

#if 0
/* Set the hidden->output weights to a least-squares fit.
 * desired and hidden are workspaces. 
 * Returns the quality of fit.
 */
static double OptimizeWeights(Classifier c, Vector desired, Matrix hidden)
{
  Obj feature, dummy;
  Obj key = Boa_Int_Create(0);
  Vector v;
  double err;
  /* Collect a vector of hidden unit reponses and desired outputs */
  /* Loop features */
  HashIter(b, entry, dummy, (HashTable)c->features->value) {
    /* Loop examples */
    int e = 0, i, h;
    feature = (Obj)entry->key;
    for(i=0;i<c->ntests;i++) {
      Obj t;
      Vector dists;
      if(c->certainty[i] == 0) continue;
#if ZERO_MEAN
      desired->d[e] = c->certainty[i]-OFFSET;
#else
      desired->d[e] = (c->certainty[i]+1)/2;
#endif
      Boa_Int_Value(key) = i;
      /* get a vector of distances */
      if(Boa_ObjGet(feature, key, 0, NULL, &t) != TPM_OK) {
	ErrorPrint(stderr);
      }
      else {
	dists = Boa_Vector_Value(t);
	/* Loop hidden units */
	for(h=0;h<c->nhidden;h++) {
	  hidden->d2[e][h] = Butterfly(1 - 2*dists->d[c->hidden_neg[h]], 
				       1 - 2*dists->d[c->hidden_pos[h]]);
	}
	Boa_Release(t);
#if BIAS
	hidden->d2[e][h] = 1;
#endif
      }
      e++;
    }
    break;
  }
  Boa_Release(key);
#if DEBUG > 1
  fprintf(stderr, "hidden:\n");
  MatrixPrint(stderr, hidden);
  fprintf(stderr, "desired:\n");
  VectorPrint(stderr, desired);
#endif
  MatrixLLS(hidden, desired->d, c->hidden_weight->d);
  v = VectorCreate(desired->len);
  MatrixVectorMultiply(v, hidden, c->hidden_weight);
  VectorSubtract(v, desired);
  err = VectorDot(v, v);
  VectorFree(v);
  return err;
}

#else

static double OptimizeWeights(Classifier c, Vector desired, Matrix hidden, 
			      Obj *distance)
{
  int e = 0, i, h;
  double err;
  Vector v;

#if CROSS_TERM
  Obj *hdistance = Allocate(c->nhidden, Obj);
  Obj dummy;
  {HashIter(b, entry, dummy, (HashTable)c->features->value) {
    Obj feature = (Obj)entry->key;
    for(h=0;h<c->nhidden;h++) {
      Obj t, key = Boa_Int_Create(c->hidden_pos[h]);
      assert(Boa_ObjGet(feature, key, 0, NULL, &t) == TPM_OK);
      Boa_Release(key);
      hdistance[h] = t;
    }
    break;
  }}
#endif

  for(i=0;i<c->ntests;i++) {
    Vector dists;
    if(c->certainty[i] == 0) continue;
    dists = Boa_Vector_Value(distance[e]);
    /* Loop hidden units */
    for(h=0;h<c->nhidden;h++) {
#if CROSS_TERM
      Vector hdists = Boa_Vector_Value(hdistance[h]);
      hidden->d2[e][h] = Butterfly(1 - 2*dists->d[c->hidden_neg[h]], 
				   1 - 2*dists->d[c->hidden_pos[h]],
				   1 - 2*hdists->d[c->hidden_neg[h]]);
#else
      hidden->d2[e][h] = Butterfly(1 - 2*dists->d[c->hidden_neg[h]], 
				   1 - 2*dists->d[c->hidden_pos[h]],
				   0);
#endif
    }
#if BIAS
    hidden->d2[e][h] = 1;
#endif
    e++;
    if(e == desired->len) break;
  }

#if CROSS_TERM
  for(h=0;h<c->nhidden;h++) {
    Boa_Release(hdistance[h]);
  }
  free(hdistance);
#endif

#if DEBUG > 1
  fprintf(stderr, "hidden:\n");
  MatrixPrint(stderr, hidden);
  fprintf(stderr, "desired:\n");
  VectorPrint(stderr, desired);
#endif
  MatrixLLS(hidden, desired->d, c->hidden_weight->d);
  v = VectorCreate(desired->len);
  MatrixVectorMultiply(v, hidden, c->hidden_weight);
  VectorSubtract(v, desired);
  err = VectorDot(v, v);
  VectorFree(v);
  return err;
}
#endif

static void Pack(Classifier c, Vector p)
{
  int h;
  /* Loop hidden units */
  for(h=0;h<c->nhidden;h++) {
    p->d[4*h] = c->hidden_pos[h] % 20;
    p->d[4*h+1] = c->hidden_pos[h] / 20;
    p->d[4*h+2] = c->hidden_neg[h] % 20;
    p->d[4*h+3] = c->hidden_neg[h] / 20;
  }
}

static void Unpack(Vector p, Classifier c)
{
  int h;
  /* Loop hidden units */
  for(h=0;h<c->nhidden;h++) {
    int x, y;

    x = (int)(p->d[4*h]+0.5);
    y = (int)(p->d[4*h+1]+0.5);
    if(x < 0) x = 0;
    if(x > 19) x = 19;
    if(y < 0) y = 0;
    if(y > 19) y = 19;
    c->hidden_pos[h] = x + y*20;

    x = (int)(p->d[4*h+2]+0.5);
    y = (int)(p->d[4*h+3]+0.5);
    if(x < 0) x = 0;
    if(x > 19) x = 19;
    if(y < 0) y = 0;
    if(y > 19) y = 19;
    c->hidden_neg[h] = x + y*20;
  }
}

typedef struct TrainData {
  Classifier c;
  Vector desired;
  Matrix hidden;
  Obj *distance, *hdistance;
} *TrainData; 

static double Error(Vector p, TrainData data)
{
  Unpack(p, data->c);
  return OptimizeWeights(data->c, data->desired, data->hidden, 
			 data->distance);
}

static void OptimizeHidden(Classifier c)
{
  int i, e;
  Obj feature, dummy;
  Obj *distance;
  Obj key;
  Vector desired;
  Matrix hidden;
  double min_err;
  Vector p;
  struct TrainData data;

  /* Count the number of examples */
  e = 0;
  for(i=0;i<c->ntests;i++) {
    if(c->certainty[i] == 0) continue;
    e++;
  }
#if DEBUG
  fprintf(stderr, "%d examples\n", e);
#endif
  if(!e) {
    memset(c->hidden_pos, 0, c->nhidden*sizeof(int));
    memset(c->hidden_neg, 0, c->nhidden*sizeof(int));
    VectorSet(c->hidden_weight, 0);
    return;
  }
  desired = VectorCreate(e);
  distance = Allocate(e, Obj);

  /* Collect desired outputs and distance vectors */
  key = Boa_Int_Create(0);
  /* Loop features */
  {HashIter(b, entry, dummy, (HashTable)c->features->value) {
    /* Loop examples */
    e = 0;
    feature = (Obj)entry->key;
    for(i=0;i<c->ntests;i++) {
      Obj t;
      if(c->certainty[i] == 0) continue;
#if ZERO_MEAN
      desired->d[e] = c->certainty[i]-OFFSET;
#else
      desired->d[e] = (c->certainty[i]+1)/2;
#endif
      Boa_Int_Value(key) = i;
      /* get a vector of distances */
      if(Boa_ObjGet(feature, key, 0, NULL, &t) != TPM_OK) {
	ErrorPrint(stderr);
	distance[e] = NULL;
      }
      else {
	distance[e] = t;
      }
      e++;
    }
    break;
  }}
  Boa_Release(key);

  data.c = c;
  data.desired = desired;
  data.distance = distance;
#if PARAMETRIC
  for(;;) {
    p = VectorCreate(c->nhidden*4);
    Pack(c, p);
    hidden = MatrixCreate(e, c->nhidden+BIAS);
    data.hidden = hidden;
    for(i=0;;i++) {
      OptPowell(p, (OptMScalarFunc*)Error, &data, &min_err);
      if(min_err < 1e-1) break;
      if(i == 20) break;
      fprintf(stderr, "restarting (min_err = %g)\n", min_err);
      VectorApply(p, (MathFunc*)Rand);
      VectorScale(p, 20);
    }
    if(i < 20) break;
    if(c->nhidden == HIDDEN_MAX) break;
    /* increase the number of hidden units */
    MatrixFree(hidden);
    VectorFree(p);
    free(c->hidden_pos);
    free(c->hidden_neg);
    VectorFree(c->hidden_weight);
    c->nhidden++;
    fprintf(stderr, "increasing nhidden to %d\n", c->nhidden);
    c->hidden_pos = Allocate(c->nhidden, int);
    c->hidden_neg = Allocate(c->nhidden, int);
    c->hidden_weight = VectorCreate(c->nhidden+BIAS);
    for(i=0;i<c->nhidden;i++) {
      c->hidden_pos[i] = RandInt(0, c->ntests-1);
      c->hidden_neg[i] = RandInt(0, c->ntests-1);
    }
  }
  Unpack(p, c);
  VectorFree(p);
  assert(min_err == OptimizeWeights(c, desired, hidden, distance));
#else
  /* nonparametric */
  free(c->hidden_pos);
  free(c->hidden_neg);
  VectorFree(c->hidden_weight);
  c->nhidden = e;
  c->hidden_pos = AllocateC(c->nhidden, int);
  c->hidden_neg = AllocateC(c->nhidden, int);
  c->hidden_weight = VectorCreate(c->nhidden+BIAS);
  hidden = MatrixCreate(e, c->nhidden+BIAS);
  h = e = 0;
  for(i=0;i<c->ntests;i++) {
    if(c->certainty[i] == 0) continue;
    if(c->certainty[i] > 0) c->hidden_pos[h++] = i;
    else c->hidden_neg[e++] = i;
  }
  for(i=e;i < c->nhidden;i++) c->hidden_neg[i] = c->hidden_neg[RandInt(0,e-1)];
  for(i=h;i < c->nhidden;i++) c->hidden_pos[i] = c->hidden_pos[RandInt(0,h-1)];
  OptimizeWeights(c, desired, hidden, distance);
#endif
  MatrixFree(hidden);

#if DEBUG
  fprintf(stderr, "min_err: %g\n", min_err);
  for(h=0;h<c->nhidden;h++) {
    fprintf(stderr, "hidden_pos[%d] = %d\nhidden_neg[%d] = %d\n", 
	    h, c->hidden_pos[h], h, c->hidden_neg[h]);
    fprintf(stderr, "hidden_weight[%d] = %g\n", h, c->hidden_weight->d[h]);
  }
#if BIAS
  fprintf(stderr, "hidden_weight[%d] = %g\n", h, c->hidden_weight->d[h]);
#endif
#endif

  for(i=0;i<desired->len;i++) Boa_Release(distance[i]);
  free(distance);
  VectorFree(desired);
}

/* index is a List of (feature, key) tuples.
 * returns a float certainty factor.
 */
static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Classifier c = (Classifier)self->value;
  List pairs;
  Obj pair;
  float weight = 0;
  int i, h;

  if(index->type != T_List) return TPM_BOTTOM;
  if(!c->updated) {
    OptimizeHidden(c);
    c->updated = 1;
  }
 
  pairs = Boa_List_Value(index);
  {ListIter(p, pair, pairs) {
    Obj feature = Boa_Tuple_Item(pair, 0);
    Obj t;
    Obj key = Boa_Tuple_Item(pair, 1);
    assert(key->type == T_Int);
    i = Boa_Int_Value(key);
    /* was this an example? */
    if(c->certainty && c->certainty[i]) {
#if ZERO_MEAN
      weight = c->certainty[i];
#else
      weight = (c->certainty[i]+1)/2;
#endif
      break;
    }
    /* get a vector of distances */
    if(Boa_ObjGet(feature, key, 0, NULL, &t) != TPM_OK) {
      ErrorPrint(stderr);
    }
    else {
      Vector dists = Boa_Vector_Value(t);
      /* loop hidden units */
      for(h=0;h<c->nhidden;h++) {
	double a;
#if CROSS_TERM
	Obj v;
	Vector hdists;
	key = Boa_Int_Create(c->hidden_pos[h]);
	assert(Boa_ObjGet(feature, key, 0, NULL, &v) == TPM_OK);
	Boa_Release(key);
	hdists = Boa_Vector_Value(v);
	a = Butterfly(1 - 2*dists->d[c->hidden_neg[h]], 
		      1 - 2*dists->d[c->hidden_pos[h]],
		      1 - 2*hdists->d[c->hidden_neg[h]]);
	Boa_Release(v);
#else
	a = Butterfly(1 - 2*dists->d[c->hidden_neg[h]], 
		      1 - 2*dists->d[c->hidden_pos[h]],
		      0);
#endif
	weight += a*c->hidden_weight->d[h];
#if DEBUG > 2
	fprintf(stderr, "neg dist = %g\npos dist = %g\n",
		dists->d[c->hidden_neg[h]], dists->d[c->hidden_pos[h]]);
	fprintf(stderr, "ratio: %g\nweight: %g\n", ratio, weight);
#endif
      }
#if BIAS
      weight += c->hidden_weight->d[h];
#endif
      Boa_Release(t);
    }
    break;
  }}

#if !ZERO_MEAN
  weight = 2*weight - 1;
#endif
  weight += OFFSET;
  /* clip */
  if(weight > 1) weight = 1;
  else if(weight < -1) weight = -1;

  BOA_TYPECHECK_GET(T_Float);
  if(value_return) *value_return = Boa_Float_Create(weight);
  return TPM_OK;
}

/* index is a List of (feature, key) tuples.
 * value is a float certainty factor.
 */
static int ObjSet(Obj self, Obj index, int flags, Obj value)
{
  Classifier c = (Classifier)self->value;
  float certainty;
  List pairs;
  Obj pair;
  if(index->type != T_List) return TPM_BOTTOM;
  pairs = Boa_List_Value(index);
  if(value->type != T_Float) return TPM_BOTTOM;
  certainty = Boa_Float_Value(value);

  {ListIter(p, pair, pairs) {
    Obj feature, t;
    int i, status;

    assert(Boa_Tuple_Length(pair) == 2);
    feature = Boa_Tuple_Item(pair, 0);

    if(Boa_ObjGet(c->features, feature, 0, NULL, &t) != TPM_OK) {
      Boa_ObjSet(c->features, feature, BOA_INSERT, Boa_Int_Create(0));
    }
    else Boa_Release(t);

    i = FeatureKeyIndex(feature, Boa_Tuple_Item(pair, 1));
    if(!c->certainty) {
      c->ntests = FeatureNumKeys(feature);
      c->certainty = AllocateC(c->ntests, float);
    }
    c->certainty[i] = certainty;
    break;
  }}
  c->updated = 0;

  return TPM_OK;
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  Generic_SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Err_StringGet, Err_StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_Classifier_Butterfly_Define(void)
{
  Boa_Vector_Pkg();
  T_Classifier_Butterfly = Boa_SimpleType_Create("classifier_butterfly", 
						 &methods, 
						 sizeof(struct Classifier), 
						 sizeof(Obj));
  return T_Classifier_Butterfly;
}
