/* Decorates a binary classifier with a metric learner.
 * Metric is separable and independent.
 */
#include <boa/boa.h>
#include <boa/vector.h>
#include "boa_learn.h"

#define DEBUG 0
#define DIMS 2

typedef struct Classifier {
  Obj subject;    /* T_Classifier */
  Obj proximity;  /* T_Proximity */
  /* metric is owned by proximity */
  Matrix metric;
  /* mean is the mean certainty for each coordinate */
  /* count is the number of occurrences of each coordinate */
  Vector count[DIMS], mean[DIMS];
  int valid;   /* true if the metric has been updated */
} *Classifier;

/* Globals *******************************************************************/
Type T_Classifier_CDM;

/* Functions *****************************************************************/

static void Init(Obj self)
{
  Classifier c = (Classifier)self->value;
#if 0
  c->subject = Boa_Create(T_Classifier_Nearest);
#else
  c->subject = Boa_Create(T_Classifier_KR);
#endif
  c->proximity = Boa_Create(T_Proximity);
}

static void Final(Obj self)
{
  Classifier c = (Classifier)self->value;
  if(c->count[0]) {
    int i;
    for(i=0;i<DIMS;i++) {
      VectorFree(c->count[i]);
      VectorFree(c->mean[i]);
    }
  }
  Boa_Release(c->subject);
  Boa_Release(c->proximity);
}

static int FeatureKeyIndex(Obj feature, Obj key) 
{
  Obj tests;
  int i;
  /* assume feature is a metric object, and key is a query */
  if(Boa_StringGet(feature, "_context/test", '/', NULL, &tests) == TPM_OK) {
    assert(Boa_ObjGetInt(tests, key, &i) == TPM_OK);
    Boa_Release(tests);
    return i;
  }
  else if(key->type == T_Int) {
    return Boa_Int_Value(key);
  } 
  fprintf(stderr, "FeatureKeyIndex: don't know how to index ");
  Boa_PrintNL(stderr, key);
  return -1;
}

static int FeatureNumKeys(Obj feature)
{
  Obj tests;
  int i;
  /* assume feature is a metric object, and key is a query */
  if(Boa_StringGet(feature, "_context/test", '/', NULL, &tests) == TPM_OK) {
    assert(Boa_StringGetInt(tests, "_length", &i) == TPM_OK);
    Boa_Release(tests);
  }
  else assert(Boa_StringGetInt(feature, "_length", &i) == TPM_OK);
  return i;
}

static Obj NewExample(Classifier c, Obj ex)
{
  Obj new_ex = Boa_Create(T_List);
  Obj tuple, key;

  /* use the key from the supplied example to create a new example */
  assert(Boa_IntGet(ex, 0, 0, NULL, &tuple) == TPM_OK);
#if 0
  key = Boa_Int_Create(FeatureKeyIndex(Boa_Tuple_Item(tuple, 0), 
				       Boa_Tuple_Item(tuple, 1)));
#else
  key = Boa_Hold(Boa_Tuple_Item(tuple, 1));
#endif
  Boa_Release(tuple);

  tuple = Boa_Create(Boa_Type_Create("(tuple 2)"));
  Boa_Tuple_Item(tuple, 0) = Boa_Hold(c->proximity);
  Boa_Tuple_Item(tuple, 1) = key;
  assert(Boa_IntSet(new_ex, 0, BOA_INSERT, tuple) == TPM_OK);
  return new_ex;
}

static void KeyToCoords(Classifier c, int key, int coords[DIMS])
{
  int i;
  /* compute the coordinates of the example */
  for(i=0;i<DIMS;i++) {
    coords[i] = key % c->count[i]->len;
    key /= c->count[i]->len;
  }
}

/* index is a List of (feature, key) tuples.
 * returns a float certainty factor.
 */
static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Classifier c = (Classifier)self->value;
  int status;
  Obj new_index;

  if(!c->metric) {
    if(value_return) *value_return = Boa_Float_Create(0);
    return TPM_OK;
  }

  if(!c->valid) {
    /* compute new metric */
    int i,j,d;
    int x[DIMS], y[DIMS];
#if DEBUG
    fprintf(stderr, "computing new metric\n");
#endif
    for(x[1]=0;x[1]<c->count[1]->len;x[1]++) {
      i = x[1] * c->count[0]->len;
      for(x[0]=0;x[0]<c->count[0]->len;x[0]++) {
	for(y[1]=0;y[1]<c->count[1]->len;y[1]++) {
	  j = y[1] * c->count[0]->len;
	  for(y[0]=0;y[0]<c->count[0]->len;y[0]++) {
	    c->metric->d2[i][j] = 0;
	    for(d=0;d<DIMS;d++) {
	      double xm = c->mean[d]->d[x[d]];
	      double ym = c->mean[d]->d[y[d]];
#if 1
	      /* CDM */
	      c->metric->d2[i][j] += (1 - xm*ym)/2/DIMS;
#else
	      /* MVDM */
	      c->metric->d2[i][j] += abs(xm - ym)/2/DIMS;
#endif
	    }
#if 0
	    fprintf(stderr, "(%d %d) = %g\n", i, j, c->metric->d2[i][j]);
#endif
	    j++;
	  }
	}
	i++;
      }
    }
#if DEBUG
    fprintf(stderr, "done\n");
#endif
    c->valid = 1;
  }

  new_index = NewExample(c, index);
  status = Boa_ObjGet(c->subject, new_index, flags, type_in_out, value_return);
  Boa_Release(new_index);
  return status;
}

/* index is a List of (feature, key) tuples.
 * value is a float certainty factor.
 */
static int ObjSet(Obj self, Obj index, int flags, Obj value)
{
  Classifier c = (Classifier)self->value;
  Obj tuple, feature, key, new_index;
  int pos, x[DIMS], i, j, status;
  float certainty;

  if(index->type != T_List) return TPM_BOTTOM;
  if(value->type != T_Float) return TPM_BOTTOM;
  certainty = Boa_Float_Value(value);

  assert(Boa_IntGet(index, 0, 0, NULL, &tuple) == TPM_OK);
  assert(Boa_Tuple_Length(tuple) == 2);
  feature = Boa_Tuple_Item(tuple, 0);
  key = Boa_Tuple_Item(tuple, 1);
  pos = FeatureKeyIndex(feature, key);

  if(!c->metric) {
    int ntests = FeatureNumKeys(feature);
    int side = isqrt(ntests);
    for(i=0;i<DIMS;i++) {
      c->count[i] = VectorCreate(side);
      VectorSet(c->count[i], 0);
      c->mean[i] = VectorCreate(side);
      VectorSet(c->mean[i], 0);
    }
    c->metric = MatrixCreate(ntests, ntests);
    /* give the metric to proximity */
    assert(Boa_StringSet(c->proximity, "_matrix", BOA_INSERT|BOA_REPLACE,
			 Boa_Matrix_Create(c->metric)) == TPM_OK);
  }

  KeyToCoords(c, pos, x);
  /* update the count and mean matrices */
  c->valid = 0;
  for(i=0;i<DIMS;i++) {
    c->mean[i]->d[x[i]] *= c->count[i]->d[x[i]];
    c->mean[i]->d[x[i]] += certainty;
    c->count[i]->d[x[i]]++;
    c->mean[i]->d[x[i]] /= c->count[i]->d[x[i]];
#if DEBUG
    for(j=0;j<c->mean[i]->len;j++) {
      fprintf(stderr, "%lg ", c->mean[i]->d[j]);
    }
    fprintf(stderr, "\n");
#endif
  }

  new_index = NewExample(c, index);
  status = Boa_ObjSet(c->subject, new_index, flags, value);
  Boa_Release(new_index);
  return status;
}

static struct Boa_Methods methods = {
  Init, Final, Empty_Clean,
  Generic_SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Err_StringGet, Err_StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_Classifier_CDM_Define(void)
{
  Boa_Vector_Pkg();
  T_Classifier_CDM = Boa_SimpleType_Create("classifier_CDM", &methods, 
					   sizeof(struct Classifier), 
					   sizeof(Obj));
  return T_Classifier_CDM;
}
