/* Converts an index into a vector of distances to all leaves in a hierarchy.
 *
 * Attributes
 *   hier
 */
#include <boa/boa.h>
#include <boa/vector.h>
#include <tpm/hier.h>

typedef struct Cophenetic {
  Obj hier;
  int *height; /* height of each hier node */
  int len;
  int *index; /* maps [0..len-1] -> hier node */
} *Cophenetic;

/* Globals *******************************************************************/
Type T_Cophenetic;

/* Functions *****************************************************************/

static void Final(Obj self)
{
  Cophenetic c = (Cophenetic)self->value;
  if(c->hier) Boa_Release(c->hier);
  if(c->index) free(c->index);
  if(c->height) free(c->height);
}

static int StringGet(Obj self, char *name, int flags,
		     Type *type_in_out, Obj *value_return)
{
  Cophenetic c = (Cophenetic)self->value;
  if(name[0] == '_') {
    if(!strcmp(name+1, "length")) {
      BOA_TYPECHECK_GET(T_Int);
      if(value_return) *value_return = Boa_Int_Create(c->len);
      return TPM_OK;
    }
  }
  return TPM_BOTTOM;
}

static int *HierHeights(Hier hier)
{
  int *height = AllocateC(hier->num_nodes, int);
  int i;
  for(i=hier->num_nodes-1;i;i--) {
    int p = HierNodeAt(hier, i)->parent;
    if(p < 0) continue;
    if(height[p] < height[i]+1) height[p] = height[i]+1;
  }
  return height;
}

/* Returns the maximum value of a leaf */
static int HierMaxLeaf(Hier hier)
{
  int i, max = 0;
  for(i=0;i<hier->num_nodes;i++) {
    HierNode node = HierNodeAt(hier, i);
    int v = *(int*)&node->data;
    if(node->children) continue;
    if(v > max) max = v;
  }
  return max;
}

static int *HierIndex(Hier hier, int len)
{
  int *index = Allocate(len, int);
  int i;
  for(i=0;i<hier->num_nodes;i++) {
    HierNode node = HierNodeAt(hier, i);
    int v = *(int*)&node->data;
    if(node->children) continue;
    index[v] = i;
  }
  return index;
}

static int StringSet(Obj self, char *name, int flags, Obj value)
{
  Cophenetic c = (Cophenetic)self->value;
  if(name[0] == '_') {
    if(!strcmp(name+1, "hier")) {
      if(value->type->type != T_Hier) return TPM_ERROR;
      /* assume leaves are integers */
      c->hier = value;
      if(c->height) free(c->height);
      c->height = HierHeights(Boa_Hier_Value(c->hier));
      c->len = HierMaxLeaf(Boa_Hier_Value(c->hier))+1;
      if(c->index) free(c->index);
      c->index = HierIndex(Boa_Hier_Value(c->hier), c->len);
      return TPM_OK;
    }
  }
  return TPM_BOTTOM;
}

static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Cophenetic c = (Cophenetic)self->value;
  if(index->type != T_Int) return TPM_BOTTOM;
  BOA_TYPECHECK_GET(T_Vector);
  if(value_return) {
    Hier hier = Boa_Hier_Value(c->hier);
    int i, j, left, right;
    Vector dists = VectorCreate(c->len);
    i = Boa_Int_Value(index);
    dists->d[i] = 0;
    left = right = i = c->index[i];
    /* climb the tree and assign distances to leaves */
    for(;;) {
      int limit;
      i = HierNodeAt(hier, i)->parent;
      if(i < 0) break;
      limit = HierRightLeaf(hier, i);
      for(j=i+1;j<=limit;j++) {
	HierNode node;
	int v;
	if((j >= left) && (j <= right)) continue;
	node = HierNodeAt(hier, j);
	if(node->children) continue;
	v = *(int*)&node->data;
	dists->d[v] = c->height[i];
      }
      left = i;
      right = limit;
    }
    /* check for nodes under diff't roots */
    for(j=0;j<hier->num_nodes;j++) {
      HierNode node;
      int v;
      if((j >= left) && (j <= right)) continue;
      node = HierNodeAt(hier, j);
      if(node->children) continue;
      v = *(int*)&node->data;
      dists->d[v] = c->height[0]+1;
    }
    *value_return = Boa_Vector_Create(dists);
  }
  return TPM_OK;
}

static struct Boa_Methods methods = {
  Empty_Init, Final, Empty_Clean,
  Generic_SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  StringGet, StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, Generic_ObjSet
};

Obj Boa_Cophenetic_Define(void)
{
  Boa_Vector_Pkg();
  Boa_Hier_Pkg();
  T_Cophenetic = Boa_SimpleType_Create("cophenetic", &methods,
				       sizeof(struct Cophenetic), sizeof(int));
  return T_Cophenetic;
}
