/* Binary classifier.
 * SetObj with a sample and a number in [-1, 1] trains the classifier.
 * GetObj with a sample returns a number in [-1, 1], where zero means 
 * "no opinion".
 *
 * A sample is a List of (feature, key) pairs, where
 * key is an index into the feature object.
 * E.g. if feature is a Hier then key is a node number.
 *
 * Uses an IdDict to store the features.
 * For each feature, there is associated information for performing
 * classification.  For example, pos/neg counts for each node of a Hier.
 *
 * This classifier picks the best covers from a hier or metric.
 * It is almost identical to SC.
 */
#include <tpm/list.h>
#include <boa/boa.h>
#include <boa/vector.h>
#include <tpm/hier.h>
#include <tpm/bitset.h>

#define DEBUG 0
#define SIZE_BIAS 0
#define RANDOM 1
/* 0 is DT, 1 means pure pos or neg (DL), 2 means pure pos (SC) */
/* 1 is not implemented yet */
#define PURE 0
/* average best instance covers together (i.e. don't pick best) */
#define AVERAGE 0
/* let some tree nodes be prohibited */
#define PROHIBIT 0

typedef enum { HIER_COVER, METRIC_COVER } CoverType;

typedef struct HierCover {
  CoverType type;
  float weight;        /* strength of this cover */
  Obj feature;         /* attribute we're covering */
  unsigned low, high;  /* range of leaves covered */
} *HierCover;

typedef struct MetricCover {
  CoverType type;
  float weight;        /* strength of this cover */
  Obj feature;         /* attribute we're covering */
  BitSet mask;
} *MetricCover;

typedef struct Cover {
  CoverType type;
  float weight;        /* strength of this cover */
} *Cover;

typedef struct Classifier {
  Obj features;
  Obj examples;    /* List of Instance */
  float total_pos, total_neg;  /* sum of positive and negative certainties */
  List covers;
} *Classifier;

/* Globals *******************************************************************/
Type T_Classifier_Cover;

/* Functions *****************************************************************/

static void CoverFree(Cover c)
{
  switch(c->type) {
  case METRIC_COVER: BitSetFree(((MetricCover)c)->mask);
  case HIER_COVER: /* do nothing */;
  }
  /* feature is not held by the cover */
  free(c);
}

static Cover CoverCopy(Cover c)
{
  switch(c->type) {
  case HIER_COVER: return (Cover)AllocCopy(1, struct HierCover, c);
  case METRIC_COVER: 
    {MetricCover m = AllocCopy(1, struct MetricCover, c);
     m->mask = BitSetCopy(m->mask);
     return (Cover)m;}
  }
  return c;
}

static void Init(Obj self)
{
  Classifier c = (Classifier)self->value;
  c->features = Boa_Create(T_IdDict);
  c->examples = Boa_Create(T_List);
}

static void Final(Obj self)
{
  Classifier c = (Classifier)self->value;
  Boa_Release(c->features);
  Boa_Release(c->examples);
  if(c->covers) ListFree(c->covers);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Classifier c = (Classifier)self->value;
  func(c->features, data);
  func(c->examples, data);
}

/* Data for Hier features */
typedef struct TreeData {
  /* sum of positive certainties under this node */
  float pos;
  /* sum of negative certainties under this node */
  float neg;
  /* weight of the node */
  float bias;
} *TreeData;

/* Find the best covers for instance.
 * Modifies *best_score and best.
 */
static void BestCoversInstance(Classifier c, Obj instance,
			       float *best_score, List best)
{
  Obj pair;
  /* Loop features */
  {ListIter(p, pair, Boa_List_Value(instance)) {
    Obj feature, t;

    feature = Boa_Tuple_Item(pair, 0);
    if(Boa_ObjGet(c->features, feature, 0, NULL, &t) != TPM_OK) {
      /* haven't seen this feature before */
      continue;
    }

    if(feature->type->type == T_Hier) {
      Hier hier = Boa_Hier_Value(feature);
      TreeData data = (TreeData)t->value;
      int i = Boa_Int_Value(Boa_Tuple_Item(pair, 1));
      /* climb the tree, scoring the nodes */
      while(i >= 0) {
	HierNode node = HierNodeAt(hier, i);
#if PURE == 2
	float score = (data[i].neg < 0) ? -1.0/0 : data[i].pos;
#else
	/* if one of the denominators is 0, then the score is NaN, which is
         * what we want, since the cover is a no-op.
         */
	float rsum = data[i].pos + data[i].neg;
	float rsize = data[i].pos - data[i].neg;
	float rmean = rsum / rsize;
	float lsum = c->total_pos + c->total_neg - rsum;
	float lsize = c->total_pos - c->total_neg - rsize;
	float lmean = lsum / lsize;
	/* score is sample variance, modified to range 0 (worst) to 1 (best) */
	float score = 1 - (lsize * (1 + lmean) * (1 - lmean) + 
			   rsize * (1 + rmean) * (1 - rmean)) / 
			     (lsize + rsize);
#if PURE == 1
	/* one of the bins must be pure */
	if((data[i].pos > 0) && (data[i].neg < 0) &&
	   (c->total_pos > data[i].pos) && (c->total_neg < data[i].neg)) 
	  score = -1.0/0;
#endif
#endif
	score *= data[i].bias;
#if DEBUG
	fprintf(stderr, "%p %d sc: %g pos: %g neg: %g bias: %g\n",
		feature, i, score,
		data[i].pos, data[i].neg, data[i].bias);
#endif
	if(score > *best_score) {
	  *best_score = score;
	  ListFlush(best);
	}
	if(score == *best_score) {
	  HierCover c = Allocate(1, struct HierCover);
	  c->type = HIER_COVER;
	  c->feature = feature;  /* XXX not held */
	  c->low = i;
	  if(data[i].pos > -data[i].neg) c->weight = score;
	  else c->weight = -score;
	  ListAddFront(best, c);
	}
	i = node->parent;
      }
    }
    else {
      Vector dists;
      int i, ntests, radius;
      float score, pos, neg, weight;
      float *certainty = (float*)t->value;
      int *p; /* sorting permutation */

      ntests = Boa_Array_Length(t->type);
      Boa_Release(t);
      /* assume feature is a metric object, and value is query */
      /* assume features, test fields have been filled in */
      /* get a vector of distances */
      if(Boa_ObjGet(feature, Boa_Tuple_Item(pair, 1), 0, NULL, &t) != TPM_OK) {
	ErrorPrint(stderr);
	continue;
      }
      dists = Boa_Vector_Value(t);

      /* compute the sorting permutation */
      p = SortingPermutation(dists->d, dists->len, sizeof(double), DoubleCmp, 0);
#if PURE == 2
      /* find closest negative example */
      /* score is # pos */
      score = 0;
      for(radius=0;; radius = i) {
	float subtotal = 0;
	/* can't split between equal distances */
	for(i=radius; 
	    (i < ntests) && (dists->d[p[i]] == dists->d[p[radius]]); i++) {
	  if(certainty[p[i]] < 0) goto done;
	  subtotal += certainty[p[i]];
	}
	score += subtotal;
      }
    done:;
#if SIZE_BIAS
      s = (float)radius / ntests;
      s = s*0.8 + 0.1;
      score *= s * (1-s);
#endif
#else
      /* score is sample variance */
      pos = neg = 0;
      score = 0;
      radius = 0;
      weight = 0;
      for(i=0;i<ntests;) {
	float current;
	int j = i;
	/* can't split between equal distances */
	while((i < ntests) && (dists->d[p[i]] == dists->d[p[j]])) {
	  if(certainty[p[i]] < 0) neg += certainty[p[i]];
	  else pos += certainty[p[i]];
	  i++;
	}
#if PURE == 1
	/* one of the bins must be pure */
	if((pos > 0) && (neg < 0) && 
	   (c->total_pos > pos) && (c->total_neg < neg)) continue;
#endif
	current = pos * neg / (neg - pos);
	current += (c->total_pos - pos) * (c->total_neg - neg) /
	  (c->total_neg - neg - c->total_pos + pos);
	current = 1 - 4 * current / (c->total_pos - c->total_neg);
#if SIZE_BIAS
	s = (float)i / ntests;
	s = s*0.8 + 0.1;
	current *= s * (1-s);
#endif
	if(current > score) {
	  score = current;
	  radius = i;
	  weight = pos > -neg;
	}
      }
#endif
      if((score > 0) && (score > *best_score)) {
	*best_score = score;
	ListFlush(best);
      }
      if(score == *best_score) {
	MetricCover c = Allocate(1, struct MetricCover);
	c->type = METRIC_COVER;
	c->feature = feature;
	c->mask = BitSetCreate(ntests);
	for(i=0;i<radius;i++) BitSetInsert(c->mask, p[i]);
#if DEBUG
	fprintf(stderr, "%p sc: %g radius: %d\n", feature, score, radius);
#endif
	if(weight) c->weight = score;
	else c->weight = -score;
	ListAddFront(best, c);
      }
      free(p);
    }
    Boa_Release(t);
  }}
}

static Cover InitCover(Cover cover)
{
  if(cover->type == HIER_COVER) {
    HierCover h = (HierCover)cover;
    /* post-process the cover */
    h->high = HierRightLeaf(Boa_Hier_Value(h->feature), h->low);
#if DEBUG
    fprintf(stderr, "cover: %p %d-%d %g\n", 
	    h->feature, h->low, h->high, h->weight);
#endif
  }
  else {
#if DEBUG
    MetricCover m = (MetricCover)cover;
    fprintf(stderr, "cover: %p %d %g\n", 
	    m->feature, BitSetSize(m->mask), m->weight);
#endif
  }
  return cover;
}

/* Picks a best cover for the examples */
static Cover BestCover(Classifier c)
{
  Cover cover;
  Obj instance;
  int i;
  float best_score = -1.0/0;
  List best = ListCreate((FreeFunc*)CoverFree);
  
  /* loop positive examples */
  {ListIter(p, instance, Boa_List_Value(c->examples)) {
    /* find the best covers for the example */
    BestCoversInstance(c, instance, &best_score, best);
  }}

  /* no covers? */
  if(best_score == -1.0/0) {
    /* make a dummy cover */
    HierCover c = Allocate(1, struct HierCover);
    c->type = HIER_COVER;
    c->high = -1;
#if DEBUG
    fprintf(stderr, "no covers\n");
#endif
    ListFree(best);
    return (Cover)c;
  }

#if RANDOM
  i = RandInt(0, ListSize(best)-1);
#else
  i = ListSize(best)-1;
#endif
  /* make a copy so ListFree(best) doesn't clobber it */
  cover = CoverCopy((Cover)ListValueAtIndex(best, i));
  ListFree(best);
  cover->weight /= fabs(cover->weight);
  return InitCover(cover);
}

/* Collects all best covers for the examples */
static List BestCovers(Classifier c)
{
  Cover cover;
  Obj instance;
  float best_score = -1.0/0, total;
  List best = ListCreate((FreeFunc*)CoverFree);
  
  /* loop positive examples */
  {ListIter(p, instance, Boa_List_Value(c->examples)) {
#if AVERAGE
    float this_best_score = -1.0/0;
    List this_best = ListCreate((FreeFunc*)CoverFree);
    /* find the best covers for the example */
    BestCoversInstance(c, instance, &this_best_score, this_best);
    if(this_best_score > best_score) best_score = this_best_score;
    ListAppend(best, this_best);
#else
    /* find the best covers for the example */
    BestCoversInstance(c, instance, &best_score, best);
#endif
  }}

  /* no covers? */
  if(best_score == -1.0/0) {
    /* make a dummy cover */
    HierCover c = Allocate(1, struct HierCover);
    c->type = HIER_COVER;
    c->high = -1;
#if DEBUG
    fprintf(stderr, "no covers\n");
#endif
    ListAddFront(best, c);
  }

#if DEBUG
  fprintf(stderr, "%d covers\n", ListSize(best));
#endif
  /* compute the total weight */
  total = 0;
  {ListIter(p, cover, best) {
    total += fabs(cover->weight);
  }}
  {ListIter(p, cover, best) {
    cover->weight /= total;
    InitCover(cover);
  }}
  return best;
}

static int FeatureKeyIndex(Obj feature, Obj key) 
{
  Obj tests;
  int i;
  /* assume feature is a metric object, and key is a query */
  if(Boa_StringGet(feature, "_context/test", '/', NULL, &tests) == TPM_OK) {
    assert(Boa_ObjGetInt(tests, key, &i) == TPM_OK);
    Boa_Release(tests);
    return i;
  }
  else if(key->type == T_Int) {
    return Boa_Int_Value(key);
  } 
  fprintf(stderr, "FeatureKeyIndex: don't know how to index ");
  Boa_PrintNL(stderr, key);
  return -1;
}

static int FeatureNumKeys(Obj feature)
{
  Obj tests;
  int i;
  /* assume feature is a metric object, and key is a query */
  if(Boa_StringGet(feature, "_context/test", '/', NULL, &tests) == TPM_OK) {
    assert(Boa_StringGetInt(tests, "_length", &i) == TPM_OK);
    Boa_Release(tests);
  }
  else assert(Boa_StringGetInt(feature, "_length", &i) == TPM_OK);
  return i;
}

/* index is a List of (feature, key) tuples.
 * returns a float certainty factor.
 */
static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Classifier c = (Classifier)self->value;
  List pairs;
  Obj pair;
  float weight = 0;
  int i;
  Cover cover;

  if(index->type != T_List) return TPM_BOTTOM;
  pairs = Boa_List_Value(index);

  if(!c->covers) {
    if(AVERAGE) c->covers = BestCovers(c);
    else {
      c->covers = ListCreate((FreeFunc*)CoverFree);
      ListAddFront(c->covers, BestCover(c));
    }
  }

  {ListIter(p, cover, c->covers) {
    switch(cover->type) {
    case HIER_COVER:
      {HierCover h = (HierCover)cover;
       /* find the value for c->feature */
       ListIter(p, pair, pairs) {
	 if(Boa_Tuple_Item(pair, 0) != h->feature) continue;
	 i = Boa_Int_Value(Boa_Tuple_Item(pair, 1));
	 if((h->low > i) || (h->high < i)) weight += -h->weight;
	 else weight += h->weight;
	 break;
       }}
      break;
    case METRIC_COVER:
      {MetricCover m = (MetricCover)cover;
       /* find the value for c->feature */
       ListIter(p, pair, pairs) {
	 if(Boa_Tuple_Item(pair, 0) != m->feature) continue;
	 /* find the index of this key and probe the mask */
	 i = FeatureKeyIndex(m->feature, Boa_Tuple_Item(pair, 1));
	 if(!BitSetMember(m->mask, i)) weight += -m->weight;
	 else weight += m->weight;
	 break;
       }}
      break;
    }
  }}

#if 1
  /* force an opinion */
  if(weight == 0) weight = -1;
#endif

  BOA_TYPECHECK_GET(T_Float);
  if(value_return) *value_return = Boa_Float_Create(weight);
  return TPM_OK;
}

/* index is a List of (feature, key) tuples.
 * value is a float certainty factor.
 */
static int ObjSet(Obj self, Obj index, int flags, Obj value)
{
  Classifier c = (Classifier)self->value;
  float certainty;
  List pairs;
  Obj pair;
  if(index->type != T_List) return TPM_BOTTOM;
  pairs = Boa_List_Value(index);
  if(value->type != T_Float) return TPM_BOTTOM;
  certainty = Boa_Float_Value(value);
  /* assumes we won't get an error */
  Boa_Release(value);

  /* add this instance to the list of examples */
  ListAddFront(Boa_List_Value(c->examples), Boa_Hold(index));
  if(certainty > 0) c->total_pos += certainty;
  else c->total_neg += certainty;

  /* destroy any existing covers */
  if(c->covers) {
    ListFree(c->covers);
    c->covers = NULL;
  }

  {ListIter(p, pair, pairs) {
    Obj feature, t;
    int status, i;

    assert(Boa_Tuple_Length(pair) == 2);
    feature = Boa_Tuple_Item(pair, 0);
    status = Boa_ObjGet(c->features, feature, 0, NULL, &t);

    if(feature->type->type == T_Hier) {
      TreeData data;
      Hier hier = Boa_Hier_Value(feature);

      if(status != TPM_OK) {
	int j;
	char str[64];
	/* haven't seen this feature before */
	/* create an array of TreeData for this feature */
	sprintf(str, "(array %d (array 3 float))", hier->num_nodes);
	t = Boa_Create(Boa_Type_Create(str));
	data = (TreeData)t->value;
	/* initialize the treeData */
	for(j=0;j<hier->num_nodes;j++) { 
	  data[j].pos = 0; data[j].neg = 0; data[j].bias = 1;
	}
#if SIZE_BIAS
	for(j=hier->num_nodes;j>0;) { 
	  HierNode node = HierNodeAt(hier, --j);
	  if(node->parent >= 0) data[node->parent].bias += data[j].bias;
	}
	/* normalize the bias */
	for(j=0;j<hier->num_nodes;j++) { 
	  data[j].bias /= hier->num_nodes;
	  data[j].bias = data[j].bias*0.8 + 0.1;
	  data[j].bias *= 1 - data[j].bias;
	}
#endif
#if PROHIBIT
	for(j=0;j<hier->num_nodes;j++) { 
	  HierNode node = HierNodeAt(hier, j);
	  if(!node->children || (*(int*)&node->data == NAI)) data[j].bias = 0;
	}
#endif
	Boa_ObjSet(c->features, feature, BOA_INSERT, Boa_Hold(t));
      }
      else data = (TreeData)t->value;

      /* climb the tree, updating data fields */
      i = Boa_Int_Value(Boa_Tuple_Item(pair, 1));
      while(i >= 0) {
	HierNode node = HierNodeAt(hier, i);
	if(certainty > 0) data[i].pos += certainty;
	else data[i].neg += certainty;
	i = node->parent;
      }
    }
    else {
      float *data;
      i = FeatureKeyIndex(feature, Boa_Tuple_Item(pair, 1));
      if(status != TPM_OK) {
	/* haven't seen this feature before */
	/* create an array of certainties for this feature (brute force) */
	int ntests = FeatureNumKeys(feature);
	char str[64];
	sprintf(str, "(array %d float)", ntests);
	t = Boa_Create(Boa_Type_Create(str));
	data = (float*)t->value;
	memset(data, 0, ntests*sizeof(float));
	Boa_ObjSet(c->features, feature, BOA_INSERT, Boa_Hold(t));
      }
      else data = (float*)t->value;
      data[i] = certainty;
    }
    Boa_Release(t);
  }} 
  return TPM_OK;
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  Generic_SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Err_StringGet, Err_StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_Classifier_Cover_Define(void)
{
  Boa_Vector_Pkg();
  T_Classifier_Cover = Boa_SimpleType_Create("classifier_cover", &methods, 
					     sizeof(struct Classifier), 
					     sizeof(Obj));
  return T_Classifier_Cover;
}
