/* Converts an index into a vector of distances to all points in a discrete
 * Euclidean space.
 * Attributes
 *   xlen    Number of values along x dimension
 *   ylen    Number of values along y dimension
 */
#include <boa/boa.h>
#include <boa/vector.h>

#define SCALE 5
#define PROB 1

typedef struct Euclid {
  int xlen, ylen;
} *Euclid;

/* Globals *******************************************************************/
Type T_Euclid;

/* Functions *****************************************************************/

static int StringGet(Obj self, char *name, int flags,
		     Type *type_in_out, Obj *value_return)
{
  Euclid e = (Euclid)self->value;
  if(name[0] == '_') {
    if(!strcmp(name+1, "length")) {
      BOA_TYPECHECK_GET(T_Int);
      if(value_return) *value_return = Boa_Int_Create(e->xlen * e->ylen);
      return TPM_OK;
    }
  }
  return TPM_BOTTOM;
}

static int StringSet(Obj self, char *name, int flags, Obj value)
{
  Euclid e = (Euclid)self->value;
  if(name[0] == '_') {
    if(!strcmp(name+1, "xlen")) {
      if(value->type != T_Int) return TPM_ERROR;
      e->xlen = Boa_Int_Value(value);
      Boa_Release(value);
      return TPM_OK;
    }
    else if(!strcmp(name+1, "ylen")) {
      if(value->type != T_Int) return TPM_ERROR;
      e->ylen = Boa_Int_Value(value);
      Boa_Release(value);
      return TPM_OK;
    }
  }
  return TPM_BOTTOM;
}

static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Euclid e = (Euclid)self->value;
  /* must be an integer */
  if(index->type != T_Int) return TPM_BOTTOM;
  BOA_TYPECHECK_GET(T_Vector);
  if(value_return) {
    Vector result = VectorCreate(e->xlen * e->ylen);
    int iy = Boa_Int_Value(index) / e->xlen;
    int ix = Boa_Int_Value(index) % e->xlen;
    int i = 0;
    int y,x;
    for(y=0;y<e->ylen;y++) {
#if PROB
      double dy = (double)(y - iy)/e->ylen*SCALE;
#else
      double dy = y - iy;
#endif
      for(x=0;x<e->xlen;x++) {
#if PROB
	double dx = (double)(x - ix)/e->xlen*SCALE;
	result->d[i++] = (1 - exp(-(dx*dx + dy*dy)))/2;
#else
	double dx = x - ix;
	result->d[i++] = dx*dx + dy*dy;
#endif
      }
    }
    *value_return = Boa_Vector_Create(result);
  }
  return TPM_OK;
}

static struct Boa_Methods methods = {
  Empty_Init, Empty_Final, Empty_Clean,
  Generic_SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  StringGet, StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, Generic_ObjSet
};

Obj Boa_Euclid_Define(void)
{
  Boa_Vector_Pkg();
  T_Euclid = Boa_SimpleType_Create("euclid", &methods,
				   sizeof(struct Euclid), sizeof(int));
  return T_Euclid;
}
