/* Converts Objs into Lists of (attr, value) pairs to give to Classifier.
 * Attributes
 *   features    Ord of Hiers
 *   compare     Comparison function between leaf data and index 
 *               (defaults to IntCmp)
 */
#include <tpm/list.h>
#include <boa/boa.h>
#include <tpm/hier.h>

#define VERBOSE 0

typedef struct Map {
  Obj features;  /* Ord */
  CmpFunc *compare;
} *Map;

/* Globals *******************************************************************/
Type T_FeatureMap;

/* Functions *****************************************************************/

static void Init(Obj self)
{
  Map m = (Map)self->value;
  m->features = Boa_Create(T_Ord);
  m->compare = IntCmp;
}

static void Final(Obj self)
{
  Map m = (Map)self->value;
  Boa_Release(m->features);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Map m = (Map)self->value;
  func(m->features, data);
}

static int SPrint(Obj self, StrStream ss)
{
  Map m = (Map)self->value;
  StrStreamPuts(ss, "featureMap");
  return Boa_SPrint(m->features, ss);
}

static int StringGet(Obj self, char *name, int flags,
		     Type *type_in_out, Obj *value_return)
{
  Map m = (Map)self->value;
  if(flags & BOA_SPLIT) 
    return Generic_SplitGet(self, name, flags, 
			    type_in_out, value_return);
  if(name[0] == '_') {
    if(!strcmp(name+1, "features")) {
      BOA_TYPECHECK_GET(m->features);
      if(value_return) {
	Boa_Hold(m->features);
	*value_return = m->features;
      }
      return TPM_OK;
    }
    else if(!strcmp(name+1, "compare")) {
      BOA_TYPECHECK_GET(T_Ptr);
      if(value_return) *value_return = Boa_Ptr_Create((void*)m->compare);
      return TPM_OK;
    }
  }
  return Generic_StringGet(self, name, flags, type_in_out, value_return);
}

static int StringSet(Obj self, char *name, int flags, Obj value)
{
  Map m = (Map)self->value;
  if(name[0] == '_') {
    if(!strcmp(name+1, "features")) {
      if(value->type != T_Ord) return TPM_ERROR;
      Boa_Release(m->features);
      m->features = value;
      return TPM_OK;
    }
    else if(!strcmp(name+1, "compare")) {
      if(value->type != T_Ptr) return TPM_BOTTOM;
      m->compare = (CmpFunc*)Boa_Ptr_Value(value);
      Boa_Release(value);
      return TPM_OK;
    }
  }
  return TPM_BOTTOM;
}

/* Loop the leaves of all trees, looking for a match with index.
 */
static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Map m = (Map)self->value;
  int f, i;
  Type pairType = Boa_Type_Create("(tuple 2)");
  List pairs;
  *value_return = Boa_Create(T_List);
  pairs = Boa_List_Value(*value_return);
  for(f=0;f<Boa_Ord_Length(m->features);f++) {
    Obj feature = Boa_Ord_Item(m->features, f);
    if(feature->type->type == T_Hier) {
      Hier hier = Boa_Hier_Value(feature);
      for(i=0;i<hier->num_nodes;i++) {
	HierNode node = HierNodeAt(hier, i);
	if(node->children) continue;
	if(!m->compare(&node->data, index->value)) {
	  Obj pair = Boa_Create(pairType);
	  Boa_Tuple_Item(pair, 0) = Boa_Hold(feature);
	  Boa_Tuple_Item(pair, 1) = Boa_Int_Create(i);
	  ListAddFront(pairs, pair);
#if VERBOSE
	  fprintf(stderr, "featureMap: ");
	  Boa_Print(stderr, index);
	  fprintf(stderr, " -> %d\n", i);
#endif
	}
      }
    }
    else {
#if 1
      /* just pair the feature with the index */
      Obj pair = Boa_Create(pairType);
      Boa_Tuple_Item(pair, 0) = Boa_Hold(feature);
      /* must stabilize because index may be temporary for this call */
      Boa_Tuple_Item(pair, 1) = Boa_Stabilize(index);
      ListAddFront(pairs, pair);
#else
      fprintf(stderr, "FeatureMap: unknown feature type ");
      Boa_PrintNL(stderr, feature->type);
#endif
    }
  }
  return TPM_OK;
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  StringGet, StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, Generic_ObjSet
};

Obj Boa_FeatureMap_Define(void)
{
  T_FeatureMap = Boa_SimpleType_Create("featureMap", &methods,
				       sizeof(struct Map), sizeof(Ptr));
  return T_FeatureMap;
}
