/* Boa adapter for the Hier data type.
 * A Hier is a compact representation of a tree with equal-sized data
 * at each node.
 * Requires: Int, ParamType.
 */
#include <boa/boa.h>
#include "hier.h"

#define length Boa_Int_Value(Boa_PType_Item(self->type, 0))
#define subtype Boa_PType_Item(self->type, 1)

/* Globals *******************************************************************/
Type T_Hier = NULL;

/* Functions *****************************************************************/

static void Init(Obj self)
{
  Hier hier = (Hier)self->value;
  hier->num_nodes = length;
  hier->data_size = TypeSize(subtype);
  /* may have to use padding to make the nodes aligned */
  if(hier->data_size % sizeof(int)) 
    hier->pad = sizeof(int) - (hier->data_size % sizeof(int));
  else 
    hier->pad = 0;
  hier->node = Allocate(hier->num_nodes * HierNodeSize(hier), char);
}

static void Final(Obj self)
{
  Hier hier = (Hier)self->value;
  free(hier->node);
}

static int SPrint(Obj self, StrStream ss)
{
  return Boa_SPrint(self->type, ss);
}

static int IntGet(Obj self, int index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Hier hier = (Hier)self->value;
  if((index < 0) || (index >= length)) return TPM_ERROR;
  /* length is immutable */
  if(flags & BOA_DELETE) { ErrorString = BOA_EFLAGS; return TPM_ERROR; }
  BOA_TYPECHECK_GET(subtype);
  if(value_return) {
    Obj t = Boa_CreateFromData(subtype, NULL);
    t->value = &HierNodeAt(hier, index)->data;
    /* Return a shallow copy, so that changes do not show thru. */
    *value_return = BoaCopyShallow(t);
    Boa_Release(t);
  }
  return TPM_OK;
}

static int IndexObj(Obj self, Obj key)
{
  Hier hier = (Hier)self->value;
  int i;
  Obj t;
  /* if types don't match, we're done */
  Boa_PrintNL(stderr, key);
  if(Boa_Compare(subtype, key->type)) return NAI;
  t = Boa_CreateFromData(subtype, NULL);
  for(i=0;i<hier->num_nodes;i++) {
    HierNode node = HierNodeAt(hier, i);
    if(node->children) continue;
    t->value = &node->data;
    if(!Boa_Compare(t, key)) {
      Boa_Release(t);
      return i;
    }
  }
  Boa_Release(t);
  return NAI;
}
static Obj IndexInt(Obj self, int index)
{
  Hier hier = (Hier)self->value;
  Obj t;
  if((index < 0) || (index >= length)) return NULL;
  /* kludge */
  if(HierNodeAt(hier, index)->children) return NULL;
  t = Boa_CreateFromData(subtype, NULL);
  t->value = &HierNodeAt(hier, index)->data;
  return t;
}

/* convert name to integer and forward to IntGet */
static int StringGet(Obj self, char *name, int flags,
		     Type *type_in_out, Obj *value_return)
{
  int index;
  if(name[0] == '_') {
    if(!strcmp(name+1, "length")) {
      return Boa_IntGet(self->type, 0, flags, type_in_out, value_return);
    }
    else if(!strcmp(name+1, "index")) {
      Obj t = Boa_Create(T_Indexer);
      Boa_Indexer_Data(t) = Boa_Hold(self);
      Boa_Indexer_ObjFunc(t) = IndexObj;
      Boa_Indexer_IntFunc(t) = IndexInt;
      *value_return = t;
      return TPM_OK;
    }
  }
  index = atoi(name);
  if(!index && (name[0] != '0')) return TPM_BOTTOM;
  return IntGet(self, index, flags, type_in_out, value_return);
}

static struct Boa_Methods methods = {
  Init, Final, Empty_Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  StringGet, Err_StringSet,
  IntGet, Err_IntSet,
  Generic_ObjGet, Generic_ObjSet
};

/* assumes num_args and args have already been filled in */
static void Type_Init(Obj self)
{
  Boa_ParamTypeInstance data = (Boa_ParamTypeInstance)self->value;
  data->methods = &methods;
  data->size = sizeof(struct Hier);
  data->align = sizeof(Ptr);
}

Obj Boa_Hier_Define(void)
{
  T_Hier = Boa_ParamType_Create("hier", Type_Init, 2, T_Int, NULL);
  return T_Hier;
}
