/* Binary classifier.
 * SetObj with a sample and a number in [-1, 1] trains the classifier.
 * GetObj with a sample returns a number in [-1, 1], where zero means 
 * "no opinion".
 *
 * A sample is a List of (feature, key) pairs, where
 * key is an index into the feature object.
 * E.g. if feature is a Hier then key is a node number.
 *
 * Uses an IdDict to store the features.
 * Assumes all features are metrics.
 *
 * This classifier uses a kernel estimate as the prior for Bayesian learning.
 */
#include <tpm/list.h>
#include <boa/boa.h>
#include <boa/vector.h>

/* Smoothing parameters */
#define SCALE 75

typedef struct Classifier {
  Obj features;   /* holds array of certainties */
  Vector weight;  /* output weights */
} *Classifier;

/* Globals *******************************************************************/
Type T_Classifier_Kernel;

/* Functions *****************************************************************/

static void Init(Obj self)
{
  Classifier c = (Classifier)self->value;
  c->features = Boa_Create(T_IdDict);
  c->weight = NULL;
}

static void Final(Obj self)
{
  Classifier c = (Classifier)self->value;
  Boa_Release(c->features);
  if(c->weight) VectorFree(c->weight);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Classifier c = (Classifier)self->value;
  func(c->features, data);
}

static int FeatureKeyIndex(Obj feature, Obj key) 
{
  Obj tests;
  int i;
  /* assume feature is a metric object, and key is a query */
  if(Boa_StringGet(feature, "_context/test", '/', NULL, &tests) == TPM_OK) {
    assert(Boa_ObjGetInt(tests, key, &i) == TPM_OK);
    Boa_Release(tests);
    return i;
  }
  else if(key->type == T_Int) {
    return Boa_Int_Value(key);
  } 
  fprintf(stderr, "FeatureKeyIndex: don't know how to index ");
  Boa_PrintNL(stderr, key);
  return -1;
}

static int FeatureNumKeys(Obj feature)
{
  Obj tests;
  int i;
  /* assume feature is a metric object, and key is a query */
  if(Boa_StringGet(feature, "_context/test", '/', NULL, &tests) == TPM_OK) {
    assert(Boa_StringGetInt(tests, "_length", &i) == TPM_OK);
    Boa_Release(tests);
  }
  else assert(Boa_StringGetInt(feature, "_length", &i) == TPM_OK);
  return i;
}

#define COMBINE_ODDS(a, b) ((a) + (b)) / ((a) * (b) + 1)

/* index is a List of (feature, key) tuples.
 * returns a float certainty factor.
 */
static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Classifier c = (Classifier)self->value;
  List pairs;
  Obj pair;
  float weight = 0;
  int i;

  if(index->type != T_List) return TPM_BOTTOM;
  pairs = Boa_List_Value(index);
  {ListIter(p, pair, pairs) {
    Obj feature = Boa_Tuple_Item(pair, 0);
    Obj t;

    if(Boa_ObjGet(c->features, feature, 0, NULL, &t) != TPM_OK) {
      /* haven't seen this feature before */
      continue;
    }
    else {
      Matrix kernel;
      int ntests;
      int col = FeatureKeyIndex(feature, Boa_Tuple_Item(pair, 1));
      float *certainty = (float*)t->value;

      ntests = Boa_Array_Length(t->type);
      Boa_Release(t);
      assert(Boa_StringGet(feature, "_matrix", 0, NULL, &t) == TPM_OK);
      kernel = Boa_Matrix_Value(t);

      if(!c->weight) {
	float total_cert = 0;
	int j, step = ntests;
	Vector likelihood = VectorCreate(kernel->rows);
	c->weight = VectorCreate(ntests);
	VectorSet(c->weight, 0);
	for(j=0;j<ntests;j++) total_cert += abs(certainty[j]);
	for(j=0;j<ntests;) {
	  float sum = 0;
	  float cert_sum = 0;
	  double max_like = -1.0/0;
	  int max_kernel;
	  int k;
	  VectorSet(likelihood, 0);
	  for(k=0;k<step && j<ntests;j++) {
	    if(certainty[j] == 0) continue;
	    k++;
	    cert_sum += abs(certainty[j]);
	    for(i=0;i<kernel->rows;i++) {
	      float f = certainty[j] * kernel->d2[i][j];
	      if(f < 0) likelihood->d[i] += f;
	    }
	  }
	  if(cert_sum == 0) continue;
	  for(i=0;i<kernel->rows;i++) {
	    /* set up plan B */
	    if(likelihood->d[i] > max_like) {
	      max_like = likelihood->d[i];
	      max_kernel = i;
	    }
	    sum += likelihood->d[i] = exp(SCALE*likelihood->d[i]/cert_sum);
#if SYMMETRIC
	    sum += exp(-SCALE) / likelihood->d[i];
#endif
	  }
	  if(sum > 0) {
	    for(i=0;i<kernel->rows;i++) {
	      for(k=0;k<ntests;k++) {
		float w = kernel->d2[i][k] / total_cert * cert_sum
#if SYMMETRIC
		  * (likelihood->d[i] - exp(-SCALE) / likelihood->d[i]) / sum;
#else
		  * likelihood->d[i] / sum;
#endif
#if 1
		c->weight->d[k] += w;
#else
		c->weight->d[k] = COMBINE_ODDS(c->weight->d[k], w);
#endif
	      }
	    }
	  }
	  else {
	    for(k=0;k<ntests;k++) {
	      c->weight->d[k] += kernel->d2[max_kernel][k] 
		/ total_cert * cert_sum;
	    }
	  }
	}
	VectorFree(likelihood);
      }
      if(certainty[col] != 0) weight = certainty[col];
      else weight = c->weight->d[col];

      Boa_Release(t);  /* kernel */
    }
  }}

  /* clip */
  if(weight > 1) weight = 1;
  else if(weight < -1) weight = -1;

  BOA_TYPECHECK_GET(T_Float);
  if(value_return) *value_return = Boa_Float_Create(weight);
  return TPM_OK;
}

/* index is a List of (feature, key) tuples.
 * value is a float certainty factor.
 */
static int ObjSet(Obj self, Obj index, int flags, Obj value)
{
  Classifier c = (Classifier)self->value;
  float certainty;
  List pairs;
  Obj pair;
  if(index->type != T_List) return TPM_BOTTOM;
  pairs = Boa_List_Value(index);
  if(value->type != T_Float) return TPM_BOTTOM;
  certainty = Boa_Float_Value(value);

  if(c->weight) {
    VectorFree(c->weight);
    c->weight = NULL;
  }

  {ListIter(p, pair, pairs) {
    Obj feature, t;
    float *data;
    int i, status;

    assert(Boa_Tuple_Length(pair) == 2);
    feature = Boa_Tuple_Item(pair, 0);
    status = Boa_ObjGet(c->features, feature, 0, NULL, &t);

    i = FeatureKeyIndex(feature, Boa_Tuple_Item(pair, 1));
    if(status != TPM_OK) {
      /* haven't seen this feature before */
      /* create an array of certainties for this feature (brute force) */
      int ntests = FeatureNumKeys(feature);
      char str[64];
      sprintf(str, "(array %d float)", ntests);
      t = Boa_Create(Boa_Type_Create(str));
      data = (float*)t->value;
      memset(data, 0, ntests*sizeof(float));
      Boa_ObjSet(c->features, feature, BOA_INSERT, Boa_Hold(t));
    }
    else data = (float*)t->value;
    data[i] = certainty;

    Boa_Release(t);
  }}

  return TPM_OK;
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  Generic_SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Err_StringGet, Err_StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_Classifier_Kernel_Define(void)
{
  Boa_Vector_Pkg();
  T_Classifier_Kernel = Boa_SimpleType_Create("classifier_kernel", &methods, 
					       sizeof(struct Classifier), 
					       sizeof(Obj));
  return T_Classifier_Kernel;
}
