/* Binary classifier.
 * SetObj with a sample and a number in [-1, 1] trains the classifier.
 * GetObj with a sample returns a number in [-1, 1], where zero means 
 * "no opinion".
 *
 * A sample is a List of (feature, key) pairs, where
 * key is an index into the feature object.
 * E.g. if feature is a Hier then key is a node number.
 *
 * Uses an IdDict to store the features.
 * Assumes all features are metrics.
 *
 * This classifier picks the class of the closest example.
 */
#include <tpm/list.h>
#include <boa/boa.h>
#include <boa/vector.h>

#define AVERAGE 0
#define VOTE 1
#define NBRS 1

typedef struct Classifier {
  Obj features;  /* holds array of certainties */
} *Classifier;

/* Globals *******************************************************************/
Type T_Classifier_Nearest;

/* Functions *****************************************************************/

static void Init(Obj self)
{
  Classifier c = (Classifier)self->value;
  c->features = Boa_Create(T_IdDict);
}

static void Final(Obj self)
{
  Classifier c = (Classifier)self->value;
  Boa_Release(c->features);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Classifier c = (Classifier)self->value;
  func(c->features, data);
}

static int FeatureKeyIndex(Obj feature, Obj key) 
{
  Obj tests;
  int i;
  /* assume feature is a metric object, and key is a query */
  if(Boa_StringGet(feature, "_context/test", '/', NULL, &tests) == TPM_OK) {
    assert(Boa_ObjGetInt(tests, key, &i) == TPM_OK);
    Boa_Release(tests);
    return i;
  }
  else if(key->type == T_Int) {
    return Boa_Int_Value(key);
  } 
  fprintf(stderr, "FeatureKeyIndex: don't know how to index ");
  Boa_PrintNL(stderr, key);
  return -1;
}

static int FeatureNumKeys(Obj feature)
{
  Obj tests;
  int i;
  /* assume feature is a metric object, and key is a query */
  if(Boa_StringGet(feature, "_context/test", '/', NULL, &tests) == TPM_OK) {
    assert(Boa_StringGetInt(tests, "_length", &i) == TPM_OK);
    Boa_Release(tests);
  }
  else assert(Boa_StringGetInt(feature, "_length", &i) == TPM_OK);
  return i;
}

#define COMBINE_ODDS(a, b) ((a) + (b)) / ((a) * (b) + 1)

/* index is a List of (feature, key) tuples.
 * returns a float certainty factor.
 */
static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Classifier c = (Classifier)self->value;
  List pairs;
  Obj pair;
  float weight = 0, total;
  int i, k;
  double distance[NBRS];
  float nbr_weight[NBRS];

  if(index->type != T_List) return TPM_BOTTOM;
  for(k=0;k<NBRS;k++) { distance[k] = 1.0/0; nbr_weight[k] = 0; }
 
  pairs = Boa_List_Value(index);
  {ListIter(p, pair, pairs) {
    Obj feature = Boa_Tuple_Item(pair, 0);
    Obj t;

    if(Boa_ObjGet(c->features, feature, 0, NULL, &t) != TPM_OK) {
      /* haven't seen this feature before */
      continue;
    }
    else {
      Vector dists;
      int ntests;
      float *certainty = (float*)t->value;
      Obj key = Boa_Tuple_Item(pair, 1);
      int index = FeatureKeyIndex(feature, key);

      ntests = Boa_Array_Length(t->type);
      Boa_Release(t);
      /* get a vector of distances */
      if(Boa_ObjGet(feature, key, 0, NULL, &t) != TPM_OK) {
	ErrorPrint(stderr);
	continue;
      }
      dists = Boa_Vector_Value(t);

      if(AVERAGE) {
	float this_weight = 0;
	for(i=0;i<ntests;i++) {
	  float new_weight;
	  if(certainty[i] == 0) continue;
	  /* assume that dists contains Pr(c(x) != c(y)) */
	  new_weight = -(2*dists->d[i] - 1)*certainty[i];
#if 0
	  this_weight = COMBINE_ODDS(this_weight, new_weight);
#else
	  this_weight += new_weight;
#endif
	}
	weight += this_weight;
      }
      else {
	for(i=0;i<ntests;i++) {
	  if(certainty[i] == 0) continue;
	  /* find the closest neighbor k that i is closer than */
	  for(k=NBRS;k>0;k--) if(dists->d[i] > distance[k-1]) break;
	  if(k == NBRS) continue; /* too far */
	  if(VOTE && (distance[k] == dists->d[i])) {
	    /* accumulate the weight of equidistant neighbors */
	    nbr_weight[k] += certainty[i];
	  }
	  else {
	    int j;
	    /* shift up */
	    for(j=NBRS-1;j>k;j--) {
	      distance[j] = distance[j-1];
	      nbr_weight[j] = nbr_weight[j-1];
	    }
	    nbr_weight[k] = certainty[i];
	    distance[k] = dists->d[i];
	  }
	}
      }
      /* label examples as given */
      if(certainty[index] != 0) nbr_weight[0] = certainty[index];
      Boa_Release(t);
    }
  }}


#if AVERAGE
  /* clip */
  if(weight > 1) weight = 1;
  else if(weight < -1) weight = -1;
#else
  /* vote the nearest neighbors */
  total = 0;
  for(k=0;k<NBRS;k++) {
    weight += nbr_weight[k] * (exp(-k) - exp(-k-1));
    /* this is Kong and Deitterich's infinite bootstrap formula */
    total += exp(-k) - exp(-k-1);
  }
  weight /= total;

  if(distance[0] == 0) weight = nbr_weight[0];
#endif

  BOA_TYPECHECK_GET(T_Float);
  if(value_return) *value_return = Boa_Float_Create(weight);
  return TPM_OK;
}

/* index is a List of (feature, key) tuples.
 * value is a float certainty factor.
 */
static int ObjSet(Obj self, Obj index, int flags, Obj value)
{
  Classifier c = (Classifier)self->value;
  float certainty;
  List pairs;
  Obj pair;
  if(index->type != T_List) return TPM_BOTTOM;
  pairs = Boa_List_Value(index);
  if(value->type != T_Float) return TPM_BOTTOM;
  certainty = Boa_Float_Value(value);

  {ListIter(p, pair, pairs) {
    Obj feature, t;
    float *data;
    int i, status;

    assert(Boa_Tuple_Length(pair) == 2);
    feature = Boa_Tuple_Item(pair, 0);
    status = Boa_ObjGet(c->features, feature, 0, NULL, &t);

    i = FeatureKeyIndex(feature, Boa_Tuple_Item(pair, 1));
    if(status != TPM_OK) {
      /* haven't seen this feature before */
      /* create an array of certainties for this feature (brute force) */
      int ntests = FeatureNumKeys(feature);
      char str[64];
      sprintf(str, "(array %d float)", ntests);
      t = Boa_Create(Boa_Type_Create(str));
      data = (float*)t->value;
      memset(data, 0, ntests*sizeof(float));
      Boa_ObjSet(c->features, feature, BOA_INSERT, Boa_Hold(t));
    }
    else data = (float*)t->value;
    data[i] = certainty;

    Boa_Release(t);
  }}

  return TPM_OK;
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  Generic_SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Err_StringGet, Err_StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_Classifier_Nearest_Define(void)
{
  Boa_Vector_Pkg();
  T_Classifier_Nearest = Boa_SimpleType_Create("classifier_nearest", &methods, 
					       sizeof(struct Classifier), 
					       sizeof(Obj));
  return T_Classifier_Nearest;
}
