/* Binary classifier.
 * SetObj with a sample and a number in [-1, 1] trains the classifier.
 * GetObj with a sample returns a number in [-1, 1], where zero means 
 * "no opinion".
 *
 * A sample is a List of (feature, key) pairs, where
 * key is an index into the feature object.
 * E.g. if feature is a Hier then key is a node number.
 *
 * Uses an IdDict to store the features.
 * For each feature, there is associated information for performing
 * classification.  For example, pos/neg counts for each node of a Hier.
 *
 * This classifier is a conservative version of SC, with an optional size bias.
 */
#include <tpm/list.h>
#include <boa/boa.h>
#include <boa/vector.h>
#include <tpm/hier.h>
#include <tpm/bitset.h>

#define DEBUG 0
/* SIZE_BIAS doesn't work for metrics yet */
#define SIZE_BIAS 1
#define RANDOM 1

typedef enum { HIER_COVER, METRIC_COVER } CoverType;

typedef struct HierCover {
  CoverType type;
  float weight;        /* strength of this cover */
  Obj feature;         /* attribute we're covering */
  unsigned low, high;  /* range of leaves covered */
} *HierCover;

typedef struct MetricCover {
  CoverType type;
  float weight;        /* strength of this cover */
  Obj feature;         /* attribute we're covering */
  BitSet mask;
} *MetricCover;

typedef struct Cover {
  CoverType type;
  float weight;        /* strength of this cover */
} *Cover;

typedef struct Classifier {
  Obj features;
  Obj pos_ex;   /* List of Instance */
  List covers;   /* List of Cover */
} *Classifier;

/* Globals *******************************************************************/
Type T_Classifier_SC;

/* Functions *****************************************************************/

static void CoverFree(Cover c)
{
  switch(c->type) {
  case METRIC_COVER: BitSetFree(((MetricCover)c)->mask);
  case HIER_COVER: /* do nothing */;
  }
  /* feature is not held by the cover */
  free(c);
}

static Cover CoverCopy(Cover c)
{
  switch(c->type) {
  case HIER_COVER: return (Cover)AllocCopy(1, struct HierCover, c);
  case METRIC_COVER: 
    {MetricCover m = AllocCopy(1, struct MetricCover, c);
     m->mask = BitSetCopy(m->mask);
     return (Cover)m;}
  }
  return c;
}

static void Init(Obj self)
{
  Classifier c = (Classifier)self->value;
  c->features = Boa_Create(T_IdDict);
  c->pos_ex = Boa_Create(T_List);
#if 0
  c->covers = Boa_Create(T_List);
#else
  c->covers = ListCreate((FreeFunc*)CoverFree);
#endif
}

static void Final(Obj self)
{
  Classifier c = (Classifier)self->value;
  Boa_Release(c->features);
  Boa_Release(c->pos_ex);
#if 0
  Boa_Release(c->covers);
#endif
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Classifier c = (Classifier)self->value;
  func(c->features, data);
  func(c->pos_ex, data);
#if 0
  func(c->covers, data);
#endif
}

typedef struct TreeData {
  /* number of pos under this node */
  float pos;
  /* prob of not having a neg under this node */
  float neg;
  /* weight of the node */
  float bias;
} *TreeData;

#if 0
/* Returns the data of all instances in the class, i.e. with weight > 0.
 * classifier is a Classifier.
 * features is an Ord.
 * If compare != NULL, only instance data matching item will be returned.
 */
List BoaClassifier_AllMatching(Obj self, Obj features, 
			       CmpFunc *compare,
			       Any item, CmpFunc *include_compare)
{
  return NULL;
}
#endif

/* Find the best covers for instance.
 * Modifies *best_score and best.
 */
static void BestCovers(Classifier c, Obj instance, 
		       float *best_score, List best)
{
  Obj pair;
  /* Loop leaves */
  {ListIter(p, pair, Boa_List_Value(instance)) {
    Obj feature, t;

    feature = Boa_Tuple_Item(pair, 0);
    if(Boa_ObjGet(c->features, feature, 0, NULL, &t) 
       != TPM_OK) {
      /* haven't seen this feature before */
      continue;
    }

    if(feature->type->type == T_Hier) {
      Hier hier = Boa_Hier_Value(feature);
      TreeData data = (TreeData)t->value;
      int i = Boa_Int_Value(Boa_Tuple_Item(pair, 1));
      /* climb the tree, looking for nodes with pos > neg */
      while(i >= 0) {
	HierNode node = HierNodeAt(hier, i);
	float score = data[i].pos * data[i].neg * data[i].bias;
#if DEBUG
	fprintf(stderr, "%p %d sc: %g pos: %g neg: %g bias: %g\n",
		feature, i, score,
		data[i].pos, data[i].neg, data[i].bias);
#endif
	if(score > *best_score) {
	  *best_score = score;
	  ListFlush(best);
	}
	if(score == *best_score) {
	  HierCover c = Allocate(1, struct HierCover);
	  c->type = HIER_COVER;
	  c->feature = feature;  /* XXX not held */
	  c->low = i;
	  c->weight = 1;  /* until I figure out what else to use */
	  ListAddFront(best, c);
	}
	i = node->parent;
      }
    }
    else {
      Vector dists;
      int i, ntests;
      double min_dist = 1.0/0;
      float score;
      float *certainty = (float*)t->value;
      ntests = Boa_Array_Length(t->type);
      /* assume feature is a metric object, and value is query */
      /* assume features, test fields have been filled in */
      /* get a vector of distances */
      if(Boa_ObjGet(feature, Boa_Tuple_Item(pair, 1), 0, NULL, &t) != TPM_OK) {
	ErrorPrepend("metric failed");
	ErrorPrint(stderr);
	continue;
      }
      dists = Boa_Vector_Value(t);

      /* find closest negative example */
      for(i=0;i<ntests;i++) {
	if((certainty[i] < 0) && (dists->d[i] < min_dist)) {
	  min_dist = dists->d[i];
	}
      }

      /* cover is [0, min_dist) */
      /* calculate the score of the cover (number of pos) */
      score = 0;
      for(i=0;i<ntests;i++) {
	if(dists->d[i] < min_dist) score += certainty[i];
      }
      if(score > *best_score) {
	*best_score = score;
	ListFlush(best);
      }
      if(score == *best_score) {
	MetricCover c = Allocate(1, struct MetricCover);
	c->type = METRIC_COVER;
	c->feature = feature;
	c->mask = BitSetCreate(ntests);
	for(i=0;i<ntests;i++) {
	  if(dists->d[i] < min_dist) BitSetInsert(c->mask, i);
	}
	c->weight = 1;  /* until I figure out what else to use */
	ListAddFront(best, c);
#if DEBUG
	fprintf(stderr, "min_dist: %g\n", min_dist);
#endif
      }
    }
    Boa_Release(t);
  }}
}

static int FeatureKeyIndex(Obj feature, Obj key) 
{
  Obj tests;
  int i;
  /* assume feature is a metric object, and key is a query */
  if(Boa_StringGet(feature, "_context/test", '/', NULL, &tests) == TPM_OK) {
    assert(Boa_ObjGetInt(tests, key, &i) == TPM_OK);
    Boa_Release(tests);
    return i;
  }
  else if(key->type == T_Int) {
    return Boa_Int_Value(key);
  } 
  fprintf(stderr, "FeatureKeyIndex: don't know how to index ");
  Boa_PrintNL(stderr, key);
  return -1;
}

static int FeatureNumKeys(Obj feature)
{
  Obj tests;
  int i;
  /* assume feature is a metric object, and key is a query */
  if(Boa_StringGet(feature, "_context/test", '/', NULL, &tests) == TPM_OK) {
    assert(Boa_StringGetInt(tests, "_length", &i) == TPM_OK);
    Boa_Release(tests);
  }
  else assert(Boa_StringGetInt(feature, "_length", &i) == TPM_OK);
  return i;
}

/* Returns 0 if cover covers instance */
static int NotCovered(Obj instance, Cover cover)
{
  Obj pair;
  ListIter(p, pair, Boa_List_Value(instance)) {
    Obj feature;
    int i;

    assert(Boa_Tuple_Length(pair) == 2);
    feature = Boa_Tuple_Item(pair, 0);

    switch(cover->type) {
    case HIER_COVER:
      {HierCover c = (HierCover)cover;
       if(feature != c->feature) continue;
       i = Boa_Int_Value(Boa_Tuple_Item(pair, 1));
       if((c->low > i) || (c->high < i)) continue;}
      break;
    case METRIC_COVER:
      {MetricCover c = (MetricCover)cover;
       if(feature != c->feature) continue;
       /* find the index of this key and probe the mask */
       i = FeatureKeyIndex(feature, Boa_Tuple_Item(pair, 1));
       if(!BitSetMember(c->mask, i)) continue;}
      break;
    }
    return 0;
  }
  return 1;
}

static void CollectCovers(Classifier c)
{
  Obj instance;
  List pos_ex;
  int i;
  
  pos_ex = ListCopy(Boa_List_Value(c->pos_ex), NULL);
  ListFreeFunc(pos_ex) = NULL;

  /* loop positive examples */
  while(ListSize(pos_ex)) {
    Cover cover;
    float best_score = 0;
    List best = ListCreate((FreeFunc*)CoverFree);
    instance = (Obj)ListFront(pos_ex)->data;
    /* find the best covers for the example */
    BestCovers(c, instance, &best_score, best);
    /* was this example disallowed? */
    if(best_score == 0) {
      fprintf(stderr, "disallowed\n");
      ListFree(best);
      ListRemoveFront(pos_ex, NULL);
      continue;
    }
#if RANDOM
    i = RandInt(0, ListSize(best)-1);
#else
    i = ListSize(best)-1;
#endif
    /* make a copy so ListFree(best) doesn't clobber it */
    cover = CoverCopy((Cover)ListValueAtIndex(best, i));
    ListFree(best);
    if(cover->type == HIER_COVER) {
      HierCover h = (HierCover)cover;
      /* post-process the cover */
      h->high = HierRightLeaf(Boa_Hier_Value(h->feature), h->low);
#if DEBUG
      fprintf(stderr, "cover: %p %d-%d %g\n", 
	      h->feature, h->low, h->high, best_score);
#endif
    }
    else {
#if DEBUG
      MetricCover m = (MetricCover)cover;
      fprintf(stderr, "cover: %p %d %g\n", 
	      m->feature, BitSetSize(m->mask), best_score);
#endif
    }
    ListAddFront(c->covers, cover);

#if 1
    /* remove all covered examples */
    ListRemoveAll(pos_ex, cover, (CmpFunc*)NotCovered);
#else
    /* remove just this example */
    ListRemoveFront(pos_ex, NULL);
#endif
  }
  ListFree(pos_ex);
}

/* index is a List of (feature, key) tuples.
 * returns a float certainty factor.
 */
static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Classifier c = (Classifier)self->value;
  List pairs;
  Obj pair;
  float not_for, not_against;
  if(index->type != T_List) return TPM_BOTTOM;
  pairs = Boa_List_Value(index);

  if(ListEmpty(c->covers)) CollectCovers(c);

  not_for = not_against = 1;
  {ListIter(p, pair, pairs) {
    Obj feature;

    assert(Boa_Tuple_Length(pair) == 2);
    feature = Boa_Tuple_Item(pair, 0);

    /* see if there are covers for this pair */
    {Cover cover; ListIter(p2, cover, c->covers) {
      int i;
      switch(cover->type) {
      case HIER_COVER:
	{HierCover c = (HierCover)cover;
	 if(feature != c->feature) continue;
	 i = Boa_Int_Value(Boa_Tuple_Item(pair, 1));
	 if((c->low > i) || (c->high < i)) continue;}
	break;
      case METRIC_COVER:
	{MetricCover c = (MetricCover)cover;
	 if(feature != c->feature) continue;
	 /* find the index of this key and probe the mask */
	 i = FeatureKeyIndex(feature, Boa_Tuple_Item(pair, 1));
	 if(!BitSetMember(c->mask, i)) continue;}
	break;
      }
      /* yes */
      not_for *= 1 - cover->weight;
      not_against *= cover->weight;
      goto found;
    }}
  found:;
  }}

#if 1
  /* force an opinion */
  if(not_against - not_for == 0) not_against = 0;
#endif

  BOA_TYPECHECK_GET(T_Float);
  if(value_return) *value_return = Boa_Float_Create(not_against - not_for);
  return TPM_OK;
}

/* index is a List of (feature, key) tuples.
 * value is a float certainty factor.
 */
static int ObjSet(Obj self, Obj index, int flags, Obj value)
{
  Classifier c = (Classifier)self->value;
  float certainty;
  List pairs;
  Obj pair;
  if(index->type != T_List) return TPM_BOTTOM;
  pairs = Boa_List_Value(index);
  if(value->type != T_Float) return TPM_BOTTOM;
  certainty = Boa_Float_Value(value);
  /* assumes we won't get an error */
  Boa_Release(value);

  if(certainty > 0) {
    /* add this instance to the list of pos ex */
    ListAddFront(Boa_List_Value(c->pos_ex), Boa_Stabilize(index));
  }
  /* flush the cover list */
  ListFlush(c->covers);
  /* rest is verbatim from SSC */

  {ListIter(p, pair, pairs) {
    Obj feature, t;
    int status, i;

    assert(Boa_Tuple_Length(pair) == 2);
    feature = Boa_Tuple_Item(pair, 0);
    status = Boa_ObjGet(c->features, feature, 0, NULL, &t);

    if(feature->type->type == T_Hier) {
      TreeData data;
      Hier hier = Boa_Hier_Value(feature);

      if(status != TPM_OK) {
	int j;
	char str[64];
	/* haven't seen this feature before */
	/* create an array of TreeData for this feature */
	sprintf(str, "(array %d (array 3 float))", hier->num_nodes);
	t = Boa_Create(Boa_Type_Create(str));
	data = (TreeData)t->value;
	/* initialize the treeData */
	/* first set bias to size of each node */
	for(j=0;j<hier->num_nodes;j++) { 
	  data[j].pos = 0; data[j].neg = 1; data[j].bias = 1;
	}
#if SIZE_BIAS
	for(j=hier->num_nodes;j>0;) { 
	  HierNode node = HierNodeAt(hier, --j);
	  if(node->parent >= 0) data[node->parent].bias += data[j].bias;
	}
	/* normalize the bias */
	for(j=0;j<hier->num_nodes;j++) { 
	  data[j].bias /= hier->num_nodes;
	  data[j].bias = data[j].bias*0.8 + 0.1;
	  data[j].bias *= 1 - data[j].bias;
	}
#endif
	Boa_ObjSet(c->features, feature, BOA_INSERT, Boa_Hold(t));
      }
      else data = (TreeData)t->value;

      /* climb the tree, updating data fields */
      i = Boa_Int_Value(Boa_Tuple_Item(pair, 1));
      while(i >= 0) {
	HierNode node = HierNodeAt(hier, i);
	if(certainty > 0) data[i].pos += certainty;
	else data[i].neg *= 1 + certainty;
	i = node->parent;
      }
    }
    else {
      float *data;
      /* haven't seen this feature before */
      i = FeatureKeyIndex(feature, Boa_Tuple_Item(pair, 1));
      if(status != TPM_OK) {
	/* create an array of certainties for this feature (brute force) */
	int ntests = FeatureNumKeys(feature);
	char str[64];
	sprintf(str, "(array %d float)", ntests);
	t = Boa_Create(Boa_Type_Create(str));
	data = (float*)t->value;
	memset(data, 0, ntests*sizeof(float));
	Boa_ObjSet(c->features, feature, BOA_INSERT, Boa_Hold(t));
      }
      else data = (float*)t->value;
      data[i] = certainty;
    }
    Boa_Release(t);
  }} 
  return TPM_OK;
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  Generic_SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Err_StringGet, Err_StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_Classifier_SC_Define(void)
{
  T_Classifier_SC = Boa_SimpleType_Create("classifier_sc", &methods, 
					  sizeof(struct Classifier), 
					  sizeof(Obj));
  return T_Classifier_SC;
}
