/* Binary classifier.
 * SetObj with a sample and a number in [-1, 1] trains the classifier.
 * GetObj with a sample returns a number in [-1, 1], where zero means 
 * "no opinion".
 *
 * A sample is a List of (feature, key) pairs, where
 * key is an index into the feature object.
 * E.g. if feature is a Hier then key is a node number.
 *
 * Uses an IdDict to store the features.
 * Assumes all features are metrics.
 *
 * This classifier performs nearest neighbor, but with a linear combination
 * of the metrics.  The combination is found by linear projection of the 
 * "empirical" metric.
 */
#include <tpm/list.h>
#include <boa/boa.h>
#include <boa/vector.h>

#define VOTE 1
#define AVERAGE 0

typedef struct Classifier {
  /* each feature stores a tuple of:
   * 1. array of certainties 
   * 2. weight in the combination
   */
  Obj features;  
} *Classifier;

/* Globals *******************************************************************/
Type T_Classifier_Select;

/* Functions *****************************************************************/

static void Init(Obj self)
{
  Classifier c = (Classifier)self->value;
  c->features = Boa_Create(T_IdDict);
}

static void Final(Obj self)
{
  Classifier c = (Classifier)self->value;
  Boa_Release(c->features);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Classifier c = (Classifier)self->value;
  func(c->features, data);
}

static int FeatureKeyIndex(Obj feature, Obj key) 
{
  Obj tests;
  int i;
  /* assume feature is a metric object, and key is a query */
  if(Boa_StringGet(feature, "_context/test", '/', NULL, &tests) == TPM_OK) {
    assert(Boa_ObjGetInt(tests, key, &i) == TPM_OK);
    Boa_Release(tests);
    return i;
  }
  else if(key->type == T_Int) {
    return Boa_Int_Value(key);
  } 
  fprintf(stderr, "FeatureKeyIndex: don't know how to index ");
  Boa_PrintNL(stderr, key);
  return -1;
}

static int FeatureNumKeys(Obj feature)
{
  Obj tests;
  int i;
  /* assume feature is a metric object, and key is a query */
  if(Boa_StringGet(feature, "_context/test", '/', NULL, &tests) == TPM_OK) {
    assert(Boa_StringGetInt(tests, "_length", &i) == TPM_OK);
    Boa_Release(tests);
  }
  else assert(Boa_StringGetInt(feature, "_length", &i) == TPM_OK);
  return i;
}

/* index is a List of (feature, key) tuples.
 * returns a float certainty factor.
 */
static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Classifier c = (Classifier)self->value;
  List pairs;
  Obj pair;
  float weight = 0;
  int i;
  Vector total_dist = NULL;
  Obj certaintyObj = NULL;

  if(index->type != T_List) return TPM_BOTTOM;

  /* add the weighted distance vectors together */
  pairs = Boa_List_Value(index);
  {ListIter(p, pair, pairs) {
    Obj feature = Boa_Tuple_Item(pair, 0);
    Obj t;

    if(Boa_ObjGet(c->features, feature, 0, NULL, &t) != TPM_OK) {
      /* haven't seen this feature before */
      continue;
    }
    else {
      Vector dists;
      double fcorr = Boa_Double_Value(Boa_Tuple_Item(t,1));
      double fsum = Boa_Double_Value(Boa_Tuple_Item(t,2));
      if(!certaintyObj) {
	assert(Boa_IntGet(t, 0, 0, NULL, &certaintyObj) == TPM_OK);
      }
      Boa_Release(t);

      /* get a vector of distances */
      if(Boa_ObjGet(feature, Boa_Tuple_Item(pair, 1), 0, NULL, &t) != TPM_OK) {
	ErrorPrint(stderr);
	continue;
      }
      dists = Boa_Vector_Value(t);
      fprintf(stderr, "fcorr: %g  fsum: %g  fcorr/sqrt(fsum): %g\n",
	      fcorr, fsum, fcorr/sqrt(fsum));
      VectorScale(dists, 2);
      VectorAddValue(dists, -1); /* debias */
      if(fsum > 0) VectorScale(dists, fcorr/sqrt(fsum));
      if(!total_dist) total_dist = VectorCopy(dists);
      else VectorAdd(total_dist, dists);
      Boa_Release(t);
    }
  }}

  /* find the minimum */
  if(certaintyObj && total_dist) {
    float *certainty = (float*)certaintyObj->value;
    double min_d = 1.0/0;
    VectorAddValue(total_dist, 1); /* rebias */
    VectorScale(total_dist, 0.5);
    for(i=0;i<total_dist->len;i++) {
      if(certainty[i] == 0) continue;
      if(total_dist->d[i] < min_d) {
	min_d = total_dist->d[i];
	weight = certainty[i];
      }
#if VOTE
      else if(total_dist->d[i] == min_d) {
	weight += certainty[i];
      }
#endif
    }
    VectorFree(total_dist);
  }
  if(certaintyObj) Boa_Release(certaintyObj);

  /* clip */
  if(weight > 1) weight = 1;
  else if(weight < -1) weight = -1;

  BOA_TYPECHECK_GET(T_Float);
  if(value_return) *value_return = Boa_Float_Create(weight);
  return TPM_OK;
}

/* index is a List of (feature, key) tuples.
 * value is a float certainty factor.
 */
static int ObjSet(Obj self, Obj index, int flags, Obj value)
{
  Classifier c = (Classifier)self->value;
  float certainty;
  List pairs;
  Obj pair;
  if(index->type != T_List) return TPM_BOTTOM;
  pairs = Boa_List_Value(index);
  if(value->type != T_Float) return TPM_BOTTOM;
  certainty = Boa_Float_Value(value);

  {ListIter(p, pair, pairs) {
    Obj feature, t;
    float *data;
    int i, status;
    int ntests;

    assert(Boa_Tuple_Length(pair) == 2);
    feature = Boa_Tuple_Item(pair, 0);
    status = Boa_ObjGet(c->features, feature, 0, NULL, &t);

    ntests = FeatureNumKeys(feature);
    i = FeatureKeyIndex(feature, Boa_Tuple_Item(pair, 1));
    if(status != TPM_OK) {
      /* haven't seen this feature before */
      /* create an array of certainties for this feature (brute force) */
      char str[64];
      sprintf(str, "(array %d float)", ntests);
      t = Boa_Create(Boa_Type_Create(str));
      data = (float*)t->value;
      memset(data, 0, ntests*sizeof(float));
      pair = Boa_Create(Boa_Type_Create("(tuple 2)"));
      Boa_Tuple_Item(pair, 0) = t;
      Boa_Tuple_Item(pair, 1) = Boa_Double_Create(0);  /* feature corr */
      Boa_Tuple_Item(pair, 2) = Boa_Double_Create(0);  /* feature sum */
      assert(Boa_ObjSet(c->features, feature, BOA_INSERT, pair) == TPM_OK);
    }
    else {
      Vector dists;
      int j;
      double *fcorrp, *fsump;
      data = (float*)Boa_Tuple_Item(t,0)->value;
      fcorrp = &Boa_Double_Value(Boa_Tuple_Item(t,1));
      fsump = &Boa_Double_Value(Boa_Tuple_Item(t,2));
      Boa_Release(t);
      /* get a vector of distances */
      if(Boa_ObjGet(feature, Boa_Tuple_Item(pair, 1), 0, NULL, &t) != TPM_OK) {
	ErrorPrint(stderr);
	continue;
      }
      dists = Boa_Vector_Value(t);
      /* adjust the feature weight */
      for(j=0;j<ntests;j++) {
	double d = dists->d[j] * data[j];
	if(j == i) continue;
	*fcorrp -= d * certainty;
	*fsump += d * d;
      }
      Boa_Release(t);
    }
    data[i] = certainty;
  }}

  return TPM_OK;
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  Generic_SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Err_StringGet, Err_StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_Classifier_Select_Define(void)
{
  Boa_Vector_Pkg();
  T_Classifier_Select = Boa_SimpleType_Create("classifier_select", &methods, 
					      sizeof(struct Classifier), 
					      sizeof(Obj));
  return T_Classifier_Select;
}
