/* Binary classifier.
 * SetObj with a sample and a number in [-1, 1] trains the classifier.
 * GetObj with a sample returns a float in [-1, 1], where zero means 
 * "no opinion".
 *
 * A sample is a List of (feature, key) pairs, where
 * key is an index into the feature object.
 * E.g. if feature is a Hier then key is a node number.
 *
 * Uses an IdDict to store the features.
 * For each feature, there is associated information for performing
 * classification.  For example, pos/neg counts for each node of a Hier.
 *
 * This classifier works like a Bayesian tree but without hidden nodes.
 * Each edge x->y of the SpanTree is assumed to contain p(y=0|x=1).
 * It uses an upward/downward algorithm to compute probabilities.  
 * Uses ratios of alpha and beta values to avoid numerical underflow.
 *
 * Each feature is associated an array of child lists and an updated flag.
 */
#include <tpm/list.h>
#include <boa/boa.h>
#include <tpm/spanTree.h>

#define VERBOSE 0

typedef struct Classifier {
  Obj features;
} *Classifier;

/* Globals *******************************************************************/
Type T_Classifier_SpanTree;

/* Functions *****************************************************************/

static void Init(Obj self)
{
  Classifier c = (Classifier)self->value;
  c->features = Boa_Create(T_IdDict);
}

static void Final(Obj self)
{
  Classifier c = (Classifier)self->value;
  /* XXX go through and manually free all the child lists */
  Boa_Release(c->features);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Classifier c = (Classifier)self->value;
  func(c->features, data);
}

typedef struct TreeData {
  List children;	 /* list of node numbers */
  double alpha, beta, b; /* ratios F/T */
  /* alpha is "prior" probability coming from above.
   * beta is "likelihood" coming from self and below. 
   * b is observation at self.
   */
  /* double-precision is needed for large runs */
} *TreeData;

#define NodeProb(i) (1.0/(1+data[i].alpha*data[i].beta))

/* Return the new value of the probability after crossing a link. */
static double CrossLink(double prob, double link)
{
  if(!finite(link)) return prob;
  else if(!finite(prob)) return link;
  else return (1 + link*prob) / (link + prob);
}

/* Compute beta value of yourself. */
static void ComputeBeta(SpanTree st, TreeData data, int i)
{
  data[i].beta = data[i].b;
  if(data[i].children) {
    int child;
    ListIterInt(p, child, data[i].children) {
      float edge = (1 - st->node[child].edge) / st->node[child].edge;
      ComputeBeta(st, data, child);
      data[i].beta *= CrossLink(data[child].beta, edge);
    }
  }
}

/* Compute alpha values of your children. */
static void ComputeAlpha(SpanTree st, TreeData data, int i)
{
  int child, child1;
  if(!data[i].children) return;

  {ListIterInt(q, child1, data[i].children) {
    float edge1 = (1 - st->node[child1].edge) / st->node[child1].edge;
    double a = data[i].alpha * data[i].b;
    if(!finite(a)) 
      data[child1].alpha = edge1;
    else {
      /* repeat the beta calculation, but save intermediate results */
      ListIterInt(p, child, data[i].children) {
	float edge = (1 - st->node[child].edge) / st->node[child].edge;
	if(child == child1) continue;
	a *= CrossLink(data[child].beta, edge);
      }
      data[child1].alpha = CrossLink(a, edge1);
    }
    ComputeAlpha(st, data, child1);
  }}
}

/* Compute alpha and beta numbers for every node in the spanning tree.
 */
static void ComputeProbs(SpanTree st, TreeData data)
{
  static int ctr = 0;

  ComputeBeta(st, data, 0);
  data[0].alpha = 1;
  ComputeAlpha(st, data, 0);

  if(VERBOSE && !(ctr++ % 1)) {
    /* display results */
    int i;
    for(i=0;i<st->num_nodes;i++) {
      fprintf(stderr, "%d: %g %g %g\n", i, data[i].beta, data[i].alpha, 
	      NodeProb(i));
    }
  }
}

#if 0
/* Returns the data of all instances in the class, i.e. with weight > 0.
 * classifier is a Classifier.
 * features is an Ord.
 * If compare != NULL, only instance data matching item will be returned.
 */
List BoaClassifier_AllMatching(Obj self, Obj features, 
			       CmpFunc *compare,
			       Any item, CmpFunc *include_compare)
{
  Classifier c = (Classifier)self->value;
  int f, i;
  double not_for, not_against;
  List in = ListCreate(NULL), out = ListCreate(NULL);

  /* loop features */
  for(f=0;f<Boa_Ord_Length(features);f++) {
    Obj feature = Boa_Ord_Item(features, f);
    Obj t;
    if(Boa_ObjGet(c->features, feature, 0, NULL, &t) 
       != TPM_OK) {
      /* haven't seen this feature before */
      continue;
    }

    if(feature->type->type == T_Hier) {
      /* Loop all nodes and collect in/out */
      int i;
      Hier hier = Boa_Hier_Value(feature);
      TreeData data = (TreeData)t->value;
      for(i=0;i<hier->num_nodes;i++) {
	HierNode node = HierNodeAt(hier, i);
	double p;
	if(node->children) continue;
	if(include_compare && include_compare(&node->data, item)) continue;
	not_for = not_against = 1;
	p = NodeProb(i);
	not_for *= 1-p;
	not_against *= p;
	if(not_against > not_for) ListAddFront(in, &node->data);
	else if(not_for > not_against) ListAddFront(out, &node->data);
      }
    }
    else {
      fprintf(stderr, "Classifier: unknown feature type ");
      Boa_PrintNL(stderr, feature->type);
    }
    Boa_Release(t);
  }
  /* subtract out from in */
  {Any data; ListIter(p, data, out) {
    ListRemoveAll(in, data, compare);
  }}
  ListFree(out);
  return in;
}
#endif

/* index is a List of (feature, key) tuples.
 * returns a float certainty factor.
 */
static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Classifier c = (Classifier)self->value;
  List pairs;
  Obj pair;
  double not_for, not_against;
  if(index->type != T_List) return TPM_BOTTOM;
  pairs = Boa_List_Value(index);

  not_for = not_against = 1;
  {ListIter(p, pair, pairs) {
    Obj feature, t;

    assert(Boa_Tuple_Length(pair) == 2);
    feature = Boa_Tuple_Item(pair, 0);
    if(Boa_ObjGet(c->features, feature, 0, NULL, &t) 
       != TPM_OK) {
      /* haven't seen this feature before */
      continue;
    }

    if(feature->type->type == T_SpanTree) {
      TreeData data = (TreeData)Boa_Tuple_Item(t, 0)->value;
      int i = Boa_Int_Value(Boa_Tuple_Item(pair, 1));
      double p;
      if((i < 0) || (i >= Boa_SpanTree_Value(feature)->num_nodes)) {
	ErrorSet("index %d is out of range", i);
	Boa_Release(t);
	return TPM_BOTTOM;
      }
      if(Boa_Int_Value(Boa_Tuple_Item(t, 1)) == 0) {
	/* recompute probabilities */
	ComputeProbs(Boa_SpanTree_Value(feature), data);
	Boa_Int_Value(Boa_Tuple_Item(t, 1)) = 1;
      }
      p = NodeProb(i);
      not_for *= 1-p;
      not_against *= p;
    }
    else {
      fprintf(stderr, "Classifier: unknown feature type ");
      Boa_PrintNL(stderr, feature->type);
    }
    Boa_Release(t);
  }}

#if 0
  /* force 1 or -1 */
  if(not_against > not_for) { not_against = 1; not_for = 0; }
  else { not_against = 0; not_for = 1; }
#endif

  BOA_TYPECHECK_GET(T_Float);
  if(value_return) *value_return = Boa_Float_Create(not_against - not_for);
  return TPM_OK;
}

/* index is a List of (feature, key) tuples.
 * value is a float certainty factor.
 */
static int ObjSet(Obj self, Obj index, int flags, Obj value)
{
  Classifier c = (Classifier)self->value;
  float certainty;
  List pairs;
  Obj pair;
  if(index->type != T_List) return TPM_BOTTOM;
  pairs = Boa_List_Value(index);
  if(value->type != T_Float) return TPM_BOTTOM;
  certainty = Boa_Float_Value(value);
  /* assumes we won't get an error */
  Boa_Release(value);

  {ListIter(p, pair, pairs) {
    Obj feature, t;
    int i;
    SpanTree st;
    TreeData data;

    assert(Boa_Tuple_Length(pair) == 2);
    feature = Boa_Tuple_Item(pair, 0);
    assert(feature->type->type == T_SpanTree);
    st = Boa_SpanTree_Value(feature);
    i = Boa_Int_Value(Boa_Tuple_Item(pair, 1));

    if(Boa_ObjGet(c->features, feature, 0, NULL, &t) 
       != TPM_OK) {
      int j;
      char str[64];
      /* haven't seen this feature before */
      /* create an array of TreeData for this feature */
      sprintf(str, "(array %d (array %d char))", 
	      st->num_nodes, sizeof(struct TreeData));
      t = Boa_Create(Boa_Type_Create(str));
      data = (TreeData)t->value;
      /* initialize the treeData */
      for(j=0;j<st->num_nodes;j++) {
	SpanTreeNode node = &st->node[j];
	/* add myself as a child of my parent */
	if(node->parent >= 0) {
	  if(!data[node->parent].children) {
	    /* don't create a child list until we have to */
	    data[node->parent].children = ListCreate(NULL);
	  }
	  ListAddFront(data[node->parent].children, (Any)j);
	}
	data[j].b = 1;
      }
      pair = Boa_Create(Boa_Type_Create("(tuple 2)"));
      Boa_Tuple_Item(pair, 0) = t;
      Boa_Tuple_Item(pair, 1) = Boa_Int_Create(0);  /* updated flag */
      Boa_ObjSet(c->features, feature, BOA_INSERT, pair);
    }
    else {
      data = (TreeData)Boa_Tuple_Item(t, 0)->value;
      Boa_Int_Value(Boa_Tuple_Item(t, 1)) = 0; /* clear updated flag */
    }

    /* update the beta value of this node */
    if(certainty <= 0) data[i].b = 1/(1+certainty);
    else data[i].b = 1 - certainty;
  }} 
  return TPM_OK;
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  Generic_SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Err_StringGet, Err_StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_Classifier_SpanTree_Define(void)
{
  T_Classifier_SpanTree = Boa_SimpleType_Create("classifier_spanTree", &methods, 
					      sizeof(struct Classifier), 
					      sizeof(Obj));
  return T_Classifier_SpanTree;
}
