/* Binary classifier.
 * SetObj with a sample and a number in [-1, 1] trains the classifier.
 * GetObj with a sample returns a number in [-1, 1], where zero means 
 * "no opinion".
 *
 * A sample is a List of (feature, key) pairs, where
 * key is an index into the feature object.
 * E.g. if feature is a Hier then key is a node number.
 *
 * Uses an IdDict to store the features.
 * For each feature, there is associated information for performing
 * classification.  For example, pos/neg counts for each node of a Hier.
 *
 * This classifier is an optimistic version of SC.  Each node containing the 
 * query is checked for its probability of accepting/rejecting the query.
 */
#include <tpm/list.h>
#include <boa/boa.h>
#include <tpm/hier.h>

#define DEBUG 0

typedef struct Classifier {
  Obj features;
} *Classifier;

/* Globals *******************************************************************/
Type T_Classifier_SSC = NULL;

/* Functions *****************************************************************/

static void Init(Obj self)
{
  Classifier c = (Classifier)self->value;
  c->features = Boa_Create(T_IdDict);
}

static void Final(Obj self)
{
  Classifier c = (Classifier)self->value;
  Boa_Release(c->features);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Classifier c = (Classifier)self->value;
  func(c->features, data);
}

typedef struct TreeData {
  /* probability of not having a pos and not having a neg, resp */
  float pos, neg;
} *TreeData;

static void HierInstanceWeight(Hier hier, int i, TreeData data,
			       float *not_for, float *not_against)
{
  /* climb the tree, collecting probabilities */
  while(i >= 0) {
    HierNode node = HierNodeAt(hier, i);
    /* not in if any neg or no pos */
    *not_for *= 1 - data[i].neg * (1 - data[i].pos);
    /* not against if no neg or any pos */
    *not_against *= 1 - (1 - data[i].neg) * data[i].pos;
    i = node->parent;
#if DEBUG
    fprintf(stderr, "not_for: %g not_against: %g\n", *not_for, *not_against);
#endif
  }
}

#if 0
/* Returns the data of all instances in the class, i.e. with weight > 0.
 * classifier is a Classifier.
 * features is an Ord.
 * If compare != NULL, only instance data matching item will be returned.
 */
List BoaClassifier_AllMatching(Obj self, Obj features, 
			       CmpFunc *compare,
			       Any item, CmpFunc *include_compare)
{
  Classifier c = (Classifier)self->value;
  int f, i;
  float not_for, not_against;
  List in = ListCreate(NULL), out = ListCreate(NULL);

  /* loop features */
  for(f=0;f<Boa_Ord_Length(features);f++) {
    Obj feature = Boa_Ord_Item(features, f);
    Obj t;
    if(Boa_ObjGet(c->features, feature, 0, NULL, &t) != TPM_OK) {
      /* haven't seen this feature before */
      continue;
    }

    if(feature->type->type == T_Hier) {
      /* Loop all nodes and collect in/out */
      int i;
      Hier hier = Boa_Hier_Value(feature);
      TreeData data = (TreeData)t->value;
      for(i=0;i<hier->num_nodes;i++) {
	HierNode node = HierNodeAt(hier, i);
	if(node->children) continue;
	if(include_compare && include_compare(&node->data, item)) continue;
	not_for = not_against = 1;
	HierInstanceWeight(hier, i, data, &not_for, &not_against);
	if(not_against > not_for) ListAddFront(in, &node->data);
	else if(not_for > not_against) ListAddFront(out, &node->data);
      }
    }
    else {
      fprintf(stderr, "Classifier: unknown feature type ");
      Boa_PrintNL(stderr, feature->type);
    }
    Boa_Release(t);
  }
#if 0
  /* subtract out from in */
  {Any data; ListIter(p, data, out) {
    ListRemoveAll(in, data, compare);
  }}
#endif
  ListFree(out);
  return in;
}
#endif

/* index is a List of (feature, key) tuples.
 * returns a float certainty factor.
 */
static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Classifier c = (Classifier)self->value;
  List pairs;
  Obj pair;
  float not_for, not_against;
  if(index->type != T_List) return TPM_BOTTOM;
  pairs = Boa_List_Value(index);

  not_for = not_against = 1;
  {ListIter(p, pair, pairs) {
    Obj feature, t;

    assert(Boa_Tuple_Length(pair) == 2);
    feature = Boa_Tuple_Item(pair, 0);
    if(Boa_ObjGet(c->features, feature, 0, NULL, &t) 
       != TPM_OK) {
      /* haven't seen this feature before */
      continue;
    }

    if(feature->type->type == T_Hier) {
#if 0
      float nf, na, p;
      nf = na = 1;
      HierInstanceWeight(Boa_Hier_Value(feature), 
			 Boa_Int_Value(Boa_Tuple_Item(pair, 1)),
			 (TreeData)t->value,
			 &nf, &na);
      p = (na - nf + 1)/2;
      not_for *= 1-p;
      not_against *= p;
#else
      HierInstanceWeight(Boa_Hier_Value(feature), 
			 Boa_Int_Value(Boa_Tuple_Item(pair, 1)),
			 (TreeData)t->value,
			 &not_for, &not_against);
#endif
    }
    else {
      fprintf(stderr, "Classifier: unknown feature type ");
      Boa_PrintNL(stderr, feature->type);
    }
    Boa_Release(t);
  }}

  BOA_TYPECHECK_GET(T_Float);
  if(value_return) *value_return = Boa_Float_Create(not_against - not_for);
  return TPM_OK;
}

/* index is a List of (feature, key) tuples.
 * value is a float certainty factor.
 */
static int ObjSet(Obj self, Obj index, int flags, Obj value)
{
  Classifier c = (Classifier)self->value;
  float certainty;
  List pairs;
  Obj pair;
  if(index->type != T_List) return TPM_BOTTOM;
  pairs = Boa_List_Value(index);
  if(value->type != T_Float) return TPM_BOTTOM;
  certainty = Boa_Float_Value(value);
  /* assumes we won't get an error */
  Boa_Release(value);

  {ListIter(p, pair, pairs) {
    Obj feature, t;
    int i;
    Hier hier;
    TreeData data;

    assert(Boa_Tuple_Length(pair) == 2);
    feature = Boa_Tuple_Item(pair, 0);
    assert(feature->type->type == T_Hier);
    hier = Boa_Hier_Value(feature);
    i = Boa_Int_Value(Boa_Tuple_Item(pair, 1));

    if(Boa_ObjGet(c->features, feature, 0, NULL, &t) 
       != TPM_OK) {
      int j;
      char str[64];
      /* haven't seen this feature before */
      /* create an array of pos,neg counts for this feature */
      sprintf(str, "(array %d (array 2 float))", hier->num_nodes);
      t = Boa_Create(Boa_Type_Create(str));
      data = (TreeData)t->value;
      /* initialize the treeData */
      for(j=0;j<hier->num_nodes;j++) data[j].pos = data[j].neg = 1;
      Boa_ObjSet(c->features, feature, BOA_INSERT, Boa_Hold(t));
    }
    else data = (TreeData)t->value;

    /* climb the tree, updating probabilities */
    while(i >= 0) {
      HierNode node = HierNodeAt(hier, i);
      if(certainty > 0) data[i].pos *= 1 - certainty;
      else data[i].neg *= 1 + certainty;
      i = node->parent;
    }
    Boa_Release(t);
  }} 
  return TPM_OK;
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  Generic_SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Err_StringGet, Err_StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_Classifier_SSC_Define(void)
{
  T_Classifier_SSC = Boa_SimpleType_Create("classifier_ssc", &methods, 
					   sizeof(struct Classifier), 
					   sizeof(Obj));
  return T_Classifier_SSC;
}
