/* Binary classifier.
 * SetObj with a sample and a number in [-1, 1] trains the classifier.
 * GetObj with a sample returns a number in [-1, 1], where zero means 
 * "no opinion".
 *
 * A sample is a List of (feature, key) pairs, where
 * key is an index into the feature object.
 * E.g. if feature is a Hier then key is a node number.
 *
 * This is a generic tree-structured classifier.  It uses another classifier
 * for the splits, which are constructed greedily as in CLS.
 */
#include <tpm/tree.h>
#include "boa_learn.h"

#define DEBUG 0
#define TOLERANCE 1e-2
/* leaf variance */
#define STOP 1e-2

typedef struct Classifier {
  Tree tree;     /* classification tree; constant classifier at each node */
  Obj examples;  /* List of instances */
  Obj weights;   /* List of weights (one for each instance) */
  Type splitter; /* Type of the split classifier */
} *Classifier;

/* Globals *******************************************************************/
Type T_Classifier_Tree;

/* Functions *****************************************************************/

typedef struct NodeData {
  Obj split;     /* classifier: weight -1 means left, 1 means right */
  float average; /* average weight of an example reaching this node */
} *NodeData;

static void NodeDataFree(NodeData data)
{
  if(data->split) Boa_Release(data->split);
  free(data);
}

static void Init(Obj self)
{
  Classifier c = (Classifier)self->value;
  c->splitter = Boa_Hold(T_Classifier_Cover);
  c->examples = Boa_Create(T_List);
  c->weights = Boa_Create(T_List);
}

static void Final(Obj self)
{
  Classifier c = (Classifier)self->value;
  if(c->tree) TreeFree(c->tree, (FreeFunc*)NodeDataFree);
  Boa_Release(c->examples);
  Boa_Release(c->weights);
  Boa_Release(c->splitter);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Classifier c = (Classifier)self->value;
  func(c->examples, data);
  func(c->weights, data);
  func(c->splitter, data);
}

/* compute the average weight */
static float AverageWeight(List weights)
{
  Obj w;
  float pos, neg;
  pos = neg = 0;
  {ListIter(p, w, weights) {
    float weight = Boa_Float_Value(w);
    if(weight > 0) pos += weight;
    else neg += weight;
  }}
  /* if the total membership in this node is tiny, just pick 1 or -1. */
  if((pos > -neg) && (pos + neg < TOLERANCE)) {
    return 1;
  }
  else if((-neg > pos) && (- neg - pos < TOLERANCE)) {
    return -1;
  }
  return (pos + neg) / (pos - neg);
}

static Tree BuildTree(List examples, List weights, Type splitter)
{
  NodeData data = AllocateC(1, struct NodeData);
  Tree tree = TreeCreate(data);
  float variance, weight;
  Obj e, w;
  List w1, w0;

  if(ListEmpty(weights)) return tree;

  data->average = AverageWeight(weights);
  variance = (1 + data->average) * (1 - data->average);

  /* time to stop? */
#if DEBUG
  fprintf(stderr, "average: %g variance: %g\n", data->average, variance);
#endif
  if(variance <= STOP) return tree;

  /* create a split */
  data->split = Boa_Create(splitter);
  {ListNode p,q;
   for(p = ListFront(examples), q = ListFront(weights);
       p && q;
       p = p->next, q = q->next) {
     w = (Obj)q->data;
     e = (Obj)p->data;
     assert(Boa_ObjSet(data->split, e, 0, Boa_Hold(w)) == TPM_OK);
  }}

  /* create two new weight lists by running the examples through the split */
  w0 = ListCreate((FreeFunc*)Boa_Free);
  w1 = ListCreate((FreeFunc*)Boa_Free);
  {ListNode p,q;
   for(p = ListFront(examples), q = ListFront(weights);
       p && q;
       p = p->next, q = q->next) {
     e = (Obj)p->data;
     assert(Boa_ObjGet(data->split, e, 0, &T_Float, &w) == TPM_OK);
     weight = Boa_Float_Value(w);
     Boa_Release(w);
     w = (Obj)q->data;
#if DEBUG
     fprintf(stderr, "%g ", Boa_Float_Value(w));
#endif
     ListAddRear(w1, Boa_Float_Create((1 + weight)/2 * Boa_Float_Value(w)));
     ListAddRear(w0, Boa_Float_Create((1 - weight)/2 * Boa_Float_Value(w)));
  }}
#if DEBUG
  fprintf(stderr, "\n");
#endif

  /* are we making progress? */
  if((AverageWeight(w0) == data->average) || 
     (AverageWeight(w1) == data->average)) {
    /* no */
    Boa_Release(data->split);
    data->split = NULL;
    ListFree(w0);
    ListFree(w1);
    return tree;
  }

  /* recurse on each list and add them as children */
  TreeAddChild(tree, BuildTree(examples, w0, splitter));
  TreeAddChild(tree, BuildTree(examples, w1, splitter));
  ListFree(w0);
  ListFree(w1);

  return tree;
}

static float Classify(Tree tree, Obj instance)
{
  NodeData data = (NodeData)tree->data;
  Obj w;
  float weight, left, right;

  /* if this a leaf? */
  if(TreeLeaf(tree)) {
    /* output the average weight */
#if DEBUG
    fprintf(stderr, "leaf: %g\n", data->average);
#endif
    return data->average;
  }

  /* ask the split for the probability of going right */
  assert(Boa_ObjGet(data->split, instance, 0, &T_Float, &w) == TPM_OK);
  weight = Boa_Float_Value(w);
  Boa_Release(w);

  if(weight < 1-TOLERANCE) {
    /* classify down the left child */
    left = Classify(tree->child, instance);
  }
  else left = 0;
  if(weight > -1+TOLERANCE) {
    /* classify down the right child */
    right = Classify(tree->child->sibling, instance);
  }
  else right = 0;
#if DEBUG
  fprintf(stderr, "balance: %g ", weight);
#endif
  weight = ((1 - weight) * left + (1 + weight) * right) / 2;
#if DEBUG
  fprintf(stderr, "average: %g actual: %g\n", data->average, weight);
#endif
  return weight;
}

/* index is a List of (feature, key) tuples.
 * returns a float certainty factor.
 */
static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Classifier c = (Classifier)self->value;

  /* this is the only possible error */
  if(index->type != T_List) return TPM_BOTTOM;

  /* build a tree if there isn't one already */
  if(!c->tree) c->tree = BuildTree(Boa_List_Value(c->examples), 
				   Boa_List_Value(c->weights), c->splitter);

  BOA_TYPECHECK_GET(T_Float);
  if(value_return) *value_return = Boa_Float_Create(Classify(c->tree, index));
  return TPM_OK;
}

/* index is a List of (feature, key) tuples.
 * value is a float certainty factor.
 */
static int ObjSet(Obj self, Obj index, int flags, Obj value)
{
  Classifier c = (Classifier)self->value;

  if(index->type != T_List) return TPM_BOTTOM;
  if(value->type != T_Float) return TPM_BOTTOM;

  /* destroy any existing tree */
  if(c->tree) {
    TreeFree(c->tree, (FreeFunc*)NodeDataFree);
    c->tree = NULL;
  }

  /* add to the example list */
  ListAddRear(Boa_List_Value(c->examples), Boa_Hold(index));
  ListAddRear(Boa_List_Value(c->weights), value);

  return TPM_OK;
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  Generic_SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Err_StringGet, Err_StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_Classifier_Tree_Define(void)
{
  T_Classifier_Tree = Boa_SimpleType_Create("classifier_tree", &methods, 
					   sizeof(struct Classifier), 
					   sizeof(Obj));
  return T_Classifier_Tree;
}
