/* Binary classifier.
 * SetObj with a sample and a number in [-1, 1] trains the classifier.
 * GetObj with a sample returns a number in [-1, 1], where zero means 
 * "no opinion".
 *
 * A sample is a List of (feature, key) pairs, where
 * key is an index into the feature object.
 * E.g. if feature is a Hier then key is a node number.
 *
 * Uses an IdDict to store the features.
 * Assumes all features are metrics.
 *
 * This classifier votes triangles formed by the query and all pairs
 * of samples.  Doesn't handle the case where there is only 1 example.
 */
#include <tpm/list.h>
#include <boa/boa.h>
#include <boa/vector.h>

#define RANDOM 0

typedef struct Classifier {
  Obj features;  /* holds array of certainties */
} *Classifier;

/* Globals *******************************************************************/
Type T_Classifier_Triangle;

/* Functions *****************************************************************/

static void Init(Obj self)
{
  Classifier c = (Classifier)self->value;
  c->features = Boa_Create(T_IdDict);
}

static void Final(Obj self)
{
  Classifier c = (Classifier)self->value;
  Boa_Release(c->features);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Classifier c = (Classifier)self->value;
  func(c->features, data);
}

static int FeatureKeyIndex(Obj feature, Obj key) 
{
  Obj tests;
  int i;
  /* assume feature is a metric object, and key is a query */
  if(Boa_StringGet(feature, "_context/test", '/', NULL, &tests) == TPM_OK) {
    assert(Boa_ObjGetInt(tests, key, &i) == TPM_OK);
    Boa_Release(tests);
    return i;
  }
  else if(key->type == T_Int) {
    return Boa_Int_Value(key);
  } 
  fprintf(stderr, "FeatureKeyIndex: don't know how to index ");
  Boa_PrintNL(stderr, key);
  return -1;
}

static int FeatureNumKeys(Obj feature)
{
  Obj tests;
  int i;
  /* assume feature is a metric object, and key is a query */
  if(Boa_StringGet(feature, "_context/test", '/', NULL, &tests) == TPM_OK) {
    assert(Boa_StringGetInt(tests, "_length", &i) == TPM_OK);
    Boa_Release(tests);
  }
  else assert(Boa_StringGetInt(feature, "_length", &i) == TPM_OK);
  return i;
}

#define COMBINE_ODDS(a, b) ((a) + (b)) / ((a) * (b) + 1)

/* index is a List of (feature, key) tuples.
 * returns a float certainty factor.
 */
static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Classifier c = (Classifier)self->value;
  List pairs;
  Obj pair;
  float weight = 0;
  int i, j;

  if(index->type != T_List) return TPM_BOTTOM;
  pairs = Boa_List_Value(index);
  {ListIter(p, pair, pairs) {
    Obj feature = Boa_Tuple_Item(pair, 0);
    Obj key = Boa_Tuple_Item(pair, 1);
    Obj t;

    if(Boa_ObjGet(c->features, feature, 0, NULL, &t) != TPM_OK) {
      /* haven't seen this feature before */
      continue;
    }
    else {
      Vector dists;
      float *certainty = (float*)Boa_Tuple_Item(t, 0)->value;
      Obj dist_list = Boa_Tuple_Item(t, 1);
      Obj examples = Boa_Tuple_Item(t, 2);
      float this_weight = 0;

      Boa_Release(t);
      /* get a vector of distances */
      if(Boa_ObjGet(feature, key, 0, NULL, &t) != TPM_OK) {
	ErrorPrint(stderr);
	continue;
      }
      dists = Boa_Vector_Value(t);

      /* loop examples */
      for(i=1;i<Boa_Index_Length(examples);i++) {
	Obj ex1 = Boa_Index_Item(examples, i);
	int ex1_index = FeatureKeyIndex(feature, ex1);
	Vector ex_dists;
	Obj o;
	assert(Boa_IntGet(dist_list, i, 0, NULL, &o) == TPM_OK);
	ex_dists = Boa_Vector_Value(o);
	Boa_Release(o);
	/* loop previous examples */
	for(j=0;j<i;j++) {
	  Obj ex2 = Boa_Index_Item(examples, j);
	  int ex2_index = FeatureKeyIndex(feature, ex2);
	  float ex_dist = ex_dists->d[ex2_index];
	  float all = (2 - ex_dist -
		       dists->d[ex1_index] - dists->d[ex2_index]) / 2;
	  float w;
	  /* assume certainties are 1 or -1 */
	  if(certainty[ex1_index] == certainty[ex2_index]) {
	    w = all / (1 - ex_dist);
	  }
	  else if(ex_dist > 0) {
	    w = (1 - dists->d[ex1_index] - all) / ex_dist;
	  }
	  else {
	    /* ex1 and ex2 are the same distance, so no combination */
	    w = 1 - dists->d[ex1_index];
	  }
#if 0
	  fprintf(stderr, "ex_dist = %g  all = %g  w = %g\n", ex_dist, all, w);
#endif
	  w = (2*w - 1) * certainty[ex1_index];
	  this_weight = COMBINE_ODDS(this_weight, w);
	}
      }
      weight += this_weight;
      Boa_Release(t);
    }
  }}

  /* clip */
  if(weight > 1) weight = 1;
  else if(weight < -1) weight = -1;

#if RANDOM
  /* random choice */
  weight = (Rand() <= (weight+1)/2) ? 1 : -1;
#endif

  BOA_TYPECHECK_GET(T_Float);
  if(value_return) *value_return = Boa_Float_Create(weight);
  return TPM_OK;
}

/* index is a List of (feature, key) tuples.
 * value is a float certainty factor.
 */
static int ObjSet(Obj self, Obj index, int flags, Obj value)
{
  Classifier c = (Classifier)self->value;
  float certainty;
  List pairs;
  Obj pair;
  if(index->type != T_List) return TPM_BOTTOM;
  pairs = Boa_List_Value(index);
  if(value->type != T_Float) return TPM_BOTTOM;
  certainty = Boa_Float_Value(value);

  {ListIter(p, pair, pairs) {
    Obj feature, key, t;
    float *data;
    List dist_list;
    Obj examples;
    int i, status;

    assert(Boa_Tuple_Length(pair) == 2);
    feature = Boa_Tuple_Item(pair, 0);
    key = Boa_Tuple_Item(pair, 1);
    status = Boa_ObjGet(c->features, feature, 0, NULL, &t);

    if(status != TPM_OK) {
      /* haven't seen this feature before */
      /* create an array of certainties for this feature (brute force) */
      int ntests = FeatureNumKeys(feature);
      char str[64];
      sprintf(str, "(array %d float)", ntests);
      t = Boa_Create(Boa_Type_Create(str));
      data = (float*)t->value;
      memset(data, 0, ntests*sizeof(float));
      pair = Boa_Create(Boa_Type_Create("(tuple 3)"));
      Boa_Tuple_Item(pair, 0) = t;
      /* create a list of distance vectors */
      Boa_Tuple_Item(pair, 1) = Boa_Create(T_List);
      dist_list = Boa_List_Value(Boa_Tuple_Item(pair, 1));
      /* create an example index */
      examples = Boa_Tuple_Item(pair, 2) = Boa_Create(T_Index);
      Boa_ObjSet(c->features, feature, BOA_INSERT, pair);
    }
    else {
      data = (float*)Boa_Tuple_Item(t, 0)->value;
      dist_list = Boa_List_Value(Boa_Tuple_Item(t, 1));
      examples = Boa_Tuple_Item(t, 2);
      Boa_Release(t);
    }

    i = FeatureKeyIndex(feature, key);
    data[i] = certainty;

    /* compute distance to all previous examples */
    if(Boa_ObjGet(feature, key, 0, NULL, &t) != TPM_OK) {
      ErrorPrint(stderr);
      continue;
    }
    ListAddRear(dist_list, t);

    /* add this instance to the index of examples */
    assert(Boa_IntSet(examples, Boa_Index_Length(examples), BOA_INSERT, 
		      Boa_Hold(key)) == TPM_OK);
  }}

  return TPM_OK;
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  Generic_SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Err_StringGet, Err_StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_Classifier_Triangle_Define(void)
{
  Boa_Vector_Pkg();
  T_Classifier_Triangle = Boa_SimpleType_Create("classifier_triangle", 
						&methods, 
						sizeof(struct Classifier), 
						sizeof(Obj));
  return T_Classifier_Triangle;
}
