/* Binary classifier.
 * SetObj with a sample and a number in [-1, 1] trains the classifier.
 * GetObj with a sample returns a number in [-1, 1], where zero means 
 * "no opinion".
 *
 * A sample is a List of (feature, key) pairs, where
 * key is an index into the feature object.
 * E.g. if feature is a Hier then key is a node number.
 *
 * Uses an IdDict to store the features.
 * Assumes all features are metrics.
 *
 * This classifier is a two-layer sigmoidal network.
 */
#include "twolayer.h"
#include <boa/boa.h>
#include <boa/vector.h>

#define INPUT_MAX 40
#define HIDDEN_MAX 400
#define BINARY_ENCODING 0

typedef struct Classifier {
  TwoLayerNormalized nnet;
  int updated;
} *Classifier;

/* Globals *******************************************************************/
Type T_Classifier_TwoLayer;

/* Functions *****************************************************************/

static void Init(Obj self)
{
  Classifier c = (Classifier)self->value;
#if BINARY_ENCODING
  c->nnet = TwoLayerNormalizedCreate(TwoLayerCreate(10, 1, 1));
#else
  c->nnet = TwoLayerNormalizedCreate(TwoLayerCreate(2, 1, 1));
#endif
}

static void Final(Obj self)
{
  Classifier c = (Classifier)self->value;
  TwoLayerNormalizedFree(c->nnet);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  Classifier c = (Classifier)self->value;
}

/* Modifies x to be the encoding of i. */
static void EncodeInput(int key, Vector x)
{
#if BINARY_ENCODING
  int i;
  i = key % INPUT_MAX;
  x->d[0] = i & 1;
  x->d[1] = (i >> 1) & 1;
  x->d[2] = (i >> 2) & 1;
  x->d[3] = (i >> 3) & 1;
  x->d[4] = (i >> 4) & 1;
  i = key / INPUT_MAX;
  x->d[5] = i & 1;
  x->d[6] = (i >> 1) & 1;
  x->d[7] = (i >> 2) & 1;
  x->d[8] = (i >> 3) & 1;
  x->d[9] = (i >> 4) & 1;
#else
  x->d[0] = key % INPUT_MAX;
  x->d[1] = key / INPUT_MAX;
#endif
}

/* index is a List of (feature, key) tuples.
 * returns a float certainty factor.
 */
static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  Classifier c = (Classifier)self->value;
  List pairs;
  Obj pair;
  float weight = 0;

  if(index->type != T_List) return TPM_BOTTOM;

  if(!c->updated) {
    int i,k;
    for(;;) {
      int ninput, noutput, nhidden;
      for(k=0;;k++) {
	if(TwoLayerNormalizedTrain(c->nnet) == TPM_OK) break;
	if(k == 3) break;
	fprintf(stderr, "retrying\n");
	/* reset the weights */
	for(i=0;i<c->nnet->net->IH->len;i++) {
	  c->nnet->net->IH->d[i] = (Rand()-0.5)/1000;
	}
	for(i=0;i<c->nnet->net->HO->len;i++) {
	  c->nnet->net->HO->d[i] = (Rand()-0.5)/1000;
	}
      }
      if(k < 3) break;
      /* increase the number of hidden units */
      ninput = c->nnet->net->IH->cols-1;
      noutput = c->nnet->net->HO->rows;
      nhidden = c->nnet->net->IH->rows;
      if(nhidden == HIDDEN_MAX) break;
      TwoLayerFree(c->nnet->net);
      c->nnet->net = TwoLayerCreate(ninput, noutput, nhidden+1);
      fprintf(stderr, "increasing nhidden to %d\n", nhidden+1);
    }
#if 0
    MatrixPrint(stderr, c->nnet->net->IH);
    MatrixPrint(stderr, c->nnet->net->HO);
#endif
    c->updated = 1;
  }

  pairs = Boa_List_Value(index);
  {ListIter(p, pair, pairs) {
    Vector x = VectorCreate(c->nnet->net->IH->cols-1);
    Vector y = VectorCreate(c->nnet->net->HO->rows);
    Obj key = Boa_Tuple_Item(pair, 1);
    int i, found = 0;
    assert(key->type == T_Int);
    i = Boa_Int_Value(key);
    EncodeInput(i, x);
    /* was this an example? */
    {TwoLayerExample ex; ListIter(p, ex, c->nnet->examples) {
      if(!VectorCompare(ex->x, x)) {
	VectorMove(y, ex->y);
	found = 1;
	break;
      }
    }}
    if(!found) TwoLayerNormalizedEval(c->nnet, x, y);
    weight = y->d[0] * 2 - 1;
    VectorFree(x);
    VectorFree(y);
  }}


  /* clip */
  if(weight > 1) weight = 1;
  else if(weight < -1) weight = -1;

  BOA_TYPECHECK_GET(T_Float);
  if(value_return) *value_return = Boa_Float_Create(weight);
  return TPM_OK;
}

/* index is a List of (feature, key) tuples.
 * value is a float certainty factor.
 */
static int ObjSet(Obj self, Obj index, int flags, Obj value)
{
  Classifier c = (Classifier)self->value;
  float certainty;
  List pairs;
  Obj pair;
  if(index->type != T_List) return TPM_BOTTOM;
  pairs = Boa_List_Value(index);
  if(value->type != T_Float) return TPM_BOTTOM;
  certainty = Boa_Float_Value(value);

  {ListIter(p, pair, pairs) {
    Vector x = VectorCreate(c->nnet->net->IH->cols-1);
    Vector y = VectorCreate(c->nnet->net->HO->rows);
    Obj key = Boa_Tuple_Item(pair, 1);
    int i;
    assert(key->type == T_Int);
    i = Boa_Int_Value(key);
    EncodeInput(i, x);
    y->d[0] = (certainty + 1)/2;
    TwoLayerNormalizedAddExample(c->nnet, TwoLayerExampleCreate(x, y));
    break;
  }}
  c->updated = 0;

  return TPM_OK;
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  Generic_SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  Err_StringGet, Err_StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, ObjSet
};

Obj Boa_Classifier_TwoLayer_Define(void)
{
  Boa_Vector_Pkg();
  T_Classifier_TwoLayer = Boa_SimpleType_Create("classifier_twoLayer", &methods, 
					      sizeof(struct Classifier), 
					      sizeof(Obj));
  return T_Classifier_TwoLayer;
}
