/* Definition of the learn package. Add it to Tcl with
 *   Learntcl_Init(interp);
 * Commands:
 *   create features
 *   mcreate classifier
 *
 * Features is an Ord which you create using the Boa package.
 * A classifier allows you to set integers to float certainties
 * and get the certainty for other integers.
 */
#include <tcl.h>
#include "tcl_macros.h"
#include <tpm/boa_learn.h>
#include <tpm/matrix.h>

/* Prototypes ****************************************************************/
int Learntcl_Init(Tcl_Interp *interp);

/* Functions *****************************************************************/
static int LearnCmd(ClientData clientData, Tcl_Interp *interp,
		    int argc, char *argv[])
{
  Obj obj;
  Err(argc < 2, "missing directive: must be one of: create, mcreate");
  if(!strcmp(argv[1], "getAll")) {
    static Obj tests = NULL;
    int i;
    Vector v;
    Obj results;
    Err(argc != 4, "usage: learn getAll classifier vector");
    GetPtr(argv[2], obj);
    GetPtr(argv[3], v);
    if(!tests) {
      tests = Boa_Create(T_Index);
      for(i=0;i<v->len;i++) {
	assert(Boa_IntSet(tests, i, BOA_INSERT, Boa_Int_Create(i)) == TPM_OK);
      }
    }
    /* try to make the requests as a unit */
    if(0 && Boa_ObjGet(obj, tests, 0, &T_Index, &results) == TPM_OK) {
      fprintf(stderr, "block\n");
      for(i=0;i<v->len;i++) {
	v->d[i] = Boa_Float_Value(Boa_Index_Item(results, i));
      }
      Boa_Release(results);
    }
    else {
      /* failed; do them individually */
      for(i=0;i<v->len;i++) {
	Obj t;
	Obj instance = Boa_Index_Item(tests, i);
	if(Boa_ObjGet(obj, instance, 0, &T_Float, &t) != TPM_OK) {
	  ErrorPrepend("Classifier get failed for instance %d", i);
	  ErrorPrint(stderr);
	  continue;
	}
	v->d[i] = Boa_Float_Value(t);
	Boa_Release(t);
      }
    }
  }
  else if(!strcmp(argv[1], "errors")) {
    Vector v, guess, ideal;
    int i, count = 0;
    Err(argc != 5, "usage: learn errors guess ideal result");
    GetPtr(argv[2], guess);
    GetPtr(argv[3], ideal);
    GetPtr(argv[4], v);
    for(i=0;i<v->len;i++) {
      if(guess->d[i] * ideal->d[i] <= 0) {
	v->d[i] = 1;
	count++;
      }
      else v->d[i] = 0;
    }
    ReturnInt(count);
  }
  else if(!strcmp(argv[1], "accumulate")) {
    Matrix cooc, weights;
    int i,j;
    Err(argc != 4, "usage: learn accumulate cooc weights");
    GetPtr(argv[2], cooc);
    GetPtr(argv[3], weights);
    for(i=0;i<weights->len;i++) {
      if(weights->d[i] > 0) {
	for(j=i;j<weights->len;j++) {
	  /* count disagreements */
	  if((weights->d[j] < 0) || 
	     ((weights->d[j] == 0) && (Rand() > 0.5))) {
	    cooc->d2[i][j]++;
	    cooc->d2[j][i]++;
	  }
	}
      }
      else if(weights->d[i] < 0) {
	for(j=i;j<weights->len;j++) {
	  if((weights->d[j] > 0) ||
	     ((weights->d[j] == 0) && (Rand() > 0.5))) {
	    cooc->d2[i][j]++;
	    cooc->d2[j][i]++;
	  }
	}
      }
    }
  }
  else if(!strcmp(argv[1], "entropy")) {
    /* Return the average entropy of an element in the matrix */
    Matrix cooc;
    double sum = 0;
    int i;
    Err(argc != 3, "usage: learn entropy cooc");
    GetPtr(argv[2], cooc);
    for(i=0;i<cooc->len;i++) {
      if((cooc->d[i] > 0) && (cooc->d[i] < 1)) {
	sum -= cooc->d[i] * log(cooc->d[i]) + 
	  (1 - cooc->d[i]) * log(1 - cooc->d[i]);
      }
    }
    ReturnDouble(sum / cooc->len);
  }
  else if(!strcmp(argv[1], "create")) {
    Obj features; /* Ord */
    Type classifier_t;
    Err(argc < 3, "usage: learn create features ?type?");
    GetPtr(argv[2], features);
    Err(features->type != T_Ord, "features must be an Ord");
    if(argc > 3) GetPtr(argv[3], classifier_t);
    else classifier_t = NULL;
    obj = Boa_Classifier_Create(classifier_t, features, NULL, NULL);
    ReturnPtr(obj);
  }
  else if(!strcmp(argv[1], "mcreate")) {
    Obj classifier;
    Err(argc != 3, "usage: learn mcreate classifier");
    GetPtr(argv[2], classifier);
    obj = Boa_MultiClass_Create(classifier);
    ReturnPtr(obj);
  }
  else Err(1, "unknown directive");
  return TCL_OK;
}

int Learntcl_Init(Tcl_Interp *interp)
{
  if(!Tcl_PkgRequire(interp, "boa", "1", 0)) return TCL_ERROR;
  Boa_Hier_Pkg();
#if 0
  Boa_SpanTree_Pkg();
#endif
  Boa_Learn_Pkg();
  Tcl_CreateCommand(interp, "learn", LearnCmd, NULL, NULL);
  return Tcl_PkgProvide(interp, "learnTcl", "1");
}
