/* Module for two-layer sigmoidal networks. */
#include "twolayer.h"
#include <tpm/optim.h>

#define DEBUG 0
/* 1 for logistic output units */
#define LOGISTIC 0

/* Prototypes ****************************************************************/
TwoLayer TwoLayerCreate(int num_input, int num_output, int num_hidden);
void TwoLayerFree(TwoLayer net);
void TwoLayerEval(TwoLayer net, Vector x, Vector y);
double TwoLayerGradient(TwoLayer net, Vector x, Vector y, Matrix IHd, Matrix HOd);
void TwoLayerVHessian(TwoLayer net, Vector x, Vector y, 
		      Matrix IHv, Matrix HOv, Matrix IHh, Matrix HOh);

TwoLayerExample TwoLayerExampleCreate(Vector x, Vector y);
void TwoLayerExampleFree(TwoLayerExample ex);

int TwoLayerTrainSteepest(TwoLayer net, List examples, double rate, int epochs);
int TwoLayerTrainSCG(TwoLayer net, List examples, int epochs);
int TwoLayerTrainCG(TwoLayer net, List examples);

TwoLayerNormalized TwoLayerNormalizedCreate(TwoLayer net);
void TwoLayerNormalizedFree(TwoLayerNormalized nnet);
void TwoLayerNormalizedAddExample(TwoLayerNormalized nnet, TwoLayerExample ex);
void TwoLayerNormalizedEval(TwoLayerNormalized nnet, Vector x, Vector y);
int TwoLayerNormalizedTrain(TwoLayerNormalized nnet);

/* Functions *****************************************************************/

TwoLayer TwoLayerCreate(int num_input, int num_output, int num_hidden)
{
  TwoLayer net = Allocate(1, struct TwoLayer);
  net->IH = MatrixCreate(num_hidden, num_input + 1);
  net->HO = MatrixCreate(num_output, num_hidden + 1);
  /* Initialize to small random values. */
  VectorApply((Vector)net->IH, (MathFunc*)Rand);
  VectorAddValue((Vector)net->IH, -0.5);
  VectorScale((Vector)net->IH, 0.01*net->IH->cols);
  VectorApply((Vector)net->HO, (MathFunc*)Rand);
  VectorAddValue((Vector)net->HO, -0.5);
  VectorScale((Vector)net->HO, 0.01*net->HO->cols);
  return net;
}

void TwoLayerFree(TwoLayer net)
{
  MatrixFree(net->IH);
  MatrixFree(net->HO);
  free(net);
}

static double Logit(double x)
{
  return (double)1/(1+exp(-x));
}

/* Mutates y to be the output of the network on x. */
void TwoLayerEval(TwoLayer net, Vector x, Vector y)
{
  Vector h = VectorCreate(net->HO->cols);

  /* xb is x plus a bias element */
  Vector xb = VectorCreate(x->len + 1);
  VectorMove(xb, x);
  xb->d[xb->len-1] = 1;
  /* this will set all but the last element of h */
  MatrixVectorMultiply(h, net->IH, xb);
  VectorFree(xb);
  VectorApply(h, (MathFunc*)Logit);

  /* set the last element to 1 (bias) */
  h->d[h->len-1] = 1;
  MatrixVectorMultiply(y, net->HO, h);
  VectorFree(h);
#if LOGISTIC
  VectorApply(y, (MathFunc*)Logit);
#endif
}

/* Adds to IHg and HOg the error gradient for desired mapping (x,y).
 * Returns the magnitude of the output error.
 */
double TwoLayerGradient(TwoLayer net, Vector x, Vector y, Matrix IHg, Matrix HOg)
{
  int i,j;
  double error;
  Vector hd = VectorCreate(net->IH->rows);
  /* 1. Compute and save a forward pass **************************************/
  Vector yd = VectorCreate(y->len);
  Vector h = VectorCreate(net->HO->cols);

  /* xb is x plus a bias element */
  Vector xb = VectorCreate(x->len + 1);
  VectorMove(xb, x);
  xb->d[xb->len-1] = 1;
  /* this will set all but the last element of h */
  MatrixVectorMultiply(h, net->IH, xb);
  VectorApply(h, (MathFunc*)Logit);

  /* set the last element to 1 (bias) */
  h->d[h->len-1] = 1;
  MatrixVectorMultiply(yd, net->HO, h);
#if LOGISTIC
  VectorApply(yd, (MathFunc*)Logit);
#endif

  /* 2. Propagate errors backward ********************************************/

  assert(IHg->rows == net->IH->rows);
  assert(IHg->cols == net->IH->cols);
  assert(HOg->rows == net->HO->rows);
  assert(HOg->cols == net->HO->cols);

  /* compute difference betw y and yd. */
  VectorSubtract(yd, y);
  error = VectorDot(yd, yd);

#if DEBUG > 1
  fprintf(stderr, "Error at output: ");
  VectorPrint(stderr, yd);
#endif

  /* assign credit to elements of HO. */
  for(i=0;i<HOg->rows;i++) {
    for(j=0;j<HOg->cols;j++) {
      HOg->d2[i][j] += h->d[j] * yd->d[i];
    }
  }

  /* propagate backward through HO (ignoring bias) */
  /* this is same as MatrixVectorMultiply with transpose(HO) */
  for(i=0;i<net->HO->cols-1;i++) {
    double sum = 0;
    for(j=0;j<net->HO->rows;j++) {
      sum += net->HO->d2[j][i] * yd->d[j];
    }
    hd->d[i] = sum;
  }
  VectorFree(yd);
  /* multiply by derivative of logistic = h*(1-h) (omitting last elt) */
  h->len--;
  VectorMultiply(hd, h);
  VectorScale(h, -1);
  VectorAddValue(h, 1);
  VectorMultiply(hd, h);
  h->len++;
  VectorFree(h);

#if DEBUG > 1
  fprintf(stderr, "Error at hidden: ");
  VectorPrint(stderr, hd);
#endif

  /* assign credit to elements of IH. */
  for(i=0;i<IHg->rows;i++) {
    for(j=0;j<IHg->cols;j++) {
      IHg->d2[i][j] += xb->d[j] * hd->d[i];
    }
  }
  VectorFree(hd);
  VectorFree(xb);
  return error;
}

/* Adds to IHh and HOh the error Hessian for desired mapping (x,y)
 * multiplied by the vector IHv or HOv, respectively.
 * The algorithm comes from Bishop, "Neural Networks for Pattern Recognition,"
 * page 159.
 */
void TwoLayerVHessian(TwoLayer net, Vector x, Vector y, 
		      Matrix IHv, Matrix HOv, Matrix IHh, Matrix HOh)
{
  Vector dh;
  Vector hd = VectorCreate(net->IH->rows);
  Vector rh = VectorCreate(net->HO->cols);
  Vector ry = VectorCreate(y->len);
  Vector t;
  int i,j;

  /* 1. Compute and save a forward pass **************************************/
  Vector yd = VectorCreate(y->len);
  Vector h = VectorCreate(net->HO->cols);

  /* xb is x plus a bias element */
  Vector xb = VectorCreate(x->len + 1);
  VectorMove(xb, x);
  xb->d[xb->len-1] = 1;
  /* this will set all but the last element of h */
  MatrixVectorMultiply(h, net->IH, xb);
  VectorApply(h, (MathFunc*)Logit);

  /* set the last element to 1 (bias) */
  h->d[h->len-1] = 1;
  MatrixVectorMultiply(yd, net->HO, h);
#if LOGISTIC
  VectorApply(yd, (MathFunc*)Logit);
#endif
  VectorSubtract(yd, y);

  /* Compute a forward Hessian pass ******************************************/
  assert(IHh->rows == net->IH->rows);
  assert(IHh->cols == net->IH->cols);
  assert(HOh->rows == net->HO->rows);
  assert(HOh->cols == net->HO->cols);

  assert(IHv->rows == net->IH->rows);
  assert(IHv->cols == net->IH->cols);
  assert(HOv->rows == net->HO->rows);
  assert(HOv->cols == net->HO->cols);

  /* dh = h*(1-h) */
  dh = VectorCopy(h);
  VectorScale(dh, -1);
  VectorAddValue(dh, 1);
  VectorMultiply(dh, h);

  /* this will set all but the last element of rh */
  MatrixVectorMultiply(rh, IHv, xb);
  rh->d[rh->len-1] = 0;
  VectorMultiply(rh, dh);

  t = VectorCreate(y->len);
  MatrixVectorMultiply(t, net->HO, rh);
  MatrixVectorMultiply(ry, HOv, h);
  VectorAdd(ry, t);
  VectorFree(t);
#if LOGISTIC
  VectorAdd(yd, y);
  VectorMultiply(ry, yd);
  VectorScale(yd, -1);
  VectorAddValue(yd, 1);
  VectorMultiply(ry, yd);
  VectorAddValue(yd, -1);
  VectorScale(yd, -1);
  VectorSubtract(yd, y);
#endif

  /* Compute a backward Hessian pass *****************************************/
  /* assign credit to elements of HO. */
  for(i=0;i<HOh->rows;i++) {
    for(j=0;j<HOh->cols;j++) {
      HOh->d2[i][j] += ry->d[i] * h->d[j] + yd->d[i] * rh->d[j];
    }
  }

  /* this is same as MatrixVectorMultiply with transpose(HO) */
  for(i=0;i<net->HO->cols-1;i++) {
    double sum = 0;
    for(j=0;j<net->HO->rows;j++) {
      sum += net->HO->d2[j][i] * yd->d[j];
    }
    hd->d[i] = sum;
  }
  VectorMultiply(hd, rh);
  VectorFree(rh);
  /* multiply by second derivative of logistic (omitting last elt) */
  h->len--;
  VectorScale(h, -2);
  VectorAddValue(h, 1);
  VectorMultiply(hd, h);
  h->len++;
  VectorFree(h);

  /* this is same as MatrixVectorMultiply with transpose(HO) */
  for(i=0;i<net->HO->cols-1;i++) {
    double sum = 0;
    for(j=0;j<net->HO->rows;j++) {
      sum += net->HO->d2[j][i] * ry->d[j] + 
	     HOv->d2[j][i] * yd->d[j];
    }
    hd->d[i] += sum;
  }
  VectorFree(ry);
  VectorFree(yd);
  VectorMultiply(hd, dh);
  VectorFree(dh);

  /* assign credit to elements of IH. */
  for(i=0;i<IHh->rows;i++) {
    for(j=0;j<IHh->cols;j++) {
      IHh->d2[i][j] += xb->d[j] * hd->d[i];
    }
  }
  VectorFree(hd);
  VectorFree(xb);
}


/* TwoLayerExample ***********************************************************/

TwoLayerExample TwoLayerExampleCreate(Vector x, Vector y)
{
  TwoLayerExample ex = Allocate(1, struct TwoLayerExample);
  ex->x = x;
  ex->y = y;
  return ex;
}

void TwoLayerExampleFree(TwoLayerExample ex)
{
  VectorFree(ex->x);
  VectorFree(ex->y);
  free(ex);
}

/* Training ******************************************************************/

/* Train a network via stochastic steepest descent. 
 * Returns TPM_OK on success.
 */
int TwoLayerTrainSteepest(TwoLayer net, List examples, double rate, 
			  int epochs)
{
  int epoch;
  TwoLayerExample ex;
  Matrix IHg, HOg;
  IHg = MatrixCreate(net->IH->rows, net->IH->cols);
  HOg = MatrixCreate(net->HO->rows, net->HO->cols);

  for(epoch=1;epoch < epochs;epoch++) {
    double sum = 0, error = 0;
#if DEBUG
    fprintf(stderr, "epoch %d\n", epoch);
#endif
    {ListIter(p, ex, examples) {
      VectorSet((Vector)IHg, 0);
      VectorSet((Vector)HOg, 0);
      error += TwoLayerGradient(net, ex->x, ex->y, IHg, HOg);
      sum += VectorMag((Vector)IHg) + VectorMag((Vector)HOg);
      VectorScale((Vector)IHg, -rate);
      VectorScale((Vector)HOg, -rate);
      VectorAdd((Vector)net->IH, (Vector)IHg);
      VectorAdd((Vector)net->HO, (Vector)HOg);
    }}
#if DEBUG
    fprintf(stderr, "error = %g\n", error);
#endif
    if(error < 1e-4) break;  /* good enough */
    if(sum < rate/1000) break;
  }
  MatrixFree(IHg);
  MatrixFree(HOg);
  if(epoch == epochs) return TPM_ERROR;
  return TPM_OK;
}

/* Modifies dir to be the new conjugate direction. */
static void CGDirection(Vector dir, Vector grad, Vector old_grad)
{
  double beta = (VectorDot(grad, grad) - VectorDot(grad, old_grad)) /
                VectorDot(old_grad, old_grad);
  VectorScale(dir, beta);
  VectorSubtract(dir, grad);
}

/* Train a network via scaled conjugate gradient. */
int TwoLayerTrainSCG(TwoLayer net, List examples, int epochs)
{
  int epoch, i, loop;
  TwoLayerExample ex;
  double error, new_error, error_delta;
  double scale = 1e-10, step;
  double delta, mu, dir_mag;
  Matrix IH_dir, HO_dir, IH_grad, HO_grad, IH_old_grad, HO_old_grad;
  Matrix IH_dir_hess, HO_dir_hess;
  IH_dir = MatrixCreate(net->IH->rows, net->IH->cols);
  HO_dir = MatrixCreate(net->HO->rows, net->HO->cols);
  IH_grad = MatrixCreate(net->IH->rows, net->IH->cols);
  HO_grad = MatrixCreate(net->HO->rows, net->HO->cols);
  IH_old_grad = MatrixCreate(net->IH->rows, net->IH->cols);
  HO_old_grad = MatrixCreate(net->HO->rows, net->HO->cols);
  IH_dir_hess = MatrixCreate(net->IH->rows, net->IH->cols);
  HO_dir_hess = MatrixCreate(net->HO->rows, net->HO->cols);
  VectorSet((Vector)IH_dir, 0);
  VectorSet((Vector)HO_dir, 0);
  /* initialize to anything but zero */
  VectorSet((Vector)IH_old_grad, 1);
  VectorSet((Vector)HO_old_grad, 1);

  for(epoch=1;epoch < epochs;epoch++) {
#if DEBUG
    fprintf(stderr, "epoch %d\n", epoch);
#endif
#if DEBUG > 1
    fprintf(stderr, "network:\n");
    MatrixPrint(stderr, net->IH);
    MatrixPrint(stderr, net->HO);
#endif

#if 0
    /* add noise to the weights */
    for(i=0;i<net->IH->len;i++) net->IH->d[i] += (Rand()-0.5)/epoch;
    for(i=0;i<net->HO->len;i++) net->HO->d[i] += (Rand()-0.5)/epoch;
#endif

    /* Compute the batch gradient */
    VectorSet((Vector)IH_grad, 0);
    VectorSet((Vector)HO_grad, 0);
    error = 0;
    {ListIter(p, ex, examples) {
      error += TwoLayerGradient(net, ex->x, ex->y, IH_grad, HO_grad);
    }}
#if DEBUG
    fprintf(stderr, "error = %g\n", error);
#endif
    if(error < 1e-2) break;
    if(VectorDot((Vector)IH_grad, (Vector)IH_grad) + 
       VectorDot((Vector)HO_grad, (Vector)HO_grad) < 1e-10) break;

    /* Compute new direction */
    CGDirection((Vector)IH_dir, (Vector)IH_grad, (Vector)IH_old_grad);
    CGDirection((Vector)HO_dir, (Vector)HO_grad, (Vector)HO_old_grad);
    VectorMove(IH_old_grad, IH_grad);
    VectorMove(HO_old_grad, HO_grad);
#if DEBUG > 1
    fprintf(stderr, "dir:\n");
    MatrixPrint(stderr, IH_dir);
    MatrixPrint(stderr, HO_dir);
#endif

    /* Compute direction times Hessian */
    VectorSet((Vector)IH_dir_hess, 0);
    VectorSet((Vector)HO_dir_hess, 0);
    {ListIter(p, ex, examples) {
      TwoLayerVHessian(net, ex->x, ex->y, IH_dir, HO_dir, 
		       IH_dir_hess, HO_dir_hess);
    }}

    delta = VectorDot((Vector)IH_dir_hess, (Vector)IH_dir) + 
            VectorDot((Vector)HO_dir_hess, (Vector)HO_dir);
    dir_mag = VectorDot((Vector)IH_dir, (Vector)IH_dir) + 
              VectorDot((Vector)HO_dir, (Vector)HO_dir);
    mu = VectorDot((Vector)IH_dir, (Vector)IH_grad) + 
         VectorDot((Vector)HO_dir, (Vector)HO_grad);
    /* Loop until we decrease the error */
    loop = 0;
    do {
      if(++loop > 1000) {
	/* we're stuck */
	epoch = epochs;
	break;
      }
      /* Compute step size */
      if(delta < -scale*dir_mag) scale = -2*delta / dir_mag;
      step = -mu / (delta + scale*dir_mag);
#if DEBUG > 1
      fprintf(stderr, "step = %g, scale = %g\n", step, scale);
#endif

      /* Update weights */
      for(i=0;i<net->IH->len;i++) net->IH->d[i] += step * IH_dir->d[i];
      for(i=0;i<net->HO->len;i++) net->HO->d[i] += step * HO_dir->d[i];

      /* Compute new error */
      VectorSet((Vector)IH_grad, 0);
      VectorSet((Vector)HO_grad, 0);
      new_error = 0;
      {ListIter(p, ex, examples) {
	new_error += TwoLayerGradient(net, ex->x, ex->y, IH_grad, HO_grad);
      }}
#if DEBUG > 1
      fprintf(stderr, "new_error = %g\n", new_error);
#endif

      /* Update scale parameter */
      error_delta = 2*(delta+scale*dir_mag)*(error - new_error)/mu/mu;
#if DEBUG > 1
      fprintf(stderr, "error_delta = %g\n", error_delta);
#endif
      if(error_delta > 0.75) scale /= 2;
      else if(error_delta < 0.25) {
	scale *= 4;
	if(error_delta < 0) {
	  /* Restore old weights */
	  for(i=0;i<net->IH->len;i++) net->IH->d[i] -= step * IH_dir->d[i];
	  for(i=0;i<net->HO->len;i++) net->HO->d[i] -= step * HO_dir->d[i];
	}
      }
    } while(error_delta < 0);
  }
  MatrixFree(IH_dir);
  MatrixFree(HO_dir);
  MatrixFree(IH_grad);
  MatrixFree(HO_grad);
  MatrixFree(IH_old_grad);
  MatrixFree(HO_old_grad);
  MatrixFree(IH_dir_hess);
  MatrixFree(HO_dir_hess);
  if(epoch >= epochs) {
    fprintf(stderr, "Warning: SCG exceeded maximum number of iterations (%d)\n", epochs);
    return TPM_ERROR;
  }
  return TPM_OK;
}

typedef struct TrainData {
  TwoLayer net;
  List examples;
  Vector yd; /* scratch area */
  Matrix IHg, HOg; /* scratch area */
} *TrainData;

static double Error(Vector weights, TrainData data)
{
  double error = 0;
  TwoLayerExample ex;
  VectorUnpack(weights, (Vector)data->net->IH, (Vector)data->net->HO);
  {ListIter(p, ex, data->examples) {
    TwoLayerEval(data->net, ex->x, data->yd);
    VectorSubtract(data->yd, ex->y);
    error += VectorDot(data->yd, data->yd);
  }}
#if DEBUG
  fprintf(stderr, "error: %g\n", error);
#endif
  return error;
}

static void Gradient(Vector weights, Vector grad, TrainData data)
{
  TwoLayerExample ex;
  VectorUnpack(weights, (Vector)data->net->IH, (Vector)data->net->HO);
  VectorSet((Vector)data->IHg, 0);
  VectorSet((Vector)data->HOg, 0);
  {ListIter(p, ex, data->examples) {
    TwoLayerGradient(data->net, ex->x, ex->y, data->IHg, data->HOg);
  }}
  VectorPack((Vector)data->IHg, (Vector)data->HOg, grad);
}

/* Train a network via conjugate gradient. */
int TwoLayerTrainCG(TwoLayer net, List examples)
{
  double fmin;
  Vector weights;
  struct TrainData data;
  data.net = net;
  data.examples = examples;
  data.yd = VectorCreate(net->HO->rows);
  data.IHg = MatrixCreate(net->IH->rows, net->IH->cols);
  data.HOg = MatrixCreate(net->HO->rows, net->HO->cols);

  weights = VectorCreate(net->IH->len + net->HO->len);
  VectorPack((Vector)net->IH, (Vector)net->HO, weights);
  OptCG(weights, (OptMScalarFunc*)Error, &data, 
	(OptVectorFunc*)Gradient, &data, &fmin);
  VectorUnpack(weights, (Vector)net->IH, (Vector)net->HO);
  VectorFree(weights);
  MatrixFree(data.IHg);
  MatrixFree(data.HOg);
  VectorFree(data.yd);
  if(fmin > 1e-1) return TPM_ERROR;
  return TPM_OK;
}

/* TwoLayerNormalized ********************************************************/

/* Takes ownership of net. */
TwoLayerNormalized TwoLayerNormalizedCreate(TwoLayer net)
{
  TwoLayerNormalized nnet = Allocate(1, struct TwoLayerNormalized);
  nnet->net = net;
  nnet->examples = ListCreate((FreeFunc*)TwoLayerExampleFree);
  nnet->input_mean = VectorCreate(net->IH->cols-1);
  VectorSet(nnet->input_mean, 0);
  nnet->input_var = VectorCreate(net->IH->cols-1);
  VectorSet(nnet->input_var, 0);
  return nnet;
}

void TwoLayerNormalizedFree(TwoLayerNormalized nnet)
{
  TwoLayerFree(nnet->net);
  ListFree(nnet->examples);
  VectorFree(nnet->input_mean);
  VectorFree(nnet->input_var);
  free(nnet);
}

/* Takes ownership of ex. */
void TwoLayerNormalizedAddExample(TwoLayerNormalized nnet, TwoLayerExample ex)
{
  Vector delta = VectorCreate(nnet->input_mean->len); /* change in mean */
  Vector sqdelta = VectorCreate(nnet->input_mean->len);

  /* compute new mean */
  VectorMove(delta, ex->x);
  VectorSubtract(delta, nnet->input_mean);
  ListAddRear(nnet->examples, ex);
  VectorScale(delta, 1.0/ListSize(nnet->examples));
  VectorAdd(nnet->input_mean, delta);

  if(ListSize(nnet->examples) > 1) {
    /* compute new variance */
    VectorMove(sqdelta, ex->x);
    VectorSubtract(sqdelta, nnet->input_mean);
    VectorMultiply(sqdelta, sqdelta);
    VectorSubtract(sqdelta, nnet->input_var);
    VectorScale(sqdelta, 1.0/(ListSize(nnet->examples)-1));
    VectorAdd(nnet->input_var, sqdelta);
    VectorMultiply(delta, delta);
    VectorAdd(nnet->input_var, delta);
  }
#if 1
  fprintf(stderr, "input_mean: ");
  VectorPrint(stderr, nnet->input_mean);
  fprintf(stderr, "input_var: ");
  VectorPrint(stderr, nnet->input_var);
#endif

  VectorFree(delta);
  VectorFree(sqdelta);
}

void TwoLayerNormalizedEval(TwoLayerNormalized nnet, Vector x, Vector y)
{
  int i;
  Vector nx = VectorCreate(x->len);
  VectorMove(nx, x);
  VectorSubtract(nx, nnet->input_mean);
  for(i=0;i<nnet->input_var->len;i++) {
    if(nnet->input_var->d[i] > 1e-20)
      nx->d[i] /= sqrt(nnet->input_var->d[i]);
  }
  TwoLayerEval(nnet->net, nx, y);
  VectorFree(nx);
}

/* Train the network by normalizing the examples and 
 * alternating between CG and Steepest. 
 */
int TwoLayerNormalizedTrain(TwoLayerNormalized nnet)
{
  int i, status;
  List new_examples = ListCreate((FreeFunc*)TwoLayerExampleFree);
  TwoLayerExample ex;
  ListIter(p, ex, nnet->examples) {
    Vector nx = VectorCreate(ex->x->len);
    VectorMove(nx, ex->x);
    VectorSubtract(nx, nnet->input_mean);
    for(i=0;i<nnet->input_var->len;i++) {
      if(nnet->input_var->d[i] > 1e-20)
	nx->d[i] /= sqrt(nnet->input_var->d[i]);
    }
    ListAddRear(new_examples, TwoLayerExampleCreate(nx, VectorCopy(ex->y)));
  }
  for(i=0;i<3;i++) {
    status = TwoLayerTrainCG(nnet->net, new_examples);
    if(status == TPM_OK) break;
    status = TwoLayerTrainSteepest(nnet->net, new_examples, 0.1, 100);
    if(status == TPM_OK) break;
  }
  ListFree(new_examples);
  return status;
}

