#include "optim.h"

#define GOLD 1.6180339887499
#define GLIMIT 100
#define TINY 1e-20

/* Modifies *a, *b, and *c so that
 * *a < *c < *b and f(*a) > f(*c) < f(*b)
 */
static void Bracket(double *a, double *b, double *c, 
		    OptScalarFunc *f, Any data)
{
  double t;
  double fa, fb, fc;
  fa = f(*a, data);
  fb = f(*b, data);
  if(fb > fa) {
    t = *a; *a = *b; *b = t;
    t = fa; fa = fb; fb = t;
  }
  *c = *b + GOLD*(*b - *a);
  fc = f(*c, data);
  while(fb > fc) {
    double r = (*b - *a)*(fb - fc);
    double q = (*b - *c)*(fb - fa);
    double u, ulim, fu;
    if(fabs(q-r) >= TINY) u = q-r;
    else if(q > r) u = TINY;
    else u = -TINY;
    u = *b - ((*b - *c)*q - (*b - *a)*r)/2/u;
    ulim = *b + GLIMIT*(*c - *b);
    if((*b - u)*(u - *c) > 0) {
      fu = f(u, data);
      if(fu < fc) {
	*a = *b; *b = u;
	fa = fb; fb = fu;
	break;
      }
      else if(fu > fb) {
	*c = u;
	fc = fu;
	break;
      }
      u = *c + GOLD*(*c - *b);
      fu = f(u, data);
    }
    else if((*c - u)*(u-ulim) > 0) {
      fu = f(u, data);
      if(fu < fc) {
	*b = *c; *c = u; u = *c + GOLD*(*c - *b);
	fb = fc; fc = fu; fu = f(u, data);
      }
    }
    else if((u-ulim)*(ulim - *c) >= 0) {
      u = ulim;
      fu = f(u, data);
    }
    else {
      u = *c + GOLD*(*c - *b);
      fu = f(u, data);
    }
    *a = *b; *b = *c; *c = u;
    fa = fb; fb = fc; fc = fu;
  }
#if 1
  fprintf(stderr, "Bracket returns (%g, %g, %g)\n", *a, *b, *c);
#endif
  t = *c; *c = *b; *b = t;
}

#define ITER_MAX 100
#define CGOLD 0.38196601125011
#define ZEPS 1e-10

/* Minimize the scalar function f in the interval [a,b] via Brent's method.
 * Requires a < b.
 * Modifies *fmin_return to be the ordinate of the minimum.
 * Returns the abscissa of the minimum.
 * Algorithm taken from Numerical Recipes and Matlab optimization toolbox.
 */
double OptBrent(double a, double b, OptScalarFunc *f, Any data, 
		double *fmin_return)
{
  double tol = 1e-4;
  double d = 0, e = 0;
  double v, w, x;
  double fx, fv, fw;
  double u, fu;
  int iter;
#if 1
  x = a + CGOLD*(b-a); /* golden section to get third point */
#else
  Bracket(&a, &b, &x, f, data);
#endif
  w = v = x;
  fx = f(x, data), fv = fx, fw = fx;

  for(iter=0;iter<ITER_MAX;iter++) {
    int golden_section_step = 1;
    double xm = (a+b)/2;
    double tol1 = ZEPS*fabs(x)+tol/3;
    double tol2 = 2*tol1;
    if(fabs(x-xm) <= (tol2-(b-a)/2)) break;

    /* Construct a trial parabolic fit */
    if(fabs(e) > tol1) {
      double r = (x-w)*(fx-fv);
      double q = (x-v)*(fx-fw);
      double p = (x-v)*q - (x-w)*r;
      q = 2*(q-r);
      if(q > 0) p = -p;
      q = fabs(q);
      r = e; e = d;
      /* Is the parabola acceptable? */
      if( (fabs(p) < fabs(0.5*q*r)) && (p > q*(a-x)) && (p < q*(b-x)) ) {
	/* Yes, take the parabolic step */
	d = p/q;
	u = x+d;
	if((u-a < tol2) || (b-u < tol2)) {
	  d = tol1;
	  if(xm < x) d = -d;
	}
	golden_section_step = 0;
      }
    }

    if(golden_section_step) {
      /* Take the golden section step */
      if(x >= xm) e = a-x;
      else e = b-x;
      d = CGOLD * e;
    }

    /* Evaluate f at x+d as long as fabs(d) > tol1 */
    if(fabs(d) >= tol1) u = d;
    else if(d >= 0) u = tol1;
    else u = -tol1;
    u += x;
    fu = f(u, data);

    if(fu <= fx) {
      if(u >= x) a = x;
      else b = x;
      v = w; w = x; x = u;
      fv = fw; fw = fx; fx = fu;
    }
    else {
      if(u < x) a = u;
      else b = u;
      if(fu <= fw || w == x) {
	v = w; w = u;
	fv = fw; fw = fu;
      }
      else if(fu <= fv || v == x || v == w) {
	v = u;
	fv = fu;
      }
    }
  }
  if(iter == ITER_MAX) {
    fprintf(stderr, "Warning: OptBrent exceeded maximum number of iterations\n");
  }
  *fmin_return = fx;
  return x;
}
