#include "optim.h"

#define DEBUG 0
/* #define MIN_ZERO 1e-1 */

/* Prototypes ****************************************************************/

void OptCG(Vector x, OptMScalarFunc *f, Any fdata, OptVectorFunc *g, Any gdata,
	   double *fmin_return);
void OptPowell(Vector x, OptMScalarFunc *f, Any data, double *fmin_return);

/* Functions *****************************************************************/

typedef struct LineMinDataRec {
  Vector x, dir;
  Vector base;
  OptMScalarFunc *f;
  Any fdata;
} *LineMinData;

/* Returns f(x + a*dir) */
static double LineMinFunc(double a, LineMinData data)
{
  int i;
  for(i=0;i<data->x->len;i++) {
    data->x->d[i] = data->base->d[i] + a * data->dir->d[i];
  }
  return data->f(data->x, data->fdata);
}

/* Modifies x to be the minimum of f along the direction dir. */
static double LineMin(Vector x, Vector dir, OptMScalarFunc *f, Any fdata,
		      double *fmin_return)
{
  int i;
  double a;
  struct LineMinDataRec data;
  data.x = x;
  data.dir = dir;
  data.base = VectorCopy(x);
  data.f = f;
  data.fdata = fdata;
  a = OptBrent(-2, 2, (OptScalarFunc*)LineMinFunc, &data, fmin_return);
  for(i=0;i<x->len;i++) {
    x->d[i] = data.base->d[i] + a * dir->d[i];
  }
  VectorFree(data.base);
#if DEBUG
  fprintf(stderr, "LineMin step: %g\n", a);
#endif
  return a;
}

#define ITER_MAX 500
#define EPS 1e-10

/* Minimize a multidimensional scalar function starting at x,
 * using the vector gradient function g.
 * Modifies x to be the minimum and *fmin_return to be the function value at
 * the minimum.
 */
void OptCG(Vector x, OptMScalarFunc *f, Any fdata, OptVectorFunc *g, Any gdata,
	   double *fmin_return)
{
  int iter;
  double ftol = 1e-10;
  double grad_mag, old_grad_mag;
  Vector dir = VectorCreate(x->len);
  Vector grad = VectorCreate(x->len);
  Vector old_grad = VectorCreate(x->len);
  double fx = f(x, fdata);
  g(x, old_grad, gdata);
  old_grad_mag = VectorDot(old_grad, old_grad);
  VectorMove(dir, old_grad);
  VectorScale(dir, -1);

  for(iter=0;iter<ITER_MAX;iter++) {
    double beta;
#if DEBUG
    fprintf(stderr, "x: ");
    VectorPrint(stderr, x);
    fprintf(stderr, "dir: ");
    VectorPrint(stderr, dir);
    fprintf(stderr, "old_grad: ");
    VectorPrint(stderr, old_grad);
#endif
    /* Minimize along dir */
    LineMin(x, dir, f, fdata, fmin_return);
    if(2*fabs(*fmin_return - fx) <= ftol*(fabs(*fmin_return)+fabs(fx)+EPS))
      break;
#ifdef MIN_ZERO
    /* this only works if we are minimizing an error function */
    if(*fmin_return < MIN_ZERO) break;
#endif

    /* Compute new conjugate direction */
    fx = *fmin_return;
    g(x, grad, gdata);
    grad_mag = VectorDot(grad, grad);
    beta = (grad_mag - VectorDot(grad, old_grad)) / old_grad_mag;
    VectorScale(dir, beta);
    VectorSubtract(dir, grad);

    VectorMove(old_grad, grad);
    old_grad_mag = grad_mag;
  }
  VectorFree(dir);
  VectorFree(grad);
  VectorFree(old_grad);
  if(iter == ITER_MAX) {
    fprintf(stderr, "Warning: OptCG exceeded maximum number of iterations\n");
  }
}

#undef ITER_MAX
#define ITER_MAX 100

#include <tpm/matrix.h>

/* Minimize a multidimensional scalar function starting at x.
 * Modifies x to be the minimum and *fmin_return to be the function value at
 * the minimum.
 * Always evaluates f with x, so that x may be a Matrix or Volume.
 */
void OptPowell(Vector x, OptMScalarFunc *f, Any data, double *fmin_return)
{
  double ftol = 1e-10;
  Vector old_x = VectorCreate(x->len);
  Vector ext_x = VectorCreate(x->len);
  Vector dir = VectorCreate(x->len);
  Matrix dirs = MatrixCreate(x->len, x->len);
  int i, iter;

  /* Initialize the directions to the unit vectors */
  MatrixIdentity(dirs);
  *fmin_return = f(x, data);
  VectorMove(old_x, x);

  for(iter=0;iter<ITER_MAX;iter++) {
    double fx = *fmin_return, fex;
    int i_max = 0;
    double delta_max = 0;
    double a;
#if DEBUG
    VectorPrint(stderr, x);
#endif
    /* Minimize along each direction, remembering the direction of greatest
     * function decrease.
     */
    for(i=0;i<x->len;i++) {
      double delta;
      double old_min = *fmin_return;
      MatrixGetRow(dirs, i, dir);
      a = LineMin(x, dir, f, data, fmin_return);
      VectorScale(dir, a);
      delta = fabs(old_min - *fmin_return);
      if(delta > delta_max) { delta_max = delta; i_max = i; }
    }
    if(2*fabs(fx - *fmin_return) <= ftol*(fabs(fx)+fabs(*fmin_return))) break;
#ifdef MIN_ZERO
    /* this only works if we are minimizing an error function */
    if(*fmin_return < MIN_ZERO) break;
#endif
    /* Construct new direction from old_x to x. */
    VectorMove(dir, x);
    VectorSubtract(dir, old_x);
    VectorMove(old_x, x);
    /* And extrapolate it. */
    VectorMove(ext_x, x);
    VectorAdd(x, dir);
    /* Good extrapolation? */
    fex = f(x, data);
    VectorMove(x, ext_x);
    if(fex < fx) {
      double t = fx - *fmin_return - delta_max;
      t = 2*(fx - 2*(*fmin_return) + fex)*t*t - delta_max*(fx-fex)*(fx-fex);
      if(t < 0) {
	a = LineMin(x, dir, f, data, fmin_return);
	VectorScale(dir, a);
	/* Replace i_max with the last. */
	memcpy(dirs->d2[i_max], dirs->d2[dirs->rows-1], 
	       dirs->cols*sizeof(double));
	/* Replace the last with dir. */
	MatrixSetRow(dirs, dirs->rows-1, dir);
      }
    }
  }
  if(iter == ITER_MAX) {
    fprintf(stderr, "Warning: OptPowell exceeded maximum number of iterations\n");
  }
  VectorFree(old_x);
  VectorFree(ext_x);
  VectorFree(dir);
  MatrixFree(dirs);
}
