#include <tpm/tpm.h>
#include "matrix.h"

/* Prototypes ****************************************************************/
void VectorDCT(Vector v);
void MatrixDCT(Matrix m);

/* Functions *****************************************************************/

/* Brute force DCT in place. */
static void DCT1D(double *a, int n)
{
  int i,x;
  double *f = AllocCopy(n, double, a);
  for(i=0;i<n;i++) {
    double sum = 0;
    for(x=0;x<n;x++) {
      sum += f[x]*cos(i*(x+0.5)*M_PI/n);
    }
    a[i] = 2*sum;
  }
  free(f);
}

#define SIN_1_4 0.7071067811856476
#define COS_1_4 SIN_1_4

#define SIN_1_8 0.3826834323650898
#define COS_1_8 0.9238795325112870
#define SIN_3_8 COS_1_8
#define COS_3_8 SIN_1_8

#define SIN_1_16 0.1950903220161282
#define COS_1_16 0.9807852804032300
#define SIN_7_16 COS_1_16
#define COS_7_16 SIN_1_16

#define SIN_3_16 0.5555702330196022
#define COS_3_16 0.8314696123025450
#define SIN_5_16 COS_3_16
#define COS_5_16 SIN_3_16

/* Fast DCT in place when a is 8 elements long. */
static void DCT1D_8(double *a)
{
  double t0, t1, t2, t3, t4, t5, t6, t7;
  double t10, t11, t12, t13;
  double t14, t15, t16, t17;
  double t25, t26;

  t0 = a[7] + a[0];
  t1 = a[6] + a[1];
  t2 = a[5] + a[2];
  t3 = a[4] + a[3];
  t4 = a[3] - a[4];
  t5 = a[2] - a[5];
  t6 = a[1] - a[6];
  t7 = a[0] - a[7];

  t10 = t3 + t0;
  t11 = t2 + t1;
  t12 = t1 - t2;
  t13 = t0 - t3;

  a[0] = (t10 + t11) * 2;
  a[4] = (t10 - t11) * COS_1_4 * 2;
  a[2] = (t13*COS_1_8 + t12*SIN_1_8) * 2;
  a[6] = (t13*SIN_1_8 - t12*COS_1_8) * 2;

  t16 = (t6 + t5) * SIN_1_4;
  t15 = (t6 - t5) * COS_1_4;

  t14 = t4 + t15;
  t25 = t4 - t15;
  t26 = t7 - t16;
  t17 = t7 + t16;

  a[1] = (t17*COS_1_16 + t14*SIN_1_16) * 2;
  a[7] = (t17*COS_7_16 - t14*SIN_7_16) * 2;
  a[3] = (t26*COS_3_16 - t25*SIN_3_16) * 2;
  a[5] = (t26*COS_5_16 + t25*SIN_5_16) * 2;
}

/* Computes the DCT in place. */
void VectorDCT(Vector v)
{
  if(v->len == 8) DCT1D_8(v->d);
  else DCT1D(v->d, v->len);
}

/* Computes the DCT in place. */
void MatrixDCT(Matrix m)
{
  int i;
  if((m->rows == 8) && (m->cols == 8)) {
    for(i=0;i<m->rows;i++) DCT1D_8(m->d2[i]);
    MatrixTranspose(m);
    for(i=0;i<m->rows;i++) DCT1D_8(m->d2[i]);
    MatrixTranspose(m);
  }
  else {
    for(i=0;i<m->rows;i++) DCT1D(m->d2[i], m->cols);
    MatrixTranspose(m);
    for(i=0;i<m->rows;i++) DCT1D(m->d2[i], m->cols);
    MatrixTranspose(m);
  }
}
