#include <tpm/tpm.h>
#include "matrix.h"

/* Prototypes ****************************************************************/
void MatrixDFT(Matrix R, Matrix I);
void VectorDFT(Vector R, Vector I);

/* Functions *****************************************************************/

/* Fast DFT of a, in place. 
 * The length of both vectors, n, must be a power of 2. 
 */
static void DFT1D(double *R, double *I, int n)
{
  double u_r, u_i, w_r, w_i, t_r, t_i;
  int nv2, nm1, j, i, k;
  double ang;
  int ip, le, le1;

  nv2 = n / 2;
  nm1 = n - 1;
  j = 0;
  for(i = 0; i < nm1; i++) {
    if(i < j) { 
      t_r = R[i]; R[i] = R[j]; R[j] = t_r; 
      t_i = I[i]; I[i] = I[j]; I[j] = t_i; 
    }
    k = nv2;
    for(; (k - 1) < j;) { j = j - k; k = k / 2; }
    j = j + k;
  }
  for(le=2, le1=1; le1 < n; le<<=1, le1<<=1) {
    u_r = 1.0;
    u_i = 0.0;
    ang = (double)M_PI / le1;
    w_r = cos(ang);
    w_i = sin(ang);
    for(j = 0; j < le1; j++) {
      for(i = j; i < n; i += le) {
	ip = i + le1;
	/* t = a[ip] * u; */
	t_r = R[ip] * u_r - I[ip] * u_i;
	t_i = R[ip] * u_i + I[ip] * u_r;
	/* a[ip] = a[i] - t; */
	R[ip] = R[i] - t_r;
	I[ip] = I[i] - t_i;
	/* a[i] += t; */
	R[i] += t_r;
	I[i] += t_i;
      }
      /* u *= w; */
      t_r = u_r;
      t_i = u_i;
      u_r = t_r * w_r - t_i * w_i;
      u_i = t_r * w_i + t_i * w_r;
    }
  }
}

/* Computes the DFT in place.  The length of R and I must be a power of two. 
 * If I is NULL, it is interpreted as a zero matrix.
 * If I is 1, then the magnitude of the DFT is placed in R.
 */
void VectorDFT(Vector R, Vector I)
{
  if(I == (Vector)1) {
    I = VectorCreate(R->len);
    VectorSet(I, 0);
    VectorDFT(R, I);
    VectorHypot(R, I);
    VectorFree(I);
    return;
  }
  else if(!I) {
    I = VectorCreate(R->len);
    VectorSet(I, 0);
    VectorDFT(R, I);
    VectorFree(I);
    return;
  }
  assert(R->len == I->len);
  DFT1D(R->d, I->d, R->len);
}

/* Computes the DFT in place.  R and I must have rows and cols a power of two
 * (but they do not have to be square).
 * If I is NULL, it is interpreted as a zero matrix.
 * If I is 1, then the magnitude of the DFT is placed in R.
 */
void MatrixDFT(Matrix R, Matrix I)
{
  int i;
  if(I == (Matrix)1) {
    I = MatrixCreate(R->rows, R->cols);
    VectorSet((Vector)I, 0);
    MatrixDFT(R, I);
    VectorHypot((Vector)R, (Vector)I);
    MatrixFree(I);
    return;
  }
  else if(!I) {
    I = MatrixCreate(R->rows, R->cols);
    VectorSet((Vector)I, 0);
    MatrixDFT(R, I);
    MatrixFree(I);
    return;
  }
  assert(R->rows == I->rows);
  assert(R->cols == I->cols);
  /* Compute 2-D DFT by 1-D DFTs */
  /* Row DFT */
  for(i = 0; i < R->rows; i++) DFT1D(R->d2[i], I->d2[i], R->cols);
  MatrixTranspose(R);
  MatrixTranspose(I);
  /* Col DFT */
  for(i = 0; i < R->rows; i++) DFT1D(R->d2[i], I->d2[i], R->cols);
  MatrixTranspose(R);
  MatrixTranspose(I);
}
