/* Matrix implementation.
 * A Matrix is a structure of {d, len, rows, d2, cols.}
 */

#include <tpm/tpm.h>
#include "matrix.h"

/* Prototypes ****************************************************************/

Matrix MatrixCreate(int rows, int cols);
Matrix MatrixFromData(int rows, int cols, double *data);
void MatrixFree(Matrix m);
Matrix MatrixCopy(Matrix m);
void MatrixIdentity(Matrix m);
Matrix MatrixFromVector(Vector v, int rows);
void MatrixGetRow(Matrix m, int row, Vector v);
void MatrixSetRow(Matrix m, int row, Vector v);
Vector MatrixRow(Matrix m, int row);
Vector MatrixColumn(Matrix m, int col);
Vector MatrixDiagonal(Matrix m);
Matrix MatrixDeleteRow(Matrix m, int row);
void MatrixReshape(Matrix m, int rows, int cols);

void MatrixTranspose(Matrix m);
Matrix MatrixTransposeCopy(Matrix m);
void MatrixFlipV(Matrix m);
void MatrixFlipH(Matrix m);
Matrix MatrixFilter(Matrix m, Matrix f);
void MatrixColumnSum(Matrix m, Vector sum);
void MatrixSubtractRow(Matrix m, double *row);

void MatrixTransposeTimesSelf_Slow(Matrix r, Matrix a);
void MatrixSelfTimesTranspose_Slow(Matrix r, Matrix a);
void MatrixMultiply_Slow(Matrix dest, Matrix a, Matrix b);
void MatrixVectorMultiply_Slow(Vector dest, Matrix a, Vector v);

void MatrixPrint(FILE *fp, Matrix m);
Matrix MatrixScan(FILE *fp);
Matrix MatrixReadBinary(FILE *fp);

/* Functions *****************************************************************/

/* Returns a Matrix whose data is given. The array becomes property 
 * of the Matrix and will be freed when the Matrix is. The data must be
 * rows*cols elements long.
 */
Matrix MatrixFromData(int rows, int cols, double *data)
{
  Matrix m;
  int i;

  m = Allocate(1, struct MatrixRec);
  m->rows = rows;
  m->cols = cols;
  m->len = rows*cols;
  m->d = data;
  m->d2 = Allocate(rows, double*);
  for(i=0;i<rows;i++) m->d2[i] = data + i*cols;
  return m;
}

/* Returns a new, uninitialized Matrix whose dimensions are given. */
Matrix MatrixCreate(int rows, int cols)
{
  double *data = Allocate(rows*cols, double);
#if 0
  if(!data) {
    fprintf(stderr, "Cannot allocate %d by %d matrix\n", rows, cols);
    return NULL;
  }
#endif
  return MatrixFromData(rows, cols, data);
}

/* Frees all storage associated with m. */
void MatrixFree(Matrix m)
{
  free(m->d2);
  free(m->d);
  free(m);
}

/* Returns a new Matrix whose data is identical to m. */
Matrix MatrixCopy(Matrix m)
{
  Matrix result = MatrixCreate(m->rows, m->cols);
  MatrixMove(result, m);
  return result;
}

/* Modifies m to be the identity matrix. */
void MatrixIdentity(Matrix m)
{
  int i, rank = min(m->rows, m->cols);
  VectorSet((Vector)m, 0);
  for(i=0;i<rank;i++) m->d2[i][i] = 1;
}

/* Returns a new Matrix whose data is identical to v 
 * and whose shape is rows by (v->len/rows).
 */
Matrix MatrixFromVector(Vector v, int rows)
{
  Matrix result = MatrixCreate(rows, v->len/rows);
  MatrixMove(result, v);
  return result;
}

/* Returns a row of m. */
Vector MatrixRow(Matrix m, int row)
{
  Vector v = VectorCreate(m->cols);
  memcpy(v->d, m->d2[row], v->len*sizeof(double));
  return v;
}

void MatrixGetRow(Matrix m, int row, Vector v)
{
  memcpy(v->d, m->d2[row], v->len*sizeof(double));
}
void MatrixSetRow(Matrix m, int row, Vector v)
{
  memcpy(m->d2[row], v->d, v->len*sizeof(double));
}

/* Returns column of m. */
Vector MatrixColumn(Matrix m, int col)
{
  Vector v = VectorCreate(m->rows);
  int i;
  for(i=0;i<v->len;i++) v->d[i] = m->d2[i][col];
  return v;
}

/* Returns the diagonal of m. */
Vector MatrixDiagonal(Matrix m)
{
  Vector v = VectorCreate(min(m->rows, m->cols));
  int i;
  for(i=0;i<v->len;i++) v->d[i] = m->d2[i][i];
  return v;
}

/* Returns a new matrix identical to m except missing row ROW. */
Matrix MatrixDeleteRow(Matrix m, int row)
{
  Matrix result = MatrixCreate(m->rows-1, m->cols);
  int len;
  /* Copy top half */
  len = row*m->cols;
  memcpy(result->d, m->d, len*sizeof(double));
  /* Copy bottom half */
  memcpy(result->d + len, m->d + len + m->cols, 
	 (m->rows - row - 1)*m->cols*sizeof(double));
  return result;
}

void MatrixReshape(Matrix m, int rows, int cols)
{
  int i;
  m->rows = rows;
  m->cols = cols;
  Reallocate(m->d2, m->rows, double*);
  for(i=0;i<m->rows;i++) m->d2[i] = m->d + i*m->cols;
}

/*****************************************************************************/

/* Transposes a matrix in place. */
void MatrixTranspose(Matrix m)
{
  int i,j;
  if(m->rows == m->cols) {
    for(i=0;i<m->rows;i++) {
      for(j=0;j<i;j++) {
	double t = m->d2[i][j];
	m->d2[i][j] = m->d2[j][i];
	m->d2[j][i] = t;
      }
    }
  }
  else {
    Matrix t = MatrixCopy(m);
    MatrixReshape(m, t->cols, t->rows);
    for(i=0;i<m->rows;i++) {
      for(j=0;j<m->cols;j++) {
	m->d2[i][j] = t->d2[j][i];
      }
    }
    MatrixFree(t);
  }
}

Matrix MatrixTransposeCopy(Matrix m)
{
  int i,j;
  Matrix t = MatrixCreate(m->cols, m->rows);
  for(i=0;i<t->rows;i++) {
    for(j=0;j<t->cols;j++) {
      t->d2[i][j] = m->d2[j][i];
    }
  }
  return t;
}

void MatrixFlipV(Matrix m)
{
#define CopyRow(dest, src) memcpy(dest, src, m->cols*sizeof(double))
  int i;
  double *t = Allocate(m->cols, double);
  for(i=0;i<m->rows/2;i++) {
    CopyRow(t, m->d2[i]);
    CopyRow(m->d2[i], m->d2[m->rows-1-i]);
    CopyRow(m->d2[m->rows-1-i], t);
  }
  free(t);
#undef CopyRow
}

void MatrixFlipH(Matrix m)
{
  int i,j;
  for(i=0;i<m->rows;i++) {
    for(j=0;j<m->cols/2;j++) {
      double t = m->d2[i][j];
      m->d2[i][j] = m->d2[i][m->cols-1-j];
      m->d2[i][m->cols-1-j] = t;
    }
  }
}

Matrix MatrixFilter(Matrix m, Matrix f)
{
  return NULL;
}

void MatrixColumnSum(Matrix m, Vector sum)
{
  int i,j;
  assert(sum->len == m->cols);
  VectorSet(sum, 0);
  for(i=0;i<m->rows;i++) {
    for(j=0;j<m->cols;j++) {
      sum->d[j] += m->d2[i][j];
    }
  }
}

/* Subtract the values of row from each row of m. */
void MatrixSubtractRow(Matrix m, double *row)
{
  int i,j;
  for(i=0;i<m->rows;i++) {
    for(j=0;j<m->cols;j++) {
      m->d2[i][j] -= row[j];
    }
  }
}

/*****************************************************************************/
/* These are all "slow" because the BLAS-based routines in mlapack.c
 * are faster. 
 */

/* r = a'*a.  Hence r must be square with side a->cols.
 * r and a may be the same matrix.
 */
void MatrixTransposeTimesSelf_Slow(Matrix r, Matrix a)
{
  Matrix t = r;
  int i,j,k;
  assert(r->rows == r->cols);
  assert(r->cols == a->cols);
  if(a == r) r = MatrixCreate(a->cols, a->cols);
  for(i=0;i<a->cols;i++) {
    for(j=0;j<a->cols;j++) {
      r->d2[i][j] = 0;
      for(k=0;k<a->rows;k++) {
	r->d2[i][j] += a->d2[k][i] * a->d2[k][j];
      }
    }
  }
  if(a == t) {
    MatrixMove(t, r);
    MatrixFree(r);
  }
}

/* r = a*a'.  Hence r must be square with side a->rows.
 * r and a may be the same matrix.
 */
void MatrixSelfTimesTranspose_Slow(Matrix r, Matrix a)
{
  Matrix t = r;
  int i,j,k;
  assert(r->rows == r->cols);
  assert(r->rows == a->rows);
  if(a == r) r = MatrixCreate(a->rows, a->rows);
  for(i=0;i<a->rows;i++) {
    for(j=0;j<a->rows;j++) {
      r->d2[i][j] = 0;
      for(k=0;k<a->cols;k++) {
	r->d2[i][j] += a->d2[i][k] * a->d2[j][k];
      }
    }
  }
  if(a == t) {
    MatrixMove(t, r);
    MatrixFree(r);
  }
}

/* Modifies dest to be the matrix product ab. 
 * dest may be the same as a or b (or both).
 * Requires: dest->rows == a->rows, dest->cols == b->cols, a->cols == b->rows
 */
void MatrixMultiply_Slow(Matrix dest, Matrix a, Matrix b)
{
  int i,j,k;
  assert(dest->rows == a->rows);
  assert(dest->cols == b->cols);
  assert(a->cols == b->rows);
  if((dest == a) && (dest != b)) {
    /* dest = dest*b */
    Vector row = VectorCreate(dest->cols);
    for(i=0;i<dest->rows;i++) {
      for(j=0;j<dest->cols;j++) {
	double sum = 0;
	for(k=0;k<b->rows;k++) {
	  sum += a->d2[i][k] * b->d2[k][j];
	}
	row->d[j] = sum;
      }
      /* clobber row i of a only after we are done with it */
      memcpy(dest->d2[i], row->d, row->len*sizeof(double));
    }
    VectorFree(row);
  }
  else if((dest == b) && (dest != a)) {
    /* dest = a*dest */
    Vector col = VectorCreate(dest->rows);
    for(i=0;i<dest->cols;i++) {
      for(j=0;j<dest->rows;j++) {
	double sum = 0;
	for(k=0;k<a->cols;k++) {
	  sum += a->d2[j][k] * b->d2[k][i];
	}
	col->d[j] = sum;
      }
      /* clobber col i of b only after we are done with it */
      for(j=0;j<dest->rows;j++) {
	dest->d2[j][i] = col->d[j];
      }
    }
    VectorFree(col);
  }
  else if((dest == a) && (dest == b)) {
    /* dest = a^2 */
    Matrix t = MatrixCopy(a);
    MatrixMultiply_Slow(dest, a, t);
    MatrixFree(t);
  }
  else {
    for(i=0;i<dest->rows;i++) {
      for(j=0;j<dest->cols;j++) {
	double sum = 0;
	for(k=0;k<b->rows;k++) {
	  sum += a->d2[i][k] * b->d2[k][j];
	}
	dest->d2[i][j] = sum;
      }
    }
  }
}

/* Modifies dest to be the product Av.
 * dest may be the same as v.
 * Requires: dest->len == a->rows, a->cols == v->len
 */
void MatrixVectorMultiply_Slow(Vector dest, Matrix a, Vector v)
{
  Vector t = dest;
  int i,j;
  assert(dest->len >= a->rows);
  assert(a->cols == v->len);
  if(dest == v) {
    t = VectorCreate(dest->len);
  }
  for(i=0;i<a->rows;i++) {
    double sum = 0;
    for(j=0;j<v->len;j++) {
      sum += a->d2[i][j] * v->d[j];
    }
    t->d[i] = sum;
  }
  if(t != dest) {
    VectorMove(dest, t);
    VectorFree(t);
  }
}

/*****************************************************************************/

/* Write a character representation of m to fp. This format was chosen so
 * that row Matrices print just like Vectors.
 */
void MatrixPrint(FILE *fp, Matrix m)
{
  int i,j;
  for(i=0;i<m->rows;i++) {
    for(j=0;j<m->cols;j++) {
      fprintf(fp, "%lg ", m->d2[i][j]);
    }
    fprintf(fp, "\n");
  }
}

/* Read a matrix in ASCII form, i.e. one row per line. 
 * Infers the width from the length of the first line, and the height
 * from the number of lines.
 */
Matrix MatrixScan(FILE *fp)
{
  Vector v = VectorCreate(32);
  Matrix mat;
  int rows, cols;
  double d;
  int i = 0;

  /* Use the first line to determine the number of columns */
  for(;;) {
    int c = getc(fp);
    if(c == '\n') break;
    if(isspace(c)) continue;
    ungetc(c, fp);
    if(fscanf(fp, "%lg", &d) < 1) {
      ErrorSet("MatrixScan expected a real number");
      VectorFree(v);
      return NULL;
    }
    if(i == v->len) {
      VectorResize(v, v->len * 2);
    }
    v->d[i++] = d;
  }
  cols = i;
  
  /* Read the rest of the lines */
  for(rows=1;;rows++) {
    int j;
    for(j = 0; j < cols; j++) {
      if(fscanf(fp, "%lg", &d) < 1) {
	if(!feof(fp)) {
	  ErrorSet("MatrixScan expected a real number");
	  VectorFree(v);
	  return NULL;
	}
	else if(j > 0) {
	  ErrorSet("unexpected end of file in MatrixScan");
	  VectorFree(v);
	  return NULL;
	}
	break;
      }
      if(i == v->len) {
	VectorResize(v, v->len * 2);
      }
      v->d[i++] = d;
    }
    if(feof(fp)) break;
  }
  if(rows*cols != v->len) {
    VectorResize(v, rows*cols);
  }
  mat = MatrixFromData(rows, cols, v->d);
  free(v);
  return mat;
}

Matrix MatrixReadBinary(FILE *fp)
{
  Matrix m;
  int rows, cols;
  /* read rows and cols */
  if(fscanf(fp, "%d %d", &rows, &cols) < 2) {
    fprintf(stderr, "bad matrix header\n");
    return NULL;
  }
  /* eat a newline */
  while(fgetc(fp) != '\n');
  m = MatrixCreate(rows, cols);
  if(fread(m->d, sizeof(double), m->len, fp) < m->len) {
    fprintf(stderr, "eof while reading matrix data (%d by %d)\n", rows, cols);
    MatrixFree(m);
    return NULL;
  }
  AdjustBuffer(m->d, m->len, sizeof(double));
  return m;
}

/*****************************************************************************/
