#ifndef MATRIX_H_INCLUDED
#define MATRIX_H_INCLUDED

#include <tpm/vector.h>

/* Designed for binary compatibility with Vectors.
 * The data is addressable both as a
 * two-dimensional array (m->d2[row][col]) of dimensions (rows, cols)
 * or a row-major array (m->d[index]) of dimension len = rows * cols.
 * m->d == m->d2[0].
 */
typedef struct MatrixRec {
  double *d;
  int len;
  int cols;
  double **d2;
  int rows;
} *Matrix;

#include <tpm/volume.h>

/* m1 gets the contents of m2. m1 must be already allocated and have exactly
 * the same dimensions as m2.
 */
#define MatrixMove(m1, m2) memcpy((m1)->d, (m2)->d, (m2)->len*sizeof(double))

#ifdef __cplusplus
extern "C" {
#endif

/* matrix.c ******************************************************************/

Matrix MatrixCreate(int rows, int cols);
Matrix MatrixFromData(int rows, int cols, double *data);
void MatrixFree(Matrix m);
Matrix MatrixCopy(Matrix m);
void MatrixIdentity(Matrix m);
Matrix MatrixFromVector(Vector v, int rows);
void MatrixGetRow(Matrix m, int row, Vector v);
void MatrixSetRow(Matrix m, int row, Vector v);
Vector MatrixRow(Matrix m, int row);
Vector MatrixColumn(Matrix m, int col);
Vector MatrixDiagonal(Matrix m);
Matrix MatrixDeleteRow(Matrix m, int row);
void MatrixReshape(Matrix m, int rows, int cols);

void MatrixTranspose(Matrix m);
Matrix MatrixTransposeCopy(Matrix m);
void MatrixFlipV(Matrix m);
void MatrixFlipH(Matrix m);
Matrix MatrixFilter(Matrix m, Matrix f);
void MatrixColumnSum(Matrix m, Vector sum);
void MatrixSubtractRow(Matrix m, double *row);

void MatrixTransposeTimesSelf_Slow(Matrix r, Matrix a);
void MatrixSelfTimesTranspose_Slow(Matrix r, Matrix a);
void MatrixMultiply_Slow(Matrix dest, Matrix a, Matrix b);
void MatrixVectorMultiply_Slow(Vector dest, Matrix a, Vector v);

#ifdef EOF
void MatrixPrint(FILE *fp, Matrix m);
Matrix MatrixScan(FILE *fp);
Matrix MatrixReadBinary(FILE *fp);
#endif

/* mmath.c *******************************************************************/

double VectorMahalanobis(Vector v1, Vector v2, Matrix icov);
double VectorDivergence(Vector v1, Vector v2, Matrix icov1, Matrix icov2,
			Matrix cov1, Matrix cov2);

/* mlapack.c *****************************************************************/

void MatrixTransposeTimesSelf(Matrix r, Matrix a);
void MatrixSelfTimesTranspose(Matrix r, Matrix a);
void MatrixMultiply(Matrix dest, Matrix a, Matrix b);
void MatrixVectorMultiply(Vector dest, Matrix a, Vector v);
double MatrixDeterminant(Matrix A);
int MatrixSolve(Matrix A, double *B);
int MatrixLLS(Matrix A, double *B, double *x);
int MatrixInvert(Matrix A);
int MatrixSVD(Matrix A, Matrix U, Matrix V, Vector S);
int MatrixPseudoInvert(Matrix A);
int MatrixEigenvalues(Matrix A, double *vr, double *vi);
int MatrixEigenvectorsSym(Matrix A, double *v);
int MatrixEigenvectorsSymX(Matrix A, double *v, int count, Matrix V);
int MatrixCholesky(Matrix A);

/* mfft.c ********************************************************************/

void MatrixFourier(Matrix A, Matrix *B);
void MatrixTwoDimFourier(Matrix A, Matrix *B);

/* matrixEv.c ****************************************************************/

void MatrixEvDecomposeK(Matrix m, int k);
void MatrixEvDecompose(Matrix m);
void MatrixEvCompose(Matrix m, Matrix r);
void MatrixEvInvert(Matrix m);
void MatrixEvPow(Matrix m, double e);
double MatrixEvQuad(Matrix m, Vector v);
double VectorMahalanobisEv(Vector v1, Vector v2, Matrix icov);
double VectorDivergenceEv(Vector v1, Vector v2, Matrix icov1, Matrix icov2);

/* Fortran routines in LAPACK ************************************************/
#define F77_FCN(f) f##_
int F77_FCN(dgetrf)(const int*, const int*, double*, int*, int*, int*);
int F77_FCN(dgetrs)(const char*, const int*, const int*, double*, 
                    const int*, int*, double*, const int*, int*);
int F77_FCN(dgelsx)(const int *m, const int *n, const int *nrhs,
		    double *A, const int *lda, double *B, const int *ldb,
		    int *jpvt, const double *rcond, int *rank,
		    double *work, int *info);
int F77_FCN(dgetri)(const int*, double*, const int*, int*, double*, 
                    const int*, int*);
int F77_FCN(dgeev)(const char *jobvl, const char *jobvr, const int *n, 
                   double *A, const int *lda, double *wr, double *wi, 
                   double *vl, const int *ldvl, double *vr, const int *ldvr, 
                   double *work, int *lwork, int *info);
int F77_FCN(dsyev)(const char *jobz, const char *uplo, const int *n,
		   double *A, const int *lda, double *w,
                   double *work, int *lwork, int *info);
int F77_FCN(dsyevx)(const char *jobz, const char *range, const char *uplo, 
		    const int *n, double *A, const int *lda, 
		    const double *vl, const double *vu,
		    const int *il, const int *iu, const double *abstol,
		    int *m, double *w, double *z, const int *ldz,
		    double *work, const int *lwork, int *iwork, 
		    int *ifail, int *info);
int F77_FCN(dpotrf)(const char *uplo, const int *n, double *A,
                    const int *lda, int *info);
int F77_FCN(dgesvd)(const char *jobu, const char *jobvt, 
		    const int *m, const int *n, double *A,
		    const int *lda, double *s, double *U, const int *ldu,
		    double *VT, const int *ldvt, 
		    double *work, const int *lwork, int *info);
void F77_FCN(dgemm)(const char *transa, const char *transb,
		    const int *m, const int *n, const int *k,
		    const double *alpha, double *A,
		    const int *lda, double *B, const int *ldb,
		    const double *beta, double *C,
		    const int *ldc);
void F77_FCN(dgemv)(const char *trans, const int *m, const int *n,
		    const double *alpha, double *A, const int *lda,
		    double *x, const int *incx, const double *beta,
		    double *y, const int *incy);

/* dft.c *********************************************************************/
void MatrixDFT(Matrix R, Matrix I);
void VectorDFT(Vector R, Vector I);

/* dct.c *********************************************************************/
void VectorDCT(Vector v);
void MatrixDCT(Matrix m);

#ifdef __cplusplus
}
#endif

#endif
