/* Spectral representation of positive definite matrices.
 * Equal to the eigenvector matrix (eigenvectors are rows) times
 * the square root of the eigenvalue matrix.
 * Low-order eigenvectors can be omitted, so the spectral representation 
 * may have fewer rows than the original matrix.
 * Allows efficient exponentiation (e.g. inverse) and evaluation of 
 * quadratic forms (e.g. Mahalanobis distance).
 */
#include <tpm/tpm.h>
#include "matrix.h"

/* Prototypes ****************************************************************/

void MatrixEvDecompose(Matrix m);
void MatrixEvDecomposeK(Matrix m, int k);
void MatrixEvCompose(Matrix m, Matrix r);
void MatrixEvInvert(Matrix m);
void MatrixEvPow(Matrix m, double e);
double MatrixEvQuad(Matrix m, Vector v);
double VectorMahalanobisEv(Vector v1, Vector v2, Matrix icov);
double VectorDivergenceEv(Vector v1, Vector v2, Matrix icov1, Matrix icov2);

/* Functions *****************************************************************/

#ifndef NO_LAPACK

/* Convert a positive definite matrix, e.g. a covariance matrix, into
 * the spectral representation.
 */
void MatrixEvDecompose(Matrix m)
{
  int i,j;
  double *v = Allocate(m->rows, double);
  MatrixEigenvectorsSym(m, v);
  /* take the root of the eigenvalues */
  for(i=0; i < m->rows; i++) {
    v[i] = sqrt(v[i]);
  }
  /* scale the eigenvectors by the rooted eigenvalues */
  for(i=0; i < m->rows; i++) {
    for(j=0; j < m->cols; j++) {
      m->d2[i][j] *= v[i];
    }
  }
  free(v);
}

/* Same as MatrixEvDecompose but only stores the top k eigenvectors,
 * i.e. the representation only has k rows.
 */
void MatrixEvDecomposeK(Matrix m, int k)
{
  MatrixEvDecompose(m);
  MatrixFlipV(m);
  m->rows = k;
}

#endif

/* Convert from spectral representation to regular form, placing result in r.
 * R = M'*M.  Hence R must be square with side M->cols.
 * M and R may be the same matrix.
 */
void MatrixEvCompose(Matrix m, Matrix r)
{
  Matrix t = r;
  int i,j,k;
  if(m == r) r = MatrixCreate(m->cols, m->cols);
  for(i=0;i<m->cols;i++) {
    for(j=0;j<m->cols;j++) {
      r->d2[i][j] = 0;
      for(k=0;k<m->rows;k++) {
	r->d2[i][j] += m->d2[k][i] * m->d2[k][j];
      }
    }
  }
  if(m == t) {
    MatrixMove(t, r);
    MatrixFree(r);
  }
}

/* Invert a matrix in spectral form.  Just divide each row by the square
 * of its norm.
 */
void MatrixEvInvert(Matrix m)
{
  int i,j;
  for(i=0;i<m->rows;i++) {
    double sum = 0;
    for(j=0;j<m->cols;j++) {
      sum += m->d2[i][j] * m->d2[i][j];
    }
    for(j=0;j<m->cols;j++) {
      m->d2[i][j] /= sum;
    }
  }
}

/* Exponentiate a matrix in spectral form.  Just multiply each row by the
 * square of its norm raised to (e-1)/2.
 */
void MatrixEvPow(Matrix m, double e)
{
  int i,j;
  for(i=0;i<m->rows;i++) {
    double sum = 0;
    for(j=0;j<m->cols;j++) {
      sum += m->d2[i][j] * m->d2[i][j];
    }
    sum = pow(sum, (e-1)/2);
    for(j=0;j<m->cols;j++) {
      m->d2[i][j] *= sum;
    }
  }
}

/* Evaluate the quadratic form v'*M*v.
 * This reduces to (S*v)'*(S*v), where S is the spectral form of M,
 * which is just the squared norm of S*v.
 */
double MatrixEvQuad(Matrix m, Vector v)
{
  int i,j;
  double sum = 0;

  for(i=0;i<m->rows;i++) {
    double d = 0;
    for(j=0;j<m->cols;j++) {
      d += v->d[j] * m->d2[i][j];
    }
    sum += d*d;
  }
  return sum;
}

#define LOGTWOPI   1.8378770664093454

/* Returns the Mahalanobis distance between v1 and v2, where
 * icov is the spectral form of the feature covariance matrix.
 */
double VectorMahalanobisEv(Vector v1, Vector v2, Matrix icov)
{
  double sum = 0;
  int i,j;

  /* Loop eigenvectors */
  for(i=0;i<icov->rows;i++) {
    double d = 0;
    double mag = 0;
    /* take dot product of v1 and v2 with the current eigenvector (icov->d2[i])
     * and subtract the results.
     */
    for(j=0;j<icov->cols;j++) {
      d += (v1->d[j] - v2->d[j]) * icov->d2[i][j];
      mag += icov->d2[i][j]*icov->d2[i][j];
    }
    sum += d*d;
    /* mag is now the inverse eigenvalue, i.e. the inverse variance.
     * subtract the log to take care of the inversion.
     */
    sum += LOGTWOPI - log(mag);
  }
  return sum/2;
}

/* Just add the Mahalanobis distance each way.  This is an approximation. */
double VectorDivergenceEv(Vector v1, Vector v2, Matrix icov1, Matrix icov2)
{
  return VectorMahalanobisEv(v1, v2, icov1) + 
         VectorMahalanobisEv(v2, v1, icov2);
}
