/* Matrix implementation.
 * A Matrix is a structure of {d, len, rows, d2, cols.}
 */

#include <tpm/tpm.h>
#include "matrixI.h"

/* Prototypes ****************************************************************/

MatrixI MatrixCreate(int rows, int cols);
MatrixI MatrixFromData(int rows, int cols, int *data);
void MatrixIFree(MatrixI m);
MatrixI MatrixICopy(MatrixI m);
MatrixI MatrixIFromVector(VectorI v, int rows);
VectorI MatrixIRow(MatrixI m, int row);
VectorI MatrixIColumn(MatrixI m, int col);

void MatrixIPrint(FILE *fp, MatrixI m);
MatrixI MatrixIScan(FILE *fp);
void MatrixIPrintBinary(FILE *fp, MatrixI m);
MatrixI MatrixIReadBinary(FILE *fp);

/* Functions *****************************************************************/

/* Returns a Matrix whose data is given. The array becomes property 
 * of the Matrix and will be freed when the Matrix is. The data must be
 * rows*cols elements long.
 */
MatrixI MatrixIFromData(int rows, int cols, int *data)
{
  MatrixI m;
  int i;

  m = Allocate(1, struct MatrixIRec);
  m->rows = rows;
  m->cols = cols;
  m->len = rows*cols;
  m->d = data;
  m->d2 = Allocate(rows, int*);
  for(i=0;i<rows;i++) m->d2[i] = data + i*cols;
  return m;
}

/* Returns a new, uninitialized Matrix whose dimensions are given. */
MatrixI MatrixICreate(int rows, int cols)
{
  int *data = Allocate(rows*cols, int);
  if(!data) {
    fprintf(stderr, "Cannot allocate %d by %d matrix\n", rows, cols);
    return NULL;
  }
  return MatrixIFromData(rows, cols, data);
}

/* Frees all storage associated with m. */
void MatrixIFree(MatrixI m)
{
  free(m->d2);
  free(m->d);
  free(m);
}

/* Returns a new Matrix whose data is identical to m. */
MatrixI MatrixICopy(MatrixI m)
{
  MatrixI result = MatrixICreate(m->rows, m->cols);
  MatrixIMove(result, m);
  return result;
}

/* Returns a new Matrix whose data is identical to v 
 * and whose shape is rows by (v->len/rows).
 */
MatrixI MatrixIFromVector(VectorI v, int rows)
{
  MatrixI result = MatrixICreate(rows, v->len/rows);
  MatrixIMove(result, v);
  return result;
}

/* Returns a row of m. */
VectorI MatrixIRow(MatrixI m, int row)
{
  VectorI v = VectorICreate(m->cols);
  memcpy(v->d, m->d2[row], v->len*sizeof(int));
  return v;
}

/* Returns column of m. */
VectorI MatrixIColumn(MatrixI m, int col)
{
  VectorI v = VectorICreate(m->rows);
  int i;
  for(i=0;i<v->len;i++) v->d[i] = m->d2[i][col];
  return v;
}

/*****************************************************************************/

/* Write a character representation of m to fp. This format was chosen so
 * that row Matrices print just like Vectors.
 */
void MatrixIPrint(FILE *fp, MatrixI m)
{
  int i,j;
  for(i=0;i<m->rows;i++) {
    for(j=0;j<m->cols;j++) {
      fprintf(fp, "%d ", m->d2[i][j]);
    }
    fprintf(fp, "\n");
  }
}

/* Write a matrix in binary form, with header "rows (cols int)\n". */
void MatrixIPrintBinary(FILE *fp, MatrixI m)
{
  fprintf(fp, "%d %d\n", m->rows, m->cols);
  fwrite(m->d, sizeof(int), m->len, fp);
}

/* Read a matrix in ASCII form, i.e. one row per line. 
 * Infers the width from the length of the first line, and the height
 * from the number of lines.
 */
MatrixI MatrixIScan(FILE *fp)
{
  VectorI v = VectorICreate(32);
  MatrixI mat;
  int rows, cols;
  int d;
  int i = 0;

  /* Use the first line to determine the number of columns */
  for(;;) {
    int c = getc(fp);
    if(c == '\n') break;
    if(isspace(c)) continue;
    ungetc(c, fp);
    if(fscanf(fp, "%d", &d) < 1) {
      ErrorSet("MatrixIScan expected an integer");
      VectorIFree(v);
      return NULL;
    }
    if(i == v->len) {
      VectorIResize(v, v->len * 2);
    }
    v->d[i++] = d;
  }
  cols = i;
  
  /* Read the rest of the lines */
  for(rows=1;;rows++) {
    int j;
    for(j = 0; j < cols; j++) {
      if(fscanf(fp, "%d", &d) < 1) {
	if(!feof(fp)) {
	  ErrorSet("MatrixIScan expected an integer");
	  VectorIFree(v);
	  return NULL;
	}
	else if(j > 0) {
	  ErrorSet("unexpected end of file in MatrixIScan");
	  VectorIFree(v);
	  return NULL;
	}
	break;
      }
      if(i == v->len) {
	VectorIResize(v, v->len * 2);
      }
      v->d[i++] = d;
    }
    if(feof(fp)) break;
  }
  if(rows*cols != v->len) {
    VectorIResize(v, rows*cols);
  }
  mat = MatrixIFromData(rows, cols, v->d);
  free(v);
  return mat;
}

MatrixI MatrixIReadBinary(FILE *fp)
{
  MatrixI m;
  int rows, cols;
  /* read rows and cols */
  if(fscanf(fp, "%d %d", &rows, &cols) < 2) {
    fprintf(stderr, "bad matrix header\n");
    return NULL;
  }
  /* eat a newline */
  while(fgetc(fp) != '\n');
  m = MatrixICreate(rows, cols);
  if(fread(m->d, sizeof(int), m->len, fp) < m->len) {
    fprintf(stderr, "eof while reading matrix data (%d by %d)\n", rows, cols);
    MatrixIFree(m);
    return NULL;
  }
  AdjustBuffer(m->d, m->len, sizeof(int));
  return m;
}

/*****************************************************************************/
