/* Sparse matrix implementation.
 * Binary-compatible with a vector, but not a regular matrix.
 * A sparse matrix is a structure of (d, len, cols, col, rows, row).
 * d is the data in the matrix.  Sentinel value at the end.
 * len is the size of d, i.e. the number of items in the matrix.
 * cols can be used however you wish.
 * col is len elements long, and specifies the column number of each element
 * in d.
 * start, end are [2*(rows+1)] elements long, and specify the starting and 
 * ending location of each row in d.  (Extra sentinel row at the end.)
 */
#include <tpm/tpm.h>
#include "matrixSparseP.h"

/* this code doesn't actually care what the element type is. */
typedef void *EltT;

/* Prototypes ****************************************************************/

MatrixSparseP MatrixSparsePCreate(int rows, int cols);
void MatrixSparsePFree(MatrixSparseP m);
int MatrixSparsePIndex(MatrixSparseP m, int r, int c);
void MatrixSparsePDelete(MatrixSparseP m, int r, int c);
void MatrixSparsePInsert(MatrixSparseP m, int r, int c, EltT value);

/* Functions *****************************************************************/

/* Allocates a matrix capable of storing rows*cols elements, 
 * in any arrangement.
 */
MatrixSparseP MatrixSparsePCreate(int rows, int cols)
{
  int i;
  MatrixSparseP m = Allocate(1, struct MatrixSparsePRec);
  m->rows = rows;
  m->cols = cols;
  m->size = m->len = m->rows * m->cols;
  if(!m->size) m->size = 16;
  m->start = Allocate(rows+1, int);
  m->end = Allocate(rows+1, int);
  for(i=0;i<=m->rows;i++) {
    m->start[i] = i * cols;
    m->end[i] = i * cols;
  }
  m->end[m->rows]++; /* for the sentinel */
  m->col = Allocate(m->size+1, int);
  for(i=0;i<m->size;i++) m->col[i] = NAI;
  m->d = Allocate(m->size+1, EltT);
  m->d[m->len] = 0; /* value of missing items */
  return m;
}

void MatrixSparsePFree(MatrixSparseP m)
{
  free(m->d);
  free(m->col);
  free(m->start);
  free(m->end);
  free(m);
}

/* Returns an index into the d/col arrays for (r,c),
 * or m->len if the entry doesn't exist.
 */
int MatrixSparsePIndex(MatrixSparseP m, int r, int c)
{
  int i;
#if 1
  /* same as below, but produces better code. */
  int e = m->end[r];
  int t = m->col[e];
  m->col[e] = c;
  for(i = m->start[r]; ; i++) {
    if(m->col[i] == c) {
      m->col[e] = t;
      if(i < e) return i;
      return m->len;
    }
  }
  return m->len;
#else
  for(i = m->start[r]; i < m->end[r]; i++)
    if(m->col[i] == c) return i;
  return m->len;
#endif
}

void MatrixSparsePDelete(MatrixSparseP m, int r, int c)
{
  int i, e;
  i = MatrixSparsePIndex(m, r, c); 
  if(i == m->len) return;
  e = m->end[r];
  e = m->end[r] = e - 1;
  if(i < e) {
    m->col[i] = m->col[e];
    m->d[i] = m->d[e];
  }
}

static void Enlarge(MatrixSparseP m)
{
  fprintf(stderr, "Enlarge\n");
  m->len++;
  if(m->len > m->size) {
    m->size += m->size/2;
    Reallocate(m->col, m->size+1, int);
    Reallocate(m->d, m->size+1, EltT);
    if(!m->col || !m->d) 
      fprintf(stderr, "MatrixSparsePInsert: Cannot allocate memory\n");
  }
}

void MatrixSparsePInsert(MatrixSparseP m, int r, int c, EltT value)
{
  int i, e;
  e = m->end[r];
  /* free space on this row? */
  if(e < m->start[r+1]) {
    /* yes */
    m->end[r] = e+1;
    m->col[e] = c;
    m->d[e] = value;
  }
  else {
    /* no; search for free space */
    int tr;
    for(tr = r+1; tr < m->rows; tr++) {
      if(m->end[tr] < m->start[tr+1]) {
	/* found some */
	break;
      }
    }
    if(tr == m->rows) {
      /* no free space */
      Enlarge(m);
    }
    i = m->end[tr];
    m->end[tr] = i + 1;
    while(tr > r) {
      e = m->start[tr];
      m->start[tr] = e + 1;
      m->d[i] = m->d[e];
      m->col[i] = m->col[e];
      tr--;
      m->end[tr] = e + 1;
      i = e;
    }
    m->d[i] = value;
    m->col[i] = c;
  }
}
