/* Wrappers for the libbihar FFT routines. 
 * Must link with -lbihar -lblas -lf2c -lm
 */
/* Should add support for the FFT routines in SGI complib
 */

#include <math.h>
#include "vector.h"
#include "matrix.h"
#include <tpm/tpm.h>

typedef struct {
  double real, imag;
} Complex;

/* Fortran routines in FFTPACK ***********************************************/
int F77_FCN(cffti)(const int *n, Complex *wsave);
int F77_FCN(cfftf)(const int *n, Complex *c, Complex *wsave);

/* Fortran routines in BIHAR *************************************************/
int F77_FCN(dcffti)(const int *n, Complex *wsave);
int F77_FCN(dcfftf)(const int *n, Complex *c, Complex *wsave);

/* Prototypes ****************************************************************/

void MatrixFourier(Matrix A, Matrix *B);
void MatrixTwoDimFourier(Matrix A, Matrix *B);

/* Functions *****************************************************************/

/* Returns the Complex equivalent of A, transposed. */
static Complex *MakeComplex(Matrix A)
{
  int i,j,k;
  Complex *result = Allocate(A->len, Complex);
  for(k=j=0;j < A->cols;j++) {
    for(i=0;i < A->rows;i++) {
      result[k].real = A->d2[i][j];
      result[k].imag = 0;
      k++;
    }
  }
  return result;
}

/* Modifies A to be the real part of the Fourier transform of A.
 * If B non-NULL, modifies it to be the imaginary part.
 * Exception:
 *   If B is 1, makes A the magnitude of the Fourier transform of A.
 */
void MatrixFourier(Matrix A, Matrix *B)
{
  Complex *wsave;
  int i,j,k, rows = A->rows;
  Complex *CA = MakeComplex(A);

  /* perform FFT on complex matrix CA */
  wsave = Allocate(4 * A->rows + 15, Complex);
  F77_FCN(dcffti)(&rows, wsave);
  for(i=0;i < A->rows;i++)
    F77_FCN(dcfftf)(&rows, &CA[A->cols*i], wsave);
  free(wsave);

  if(B == (Matrix*)1) {
    /* set A to the magnitude of CA, transposed. */
    for(k=j=0;j < A->cols;j++) {
      for(i=0;i < A->rows;i++) {
	A->d2[i][j] = hypot(CA[k].real, CA[k].imag);
	k++;
      }
    }
  }
  else {
    /* set A to real part of CA, B to imag part, transposed. */
    if(B) *B = MatrixCreate(A->rows, A->cols);
    for(k=j=0;j < A->cols;j++) {
      for(i=0;i < A->rows;i++) {
	A->d2[i][j] = CA[k].real;
	if(B) (*B)->d2[i][j] = CA[k].imag;
	k++;
      }
    }
  }
  free(CA);
}

/* Same contract as MatrixFourier, but FFT2 is taken. */
void MatrixTwoDimFourier(Matrix A, Matrix *B)
{
  Complex *wsave;
  int i,j, rows = A->rows;
  Complex *CA = MakeComplex(A);
  Complex c;

  /* perform FFT on complex matrix CA */
  wsave = Allocate(4 * A->rows + 15, Complex);
  F77_FCN(dcffti)(&rows, wsave);
  for(i=0;i < A->rows;i++)
    F77_FCN(dcfftf)(&rows, &CA[A->cols*i], wsave);

  /* transpose CA */
#define swap(a,b) c=a;a=b;b=c
  for(i=0;i<A->rows;i++) {
    for(j=i+1;j<A->cols;j++) {
      swap(CA[i*A->cols+j], CA[j*A->cols+i]);
    }
  }

  /* FFT again */
  F77_FCN(dcffti)(&rows, wsave);
  for(i=0;i < A->rows;i++)
    F77_FCN(dcfftf)(&rows, &CA[A->cols*i], wsave);
  free(wsave);

  if(B == (Matrix*)1) {
    /* set A to the magnitude of CA */
    for(i=0;i < A->rows*A->cols;i++) {
      A->d[i] = hypot(CA[i].real, CA[i].imag);
    }
  }
  else {
    /* set A to real part of CA, B to imag part. */
    if(B) *B = MatrixCreate(A->rows, A->cols);
    for(i=0;i < A->len;i++) {
      A->d[i] = CA[i].real;
      if(B) (*B)->d[i] = CA[i].imag;
    }
  }
  free(CA);
}
