/* Routines that require the LAPACK library. 
 * Must link with -llapack -lblas -lf2c -lm
 */
/* The Fortran routines expect column-major matrices, but our Matrices are
 * row-major. These functions try to hide this difference as much as possible.
 */

#include <math.h>
#include "matrix.h"
#include <tpm/tpm.h>

/* Prototypes ****************************************************************/

void MatrixMultiply(Matrix dest, Matrix A, Matrix B);
void MatrixTransposeTimesSelf(Matrix dest, Matrix A);
void MatrixSelfTimesTranspose(Matrix dest, Matrix A);
void MatrixVectorMultiply(Vector dest, Matrix A, Vector v);
double MatrixDeterminant(Matrix A);
int MatrixSolve(Matrix A, double *B);
int MatrixLLS(Matrix A, double *B, double *x);
int MatrixInvert(Matrix A);
int MatrixSVD(Matrix A, Matrix U, Matrix V, Vector S);
int MatrixPseudoInvert(Matrix A);
int MatrixEigenvalues(Matrix A, double *vr, double *vi);
int MatrixEigenvectorsSym(Matrix A, double *v);
int MatrixEigenvectorsSymX(Matrix A, double *v, int count, Matrix V);
int MatrixCholesky(Matrix A);

/* Functions *****************************************************************/

/* Modifies dest to be the matrix product AB. 
 * dest may be the same as A or B (or both).
 * Requires: dest->rows == A->rows, dest->cols == B->cols, A->cols == B->rows
 * This routine uses BLAS.
 */
void MatrixMultiply(Matrix dest, Matrix A, Matrix B)
{
  char trans = 'n';
  double zero = 0, one = 1;
  Matrix T = dest;
  assert(dest->rows == A->rows);
  assert(dest->cols == B->cols);
  assert(A->cols == B->rows);
  /* LAPACK requires the output to be distinct from the inputs. */
  if((dest == A) || (dest == B)) {
    T = MatrixCreate(dest->rows, dest->cols);
  }
  F77_FCN(dgemm)(&trans, &trans, &dest->cols, &dest->rows, &B->rows,
		 &one, B->d, &B->cols, A->d, &A->cols, &zero, T->d,
		 &dest->cols);
  if(T != dest) {
    MatrixMove(dest, T);
    MatrixFree(T);
  }
}

/* dest = A'*A.  Hence dest must be square with side A->cols.
 * dest and A may be the same matrix.
 */
void MatrixTransposeTimesSelf(Matrix dest, Matrix A)
{
  char transa = 'n', transb = 't';
  double zero = 0, one = 1;
  Matrix T = dest;
  assert(dest->rows == dest->cols);
  assert(dest->cols == A->cols);
  if(dest == A) {
    T = MatrixCreate(dest->rows, dest->cols);
  }
  F77_FCN(dgemm)(&transa, &transb, &dest->cols, &dest->rows, &A->rows,
		 &one, A->d, &A->cols, A->d, &A->cols, &zero, T->d,
		 &dest->cols);
  if(T != dest) {
    MatrixMove(dest, T);
    MatrixFree(T);
  }
}

/* dest = A*A'.  Hence dest must be square with side A->rows.
 * dest and A may be the same matrix.
 * Same as above except transa and transb are swapped.
 */
void MatrixSelfTimesTranspose(Matrix dest, Matrix A)
{
  char transa = 't', transb = 'n';
  double zero = 0, one = 1;
  Matrix T = dest;
  assert(dest->rows == dest->cols);
  assert(dest->cols == A->rows);
  if(dest == A) {
    T = MatrixCreate(dest->rows, dest->cols);
  }
  F77_FCN(dgemm)(&transa, &transb, &dest->cols, &dest->rows, &A->cols,
		 &one, A->d, &A->cols, A->d, &A->cols, &zero, T->d,
		 &dest->cols);
  if(T != dest) {
    MatrixMove(dest, T);
    MatrixFree(T);
  }
}

/* Modifies dest to be the product Av.
 * dest may be the same as v.
 * Requires: dest->len == A->rows, A->cols == v->len
 */
void MatrixVectorMultiply(Vector dest, Matrix A, Vector v)
{
  char trans = 't';
  double zero = 0, one = 1;
  int i_one = 1;
  Vector t = dest;
  assert(dest->len >= A->rows);
  assert(A->cols == v->len);
  if(dest == v) {
    t = VectorCreate(dest->len);
  }
  F77_FCN(dgemv)(&trans, &A->cols, &A->rows, &one, A->d, &A->cols,
		 v->d, &i_one, &zero, t->d, &i_one);
  if(t != dest) {
    VectorMove(dest, t);
    VectorFree(t);
  }
}

/* Returns the determinant of A. */
double MatrixDeterminant(Matrix A)
{
  int   *ipvt;
  int    info;
  double det;
  int    c1, neg;
  Matrix tmp;

  /* special cases */
  if(A->rows != A->cols) return 0;
  if(A->rows == 0) return 0;
  else if(A->rows == 1) return A->d[0];
  else if(A->rows == 2) {
    return A->d2[0][0] * A->d2[1][1] - A->d2[1][0] * A->d2[0][1];
  }

  tmp = MatrixCopy(A);
  ipvt = Allocate(tmp->rows, int);
  /* Note rows and cols are reversed */
  F77_FCN(dgetrf)(&tmp->cols, &tmp->rows, tmp->d,
                  &tmp->rows, ipvt, &info);
  if(info > 0) {
    /* singular matrix */
    free(ipvt);
    MatrixFree(tmp);
    return 0.0;
  }

  /* Take the product of the diagonal elements */
  det = 1;
  neg = 0;
  for (c1 = 0; c1 < tmp->rows; c1++) {
    det *= tmp->d2[c1][c1];
    if (ipvt[c1] != (c1+1)) neg = !neg;
  }
  free(ipvt);
  MatrixFree(tmp);

  /* Since tmp is an LU decomposition of a rowwise permutation of A,
     multiply by appropriate sign */
  return neg?-det:det;
}

/* Modifies B to have the solution x to A * x = B. 
 * Returns 0 if an error occurred.
 */
int MatrixSolve(Matrix A, double *B)
{
  int *ipvt;
  int info;
  Matrix tmp;
  char job;
  int nrhs;

  tmp = MatrixCopy(A);
  ipvt = Allocate(tmp->rows, int);
  /* Note width and height are reversed */
  F77_FCN(dgetrf)(&tmp->cols, &tmp->rows, tmp->d,
		  &tmp->rows, ipvt, &info);
  if(info > 0) {
    fprintf(stderr, "MatrixSolve: dgetrf reports singular matrix\n");
    free(ipvt);
    return 0;
  }
  else {
    job = 'T';
    nrhs = 1;
    F77_FCN(dgetrs)(&job, &tmp->cols, &nrhs, tmp->d, &tmp->rows,
		    ipvt, B, &tmp->rows, &info);
  }
  free(ipvt);
  MatrixFree(tmp);
  return 1;
}

/* Same as MatrixSolve but allows rectangular rank-deficient A.
 * Returns the minimum-norm or least-squares solution.
 * Solution goes into x, with B unchanged.
 */
int MatrixLLS(Matrix A, double *B, double *x)
{
  Matrix tmp = MatrixTransposeCopy(A);
  int *ipvt = Allocate(A->cols, int);
  double rcond = 1e-5;
  int nrhs = 1;
  int ldb = max(A->rows, A->cols);
  int rank = min(A->rows, A->cols);
  double *work = Allocate(max(rank+3*A->cols, 2*rank+nrhs), double);
  int info;
  double *v = AllocCopy(ldb, double, B);

  /* Note rows and cols are reversed */
  F77_FCN(dgelsx)(&A->rows, &A->cols, &nrhs, tmp->d, &A->rows,
		  v, &ldb, ipvt, &rcond, &rank, work, &info);
  memcpy(x, v, A->cols*sizeof(double));
  free(v);
  free(ipvt);
  free(work);
  MatrixFree(tmp);
  return !info;
}

/* Modifies A to be the inverse of A: inv(A) * A = I. 
 * Requires that A be full rank; it need not be square.
 * Returns 0 if an error occurred.
 */
int MatrixInvert(Matrix A)
{
  int *ipvt;
  int info, lwork;
  double *work;

  if(A->rows > A->cols) {
    Matrix K = MatrixCreate(A->cols, A->cols);
    MatrixTransposeTimesSelf(K, A);
    MatrixInvert(K);
    MatrixTranspose(A);
    MatrixMultiply(A, K, A);
    MatrixFree(K);
    return 1;
  }
  else if(A->rows < A->cols) {
    Matrix K = MatrixCreate(A->rows, A->rows);
    MatrixSelfTimesTranspose(K, A);
    MatrixInvert(K);
    MatrixTranspose(A);
    MatrixMultiply(A, A, K);
    MatrixFree(K);
    return 1;
  }

  ipvt = Allocate(A->rows, int);
  /* Note width and height are reversed */
  F77_FCN(dgetrf)(&A->cols, &A->rows, A->d,
		  &A->rows, ipvt, &info);
  if(info > 0) {
    fprintf(stderr, "MatrixInvert: dgetrf reports singular matrix\n");
    free(ipvt);
    return 0;
  }
  else {
    lwork = A->rows * 3;
    work = Allocate(lwork, double);
    F77_FCN(dgetri)(&A->rows, A->d, &A->cols, ipvt,
		    work, &lwork, &info);
    free(work);
    if(info > 0) {
      fprintf(stderr, "MatrixInvert: dgetri reports singular matrix\n");
      free(ipvt);
      return 0;
    }
  }
  free(ipvt);
  return 1;
}

/* Fills U, V, and S with the SVD of A.
 * Requires: 
 *   U->rows == A->rows
 *   U->cols == min(A->rows, A->cols)
 *   V->rows == min(A->rows, A->cols)
 *   V->cols == A->cols
 *   S->len == min(A->rows, A->cols)
 */
int MatrixSVD(Matrix A, Matrix U, Matrix V, Vector S)
{
  char jobu = 'S';
  char jobvt = 'S';
  double *dup;
  int lwork;
  double *work;
  int info;
  int rank = min(A->rows, A->cols);

  assert(U->rows == A->rows);
  assert(U->cols == rank);
  assert(V->rows == A->cols);
  assert(V->cols == rank);
  assert(S->len == rank);

  /* make a copy since dgesvd clobbers A */
  dup = Allocate(A->len, double);
  memcpy(dup, A->d, A->len * sizeof(double));

  lwork = 5 * max(A->rows, A->cols);
  work = Allocate(lwork, double);

  /* swap U and V because we're implicitly computing SVD(A') */
  F77_FCN (dgesvd) (&jobu, &jobvt, &A->cols, &A->rows, dup, &A->cols,
		    S->d, V->d, &V->rows, U->d, &U->cols, work, &lwork,
		    &info);
  free(work);
  free(dup);
  if(info > 0) {
    fprintf(stderr, "MatrixSVD failed\n");
    return 0;
  }
  return 1;
}


/* Modifies A to be the pseudo-inverse of A: pinv(A) * A = I. 
 * Returns 0 if an error occurred.
 */
int MatrixPseudoInvert(Matrix A)
{
  double tol;
  int rank = min(A->rows, A->cols);
  Matrix U = MatrixCreate(A->rows, rank);
  Matrix V = MatrixCreate(A->cols, rank);
  Vector S = VectorCreate(rank);
  int i,j;

  if(!MatrixSVD(A, U, V, S)) return 0;
  tol = max(A->rows, A->cols) * S->d[0] * 1e-15;
  /* invert the components of S less than tol */
  for(i=0;i<rank;i++) {
    if(S->d[i] > tol) S->d[i] = 1.0 / S->d[i];
    else S->d[i] = 0.0;
    /* scale col i of U */
    for(j=0;j<U->rows;j++) {
      U->d2[j][i] *= S->d[i];
    }
  }
  VectorFree(S);
  MatrixTranspose(U);
  /* swap rows and cols of A to make MatrixMultiply happy */
  MatrixReshape(A, V->rows, U->cols);
  MatrixMultiply(A, V, U);
  MatrixFree(U);
  MatrixFree(V);
  return 1;
}

/* Fills vr and vi with the real and imaginary parts of the eigenvalues of A'.
 * If vr or vi is NULL, that part of the result will not be returned.
 * Returns 0 if an error occurred.
 * Requires A square but not necessarily symmetric.
 */
int MatrixEigenvalues(Matrix A, double *vr, double *vi)
{
  char jobvl = 'N', jobvr = 'N';
  int rank = A->rows;
  double *dup;
  int one = 1;
  int lwork;
  double *work;
  int info;
  double *wr, *wi;

  if(vr == NULL) wr = Allocate(rank, double);
  else wr = vr;
  if(vi == NULL) wi = Allocate(rank, double);
  else wi = vi;

  /* make a copy since dgeev clobbers A */
  dup = Allocate(A->len, double);
  memcpy(dup, A->d, A->len * sizeof(double));

  lwork = 3 * A->rows;
  work = Allocate(lwork, double);

  F77_FCN (dgeev) (&jobvl, &jobvr, &rank, dup, &A->rows,
		   wr, wi, 
		   NULL /* left eigenvectors */, &one, 
		   NULL /* right eigenvectors */, &one, work, &lwork, &info);
  free(dup);
  free(work);
  if(vr == NULL) free(wr);
  if(vi == NULL) free(wi);
  if(info > 0) {
    fprintf(stderr, "MatrixEigenvalues failed\n");
    return 0;
  }
  return 1;
}

/* Modifies real, symmetric A to be its eigenvectors (as rows).
 * If v != NULL, it is filled with the corresponding eigenvalues in 
 * ascending order.
 * The DSYEV driver routine uses DSYTRD (tridiagonal reduction) followed by
 * DORGTR (QR) and DSTEQR (eigenvectors from QR).
 */
int MatrixEigenvectorsSym(Matrix A, double *v)
{
  char jobz = 'V';
  char uplo = 'U';
  int rank = A->rows;
  int lwork;
  double *work;
  int info;
  double *w;
  
  if(v == NULL) w = Allocate(rank, double);
  else w = v;

  lwork = 3 * A->rows;
  work = Allocate(lwork, double);

  F77_FCN (dsyev) (&jobz, &uplo, &rank, A->d, &A->rows, w, 
		   work, &lwork, &info);
  free(work);
  if(v == NULL) free(w);
  if(info > 0) {
    fprintf(stderr, "MatrixEigenvectorsSym failed\n");
    return 0;
  }
  return 1;
}

/* Computes count eigenvectors corresponding to the largest eigenvalues.
 * The eigenvalues are returned in V as rows.
 * If v != NULL, it is filled with the corresponding eigenvalues in 
 * ascending order.
 * The real, symmetric matrix A is destroyed.
 */
int MatrixEigenvectorsSymX(Matrix A, double *v, int count, Matrix V)
{
  char jobz = 'V';
  char range = 'I';
  char uplo = 'U';
  int rank = A->rows;
  double *dup;
  int lwork;
  double *work;
  int *iwork;
  int *ifail;
  int info;
  double *w;
#ifdef hpux
  double abstol = 2*1.76126e-312;   /* 2*DLAMCH('S') */
#else
  double abstol = 2*2.122e-314;   /* 2*DLAMCH('S') */
#endif
  int dummy;
  
  if(v == NULL) w = Allocate(rank, double);
  else w = v;

#if 0
  /* make a copy since dsyevx clobbers A */
  dup = Allocate(A->len, double);
  memcpy(dup, A->d, A->len * sizeof(double));
#else
  dup = A->d;
#endif

  lwork = 8 * A->rows;
  work = Allocate(lwork, double);
  iwork = Allocate(5 * A->rows, int);
  ifail = Allocate(A->rows, int);

  count = rank - count + 1;
  F77_FCN (dsyevx) (&jobz, &range, &uplo, &rank, dup, &A->rows, 
		    NULL, NULL, &count, &rank, &abstol, &dummy,
		    w, V->d, &V->cols,
		    work, &lwork, iwork, 
		    ifail, &info);
  free(ifail);
  free(iwork);
  free(work);

#if 0
  free(dup);
#endif

  if(v == NULL) free(w);
  if(info > 0) {
    fprintf(stderr, "MatrixEigenvectorsSym failed\n");
    return 0;
  }
  return 1;
}

/* Modifies A to be lower triangular R such that R * R' = A. 
 * Returns 0 if an error occurred.
 */
int MatrixCholesky(Matrix A)
{
  char uplo = 'U';
  int rank = A->rows;
  int info;
  int i,j;

  F77_FCN (dpotrf) (&uplo, &rank, A->d, &A->rows, &info);
  if(info > 0) {
    fprintf(stderr, "MatrixCholesky: matrix is not positive definite\n");
    return 0;
  }
  /* clear out the upper triangular */
  for(i=0;i<A->rows;i++)
    for(j=i+1;j<A->cols;j++)
      A->d2[i][j] = 0.0;
  return 1;
}
