/* These functions are distinguished by requiring libm.a */

#include <math.h>
#include "matrix.h"
#include <errno.h>

/* Prototypes ****************************************************************/

double VectorMahalanobis(Vector v1, Vector v2, Matrix icov);
double VectorDivergence(Vector v1, Vector v2, Matrix icov1, Matrix icov2,
			Matrix cov1, Matrix cov2);

/* Functions *****************************************************************/

/* Returns the Mahalanobis distance between v1 and v2, where
 * icov is the inverse of the feature covariance matrix.
 */
double VectorMahalanobis(Vector v1, Vector v2, Matrix icov)
{
  Vector d;
  double total, subtotal;
  int i,j;

  /* Compute difference vector d = (v2 - v1) */
  d = VectorCopy(v2);
  VectorSubtract(d, v1);

  total = 0.0;
  /* Compute dT * icov * d */
  for(i=0;i<d->len;i++) {
    subtotal = 0.0;
    for(j=0;j<d->len;j++) {
      subtotal += icov->d2[i][j] * d->d[j];
    }
    total += subtotal * d->d[i];
  }
  VectorFree(d);

  return sqrt(total);
}

/* Returns the Gaussian divergence between v1 and v2, where
 * cov is the feature covariance matrix and icov is its inverse.
 */
double VectorDivergence(Vector v1, Vector v2, Matrix icov1, Matrix icov2,
			Matrix cov1, Matrix cov2)
{
  Vector d;
  Matrix k;
  double total, subtotal;
  int i,j;

  total = 0.0;
  /* Compute trace term */
  for(i=0;i<cov1->len;i++) {
    total += cov1->d[i] * icov2->d[i];
    total += icov1->d[i] * cov2->d[i];
  }
  total -= 2*v1->len;

  /* Compute difference vector d = (v2 - v1) */
  d = VectorCopy(v2);
  VectorSubtract(d, v1);

  /* Compute combined covariance k = icov1 + icov2 */
  k = MatrixCopy(icov1);
  VectorAdd((Vector)k, (Vector)icov2);

  /* Compute dT * k * d */
  for(i=0;i<d->len;i++) {
    subtotal = 0.0;
    for(j=0;j<d->len;j++) {
      subtotal += k->d2[i][j] * d->d[j];
    }
    total += subtotal * d->d[i];
  }
  total /= 2;
  MatrixFree(k);
  VectorFree(d);

  /* watch out for numerical error */
  if((total < 0) && (total > -1e-10)) total = 0;
  else if((total < 0) || (total != total)) {
    /* trap here instead of in sqrt(), which will spew messages */
    errno = EDOM; return 0;
  }
  return sqrt(total);
}
