/* Definition of the array package. Add it to Tcl with
 *   Arraytcl_Init(interp);
 * Vector commands:
 *   create len
 *   free ptr
 *   copy ptr
 *   move src dest
 *   size ptr
 *   set ptr list
 *   get ptr
 *   at ptr index ?value?
 *   fill ptr value
 *   scale ptr value
 *   addValue ptr value
 *   sum ptr
 *   rescale ptr min max
 *   truncate ptr min max
 *   trunc_rescale ptr tmin tmax rmin rmax
 *   add ptr v
 *   sub ptr v
 *   mul ptr v
 *   div ptr v
 *   dot ptr v
 *   compare ptr v
 *   log ptr
 *   read file
 *   write ptr file
 * Matrix commands:
 *   create rows cols
 *   free ptr
 *   copy ptr
 *   move src dest
 *   size ptr
 *   set ptr list
 *   get ptr
 *   at ptr row col ?value?
 *   image ptr photo
 *   read file
 *   write ptr file
 *   invert ptr
 *   determinant ptr
 */
#include <tk.h>
#include "tcl_macros.h"
#include <tpm/tpm.h>
#include <tpm/matrix.h>
#include <tpm/heap.h>

/* Prototypes ****************************************************************/
int Matrixtcl_Init(Tcl_Interp *interp);

/* private */
static int VectorCmd(ClientData clientData, Tcl_Interp *interp,
		     int argc, char *argv[]);
static int MatrixCmd(ClientData clientData, Tcl_Interp *interp,
		     int argc, char *argv[]);
static void VectorTruncRescale(Vector v, double trunc_min, double trunc_max, 
			       double new_min, double new_max);

/* Functions *****************************************************************/

static void ReturnVector(Tcl_Interp *interp, Vector vector)
{
  Tcl_DString ds;
  char str[20];
  int i;

  /* Collect the result in a DString */
  Tcl_DStringInit(&ds);
  for(i=0;i<vector->len;i++) {
    sprintf(str, "%g", vector->d[i]);
    Tcl_DStringAppendElement(&ds, str);
  }
  Tcl_DStringResult(interp, &ds);
  Tcl_DStringFree(&ds);
}

static int VectorCmd(ClientData clientData, Tcl_Interp *interp,
		     int argc, char *argv[])
{
  Vector vector;
  Err(argc < 2, "missing directive: must be one of: create, free, size, set, get, at, fill, rescale, truncate, trunc_rescale, read, write");
  if(!strcmp(argv[1], "create")) {
    int len;
    Err(argc != 3, "usage: vector create len");
    GetInt(argv[2], len);
    vector = VectorCreate(len);
    ReturnPtr(vector);
  }
  else if(!strcmp(argv[1], "free")) {
    Err(argc != 3, "usage: vector free ptr");
    GetPtr(argv[2], vector);
    VectorFree(vector);
  }
  else if(!strcmp(argv[1], "copy")) {
    Err(argc != 3, "usage: vector copy ptr");
    GetPtr(argv[2], vector);
    ReturnPtr(VectorCopy(vector));
  }
  else if(!strcmp(argv[1], "move")) {
    Vector result;
    Err(argc != 4, "usage: vector move dest src");
    GetPtr(argv[2], result);
    GetPtr(argv[3], vector);
    VectorMove(result, vector);
    ReturnPtr(result);
  }
  else if(!strcmp(argv[1], "size")) {
    Err(argc != 3, "usage: vector size ptr");
    GetPtr(argv[2], vector);
    sprintf(interp->result, "%d", vector->len);
  }
  else if(!strcmp(argv[1], "set")) {
    int i;
    Err(argc != 4, "usage: vector set ptr list");
    GetPtr(argv[2], vector);
    if(Tcl_SplitList(interp, argv[3], &argc, &argv) != TCL_OK) 
      return TCL_ERROR;
    for(i=0;i<argc;i++) {
      GetDouble(argv[i], vector->d[i]);
    }
  }
  else if(!strcmp(argv[1], "get")) {
    Err(argc != 3, "usage: vector get ptr");
    GetPtr(argv[2], vector);
    ReturnVector(interp, vector);
  }
  else if(!strcmp(argv[1], "at")) {
    int i;
    Err((argc < 4) || (argc > 5), "usage: vector at ptr index ?value?");
    GetPtr(argv[2], vector);
    GetInt(argv[3], i);
    Err((i < 0) || (i >= vector->len), "index out of bounds");
    if(argc == 5) {
      GetDouble(argv[4], vector->d[i]);
    }
    sprintf(interp->result, "%g", vector->d[i]);
  }
  else if(!strcmp(argv[1], "fill")) {
    double value;
    Err(argc != 4, "usage: vector fill ptr value");
    GetPtr(argv[2], vector);
    GetDouble(argv[3], value);
    VectorSet(vector, value);
  }
  else if(!strcmp(argv[1], "random")) {
    int i;
    Err(argc != 3, "usage: vector random ptr");
    GetPtr(argv[2], vector);
    for(i=0;i<vector->len;i++) vector->d[i] = Rand();
  }
  else if(!strcmp(argv[1], "scale")) {
    double value;
    Err(argc != 4, "usage: vector scale ptr value");
    GetPtr(argv[2], vector);
    GetDouble(argv[3], value);
    VectorScale(vector, value);
  }
  else if(!strcmp(argv[1], "addValue")) {
    double value;
    Err(argc != 4, "usage: vector addValue ptr value");
    GetPtr(argv[2], vector);
    GetDouble(argv[3], value);
    VectorAddValue(vector, value);
  }
  else if(!strcmp(argv[1], "sum")) {
    Err(argc != 3, "usage: vector sum ptr");
    GetPtr(argv[2], vector);
    ReturnDouble(VectorSum(vector));
  }
  else if(!strcmp(argv[1], "mag")) {
    Err(argc != 3, "usage: vector mag ptr");
    GetPtr(argv[2], vector);
    ReturnDouble(VectorMag(vector));
  }
  else if(!strcmp(argv[1], "entropy")) {
    Err(argc != 3, "usage: vector entropy ptr");
    GetPtr(argv[2], vector);
    ReturnDouble(VectorEntropy(vector));
  }
  else if(!strcmp(argv[1], "minmax")) {
    double min, max;
    Err(argc != 3, "usage: vector minmax ptr");
    GetPtr(argv[2], vector);
    VectorMinMax(vector, &min, &max);
    sprintf(interp->result, "%g %g", min, max);
  }
  else if(!strcmp(argv[1], "rescale")) {
    double new_min, new_max;
    Err(argc != 5, "usage: vector rescale ptr min max");
    GetPtr(argv[2], vector);
    GetDouble(argv[3], new_min);
    GetDouble(argv[4], new_max);
    VectorRescale(vector, new_min, new_max);
  }
  else if(!strcmp(argv[1], "truncate")) {
    double new_min, new_max;
    Err(argc != 5, "usage: vector trunc ptr min max");
    GetPtr(argv[2], vector);
    GetDouble(argv[3], new_min);
    GetDouble(argv[4], new_max);
    VectorTruncate(vector, new_min, new_max);
  }
  else if(!strcmp(argv[1], "trunc_rescale")) {
    double trunc_min, trunc_max, new_min, new_max;
    Err(argc != 7, "usage: vector trunc_rescale ptr tmin tmax rmin rmax");
    GetPtr(argv[2], vector);
    GetDouble(argv[3], trunc_min);
    GetDouble(argv[4], trunc_max);
    GetDouble(argv[5], new_min);
    GetDouble(argv[6], new_max);
    VectorTruncRescale(vector, trunc_min, trunc_max, new_min, new_max);
  }
  else if(!strcmp(argv[1], "add")) {
    Vector result;
    Err(argc != 4, "usage: vector add v1 v2");
    GetPtr(argv[2], result);
    GetPtr(argv[3], vector);
    VectorAdd(result, vector);
  }
  else if(!strcmp(argv[1], "sub")) {
    Vector result;
    Err(argc != 4, "usage: vector sub v1 v2");
    GetPtr(argv[2], result);
    GetPtr(argv[3], vector);
    VectorSubtract(result, vector);
  }
  else if(!strcmp(argv[1], "mul")) {
    Vector result;
    Err(argc != 4, "usage: vector mul v1 v2");
    GetPtr(argv[2], result);
    GetPtr(argv[3], vector);
    VectorMultiply(result, vector);
  }
  else if(!strcmp(argv[1], "div")) {
    Vector result;
    Err(argc != 4, "usage: vector div v1 v2");
    GetPtr(argv[2], result);
    GetPtr(argv[3], vector);
    VectorDivide(result, vector);
  }
  else if(!strcmp(argv[1], "dot")) {
    Vector result;
    Err(argc != 4, "usage: vector dot v1 v2");
    GetPtr(argv[2], result);
    GetPtr(argv[3], vector);
    ReturnDouble(VectorDot(result, vector));
  }
  else if(!strcmp(argv[1], "compare")) {
    Vector result;
    Err(argc != 4, "usage: vector compare v1 v2");
    GetPtr(argv[2], result);
    GetPtr(argv[3], vector);
    ReturnInt(VectorCompare(result, vector));
  }
  else if(!strcmp(argv[1], "distance")) {
    Vector other;
    Err(argc != 4, "usage: vector distance v1 v2");
    GetPtr(argv[2], other);
    GetPtr(argv[3], vector);
    ReturnDouble(VectorDistance(other, vector));
  }
  else if(!strcmp(argv[1], "rel_entropy")) {
    Vector other;
    Err(argc != 4, "usage: vector rel_entropy v1 v2");
    GetPtr(argv[2], other);
    GetPtr(argv[3], vector);
    ReturnDouble(VectorRelEntropy(other, vector));
  }
  else if(!strcmp(argv[1], "log")) {
    Err(argc != 3, "usage: vector log v");
    GetPtr(argv[2], vector);
    VectorApply(vector, log);
  }
  else if(!strcmp(argv[1], "read")) {
    Tcl_Channel chan;
    Tcl_DString ds;
    int len;
    /* read a vector from a file */
    Err(argc != 3, "usage: vector read file");
    GetChannel(argv[2], chan);
    /* read length */
    Tcl_DStringInit(&ds);
    if(Tcl_Gets(chan, &ds) == -1) {
      Tcl_PosixError(interp);
      return TCL_ERROR;
    }
    Err(sscanf(Tcl_DStringValue(&ds), "%d", &len) != 1, "bad vector header");
    Tcl_DStringFree(&ds);
    vector = VectorCreate(len);
    /* read data */
    if(Tcl_Read(chan, (char*)vector->d, vector->len*sizeof(double)) == -1) {
      VectorFree(vector);
      Tcl_PosixError(interp);
      return TCL_ERROR;
    }
    ReturnPtr(vector);
  }
  else if(!strcmp(argv[1], "write")) {
    Tcl_Channel chan;
    char str[16];
    /* write a vector to a file */
    Err(argc != 4, "usage: vector write ptr file");
    GetPtr(argv[2], vector);
    GetChannel(argv[3], chan);
    /* write length */
    sprintf(str, "%d\n", vector->len);
    if(Tcl_Write(chan, str, strlen(str)) == -1) {
      Tcl_PosixError(interp);
      return TCL_ERROR;
    }
    /* write data */
    if(Tcl_Write(chan, (char*)vector->d, vector->len*sizeof(double)) == -1) {
      Tcl_PosixError(interp);
      return TCL_ERROR;
    }
  }
  else if(!strcmp(argv[1], "maxk")) {
    Tcl_DString ds;
    char str[20];
    Vector vector;
    Vector v;
    int k;
    Err(argc != 4, "usage: vector maxk ptr k");
    GetPtr(argv[2], vector);
    GetInt(argv[3], k);

    v = VectorCopy(vector);

    /* Collect the result in a DString */
    Tcl_DStringInit(&ds);
    while(k--) {
      double best = -HUGE_VAL;
      int i, best_i;
      for(i=0;i<v->len;i++) {
	if(v->d[i] > best) {
	  best = v->d[i];
	  best_i = i;
	}
      }
      sprintf(str, "%d", best_i);
      Tcl_DStringAppendElement(&ds, str);
      v->d[best_i] = -HUGE_VAL;
    }
    Tcl_DStringResult(interp, &ds);
    Tcl_DStringFree(&ds);
    VectorFree(v);
  }
  else if(!strcmp(argv[1], "sample")) {
    Vector vector;
    Err(argc != 3, "usage: vector sample ptr");
    GetPtr(argv[2], vector);
    ReturnInt(VectorSample(vector));
  }
  else {
    Err(1, "unknown directive");
  }
  return TCL_OK;
}

static int MatrixCmd(ClientData clientData, Tcl_Interp *interp,
		     int argc, char *argv[])
{
  Matrix matrix;
  Err(argc < 2, "missing directive: must be one of: create, free, size, set, get, at, image, zoom, read, write");
  if(!strcmp(argv[1], "create")) {
    int rows, cols;
    Err(argc != 4, "usage: matrix create rows cols");
    GetInt(argv[2], rows);
    GetInt(argv[3], cols);
    matrix = MatrixCreate(rows, cols);
    ReturnPtr(matrix);
  }
  else if(!strcmp(argv[1], "free")) {
    Err(argc != 3, "usage: matrix free ptr");
    GetPtr(argv[2], matrix);
    MatrixFree(matrix);
  }
  else if(!strcmp(argv[1], "copy")) {
    Err(argc != 3, "usage: matrix copy ptr");
    GetPtr(argv[2], matrix);
    ReturnPtr(MatrixCopy(matrix));
  }
  else if(!strcmp(argv[1], "move")) {
    Matrix result;
    Err(argc != 4, "usage: matrix move dest src");
    GetPtr(argv[2], result);
    GetPtr(argv[3], matrix);
    MatrixMove(result, matrix);
    ReturnPtr(result);
  }
  else if(!strcmp(argv[1], "size")) {
    Err(argc != 3, "usage: matrix size ptr");
    GetPtr(argv[2], matrix);
    sprintf(interp->result, "%d %d", matrix->rows, matrix->cols);
  }
  else if(!strcmp(argv[1], "rows")) {
    Err(argc != 3, "usage: matrix rows ptr");
    GetPtr(argv[2], matrix);
    ReturnInt(matrix->rows);
  }
  else if(!strcmp(argv[1], "cols")) {
    Err(argc != 3, "usage: matrix cols ptr");
    GetPtr(argv[2], matrix);
    ReturnInt(matrix->cols);
  }
  else if(!strcmp(argv[1], "row")) {
    int row;
    Err(argc != 4, "usage: matrix row ptr row");
    GetPtr(argv[2], matrix);
    GetInt(argv[3], row);
    ReturnPtr(MatrixRow(matrix, row));
  }
  else if(!strcmp(argv[1], "delrow")) {
    int row;
    Err(argc != 4, "usage: matrix delrow ptr row");
    GetPtr(argv[2], matrix);
    GetInt(argv[3], row);
    ReturnPtr(MatrixDeleteRow(matrix, row));
  }
  else if(!strcmp(argv[1], "identity")) {
    int i;
    Err(argc != 3, "usage: matrix identity ptr");
    GetPtr(argv[2], matrix);
    VectorSet((Vector)matrix, 0);
    for(i=0;i<matrix->rows;i++) matrix->d2[i][i] = 1;
  }
  else if(!strcmp(argv[1], "set")) {
    int i;
    Err(argc != 4, "usage: matrix set ptr list");
    GetPtr(argv[2], matrix);
    if(Tcl_SplitList(interp, argv[3], &argc, &argv) != TCL_OK) 
      return TCL_ERROR;
    Err(argc > matrix->len, "argument list too long for [matrix set]");
    for(i=0;i<argc;i++) {
      GetDouble(argv[i], matrix->d[i]);
    }
  }
  else if(!strcmp(argv[1], "get")) {
    Tcl_DString ds;
    char str[20];
    int i,j;
    Err(argc != 3, "usage: matrix get ptr");
    GetPtr(argv[2], matrix);

    /* Collect the result in a DString */
    Tcl_DStringInit(&ds);
    for(i=0;i<matrix->rows;i++) {
      Tcl_DStringStartSublist(&ds);
      for(j=0;j<matrix->cols;j++) {
	sprintf(str, "%g", matrix->d2[i][j]);
	Tcl_DStringAppendElement(&ds, str);
      }
      Tcl_DStringEndSublist(&ds);
    }
    Tcl_DStringResult(interp, &ds);
    Tcl_DStringFree(&ds);
  }
  else if(!strcmp(argv[1], "at")) {
    int row, col;
    Err((argc < 5) || (argc > 6), "usage: matrix at ptr row col ?value?");
    GetPtr(argv[2], matrix);
    GetInt(argv[3], row);
    Err((row < 0) || (row >= matrix->rows), "row out of bounds");
    GetInt(argv[4], col);
    Err((col < 0) || (col >= matrix->cols), "col out of bounds");
    if(argc == 6) {
      GetDouble(argv[5], matrix->d2[row][col]);
    }
    sprintf(interp->result, "%g", matrix->d2[row][col]);
  }
  else if(!strcmp(argv[1], "image")) {
    Tk_PhotoHandle image;
    Tk_PhotoImageBlock block;
    int i;
    Err(argc != 4, "usage: matrix image ptr photo");
    GetPtr(argv[2], matrix);
    /* is this a valid image name? */
    if((image = Tk_FindPhoto(argv[3])) == NULL) {
      Tcl_AppendResult(interp, "image \"", argv[3],
		       "\" doesn't exist", (char *) NULL);
      return TCL_ERROR;
    }
    /* put the matrix data into a Tk_PhotoImageBlock */
    block.width = matrix->cols;
    block.height = matrix->rows;
    block.pitch = block.width;
    block.pixelSize = 1;
    block.offset[0] = 0;
    block.offset[1] = 0;
    block.offset[2] = 0;
    block.pixelPtr = Allocate(block.width * block.height, unsigned char);
    /* fill the block */
    for(i=0;i<matrix->len;i++) {
      double v = matrix->d[i];
      /* truncate to 0..255 */
      if(v > 255) v = 255;
      else if(v < 0) v = 0;
      block.pixelPtr[i] = (unsigned char)v;
    }
    Tk_PhotoSetSize(image, block.width, block.height);
    Tk_PhotoPutBlock(image, &block, 0, 0, block.width, block.height);
    free(block.pixelPtr);
  }
  else if(!strcmp(argv[1], "rgb")) {
    Matrix red, green, blue;
    Tk_PhotoHandle image;
    Tk_PhotoImageBlock block;
    int i;
    Err(argc != 6, "usage: matrix rgb red green blue photo");
    GetPtr(argv[2], red);
    GetPtr(argv[3], green);
    GetPtr(argv[4], blue);
    /* is this a valid image name? */
    if((image = Tk_FindPhoto(argv[5])) == NULL) {
      Tcl_AppendResult(interp, "image \"", argv[5],
		       "\" doesn't exist", (char *) NULL);
      return TCL_ERROR;
    }
    /* put the matrix data into a Tk_PhotoImageBlock */
    block.width = red->cols;
    block.height = red->rows;
    block.pitch = block.width * 3;
    block.pixelSize = 3;
    block.offset[0] = 0;
    block.offset[1] = 1;
    block.offset[2] = 2;
    block.pixelPtr = Allocate(block.pitch * block.height, unsigned char);
    /* fill the block */
    for(i=0;i<red->len;i++) {
      double v = red->d[i];
      /* truncate to 0..255 */
      if(v > 255) v = 255;
      else if(v < 0) v = 0;
      block.pixelPtr[i*3+0] = (unsigned char)v;
    }
    for(i=0;i<green->len;i++) {
      double v = green->d[i];
      /* truncate to 0..255 */
      if(v > 255) v = 255;
      else if(v < 0) v = 0;
      block.pixelPtr[i*3+1] = (unsigned char)v;
    }
    for(i=0;i<blue->len;i++) {
      double v = blue->d[i];
      /* truncate to 0..255 */
      if(v > 255) v = 255;
      else if(v < 0) v = 0;
      block.pixelPtr[i*3+2] = (unsigned char)v;
    }
    Tk_PhotoSetSize(image, block.width, block.height);
    Tk_PhotoPutBlock(image, &block, 0, 0, block.width, block.height);
    free(block.pixelPtr);
  }
  else if(!strcmp(argv[1], "zoom")) {
    double value;
    Err(argc != 4, "usage: matrix zoom ptr value");
    GetPtr(argv[2], matrix);
    GetDouble(argv[3], value);
/*
    MatrixZoom(matrix, value);
*/
  }
  else if(!strcmp(argv[1], "read")) {
    Tcl_Channel chan;
    Tcl_DString ds;
    int rows, cols;
    /* read a matrix from a file */
    Err(argc != 3, "usage: matrix read file");
    GetChannel(argv[2], chan);
    /* read length */
    Tcl_DStringInit(&ds);
    if(Tcl_Gets(chan, &ds) == -1) {
      Tcl_PosixError(interp);
      return TCL_ERROR;
    }
    Err(sscanf(Tcl_DStringValue(&ds), "%d %d", &rows, &cols) != 2,
	"bad matrix header");
    Tcl_DStringFree(&ds);
    matrix = MatrixCreate(rows, cols);
    /* read data */
    if(Tcl_Read(chan, (char*)matrix->d, matrix->len*sizeof(double)) == -1) {
      MatrixFree(matrix);
      Tcl_PosixError(interp);
      return TCL_ERROR;
    }
    AdjustBuffer(matrix->d, matrix->len, sizeof(double));
    ReturnPtr(matrix);
  }
  else if(!strcmp(argv[1], "write")) {
    Tcl_Channel chan;
    char str[16];
    /* write a matrix to a file */
    Err(argc != 4, "usage: matrix write ptr file");
    GetPtr(argv[2], matrix);
    GetChannel(argv[3], chan);
    /* write size */
    sprintf(str, "%d %d\n", matrix->rows, matrix->cols);
    if(Tcl_Write(chan, str, strlen(str)) == -1) {
      Tcl_PosixError(interp);
      return TCL_ERROR;
    }
    /* write data */
    AdjustBuffer(matrix->d, matrix->len, sizeof(double));
    if(Tcl_Write(chan, (char*)matrix->d, matrix->len*sizeof(double)) 
       != matrix->len*sizeof(double)) {
      AdjustBuffer(matrix->d, matrix->len, sizeof(double));
      Tcl_PosixError(interp);
      return TCL_ERROR;
    }
    AdjustBuffer(matrix->d, matrix->len, sizeof(double));
  }
#ifndef NO_LAPACK
  /* these require lapack */
  else if(!strcmp(argv[1], "invert")) {
    Err(argc != 3, "usage: matrix invert ptr");
    GetPtr(argv[2], matrix);
    MatrixInvert(matrix);
  }
  else if(!strcmp(argv[1], "determinant")) {
    Err(argc != 3, "usage: matrix determinant ptr");
    GetPtr(argv[2], matrix);
    ReturnDouble(MatrixDeterminant(matrix));
  }
#endif
  else {
    Err(1, "unknown directive");
  }
  return TCL_OK;
}

int Matrixtcl_Init(Tcl_Interp *interp)
{
  Tcl_CreateCommand(interp, "vector", VectorCmd, NULL, NULL);
  Tcl_CreateCommand(interp, "matrix", MatrixCmd, NULL, NULL);
  return Tcl_PkgProvide(interp, "matrixTcl", "1");
}

/* Truncate the values in a vector to [trunc_min, trunc_max] and then
 * rescale to range from new_min to new_max.
 */
static void VectorTruncRescale(Vector v, double trunc_min, double trunc_max, 
			       double new_min, double new_max)
{
  int i;
  double factor;

  factor = (new_max - new_min) / (trunc_max - trunc_min);
  for(i=0;i<v->len;i++) {
    if(v->d[i] <= trunc_min) v->d[i] = new_min;
    else if(v->d[i] >= trunc_max) v->d[i] = new_max;
    else 
      v->d[i] = (v->d[i] - trunc_min) * factor + new_min;
  }
}

