/* Vector implementation.
 * A Vector is a structure of {d, len}.
 */

#include "vector.h"
#include <tpm/tpm.h>

/* Prototypes ****************************************************************/

Vector VectorCreate(int length);
Vector VectorFromData(int length, double *data);
void VectorFree(Vector v);
Vector VectorCopy(Vector v);
void VectorResize(Vector v, int len);

void VectorSet(Vector v, double x);
void VectorAddValue(Vector v, double x);
void VectorScale(Vector v, double x);
void VectorApply(Vector v, MathFunc *func);

void VectorFlip(Vector v);
double VectorSum(Vector v);
int VectorMaximum(Vector v);
void VectorMaximumK(Vector v, int k, double *value_return);
void VectorMinMax(Vector v, double *min_return, double *max_return);
void VectorRescale(Vector v, double new_min, double new_max);
void VectorTruncate(Vector v, double new_min, double new_max);

void VectorAdd(Vector v1, Vector v2);
void VectorSubtract(Vector v1, Vector v2);
void VectorMultiply(Vector v1, Vector v2);
void VectorDivide(Vector v1, Vector v2);
double VectorDot(Vector v1, Vector v2);
int VectorCompare(Vector v1, Vector v2);

void VectorPack(Vector a, Vector b, Vector c);
void VectorUnpack(Vector c, Vector a, Vector b);

void VectorNormalizePMF(Vector p);
int VectorSample(Vector p);
double VectorChi(Vector v1, Vector v2);

void VectorPrint(FILE *fp, Vector v);
void VectorPrintNL(FILE *fp, Vector v);

/* Functions *****************************************************************/

/* Returns a Vector whose data is given. The array becomes property
 * of the Vector and will be freed when the Vector is. The data must be
 * length elements long.
 */
Vector VectorFromData(int length, double *data)
{
  Vector v;

  v = Allocate(1, struct VectorRec);
  v->len = length;
  v->d = data;
  return v;
}

/* Returns a new, uninitialized Vector whose length is given. */
Vector VectorCreate(int length)
{
  double *data = Allocate(length, double);
  if(!data) {
    fprintf(stderr, "Cannot allocate %d element vector\n", length);
    return NULL;
  }
  return VectorFromData(length, data);
}

/* Frees all storage associated with v. */
void VectorFree(Vector v)
{
  free(v->d);
  free(v);
}

/* Returns a new Vector whose data is identical to v. */
Vector VectorCopy(Vector v)
{
  Vector result = VectorCreate(v->len);
  VectorMove(result, v);
  return result;
}

/* If len < v->len, truncates v.
 * If len > v->len, adds garbage to the end of v. 
 */
void VectorResize(Vector v, int len)
{
  v->d = Reallocate(v->d, len, double);
  v->len = len;
}

/*****************************************************************************/

/* Set all elements of v to x. */
void VectorSet(Vector v, double x)
{
  int i;
  for(i=0;i<v->len;i++) v->d[i] = x;
}

/* Add x to all elements of v. */
void VectorAddValue(Vector v, double x)
{
  int i;
  for(i=0;i<v->len;i++) v->d[i] += x;
}

/* Multiply all elements of v by x. */
void VectorScale(Vector v, double x)
{
  int i;
  for(i=0;i<v->len;i++) v->d[i] *= x;
}

/* Apply func to all elements of v. */
void VectorApply(Vector v, MathFunc *func)
{
  int i;
  for(i=0;i<v->len;i++) v->d[i] = func(v->d[i]);
}

/*****************************************************************************/

/* Changes [1 2 3] to [3 2 1] */
void VectorFlip(Vector v)
{
  int i;
  for(i=0;i<v->len/2;i++) {
    double t;
    t = v->d[i];
    v->d[i] = v->d[v->len-1-i];
    v->d[v->len-1-i] = t;
  }
}

/* Returns the sum of all elements of v. */
double VectorSum(Vector v)
{
  int i;
  double sum = 0.0;
  for(i=0;i<v->len;i++) sum += v->d[i];
  return sum;
}

/* Returns the index of the maximum element of v. */
int VectorMaximum(Vector v)
{
  double best = v->d[0];
  int best_i = 0;
  int i;
  for(i=1;i<v->len;i++) {
    if(v->d[i] > best) {
      best = v->d[i];
      best_i = i;
    }
  }
#if 0
  if(!finite(best)) {
    fprintf(stderr, "Warning: VectorMaximum returns NaN\n");
  }
#endif
  return best_i;
}

/* Modifies value_return[k] to be the kth largest value in the vector. 
 * value_return[0] is the largest value.
 * If there are fewer than k unique values in the vector, the smallest
 * value will be repeated in value_return.
 * This routine is efficient for small k.
 */
void VectorMaximumK(Vector v, int k, double *value_return)
{
  int n;
  double best, upper;
  if(k < 1) return;
  n = VectorMaximum(v);
  /* the maximum value is now an upper bound */
  upper = value_return[0] = v->d[n];
  for(n=1;n<k;n++) {
    int i = 1;
    best = v->d[0];
    if(best >= upper) {
      /* initialize best to be smaller than the upper bound */
      for(;i<v->len;i++) {
	if(v->d[i] < upper) {
	  best = v->d[i];
	  i++;
	  break;
	}
      }
    }
    /* find elements greater than best but less than the upper bound */
    for(;i<v->len;i++) {
      if((v->d[i] > best) && (v->d[i] < upper)) {
	best = v->d[i];
      }
    }
    upper = value_return[n] = best;
  }
}

/* Sets *min_return and *max_return to the minimum and maximum values of v. */
void VectorMinMax(Vector v, double *min_return, double *max_return)
{
  int i;
  *min_return = *max_return = v->d[0];
  for(i=0;i<v->len;i++) {
    if(v->d[i] < *min_return) *min_return = v->d[i];
    else if(v->d[i] > *max_return) *max_return = v->d[i];
  }
}

/* Applies a linear scale to change the range of v to be [new_min, new_max]. */
void VectorRescale(Vector v, double new_min, double new_max)
{
  double old_min, old_max;
  VectorMinMax(v, &old_min, &old_max);
  VectorAddValue(v, -old_min);
  if(old_max != old_min)
    VectorScale(v, (new_max - new_min) / (old_max - old_min));
  VectorAddValue(v, new_min);
}

/* Set anything below new_min to new_min and the corresponding for new_max. */
void VectorTruncate(Vector v, double new_min, double new_max)
{
  int i;
  for(i=0;i<v->len;i++) {
    if(v->d[i] < new_min) v->d[i] = new_min;
    else if(v->d[i] > new_max) v->d[i] = new_max;
  }
}

/*****************************************************************************/

/* Add corresponding elements of v2 to v1. */
void VectorAdd(Vector v1, Vector v2)
{
  int i;
  for(i=0;i<v1->len;i++) v1->d[i] += v2->d[i];
}

/* Subtract corresponding elements of v2 from v1. */
void VectorSubtract(Vector v1, Vector v2)
{
  int i;
  for(i=0;i<v1->len;i++) v1->d[i] -= v2->d[i];
}

/* Multiply corresponding elements of v1 by v2. */
void VectorMultiply(Vector v1, Vector v2)
{
  int i;
  for(i=0;i<v1->len;i++) v1->d[i] *= v2->d[i];
}

/* Divide corresponding elements of v1 by v2. */
void VectorDivide(Vector v1, Vector v2)
{
  int i;
  for(i=0;i<v1->len;i++) v1->d[i] /= v2->d[i];
}

/* Compute vector dot product = VectorSum(VectorMultiply(v1,v2))
 * of v1 and v2. 
 */
double VectorDot(Vector v1, Vector v2)
{
  int i;
  double sum = 0.0;
  for(i=0;i<v1->len;i++) sum += v1->d[i] * v2->d[i];
  return sum;
}

/* Returns 0 iff v1 and v2 are equal vectors. */
int VectorCompare(Vector v1, Vector v2)
{
  int i;
  for(i=0;i<v1->len;i++) if(v1->d[i] != v2->d[i]) return 1;
  return 0;
}

/* Copies the components of a and b consecutively into c.
 * Requires c->len >= a->len + b->len 
 */
void VectorPack(Vector a, Vector b, Vector c)
{
  int i,j=0;
  for(i=0;i<a->len;i++) c->d[j++] = a->d[i];
  for(i=0;i<b->len;i++) c->d[j++] = b->d[i];
}

/* Copies the components of c into a and b.
 * Requires c->len >= a->len + b->len
 */
void VectorUnpack(Vector c, Vector a, Vector b)
{
  int i,j=0;
  for(i=0;i<a->len;i++) a->d[i] = c->d[j++];
  for(i=0;i<b->len;i++) b->d[i] = c->d[j++];
}

/* Scales p so that VectorSum(p) == 1. 
 * This routine is safer than VectorScale(p, 1.0/VectorSum(p)).
 */
void VectorNormalizePMF(Vector p)
{
  double sum = VectorSum(p);
  if(sum > 1e-300) VectorScale(p, 1.0/sum);
  else {
    int i, best = 0;
    for(i=1;i<p->len;i++) {
      if(p->d[i] > p->d[best]) best = i;
    }
    if(p->d[best] == 0) {
      VectorSet(p, 1.0/p->len);
    }
    else {
      VectorSet(p, 0);
      p->d[best] = 1;
    }
  }
}

/* Sample from p as if it was a scalar probability distribution. 
 * Requires: VectorSum(p) == 1.
 */
int VectorSample(Vector p)
{
  int i;
  double r = Rand();
  for(i=0;i<p->len;i++) {
    r -= p->d[i];
    if(r < 1e-5) break;
  }
  if(i == p->len) {
    fprintf(stderr, "VectorSample failed.  VectorSum(p) == %g\n", 
	    VectorSum(p));
    return 0;
  }
  return i;
}

/* Returns the Chi-squared distance between PMFs.
 * Can be used as an approximation to VectorRelEntropy(v1, v2).
 */
double VectorChi(Vector v1, Vector v2)
{
  double d = 0;
  int i;
  for(i=0;i<v1->len;i++) {
    double f = v1->d[i] - v2->d[i];
    if(f > 0) d += f*f/v2->d[i];
  }
  return d/2;
}

/*****************************************************************************/

/* Write a character representation of v to fp. */
void VectorPrint(FILE *fp, Vector v)
{
  int i;
  for(i=0;i<v->len;i++) fprintf(fp, "%lg ", v->d[i]);
}

void VectorPrintNL(FILE *fp, Vector v)
{
  VectorPrint(fp, v);
  fprintf(fp, "\n");
}
