/* VectorI implementation.
 * A VectorI is a structure of {d, len}.
 */

#include "vectorI.h"
#include <tpm/tpm.h>

/* Prototypes ****************************************************************/

VectorI VectorICreate(int length);
VectorI VectorIFromData(int length, int *data);
void VectorIFree(VectorI v);
VectorI VectorICopy(VectorI v);
void VectorIResize(VectorI v, int len);

void VectorISet(VectorI v, int x);
void VectorIAddValue(VectorI v, int x);
void VectorIScale(VectorI v, int x);

double VectorISum(VectorI v);
void VectorIMinMax(VectorI v, int *min_return, int *max_return);
void VectorITruncate(VectorI v, int new_min, int new_max);

void VectorIAdd(VectorI v1, VectorI v2);
void VectorISubtract(VectorI v1, VectorI v2);
void VectorIMultiply(VectorI v1, VectorI v2);
void VectorIDivide(VectorI v1, VectorI v2);

void VectorIPrint(FILE *fp, VectorI v);

/* Functions *****************************************************************/

/* Returns a Vector whose data is given. The array becomes property
 * of the Vector and will be freed when the Vector is. The data must be
 * length elements long.
 */
VectorI VectorIFromData(int length, int *data)
{
  VectorI v;

  v = Allocate(1, struct VectorIRec);
  v->len = length;
  v->d = data;
  return v;
}

/* Returns a new, uninitialized Vector whose length is given. */
VectorI VectorICreate(int length)
{
  int *data = Allocate(length, int);
  if(!data) {
    fprintf(stderr, "Cannot allocate %d element vector\n", length);
    return NULL;
  }
  return VectorIFromData(length, data);
}

/* Frees all storage associated with v. */
void VectorIFree(VectorI v)
{
  free(v->d);
  free(v);
}

/* Returns a new Vector whose data is identical to v. */
VectorI VectorICopy(VectorI v)
{
  VectorI result = VectorICreate(v->len);
  VectorIMove(result, v);
  return result;
}

/* If len < v->len, truncates v.
 * If len > v->len, adds garbage to the end of v. 
 */
void VectorIResize(VectorI v, int len)
{
  v->d = Reallocate(v->d, len, int);
  v->len = len;
}

/*****************************************************************************/

/* Set all elements of v to x. */
void VectorISet(VectorI v, int x)
{
  int i;
  for(i=0;i<v->len;i++) v->d[i] = x;
}

/* Add x to all elements of v. */
void VectorIAddValue(VectorI v, int x)
{
  int i;
  for(i=0;i<v->len;i++) v->d[i] += x;
}

/* Multiply all elements of v by x. */
void VectorIScale(VectorI v, int x)
{
  int i;
  for(i=0;i<v->len;i++) v->d[i] *= x;
}

/*****************************************************************************/

/* Returns the sum of all elements of v. */
double VectorISum(VectorI v)
{
  int i;
  double sum = 0.0;
  for(i=0;i<v->len;i++) sum += v->d[i];
  return sum;
}

/* Sets *min_return and *max_return to the minimum and maximum values of v. */
void VectorIMinMax(VectorI v, int *min_return, int *max_return)
{
  int i;
  *min_return = *max_return = v->d[0];
  for(i=0;i<v->len;i++) {
    if(v->d[i] < *min_return) *min_return = v->d[i];
    else if(v->d[i] > *max_return) *max_return = v->d[i];
  }
}

/* Set anything below new_min to new_min and the corresponding for new_max. */
void VectorITruncate(VectorI v, int new_min, int new_max)
{
  int i;
  for(i=0;i<v->len;i++) {
    if(v->d[i] < new_min) v->d[i] = new_min;
    else if(v->d[i] > new_max) v->d[i] = new_max;
  }
}

/*****************************************************************************/

/* Add corresponding elements of v2 to v1. */
void VectorIAdd(VectorI v1, VectorI v2)
{
  int i;
  for(i=0;i<v1->len;i++) v1->d[i] += v2->d[i];
}

/* Subtract corresponding elements of v2 from v1. */
void VectorISubtract(VectorI v1, VectorI v2)
{
  int i;
  for(i=0;i<v1->len;i++) v1->d[i] -= v2->d[i];
}

/* Multiply corresponding elements of v1 by v2. */
void VectorIMultiply(VectorI v1, VectorI v2)
{
  int i;
  for(i=0;i<v1->len;i++) v1->d[i] *= v2->d[i];
}

/* Divide corresponding elements of v1 by b2. */
void VectorIDivide(VectorI v1, VectorI v2)
{
  int i;
  for(i=0;i<v1->len;i++) v1->d[i] /= v2->d[i];
}

/*****************************************************************************/

/* Write a character representation of v to fp. */
void VectorIPrint(FILE *fp, VectorI v)
{
  int i;
  for(i=0;i<v->len;i++) fprintf(fp, "%d ", v->d[i]);
  fprintf(fp, "\n");
}

