/* These functions are distinguished by requiring libm.a */

#include "vector.h"
#include <math.h>

/* Prototypes ****************************************************************/

double VectorMag(Vector v);
double VectorDistance(Vector v1, Vector v2);
double VectorWeightedDistance(Vector v1, Vector v2, Vector w);
double VectorDistanceLp(Vector v1, Vector v2, double p);
void VectorHypot(Vector A, Vector B);
double VectorEntropy(Vector v);
double VectorRelEntropy(Vector p, Vector q);

/* Functions *****************************************************************/

/* Returns sqrt(VectorDot(v, v)) */
double VectorMag(Vector v)
{
  int i;
  double sum = 0.0;
  for(i=0;i<v->len;i++) sum += v->d[i] * v->d[i];
  return sqrt(sum);
}

/* Returns the Euclidean distance = VectorMag(VectorSubtract(v1, v2))
 * between v1 and v2. 
 */
double VectorDistance(Vector v1, Vector v2)
{
  int i;
  double sum = 0.0;
  for(i=0;i<v1->len;i++) {
    double d = v1->d[i] - v2->d[i];
    sum += d*d;
  }
  return sqrt(sum);
}

/* Returns the weighted Euclidean distance between v1 and v2
 * = sqrt(sum w[i]*(v1[i] - v2[i])^2)
 */
double VectorWeightedDistance(Vector v1, Vector v2, Vector w)
{
  int i;
  double sum = 0.0;
  for(i=0;i<v1->len;i++) {
    double d = v1->d[i] - v2->d[i];
    sum += d*d*w->d[i];
  }
  return sqrt(sum);
}

/* Returns the Lp norm of v1 - v2. */
#define MLM_HACK 1
double VectorDistanceLp(Vector v1, Vector v2, double p)
{
  int i;
  double sum = 0.0;
  if(p == 2) {
    for(i=0;i<v1->len;i++) {
      double d = v1->d[i] - v2->d[i];
      sum += d*d;
    }
#if MLM_HACK
    sum /= v1->len;
#endif
    return sqrt(sum);
  }
  if(p == 1) {
    for(i=0;i<v1->len;i++) {
      double d = fabs(v1->d[i] - v2->d[i]);
      sum += d;
    }
#if MLM_HACK
    sum /= v1->len;
#endif
    return sum;
  }
  if(p == 0.5) {
    for(i=0;i<v1->len;i++) {
      double d = fabs(v1->d[i] - v2->d[i]);
      sum += sqrt(d);
    }
#if MLM_HACK
    sum /= v1->len;
#endif
    return sum*sum;
  }
  /* general case */
  for(i=0;i<v1->len;i++) {
    double d = fabs(v1->d[i] - v2->d[i]);
    sum += pow(d, p);
  }
#if MLM_HACK
  sum /= v1->len;
#endif
  return pow(sum, 1.0/p);
}

/* Sets A[i] to hypot(A[i], B[i]), i.e. computes elementwise magnitude
 * of the complex vector A + Bi.
 */
void VectorHypot(Vector A, Vector B)
{
  int i;
  for(i=0;i<A->len;i++) {
    A->d[i] = hypot(A->d[i], B->d[i]);
  }
}

#include <stdio.h>

/* Returns the sum of v[i] * -log(v[i]). */
double VectorEntropy(Vector v)
{
  int i;
  double sum = 0.0;
  for(i=0;i<v->len;i++) {
    if(v->d[i] > 0) {
      sum += v->d[i] * log(v->d[i]);
    }
    else if(v->d[i] < 0) {
      fprintf(stderr, "Warning: VectorEntropy encountered negative element (v[%d] = %g)\n", i, v->d[i]);
    }
  }
  return -sum;
}

/* Returns the sum of p[i] * log(p[i]/q[i]) */
double VectorRelEntropy(Vector p, Vector q)
{
  int i;
  double sum = 0.0;
  for(i=0;i<p->len;i++) {
    if(p->d[i] > 0) {
      sum += p->d[i] * log(p->d[i] / q->d[i]);
    }
  }
  return sum;
}

/*****************************************************************************/

/* A LogDouble is simply a double represented by its logarithm.
 * Multiplication becomes addition.
 * exp and log are unchanged.
 * These routines avoid computing exp(x), since that would often
 * produce overflow or underflow.
 */

#include "logdouble.h"

double LogDoubleAdd(double a, double b)
{
  /* make a the larger number */
  if(a < b) {
    double t = a; a = b; b = t;
  }
  /* log(exp(a) + exp(b)) = a + log(1 + exp(b-a)) */
  if(!finite(b)) return a;
  return a + log(1 + exp(b-a));
}

#include <stdio.h>

void LogDoublePrint(FILE *fp, double x)
{
  double base10 = x / M_LN10;
  int int_part = (int)base10;
  double frac_part = base10 - int_part;
  fprintf(fp, "%f", exp(frac_part * M_LN10));
  if(int_part > 0) {
    fprintf(fp, "e+%d", int_part);
  }
  else if(int_part < 0) {
    fprintf(fp, "e%d", int_part);
  }
}
