/* Volume implementation.
 * A Volume is a structure of {d, len, xdim, d2, zydim, ydim, d3, zdim}
 */

#include "matrix.h"
#include <tpm/tpm.h>

/* Prototypes ****************************************************************/

Volume VolumeCreate(int zdim, int ydim, int xdim);
Volume VolumeFromData(int zdim, int ydim, int xdim, double *data);
void VolumeFree(Volume v);
Volume VolumeCopy(Volume v);
Volume VolumeFromMatrix(Matrix m, int zdim);
Matrix MatrixFromVolume(Volume v, int z);

Matrix VolumeGray(Volume v);
void VolumePrint(FILE *fp, Volume v);

/* Functions *****************************************************************/

/* Returns a Volume whose data is given. The array becomes property 
 * of the Volume and will be freed when the Volume is. The data must be
 * zdim*ydim*xdim elements long.
 */
Volume VolumeFromData(int zdim, int ydim, int xdim, double *data)
{
  Volume v;
  int i;

  v = Allocate(1, struct VolumeRec);
  v->zdim = zdim;
  v->ydim = ydim;
  v->xdim = xdim;
  v->zydim = zdim*ydim;
  v->len = v->zydim * xdim;

  v->d = data;
  v->d2 = Allocate(v->zydim, double*);
  for(i=0;i<v->zydim;i++) v->d2[i] = &v->d[i*xdim];
  v->d3 = Allocate(v->zdim, double**);
  for(i=0;i<v->zdim;i++) v->d3[i] = &v->d2[i*ydim];
  return v;
}

/* Returns a new, uninitialized Volume whose dimensions are given. */
Volume VolumeCreate(int zdim, int ydim, int xdim)
{
  double *data = Allocate(zdim*ydim*xdim, double);
  if(!data) {
    fprintf(stderr, "Cannot allocate %d by %d by %d volume\n", 
	    zdim, ydim, xdim);
    return NULL;
  }
  return VolumeFromData(zdim, ydim, xdim, data);
}

/* Frees all storage associated with v. */
void VolumeFree(Volume v)
{
  free(v->d3);
  free(v->d2);
  free(v->d);
  free(v);
}

/* Returns a new Volume whose data is identical to v. */
Volume VolumeCopy(Volume v)
{
  Volume result = VolumeCreate(v->zdim, v->ydim, v->xdim);
  VolumeMove(result, v);
  return result;
}

/* Returns a new Volume whose data is identical to m
 * and whose shape is zdim by (m->rows/zdim) by m->cols
 */
Volume VolumeFromMatrix(Matrix m, int zdim)
{
  Volume result = VolumeCreate(zdim, m->rows/zdim, m->cols);
  VolumeMove(result, m);
  return result;
}

/* Returns a z slice of a volume. */
Matrix MatrixFromVolume(Volume v, int z)
{
  Matrix m = MatrixCreate(v->ydim, v->xdim);
  memcpy(m->d, v->d3[z][0], m->len*sizeof(double));
  return m;
}

/*****************************************************************************/

/* Returns the grey-scale matrix corresponding to an RGB volume. */
Matrix VolumeGray(Volume v)
{
  Matrix m;
  int i;
  if(v->zdim == 1) return MatrixFromVolume(v, 0);
  m = MatrixCreate(v->ydim, v->xdim);
  for(i=0;i<m->len;i++) {
    m->d[i] = 0.30 * v->d3[0][0][i] + 
              0.59 * v->d3[1][0][i] + 
	      0.11 * v->d3[2][0][i];
  }
  return m;
}

/*****************************************************************************/

/* Write a character representation of v to fp. This format was chosen
 * so that Volumes of depth 1 print just like Matrices.
 */
void VolumePrint(FILE *fp, Volume v)
{
  int i,j,k;
  for(i=0;i<v->zdim;i++) {
    for(j=0;j<v->ydim;j++) {
      for(k=0;k<v->xdim;k++) {
	fprintf(fp, "%lg ", v->d3[i][j][k]);
      }
      fprintf(fp, "\n");
    }
    if(i < v->zdim - 1) fprintf(fp, "\n");
  }
}

/*****************************************************************************/
