#ifndef VOLUME_H_INCLUDED
#define VOLUME_H_INCLUDED

/* Designed for bytewise compatibility with Matrices and Vectors.
 * The data is addressable as a
 * three-dimensional array (m->d3[z][y][x]) of dimensions (zdim, ydim, xdim)
 * or a two-dimensional array (m->d2[zy][x]) of dimensions (zdim*ydim, xdim)
 * or an array (m->d[index]) of dimension len = zdim * ydim * xdim.
 *
 * How to interpret a Volume as a Matrix:
 * If the Volume contains a movie sequence, and the 
 * Z coordinate of the Volume is frame number, then the Matrix
 * interpretation is the vertical concatenation of the frames, i.e.
 * the rows are the scanlines of the movie in temporal order.
 * This is the interpretation that will be used when the Volume is
 * passed to a Matrix procedure.
 */
typedef struct VolumeRec {
  double *d;
  int len;
  int xdim;
  double **d2;
  int zydim;
  int ydim;
  double ***d3;
  int zdim;
} *Volume;

/* v1 gets the contents of v2. v1 must be already allocated and have exactly
 * the same dimensions as v2.
 */
#define VolumeMove(v1, v2) memcpy((v1)->d, (v2)->d, (v2)->len*sizeof(double))

#ifdef __cplusplus
extern "C" {
#endif

/* volume.c ******************************************************************/

Volume VolumeCreate(int zdim, int ydim, int xdim);
Volume VolumeFromData(int zdim, int ydim, int xdim, double *data);
void VolumeFree(Volume v);
Volume VolumeCopy(Volume v);
#ifdef MATRIX_H_INCLUDED
Volume VolumeFromMatrix(Matrix m, int zdim);
Matrix MatrixFromVolume(Volume v, int z);
#endif

Matrix VolumeGray(Volume v);
#ifdef EOF
void VolumePrint(FILE *fp, Volume v);
#endif

#ifdef __cplusplus
}
#endif

#endif
