/* A BitSet is a subset of the non-negative integers, stored compactly.
 */
#include <tpm/tpm.h>
#include <tpm/bitset.h>

/* Prototypes ****************************************************************/
BitSet BitSetCreate(int length);
void BitSetFree(BitSet s);
BitSet BitSetCopy(BitSet s);
void BitSetInsert(BitSet s, int i);
void BitSetDelete(BitSet s, int i);
int BitSetMember(BitSet s, int i);
int BitSetSize(BitSet s);
void BitSetUnion(BitSet result, BitSet s);
void BitSetIntersect(BitSet result, BitSet s);
void BitSetXor(BitSet result, BitSet s);

/* Functions *****************************************************************/

/* Create a new, empty BitSet with initial buffer size 'length'. */
BitSet BitSetCreate(int length)
{
  BitSet set = Allocate(1, struct BitSet);
  if(length <= 0) length = 1;
  set->length = (length+7)/8;
  set->mask = AllocateC(set->length, unsigned char);
  return set;
}

void BitSetFree(BitSet s)
{
  free(s->mask);
  free(s);
}

BitSet BitSetCopy(BitSet s)
{
  BitSet result = AllocCopy(1, struct BitSet, s);
  result->mask = AllocCopy(s->length, unsigned char, s->mask);
  return result;
}

/* Add i to the set, enlarging the buffer if necessary. */
void BitSetInsert(BitSet s, int i)
{
  if(i < 0) return;
  if(i/8 >= s->length) {
    s->length = 1 + i/8;
    Reallocate(s->mask, s->length, unsigned char);
  }
  s->mask[i/8] |= 0x80 >> (i%8);
}

/* Remove i from the set. */
void BitSetDelete(BitSet s, int i)
{
  if((i < 0) || (i/8 >= s->length)) return;
  s->mask[i/8] &= ~(0x80 >> (i%8));
}

/* Return 1 if i is in the set. */
int BitSetMember(BitSet s, int i)
{
  if((i < 0) || (i/8 >= s->length)) return 0;
  return s->mask[i/8] & (0x80 >> (i%8));
}

/* Return the number of elements in the set */
int BitSetSize(BitSet s)
{
  int i, result = 0;
  for(i=0;i<s->length;i++) {
    result += NumBits(s->mask[i]);
  }
  return result;
}

/* Modify result to contain the union with s. */
void BitSetUnion(BitSet result, BitSet s)
{
  int i;
  /* Enlarge result to match s */
  if(result->length < s->length) {
    result->length = s->length;
    Reallocate(result->mask, result->length, unsigned char);
  }
  for(i=0;i<s->length;i++) {
    result->mask[i] |= s->mask[i];
  }
}

/* Modify result to contain the intersection with s. */
void BitSetIntersect(BitSet result, BitSet s)
{
  int i;
  for(i=0;i<result->length;i++) {
    result->mask[i] &= s->mask[i];
  }
}

/* Modify result to contain the symmetric difference with s. */
void BitSetXor(BitSet result, BitSet s)
{
  int i;
  /* Enlarge result to match s */
  if(result->length < s->length) {
    result->length = s->length;
    Reallocate(result->mask, result->length, unsigned char);
  }
  for(i=0;i<s->length;i++) {
    result->mask[i] ^= s->mask[i];
  }
}
