#include <tpm/cache.h>

#define VERBOSE 0

/* Prototypes ****************************************************************/

Cache CacheCreate(CacheableFunc *f, Any data,
		  HashFunc *hash, CmpFunc *compare,
		  CopyFunc *key_copy, CopyFunc *value_copy, 
		  FreeFunc *key_free, FreeFunc *value_free,
		  unsigned item_limit);
void CacheFree(Cache c);
Any CacheGet(Any key, Cache c);
Any CachePeek(Any key, Cache c);

/* Functions *****************************************************************/

Cache CacheCreate(CacheableFunc *f, Any data,
		  HashFunc *hash, CmpFunc *compare,
		  CopyFunc *key_copy, CopyFunc *value_copy, 
		  FreeFunc *key_free, FreeFunc *value_free,
		  unsigned item_limit)
{
  Cache c = Allocate(1, struct Cache);
  c->f = f;
  c->data = data;
  c->key_copy = key_copy;
  c->value_copy = value_copy;
  c->dict = HashTableCreate(hash, compare, key_free, value_free);
  /* keys are held by the HashTable */
  c->queue = ListCreate(NULL);
  c->item_limit = item_limit;
  c->hits = c->misses = 0;
  return c;
}

void CacheFree(Cache c)
{
  if(VERBOSE) {
    fprintf(stderr, "hit ratio = %g (%d of %d)\n", 
	    (double)c->hits/(c->hits + c->misses), 
	    c->hits, c->hits + c->misses);
  }
  ListFree(c->queue);
  HashTableFree(c->dict);
  free(c);
}

static void Uncache(Cache c)
{
  Any key;
  HashEntry *e;
  assert(ListSize(c->queue) > 0);
#if 1
  ListRemoveRear(c->queue, &key);
#else
  /* Removing the front is better for sequential access patterns. */
  ListRemoveFront(c->queue, &key);
#endif
  if(VERBOSE > 1) {
    fprintf(stderr, "uncaching %s\n", (char*)key);
  }
  e = HashTableFindEntry(c->dict, key);
  assert(e);
  HashTableRemoveEntry(c->dict, e);
}

static void Shrink(Cache c)
{
  while(ListSize(c->queue) > c->item_limit) Uncache(c);
}

/* Arguments are reversed so that CacheGet is a CacheableFunc. */
Any CacheGet(Any key, Cache c)
{
  int found;
  HashEntry *e = HashTableAddEntry(c->dict, key, &found);
  if(!found) {
    Shrink(c);
    if(c->key_copy) e->key = c->key_copy(key);
    e->value = c->f(key, c->data);
    ListAddFront(c->queue, e->key);
    c->misses++;
    if(VERBOSE > 1) {
      fprintf(stderr, "cache miss\n");
    }
  }
  else {
    c->hits++;
    if(VERBOSE > 1) {
      fprintf(stderr, "cache hit\n");
    }
  }
  if(c->value_copy) return c->value_copy(e->value);
  return e->value;
}

/* Like CacheGet except returns NULL if key is not in the cache. */
Any CachePeek(Any key, Cache c)
{
  int found;
  HashEntry *e = HashTableAddEntry(c->dict, key, &found);
  if(!found) return NULL;
  else {
    c->hits++;
    if(VERBOSE > 1) {
      fprintf(stderr, "cache hit\n");
    }
  }
  if(c->value_copy) return c->value_copy(e->value);
  return e->value;
}
