/* Digraph implementation.
 * Digraphs are implemented as a list of nodes, which hold data and a 
 * list of outgoing edges. An edge holds data and a destination node.
 */

#include <tpm/digraph.h>

/* Prototypes ****************************************************************/

Digraph DigraphCreate(FreeFunc *nf, FreeFunc *ef);
void DigraphFree(Digraph dg);
DigraphNode DigraphAddNode(Digraph dg, void *data);
DigraphNode DigraphGetNode(Digraph dg, void *data, CmpFunc *cf);
void DigraphAddEdge(DigraphNode n1, DigraphNode n2, void *data);
void *DigraphGetEdge(DigraphNode n1, DigraphNode n2);
void DigraphReplaceEdge(Digraph dg, DigraphNode n1, DigraphNode n2, 
			void *data);

/* Private */
static void DigraphNodeFree(DigraphNode dgn);

/* Functions *****************************************************************/

static void DigraphNodeFree(DigraphNode dgn)
{
  ListFree(dgn->edges);
  free(dgn);
}

/* Create an empty Digraph, using nf as the free function for node data
 * and ef as the free function for edge data.
 * The free functions will be used when nodes or edges are removed
 * and when the digraph is freed.
 * If a free function is specified as NULL, then the module will not
 * attempt to free the data.
 */
Digraph DigraphCreate(FreeFunc *nf, FreeFunc *ef)
{
  Digraph dg = Allocate(1, struct DigraphStruct);
  dg->nodes = ListCreate((FreeFunc*)DigraphNodeFree);
  dg->nf = nf;
  dg->ef = ef;
  return dg;
}

/* Free the storage used by a Digraph, including the node and edge data,
 * if the corresponding free functions exist.
 */
void DigraphFree(Digraph dg)
{
  DigraphNode node;
  DigraphEdge edge;

  /* free the node and edge data */
  if(dg->nf || dg->ef) {
    ListNode ptr;
    ListIterate(ptr, dg->nodes) {
      node = (DigraphNode)ptr->data;
      if(dg->nf) dg->nf(node->data);
      {ListNode ep;ListIterate(ep, node->edges) {
	edge = (DigraphEdge)ep->data;
	if(dg->ef) dg->ef(edge->data);
      }}
    }
  }
  ListFree(dg->nodes);
  free(dg);
}

/* Create a new node with the specified data value.
 * The node will have no links from or to it.
 */
DigraphNode DigraphAddNode(Digraph dg, void *data)
{
  DigraphNode node = Allocate(1, struct DigraphNodeStruct);
  node->data = data;
  /* DigraphEdge can be freed with GenericFree */
  node->edges = ListCreate(GenericFree);
  ListAddRear(dg->nodes, node);
  return node;
}

/* Returns the node in the Digraph whose data matches "data", using
 * compare function cf. Same semantics as ListIndexAtValue.
 */
DigraphNode DigraphGetNode(Digraph dg, void *data, CmpFunc *cf)
{
  ListNode ptr;
  DigraphNode node;

  ListIterate(ptr, dg->nodes) {
    node = (DigraphNode)ptr->data;
    if(cf) {
      if(!cf(node->data, data)) return node;
    }
    else if(node->data == data) return node;
  }
  return NULL;
}

/* If dg has no edge from n1 to n2, then one is added, with the given data.
 * Otherwise, the data in that edge is replaced with the given data.
 * The old data is freed with ef, if there is one.
 */
void DigraphReplaceEdge(Digraph dg, DigraphNode n1, DigraphNode n2, void *data)
{
  ListNode ptr;
  DigraphEdge edge;

  /* does the edge already exist? */
  ListIterate(ptr, n1->edges) {
    edge = (DigraphEdge)ptr->data;
    if(edge->to == n2) {
      if(dg->ef) dg->ef(edge->data);
      edge->data = data;
      return;
    }
  }
  DigraphAddEdge(n1, n2, data);
}

/* Add an edge from n1 to n2, with the given data.
 * Will create a duplicate edge if one already exists.
 * Use DigraphReplaceEdge to avoid duplicates.
 */
void DigraphAddEdge(DigraphNode n1, DigraphNode n2, void *data)
{
  DigraphEdge edge = Allocate(1, struct DigraphEdgeStruct);
  edge->data = data;
  edge->to = n2;
  ListAddRear(n1->edges, edge);
}

/* Returns the data of the edge connecting n1 and n2.
 * If there is no such edge, returns NULL.
 */
void *DigraphGetEdge(DigraphNode n1, DigraphNode n2)
{
  ListNode ptr;
  DigraphEdge edge;

  ListIterate(ptr, n1->edges) {
    edge = (DigraphEdge)ptr->data;
    /* pointer equality */
    if(edge->to == n2) return edge->data;
  }
  /* no such edge */
  return NULL;
}
