/* Doubly-linked circular list implementation.
 * Each node has data and prev/next links.
 * A sentinel node is used for the head; its next points to the first node
 * and its prev points to the last node in the list.
 * The sentinel also stores useful data about the list.
 */
#include "dlist.h"

/* Prototypes ****************************************************************/

DList DListCreate(FreeFunc *ff);
void DListRemoveAll(DList list);
void DListFree(DList list);

void DListAddFront(DList list, Any value);
void DListAddRear(DList list, Any value);
void DListRemove(DList list, DList node, Any *value_return);
void DListRemoveMatching(DList list, Any value, CmpFunc *compare);

/* Functions *****************************************************************/

DList DListCreate(FreeFunc *ff)
{
  DList list = Allocate(1, struct DList);
  DListData data = Allocate(1, struct DListData);
  data->ff = ff;
  data->size = 0;
  list->data = data;
  list->prev = list->next = list;
  return list;
}

void DListRemoveAll(DList list)
{
  DList node;
  assert(list);
  for(node = list->next; node != list;) {
    DList temp = node;
    node = node->next;
    DListRemove(list, temp, NULL);
  }
}

void DListFree(DList list)
{
  DListRemoveAll(list);
  /* free the sentinel */
  free(list->data);
  free(list);
}

void DListAddFront(DList list, Any value)
{
  DListData data = (DListData)list->data;
  DList node = Allocate(1, struct DList);
  node->data = value;
  list->next->prev = node;
  node->next = list->next;
  node->prev = list;
  list->next = node;
  data->size++;
}

void DListAddRear(DList list, Any value)
{
  DListData data = (DListData)list->data;
  DList node = Allocate(1, struct DList);
  node->data = value;
  list->prev->next = node;
  node->prev = list->prev;
  node->next = list;
  list->prev = node;
  data->size++;
}

/* Removes node from list */
void DListRemove(DList list, DList node, Any *value_return)
{
  DListData data = (DListData)list->data;
  if(node == list) return;
  data->size--;
  if(value_return) *value_return = node->data;
  else if(data->ff) data->ff(node->data);
  node->prev->next = node->next;
  node->next->prev = node->prev;
  free(node);
}

/* Remove all nodes such that compare(data, value) == 0. 
 * If compare == NULL, use pointer equality.
 */
void DListRemoveMatching(DList list, Any value, CmpFunc *compare)
{
  DList node;
  for(node = DListFront(list); node != list;) {
    if((node->data == value) ||
       (compare && !compare(node->data, value))) {
      DList temp = node;
      node->prev->next = node->next;
      node->next->prev = node->prev;
      if(DListFreeFunc(list)) DListFreeFunc(list)(node->data);
      DListSize(list)--;
      node = node->next;
      free(temp);
    }
    else node = node->next;
  }
}
