/* Machine-independent dynamic loading.
 */
#include "tpm.h"
#include <sys/stat.h>

#define USE_HASH 0
#if USE_HASH
#include <tpm/hash.h>
#endif

/* Globals *******************************************************************/

/* these variables are never freed */
static char *searchPath = NULL;
#if USE_HASH
static HashTable handles = NULL;
#endif

/* Prototypes ****************************************************************/

char *DynAppendPath(char *path);
void *DynLoad(char *name);
void *DynSymbol(char *libname, char *symbol);
const char *DynError(void);

/* Functions *****************************************************************/

#ifdef hpux
#include <dl.h>

static void *load(char *path)
{
  /* DYNAMIC_PATH and the +s option allow searching SHLIB_PATH for the 
     library.  Since our paths are absolute, it isn't necessary. */
  /* If BIND_IMMEDIATE is used, we better use BIND_NONFATAL also. */
  return shl_load(path, BIND_FIRST | BIND_VERBOSE, 0L);
}

static int unload(void *handle)
{
  return shl_unload(handle);
}

/* Find a symbol in a library.
 * If handle is NULL, searches all loaded libraries and the static program.
 * Returns NULL on failure, otherwise a pointer to the symbol.
 */
static void *findsym(void *handle, char *symbol)
{
  int result;
  void *value;

  result = shl_findsym((shl_t*)&handle, symbol, TYPE_UNDEFINED, &value);
  if(result < 0) return NULL;
  return value;
}

const char *DynError(void)
{
  extern char *sys_errlist[];
  return sys_errlist[errno];
}

#else
#include <dlfcn.h>

#ifdef sun
#define RTLD_LAZY 1
#define RTLD_NOW 2
#endif
#ifndef RTLD_GLOBAL
#define RTLD_GLOBAL 0
#endif

static void *load(char *path)
{
  return dlopen(path, RTLD_LAZY);
}

static int unload(void *handle)
{
  int result = dlclose(handle);
  if(result != 0) return -1;
  return result;
}

static void *findsym(void *handle, char *symbol)
{
  if(handle == NULL) {
#if USE_HASH
    if(handles) {
      HashIter(b, e, handle, handles) {
	void *result = dlsym(handle, symbol);
	if(result) return result;
      }
    }
    return NULL;
#else
    handle = dlopen(NULL, RTLD_LAZY | RTLD_GLOBAL);
#endif
  }
  return dlsym(handle, symbol);
}

const char *DynError(void)
{
  return dlerror();
}

#endif

/* Returns the allocated full pathname of the first file matching name
 * given the searchPath. Returns NULL if no such file was found.
 */
static char *FindFile(char *name, char *searchPath)
{
  char *pathName;
  struct stat statInfo;

  /* relative path? */
  if (*name != '/') {
    int found = 0;
    char *path, *s;

    if(!searchPath) s = strdup("");
    else s = strdup(searchPath);
    /* loop the searchPath entries */
    for(path = strtok(s, ":"); path; path = strtok(NULL, ":")) {
      int length = strlen(path) + strlen(name) + 2;
      pathName = (char*)malloc(length);
      sprintf(pathName, "%s/%s", path, name);
      if(stat(pathName, &statInfo) >= 0) {
	found = 1;
	break;
      }
      free(pathName);
    }
    free(s);
    if(!found) return NULL;
  } 
  /* absolute path; don't use searchPath */
  else {
    if(stat(name, &statInfo) < 0) return NULL;
    pathName = strdup(name);
  }
  return (pathName);
}

char *DynAppendPath(char *path)
{
  /* has the searchPath been initialized yet? */
  if(!searchPath) {
    /* try to read it out of the LD_LIBRARY_PATH variable */
    char *root = getenv("LD_LIBRARY_PATH");
    if(root) searchPath = strdup(root);
    else searchPath = strdup("/lib:/usr/lib:/usr/local/lib");
  }

  if(!path) return searchPath;
  searchPath = (char*)realloc(searchPath, strlen(searchPath)+strlen(path)+2);
  sprintf(searchPath, "%s:%s", searchPath, path);
  return searchPath;
}

/* Loads the named library into the current process using the current
 * search path. Doesn't load libraries more than once. 
 * Returns a system-dependent handle to the library, or
 * NULL if there is no such library.
 * If name is NULL, initializes the handle table.
 */
void *DynLoad(char *name)
{
  char *pathName;
  void *handle;
#if USE_HASH
  HashEntry *e;
#endif

  /* make sure the searchPath is initialized */
  DynAppendPath(NULL);

#if USE_HASH
  /* make sure the handle table is initialized */
  if(!handles) {
    handles = HashTableCreate(HASH_WORD, NULL);
#ifndef hpux
    /* define a handle for the static program */
    e = HashTableAddEntry(handles, "", &found);
    e->value = dlopen(NULL, RTLD_NOW);
#endif    
  }
#endif
  if(!name) return NULL;

  pathName = FindFile(name, searchPath);
  if(!pathName) return NULL;
  
#if USE_HASH
  e = HashTableAddEntry(handles, pathName, &found);
  if(!found) e->value = load(pathName);
  free(pathName);
  return e->value;
#else
  handle = load(pathName);
  free(pathName);
  return handle;
#endif
}

void *DynSymbol(char *libname, char *symbol)
{
  void *handle = DynLoad(libname);
  return findsym(handle, symbol);
}
