/* Hash table implementation. Based on TCL hash tables.
 * Tables are indexed by arbitrary objects, given hash and comparison fcns.
 * Hash tables automatically grow to keep the value/bucket ratio below
 * the REBUILD_MULTIPLIER. However, they will not shrink.
 */

#include <tpm/hash.h>

#define INITIAL_BUCKETS 4
#define REBUILD_MULTIPLIER 3
/* 2^OCCUPANCY_SHIFT must not exceed INITIAL_BUCKETS */
#define OCCUPANCY_SHIFT 2

/* Prototypes ****************************************************************/

HashTable HashTableCreate(HashFunc *hash, CmpFunc *compare, 
			    FreeFunc *key_free, FreeFunc *value_free);
void HashTableFlush(HashTable ht);
void HashTableFree(HashTable ht);
HashTable HashTableCopy(HashTable ht, 
			  CopyFunc *key_copy, CopyFunc *value_copy);
int HashTableCompare(HashTable a, HashTable b, CmpFunc *compare);

HashEntry *HashTableAddEntry(HashTable ht, Any key, int *found_return);
HashEntry *HashTableFindEntry(HashTable ht, Any key);
void HashTableRemoveEntry(HashTable ht, HashEntry *entry);
void HashTableRemoveNoFree(HashTable ht, HashEntry *entry);
int HashTableRemoveValueAll(HashTable ht, Any value, CmpFunc *cf);
HashEntry *HashTableNextEntry(HashTable ht, HashEntry *entry);
int HashTableIndexEntry(HashTable ht, HashEntry *entry);
HashEntry *HashTableEntryAtIndex(HashTable ht, int index);

/* Private */
static void RebuildTable(HashTable ht);

/* Functions *****************************************************************/

/* used by HashTableCreate and HashTableFlush */
static HashTable initialize(HashTable ht)
{
  int i;
  ht->num_buckets = INITIAL_BUCKETS;
  ht->num_entries = 0;
  ht->occupancy = AllocateC(ht->num_buckets >> OCCUPANCY_SHIFT, int);
  ht->buckets = Allocate(ht->num_buckets, HashEntry *);
  if(!ht->buckets) {
    fprintf(stderr, "Out of memory; cannot create hash table\n");
    return NULL;
  }
  for(i=0;i<ht->num_buckets;i++) ht->buckets[i] = NULL;
  return ht;
}

/* Create an empty hash table.
 * If value_free != NULL, it will be used to free values in the table.
 * If key_free != NULL, it will be used to free keys in the table.
 */
HashTable HashTableCreate(HashFunc *hash, CmpFunc *compare, 
			    FreeFunc *key_free, FreeFunc *value_free)
{
  HashTable ht = Allocate(1, struct HashTable);
  ht->hash = hash;
  ht->compare = compare;
  ht->key_free = key_free;
  ht->value_free = value_free;
  return initialize(ht);
}

/* used by HashTableFlush and HashTableFree */
static void HashTableRemoveAll(HashTable ht)
{
  int b;
  HashEntry **ep, *t;

  if(!ht) return;
  for(b=0;b<ht->num_buckets;b++) {
    for(ep=&ht->buckets[b];*ep;) {
      /* have to keep the ht consistent, since ht->free may inspect us */
      t = *ep;
      *ep = t->next;
      ht->num_entries--;
      if(ht->key_free) ht->key_free(t->key);
      if(ht->value_free) ht->value_free(t->value);
      free(t);
    }
  }
}

/* Destroys the contents of the HashTable, but not the HashTable itself. */
void HashTableFlush(HashTable ht)
{
  if(!ht) return;
  HashTableRemoveAll(ht);
  free(ht->buckets);
  free(ht->occupancy);
  initialize(ht);
}

/* Destroys a HashTable, using the ff to free the stored values. */
void HashTableFree(HashTable ht)
{
  if(!ht) return;
  HashTableRemoveAll(ht);
  free(ht->buckets);
  free(ht->occupancy);
  free(ht);
}

/* Returns a new HashTable with the same contents as ht.
 * Uses the provided functions to copy the keys and values.
 * If a function is NULL, just copies the pointer.
 */
HashTable HashTableCopy(HashTable ht, 
			  CopyFunc *key_copy, CopyFunc *value_copy)
{
  HashTable new = HashTableCreate(ht->hash, ht->compare, 
				  ht->key_free, ht->value_free);
  int b;
  HashEntry *e;
  HashIterate(b, e, ht) {
    HashEntry *entry = HashTableAddEntry(new, e->key, NULL);
    if(key_copy) entry->key = key_copy(e->key);
    entry->value = value_copy ? value_copy(e->value) : e->value;
  }
  return new;
}

/* Returns 0 if two HashTables are equal, using the given comparison
 * function between values.
 * If compare == NULL, uses pointer equality.
 */
int HashTableCompare(HashTable a, HashTable b, CmpFunc *compare)
{
  Any t;
  if(a->num_entries != b->num_entries) return 1;
  {HashIter(bucket, ea, t, a) {
    /* look up this entry in the other table */
    HashEntry *eb = HashTableFindEntry(b, ea->key);
    if(!eb) return 1;
    if(compare && compare(eb->value, ea->value)) return 1;
    if(!compare && (eb->value != ea->value)) return 1;
  }}
  return 0;
}

/* Used by FindEntry and LookupEntry. 
 * Returns the bucket index and HashEntry corresponding to key.
 */
static HashEntry *LookupEntry(HashTable ht, Any key, int *bucket)
{
  int b;
  HashEntry *e;

  if(ht->num_entries == 1) {
    /* special case: table only has one entry (occurs frequently in boa) 
     * don't hash the input unless we have to.
     */
    for(b=0;b<ht->num_buckets-1;b++)
      if(ht->buckets[b]) break;
    for(e = ht->buckets[b]; e; e=e->next)
      if(!ht->compare(key, e->key)) break;
    if(e) {
      *bucket = b;
      return e;
    }
    *bucket = ht->hash(key) & (ht->num_buckets - 1);
    return NULL;
  }

  b = ht->hash(key) & (ht->num_buckets - 1);
  for(e = ht->buckets[b]; e; e=e->next)
    if(!ht->compare(key, e->key)) break;
  *bucket = b;
  return e;
}

/* If ht has no entries matching the key, returns a new HashEntry 
 * corresponding to that key (which is now owned by the hash table) 
 * into which data may be placed, and sets *found_return to 0.
 * Otherwise, returns the matching entry and sets *found_return to 1.
 * found_return may be NULL.
 */
HashEntry *HashTableAddEntry(HashTable ht, Any key, int *found_return)
{
  int b;
  HashEntry *e;
  if(!ht) return NULL;
  e = LookupEntry(ht, key, &b);
  if(!e) {
    if(found_return) *found_return = 0;
    e = Allocate(1, HashEntry);
    if(!e) {
      fprintf(stderr, "Out of memory; cannot add hash entry\n");
      return NULL;
    }
    e->key = key;
    e->value = NULL;
    e->bucket = b;
    e->next = ht->buckets[b];
    ht->buckets[b] = e;

    ht->num_entries++;
    if(ht->num_entries >= REBUILD_MULTIPLIER * ht->num_buckets) 
      RebuildTable(ht);
    else
      ht->occupancy[b >> OCCUPANCY_SHIFT]++;
  }
  else if(found_return) *found_return = 1;
  return e;
}

/* Returns the HashEntry matching the key, or NULL if there is none. */
HashEntry *HashTableFindEntry(HashTable ht, Any key)
{
  int b;
  if(!ht) return NULL;
  return LookupEntry(ht, key, &b);
}

/* Removes the entry from the hash table, freeing the entry data.
 */
void HashTableRemoveEntry(HashTable ht, HashEntry *entry)
{
  HashEntry **e;

  if(!ht) return;
  for(e = &ht->buckets[entry->bucket]; *e != entry; e = &(*e)->next);
  *e = entry->next;
  ht->num_entries--;
  ht->occupancy[entry->bucket >> OCCUPANCY_SHIFT]--;
  if(ht->key_free) ht->key_free(entry->key);
  if(ht->value_free) ht->value_free(entry->value);
  free(entry);
}

/* Removes the entry from the hash table, without freeing the entry data. */
void HashTableRemoveNoFree(HashTable ht, HashEntry *entry)
{
  FreeFunc *free = ht->value_free;
  ht->value_free = NULL;
  HashTableRemoveEntry(ht, entry);
  ht->value_free = free;
}

/* Remove all entries such that cf(data, value) == 0.
 * If cf == NULL, uses pointer equality (data == value).
 * Returns the number of nodes removed.
 */
int HashTableRemoveValueAll(HashTable ht, Any value, CmpFunc *cf)
{
  int b, count = 0;
  HashEntry **ep, *t;
  for(b = 0; b < ht->num_buckets; b++) {
    for(ep = &ht->buckets[b]; *ep;) {
      if((cf && !cf((*ep)->value, value)) ||
	 (!cf && ((*ep)->value == value))) {
	t = *ep;
	*ep = t->next;
	ht->num_entries--;
	ht->occupancy[b >> OCCUPANCY_SHIFT]--;
	if(ht->key_free) ht->key_free(t->key);
	if(ht->value_free) ht->value_free(t->value);
	free(t);
	count++;
      }
      else ep = &(*ep)->next;
    }
  }
  return count;
}

/* used to expand the table. */
static void RebuildTable(HashTable ht)
{
  HashEntry **old_buckets;
  HashEntry *e, *t;
  int i,b, old_size;

  old_size = ht->num_buckets;
  old_buckets = ht->buckets;

  ht->num_buckets *= 4;
  free(ht->occupancy);
  ht->occupancy = AllocateC(ht->num_buckets >> OCCUPANCY_SHIFT, int);
  ht->buckets = Allocate(ht->num_buckets, HashEntry *);
  if(!ht->buckets) {
    fprintf(stderr, "Out of memory; cannot rebuild hash table\n");
    ht->buckets = old_buckets;
    ht->num_buckets = old_size;
    return;
  }
  for(i=0;i<ht->num_buckets;i++) ht->buckets[i] = NULL;
  
  /* rehash all entries */
  for(i=0;i<old_size;i++) {
    for(e=old_buckets[i];e;e=t) {
      b = ht->hash(e->key) & (ht->num_buckets - 1);
      e->bucket = b;
      t = e->next;
      e->next = ht->buckets[b];
      ht->buckets[b] = e;
      ht->occupancy[b >> OCCUPANCY_SHIFT]++;
    }
  }
  free(old_buckets);
}

/* Returns the HashEntry immediately after entry in the HashTable,
 * or NULL if no such entry. 
 * If entry == NULL, returns the first entry in the table.
 */
HashEntry *HashTableNextEntry(HashTable ht, HashEntry *entry)
{
  int b;
  HashEntry *e;

  if(!entry) b = 0;
  else { 
    if(entry->next) return entry->next;
    b = entry->bucket + 1;
  }
  for(e = NULL; b < ht->num_buckets; b++) {
    if(e = ht->buckets[b]) break;
  }
  return e;
}

int HashTableIndexEntry(HashTable ht, HashEntry *entry)
{
  int b, eb, index=0;
  HashEntry *e;

  eb = entry->bucket >> OCCUPANCY_SHIFT;
  for(b=0;b<eb;b++) {
    index += ht->occupancy[b];
  }
  b <<= OCCUPANCY_SHIFT;
  for(;b<entry->bucket;b++) {
    for(e = ht->buckets[b]; e; e = e->next) index++;
  }
  for(e = ht->buckets[b]; e != entry; e = e->next) index++;
  return index;
}

HashEntry *HashTableEntryAtIndex(HashTable ht, int index)
{
  int b, eb;
  HashEntry *e;

  eb = ht->num_buckets >> OCCUPANCY_SHIFT;
  for(b=0;b<eb;b++) {
    if(ht->occupancy[b] > index) break;
    index -= ht->occupancy[b];
  }
  b <<= OCCUPANCY_SHIFT;
  for(;b<ht->num_buckets;b++) {
    for(e = ht->buckets[b]; e; e = e->next) if(!index--) return e;
  }
  return NULL;
}
