/* Heap of arbitrary data, using an array.
 * Taken from Cormen, Leiserson, and Rivest.
 * Insert, Delete, and ExtractMin take O(lg n) time.
 */
#include "heap.h"

/* these all take indices in (0, length] */
#define HeapParent(i) ((i)>>1)
#define HeapLeft(i) ((i)<<1)
#define HeapRight(i) (HeapLeft(i)+1)
#define HeapValue(heap, i) (heap)->data[(i)-1]

/* Prototypes ****************************************************************/

Heap HeapFromArray(Any *data, int length, CmpFunc *cf, FreeFunc *ff, 
		   HeapIndexFunc *xf);
Heap HeapCreate(CmpFunc *cf, FreeFunc *ff, HeapIndexFunc *xf);
void HeapFree(Heap heap);
void HeapDelete(Heap heap, int i);
void HeapDeleteNoFree(Heap heap, int i);
Any HeapExtractMin(Heap heap);
void HeapInsert(Heap heap, Any value);
void HeapChangedKey(Heap heap, int i);

/* Functions *****************************************************************/

#define HeapLT(i, j) (heap->cf(HeapValue(heap, i), HeapValue(heap, j)) < 0)

static void SiftDown(Heap heap, int i)
{
  int left = HeapLeft(i);
  int right = HeapRight(i);
  int smallest;
  if((left <= heap->length) && HeapLT(left, i))
    smallest = left;
  else
    smallest = i;
  if((right <= heap->length) && HeapLT(right, smallest))
    smallest = right;
  if(smallest != i) {
    Any p = HeapValue(heap, i);
    HeapValue(heap, i) = HeapValue(heap, smallest);
    HeapValue(heap, smallest) = p;
    SiftDown(heap, smallest);
  }
  if(heap->xf) heap->xf(HeapValue(heap, i), i-1);
}

/* data is now property of the heap.
 * If you plan to insert elements, data must be realloc-able. 
 * cf is used to compare elements and ff to free them (ff may be NULL).
 */
Heap HeapFromArray(Any *data, int length, CmpFunc *cf, FreeFunc *ff, 
		   HeapIndexFunc *xf)
{
  int i;
  Heap heap = Allocate(1, struct Heap);
  heap->length = heap->bufsize = length;
  heap->data = data;
  heap->cf = cf;
  heap->ff = ff;
  heap->xf = xf;
  for(i=length; i > length/2; i--) if(xf) heap->xf(HeapValue(heap, i), i-1);
  for(; i > 0; i--) SiftDown(heap, i);
  return heap;
}

Heap HeapCreate(CmpFunc *cf, FreeFunc *ff, HeapIndexFunc *xf)
{
  return HeapFromArray(Allocate(1, Any), 0, cf, ff, xf);
}

void HeapFree(Heap heap)
{
  if(heap->ff) {
    int i;
    for(i=0;i<heap->length;i++) heap->ff(heap->data[i]);
  }
  free(heap->data);
  free(heap);
}

/* i is an index in [0, length) */
void HeapDelete(Heap heap, int i)
{
  i++;
  assert((i > 0) && (i <= heap->length));
  if(heap->ff) heap->ff(HeapValue(heap, i));
  HeapValue(heap, i) = HeapValue(heap, heap->length);
  heap->length--;
  /* don't shrink the array; insert will take care of it */
  if(i > heap->length) return;
  SiftDown(heap, i);
}

void HeapDeleteNoFree(Heap heap, int i)
{
  FreeFunc *ff = heap->ff;
  heap->ff = NULL;
  HeapDelete(heap, i);
  heap->ff = ff;
}

/* Returns NULL on underflow */
Any HeapExtractMin(Heap heap)
{
  Any value;
  if(heap->length == 0) return NULL;
  value = HeapValue(heap, 1);
  HeapDeleteNoFree(heap, 0);
  return value;
}

static void SiftUp(Heap heap, int i, Any value)
{
  for(;
      (i>1) && (heap->cf(value, HeapValue(heap, HeapParent(i))) < 0);
      i = HeapParent(i)) {
    HeapValue(heap, i) = HeapValue(heap, HeapParent(i));
    if(heap->xf) heap->xf(HeapValue(heap, i), i-1);
  }
  HeapValue(heap, i) = value;
  if(heap->xf) heap->xf(HeapValue(heap, i), i-1);
}

void HeapInsert(Heap heap, Any value)
{
  heap->length++;
  if(heap->length > heap->bufsize) {
    heap->bufsize *= 2;
    Reallocate(heap->data, heap->bufsize, Any);
  }
  else if(heap->length < heap->bufsize/4) {
    heap->bufsize /= 2;
    Reallocate(heap->data, heap->bufsize, Any);
  }
  SiftUp(heap, heap->length, value);
}

/* Reposition node i to restore the heap property. */
void HeapChangedKey(Heap heap, int i)
{
  i++;
  assert((i > 0) && (i <= heap->length));
  if((i > 1) && HeapLT(i, HeapParent(i)))
    SiftUp(heap, i, HeapValue(heap, i));
  else
    SiftDown(heap, i);
}
