#include <tpm/index.h>

/* Prototypes ****************************************************************/
Index IndexCreate(HashFunc *hash, CmpFunc *compare, FreeFunc *key_free);
void IndexFree(Index ind);
void IndexInsert(Index ind, Any key);
int IndexDelete(Index ind, int i);
Any IndexAtIndex(Index ind, int i);
int IndexAtKey(Index ind, Any key);
void IndexPrint(FILE *fp, Index ind);

/* Functions *****************************************************************/

Index IndexCreate(HashFunc *hash, CmpFunc *compare, FreeFunc *key_free)
{
  Index ind = Allocate(1, struct Index);
  ind->bufsize = 1024;
  ind->keys = Allocate(ind->bufsize, Any);
  /* keys are held in ind->keys, not the HashTable */
  ind->dict = HashTableCreate(hash, compare, NULL, NULL);
  ind->key_free = key_free;
  ind->missing = ind->length = 0;
  return ind;
}

void IndexFree(Index ind)
{
  int i;
  for(i=0;i<ind->length;i++) ind->key_free(ind->keys[i]);
  free(ind->keys);
  HashTableFree(ind->dict);
  free(ind);
}

/* Does not check for duplicates. */
void IndexInsert(Index ind, Any key)
{
  /* don't update the HashTable until we need to. */
#if 0
  HashEntry *e = HashTableAddEntry(ind->dict, key, &found);
  if(found) return;
  e->value = (Any)ind->length;
#endif
  ind->length++;
  if(ind->length > ind->bufsize) {
    ind->bufsize *= 2;
    Reallocate(ind->keys, ind->bufsize, Any);
  }
  ind->keys[ind->length-1] = key;
}

Any IndexAtIndex(Index ind, int i)
{
  if((i < 0) || (i >= ind->length)) return NULL;
  return ind->keys[i];
}

/* Returns NAI if no such key. */
int IndexAtKey(Index ind, Any key)
{
  HashEntry *e = HashTableFindEntry(ind->dict, key);
  if(e) return (int)e->value;

  if(ind->missing < ind->length) {
    /* insert any missing items */
    for(;ind->missing < ind->length;ind->missing++) {
      int found;
      e = HashTableAddEntry(ind->dict, ind->keys[ind->missing], &found);
      e->value = (Any)ind->missing;
    }
    /* and try again */
    e = HashTableFindEntry(ind->dict, key);
  }
  if(e) return (int)e->value;
  return NAI;
}

/* Deletes the key with index i. 
 * Returns nonzero if no such key.
 */
int IndexDelete(Index ind, int i)
{
  if((i < 0) || (i >= ind->length)) return TPM_ERROR;
  ind->key_free(ind->keys[i]);
  ind->length--;
  if(ind->length < ind->bufsize/4) {
    ind->bufsize /= 2;
    Reallocate(ind->keys, ind->bufsize, Any);
  }
  /* shift down */
  memmove(&ind->keys[i], &ind->keys[i+1], 
	  (ind->length-i)*sizeof(Any));
  /* fix up the HashTable */
  if(ind->missing > i) {
    ind->missing--;
    for(;i<ind->missing;i++) {
      HashEntry *e = HashTableFindEntry(ind->dict, ind->keys[i]);
      assert(e);
      e->value = (Any)i;
    }
  }
  return TPM_OK;
}

void IndexPrint(FILE *fp, Index ind)
{
  Any v;
  HashIter(b, e, v, ind->dict) {
    fprintf(fp, "%s: %d\n", (char*)e->key, (int)v);
  }
}
