/* Code for the memory checker. See memcheck.h for full description. */

#include <stdio.h>
#include <malloc.h>
#include <assert.h>

/* Memory node structure; singly linked list */
typedef struct MemNodeStruct {
  char *real_ptr,             /* Memory block UNIX knows about */
       *fake_ptr;             /* Memory block the application knows about */
  unsigned bytes;             /* Size of the block according to the app. */
  char *file;                 /* File of creation */
  unsigned line;              /* Line of creation */
  struct MemNodeStruct *next; /* List link */
} *MemNode;

/* Globals *******************************************************************/
static MemNode MemNodeList = NULL;
static unsigned max_usage = 0, cur_usage = 0;
static int abort_on_error = 1;
static int use_pickets = 1;
static int search_len = 0, searches = 0;

/* Hopefully this will be aligned on all processors */
#define PICKET_SIZE 16

/* Predicate which is true when node has not been allocated by the checker */
#define UNKNOWN(node) ((node)->real_ptr == (node)->fake_ptr)

/* Private *******************************************************************/

/* Generic error message. Aborts execution depending on abort_on_error.
 */
#ifdef __STDC__
static void show_error(char *s, char *file, unsigned line)
#else
static void show_error(s, file, line)
     char *s;
     char *file;
     unsigned line;
#endif
{
  fprintf(stderr, 
	    "*******************************************************************************\n");
  fprintf(stderr, "%s line %d: %s", file, line, s);
  fprintf(stderr, 
	  "\n*******************************************************************************\n");
  fflush(stderr);
  if(abort_on_error == 1) exit(1);
  else if(abort_on_error == 2) assert(0);
}

/* Lists active memory blocks.
 */
#ifdef __STDC__
static void show_mem_nodes(void)
#else
static void show_mem_nodes()
#endif
{
  MemNode node;

  fprintf(stderr, "Active Memory Blocks:\n");
  for(node = MemNodeList; node; node = node->next) {
    if(UNKNOWN(node))
      fprintf(stderr, "  %s line %d: size unknown\n", node->file, node->line);
    else
      fprintf(stderr, "  %s line %d: %d bytes\n", node->file, node->line, node->bytes);
  }
}

/* Add a node to the memory list.
 */
#ifdef __STDC__
static void add_mem_node(char *real_p, char *fake_p, unsigned bytes, 
			 char *file, unsigned line)
#else
static void add_mem_node(real_p, fake_p, bytes, 
			 file, line)
     char *real_p, *fake_p;
     unsigned bytes;
     char *file;
     unsigned line;
#endif
{
  MemNode node;

  node = (MemNode)malloc(sizeof(struct MemNodeStruct));
  if(!node) {
    show_error("INTERNAL: cannot allocate memory node", file, line);
    /* Here we must abort */
    exit(1);
  }
  node->real_ptr = real_p;
  node->fake_ptr = fake_p;
  node->bytes = bytes;
  node->file = file;
  node->line = line;
  node->next = MemNodeList;
  MemNodeList = node;
  cur_usage += bytes;
  if(cur_usage > max_usage) max_usage = cur_usage;
}

/* Remove a node from the memory list.
 */
#ifdef __STDC__
static void remove_mem_node(MemNode *node_p)
#else
static void remove_mem_node(node_p)
     MemNode *node_p;
#endif
{
  MemNode node;

  node = *node_p;
  cur_usage -= node->bytes;
  *node_p = node->next;
  free(node);
}

/* Find a block in the memory node list. Returns a modifiable pointer
 * to a node (useful for removal).
 */
#ifdef __STDC__
static MemNode *find_mem_node(char *p)
#else
static MemNode *find_mem_node(p)
     char *p;
#endif
{
  MemNode *node_p;

  searches++;
  for(node_p = &MemNodeList; *node_p; node_p = &(*node_p)->next) {
    search_len++;
    if((*node_p)->fake_ptr == p) return node_p;
  }
  return node_p;
}

/* Initialize the pickets for an allocated block.
 */
#ifdef __STDC__
static void write_pickets(char *fake, unsigned bytes)
#else
static void write_pickets(fake, bytes)
     char *fake;
     unsigned bytes;
#endif
{
  int i;
  char *p, *q;

  p = fake - PICKET_SIZE;
  q = fake + bytes + PICKET_SIZE - 1;
  for(i=0;i<PICKET_SIZE;i++) {
    *p++ = *q-- = i+1;
  }
}

/* Check the pickets on an allocated block.
 */
#ifdef __STDC__
static void verify(MemNode node, char *file, unsigned line)
#else
static void verify(node, file, line)
     MemNode node;
     char *file;
     unsigned line;
#endif
{
  int i,bytes;
  char msg[80]; /* Making some assumptions here */
  char *p, *q;

  if(node->fake_ptr == node->real_ptr) return;
  bytes = node->bytes;
  p = node->fake_ptr - PICKET_SIZE;
  q = node->fake_ptr + bytes + PICKET_SIZE-1;

  /* Check that pickets are intact */
  for(i=0;i<PICKET_SIZE;i++) {
    if(*p++ != i+1) {
      if(i == 0)
	sprintf(msg, "detected memory underflow by >= %d bytes", PICKET_SIZE);
      else if(i < PICKET_SIZE-1)
	sprintf(msg, "detected memory underflow by %d bytes", PICKET_SIZE-i);
      else
	sprintf(msg, "detected memory underflow by 1 byte");
      sprintf(msg, "%s\n\tin block allocated in %s line %d", 
	      msg, node->file, node->line);
      show_error(msg, file, line);
      break;
    }
    if(*q-- != i+1) {
      if(i == 0)
	sprintf(msg, "detected memory overflow by >= %d bytes", PICKET_SIZE);
      else if(i < PICKET_SIZE-1)
	sprintf(msg, "detected memory overflow by %d bytes", PICKET_SIZE-i);
      else
	sprintf(msg, "detected memory overflow by 1 byte");
      sprintf(msg, "%s\n\tin block allocated in %s line %d", 
	      msg, node->file, node->line);
      show_error(msg, file, line);
      break;
    }
  }
}

/* Public ********************************************************************/

/* Routine called by MEM_ABORT_ON_ERROR macro.
 */
#ifdef __STDC__
void _mc_abort_on_error(int flag)
#else
void _mc_abort_on_error(flag)
     int flag;
#endif
{
  abort_on_error = flag;
}

/* Routine called by MEM_USE_PICKETS macro.
 */
#ifdef __STDC__
void _mc_use_pickets(int flag)
#else
void _mc_use_pickets(flag)
     int flag;
#endif
{
  use_pickets = flag;
}

/* Routine called by MEM_VERIFY_ALL macro.
 */
#ifdef __STDC__
void _mc_verify_all(char *file, unsigned line)
#else
void _mc_verify_all(file, line)
     char *file;
     unsigned line;
#endif
{
  MemNode node;

  for(node = MemNodeList; node; node=node->next) {
    verify(node, file, line);
  }
}

/* Routine used as replacement for malloc.
 */
#ifdef __STDC__
char *_mc_malloc(unsigned bytes, char *file, unsigned line)
#else
char *_mc_malloc(bytes, file, line)
     unsigned bytes;
     char *file;
     unsigned line;
#endif
{
  char *real, *fake;
  
  if(use_pickets) {
    real = (char*)malloc(bytes + 2*PICKET_SIZE);
    fake = real + PICKET_SIZE;
  }
  else fake = real = (char*)malloc(bytes);
  if(!real) {
    show_error("malloc returned NULL", file, line);
    return real;
  }

  if(use_pickets) {
    /* Fill pickets with data */
    write_pickets(fake, bytes);
  }
  add_mem_node(real, fake, bytes, file, line);
  return fake;
}

/* Routine used as replacement for calloc.
 */
#ifdef __STDC__
char *_mc_calloc(unsigned nitems, unsigned item_size, 
		 char *file, unsigned line)
#else
char *_mc_calloc(nitems, item_size, 
		 file, line)
     unsigned nitems, item_size;
     char *file;
     unsigned line;
#endif
{
  char *real, *fake;
  int bytes, inc;
  
  if(use_pickets) {
    if(item_size >= PICKET_SIZE) inc=1;
    else inc = PICKET_SIZE / item_size + 1;
    real = (char*)calloc(nitems+2*inc, item_size);
    fake = real + item_size*inc;
  }
  else fake = real = (char*)calloc(nitems, item_size);
  bytes = nitems*item_size;
  if(!real) {
    show_error("calloc returned NULL", file, line);
    return real;
  }

  if(use_pickets) {
    /* Fill pickets with data */
    write_pickets(fake, bytes);
  }
  add_mem_node(real, fake, bytes, file, line);
  return fake;
}

/* Routine used as replacement for realloc.
 */
#ifdef __STDC__
char *_mc_realloc(char *p, unsigned bytes, char *file, unsigned line)
#else
char *_mc_realloc(p, bytes, file, line)
     char *p;
     unsigned bytes;
     char *file;
     unsigned line;
#endif
{
  char *real, *fake;
  MemNode *node;

  node = find_mem_node(p);
  if(!*node) {
    if(p) {
      show_error("attempted to realloc a non-NULL unallocated or freed memory block",
		file, line);
      return NULL;
    }
    real = NULL;
  }
  else {
    verify(*node, file, line);
    real = (*node)->real_ptr;
    remove_mem_node(node);
  }

  if(use_pickets) {
    real = (char*)realloc(real, bytes + 2*PICKET_SIZE);
    fake = real + PICKET_SIZE;
  }
  else fake = real = (char*)realloc(real, bytes);
  if(!real) {
    show_error("realloc returned NULL", file, line);
    return real;
  }

  if(use_pickets) {
    /* Fill pickets */
    write_pickets(fake, bytes);
  }
  add_mem_node(real, fake, bytes, file, line);
  return fake;
}

/* Routine used as replacement for free.
 */
#ifdef __STDC__
int _mc_free(char *p, char *file, unsigned line)
#else
int _mc_free(p, file, line)
     char *p;
     char *file;
     unsigned line;
#endif
{
  MemNode *node;

  node = find_mem_node(p);
  if(!*node) {
    show_error("attempted to free an unallocated or freed memory block", 
	      file, line);
    return 0;
  }
  verify(*node, file, line);
  free((*node)->real_ptr);
  remove_mem_node(node);
  return 1;
}

/* Routine used as replacement for strdup.
 */
#ifdef __STDC__
char *_mc_strdup(char *s, char *file, unsigned line)
#else
char *_mc_strdup(s, file, line)
     char *s;
     char *file;
     unsigned line;
#endif
{
  char *str;

  str = _mc_malloc(strlen(s)+1, file, line);
  strcpy(str, s);
  return str;
}

/* Routine called by MEM_ALLOC_NOTIFY macro.
 */
#ifdef __STDC__
void _mc_alloc_notify(char *p, char *file, unsigned line)
#else
void _mc_alloc_notify(p, file, line)
     char *p;
     char *file;
     unsigned line;
#endif
{
  add_mem_node(p, p, 0, file, line);
}

/* Routine called by MEM_FREE_NOTIFY macro.
 */
#ifdef __STDC__
void _mc_free_notify(char *p, char *file, unsigned line)
#else
void _mc_free_notify(p, file, line)
     char *p;
     char *file;
     unsigned line;
#endif
{
  MemNode *node;

  node = find_mem_node(p);
  if(!*node) {
    show_error("attempted to free an unallocated or freed memory block", 
	      file, line);
    return;
  }
  remove_mem_node(node);
}

/* Routine called by MEM_STATUS macro.
 */
#ifdef __STDC__
void _mc_status(char *file, unsigned line)
#else
void _mc_status(file, line)
     char *file;
     unsigned line;
#endif
{
  _mc_verify_all(file, line);
  fprintf(stderr, "%s line %d: Memory Status\n", file, line);
  fprintf(stderr, "  Current bytes allocated: %d\n", cur_usage);
  fprintf(stderr, "  Maximum bytes allocated: %d\n", max_usage);
  fprintf(stderr, "  Average search length:   %g\n", (double)search_len / searches);
}

/* Routine called by MEM_BLOCKS macro.
 */
#ifdef __STDC__
void _mc_mem_nodes(char *file, unsigned line)
#else
void _mc_mem_nodes(file, line)
     char *file;
     unsigned line;
#endif
{
  if(MemNodeList == NULL) {
    fprintf(stderr, "All memory blocks have been deallocated.\n");
    return;
  }
  _mc_verify_all(file, line);
  show_mem_nodes();
}

