/* Code for the memory checker. See memcheck.h for full description. */

#include <stdio.h>
#include <malloc.h>
#include <assert.h>
#include <stdlib.h> /* for size_t on sunOS */

/* Memory node structure */
typedef struct MemNodeStruct {
  char *real_ptr,             /* Memory block UNIX knows about */
       *fake_ptr;             /* Memory block the application knows about */
  unsigned bytes;             /* Size of the block according to the app. */
  char *file;                 /* File of creation */
  unsigned line;              /* Line of creation */
} *MemNode;

/* Internal HashTable implementation *****************************************/

#define Allocate(n,t) (t*)malloc((n)*sizeof(t))
typedef void FreeFunc(void *);

/* Key types */
enum {
  HASH_STRING,
  HASH_WORD
  };

typedef char *HashKey;

typedef struct HashEntryStruct {
  void *value;
  struct HashEntryStruct **bucket;
  struct HashEntryStruct *next;
  char *key;
  char keydata[sizeof(int)]; /* must be last field */
} HashEntry;

typedef struct HashTableStruct {
  int key_type;
  int num_buckets, num_entries;
  int mask, down_shift;
  HashEntry **buckets;
  FreeFunc *ff;
} *HashTable;

#define HashEntryValue(h) (h)->value
#define HashEntryKey(h) (h)->key

#define HashIterate(b, e, ht) \
  for(b=0; b<(ht)->num_buckets; b++) \
     for(e=(ht)->buckets[b]; e; e=e->next)

#define HashIter(b, e, v, ht) \
  int b; HashEntry *e; for(b=0; b<(ht)->num_buckets; b++) \
          for(e=(ht)->buckets[b]; e && (v=(e)->value,e); e=e->next)

#define INITIAL_BUCKETS 4
#define REBUILD_MULTIPLIER 3

/* used by HashTableCreate and HashTableFlush */
static HashTable initialize(HashTable ht)
{
  int i;
  ht->num_buckets = INITIAL_BUCKETS;
  ht->mask = ht->num_buckets - 1; /* assumes num_buckets is power of two */
  ht->down_shift = 28;
  ht->num_entries = 0;
  ht->buckets = Allocate(ht->num_buckets, HashEntry *);
  if(!ht->buckets) {
    fprintf(stderr, "Out of memory; cannot create hash table\n");
    return NULL;
  }
  for(i=0;i<ht->num_buckets;i++) ht->buckets[i] = NULL;
  return ht;
}

/* Create an empty hash table.
 * If key_type == 0, then it will use string keys.
 * Otherwise, it will use integer arrays of length key_type.
 * If ff != NULL, it will be used to free values in the table.
 */
static HashTable HashTableCreate(int key_type, FreeFunc *ff)
{
  HashTable ht = Allocate(1, struct HashTableStruct);
  ht->key_type = key_type;
  ht->ff = ff;
  return initialize(ht);
}

/* used by HashTableFlush and HashTableFree */
static void HashTableRemoveAll(HashTable ht)
{
  int b;
  HashEntry *e, *t;

  if(!ht) return;
  for(b=0;b<ht->num_buckets;b++) {
    for(e=ht->buckets[b];e;e=t) {
      if(ht->ff) ht->ff(e->value);
      t = e->next;
      free(e);
    }
  }
  ht->num_entries = 0;
}

/* Destroys a HashTable, using the ff to free the stored values. */
static void HashTableFree(HashTable ht)
{
  if(!ht) return;
  HashTableRemoveAll(ht);
  free(ht->buckets);
  free(ht);
}

/* taken from Tcl7.3. Produces a hash key for a string. */
static int HashString(char *string)
{
  int result = 0;
  char c;
  for(;;) {
    c = *string++;
    if(!c) break;
    result += (result<<3) + c;
  }
  return result;
}

/* Computes the hash bucket for the given key */
static int HashBucket(HashTable ht, HashKey key)
{
  int i, b;
  if(ht->key_type == HASH_STRING) b = HashString(key);
  else {
    if(ht->key_type > 1) {
      b = *(int*)key;
      for(i=1;i<ht->key_type;i++) {
	b += ((int*)key)[i];
      }
    }
    else b = (int)key;
    b = (b * 1103515245) >> ht->down_shift;
  }
  return (b & ht->mask);
}

/* Used by HashTableFindEntry and HashTableLookupEntry. 
 * Returns the bucket index and HashEntry corresponding to key.
 */
static HashEntry *LookupEntry(HashTable ht, HashKey key, int *bucket)
{
  int b, i;
  HashEntry *e;

  b = HashBucket(ht, key);
  *bucket = b;
  if(ht->key_type == HASH_STRING) {
    for(e = ht->buckets[b]; e; e=e->next)
      if(!strcmp(key, e->key)) break;
  }
  else if(ht->key_type == HASH_WORD) {
    for(e = ht->buckets[b]; e; e=e->next)
      if(key == e->key) break;
  }
  else {
    for(e = ht->buckets[b]; e; e=e->next) {
      /* compare the arrays (key, e->key) */
      for(i=0;i<ht->key_type;i++) {
	if(((int*)key)[i] != ((int*)e->key)[i]) break;
      }
      if(i == ht->key_type) break;
    }
  }
  return e;
}

/* used to expand the table. */
static void RebuildTable(HashTable ht)
{
  HashEntry **old_buckets;
  HashEntry *e, *t;
  int i,b, old_size;

  old_size = ht->num_buckets;
  old_buckets = ht->buckets;

  ht->num_buckets *= 4;
  ht->buckets = Allocate(ht->num_buckets, HashEntry *);
  if(!ht->buckets) {
    fprintf(stderr, "Out of memory; cannot rebuild hash table\n");
    ht->buckets = old_buckets;
    ht->num_buckets = old_size;
    return;
  }
  ht->down_shift -= 2;
  ht->mask = (ht->mask << 2) + 3;
  for(i=0;i<ht->num_buckets;i++) ht->buckets[i] = NULL;
  
  /* rehash all entries */
  for(i=0;i<old_size;i++) {
    for(e=old_buckets[i];e;e=t) {
      b = HashBucket(ht, e->key);
      e->bucket = &ht->buckets[b];
      t = e->next;
      e->next = *e->bucket;
      ht->buckets[b] = e;
    }
  }
  free(old_buckets);
}

/* If ht has no entries matching the key, returns a new HashEntry 
 * corresponding to that key (which has been entered into the hash table) 
 * into which data may be placed, and sets found_return to 0.
 * Otherwise, returns the matching entry and sets found_return to 1.
 */
static HashEntry *HashTableAddEntry(HashTable ht, HashKey key, 
				    int *found_return)
{
  int b;
  HashEntry *e;
  if(!ht) return NULL;
  e = LookupEntry(ht, key, &b);
  if(!e) {
    *found_return = 0;
    if(ht->key_type == HASH_STRING) {
      e = (HashEntry*)malloc(sizeof(HashEntry) + 
			     strlen(key) + 1 - sizeof(e->keydata));
      if(!e) {
	fprintf(stderr, "Out of memory; cannot add hash entry\n");
	return NULL;
      }
      e->key = e->keydata;
      strcpy(e->keydata, key);
    }
    else if(ht->key_type == HASH_WORD) {
      e = Allocate(1, HashEntry);
      if(!e) {
	fprintf(stderr, "Out of memory; cannot add hash entry\n");
	return NULL;
      }
      e->key = key;
    }
    else {
      e = (HashEntry*)malloc(sizeof(HashEntry) + 
			     ht->key_type*sizeof(int) - sizeof(e->keydata));
      if(!e) {
	fprintf(stderr, "Out of memory; cannot add hash entry\n");
	return NULL;
      }
      e->key = e->keydata;
      memcpy(e->keydata, key, ht->key_type*sizeof(int));
    }
    e->bucket = &ht->buckets[b];
    e->next = *e->bucket;
    ht->buckets[b] = e;

    ht->num_entries++;
    if(ht->num_entries >= REBUILD_MULTIPLIER * ht->num_buckets) 
      RebuildTable(ht);
  }
  else *found_return = 1;
  return e;
}

/* Returns the HashEntry matching the key, or NULL if there is none. */
static HashEntry *HashTableFindEntry(HashTable ht, HashKey key)
{
  int b;
  if(!ht) return NULL;
  return LookupEntry(ht, key, &b);
}

/* Removes the entry from the hash table, and uses the ff to free the 
 * entry data.
 */
static void HashTableRemoveEntry(HashTable ht, HashEntry *entry)
{
  HashEntry **e;

  if(!ht) return;
  for(e = entry->bucket; *e != entry; e = &(*e)->next);
  *e = entry->next;
  if(ht->ff) ht->ff(entry->value);
  free(entry);
  ht->num_entries--;
}

/* Globals *******************************************************************/
static HashTable memTable = NULL; /* holds the MemNodes */
static unsigned max_usage = 0, cur_usage = 0;
static int abort_on_error = 1;
static int use_pickets = 1;

/* Hopefully this will be aligned on all processors */
#define PICKET_SIZE 16

/* Predicate which is true when node has not been allocated by the checker */
#define UNKNOWN(node) ((node)->real_ptr == (node)->fake_ptr)

#define find_mem_node(p) HashTableFindEntry(memTable, p)

/* Private *******************************************************************/

/* Generic error message. Aborts execution depending on abort_on_error.
 */
#ifdef __STDC__
static void show_error(char *s, char *file, unsigned line)
#else
static void show_error(s, file, line)
     char *s;
     char *file;
     unsigned line;
#endif
{
  fprintf(stderr, 
	  "*******************************************************************************\n");
  fprintf(stderr, "%s line %d: %s\n", file, line, s);
  fprintf(stderr, 
	  "*******************************************************************************\n");
  fflush(stderr);
  if(abort_on_error == 1) exit(1);
  else if(abort_on_error == 2) assert(0);
}

/* Lists active memory blocks.
 */
#ifdef __STDC__
static void show_mem_nodes(void)
#else
static void show_mem_nodes()
#endif
{
  MemNode node;

  fprintf(stderr, "Active Memory Blocks:\n");
  {HashIter(b, e, node, memTable) {
    if(UNKNOWN(node))
      fprintf(stderr, "  %s line %d: size unknown\n", node->file, node->line);
    else
      fprintf(stderr, "  %s line %d: %d bytes\n", node->file, node->line, node->bytes);
  }}
}

/* Add a node to the memory list.
 */
#ifdef __STDC__
static void add_mem_node(char *real_p, char *fake_p, unsigned bytes, 
			 char *file, unsigned line)
#else
static void add_mem_node(real_p, fake_p, bytes, 
			 file, line)
     char *real_p, *fake_p;
     unsigned bytes;
     char *file;
     unsigned line;
#endif
{
  MemNode node;
  int found;

  node = (MemNode)malloc(sizeof(struct MemNodeStruct));
  if(!node) {
    show_error("INTERNAL: cannot allocate memory node", file, line);
    /* Here we must abort */
    exit(1);
  }
  node->real_ptr = real_p;
  node->fake_ptr = fake_p;
  node->bytes = bytes;
  node->file = file;
  node->line = line;
  cur_usage += bytes;
  if(cur_usage > max_usage) max_usage = cur_usage;

  if(!memTable) memTable = HashTableCreate(1, free);
  HashTableAddEntry(memTable, fake_p, &found)->value = node;
  if(found) {
    fprintf(stderr, "memchecker internal error: hash collision\n");
  }
}

/* Remove a node from the memory list.
 */
#ifdef __STDC__
static void remove_mem_node(HashEntry *e)
#else
static void remove_mem_node(e)
     HashEntry *e;
#endif
{
  cur_usage -= ((MemNode)e->value)->bytes;
  HashTableRemoveEntry(memTable, e);
  if(memTable->num_entries == 0) {
    HashTableFree(memTable);
    memTable = NULL;
  }
}

/* Initialize the pickets for an allocated block.
 */
#ifdef __STDC__
static void write_pickets(char *fake, unsigned bytes)
#else
static void write_pickets(fake, bytes)
     char *fake;
     unsigned bytes;
#endif
{
  int i;
  char *p, *q;

  p = fake - PICKET_SIZE;
  q = fake + bytes + PICKET_SIZE - 1;
  for(i=0;i<PICKET_SIZE;i++) {
    *p++ = *q-- = i+1;
  }
}

/* Check the pickets on an allocated block.
 */
#ifdef __STDC__
static void verify(MemNode node, char *file, unsigned line)
#else
static void verify(node, file, line)
     MemNode node;
     char *file;
     unsigned line;
#endif
{
  int i,bytes;
  char msg[256]; /* Making some assumptions here */
  char *p, *q;

  if(node->fake_ptr == node->real_ptr) return;
  bytes = node->bytes;
  p = node->fake_ptr - PICKET_SIZE;
  q = node->fake_ptr + bytes + PICKET_SIZE-1;

  /* Check that pickets are intact */
  for(i=0;i<PICKET_SIZE;i++) {
    if(*p++ != i+1) {
      if(i == 0)
	sprintf(msg, "detected memory underflow by >= %d bytes", PICKET_SIZE);
      else if(i < PICKET_SIZE-1)
	sprintf(msg, "detected memory underflow by %d bytes", PICKET_SIZE-i);
      else
	sprintf(msg, "detected memory underflow by 1 byte");
      sprintf(msg, "%s\n\tin block allocated in %s line %d", 
	      msg, node->file, node->line);
      show_error(msg, file, line);
      break;
    }
    if(*q-- != i+1) {
      if(i == 0)
	sprintf(msg, "detected memory overflow by >= %d bytes", PICKET_SIZE);
      else if(i < PICKET_SIZE-1)
	sprintf(msg, "detected memory overflow by %d bytes", PICKET_SIZE-i);
      else
	sprintf(msg, "detected memory overflow by 1 byte");
      sprintf(msg, "%s\n\tin block allocated in %s line %d", 
	      msg, node->file, node->line);
      show_error(msg, file, line);
      break;
    }
  }
}

/* Public ********************************************************************/

/* Routine called by MEM_ABORT_ON_ERROR macro.
 */
#ifdef __STDC__
void _mc_abort_on_error(int flag)
#else
void _mc_abort_on_error(flag)
     int flag;
#endif
{
  abort_on_error = flag;
}

/* Routine called by MEM_USE_PICKETS macro.
 */
#ifdef __STDC__
void _mc_use_pickets(int flag)
#else
void _mc_use_pickets(flag)
     int flag;
#endif
{
  use_pickets = flag;
}

/* Routine called by MEM_VERIFY_ALL macro.
 */
#ifdef __STDC__
void _mc_verify_all(char *file, unsigned line)
#else
void _mc_verify_all(file, line)
     char *file;
     unsigned line;
#endif
{
  MemNode node;

  if(!memTable) return;
  {HashIter(b, e, node, memTable) {
    verify(node, file, line);
  }}
}

/* Routine used as replacement for malloc.
 */
#ifdef __STDC__
void *_mc_malloc(size_t bytes, char *file, unsigned line)
#else
char *_mc_malloc(bytes, file, line)
     size_t bytes;
     char *file;
     unsigned line;
#endif
{
  char *real, *fake;

  if(bytes < 0) {
    show_error("negative value passed to malloc", file, line);
    return NULL;
  }
  if(use_pickets) {
    real = (char*)malloc(bytes + 2*PICKET_SIZE);
    fake = real + PICKET_SIZE;
  }
  else fake = real = (char*)malloc(bytes);
  if(!real) {
    show_error("malloc returned NULL", file, line);
    return real;
  }

  if(use_pickets) {
    /* Fill pickets with data */
    write_pickets(fake, bytes);
  }
  add_mem_node(real, fake, bytes, file, line);
  return fake;
}

/* Routine used as replacement for calloc.
 */
#ifdef __STDC__
void *_mc_calloc(size_t nitems, size_t item_size, 
		 char *file, unsigned line)
#else
char *_mc_calloc(nitems, item_size, 
		 file, line)
     size_t nitems, item_size;
     char *file;
     unsigned line;
#endif
{
  char *real, *fake;
  int bytes, inc;
  
  if((nitems < 0) || (item_size < 0)) {
    show_error("negative value passed to calloc", file, line);
    return NULL;
  }
  bytes = nitems*item_size;
  if(use_pickets) {
    /* can't use the real calloc here, so we fake it */
    real = (char*)malloc(bytes + 2*PICKET_SIZE);
    memset(real, 0, bytes + 2*PICKET_SIZE);
    fake = real + PICKET_SIZE;
  }
  else fake = real = (char*)calloc(nitems, item_size);
  if(!real) {
    show_error("calloc returned NULL", file, line);
    return real;
  }

  if(use_pickets) {
    /* Fill pickets with data */
    write_pickets(fake, bytes);
  }
  add_mem_node(real, fake, bytes, file, line);
  return fake;
}

/* Routine used as replacement for realloc.
 */
#ifdef __STDC__
void *_mc_realloc(void *p, size_t bytes, char *file, unsigned line)
#else
char *_mc_realloc(p, bytes, file, line)
     char *p;
     size_t bytes;
     char *file;
     unsigned line;
#endif
{
  char *real, *fake;
  HashEntry *entry;

  if(bytes < 0) {
    show_error("negative value passed to realloc", file, line);
    return NULL;
  }
  entry = HashTableFindEntry(memTable, p);
  if(!entry) {
    if(p) {
      show_error("attempted to realloc a non-NULL unallocated or freed memory block",
		file, line);
      return NULL;
    }
    real = NULL;
  }
  else {
    MemNode node = entry->value;
    verify(node, file, line);
    real = node->real_ptr;
    remove_mem_node(entry);
  }

  if(use_pickets) {
    real = (char*)realloc(real, bytes + 2*PICKET_SIZE);
    fake = real + PICKET_SIZE;
  }
  else fake = real = (char*)realloc(real, bytes);
  if(!real) {
    show_error("realloc returned NULL", file, line);
    return real;
  }

  if(use_pickets) {
    /* Fill pickets */
    write_pickets(fake, bytes);
  }
  add_mem_node(real, fake, bytes, file, line);
  return fake;
}

/* Routine used as replacement for free.
 */
#ifdef __STDC__
int _mc_free(void *p, char *file, unsigned line)
#else
int _mc_free(p, file, line)
     char *p;
     char *file;
     unsigned line;
#endif
{
  MemNode node;
  HashEntry *entry;

  entry = find_mem_node(p);
  if(!entry) {
    show_error("attempted to free an unallocated or freed memory block", 
	      file, line);
    return 0;
  }
  node = entry->value;
  verify(node, file, line);
  free(node->real_ptr);
  remove_mem_node(entry);
  return 1;
}

/* Routine used as replacement for strdup.
 */
#ifdef __STDC__
char *_mc_strdup(char *s, char *file, unsigned line)
#else
char *_mc_strdup(s, file, line)
     char *s;
     char *file;
     unsigned line;
#endif
{
  char *str;

  str = _mc_malloc(strlen(s)+1, file, line);
  strcpy(str, s);
  return str;
}

/* Routine called by MEM_ALLOC_NOTIFY macro.
 */
#ifdef __STDC__
void _mc_alloc_notify(void *p, char *file, unsigned line)
#else
void _mc_alloc_notify(p, file, line)
     char *p;
     char *file;
     unsigned line;
#endif
{
  add_mem_node(p, p, 0, file, line);
}

/* Routine called by MEM_FREE_NOTIFY macro.
 */
#ifdef __STDC__
void _mc_free_notify(void *p, char *file, unsigned line)
#else
void _mc_free_notify(p, file, line)
     char *p;
     char *file;
     unsigned line;
#endif
{
  HashEntry *entry;

  entry = find_mem_node(p);
  if(!entry) {
    show_error("attempted to free an unallocated or freed memory block", 
	      file, line);
    return;
  }
  remove_mem_node(entry);
}

/* Routine called by MEM_STATUS macro.
 */
#ifdef __STDC__
void _mc_status(char *file, unsigned line)
#else
void _mc_status(file, line)
     char *file;
     unsigned line;
#endif
{
  _mc_verify_all(file, line);
  fprintf(stderr, "%s line %d: Memory Status\n", file, line);
  fprintf(stderr, "  Current bytes allocated: %d\n", cur_usage);
  fprintf(stderr, "  Maximum bytes allocated: %d\n", max_usage);
  fprintf(stderr, "  Number of memory blocks: %d\n", 
	  memTable ? memTable->num_entries : 0);
}

/* Routine called by MEM_BLOCKS macro.
 */
#ifdef __STDC__
void _mc_mem_nodes(char *file, unsigned line)
#else
void _mc_mem_nodes(file, line)
     char *file;
     unsigned line;
#endif
{
  if(!memTable) {
    fprintf(stderr, "All memory blocks have been deallocated.\n");
    return;
  }
  _mc_verify_all(file, line);
  show_mem_nodes();
}

