/* 
 * NameSpaces define a root context, current context, and function for
 * handling unknown names.
 *
 * NameContexts are HashTables of objects, NameContexts, and links.
 * If the key starts with '/', the value is a NameContext.
 * Otherwise the value is an object.
 */

#include "name.h"

/* Prototypes ****************************************************************/

NameSpace NameSpaceCreate(FreeFunc *ff);
void NameSpaceFree(NameSpace ns);
NameContext NameCreateContextInContext(NameSpace ns, NameContext context, 
				       char *name);
int NameCreateObjectInContext(NameSpace ns, NameContext context, 
			      char *name, void *object);
void *NameLookupInContext(NameSpace ns, NameContext context,
			  char *name, int *type_return);
int NameRemoveInContext(NameSpace ns, NameContext context, char *name);
void NameDump(FILE *fp, NameContext context);
char *NameOfContext(NameContext context);

/* Private */
static void NameContextFree(NameContext context, FreeFunc *ff);

/* Functions *****************************************************************/

static NameContext NameContextCreate(NameContext parent)
{
  NameContext context;
  context = Allocate(1, struct NameContext);
  context->parent = parent;
  context->entries = HashTableCreate(HASH_STRING, NULL);
  return context;
}

/* Creates a new, empty NameSpace.
 * If ff != NULL, it will be used to free removed objects.
 */
NameSpace NameSpaceCreate(FreeFunc *ff)
{
  NameSpace ns = Allocate(1, struct NameSpace);
  ns->ff = ff;
  ns->root = NameContextCreate(NULL);
  ns->current = ns->root;
  ns->unknown = NULL;
  return ns;
}

static void EntryFree(HashEntry *e, FreeFunc *ff)
{
  /* is it a context? */
  if(e->key[0] == '/') NameContextFree(e->value, ff);
  else if(ff) ff(e->value);
}

static void NameContextFree(NameContext context, FreeFunc *ff)
{
  int b;
  HashEntry *e, *t;

  /* this is a special inlined version of HashTableFree */
  for(b=0;b<context->entries->num_buckets;b++) {
    for(e=context->entries->buckets[b];e;e=t) {
      EntryFree(e, ff);
      t = e->next;
      free(e);
    }
  }
  context->entries->num_buckets = 0;
  /* now that we've removed all the entries, let the library clean up */
  HashTableFree(context->entries);
  free(context);
}

/* Frees all storage associated with a NameSpace.
 * Uses the ff, if it was specified, to free the objects.
 */
void NameSpaceFree(NameSpace ns)
{
  NameContextFree(ns->root, ns->ff);
  free(ns);
}

/* Returns the context and last component of a path.
 * Examples:
 *   foo       -> current, "foo"
 *   /foo      -> root, "foo"
 *   bar/foo   -> current(bar), "foo"
 *   /bar/foo  -> root(bar), "foo"
 * Illegal:
 *   //foo
 *   foo/
 * Returns NULL if:
 *   1. The name has doubled slashes or a trailing slash.
 *   2. The name has a nonexistent context component.
 */
static char *ContextOfName(NameSpace ns, NameContext *context, 
			   char *name)
{
  HashEntry *e;
  char *str; /* copy of the name */
  char *p; /* current position in str */
  int i;
  char *last; /* last component of name */

  /* is the name empty? */
  if(!name[0]) return NULL;

  /* walk backwards through the name to find the last component */
  i = strlen(name)-1;
  for(;i>=0;i--) if(name[i] == '/') break;
  last = name+i+1; 
  /* check for double or ending slash */
  if(!*last) return NULL;

  /* if only one component, we're done */
  if(i == -1) return last;

  /* is it a simple root reference? */
  if(i == 0) { *context = ns->root; return last; }

  /* set name to be only the context */
  name[i] = '\0';

  /* is the name absolute? */
  if(name[0] == '/') {
    if(i == 1) { /* doubled leading slash */ return NULL; }
    *context = ns->root;
    str = strdup(name);
  }
  else {
    /* prefix one character */
    str = Allocate(strlen(name)+2, char);
    str[0] = '/';
    strcpy(str+1, name);
  }
  /* restore the name */
  name[i] = '/';

  /* walk down the path */
  for(p = strtok(str+1, "/");p;p = strtok(NULL, "/")) {
    /* is the component empty? */
    if(!*p) { last = NULL; break; }

    /* lookup with a leading slash */
    p[-1] = '/';
    e = HashTableFindEntry((*context)->entries, p-1);
    if(!e) { 
      /* missing context. call the unknown name function to handle it. */
      if(!ns->unknown ||
	 !(*context = ns->unknown(ns, *context, p-1, &i, ns->userData))) {
	last = NULL; break; 
      }
    }
    else *context = e->value;
  }
  free(str);
  return last;
}

/* Creates a new context using the given context.
 * <name> may be a relative or absolute path.
 * Returns the new context, or NULL if the path was improper.
 */
NameContext NameCreateContextInContext(NameSpace ns, NameContext context, 
				       char *name)
{
  HashEntry *e;
  int found;
  NameContext new;
  char *newname;

  name = ContextOfName(ns, &context, name);
  if(!name) return NULL;

  /* prefix the name with a '/' */
  newname = Allocate(strlen(name)+2, char);
  newname[0] = '/';
  strcpy(newname+1, name);

  new = NameContextCreate(context);
  e = HashTableAddEntry(context->entries, newname, &found);
  if(found) EntryFree(e, ns->ff);
  e->value = new;
  free(newname);
  return new;
}

/* Names an object using the given context.
 * <name> may be a relative or absolute path.
 * Returns 0 if the path was improper.
 */
int NameCreateObjectInContext(NameSpace ns, NameContext context, 
			      char *name, void *object)
{
  HashEntry *e;
  int found;
  
  name = ContextOfName(ns, &context, name);
  if(!name) return 0;

  e = HashTableAddEntry(context->entries, name, &found);
  if(found) EntryFree(e, ns->ff);
  e->value = object;
  return 1;
}

/* Looks up a pathname in the given context.
 * If type_return != NULL, then
 *   type_return is set to 0 if the result is an object,
 *   and 1 if the result is a context.
 * Returns NULL is there is nothing of that name.
 */
void *NameLookupInContext(NameSpace ns, NameContext context,
			  char *name, int *type_return)
{
  HashEntry *e;
  char *newname;

  name = ContextOfName(ns, &context, name);
  if(!name) return NULL;

  /* try without slash */
  e = HashTableFindEntry(context->entries, name);
  if(e) {
    /* it's an object */
    if(type_return) *type_return = 0;
    return e->value;
  }

  /* try with slash */
  newname = Allocate(strlen(name)+2, char);
  newname[0] = '/';
  strcpy(newname+1, name);

  e = HashTableFindEntry(context->entries, newname);
  free(newname);
  if(e) {
    /* it's a context */
    if(type_return) *type_return = 1;
    return e->value;
  }

  /* failed; try the unknown name function */
  if(!ns->unknown ||
     !(context = ns->unknown(ns, context, name, type_return, ns->userData)))
    return NULL;
  /* it succeeded */
  return context;
}

/* Returns 0 if the name was improper.
 */
int NameRemoveInContext(NameSpace ns, NameContext context, char *name)
{
  HashEntry *e;
  char *newname;
  int dummy;

  name = ContextOfName(ns, &context, name);
  if(!name) return 0;

 tryagain:
  /* try without slash */
  e = HashTableFindEntry(context->entries, name);
  if(e) {
    /* it's an object */
    if(ns->ff) ns->ff(e->value);
    HashTableRemoveEntry(context->entries, e);
    return 1;
  }

  /* try with slash */
  newname = Allocate(strlen(name)+2, char);
  newname[0] = '/';
  strcpy(newname+1, name);

  e = HashTableFindEntry(context->entries, newname);
  free(newname);
  if(e) {
    /* it's a context */
    NameContextFree(e->value, ns->ff);
    HashTableRemoveEntry(context->entries, e);
    return 1;
  }

  /* failed; try the unknown name function */
  if(!ns->unknown ||
     !ns->unknown(ns, context, name, &dummy, ns->userData))
    return 0;
  /* it succeeded; now delete what it created. */
  goto tryagain;
}

static void NameDumpHelper(FILE *fp, NameContext context, int indent)
{
  int b;
  HashEntry *e;

  HashIterate(b, e, context->entries) {
    int i;
    for(i=0;i<indent;i++) fputc(' ', fp);
    if(e->key[0] == '/') {
      fputs(e->key+1, fp);
      fputc('/', fp);
      fputc('\n', fp);
      NameDumpHelper(fp, e->value, indent+2);
    }
    else {
      fputs(e->key, fp);
      fputc('\n', fp);
    }
  }
}

void NameDump(FILE *fp, NameContext context)
{
  NameDumpHelper(fp, context, 0);
}

char *NameOfContext(NameContext context)
{
  int b;
  HashEntry *e;
  NameContext parent = context->parent;
  if(!parent) return "/";
  /* ask the parent who it is */
  HashIterate(b, e, parent->entries) {
    if(e->value == context) return e->key+1;
  }
  /* something went wrong */
  return NULL;
}
