/* Pathname utilities.
 * Would like to add a TclTranslate-type function.
 */
#include <tpm/tpm.h>
#include <dirent.h>
#include <sys/stat.h>
#include <pwd.h>

#if defined(sgi) || defined(linux) || defined(sparc)
#define NAMLEN(e) strlen(e->d_name)
#else
#define NAMLEN(e) e->d_namlen
#endif

/* Prototypes ****************************************************************/

char *Path_Parent(char *path);
char *Path_LastName(char *path);
char *Path_TrueName(char *path);
char *Path_Ups(char *path);
char *Path_Home(char *user);

/* Functions *****************************************************************/

/* If path denotes a file, returns the directory containing the file.
 * Otherwise returns the parent directory.
 * Caller must free the result.
 * If path is not valid, returns NULL.
 */
char *Path_Parent(char *path)
{
  struct stat st;
  if(stat(path, &st) < 0) {
    ErrorSet("cannot stat `%s'", path);
    return NULL;
  }

  /* open our parent */
  /* if it's a file, remove the last component */
  if((st.st_mode & S_IFMT) != S_IFDIR) {
    char *p = NULL, *q;
    /* set p to the last slash */
    for(q = path;*q;q++) if(*q == '/') p = q;
    if(!p) {
      /* no slashes */
      return strdup(".");
    }
    else if(p == path) {
      /* leading slash */
      return strdup("/");
    }
    else {
      *p = 0;
      q = strdup(path);
      *p = '/';
      return q;
    }
  }
  /* otherwise append "/.." */
  else {
    int len = strlen(path) + 4;
    char *result = (char*)malloc(len);
    sprintf(result, "%s/..", path);
    return result;
  }
  /* never gets here */
}

/* Returns the name of the file named by path in its parent. 
 * Caller must free the result.
 * Returns NULL on error.
 */
char *Path_LastName(char *path)
{
  struct stat st;
  int len;
  dev_t this_dev;
  ino_t this_ino;
  int mount_point;
  char *name, *parent;
  DIR *dir;
  struct dirent *ent;

  if(stat(path, &st) < 0) {
    ErrorSet("cannot stat `%s'", path);
    return NULL;
  }
  this_dev = st.st_dev;
  this_ino = st.st_ino;

  parent = Path_Parent(path);
  assert(parent);
  if(stat(parent, &st) < 0) {
    ErrorSet("cannot stat `%s'", parent);
    goto error;
  }
  mount_point = st.st_dev != this_dev;
  /* is this the root? (parent == self) */
  if(this_ino == st.st_ino && this_dev == st.st_dev) {
    free(parent);
    return strdup("/");
  }

  /* scan for our inode */
  dir = opendir(parent);
  if(dir == NULL) {
    ErrorSet("opendir failed on `%s'", parent);
    return NULL;
  }
  len = strlen(parent)+1;
  while( (ent = readdir(dir)) ) {
    /* ignore '.' and '..' */
    if(ent->d_name[0] == '.' &&
       (NAMLEN(ent) == 1 || (NAMLEN(ent) == 2 && ent->d_name[1] == '.')))
      continue;
    if(mount_point) {
      /* have to stat each entry to see if it's us */
      name = (char*)malloc(len + NAMLEN(ent) + 1);
      memcpy(name, parent, len-1);
      name[len-1] = '/';
      memcpy(name+len, ent->d_name, NAMLEN(ent)+1);
      if(lstat(name, &st) < 0) { free(name); closedir(dir); goto error; }
      free(name);
      if(st.st_dev == this_dev && st.st_ino == this_ino) break;
    }
    else if(ent->d_ino == this_ino) break;
  }
  if(!ent) { 
    closedir(dir); 
    ErrorString = "directory entry not found"; 
    goto error; 
  }

  name = (char*)malloc(NAMLEN(ent) + 1);
  memcpy(name, ent->d_name, NAMLEN(ent)+1);
  closedir(dir);
  free(parent);
  return name;
 error:
  free(parent);
  return NULL;
}

/* Returns a canonical name (without .., ., or symlinks) for path. 
 * Caller must free the result.
 * Returns NULL on error.
 * Path_TrueName("") is same as getcwd().
 */
char *Path_TrueName(char *path)
{
  struct stat st;
  char *result, *name;
  DIR *dir;
  struct dirent *ent;
  int plen, rlen, len;
  dev_t this_dev, root_dev, dot_dev;
  ino_t this_ino, root_ino, dot_ino;
  int mount_point;

  rlen = 0;
  result = strdup("/");
  if(!*path) path = strdup(".");
  else path = strdup(path);

  if(stat(path, &st) < 0) {
    ErrorSet("cannot stat path `%s'", path);
    goto error;
  }
  this_dev = st.st_dev;
  this_ino = st.st_ino;

  if(stat("/", &st) < 0) {
    ErrorString = "cannot stat /";
    goto error;
  }
  root_dev = st.st_dev;
  root_ino = st.st_ino;

  /* loop until we reach the root */
  while(root_ino != this_ino || root_dev != this_dev) {
    char *parent = Path_Parent(path);
    assert(parent);
    free(path);
    path = parent;
    plen = strlen(path) + 1;
    if(stat(path, &st) < 0) {
      ErrorSet("cannot stat parent `%s'", path);
      goto error;
    }
    dot_dev = st.st_dev;
    dot_ino = st.st_ino;
    mount_point = dot_dev != this_dev;
   
    /* scan for our inode */
    dir = opendir(path);
    if(dir == NULL) {
      ErrorSet("opendir failed on `%s'", path);
      goto error;
    }
    while( (ent = readdir(dir)) ) {
      /* ignore '.' and '..' */
      if(ent->d_name[0] == '.' &&
	 (NAMLEN(ent) == 1 || (NAMLEN(ent) == 2 && ent->d_name[1] == '.')))
	continue;
      if(mount_point) {
	/* have to stat each entry to see if it's us */
	name = (char*)malloc(plen + NAMLEN(ent) + 1);
	memcpy(name, path, plen-1);
	name[plen-1] = '/';
	memcpy(name+plen, ent->d_name, NAMLEN(ent)+1);
	if(lstat(name, &st) < 0) { free(name); closedir(dir); goto error; }
	free(name);
	if(st.st_dev == this_dev && st.st_ino == this_ino) break;
      }
      else if(ent->d_ino == this_ino) break;
    }
    if(!ent) { 
      closedir(dir); 
      ErrorString = "directory entry not found"; 
      goto error; 
    }
    
    /* put ent->d_name at the front of result */
    len = NAMLEN(ent);
    result = (char*)realloc(result, rlen + len + 2);
    if(rlen) memmove(result + len + 1, result, rlen+2);
    else result[len+1] = 0;
    memcpy(result + 1, ent->d_name, len);
    rlen += len + 1;
    closedir(dir);
#if 0
    fprintf(stderr, "%s\n", result);
#endif

    /* become our parent */
    this_dev = dot_dev;
    this_ino = dot_ino;
  }
  free(path);
  return result;
 error:
  free(path);
  free(result);
  return NULL;
}

char *Path_Ups(char *path)
{
  struct stat st;
  dev_t root_dev;
  ino_t root_ino;
  int len, plen;
  char *parent, *result;

  /* already at root? */
  if(!path[0] || (path[0] == '/' && !path[1])) return strdup("");

  parent = Path_Parent(path);
  if(!parent) return NULL;
  len = strlen(parent);
  /* remove trailing slash */
  if(parent[len-1] == '/') {
    len--;
    parent[len] = 0;
  }
  len++;
  plen = len;

  if(stat("/", &st) < 0) goto error;
  root_dev = st.st_dev;
  root_ino = st.st_ino;

  for(;;) {
    if(stat(parent, &st) < 0) {
      ErrorSet("cannot stat parent `%s'", parent);
      goto error;
    }
    if(root_ino == st.st_ino && root_dev == st.st_dev) break;
    plen += 3;
    Reallocate(parent, plen, char);
    strcat(parent, "/..");
  }
#if 0
  result = strdup(parent+len);
#else
  result = strdup(parent);
#endif
  free(parent);
  return result;
 error:
  free(parent);
  return NULL;
}

char *Path_Home(char *user)
{
  if(!user[0]) {
    return strdup(getenv("HOME"));
  }
  else {
    char *result;
    struct passwd *pwPtr = getpwnam(user);
    if(!pwPtr) result = NULL;
    else result = strdup(pwPtr->pw_dir);
    endpwent();
    return result;
  }
}
