#include <tpm/tpm.h>
#include "region.h"
#include "concave.h"

/* Prototypes ****************************************************************/

Tpm_Region *Tpm_RegionCreate(void);
void Tpm_RegionFree(Tpm_Region *r);
void Tpm_RegionSetDims(Tpm_Region *r, int width, int height);
void Tpm_RegionRectangle(Tpm_Region *r, int x, int y, int width, int height);
Tpm_Region *Tpm_RegionFromPolygon(Tpm_Polygon *p);
Tpm_Polygon *Tpm_RegionPolygon(Tpm_Region *r);
Tpm_Polygon **Tpm_RegionToPolygons(Tpm_Region *r);
Tpm_Region *Tpm_RegionIntersection(Tpm_Region *a, Tpm_Region *b);
int Tpm_RegionArea(Tpm_Region *r);
int Tpm_RegionContains(Tpm_Region *r, int x, int y);

void Tpm_RegionWritePGM(FILE *h, Tpm_Region *r);
void Tpm_RegionPrint(FILE *h, Tpm_Region *r);
void Tpm_RegionScan(FILE *h, Tpm_Region *r);

/* Functions *****************************************************************/

/* Create a new, empty region */
Tpm_Region *Tpm_RegionCreate(void)
{
  Tpm_Region *r = Allocate(1, Tpm_Region);
  r->x = r->y = r->width = r->height = 0;
  r->data = NULL;
  return r;
}

/* Destroy a region */
void Tpm_RegionFree(Tpm_Region *r)
{
  if(!r) return;
  if(r->data) GenericMatrixFree(r->data);
  free(r);
}

/* Expand and clear a region */
void Tpm_RegionSetDims(Tpm_Region *r, int width, int height)
{
  if(r->data) GenericMatrixFree(r->data);
  r->width = width;
  r->height = height;
  r->data = GenericMatrixAlloc(r->height, r->width, 1);
}

void Tpm_RegionRectangle(Tpm_Region *r, int x, int y, int width, int height)
{
  Tpm_RegionSetDims(r, width, height);
  r->x = x;
  r->y = y;
  memset(r->data[0], 1, width*height);
}

/* output a region as a PGM raw image. */
void Tpm_RegionWritePGM(FILE *h, Tpm_Region *r)
{
  int i,j;
  fprintf(h, "P5\n%d %d %d\n", r->width, r->height, 255);
  for(i=0;i<r->height;i++) {
    for(j=0;j<r->width;j++) {
      if(r->data[i][j])
	putc(255, h);
      else
	putc(0, h);
    }
  }
}

/* output a region as a binary mask. does not output position or extent. */
void Tpm_RegionPrint(FILE *h, Tpm_Region *r)
{
  int i,j;
  for(i=0;i<r->height;i++) {
    for(j=0;j<r->width;j++) {
      if(r->data[i][j])
	putc(-1, h);
      else
	putc(0, h);
    }
  }
}

/* fill in the data of a region */
void Tpm_RegionScan(FILE *h, Tpm_Region *r)
{
  int i,j;
  for(i=0;i<r->height;i++) {
    for(j=0;j<r->width;j++) {
      r->data[i][j] = !!getc(h);
    }
  }
}

/* used by Tpm_RegionFromPolygon */
static void SpanSet(Tpm_Region *r, int y, int xl, int xr)
{
  memset(&r->data[y - r->y][xl - r->x], 1, xr-xl+1);
}

/* Return a region corresponding to the interior of p. */
Tpm_Region *Tpm_RegionFromPolygon(Tpm_Polygon *p)
{
  Window w;
  Tpm_Region *r = Tpm_RegionCreate();
  int i;

  if(!p->npoints) return r;
  /* find the extent of the vertices */
  w.x0 = w.x1 = p->points[0].x;
  w.y0 = w.y1 = p->points[0].y;
  for(i=1;i<p->npoints;i++) {
    w.x0 = min(w.x0, p->points[i].x);
    w.y0 = min(w.y0, p->points[i].y);
    w.x1 = max(w.x1, p->points[i].x);
    w.y1 = max(w.y1, p->points[i].y);
  }
  r->x = w.x0; r->y = w.y0;
  Tpm_RegionSetDims(r, w.x1-w.x0+1, w.y1-w.y0+1);

  /* scan-convert the polygon */
  concave(p, &w, SpanSet, r);

  return r;
}

/* Directions:
 * 5 6 7
 * 4   0
 * 3 2 1
 */
static void GoDirection(XPoint *p, int dir, XPoint *d)
{
  switch(dir) {
  case 0:
  case 1:
  case 7:
    d->x = p->x + 1;
    break;
  case 3:
  case 4:
  case 5:
    d->x = p->x - 1;
    break;
  default:
    d->x = p->x;
  }
  switch(dir) {
  case 1:
  case 2:
  case 3:
    d->y = p->y + 1;
    break;
  case 5:
  case 6:
  case 7:
    d->y = p->y - 1;
    break;
  default:
    d->y = p->y;
  }
}

/* Returns the contour of r.
 * Assumes only one connected region without holes.
 */
Tpm_Polygon *Tpm_RegionPolygon(Tpm_Region *r)
{
  XPoint start, cur, new;
  int x, y;
  int dir, new_dir;
  Tpm_Polygon *poly = Tpm_PolygonCreate();

  /* find a starting point */
  for(y=0;y<r->height;y++) {
    for(x=0;x<r->width;x++) {
      if(r->data[y][x]) {
	start.x = x + r->x; start.y = y + r->y;
	goto found;
      }
    }
  }
  /* the region is empty */
  return poly;

 found:
  cur = start;
  dir = 1;
  do {
    Tpm_PolygonAddPoint(poly, &cur);
    new_dir = (dir+7)%8;
    for(;;) {
      GoDirection(&cur, new_dir, &new);
      if(Tpm_RegionContains(r, new.x, new.y)) {
	cur = new;
	dir = new_dir;
	break;
      }
      new_dir = (new_dir+1)%8;
      /* did we go all the way around? */
      if(new_dir == (dir+7)%8) break;
    }
  } while((cur.x != start.x) || (cur.y != start.y));

  return poly;
}

/* Returns a NULL-terminated array of Polygons,
 * which are the contours of the connected regions in r.
 * Holes are ignored.
 */
#if 0
/* Would be better if used concave() to clear the polygons. */
Tpm_Polygon **Tpm_RegionToPolygons(Tpm_Region *r)
{
  Tpm_Polygon **polys;
  int npolys;
  typedef struct {
    int y,x1,x2;
  } Scan;
  Scan *scans;
  int nscans;

  npolys = 0;
  polys = Allocate(1, Tpm_Polygon*);
  scans = Allocate(r->height*4, Scan);

  /* Find and remove polygons from r */
  for(;;) {
    XPoint start, cur, new;
    int x, y;
    int dir, new_dir;
    Tpm_Polygon *poly;

    /* find a starting point */
    for(y=0;y<r->height;y++) {
      for(x=0;x<r->width;x++) {
	if(r->data[y][x]) {
	  start.x = x + r->x; start.y = y + r->y;
	  goto found;
	}
      }
    }
    /* the region is empty */
    break;

  found:
    /* extract the polygon starting at <start> */
    poly = Tpm_PolygonCreate();
    cur = start;
    dir = 0;
    nscans = 0;
    do {
      /* add the point to the scanline list */
      /* are we going up or down? */
      if(dir < 4) { /* going down or right */
	/* find a scan with the same y coordinate */
	for(y=0;y<nscans;y++) {
	  if(scans[y].y == cur.y) {
	    /* is it unfilled and to the left? */
	    if(scans[y].x2 == -1) {
	      if(scans[y].x1 <= cur.x) {
		/* fill it */
		scans[y].x2 = cur.x;
		break;
	      }
	    }
	    /* is it intersected? */
	    else if(scans[y].x1 < cur.x && scans[y].x2 > cur.x) {
	      /* break off the right half */
	      scans[nscans].y = cur.y;
	      scans[nscans].x1 = -1;
	      scans[nscans].x2 = scans[y].x2;
	      scans[y].x2 = cur.x;
	      break;
	    }
	    /* is it adjacent? */
	    else if(scans[y].x2+1 == cur.x) {
	      scans[y].x2 = cur.x;
	      break;
	    }
	  }
	}
	if(y == nscans) {
	  /* didn't find one; create new */
	  scans[nscans].y = cur.y;
	  if(dir == 0) {
	    scans[nscans].x1 = cur.x;
	    scans[nscans].x2 = -1;
	  }
	  else {
	    scans[nscans].x1 = -1;
	    scans[nscans].x2 = cur.x;
	  }
	  nscans++;
	}
      }
      else { /* going up or left */
	/* find a scan with the same y coordinate */
	for(y=0;y<nscans;y++) {
	  if(scans[y].y == cur.y) {
	    /* is it unfilled and to the right? */
	    if(scans[y].x1 == -1) {
	      if(scans[y].x2 >= cur.x) {
		scans[y].x1 = cur.x;
		break;
	      }
	    }
	    /* is it intersected? */
	    else if(scans[y].x1 < cur.x && scans[y].x2 > cur.x) {
	      /* break off the left half */
	      scans[nscans].y = cur.y;
	      scans[nscans].x1 = scans[y].x1;
	      scans[nscans].x2 = -1;
	      scans[y].x1 = cur.x;
	      break;
	    }
	    /* is it adjacent? */
	    else if(scans[y].x1-1 == cur.x) {
	      scans[y].x1 = cur.x;
	      break;
	    }
	  }
	}
	if(y == nscans) {
	  /* didn't find one; create new */
	  scans[nscans].y = cur.y;
	  if(dir == 4) {
	    scans[nscans].x1 = -1;
	    scans[nscans].x2 = cur.x;
	  }
	  else {
	    scans[nscans].x1 = cur.x;
	    scans[nscans].x2 = -1;
	  }
	  nscans++;
	}
      }

      Tpm_PolygonAddPoint(poly, &cur);
      new_dir = (dir+7)%8;
      for(;;) {
	GoDirection(&cur, new_dir, &new);
	if(Tpm_RegionContains(r, new.x, new.y)) {
	  cur = new;
	  dir = new_dir;
	  break;
	}
	new_dir = (new_dir+1)%8;
	/* did we go all the way around? */
	if(new_dir == (dir+7)%8) break;
      }
    } while((cur.x != start.x) || (cur.y != start.y));

    /* add the polygon to the result */
    polys[npolys] = poly;
    npolys++;
    polys = (Tpm_Polygon**)realloc(polys, (npolys+1)*sizeof(Tpm_Polygon*));

    /* erase the polygon */
    for(y=0;y<nscans;y++) {
      for(x=scans[y].x1;x<=scans[y].x2;x++)
	r->data[scans[y].y - r->y][x - r->x] = 0;
    }
  }
  free(scans);
  polys[npolys] = NULL;
  return polys;
}
#endif

/* used by Tpm_RegionToPolygons */
static void SpanClear(Tpm_Region *r, int y, int xl, int xr)
{
  memset(&r->data[y - r->y][xl - r->x], 0, xr-xl+1);
}

/* Returns a NULL-terminated array of Polygons,
 * which are the contours of the connected regions in r.
 * Holes are ignored.
 */
Tpm_Polygon **Tpm_RegionToPolygons(Tpm_Region *r)
{
  Tpm_Polygon **polys;
  int npolys;
  Window w;

  npolys = 0;
  polys = Allocate(1, Tpm_Polygon*);
  w.x0 = r->x;
  w.x1 = r->x + r->width - 1;
  w.y0 = r->y;
  w.y1 = r->y + r->height - 1;

  /* Find and remove polygons from r */
  for(;;) {
    Tpm_Polygon *poly = Tpm_RegionPolygon(r);
    if(poly->npoints == 0) {
      Tpm_PolygonFree(poly);
      break;
    }

#if 0
    {FILE *h = fopen("region_foo", "w");
     Tpm_RegionWritePgm(h, r);
     fclose(h);}
#endif

    /* add the polygon to the result */
    polys[npolys] = poly;
    npolys++;
    polys = (Tpm_Polygon**)realloc(polys, (npolys+1)*sizeof(Tpm_Polygon*));

    /* erase the polygon */
    concave(poly, &w, SpanClear, r);
  }

  polys[npolys] = NULL;
  return polys;
}

/* Returns a new region which includes the points in both a and b.
 * Returns NULL if this region would be empty. 
 */
Tpm_Region *Tpm_RegionIntersection(Tpm_Region *a, Tpm_Region *b)
{
  Tpm_Region *r;
  int xa, ya, xb, yb;
  unsigned width,height;
  unsigned x,y;

  /* find the bounding box of intersection */
  if(a->x < b->x) {
    xa = b->x - a->x;
    if(xa >= a->width) return NULL;
    xb = 0;
  }
  else {
    xa = 0;
    xb = a->x - b->x;
    if(xb >= b->width) return NULL;
  }
  if(a->y < b->y) {
    ya = b->y - a->y;
    if(ya >= a->height) return NULL;
    yb = 0;
  }
  else {
    ya = 0;
    yb = a->y - b->y;
    if(yb >= b->height) return NULL;
  }
  width = min(a->width - xa, b->width - xb);
  height = min(a->height - ya, b->height - yb);

  r = Allocate(1, Tpm_Region);
  r->x = a->x + xa;
  r->y = a->y + ya;
  r->width = width;
  r->height = height;
  r->data = GenericMatrixAlloc(height, width, 1);
  for(y=0;y<height;y++) {
    for(x=0;x<width;x++) {
      r->data[y][x] = a->data[ya+y][xa+x] & b->data[yb+y][xb+x];
    }
  }
  return r;
}

/* Returns the number of points included in r. */
int Tpm_RegionArea(Tpm_Region *r)
{
  int y,x;
  unsigned count=0;
  for(y=0;y<r->height;y++) {
    for(x=0;x<r->width;x++) {
      if(r->data[y][x]) count++;
    }
  }
  return count;
}

/* Returns 1 iff r contains the point (x, y) */
int Tpm_RegionContains(Tpm_Region *r, int x, int y)
{
  x -= r->x;
  if((x < 0) || (x >= r->width)) return 0;
  y -= r->y;
  if((y < 0) || (y >= r->height)) return 0;
  return r->data[y][x];
}
