/*
 * Concave Polygon Scan Conversion
 * by Paul Heckbert
 * from "Graphics Gems", Academic Press, 1990
 */

/*
 * concave: scan convert nvert-sided concave polygon with vertices at
 * (point[i].x, point[i].y) for i in [0..nvert-1] within the window win by
 * calling spanproc for each visible span of pixels.
 * Polygon can be clockwise or counterclockwise.
 * Algorithm does uniform point sampling at pixel centers.
 * Inside-outside test done by Jordan's rule: a point is considered inside if
 * an emanating ray intersects the polygon an odd number of times.
 * spanproc should fill in pixels from xl to xr inclusive on scanline y,
 * e.g:
 *	void spanproc(int y, int xl, int xr)
 *	{
 *	    int x;
 *	    for (x=xl; x<=xr; x++)
 *		pixel_write(x, y, pixelvalue);
 *	}
 *
 *  Paul Heckbert	30 June 81, 18 Dec 89
 *
 * Modified by TPM 8/21/94 ANSI, user_data, poly, hack for last scanline
 */

#include <stdio.h>
#include <math.h>
#include <tpm/tpm.h>

#define ALLOC(ptr, type, n)  ptr = (type *)malloc((n)*sizeof(type))

typedef struct {		/* window: a discrete 2-D rectangle */
    int x0, y0;			/* xmin and ymin */
    int x1, y1;			/* xmax and ymax (inclusive) */
} Window;

typedef struct {		/* a polygon edge */
    double x;	/* x coordinate of edge's intersection with current scanline */
    double dx;	/* change in x with respect to y */
    int i;	/* edge number: edge i goes from pt[i] to pt[i+1] */
} Edge;

typedef struct {
  short x,y;
} XPoint;

typedef struct {
  int npoints;
  XPoint *points;
} Tpm_Polygon;

static int n;
static XPoint *pt;

static int nact;	       /* number of active edges */
static Edge *active;	       /* active edge list:edges crossing scanline y */

static int compare_ind(), compare_active();
static void cdelete(int i);
static void cinsert(int i, int y);

typedef void SpanProc(void *data, int y, int xl, int xr);

/* nvert = number of vertices */
/* point = vertices of polygon */
/* win = screen clipping window */
/* spanproc = called for each span of pixels */
/* user_data = passed as first arg to spanproc */
void concave(Tpm_Polygon *poly, Window *win, 
	     void(*spanproc)(), void *user_data)
{
  int k, y0, y1, y, i, j, xl, xr;
  int *ind;		/* list of vertex indices, sorted by pt[ind[j]].y */
  
  if(poly->npoints <= 0) return;

  if(poly->npoints == 1) {
    /* make a single point */
    ((SpanProc*)spanproc)(user_data, poly->points[0].y, 
			  poly->points[0].x, poly->points[0].x);
    return;
  }
  n = poly->npoints;
  pt = poly->points;
  ALLOC(ind, int, n);
  ALLOC(active, Edge, n+2); /* give a buffer of 2 */
  
  /* create y-sorted array of indices ind[k] into vertex list */
  for (k=0; k<n; k++) {
    ind[k] = k;
  }
  /* sort ind by pt[ind[k]].y */
  qsort(ind, n, sizeof(ind[0]), compare_ind);	
  
  /* start with empty active list */
  nact = 0;				
  /* ind[k] is next vertex to process */
  k = 0;				
  /* ymin of polygon */
  y0 = max(win->y0, pt[ind[0]].y);		
  /* ymax of polygon */
  y1 = min(win->y1, pt[ind[n-1]].y);

  /* step through scanlines */
  for (y=y0; y<=y1; ) {
    /* check vertices between previous scanline and current one, if any */
    for (; k<n && pt[ind[k]].y<=y; k++) {
      i = ind[k];	
      /*
       * insert or delete edges before and after vertex i (i-1 to i,
       * and i to i+1) from active list if they cross scanline y
       */
      j = i>0 ? i-1 : n-1;      /* vertex previous to i */
      if (pt[j].y < y) {        /* old edge, remove from active list */
	cdelete(j);
      }
      else if (pt[j].y > y) {   /* new edge, add to active list */
	cinsert(j, y);
      }

      j = i<n-1 ? i+1 : 0;      /* vertex next after i */
      if (pt[j].y < y) {        /* old edge, remove from active list */
	cdelete(i);
      }
      else if (pt[j].y > y) {	/* new edge, add to active list */
	cinsert(i, y);
      }

      /* sanity check */
#if 0
      printf("%d\n", nact);
#endif
      assert(nact <= n+2);
    }
    /* no active edges? */
    if((nact == 0) && (y == y0)) {
      /* must be a single horizontal line */
      active[0].dx = 0;
      active[0].x = pt[ind[0]].x;
      active[1].dx = 0;
      active[1].x = pt[ind[n-1]].x;
      nact=2;
    }

    /* sort active edge list by active[j].x */
    qsort(active, nact, sizeof(active[0]), compare_active);

    /* draw horizontal segments for scanline y */
    for (j=0; j<nact; j+=2) {	/* draw horizontal segments */
      /* span between j & j+1 if inside, 
	 span between j+1 & j+2 if outside */
      xl = (int)(active[j].x+.5);	/* left end of span */
      xr = (int)(active[j+1].x+.5);	/* right end of span */
      /* clip to the window */
      if (xl < win->x0) xl = win->x0;
      if (xr > win->x1) xr = win->x1;
      /* draw pixels in span */
      if (xl <= xr) ((SpanProc*)spanproc)(user_data, y, xl, xr);
      /* increment edge coords */
      active[j].x += active[j].dx;	
      active[j+1].x += active[j+1].dx;
    }
    if(++y > y1) break;
    /* draw horizontal segments for next scanline.
     * inefficient, but works.
     */
    for (j=0; j<nact; j+=2) {	/* draw horizontal segments */
      /* span between j & j+1 if inside, 
	 span between j+1 & j+2 if outside */
      xl = (int)(active[j].x+.5);	/* left end of span */
      xr = (int)(active[j+1].x+.5);	/* right end of span */
      /* clip to the window */
      if (xl < win->x0) xl = win->x0;
      if (xr > win->x1) xr = win->x1;
      /* draw pixels in span */
      if (xl <= xr) ((SpanProc*)spanproc)(user_data, y, xl, xr);
      /* do not increment edge coords */
    }
  }

  free(ind);
  free(active);
}

/* remove edge i from active list */
static void cdelete(int i)
{
    int j;

    for (j=0; j<nact && active[j].i!=i; j++);
    if (j>=nact) return;  /* edge not in active list; happens at win->y0 */

    nact--;
    memcpy(&active[j], &active[j+1], (nact-j)*sizeof active[0]);
}

/* append edge i to end of active list */
static void cinsert(int i, int y)
{
    int j;
    double dx;
    XPoint *p, *q;

    j = i<n-1 ? i+1 : 0;    /* vertex next after i */
    if (pt[i].y < pt[j].y) {p = &pt[i]; q = &pt[j];}
    else		   {p = &pt[j]; q = &pt[i];}
    /* initialize x position at intersection of edge with scanline y */
    if(q->y == p->y) return; /* ignore horiz edges */
    dx = (double)(q->x-p->x)/(q->y-p->y);

    active[nact].dx = dx;
    active[nact].x = dx*(y-p->y)+p->x;
    active[nact].i = i;
    nact++;
}

/* comparison routines for qsort */
static int compare_ind(int *u, int *v) {return pt[*u].y <= pt[*v].y ? -1 : 1;}
static int compare_active(Edge *u, Edge *v) {return u->x <= v->x ? -1 : 1;}
