#include "spanTree.h"

/* Prototypes ****************************************************************/
SpanTree SpanTreeCreate(int num_nodes);
void SpanTreeFree(SpanTree st);
int SpanTreeRead(SpanTree st, FILE *h);
SpanTree SpanTreeCreateFromStream(FILE *h);
SpanTree SpanTreeCreateFromFile(char *file);
int SpanTreeWrite(SpanTree st, FILE *h);
int SpanTreePrint(FILE *h, SpanTree st);
void SpanTreeDistances(SpanTree st, int i, float *dist);

/* Functions *****************************************************************/

/* Return a new uninitialized spanning tree with the given number of nodes. */
SpanTree SpanTreeCreate(int num_nodes)
{
  SpanTree st = Allocate(1, struct SpanTree);
  st->num_nodes = num_nodes;
  st->node = Allocate(num_nodes, struct SpanTreeNode);
  return st;
}

void SpanTreeFree(SpanTree st)
{
  free(st->node);
  free(st);
}

/* File format is
 * length
 * edge parent
 * ...
 * with edge and parent data stored in binary for fast I/O.
 */
/* Reads node data from file into an existing SpanTree. */
int SpanTreeRead(SpanTree st, FILE *h)
{
  int status = fread(st->node, 1, st->num_nodes*sizeof(struct SpanTreeNode), h);
  if(!status) return TPM_ERROR;
#ifdef TPM_SWAP
  assert(sizeof(int) == sizeof(float));
  SwapBuffer(st->node, st->num_nodes*sizeof(struct SpanTreeNode)/sizeof(int), 
	     sizeof(int));
#endif
  return TPM_OK;
}

SpanTree SpanTreeCreateFromStream(FILE *h)
{
  SpanTree st;
  int num_nodes;

  /* read header */
  if(fscanf(h, "%d", &num_nodes) < 1) return NULL;
  getc(h); /* absorb the newline; not safe if done by fscanf */
  st = SpanTreeCreate(num_nodes);
  if(SpanTreeRead(st, h)) {
    /* Show a message, but don't do anything */
    fprintf(stderr, "SpanTreeCreateFromStream: error reading file\n");
  }
  return st;
}

SpanTree SpanTreeCreateFromFile(char *file)
{
  FILE *h;
  SpanTree st;
  if(file[0] == '-' && !file[1]) h = stdin;
  else h = fopen(file, "r");
  if(!h) return NULL;
  st = SpanTreeCreateFromStream(h);
  fclose(h);
  return st;
}

int SpanTreeWrite(SpanTree st, FILE *h)
{
  int status;
  fprintf(h, "%d\n", st->num_nodes);
#ifdef TPM_SWAP
  assert(sizeof(int) == sizeof(float));
  SwapBuffer(st->node, st->num_nodes*sizeof(struct SpanTreeNode)/sizeof(int), 
	     sizeof(int));
#endif
  status = fwrite(st->node, 1, st->num_nodes*sizeof(struct SpanTreeNode), h);
#ifdef TPM_SWAP
  assert(sizeof(int) == sizeof(float));
  SwapBuffer(st->node, st->num_nodes*sizeof(struct SpanTreeNode)/sizeof(int), 
	     sizeof(int));
#endif
  if(!status) return TPM_ERROR;
  return TPM_OK;
}

int SpanTreePrint(FILE *h, SpanTree st)
{
  int i;
  fprintf(h, "%d nodes\n", st->num_nodes);
  for(i=0;i<st->num_nodes;i++) {
    fprintf(h, "%d: %d %g\n", i, st->node[i].parent, st->node[i].edge);
  }
  return TPM_OK;
}

#define COMBINE_MAX 1
/* Fills dist with the distance of every node to i. */
void SpanTreeDistances(SpanTree st, int i, float *dist)
{
  int j;
  float f;
  /* distance from x to its ancestors */
  float *d = Allocate(st->num_nodes, float);
  for(j=0;j<st->num_nodes;j++) {
    d[j] = 1.0/0;
  }
  /* compute costs going up */
  f = 0;
  j = i;
  while(j >= 0) {
    d[j] = f;
#if COMBINE_MAX
    f = max(f, st->node[j].edge);
#else
    f += st->node[j].edge;
#endif
    j = st->node[j].parent;
  }
  /* loop nodes */
  for(i=0;i<st->num_nodes;i++) {
    /* find an ancestor with finite d[j] */
    dist[i] = d[i];
    f = 0;
    j = i;
    while(j >= 0) {
      if(finite(d[j])) { 
#if COMBINE_MAX
	dist[i] = max(f, d[j]); 
#else
	dist[i] = f + d[j]; 
#endif
	break; 
      }
#if COMBINE_MAX
      f = max(f, st->node[j].edge);
#else
      f += st->node[j].edge;
#endif
      j = st->node[j].parent;
    }
  }
  free(d);
}
