/* Implementation of some useful stream operations.
 * In particular, implements DiskArrays, which are Streams
 * over packed arrays (e.g. a movie produced from concatenated image files).
 */

#include <tpm/tpm.h>
#include <tpm/stream.h>
#include <ctype.h>

/* Prototypes ****************************************************************/

FileHandle FileOpen(char *fname, char *mode);
void FileClose(FileHandle h);
FileHandle fdup(FileHandle h);
int getline(char *buffer, int bufsize, FILE *fp);
int FileCheckHeader(FILE *fp, char *data, int length);
size_t FileSize(FileHandle h);

void DiskArrayOpen(Stream *s, FileHandle h, int num_items, int item_size);

int Tpm_ReadIntBinary(int *value_return, Tpm_ReadCharFunc *readChar, Any data);
int Tpm_WriteIntBinary(int value, Tpm_WriteCharFunc *writeChar, Any data);
int Tpm_ReadFloatBinary(float *value_return, Tpm_ReadCharFunc *readChar,
			Any data);
int Tpm_WriteFloatBinary(float value, Tpm_WriteCharFunc *writeChar, Any data);
int Tpm_ReadDoubleBinary(double *value_return, Tpm_ReadCharFunc *readChar,
			 Any data);
int Tpm_WriteDoubleBinary(double value, Tpm_WriteCharFunc *writeChar, Any data);
int Tpm_ReadStringBinary(char **value_return, Tpm_ReadCharFunc *readChar, 
			 Any data);
int Tpm_WriteStringBinary(char *value, Tpm_WriteCharFunc *writeChar, Any data);

/* Functions *****************************************************************/

FileHandle FileOpen(char *fname, char *mode)
{
  FileHandle h;

  h = fopen(fname, mode);
  if(!h) {
    fprintf(stderr, "Cannot open file `%s': ", fname);
    perror("");
    fflush(stderr);
    exit(1);
  }
  return h;
}

void FileClose(FileHandle h)
{
  if(fclose(h) == EOF) {
    perror("Cannot close file");
    exit(1);
  }
}

/* Duplicate a file handle. Reading from the second handle before
 * closing the first can produce strange results.
 */
FileHandle fdup(FileHandle h)
{
  return fdopen(dup(fileno(h)), "r");
}

/* A safe way to read lines from a file.
 * Reads up to bufsize characters, terminating at newlines and EOF.
 * Newlines do not show up in the buffer.
 * Returns 0 on EOF.
 */
int getline(char *buffer, int bufsize, FILE *fp)
{
  int i;

  fgets(buffer, bufsize, fp);
  if(feof(fp)) return 0;
  /* Remove trailing whitespace */
  i = strlen(buffer)-1;
  for(i = strlen(buffer)-1; isspace(buffer[i]) && (i >= 0); i--);
  buffer[i+1]=0;
  return 1;
}

int FileCheckHeader(FILE *fp, char *data, int length)
{
  int i,status = TPM_OK;
  for(i=0;i<length;i++) {
    int c = getc(fp);
    if((char)c != *data) { 
      if(c != EOF) ungetc(c, fp); 
      status = TPM_ERROR; 
      break; 
    }
    data++;
  }
#if 0
  data--;
  for(;i;i--) ungetc(*data, fp);
#else
  if(i) fseek(fp, -i, SEEK_CUR);
#endif
  return status;
}

size_t FileSize(FileHandle h)
{
  size_t pos = ftell(h), len;
  fseek(h, 0, SEEK_END);
  len = ftell(h);
  fseek(h, pos, SEEK_SET);
  return len;
}

/* DiskArray *****************************************************************/

typedef struct {
  int current_pos, offset;
  int item_size, num_items;
} *DiskArrayData;

/* Reads data from the specified position in the disk array,
 * avoiding fseeks whenever it can.
 */
static int DiskArrayRead(FileHandle h, DiskArrayData data, 
			 int index, void *dest)
{
  int pos;

  if(index == -1) {
    index = (data->current_pos - data->offset) / data->item_size;
  }
  if(index >= data->num_items) return 0;
  pos = index * data->item_size + data->offset;
  if(data->current_pos != pos) {
    if(fseek(h, pos - data->current_pos, SEEK_CUR) == -1) {
      perror("DiskArrayRead");
      return 0;
    }
  }
  if(!fread(dest, data->item_size, 1, h)) return 0;
  data->current_pos = pos + data->item_size;
  return 1;
}

static int DiskArrayClose(FileHandle h, DiskArrayData data)
{
  int status;

  status = fclose(h);
  free(data);
  if(status == EOF) {
    perror("DiskArrayClose");
    return 0;
  }
  return 1;
}

/* Make s into a disk array of num_items items. 
 * Each item is exactly item_size bytes, with no padding in between.
 * Use StreamRead or StreamNext to read out of the disk array.
 */
void DiskArrayOpen(Stream *s, FileHandle h, int num_items, int item_size)
{
  DiskArrayData data = (DiskArrayData)malloc(sizeof(*data));

  s->source = h;
  data->current_pos = data->offset = ftell(h);
  data->item_size = item_size;
  data->num_items = num_items;
  s->data = data;
  s->readFunc = (StreamReadFunc*)DiskArrayRead;
  s->closeFunc = (StreamCloseFunc*)DiskArrayClose;
}

/* Character conversion ******************************************************/

int Tpm_ReadIntBinary(int *value_return, Tpm_ReadCharFunc *readChar, Any data)
{
  /* should check for EOF */
  *value_return = readChar(data) << 0;
  *value_return += readChar(data) << 8;
  *value_return += readChar(data) << 16;
  *value_return += readChar(data) << 24;
  return TPM_OK;
}

int Tpm_WriteIntBinary(int value, Tpm_WriteCharFunc *writeChar, Any data)
{
  return 
    (writeChar((value & 0x000000ff) >> 0, data) == EOF) ||
    (writeChar((value & 0x0000ff00) >> 8, data) == EOF) ||
    (writeChar((value & 0x00ff0000) >> 16, data) == EOF) ||
    (writeChar((value & 0xff000000) >> 24, data) == EOF);
}

int Tpm_ReadFloatBinary(float *value_return, Tpm_ReadCharFunc *readChar,
			Any data)
{
  char *p = (char*)value_return;
#ifdef TPM_SWAP
  *p++ = readChar(data);
  *p++ = readChar(data);
  *p++ = readChar(data);
  *p++ = readChar(data);
#else
  p += 3;
  *p-- = readChar(data);
  *p-- = readChar(data);
  *p-- = readChar(data);
  *p-- = readChar(data);
#endif
  return TPM_OK;
}

int Tpm_WriteFloatBinary(float value, Tpm_WriteCharFunc *writeChar, Any data)
{
  char *p = (char*)&value;
#ifdef TPM_SWAP
  writeChar(*p, data); p++;
  writeChar(*p, data); p++;
  writeChar(*p, data); p++;
  writeChar(*p, data); p++;
#else
  p += 3;
  writeChar(*p, data); p--;
  writeChar(*p, data); p--;
  writeChar(*p, data); p--;
  writeChar(*p, data); p--;
#endif
  return TPM_OK;
}

int Tpm_ReadDoubleBinary(double *value_return, Tpm_ReadCharFunc *readChar,
			 Any data)
{
  char *p = (char*)value_return;
#ifdef TPM_SWAP
  *p++ = readChar(data);
  *p++ = readChar(data);
  *p++ = readChar(data);
  *p++ = readChar(data);
  *p++ = readChar(data);
  *p++ = readChar(data);
  *p++ = readChar(data);
  *p++ = readChar(data);
#else
  p += 7;
  *p-- = readChar(data);
  *p-- = readChar(data);
  *p-- = readChar(data);
  *p-- = readChar(data);
  *p-- = readChar(data);
  *p-- = readChar(data);
  *p-- = readChar(data);
  *p-- = readChar(data);
#endif
  return TPM_OK;
}

int Tpm_WriteDoubleBinary(double value, Tpm_WriteCharFunc *writeChar, Any data)
{
  char *p = (char*)&value;
#ifdef TPM_SWAP
  writeChar(*p, data); p++;
  writeChar(*p, data); p++;
  writeChar(*p, data); p++;
  writeChar(*p, data); p++;
  writeChar(*p, data); p++;
  writeChar(*p, data); p++;
  writeChar(*p, data); p++;
  writeChar(*p, data); p++;
#else
  p += 7;
  writeChar(*p, data); p--;
  writeChar(*p, data); p--;
  writeChar(*p, data); p--;
  writeChar(*p, data); p--;
  writeChar(*p, data); p--;
  writeChar(*p, data); p--;
  writeChar(*p, data); p--;
  writeChar(*p, data); p--;
#endif
  return TPM_OK;
}

#include "strstream.h"
int Tpm_ReadStringBinary(char **value_return, Tpm_ReadCharFunc *readChar, 
			 Any data)
{
  StrStream ss = StrStreamCreate(16, STRSTREAM_UNLIMITED);
  int c;
  while( (c = readChar(data)) ) {
    StrStreamPutc(ss, c);
  }
  *value_return = strdup(StrStreamAsString(ss));
  StrStreamFree(ss);
  return TPM_OK;
}

int Tpm_WriteStringBinary(char *value, Tpm_WriteCharFunc *writeChar, Any data)
{
  while(*value) {
    writeChar(*value, data); value++;
  }
  return writeChar(0, data);
}
