#include <tpm/tpm.h>
#include <tpm/strstream.h>
#include <ctype.h>

/* padding used to reduce number of reallocs */
#define INCREMENT 16

/* Prototypes ****************************************************************/

StrStream StrStreamCreate(int len, int limit);
void StrStreamFree(StrStream ss);
StrStream StrStreamFromData(char *s, int len, int limit);
StrStream StrStreamFromString(char *s, int len, int limit);
char *StrStreamAsString(StrStream ss);
int StrStreamSeekR(StrStream ss, long offset, int whence);
int StrStreamSeekW(StrStream ss, long offset, int whence);

int StrStreamPut(StrStream ss, char *data, int len);
int StrStreamPutc(StrStream ss, char c);
int StrStreamPuts(StrStream ss, char *s);
int StrStreamPutInt(StrStream ss, int i);
int StrStreamPutDouble(StrStream ss, double d);

int StrStreamGet(StrStream ss, char *buffer, int len);
int StrStreamGetc(StrStream ss, char *c);
char *StrStreamGetLine(StrStream ss);
char *StrStreamGetWord(StrStream ss);
char *StrStreamGetSexp(StrStream ss);
int StrStreamGetInt(StrStream ss, int *i);
int StrStreamGetDouble(StrStream ss, double *dp);
int StrStreamUnget(StrStream ss, int i);

int StrStreamScanWord(FILE *fp, StrStream ss);
int StrStreamScanSexp(FILE *fp, StrStream ss);
int StrStreamScanLine(FILE *fp, StrStream ss);

/* Functions *****************************************************************/

/* Returns a new StrStream of the given length and length limit.
 * limit must be != 0, because the data is allocated here.
 */
StrStream StrStreamCreate(int len, int limit)
{
  if(!limit) return NULL;
  return StrStreamFromData(Allocate(len, char), len, limit);
}

/* Frees the data if ss->limit != 0 */
void StrStreamFree(StrStream ss)
{
  if(ss->limit) free(ss->data);
  free(ss);
}

/* Creates a StrStream around the data in s.
 * If limit != 0, then s belongs to the StrStream 
 * and will be freed by StrStreamFree.
 */
StrStream StrStreamFromData(char *s, int len, int limit)
{
  StrStream ss = Allocate(1, struct StrStream);
  ss->len = len;
  ss->limit = limit;
  ss->rpos = ss->wpos = 0;
  ss->data = s;
  return ss;
}

/* Same as StrStreamFromData but 
 * wpos is set to the end of s (at the null terminator).
 */
StrStream StrStreamFromString(char *s, int len, int limit)
{
  StrStream ss = StrStreamFromData(s, len, limit);
  ss->wpos = strlen(s);
  return ss;
}

char *StrStreamAsString(StrStream ss)
{
  /* append a NULL to the end of the stream, 
     if there isn't one there already */
  if(ss->data[ss->wpos]) {
    if(StrStreamPutc(ss, 0) == TPM_ERROR) return NULL;
    ss->wpos--;
  }
  return ss->data+ss->rpos;
}

static int Expand(StrStream ss, int new_size)
{
  if(new_size >= ss->len) {
    if(!ss->limit || (new_size >= ss->limit)) {
      ErrorString = TPM_EFULL;
      return TPM_ERROR;
    }
    ss->len = new_size + INCREMENT;
    if(ss->len > ss->limit) ss->len = ss->limit;
    ss->data = (char*)realloc(ss->data, ss->len);
  }
  return TPM_OK;
}

int StrStreamSeekR(StrStream ss, long offset, int whence)
{
  unsigned new_pos;
  if(!whence) new_pos = offset;
  else new_pos = ss->rpos + offset;
  if((-offset > ss->rpos) || (new_pos > ss->wpos)) return TPM_ERROR;
  ss->rpos = new_pos;
  return TPM_OK;
}

int StrStreamSeekW(StrStream ss, long offset, int whence)
{
  unsigned new_pos;
  if(!whence) new_pos = offset;
  else new_pos = ss->wpos + offset;
  if((new_pos < ss->rpos) || Expand(ss, new_pos)) return TPM_ERROR;
  ss->wpos = new_pos;
  return TPM_OK;
}

/* Put ***********************************************************************/

int StrStreamPut(StrStream ss, char *data, int len)
{
  unsigned new_pos = ss->wpos + len;
  TRY_RETURN(Expand(ss, new_pos));
  memcpy(ss->data+ss->wpos, data, len);
  ss->wpos = new_pos;
  return TPM_OK;
}

int StrStreamPutc(StrStream ss, char c)
{
  TRY_RETURN(Expand(ss, ss->wpos+1));
  ss->data[ss->wpos++] = c;
  return TPM_OK;
}

int StrStreamPuts(StrStream ss, char *s)
{
  return StrStreamPut(ss, s, strlen(s));
}

int StrStreamPutInt(StrStream ss, int i)
{
  unsigned len, f, neg = 0;

  if(i < 0) { len = 2; neg = 1; i = -i; }
  else len = 1;
  /* set len to the # of digits in i */
  for(f=1; f<=i/10; f*=10,len++);
  /* get more space, if necessary */
  TRY_RETURN(Expand(ss, ss->wpos + len));
  /* write the digits */
  if(neg) { ss->data[ss->wpos++] = '-'; len--; }
  for(;len;len--) {
    ss->data[ss->wpos++] = i/f + '0';
    i = i % f;
    f /= 10;
  }
  return TPM_OK;
}

int StrStreamPutDouble(StrStream ss, double d)
{
  /* wimping out here. just allocate a big chunk and use sprintf. */
  TRY_RETURN(Expand(ss, ss->wpos + 32));
  ss->wpos += sprintf(ss->data+ss->wpos, "%.16g", d);
  return TPM_OK;
}

/* Get ***********************************************************************/

int StrStreamGet(StrStream ss, char *buffer, int len)
{
  if(ss->rpos + len >= ss->wpos) {
    ErrorString = TPM_EEMPTY;
    return TPM_ERROR;
  }
  memcpy(buffer, ss->data+ss->rpos, len);
  ss->rpos += len;
  return TPM_OK;
}

int StrStreamGetc(StrStream ss, char *c)
{
  if(ss->rpos == ss->wpos) {
    ErrorString = TPM_EEMPTY;
    return TPM_ERROR;
  }
  *c = ss->data[ss->rpos++];
  return TPM_OK;
}

/* Allocates a new string filled with len bytes starting at data plus
 * a terminator.
 */
static char *StringFromData(char *data, int len)
{
  char *result = (char*)malloc(len+1);
  memcpy(result, data, len);
  result[len] = 0;
  return result;
}

/* Reads up to len bytes (including terminator),
 * stopping at EOF and newline.
 * Returns NULL if no data in ss.
 * Caller must free resulting string.
 */
char *StrStreamGetLine(StrStream ss)
{
  unsigned old_rpos = ss->rpos;
  if(ss->rpos == ss->wpos) {
    ErrorString = TPM_EEMPTY;
    return NULL;
  }
  while(ss->rpos < ss->wpos) {
    if(ss->data[ss->rpos] == '\n') break;
    ss->rpos++;
  }
  return StringFromData(ss->data + old_rpos, ss->rpos - old_rpos);
}

/* Reads up to len bytes (including terminator),
 * stopping at whitespace. Ignores leading whitespace.
 * Returns NULL if no data in ss.
 * Caller must free resulting string.
 */
char *StrStreamGetWord(StrStream ss)
{
  char gotchar = 0;
  unsigned old_rpos = ss->rpos;
  if(ss->rpos == ss->wpos) {
    ErrorString = TPM_EEMPTY;
    return NULL;
  }
  while(ss->rpos < ss->wpos) {
    if(isspace(ss->data[ss->rpos])) { if(gotchar) break; }
    else { gotchar = 1; ss->rpos++; }
  }
  return StringFromData(ss->data + old_rpos, ss->rpos - old_rpos);
}

/* Read a word or all characters enclosed in parentheses, nesting allowed. 
 * Leading whitespace is ignored. 
 * Returns NULL if no data in ss.
 * Caller must free resulting string.
 */
char *StrStreamGetSexp(StrStream ss)
{
  char c, depth = 0, gotchar = 0;
  unsigned old_rpos = ss->rpos;
  if(ss->rpos == ss->wpos) {
    ErrorString = TPM_EEMPTY;
    return NULL;
  }
  while(ss->rpos < ss->wpos) {
    c = ss->data[ss->rpos];
    if(!depth && isspace(c)) { if(gotchar) break; else continue; }
    gotchar = 1; ss->rpos++;
    if(c == '(') depth++;
    else if(c == ')') { if(--depth == 0) break; }
  }
  return StringFromData(ss->data + old_rpos, ss->rpos - old_rpos);
}

/* Reads an integer value into *ip.
 * Leaves ss and *ip unaltered and returns TPM_ERROR if no digits were found.
 */
int StrStreamGetInt(StrStream ss, int *ip)
{
  int neg = 0, i = 0;
  char c;
  /* skip spaces */
  do { TRY_RETURN(StrStreamGetc(ss, &c)); i++; } while(isspace(c));
  /* read sign */
  if(c == '-') { neg = 1; i++; TRY_RETURN(StrStreamGetc(ss, &c)); }
  /* read first digit */
  if(!isdigit(c)) {
    StrStreamUnget(ss, i);
    ErrorString = TPM_EINT;
    return TPM_ERROR;
  }
  i = c - '0';
  if(neg) i = -i;
  /* read remaining digits */
  while(StrStreamGetc(ss, &c) == TPM_OK) {
    if(!isdigit(c)) { /* back up */ ss->rpos--; break; }
    i = i*10 + c - '0';
  }
  *ip = i;
  return TPM_OK;
}

/* Reads a double-precision value into *dp.
 */
int StrStreamGetDouble(StrStream ss, double *dp)
{
  char *s;
  double d = strtod(StrStreamAsString(ss), &s);
  if(s == ss->data+ss->rpos) {
    ErrorString = TPM_EFLOAT;
    return TPM_ERROR;
  }
  ss->rpos = s - ss->data;
  *dp = d;
  return TPM_OK;
}

/* Unget i characters from ss.
 */
int StrStreamUnget(StrStream ss, int i)
{
  if(ss->rpos < i) {
    ErrorString = TPM_EEMPTY;
    return TPM_ERROR;
  }
  ss->rpos -= i;
  return TPM_OK;
}

/* Scan **********************************************************************/

/* Read characters up to whitespace; leading whitespace is ignored */
int StrStreamScanWord(FILE *fp, StrStream ss)
{
  int c, gotchar = 0;
  for(;;) {
    c = getc(fp);
    if(c == EOF) { if(gotchar) break; 
      ErrorString = TPM_EEMPTY; return TPM_ERROR;
    }
    if(isspace(c)) { if(gotchar) break; else continue; }
    gotchar = 1;
    TRY_RETURN(StrStreamPutc(ss, (char)c));
  }
  return TPM_OK;
}

/* Read a word or all characters enclosed in parentheses, nesting allowed. 
 * Leading whitespace is ignored.
 */
int StrStreamScanSexp(FILE *fp, StrStream ss)
{
  int c, depth=0, gotchar = 0, quote = 0;
  for(;;) {
    c = getc(fp);
    if(c == EOF) { if(!depth && !quote) break; 
      ErrorString = TPM_EEMPTY; return TPM_ERROR; 
    }
    if(!depth && !quote && isspace(c)) { if(gotchar) break; else continue; }
    gotchar = 1;
    TRY_RETURN(StrStreamPutc(ss, (char)c));
    if(c == '"') quote = !quote;
    else if(!quote) {
      if(c == '(') depth++;
      else if(c == ')') { 
	if(--depth == 0) {
	  /* eat one character of whitespace */
	  getc(fp); break; 
	}
      }
    }
  }
  return TPM_OK;
}

/* Read all characters until newline or EOF. */
int StrStreamScanLine(FILE *fp, StrStream ss)
{
  int c;
  if(feof(fp)) { ErrorString = TPM_EEMPTY; return TPM_ERROR; }
  for(;;) {
    c = getc(fp);
    if(c == EOF) break;
    if(c == '\n') break;
    TRY_RETURN(StrStreamPutc(ss, (char)c));
  }
  return TPM_OK;
}
