/* Definition of the region package. Add it to Tcl with
 *   Regiontcl_Init(interp);
 * Commands:
 *   create
 *   free r
 *   dims r ?w h?
 *   fromPolygon poly    (poly is a list of coords, as from a canvas)
 *   polygon r
 *   polygons r
 *   area r
 *   contains r x y
 */
#include <tcl.h>
#include "tcl_macros.h"
#include <tpm/tpm.h>
#include <tpm/region.h>

/* Prototypes ****************************************************************/
int Regiontcl_Init(Tcl_Interp *interp);

/* Functions *****************************************************************/

static int GetPolygon(Tcl_Interp *interp, char *s, Tpm_Polygon **poly_return)
{
  XPoint point;
  int i;
  int argc;
  char **argv;
  Tpm_Polygon *poly = Tpm_PolygonCreate();
  Err( ((Tcl_SplitList(interp, s, &argc, &argv) == TCL_ERROR) ||
	(argc % 2)),
       "bad polygon specification");
  MEM_ALLOC_NOTIFY(argv);
  /* add the values in coords to poly */
  for(i=0;i<argc;i++) {
    double f;
    GetDouble(argv[i], f); point.x = f;
    i++;
    GetDouble(argv[i], f); point.y = f;
    Tpm_PolygonAddPoint(poly, &point);
  }
  free(argv);
  *poly_return = poly;
  return TCL_OK;
}

static void ReturnPolygon(Tcl_Interp *interp, Tpm_Polygon *poly)
{
  int i;
  for(i=0;i<poly->npoints;i++) {
    char str[10];
    sprintf(str, "%d", poly->points[i].x);
    Tcl_AppendElement(interp, str);
    sprintf(str, "%d", poly->points[i].y);
    Tcl_AppendElement(interp, str);
  }
}

static void ReturnPolygons(Tcl_Interp *interp, Tpm_Polygon **polys)
{
  int i;
  Tcl_DString ds;

  Tcl_DStringInit(&ds);
  for(; *polys; polys++) {
    Tpm_Polygon *poly = *polys;
    Tcl_DStringStartSublist(&ds);
    for(i=0;i<poly->npoints;i++) {
      char str[10];
      sprintf(str, "%d", poly->points[i].x);
      Tcl_DStringAppendElement(&ds, str);
      sprintf(str, "%d", poly->points[i].y);
      Tcl_DStringAppendElement(&ds, str);
    }
    Tcl_DStringEndSublist(&ds);
  }
  Tcl_DStringResult(interp, &ds);
  Tcl_DStringFree(&ds);
}

static int RegionCmd(ClientData clientData, Tcl_Interp *interp,
		     int argc, char *argv[])
{
  Tpm_Region *r;
  Tpm_Polygon *poly;
  Err(argc < 2, "missing directive; try one of: em_E, em_M, em_V");
  if(!strcmp(argv[1], "create")) {
    Err(argc != 2, "usage: region create");
    ReturnPtr(Tpm_RegionCreate());
  }
  else if(!strcmp(argv[1], "free")) {
    Err(argc != 3, "usage: region free r");
    GetPtr(argv[2], r);
    Tpm_RegionFree(r);
  }
  else if(!strcmp(argv[1], "dims")) {
    GetPtr(argv[2], r);
    if(argc > 3) {
      int width, height;
      Err(argc != 5, "usage: region dims r width height");
      GetInt(argv[3], width);
      GetInt(argv[4], height);
      Tpm_RegionSetDims(r, width, height);
    }
    sprintf(interp->result, "%d %d", r->width, r->height);
  }
  else if(!strcmp(argv[1], "rectangle")) {
    Err(argc < 3, "usage: region rectangle r ?x y width height?");
    GetPtr(argv[2], r);
    if(argc > 3) {
      int x, y, width, height;
      Err(argc != 7, "usage: region rectangle r x y width height");
      GetInt(argv[3], x);
      GetInt(argv[4], y);
      GetInt(argv[5], width);
      GetInt(argv[6], height);
      Tpm_RegionRectangle(r, x, y, width, height);
    }
    sprintf(interp->result, "%d %d %d %d", r->x, r->y, r->width, r->height);
  }
  else if(!strcmp(argv[1], "fromPolygon")) {
    Err(argc != 3, "usage: region fromPolygon coords");
    if(GetPolygon(interp, argv[2], &poly) == TCL_ERROR) return TCL_ERROR;
    r = Tpm_RegionFromPolygon(poly);
    Tpm_PolygonFree(poly);
    ReturnPtr(r);
  }
  else if(!strcmp(argv[1], "polygon")) {
    Err(argc != 3, "usage: region polygon r");
    GetPtr(argv[2], r);
    poly = Tpm_RegionPolygon(r);
    ReturnPolygon(interp, poly);
    Tpm_PolygonFree(poly);
  }
  else if(!strcmp(argv[1], "polygons")) {
    Tpm_Polygon **polys;
    Err(argc != 3, "usage: region polygons r");
    GetPtr(argv[2], r);
    polys = Tpm_RegionToPolygons(r);
    ReturnPolygons(interp, polys);
    Tpm_PolygonsFree(polys);
  }
  else if(!strcmp(argv[1], "area")) {
    Err(argc != 3, "usage: region area r");
    GetPtr(argv[2], r);
    ReturnInt(Tpm_RegionArea(r));
  }
  else if(!strcmp(argv[1], "contains")) {
    int x, y;
    Err(argc != 5, "usage: region contains r x y");
    GetPtr(argv[2], r);
    GetInt(argv[3], x);
    GetInt(argv[4], y);
    ReturnInt(Tpm_RegionContains(r, x, y));
  }
  else Err(1, "unknown directive");
  return TCL_OK;
}

int Regiontcl_Init(Tcl_Interp *interp)
{
  Tcl_CreateCommand(interp, "region", RegionCmd, NULL, NULL);
  return Tcl_PkgProvide(interp, "regionTcl", "1");
}
