/* Miscellaneous routines.
 * Includes random number generators, create/read/write/compare functions
 * for common types, and allocation/deallocation convenience routines.
 */
#include "tpm.h"

/* Prototypes ****************************************************************/
double RandDouble(void);
float RandFloat(void);
int RandInt(int start, int end);
double RandNormal(void);
double ipow(double b, int e);
unsigned long isqrt(unsigned long x);
int gcd(int a, int b);

int *SortingPermutation(void *array, size_t nel, size_t width, 
			CmpFunc *compare, int inverse);
int NumBitsByte(unsigned char byte);
int NumBits(unsigned int x);
Association *AssociationCreate(char *name, void *value);
void AssociationFree(void *a);
int SafeAtoi(char *str, int *value_return);
void *GenericMatrixAlloc(int rows, int cols, int item_size);
void GenericMatrixFree(void *m);
void *GenericMalloc(size_t size);
FreeFunc GenericFree;

/* Random number generators **************************************************/

/* Returns a double in the range [0,1] */
double Rand(void)
{
  return (double)rand()/RAND_MAX;
}

/*
*  Generates a uniformly distributed r.v. between 0 and 1.
*  Kris Popat  6/85
*  Ref: Applied Statistics, 1982 vol 31 no 2 pp 188-190
*  Based on FORTRAN routine by H. Malvar.
*/

static long ix = 101;
static long iy = 1001;
static long iz = 10001;

float RandFloat(void)
{
  static float u;
  
  ix = 171*(ix % 177)-2*(ix/177);
  iy = 172*(iy % 176)-2*(iy/176);
  iz = 170*(iz % 178)-2*(iz/178);
  
  if (ix<0) ix = ix + 30269;
  if (iy<0) iy = iy + 30307;
  if (iz<0) iz = iz + 30323;
  
  u = ((float) ix)/30269.0 +
                ((float) iy)/30307.0 + ((float) iz)/30323.0;
  u -= (float)(int)u;
  return(u);
}

/* Generates an integer r.v. uniformly distributed in [start, end] */
int RandInt(int start, int end)
{
  return (int)(Rand()*(end-start)+start+0.5);
}

/* Generate a normally-distributed random variable using the polar method */
double RandNormal(void)
{
  double theta = 2*M_PI*Rand();
  double radius;
  do {
    radius = sqrt(-2*log(Rand()));
  } while(!finite(radius));
  return radius * cos(theta);
}

/* Integer math functions ****************************************************/

/* Fast power to integer exponent.
 * Taken from Ammeraal, "Programs and Data Structures in C".
 */
double ipow(double b, int e)
{
  int negexp = e < 0;
  double y;
  if(e == 0) return 1.0;
  if(negexp) e = -e;
  while(!(e&1)) { b*=b; e>>=1; }
  y = b;
  while(e >>= 1) {
    b *= b;
    if(e&1) y*=b;
  }
  return negexp ? 1.0/y : y;
}

/* Integer square root.
 * Adapted from Bentley, "More Programming Pearls".
 * Valid for x = 0..2147483648
 * On machines with good FPUs, is slower than (double)sqrt((double)x).
 * On machines without hardware divide (like HP), is really slow.
 */
unsigned long isqrt(unsigned long x)
{
  unsigned long z, nz;
  if(!x) return 0;
  /* Get a good approximation to the root by halving the number of bits */
  nz = x >> 2;
  z = 1;
  while(nz) { nz >>= 2; z <<= 1; }
  /* Now drive to the root with four Newton iterations */
  z = (z + x/z) >> 1;
  z = (z + x/z) >> 1;
  z = (z + x/z) >> 1;
  z = (z + (x+z-1)/z) >> 1;
  return z;
}

/* Greatest positive common factor. E.g. gcd(12, -18) == 6. */
int gcd(int a, int b)
{
  int g;
  if(b == 0) g = a;
  else g = gcd(b, a%b);
  if(g < 0) return -g;
  return g;
}

/* Create/Read/Write/Cmp/Hash functions **************************************/

void *DoubleCreate(double r)
{
  double *p = Allocate(1,double);
  *p = r;
  return (void*)p;
}

void *DoubleRead(FILE *fp)
{
  double *p = Allocate(1,double);
  int status = fscanf(fp, "%lg", p);
  if(!status || (status == EOF)) {
    fprintf(stderr, "doubleRead: Bad floating point number\n");
  }
  return (void*)p;
}

void DoubleWrite(FILE *fp, void *p)
{
  fprintf(fp, "%g", *(double*)p);
}

int DoubleCmp(const void *va, const void *vb)
{
  const double *a = va, *b = vb;
  
  if(*a < *b) return -1;
  else if(*a > *b) return 1;
  else return 0;
}

int FloatCmp(const void *va, const void *vb)
{
  const float *a = va, *b = vb;
  if(*a < *b) return -1;
  else if(*a > *b) return 1;
  else return 0;
}

void *IntCreate(int i)
{
  int *p = Allocate(1,int);
  *p = i;
  return (void*)p;
}

void *IntRead(FILE *fp)
{
  int *p = Allocate(1,int);
  fscanf(fp, "%d", p);
  return (void*)p;
}

void IntWrite(FILE *fp, void *p)
{
  fprintf(fp, "%d", *(int*)p);
}

int IntCmp(const void *va, const void *vb)
{
  const int *a = va, *b = vb;

  if(*a < *b) return -1;
  else if(*a > *b) return 1;
  else return 0;
}

void *IntCopy(void *v)
{
  return IntCreate(*(int*)v);
}

void *StringCopy(void *s)
{
  return strdup((char*)s);
}

int PtrCmp(const void *a, const void *b)
{
  if(a < b) return -1;
  else if(a > b) return 1;
  else return 0;
}

void PtrWrite(FILE *fp, void *p)
{
  fprintf(fp, "%p", *(void**)p);
}

int AllMatchCmp(const void *va, const void *vb)
{
  return 0;
}

/* taken from Tcl7.3 */
int HashString(const void *v)
{
  const char *string = (const char *)v;
  int result = 0;
  char c;
  for(;;) {
    c = *string++;
    if(!c) break;
    result += (result<<3) + c;
  }
  return result;
}
/* hash an arbitrary block of memory */
int HashMem(const char *block, int length)
{
  int i, result = 0;
  for(i=0;i<length;i++) result += (result<<3) + block[i];
  return result;
}
int HashPtr(const void *p)
{
  return HashMem((char*)&p, sizeof(Ptr));
}
int HashInts(const int *array, int length)
{
  int i, result = 0;
  for(i=0;i<length;i++) result += array[i];
  return result * 1103515245;
}
int HashInt(const void *i)
{
  return HashInts(i, 1);
}

/* Misc **********************************************************************/

/* Returns the permutation such that array[p[i]] is sorted. 
 * If inverse = 1, returns the inverse of this permutation,
 * i.e. p[i] is the rank of i in the sorted array.
 */
/* Algorithm: copy the array and pair each element with its index.
 * width must be at least sizeof(int) for proper alignment.
 * Run qsort, then set p to map the new index into the old one.
 */
int *SortingPermutation(void *array, size_t nel, size_t width, 
			CmpFunc *compare, int inverse)
{
  int i;
  int new_width = (width <= sizeof(int)) ? (width + sizeof(int)) : (2*width);
  int *result = Allocate(nel, int);
  char *sorted = (char*)malloc(nel * new_width);
  
  assert(width >= sizeof(int));
  for(i=0;i<nel;i++) {
    memcpy(sorted+i*new_width, (char*)array+i*width, width);
    *(int*)(sorted + i*new_width + width) = i;
  }
  qsort(sorted, nel, new_width, compare);
  if(inverse) {
    int j;
    for(i=0;i<nel;i++) {
      j = *(int*)(sorted + i*new_width + width);
      result[j] = i;
    }
  }
  else {
    for(i=0;i<nel;i++) result[i] = *(int*)(sorted + i*new_width + width);
  }
  free(sorted);
  return result;
}

/* Returns the number of bits set in a byte */
int NumBitsByte(unsigned char byte)
{
  int count;
  for(count = 0; byte; count += byte&1, byte >>= 1);
  return count;
}

int NumBits(unsigned int x)
{
  union {
    unsigned int v;
    unsigned char c[4];
  } u;
  assert(sizeof(int) == 4);
  u.v = x;
  return NumBitsByte(u.c[0]) + NumBitsByte(u.c[1]) + 
         NumBitsByte(u.c[2]) + NumBitsByte(u.c[3]);
}

Association *AssociationCreate(char *name, void *value)
{
  Association *a = Allocate(1, Association);
  a->name = name;
  a->value = value;
  return a;
}

void AssociationFree(void *a)
{
  free(((Association*)a)->name);
  free(a);
}

void *GenericMatrixAlloc(int rows, int cols, int item_size)
{
  char **p;
  int i;

  p = Allocate(rows, char*);
  p[0] = Allocate(rows * cols * item_size, char);
  memset(p[0], 0, rows*cols*item_size);
  for(i=1;i<rows;i++)
    p[i] = &p[0][i*cols*item_size];
  return (void*)p;
}

void GenericMatrixFree(void *m)
{
  free(((void**)m)[0]);
  free(m);
}

/* Set *value_return to atoi(str), returning an error code on failure. */
int SafeAtoi(char *str, int *value_return)
{
  *value_return = atoi(str);
  if(!*value_return && (str[0] != '0')) return TPM_BOTTOM;
  return TPM_OK;
}

void *GenericMalloc(size_t size)
{
  void *p = malloc(size);
  if(!p) {
    fprintf(stderr, "Failed allocating %d bytes\nOut of memory\n", size);
    exit(1);
  }
  return p;
}

void GenericFree(void *p)
{
  free(p);
}

/* Every ANSI compiler should have these functions!! */
#if defined(ultrix)
char *strdup(char *s)
{
  char *str;

  str = malloc(strlen(s)+1);
  strcpy(str, s);
  return str;
}
#endif

#if defined(hpux)
int finite(double x)
{
  return (x > -1.0/0) && (x < 1.0/0);
}
#endif
