/* Implementation of Trie, a rapid access key/value store that assumes
 * dense packing of keys (e.g. a word dictionary).
 * NULL denotes the empty Trie, and may be passed to any Trie function.
 * The key in a TrieEntry points to a dynamically allocated string holding
 * the remainder of the key (i.e. not including the traversed key components).
 */

#include "trie.h"

/* Prototypes ****************************************************************/

void TrieFree(Trie t, FreeFunc *ff);
Trie TrieAdd(Trie *tp, char *key, int *found);
Trie TrieFind(Trie t, char *key);

/* Functions *****************************************************************/

#define NUM_VALS 27
/* Only used internally, since NULL is a valid Trie. */
#define TrieCreate() \
  (Trie)calloc(1, sizeof(struct Trie))

/* Releases the storage associated with t.
 * If ff != NULL, it is used to free the entry values,
 * otherwise they are not freed.
 */
void TrieFree(Trie t, FreeFunc *ff)
{
  int i;
  if(!t) return;
  if(IsTrieEntry(t)) {
    free(TrieKey(t));
    if(ff) ff(TrieValue(t));
  }
  else {
    for(i=0;i<NUM_VALS;i++) TrieFree(t->link[i], ff);
  }
  free(t);
}

static TrieEntry TrieEntryCreate(char *key)
{
  TrieEntry d = Allocate(1, struct TrieEntry);
  d->nonzero = 1;
  d->key = strdup(key);
  return d;
}

/* Modifies *tp to contain an entry for key, returning the node
 * for that entry. 
 * *found is set to 1 if the node already existed, 0 otherwise.
 */
Trie TrieAdd(Trie *tp, char *key, int *found)
{
  Trie t;
  
  if(*key > NUM_VALS) {
    fprintf(stderr, "TrieAdd: key character too large\n");
    return NULL;
  }

  *found = 0;
  if(!*tp) { 
    /* Input is NULL, so create a new Trie */
    *tp = (Trie)TrieEntryCreate(key); 
    return *tp; 
  }
  t = *tp;
  if(!IsTrieEntry(t)) {
    /* We're given a Trie node */
    if(*key) {
      /* Recurse on the next character */
      return TrieAdd(&t->link[*key], key+1, found);
    }
    else {
      return TrieAdd(&t->link[*key], key, found);
    }
  }
  else if(strcmp(key, TrieKey(t))) {
    /* Reached an entry node, but keys do not match.
     * Create a new Trie node with these 
     * two keys under it.
     */
    *tp = TrieCreate();
    /* Insert old key/value */
    TrieValue(TrieAdd(&(*tp)->link[*TrieKey(t)], TrieKey(t)+1, found)) 
      = TrieValue(t);
    free(TrieKey(t));
    free(t);
    /* Insert new key */
    if(*key) {
      return TrieAdd(&(*tp)->link[*key], key+1, found);
    }
    else {
      return TrieAdd(&(*tp)->link[*key], key, found);
    }
  }
  else *found = 1; /* matched an existing entry */
  return *tp;
}

Trie TrieFind(Trie t, char *key)
{
  if(!t) return NULL;
  if(IsTrieEntry(t)) {
    if(strcmp(TrieKey(t), key)) return NULL;
    return t;
  }
  if(!*key) return t->link[0];
  if(*key > NUM_VALS) {
    fprintf(stderr, "TrieFind: key character too large\n");
    return NULL;
  }
  return TrieFind(t->link[*key], key+1);
}
