/* Definition of the Jpeg photo format.
 * Use it for reading:
 *   image create photo -file 001.jpeg
 * Or writing:
 *   image1 write data.jpg -format jpeg
 */

#include <stdio.h>
#include <stdlib.h> /* for malloc */
#include <setjmp.h>
#include <jpeglib.h>
/* tk.h must come after jpeglib.h */
#include <tk.h>

typedef struct {
  int width, height, channels;
  FILE *fp;
  struct jpeg_decompress_struct *cinfo;
} Jpeg;

static void JpegClose(Jpeg *jpeg)
{
  if(jpeg->fp) fclose(jpeg->fp);
  jpeg_destroy_decompress(jpeg->cinfo);
  free(jpeg->cinfo->err);
  free(jpeg->cinfo);
}

/*
 * ERROR HANDLING:
 *
 * The JPEG library's standard error handler (jerror.c) is divided into
 * several "methods" which you can override individually.  This lets you
 * adjust the behavior without duplicating a lot of code, which you might
 * have to update with each future release.
 *
 * Our example here shows how to override the "error_exit" method so that
 * control is returned to the library's caller when a fatal error occurs,
 * rather than calling exit() as the standard error_exit method does.
 *
 * We use C's setjmp/longjmp facility to return control.  This means that the
 * routine which calls the JPEG library must first execute a setjmp() call to
 * establish the return point.  We want the replacement error_exit to do a
 * longjmp().  But we need to make the setjmp buffer accessible to the
 * error_exit routine.  To do this, we make a private extension of the
 * standard JPEG error handler object.  (If we were using C++, we'd say we
 * were making a subclass of the regular error handler.)
 *
 * Here's the extended error handler struct:
 */

typedef struct my_error_mgr {
  struct jpeg_error_mgr pub;	/* "public" fields */

  jmp_buf setjmp_buffer;	/* for return to caller */
} *my_error_ptr;

/*
 * Here's the routine that will replace the standard error_exit method:
 */
static void my_error_exit (j_common_ptr cinfo)
{
  /* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
  my_error_ptr myerr = (my_error_ptr) cinfo->err;

#if 0
  /* Display the message. */
  (*cinfo->err->output_message) (cinfo);
#endif

  /* Return control to the setjmp point */
  longjmp(myerr->setjmp_buffer, 1);
}

/* This is verbatim from ~/src/image/jpeg.c */
static int JpegReadInit(Jpeg *jpeg)
{
  /* This struct contains the JPEG decompression parameters and pointers to
   * working space (which is allocated as needed by the JPEG library).
   */
  struct jpeg_decompress_struct *cinfo;
  /* We use our private extension JPEG error handler.
   * Note that this struct must live as long as the main JPEG parameter
   * struct, to avoid dangling-pointer problems.
   */
  my_error_ptr jerr;

#if 0
  /* Check the header before calling the library */
  if(FileCheckHeader(jpeg->fp, "\377\330\377\340\000\020JFIF\000\001", 12))
    return TPM_ERROR;
#endif

  cinfo = (struct jpeg_decompress_struct*)malloc(sizeof(struct jpeg_decompress_struct));
  jerr = (my_error_ptr)malloc(sizeof(struct my_error_mgr));

  /* Step 1: allocate and initialize JPEG decompression object */
  /* We set up the normal JPEG error routines, then override error_exit. */
  cinfo->err = jpeg_std_error(&jerr->pub);
  jerr->pub.error_exit = my_error_exit;
  /* Establish the setjmp return context for my_error_exit to use. */
  if (setjmp(jerr->setjmp_buffer)) {
    /* If we get here, the JPEG code has signaled an error.
     * We need to clean up the JPEG object and return.
     */
    return TCL_ERROR;
  }
  /* Now we can initialize the JPEG decompression object. */
  jpeg_create_decompress(cinfo);

  /* Step 2: specify data source (eg, a file) */
  jpeg_stdio_src(cinfo, jpeg->fp);

  /* Step 3: read file parameters with jpeg_read_header() */

  (void) jpeg_read_header(cinfo, TRUE);
  /* We can ignore the return value from jpeg_read_header since
   *   (a) suspension is not possible with the stdio data source, and
   *   (b) we passed TRUE to reject a tables-only JPEG file as an error.
   * See libjpeg.doc for more info.
   */

  jpeg->cinfo = cinfo;
  jpeg->width = jpeg->cinfo->image_width;
  jpeg->height = jpeg->cinfo->image_height;
  jpeg->channels = jpeg->cinfo->num_components;
  return TCL_OK;
}

/* FileMatch is used to determine which format an image file is in.
   The fileName argument is the name of the file containing the image data,
   which is open for reading as f.  The formatString argument contains the
   value given for the -format option, or NULL if the option was not
   specified.  If the data in the file appears to be in the format
   supported by this handler, the FileMatch procedure should store the
   width and height of the image in *widthPtr and *heightPtr
   respectively, and return 1.  Otherwise it should return 0.
 */
static int FileMatch(FILE *fp, char *fileName, char *formatString, 
		     int *widthPtr, int *heightPtr)
{
  Jpeg jpeg;
  jpeg.fp = fp;
  if(JpegReadInit(&jpeg)) return 0;
  *widthPtr = jpeg.width;
  *heightPtr = jpeg.height;
  jpeg.fp = NULL;
  JpegClose(&jpeg);
  return 1;
}

/* FileRead is used to read data from an image file into a photo
   image.  The interp argument is the interpreter in which the command
   was invoked to read the image; it should be used for reporting errors.
   The image data is in the file named fileName, which is open for
   reading as f.  The formatString argument contains the value given for
   the -format option, or NULL if the option was not specified.  The
   image data in the file, or a subimage of it, is to be read into the
   photo image identified by the handle imageHandle.  The subimage of the
   data in the file is of dimensions width x height and has its top-left
   corner at coordinates (srcX,srcY).  It is to be stored in the photo
   image with its top-left corner at coordinates (destX,destY) using the
   Tk_PhotoPutBlock procedure.  The return value is a standard Tcl return
   value.
 */
static int FileRead(Tcl_Interp *interp,
		    FILE *fp, char *fileName, char *formatString,
		    Tk_PhotoHandle imageHandle,
		    int destX, int destY,
		    int width, int height,
		    int srcX, int srcY)
{
  struct jpeg_decompress_struct *cinfo;
  int row_stride;
  Tk_PhotoImageBlock block;
  Jpeg jpeg;

  jpeg.fp = fp;
  if(JpegReadInit(&jpeg)) return TCL_ERROR;
  cinfo = jpeg.cinfo;

  block.width = jpeg.width;
  block.height = jpeg.height;
  block.pixelSize = jpeg.channels;
  block.pitch = jpeg.width * jpeg.channels;
  block.offset[0] = 0;
  block.offset[1] = 1;
  block.offset[2] = 2;
  block.pixelPtr = (unsigned char*)malloc(jpeg.height*block.pitch);

  Tk_PhotoExpand(imageHandle, destX + width, destY + height);

  /* Step 4: set parameters for decompression */

  /* In this example, we don't need to change any of the defaults set by
   * jpeg_read_header(), so we do nothing here.
   */

  /* Step 5: Start decompressor */

  (void) jpeg_start_decompress(cinfo);
  /* We can ignore the return value since suspension is not possible
   * with the stdio data source.
   */

  /* We may need to do some setup of our own at this point before reading
   * the data.  After jpeg_start_decompress() we have the correct scaled
   * output image dimensions available, as well as the output colormap
   * if we asked for color quantization.
   * In this example, we need to make an output work buffer of the right size.
   */ 
  /* JSAMPLEs per row in output buffer */
  row_stride = cinfo->output_width * cinfo->output_components;

  /* Step 6: while (scan lines remain to be read) */
  /*           jpeg_read_scanlines(...); */

  /* Here we use the library's state variable cinfo.output_scanline as the
   * loop counter, so that we don't have to keep track ourselves.
   */
  while (cinfo->output_scanline < cinfo->output_height) {
    /* jpeg_read_scanlines expects an array of pointers to scanlines.
     * Here the array is only one element long, but you could ask for
     * more than one scanline at a time if that's more convenient.
     */
    unsigned char *buffer = &block.pixelPtr[cinfo->output_scanline * block.pitch];
    (void) jpeg_read_scanlines(cinfo, &buffer, 1);
  }

  /* Step 7: Finish decompression */

  (void) jpeg_finish_decompress(cinfo);
  /* We can ignore the return value since suspension is not possible
   * with the stdio data source.
   */

  jpeg.fp = NULL;
  JpegClose(&jpeg);
  Tk_PhotoPutBlock(imageHandle, &block, destX, destY, 
		   block.width, block.height);
  free(block.pixelPtr);
  return TCL_OK;
}

/* FileWrite is used to write data from a photo image to a file.  The
   interp argument is the interpreter in which the command was invoked to
   write the image; it should be used for reporting errors.  The image
   data to be written are in memory and are described by the
   Tk_PhotoImageBlock structure pointed to by block; see the manual page
   FindPhoto(3) for details.  The fileName argument points to the string
   giving the name of the file in which to write the image data.  The
   formatString argument contains the value given for the -format option,
   or NULL if the option was not specified.  The format string can
   contain extra characters after the name of the format.  If
   appropriate, the FileWrite procedure may interpret these characters to
   specify further details about the image file.  The return value is a
   standard Tcl return value.  
 */
static int FileWrite(Tcl_Interp *interp, char *fileName, char *formatString,
		     Tk_PhotoImageBlock *block)
{
  return TCL_ERROR;
}

/* exported */
Tk_PhotoImageFormat FormatJpeg = {
  "jpeg", 
  FileMatch, NULL, 
  FileRead, NULL, 
  FileWrite, NULL
};
