/* Definition of the Raw photo format.
 * Use it for reading:
 *   image create photo -format "raw 128 128 3" -file 001.raw
 * Or writing:
 *   image1 write data.out -format "raw 256 256"
 * The format syntax is
 *   raw <width> <height> [channels]
 * where channels defaults to 3.
 */

#include <tk.h>
#include <stdlib.h> /* for malloc */
#include <stdio.h>
#ifndef SEEK_CUR
#define SEEK_CUR 2
#endif

/* used if channels is not specified in the format string */
#define DEFAULT_CHANNELS 3

/* Maximum amount of memory to allocate for read buffer. */
#define MAX_MEMORY (128*1024)

/* FileMatch is used to determine which format an image file is in.
   The fileName argument is the name of the file containing the image data,
   which is open for reading as f.  The formatString argument contains the
   value given for the -format option, or NULL if the option was not
   specified.  If the data in the file appears to be in the format
   supported by this handler, the FileMatch procedure should store the
   width and height of the image in *widthPtr and *heightPtr
   respectively, and return 1.  Otherwise it should return 0.
 */
static int FileMatch(FILE *fp, char *fileName, char *formatString, 
		     int *widthPtr, int *heightPtr)
{
  /* get width and height from the format, otherwise return an error */
  if(!formatString || 
     sscanf(formatString, "raw %d %d", widthPtr, heightPtr) < 2) return 0;
  return 1;
}

/* FileRead is used to read data from an image file into a photo
   image.  The interp argument is the interpreter in which the command
   was invoked to read the image; it should be used for reporting errors.
   The image data is in the file named fileName, which is open for
   reading as f.  The formatString argument contains the value given for
   the -format option, or NULL if the option was not specified.  The
   image data in the file, or a subimage of it, is to be read into the
   photo image identified by the handle imageHandle.  The subimage of the
   data in the file is of dimensions width x height and has its top-left
   corner at coordinates (srcX,srcY).  It is to be stored in the photo
   image with its top-left corner at coordinates (destX,destY) using the
   Tk_PhotoPutBlock procedure.  The return value is a standard Tcl return
   value.
 */
static int FileRead(Tcl_Interp *interp,
		    FILE *fp, char *fileName, char *formatString,
		    Tk_PhotoHandle imageHandle,
		    int destX, int destY,
		    int width, int height,
		    int srcX, int srcY)
{
  int h, total_width, total_height, channels;
  int nLines, nBytes;
  Tk_PhotoImageBlock block;
  unsigned char *pixelPtr;

  if(sscanf(formatString, "raw %d %d %d", 
	    &total_width, &total_height, &channels) < 3)
    channels = DEFAULT_CHANNELS;

  /* read in sets of scanlines */
  block.width = width;
  block.pixelSize = channels;
  block.pitch = total_width * channels;
  block.offset[0] = 0;
  block.offset[1] = 1;
  block.offset[2] = 2;

  Tk_PhotoExpand(imageHandle, destX + width, destY + height);

  /* advance to the (0, srcY) position */
  fseek(fp, srcY*block.pitch, SEEK_CUR);

  /* determine how many scanlines to read at once */
  nLines = (MAX_MEMORY + block.pitch - 1) / block.pitch;
  if(nLines > height) nLines = height;
  else if(nLines <= 0) nLines = 1;

  /* allocate a scan line buffer */
  nBytes = nLines * block.pitch;
  pixelPtr = (unsigned char *)malloc(nBytes);
  /* address the block.pixelPtr to the srcX position in the scan line buffer */
  block.pixelPtr = pixelPtr + srcX * channels;

  for(h = height; h > 0; h -= nLines) {
    if(nLines > h) {
      nLines = h;
      nBytes = nLines * block.pitch;
    }
    /* read nLines scanlines (even if we don't want all of it) */
    if(fread(pixelPtr, 1, nBytes, fp) != nBytes) goto fail;

    /* write the part we want to the image */
    block.height = nLines;
    Tk_PhotoPutBlock(imageHandle, &block, destX, destY, 
		     block.width, block.height);
    destY += nLines;
  }
  free(pixelPtr);
  return TCL_OK;

 fail:
  Tcl_AppendResult(interp, "error reading image file \"", fileName, "\": ", 
		   feof(fp) ? "not enough data" : Tcl_PosixError(interp),
		   NULL);
  free(pixelPtr);
  return TCL_ERROR;
}

/* FileWrite is used to write data from a photo image to a file.  The
   interp argument is the interpreter in which the command was invoked to
   write the image; it should be used for reporting errors.  The image
   data to be written are in memory and are described by the
   Tk_PhotoImageBlock structure pointed to by block; see the manual page
   FindPhoto(3) for details.  The fileName argument points to the string
   giving the name of the file in which to write the image data.  The
   formatString argument contains the value given for the -format option,
   or NULL if the option was not specified.  The format string can
   contain extra characters after the name of the format.  If
   appropriate, the FileWrite procedure may interpret these characters to
   specify further details about the image file.  The return value is a
   standard Tcl return value.  
 */
static int FileWrite(Tcl_Interp *interp, char *fileName, char *formatString,
		     Tk_PhotoImageBlock *block)
{
  int channels, nBytes;
  int tclfile = 0;
  Tcl_Channel chan;
  
  /* see if there is a Tcl channel open already */
  chan = Tcl_GetChannel(interp, fileName, NULL);
  if(chan) tclfile = 1;
  else {
    Tcl_ResetResult(interp);  /* clear the GetChannel error */
    chan = Tcl_OpenFileChannel(interp, fileName, "w", 0666);
    if(!chan) return TCL_ERROR;
  }
  if(sscanf(formatString, "raw %*d %*d %d", &channels) < 1)
    channels = DEFAULT_CHANNELS;
  if(channels == block->pixelSize) {
    if(block->pitch == block->width*channels) {
      /* simple case; no padding */
      nBytes = block->pitch * block->height;
      if(Tcl_Write(chan, (char*)block->pixelPtr, nBytes) != nBytes) goto fail;
    }
    else {
      unsigned char *bp = block->pixelPtr;
      int i;
      nBytes = block->width*channels;
      /* write one scan line at a time */
      for(i=0;i<block->height;i++) {
	if(Tcl_Write(chan, (char*)bp, nBytes) != nBytes) goto fail;
	bp += block->pitch;
      }
    }
  }
  else if(channels == 1) {
    /* write one scan line at a time */
    unsigned char *buffer, *bp = block->pixelPtr;
    int i, j;
    nBytes = block->width * channels;
    buffer = (unsigned char*)malloc(nBytes);
    for(i=0;i<block->height;i++) {
      for(j=0;j<block->width;j++) {
	/* NTSC gray component */
	buffer[j] = (unsigned char)
	  (0.30 * *bp + 0.59 * *(bp+1) + 0.11 * *(bp+2));
	bp += 3;
      }
      if(Tcl_Write(chan, (char*)buffer, nBytes) != nBytes) goto fail;
      bp += block->pitch - block->width * 3;
    }
    free(buffer);
  }
  else {
    Tcl_SetResult(interp, "cannot do that channel conversion", TCL_STATIC);
    return TCL_ERROR;
  }

  if(tclfile || !Tcl_Close(interp, chan)) return TCL_OK;
  chan = NULL;
  /* fall through */

 fail:
  Tcl_AppendResult(interp, "error writing \"", fileName, "\": ",
		   Tcl_PosixError(interp), NULL);
  if(!tclfile && chan) Tcl_Close(interp, chan);
  return TCL_ERROR;
}

/* exported */
Tk_PhotoImageFormat FormatRaw = {
  "raw", 
  FileMatch, NULL, 
  FileRead, NULL, 
  FileWrite, NULL
};
