/* Definition of the Sgi photo format.
 * Use it for reading:
 *   image create photo -file 001.sgi
 * Or writing:
 *   image1 write data.sgi -format sgi
 */

#include <tk.h>
#include <stdlib.h> /* for malloc */
#include <gl/image.h>

#define DEBUG 0

static void ErrorProc(char *string)
{
#if DEBUG
  fprintf(stderr, "%s\n", string);
#endif
}

/* FileMatch is used to determine which format an image file is in.
   The fileName argument is the name of the file containing the image data,
   which is open for reading as f.  The formatString argument contains the
   value given for the -format option, or NULL if the option was not
   specified.  If the data in the file appears to be in the format
   supported by this handler, the FileMatch procedure should store the
   width and height of the image in *widthPtr and *heightPtr
   respectively, and return 1.  Otherwise it should return 0.
 */
static int FileMatch(FILE *fp, char *fileName, char *formatString, 
		     int *widthPtr, int *heightPtr)
{
  IMAGE *img;
  i_seterror(ErrorProc);
  img = iopen(fileName, "r");
  if(!img) return 0;
#if DEBUG
  fprintf(stderr, "match: %d %d\n", img->xsize, img->ysize);
#endif
  *widthPtr = img->xsize;
  *heightPtr = img->ysize;
  iclose(img);
  return 1;
}

/* FileRead is used to read data from an image file into a photo
   image.  The interp argument is the interpreter in which the command
   was invoked to read the image; it should be used for reporting errors.
   The image data is in the file named fileName, which is open for
   reading as f.  The formatString argument contains the value given for
   the -format option, or NULL if the option was not specified.  The
   image data in the file, or a subimage of it, is to be read into the
   photo image identified by the handle imageHandle.  The subimage of the
   data in the file is of dimensions width x height and has its top-left
   corner at coordinates (srcX,srcY).  It is to be stored in the photo
   image with its top-left corner at coordinates (destX,destY) using the
   Tk_PhotoPutBlock procedure.  The return value is a standard Tcl return
   value.
 */
static int FileRead(Tcl_Interp *interp,
		    FILE *fp, char *fileName, char *formatString,
		    Tk_PhotoHandle imageHandle,
		    int destX, int destY,
		    int width, int height,
		    int srcX, int srcY)
{
  int z, y, x;
  unsigned short *buffer;
  Tk_PhotoImageBlock block;
  IMAGE *img;
  i_seterror(ErrorProc);
  img = iopen(fileName, "r");
  if(!img) return TCL_ERROR;
  
  /* read one scanline at a time */
  block.height = 1;
  block.width = width;
  block.pixelSize = 3;
  block.pitch = img->xsize * 3;
  block.offset[0] = 0;
  block.offset[1] = 1;
  block.offset[2] = 2;
  block.pixelPtr = (unsigned char *)malloc(block.pitch);

  Tk_PhotoExpand(imageHandle, destX + width, destY + height);

  buffer = (unsigned short*)malloc(img->xsize * sizeof(unsigned short));

  /* Loop scanlines */
  for(y=0;y<height;y++) {
    /* Loop color planes */
    for(z=0;z<img->zsize;z++) {
      /* read one row (even if we don't want all of it) */
      /* SGI images are stored upside-down so we flip y */
      getrow(img, buffer, img->ysize - 1 - (y+srcY), z);
      /* getrow gives us an array of shorts which we must convert into chars */
      for(x=0;x<img->xsize;x++) {
	block.pixelPtr[x*3 + z] = buffer[x];
      }
    }
    Tk_PhotoPutBlock(imageHandle, &block, destX, destY, 
		     block.width, block.height);
    destY += block.height;
  }
  iclose(img);
  free(buffer);
  free(block.pixelPtr);
  return TCL_OK;
}

/* FileWrite is used to write data from a photo image to a file.  The
   interp argument is the interpreter in which the command was invoked to
   write the image; it should be used for reporting errors.  The image
   data to be written are in memory and are described by the
   Tk_PhotoImageBlock structure pointed to by block; see the manual page
   FindPhoto(3) for details.  The fileName argument points to the string
   giving the name of the file in which to write the image data.  The
   formatString argument contains the value given for the -format option,
   or NULL if the option was not specified.  The format string can
   contain extra characters after the name of the format.  If
   appropriate, the FileWrite procedure may interpret these characters to
   specify further details about the image file.  The return value is a
   standard Tcl return value.  
 */
static int FileWrite(Tcl_Interp *interp, char *fileName, char *formatString,
		     Tk_PhotoImageBlock *block)
{
  return TCL_ERROR;
}

/* exported */
Tk_PhotoImageFormat FormatSgi = {
  "sgi", 
  FileMatch, NULL, 
  FileRead, NULL, 
  FileWrite, NULL
};
