#include "tk.h"

/* Prototypes ****************************************************************/

Pixmap PhotoToPixmap(Tk_Window tkwin, GC gc, Tk_PhotoHandle image);
XImage *ImageBlockToXImage(Tk_Window tkwin, Tk_PhotoImageBlock *block);

/* private */
static void XImageFill(XImage *ximage, unsigned char *dptr, int img_channels);

/* Functions *****************************************************************/

Pixmap PhotoToPixmap(Tk_Window tkwin, GC gc, Tk_PhotoHandle image)
{
  Pixmap pixmap;
  Tk_PhotoImageBlock block;
  XImage *ximage;

  Tk_PhotoGetImage(image, &block);

  /* convert the block into an ximage */
  ximage = ImageBlockToXImage(tkwin, &block);

  /* copy onto a pixmap */
  Tk_MakeWindowExist(tkwin);
  pixmap = Tk_GetPixmap(Tk_Display(tkwin), Tk_WindowId(tkwin), 
			block.width, block.height, Tk_Depth(tkwin));
  if(!pixmap) return 0;
  if(!gc) return 0;
  XPutImage(Tk_Display(tkwin), pixmap, gc, ximage,
	    0, 0, 0, 0, block.width, block.height);
  XDestroyImage(ximage);

  return pixmap;
}

XImage *ImageBlockToXImage(Tk_Window tkwin, Tk_PhotoImageBlock *block)
{
  XImage *ximage;
  int depth;

  depth = Tk_Depth(tkwin);
  ximage = XCreateImage(Tk_Display(tkwin), Tk_Visual(tkwin), depth, 
			ZPixmap, 0, NULL,
			block->width, block->height, (depth>8)?32:depth, 0);
  ximage->data = (char*)malloc(ximage->bytes_per_line * block->height);
  XImageFill(ximage, block->pixelPtr, block->pixelSize);
  return ximage;
}

/* Fill an XImage with multi-byte data */
/* If the pixel size is greater than 'channels', 
 *   the byte values will be replicated across multiple bytes.
 * If the pixel size is less than 'channels',
 *   the low order bits (or bytes) of the data will be used.
 * Should work regardless of byte ordering (MSBfirst or LSBfirst).
 */
#define LSB3to4_1(r,g,b) \
  ((r) + ((g) << 8) + ((b) << 16))
#define LSB3to4_2(r,g,b) \
  (LSB3to4_1(r,g,b) << 8)
#define MSB3to4_1(r,g,b) \
  (((r) << 16) + ((g) << 8) + (b))
#define MSB3to4_2(r,g,b) \
  (MSB3to4_1(r,g,b) << 8)
static void XImageFill(XImage *ximage, unsigned char *dptr, int img_channels)
{
  int xi_channels, i, j, image_size;
  unsigned char *xdptr;
  int pad; /* extra bytes at end of each ximage row */

  xdptr = (unsigned char *)ximage->data;
  xi_channels = ximage->bytes_per_line / ximage->width;
  pad = ximage->bytes_per_line/xi_channels % ximage->width;

#if DEBUG
  if(debug)
    printf("%d -> %d;pad = %d\n", img_channels, xi_channels, pad);
#endif

  image_size = ximage->width * ximage->height;
  if(xi_channels > 0) {
    if(xi_channels == img_channels) {
      if(pad) {
        for(i=0;i<ximage->height;i++) {
          memcpy(xdptr, dptr, ximage->width*img_channels);
          xdptr += ximage->width*img_channels + pad;
          dptr += ximage->width*img_channels;
        }
      }
      else 
        memcpy(xdptr, dptr, image_size*img_channels);
    }
    else if(xi_channels == 4) {
      if(img_channels == 3) {
#if FAST_FILL
	/* faster, but not as robust */
        if((ximage->byte_order == MSBFirst) &&
           (ximage->red_mask == 0xff0000)) {
          for(i=0;i<image_size;i++,xdptr+=4,dptr+=3) {
            *(int*)xdptr = MSB3to4_1(*dptr, *(dptr+1), *(dptr+2));
          }
        }
        else {
          for(i=0;i<image_size;i++,xdptr+=4,dptr+=3) {
            *(int*)xdptr = LSB3to4_1(*dptr, *(dptr+1), *(dptr+2));
          }
        }
#else
        if((ximage->byte_order == MSBFirst) &&
           (ximage->red_mask == 0xff0000)) {
          for(i=0;i<ximage->height;i++) {
            for(j=0;j<ximage->width;j++) {
              XPutPixel(ximage, j, i, MSB3to4_1(*dptr, *(dptr+1), *(dptr+2)
));
              dptr+=3;
            }
          }
	}
        else {
          for(i=0;i<ximage->height;i++) {
            for(j=0;j<ximage->width;j++) {
              XPutPixel(ximage, j, i, LSB3to4_1(*dptr, *(dptr+1), *(dptr+2)
));
              dptr+=3;
            }
          }
	}
#endif
      }
      else if(img_channels == 1) {
        /* put same byte in every channel */
        for(i=0;i<image_size;i++,xdptr+=4,dptr++) {
          *(int*)xdptr = MSB3to4_2(*dptr, *dptr, *dptr) + *dptr;
        }
      }
    }
    else if(xi_channels == 1) {
      /* Compute NTSC gray = 0.30 red, 0.59 green, 0.11 blue */
      for(i=0;i<ximage->height;i++) {
        for(j=0;j<ximage->width;j++) {
          *xdptr++ = (unsigned char)
            (0.30 * *dptr + 0.59 * *(dptr+1) + 0.11 * *(dptr+2));
          dptr += 3;
        }
        xdptr += pad;
      }
    }
    else {
      fprintf(stderr, "Unexpected xi_channels: %d\n", xi_channels);
    }
  }
  else {
    /* Assume two pixels per byte */
    for(i=0;i<image_size;i+=2,dptr+=2*img_channels) {
      *xdptr++ = (*dptr << 4) | *(dptr+img_channels);
    }
  }
}
