/* Miscellaneous commands. Add to Tcl with Utiltcl_Init(interp).
 * sort          Sort multiple lists according to one key list.
 * ran           Return a uniformly distributed r.v. in [0,1].
 * makesafe      Make the current interpreter a safe one.
 */

#include <tcl.h>
#include <tcl_macros.h>
#include <tpm/tpm.h>

/* Prototypes ****************************************************************/
int Utiltcl_Init(Tcl_Interp *interp);

/* private */
static Tcl_CmdProc SortCmd, MakeSafeCmd;

/* Functions *****************************************************************/
static int SCompare(char **a, char **b)
{
  return strcmp(*a, *b);
}

static int DCompare(double *a, double *b)
{
  if(*a > *b) return 1;
  else if(*a < *b) return -1;
  else return 0;
}
/* Sort value lists according to a key list */
static int SortCmd(ClientData clientData, Tcl_Interp *interp,
		   int argc, char *argv[])
{
#define pad (numeric*sizeof(double)/sizeof(char*))
  int rows, cols, i, j, width;
  char **array, **elts;
  int arg, numeric = 0;

  Err(argc < 2, "usage: sort ?-real? keys ?var? ?var? ...");
  arg = 1;
  if(!strcmp(argv[arg], "-real")) {
    Err(argc < 3, "usage: sort ?-real? keys ?var? ?var? ...");
    numeric = 1;
    arg++;
  }
  /* collect all lists into an array of structures */
  array = NULL;
  cols = argc - arg;
  rows = 0;
  for(i=0;i<cols;i++) {
    char *list = Tcl_GetVar(interp, argv[i+arg], 0);
    int argc;
    Err(!list, "arguments to sort should be variable names");
    if(Tcl_SplitList(interp, list, &argc, &elts) != TCL_OK) return TCL_ERROR;
    MEM_ALLOC_NOTIFY(elts);
    if(!rows) {
      rows = argc;
      width = cols+pad;
      /* make sure each row is aligned for a double */
      if(width*sizeof(char*) % sizeof(double)) 
	width += (sizeof(double) - 
	  (width*sizeof(char*) % sizeof(double))) / sizeof(char*);
      array = (char**)malloc(rows*width*sizeof(char*));
      if(numeric) {
	for(j=0;j<rows;j++) {
	  GetDouble(elts[j], array[j*width]);
	}
      }
    }
    else Err(argc < rows, "not enough values for keys");
    for(j=0;j<rows;j++) {
      array[j*width+i+pad] = strdup(elts[j]);
    }
    free(elts);
  }

  qsort(array, rows, width*sizeof(char*), 
	numeric?(CmpFunc*)DCompare:(CmpFunc*)SCompare);

  /* unpack the array into the tcl variables */
  elts = (char**)malloc(rows*sizeof(char*));
  for(i=0;i<cols;i++) {
    char *list;
    for(j=0;j<rows;j++) elts[j] = array[j*width+i+pad];
    list = Tcl_Merge(rows, elts);
    MEM_ALLOC_NOTIFY(list);
    for(j=0;j<rows;j++) free(elts[j]);
    Tcl_SetVar(interp, argv[i+arg], list, 0);
    free(list);
  }
  free(elts);
  free(array);

  return TCL_OK;
}

/*
*  Generates a uniformly distributed r.v. between 0 and 1.
*  Kris Popat  6/85
*  Ref: Applied Statistics, 1982 vol 31 no 2 pp 188-190
*  Based on FORTRAN routine by H. Malvar.
*/

static long ix = 101;
static long iy = 1001;
static long iz = 10001;

float ran(void)
{
  static float u;
  
  ix = 171*(ix % 177)-2*(ix/177);
  iy = 172*(iy % 176)-2*(iy/176);
  iz = 170*(iz % 178)-2*(iz/178);
  
  if (ix<0) ix = ix + 30269;
  if (iy<0) iy = iy + 30307;
  if (iz<0) iz = iz + 30323;
  
  u = ((float) ix)/30269.0 +
                ((float) iy)/30307.0 + ((float) iz)/30323.0;
  u -= (float)(int)u;
  return(u);
}

static int RanCmd(ClientData clientData, Tcl_Interp *interp,
		  int argc, char *argv[])
{
  if(argc > 1) {
    int seed;
    GetInt(argv[1], seed);
    srand(seed);
  }
  ReturnDouble(ran());
  return TCL_OK;
}

static int MakeSafeCmd(ClientData clientData, Tcl_Interp *interp,
		       int argc, char *argv[])
{
  return Tcl_MakeSafe(interp);
}

static int DerefCmd(ClientData clientData, Tcl_Interp *interp,
		    int argc, char *argv[])
{
  void **p;
  Err(argc != 2, "usage: deref ptr");
  GetPtr(argv[1], p);
  ReturnPtr(*p);
  return TCL_OK;
}

static int MemBlocksCmd(ClientData clientData, Tcl_Interp *interp,
			int argc, char *argv[])
{
  MEM_BLOCKS();
  MEM_STATUS();
  return TCL_OK;
}

static int ReadIntCmd(ClientData clientData, Tcl_Interp *interp,
			int argc, char *argv[])
{
  int i;
  Tcl_Channel chan;
  Err(argc != 2, "usage: readInt channel");
  chan = Tcl_GetChannel(interp, argv[1], NULL);
  Err(!chan, "invalid channel number");
  if(Tcl_Read(chan, (char*)&i, sizeof(int)) != sizeof(int)) {
    Err(1, "couldn't read 4 bytes");
  }
  ReturnInt(i);
  return TCL_OK;
}

static int WriteIntCmd(ClientData clientData, Tcl_Interp *interp,
		       int argc, char *argv[])
{
  int i;
  Tcl_Channel chan;
  Err(argc != 3, "usage: writeInt channel int");
  chan = Tcl_GetChannel(interp, argv[1], NULL);
  Err(!chan, "invalid channel number");
  GetInt(argv[2], i);
  if(Tcl_Write(chan, (char*)&i, sizeof(int)) != sizeof(int)) {
    Err(1, "couldn't write 4 bytes");
  }
  return TCL_OK;
}

static int ReadFloatCmd(ClientData clientData, Tcl_Interp *interp,
			int argc, char *argv[])
{
  float i;
  Tcl_Channel chan;
  Err(argc != 2, "usage: readFloat channel");
  chan = Tcl_GetChannel(interp, argv[1], NULL);
  Err(!chan, "invalid channel number");
  if(Tcl_Read(chan, (char*)&i, sizeof(float)) != sizeof(float)) {
    Err(1, "couldn't read 4 bytes");
  }
  ReturnFloat(i);
  return TCL_OK;
}

/* Read a block of data from one channel and write it to another. */
static int ReadWriteCmd(ClientData clientData, Tcl_Interp *interp,
			int argc, char *argv[])
{
  Tcl_Channel chan_r, chan_w;
  int bufsize;
  char *buffer;
  Err(argc != 3, "usage: readWrite channel_read channel_write");
  chan_r = Tcl_GetChannel(interp, argv[1], NULL);
  Err(!chan_r, "invalid channel number");
  chan_w = Tcl_GetChannel(interp, argv[2], NULL);
  Err(!chan_w, "invalid channel number");

  bufsize = Tcl_GetChannelBufferSize(chan_r);
  buffer = (char*)malloc(bufsize);
  /* keep reading bufsize characters until the input is exhausted */
  /* this idea is taken from tclIOCmd.c */
  for(;;) {
    int bytes_read = Tcl_Read(chan_r, buffer, bufsize);
    if(bytes_read < 0) break;
    Tcl_Write(chan_w, buffer, bytes_read);
    fprintf(stderr, "wrote %d bytes\n", bytes_read);
    if(bytes_read < bufsize) break;
  }
  Tcl_Flush(chan_w);
  free(buffer);
  return TCL_OK;
}

/* why fd_set is in time.h, I don't know */
#include <sys/time.h>

static int forced_write(int fd, char *buffer, int bytes)
{
  for(;;) {
    int written = write(fd, buffer, bytes);
    if(written < 0) { 
      if((errno == EAGAIN) || (errno == EWOULDBLOCK)) {
	continue;
      }
      perror("Write error"); 
      return 1; 
    }
    if(written == 0) continue;
    if(written < bytes) { buffer += written; bytes -= written; continue; }
    break;
  }
  return 0;
}

/* Copies whatever is ready at f1 to f2. */
static int block_copy(int f1, int f2)
{
  char buffer[8192];
  int bytes = read(f1, buffer, sizeof(buffer));
  if(bytes <= 0) {
    if(bytes < 0) perror("Read error");
    return 1;
  }
  if(forced_write(f2, buffer, bytes)) return 1;
#if 0
  fprintf(stderr, "wrote %d bytes\n", bytes);
#endif
  return 0;
}

static void splice_fds(int fd1, int fd2)
{
  fd_set reads;

  FD_ZERO(&reads);

  while(1) {
    FD_SET(fd1, &reads);
    FD_SET(fd2, &reads);

    if(select(max(fd1, fd2)+5, &reads, NULL, NULL, NULL) <= 0)
      return;

    if(FD_ISSET(fd1, &reads))
      if(block_copy(fd1, fd2))
	return;
    
    if(FD_ISSET(fd2, &reads))
      if(block_copy(fd2, fd1))
	return;
  }
}

static int SpliceCmd(ClientData clientData, Tcl_Interp *interp,
		     int argc, char *argv[])
{
  Tcl_Channel chan1, chan2;
  Tcl_File file1, file2;
  int fd1, fd2;
  Err(argc != 3, "usage: spliceChannels channel1 channel2");
  chan1 = Tcl_GetChannel(interp, argv[1], NULL);
  Err(!chan1, "invalid channel number");
  chan2 = Tcl_GetChannel(interp, argv[2], NULL);
  Err(!chan2, "invalid channel number");

  /* make the channels nonblocking */
  Tcl_SetChannelOption(interp, chan1, "-blocking", "0");
  Tcl_SetChannelOption(interp, chan2, "-blocking", "0");

  /* get the files */
  file1 = Tcl_GetChannelFile(chan1, TCL_READABLE);
  file2 = Tcl_GetChannelFile(chan2, TCL_READABLE);
  
  /* get the unix file descriptors */
  fd1 = (int)Tcl_GetFileInfo(file1, NULL);
  fd2 = (int)Tcl_GetFileInfo(file2, NULL);

#if !defined(linux) && !defined(sun)
  if(1) {
    /* don't create zombies */
    struct sigaction act;
    act.sa_flags = SA_NOCLDWAIT;
    act.sa_handler = NULL;
    sigaction(SIGCLD, &act, NULL);
  }
#endif

  /* fork a process to do the work */
  if(fork() == 0) {
    splice_fds(fd1, fd2);
    close(fd1);
    close(fd2);
    fprintf(stderr, "splice broken\n");
    /* become a zombie */
    exit(0);
  }
  /* close the channels */
  Tcl_UnregisterChannel(interp, chan1);
  Tcl_UnregisterChannel(interp, chan2);

  return TCL_OK;
}

int Utiltcl_Init(Tcl_Interp *interp)
{
  Tcl_CreateCommand(interp, "sort", SortCmd, NULL, NULL);
  Tcl_CreateCommand(interp, "ran", RanCmd, NULL, NULL);
  Tcl_CreateCommand(interp, "makesafe", MakeSafeCmd, NULL, NULL);
  Tcl_CreateCommand(interp, "deref", DerefCmd, NULL, NULL);
  Tcl_CreateCommand(interp, "MEM_BLOCKS", MemBlocksCmd, NULL, NULL);
  Tcl_CreateCommand(interp, "readInt", ReadIntCmd, NULL, NULL);
  Tcl_CreateCommand(interp, "writeInt", WriteIntCmd, NULL, NULL);
  Tcl_CreateCommand(interp, "readFloat", ReadFloatCmd, NULL, NULL);
  Tcl_CreateCommand(interp, "readWrite", ReadWriteCmd, NULL, NULL);
  Tcl_CreateCommand(interp, "spliceChannels", SpliceCmd, NULL, NULL);
  return Tcl_PkgProvide(interp, "utilTcl", "1");
}
