/* 
label 1 motion 2710 0 0
get 2710 2763
label -1 motion 2762 0 0
get 2710 2763
*/
/* Interactive learning with an across hier.
 * Input:
 *   "label" weight within member segment label
 *   get from to
 * Output:
 *   weight within member segment label
 */
#include <ph/photobook.h>
#include <tpm/learn.h>
#include <tpm/boa_learn.h>

#define VERBOSE 1

int AcrossDataCompare(AcrossData *a, AcrossData *b)
{
  return memcmp(a, b, sizeof(AcrossData));
}

Type AcrossDataType;

List AcrossDataInstance(AcrossData *d)
{
  List instances = ListCreate((FreeFunc*)Boa_Free);
  Obj t = Boa_CreateFromData(AcrossDataType, NULL);
  t->value = d;
  ListAddFront(instances, t);
  return instances;
}

typedef struct Range {
  int low, high;
} *Range;

int RangeContains(AcrossData *data, Range range)
{
  return (data->member < range->low) || (data->member > range->high);
}

List GetLabels(Obj classifier, int from_member, int to_member)
{
  struct Range range;
  range.low = from_member;
  range.high = to_member;
  return Boa_Classifier_AllMatching(classifier, &range, 
				    (CmpFunc*)RangeContains);
}

main(int argc, char *argv[])
{
  Obj od, across, member_list, within_list, features, classifiers, t;
  char str[128];
  int i;
  int num_words;
  char **words;

  if(argc < 3) {
    fprintf(stderr, "usage: across_interactive db hier hier ...\n");
    exit(1);
  }

  Boa_Startup();
  Boa_Tcl_Pkg();
  Boa_Hier_Pkg();
  Boa_Learn_Pkg();

  od = Boa_Create(T_ObjDir);
  sprintf(str, "%s/%s", getenv("PHOTOBOOK_DB_DIR"), argv[1]);
  Boa_StringSet(od, "_path", BOA_REPLACE, Boa_String_01(str));

  if(Boa_StringGet(od, "index", 0, NULL, &member_list) != TPM_OK) {
    ErrorPrepend("Cannot read index");
    ErrorPrint(stderr);
    exit(1);
  }
  if(Boa_StringGet(od, "within_list", 0, NULL, &within_list) != TPM_OK) {
    ErrorPrepend("Cannot read within_list");
    ErrorPrint(stderr);
    exit(1);
  }

  /* read the across directory and acrosses from the command line */
  if(Boa_StringGet(od, "across", 0, NULL, &across) != TPM_OK) {
    ErrorPrepend("Cannot read across");
    ErrorPrint(stderr);
    exit(1);
  }
  features = Boa_Create(T_Ord);
  for(i=2;i<argc;i++) {
    char *fname = argv[i];
    if(Boa_StringGet(across, fname, 0, NULL, &t) != TPM_OK) {
      ErrorPrepend("Cannot open across hier `%s'", fname);
      ErrorPrint(stderr);
      exit(1);
    }
    if(t->type->type != T_Hier) {
      fprintf(stderr, "`%s' is not a Hier\n", fname);
      exit(1);
    }
    Boa_StringSet(features, fname, BOA_INSERT, t);
  }
  Boa_Release(across);
  Boa_Release(od);

  /* create the classifiers */
  AcrossDataType = Boa_Type_Create("(array 3 int)");
  classifiers = Boa_MultiClass_Create(Boa_Classifier_Create(
                  T_Classifier_SC, features, (CmpFunc*)AcrossDataCompare, 
		  (InstanceFunc*)AcrossDataInstance));

  /* read loop */
  fprintf(stderr, "READY\n");
  for(i=0;;i++,TpmSplitFree(num_words, words)) {
    char line[128];
    if(!getline(line, 128, stdin)) break;
    /* parse into words */
    TpmSplit(line, ' ', &num_words, &words);

    if(i % 100 == 0) fprintf(stderr, "%s\n", line);

    if((num_words == 6) && !strcmp(words[0], "label")) {
      int label;
      float weight;
      AcrossData *data;
      Obj t;
      Obj classifier;

      label = atoi(words[5]);
      while(label >= Boa_Ord_Length(classifiers)) {
	Boa_MultiClass_AddClass(classifiers);
      }
      weight = atof(words[1]);

      t = Boa_Create(AcrossDataType);
      data = (AcrossData*)t->value;
      /* look up the within number */
      if(Boa_StringGetInt(within_list, words[2], &data->within) != TPM_OK) {
	fprintf(stderr, "unknown within `%s'\n", words[2]);
	continue;
      }
      /* look up the member number */
      if(Boa_StringGetInt(member_list, words[3], &data->member) != TPM_OK) {
	fprintf(stderr, "unknown member `%s'\n", words[3]);
	continue;
      }
      data->segment = atoi(words[4]);

      classifier = Boa_Ord_Item(classifiers, label);
      if(Boa_ObjSet(classifier, t, BOA_INSERT,
		    Boa_Float_Create(weight)) != TPM_OK) {
	ErrorPrepend("Classifier set failed for label %d", label);
	ErrorPrint(stderr);
      }
      Boa_Release(t);
    }
    else if((num_words == 3) && !strcmp(words[0], "get")) {
      int from, to, label;
      if(Boa_StringGetInt(member_list, words[1], &from) != TPM_OK) {
	fprintf(stderr, "unknown member `%s'\n", words[1]);
	continue;
      }
      if(Boa_StringGetInt(member_list, words[2], &to) != TPM_OK) {
	fprintf(stderr, "unknown member `%s'\n", words[2]);
	continue;
      }
      /* loop labels */
      for(label=0;label<Boa_Ord_Length(classifiers);label++) {
	List results = GetLabels(Boa_Ord_Item(classifiers, label), from, to);
	AcrossData *data;
	ListIter(p, data, results) {
	  char *within, *member;
	  /* get the within name */
	  Boa_IntGet(within_list, data->within, 0, &T_String, &t);
	  within = Boa_String_Value(t);
	  Boa_Release(t);
	  /* get the member name */
	  Boa_IntGet(member_list, data->member, 0, &T_String, &t);
	  member = Boa_String_Value(t);
	  Boa_Release(t);
	  printf("%g %s %s %d %d\n", 
		 1.0, within, member, data->segment, label);
	}
	ListFree(results);
      }
      printf("END\n");
      fflush(stdout);
    }
    else fprintf(stderr, "unrecognized command `%s'\n", line);
  }
  Boa_Release(classifiers);
  Boa_Release(within_list);
  Boa_Release(member_list);

  Boa_Shutdown();
  exit(0);
}

