/* Usage: neighbors model features index > output
 * To compute query results for each of the members listed in index,
 * using the given model.
 * "features" is the member context to give to the model.
 * Output is a line of indices per member, starting with itself.
 * Indices start at zero.
 */
#include <ph/photobook.h>

typedef struct {
  float dist;
  int index;
} DistRec;

void main(int argc, char *argv[])
{
  Obj od, model, features, tests, query;
  DistRec *recs;
  int ntests, threshold, i,j;

  if(argc < 4) {
    fprintf(stderr, "usage: neighbors model members index [threshold] > output\n");
    exit(1);
  }

  Boa_Startup();
  Boa_Vector_Pkg();
  Boa_Hier_Pkg();
  Boa_Tcl_Pkg();    /* since disk objects may need it. grrr */

  od = Boa_Create(T_ObjDir);
  if(Boa_StringGet(od, argv[1], '/', NULL, &model)) {
    ErrorPrepend("Cannot open `%s'", argv[1]);
    ErrorPrint(stderr);
    exit(1);
  }
  if(Boa_StringGet(od, argv[2], '/', NULL, &features)) {
    ErrorPrepend("Cannot open `%s'", argv[2]);
    ErrorPrint(stderr);
    exit(1);
  }
  if(Boa_StringGet(od, argv[3], '/', NULL, &tests)) {
    ErrorPrepend("Cannot open `%s'", argv[3]);
    ErrorPrint(stderr);
    exit(1);
  }

  assert(Boa_StringGetInt(tests, "_length", &ntests) == TPM_OK);
  recs = Allocate(ntests, DistRec);
  if(argc > 4) threshold = atoi(argv[4]);
  else threshold = ntests;

  Boa_StringSet(model, "features", BOA_INSERT, features);
  /* Make a dummy entries that will be mutated later.
   * This is an efficiency hack to avoid setting each time.
   */
  /* index now becomes property of the model, but we will still use it */
  Boa_StringSet(model, "test", BOA_INSERT, tests);

  for(i=0;i<ntests;i++) {
    Obj query, result;
    /* get the query */
    assert(Boa_IntGet(tests, i, 0, NULL, &query) == TPM_OK);
    fprintf(stderr, "query: ");
    Boa_PrintNL(stderr, query);
    /* make the query */
    if(Boa_ObjGet(model, query, 0, NULL, &result) != TPM_OK) {
      ErrorPrepend("Call failed");
      ErrorPrint(stderr);
      exit(1);
    }
    Boa_Release(query);
    /* result should be a vector of length ntests */
    for(j=0;j<ntests;j++) {
      recs[j].dist = Boa_Vector_Value(result)->d[j];
      recs[j].index = j;
    }
    Boa_Release(result);
    qsort(recs, ntests, sizeof(DistRec), FloatCmp);
    /* force the query to be first */
    if(recs[0].index != i) {
      for(j=1;j<ntests;j++) {
	if(recs[j].index == i) {
	  DistRec t;
	  t = recs[0];
	  recs[0] = recs[j];
	  recs[j] = t;
	  break;
	}
      }
    }
    for(j=0;j<threshold;j++) printf("%d ", recs[j].index);
    printf("\n");
  }

  free(recs);
  Boa_Release(model);
  Boa_Release(od);

  Boa_Shutdown();
  exit(0);
}
