/* why don't acrosses store member+seg?
 * because seg might be costly in general
 */
#include <tpm/list.h>
#include "photobook.h"
#include <tpm/boa_learn.h>

#define VERBOSE 1
#define BLOCK 0
#define EXCLUSIVE 1

static Obj classifiers = NULL;   /* Ord of one T_Classifier per label */
static Obj withins;              /* Tuple of within-hier directories */
static Obj member_index;
static Obj tests;		 /* List of probes for each image (Ord) */

#define SegAt(hier, i) ((Seg*)&HierNodeAt(hier, i)->data)

/* Prototypes ****************************************************************/
void Ph_AcrossInit(Obj members, Obj across, List across_list, 
		   Obj within, List within_list);
void Ph_AcrossDestroy(void);
void Ph_AcrossEnable(int tree, int flag);
List Ph_AcrossGet(int member, int label);
Seg Ph_AcrossSeg(int member, int label);
void Ph_AcrossEx(int member, int label, Seg *seg, float weight);

/* Functions *****************************************************************/

/* Returns an unheld string (caller doesn't free it) */
static char *MemberName(int index)
{
  char *name;
  Obj t;
  if(Boa_IntGet(member_index, index, 0, &T_String, &t) != TPM_OK)
    return NULL;
  name = Boa_String_Value(t);
  Boa_Release(t);
  return name;
}

static Seg *GetSeg(AcrossData *data)
{
  Seg *seg;
  Obj t;
  if((data->within < 0) || (data->within >= Boa_Tuple_Length(withins)) || 
     !Boa_Tuple_Item(withins, data->within)) {
    fprintf(stderr, "Across hierarchy wants nonexistent within-hier %d\n", data->within);
    return NULL;
  }
  if(Boa_StringGet(Boa_Tuple_Item(withins, data->within), 
		   MemberName(data->member), 
		   0, NULL, &t)) {
    ErrorPrepend("Cannot load within-hier for %s", 
		    MemberName(data->member));
    ErrorPrint(stderr);
    return NULL;
  }
  if(t->type->type != T_Hier) {
    ErrorSet("within file for %s is not a Hier",
	     MemberName(data->member));
    return NULL;
  }
  if(data->segment >= Boa_Hier_Length(t->type)) {
    fprintf(stderr, "segment %d >= %d\n", 
	    data->segment, Boa_Hier_Length(t->type));
    return NULL;
  }
  seg = SegAt(Boa_Hier_Value(t), data->segment);
  Boa_Release(t);  /* assume t persists */
  return seg;
}

/* Instance structure */
typedef struct {
  int member;
  int x,y;
} SegData;

/* Returns 0 if there is any intersection */
static int SegCompare(AcrossData *hdata, SegData *sdata)
{
  Seg *iseg;
  /* this is why member should be numeric */
  if(hdata->member != sdata->member) return 1;
  /* get the segmask */
  iseg = GetSeg(hdata);
  if(!iseg) return 1;
  return !SegIsSet(iseg, sdata->x, sdata->y);
}

static Type SegDataType;

static List SegDataInstance(AcrossData *d)
{
  List instances = ListCreate((FreeFunc*)Boa_Free);
  SegData p;
  Seg *seg;
  p.member = d->member;
  seg = GetSeg(d);
  if(!seg) return instances;
  for(p.y=0;p.y<SEG_YCELLS;p.y++) {
    for(p.x=0;p.x<SEG_XCELLS;p.x++) {
      if(SegIsSet(seg, p.x, p.y)) {
	Obj t = Boa_CreateFromData(SegDataType, AllocCopy(1, SegData, &p));
	ListAddFront(instances, t);
      }
    }
  }
  return instances;
}

/* Create a list of tests for probing an image.
 * Would be nice if this could just be a computed stream.
 */
static Obj TestList(int member)
{
  Obj list = Boa_Create(T_List);
  int x,y;
  /* loop all cells */
  for(y = 0; y < SEG_YCELLS; y++) {
    for(x = 0; x < SEG_XCELLS; x++) {
      Obj instance = Boa_Create(SegDataType);
      SegData *data = instance->value;
      data->member = member;
      data->x = x;
      data->y = y;
      ListAddFront(Boa_List_Value(list), instance);
    }
  }
  return list;
}

void Ph_AcrossInit(Obj members, Obj across, List across_list, 
		   Obj within, List within_list)
{
  Obj features, t;
  char *fname, str[64];
  int i;

  Boa_Learn_Pkg();
  Ph_AcrossDestroy();
  member_index = Boa_Hold(members);
  features = Boa_Create(T_Ord);
  {ListIter(p, fname, across_list) {
    if(Boa_StringGet(across, fname, 0, NULL, &t)) {
      ErrorPrepend("Cannot open across hier `%s'\n", fname);
      ErrorPrint(stderr);
      continue;
    }
    if(t->type->type != T_Hier) {
      fprintf(stderr, "`%s' is not a Hier\n", fname);
      continue;
    }
    Boa_StringSet(features, fname, BOA_INSERT, t);
  }}
  
  sprintf(str, "(tuple %d)", ListSize(within_list));
  withins = Boa_Create(Boa_Type_Create(str));
  i = 0;
  {ListIter(p, fname, within_list) {
    if(Boa_StringGet(within, fname, 0, NULL, &t)) {
      ErrorPrepend("Cannot open within hier `%s'\n", fname);
      ErrorPrint(stderr);
      Boa_Tuple_Item(withins, i) = NULL;
    }
    else 
      Boa_Tuple_Item(withins, i) = t;
    i++;
  }}

  /* create the classifiers */
  SegDataType = Boa_Type_Create("(array 3 int)");
  classifiers = Boa_MultiClass_Create(
    Boa_Classifier_Create(T_Classifier_SC, 
			  features, 
			  (CmpFunc*)SegCompare, 
			  (InstanceFunc*)SegDataInstance));
#if BLOCK
  /* overloaded assignment might help here */
  t = Boa_MultiClass_CreateBlock(classifiers);
  Boa_Release(classifiers);
  classifiers = t;
#endif
#if EXCLUSIVE
  t = Boa_MultiClass_CreateExclusive(classifiers);
  Boa_Release(classifiers);
  classifiers = t;
#endif
  if(BLOCK) {
    int nmembers;
    assert(Boa_StringGetInt(member_index, "_length", &nmembers) == TPM_OK);
    tests = Boa_Create(T_Ord);
    for(i=0;i<nmembers;i++) {
      Boa_StringSet(tests, MemberName(i), BOA_INSERT, TestList(i));
    }
  }
}

void Ph_AcrossDestroy(void)
{
  if(!classifiers) return;
  Boa_Release(classifiers);
  classifiers = NULL;
  Boa_Release(withins);
  Boa_Release(member_index);
  if(BLOCK) Boa_Release(tests);
}

void Ph_AcrossEnable(int tree, int flag)
{
}

/* Returns the SegUnion of all segments of member with the given label. */
Seg Ph_AcrossSeg(int member, int label)
{
  Seg seg = SegZero;
  Obj classifier;
  int status;

  if(!classifiers || (label < 0) || (label >= Boa_Ord_Length(classifiers))) 
    return seg;
  classifier = Boa_Ord_Item(classifiers, label);

  if(BLOCK) {
    SegData *data;
    Obj t, test, results;
    ListNode pt, pr;
    test = Boa_Ord_Item(tests, member);
    status = Boa_ObjGet(classifier, test, '/', NULL, &results);
    if(status == TPM_OK) {
      if(VERBOSE) fprintf(stderr, "block\n");
      for(pt = ListFront(Boa_List_Value(test)), 
	  pr = ListFront(Boa_List_Value(results));
	  pt && pr;
	  pt = pt->next, pr = pr->next) {
	if(Boa_Float_Value((Obj)pr->data) > 0) {
	  t = (Obj)pt->data;
	  data = t->value;
	  SegSet(&seg, data->x, data->y);
	}
      }
      Boa_Release(results);
    }
  }
  else status = TPM_ERROR;
  if(status != TPM_OK) {
    SegData data;
    Obj instance;

    /* Loop all cells and set those which are in the class */
    data.member = member;
    /* instance will track data */
    instance = Boa_CreateFromData(SegDataType, NULL);
    instance->value = &data;
    /* loop all cells */
    for(data.y = 0; data.y < SEG_YCELLS; data.y++) {
      for(data.x = 0; data.x < SEG_XCELLS; data.x++) {
	Obj t;
	if(Boa_ObjGet(classifier, instance, 0, &T_Float, &t) != TPM_OK) {
	  ErrorPrepend("classifier failed on (%d, %d)", data.x, data.y);
	  ErrorPrint(stderr);
	  continue;
	}
	if(Boa_Float_Value(t) > 0) SegSet(&seg, data.x, data.y);
	Boa_Release(t);
      }
    }
    Boa_Release(instance);
  }
  return seg;
}

static int MemCompare(AcrossData *data, int *member)
{
  return data->member != *member;
}

/* Returns a List of AcrossData, or NULL if the label is unknown. */
List Ph_AcrossGet(int member, int label)
{
  Obj classifier;
  if(!classifiers || (label < 0) || (label >= Boa_Ord_Length(classifiers))) 
    return NULL;
  classifier = Boa_Ord_Item(classifiers, label);
#if 1
  return NULL;
#else
  return Boa_Classifier_AllMatching(classifier, &member, (CmpFunc*)MemCompare);
#endif
}

#if 0
/* Returns a List of AcrossData, or NULL if the label is unknown. */
List Ph_AcrossGet(int member, int label)
{
  Obj classifier, instance;
  List result;
  int i, j;
  if(!classifiers || (label < 0) || (label >= Boa_Ord_Length(classifiers))) 
    return NULL;
  result = ListCreate(NULL);
  classifier = Boa_Ord_Item(classifiers, label);
  instance = Boa_CreateFromData(Boa_Type_Create("(array 3 int)"), NULL);
  /* Loop features */
  for(i=0;i<Boa_Ord_Length(features);i++) {
    /* Loop leaves */
    Hier hier = Boa_Hier_Value(Boa_Ord_Item(features, i));
    for(j=0;j<hier->num_nodes;j++) {
      Obj t;
      instance->value = &HierNodeAt(hier, j)->data;
      /* is this leaf in the class? */
      Boa_ObjGet(classifier, instance, 0, &T_Float, &t);
      if(Boa_Float_Value(t) > 0) ListAddFront(instance->value);
      Boa_Release(t);
    }
  }
  Boa_Release(instance);
  return result;
}

/* Returns the SegUnion of all segments of member with the given label. */
Seg Ph_AcrossSeg(int member, int label)
{
  Seg result = SegZero;
  AcrossData *data;
  List segments = Ph_AcrossGet(member, label);
  if(!segments) return result;
  {ListIter(p, data, segments) {
    /* Get the segment file corresponding to this item */
    Obj obj;
    if(GetSegments(data->within, member, &obj)) continue;
    SegUnion(&result, SegAt((Hier)obj->value, data->segment));
    Boa_Release(obj);
  }}
  ListFree(segments);
  return result;
}
#endif

void Ph_AcrossEx(int member, int label, Seg *seg, float weight)
{
  SegData data;
  Obj instance, classifier, value;

  if(!classifiers) {
    fprintf(stderr, "Ph_AcrossEx: not initialized\n"); 
    return;
  }
  while(label >= Boa_Ord_Length(classifiers)) {
    Boa_MultiClass_AddClass(classifiers);
  }
  classifier = Boa_Ord_Item(classifiers, label);

  data.member = member;
  /* instance will track data */
  instance = Boa_CreateFromData(SegDataType, NULL);
  instance->value = &data;

  /* make each cell of the seg an example */
  value = Boa_Float_Create(weight);
  for(data.y = 0; data.y < SEG_YCELLS; data.y++) {
    for(data.x = 0; data.x < SEG_XCELLS; data.x++) {
      if(SegIsSet(seg, data.x, data.y)) {
	Boa_Hold(value);
	Boa_ObjSet(classifier, instance, 0, value);
      }
    }
  }
  Boa_Release(value);
  Boa_Release(instance);
}
