#include <tpm/matrix.h>
#include <boa/boa.h>
#include <boa/image.h>

/* Prototypes ****************************************************************/

int Boa_ImageOhta(Obj context, Type *type_in_out, Obj *value_return);
int Boa_ImageGray(Obj context, Type *type_in_out, Obj *value_return);
int Boa_ImageZoom(Obj context, Type *type_in_out, Obj *value_return);
int Boa_ImageCooc(Obj context, Type *type_in_out, Obj *value_return);

/* Functions *****************************************************************/

/* Produces the Ohta equivalent of an RGB image.
 * Input: image
 * Output: Ohta image
 */
int Boa_ImageOhta(Obj context, Type *type_in_out, Obj *value_return)
{
  Obj t;
  TpmImage image;
  BOA_TYPECHECK_GET(T_Image);
  /* get args */
  TRY_RETURN(Boa_StringGet(context, "image", 0, &T_Image, &t));
  image = TpmImageCopy(Boa_Image_Value(t));
  TpmImageRGB2Ohta(image);
  *value_return = Boa_Image_Create(image);
  Boa_Release(t);
  return TPM_OK;
}

/* Produces the NTSC gray equivalent of an RGB image.
 * Input: image
 * Output: gray image
 */
int Boa_ImageGray(Obj context, Type *type_in_out, Obj *value_return)
{
  Obj t;
  BOA_TYPECHECK_GET(T_Image);
  /* get args */
  TRY_RETURN(Boa_StringGet(context, "image", 0, &T_Image, &t));
  *value_return = Boa_Image_Create(TpmImageGray(Boa_Image_Value(t)));
  Boa_Release(t);
  return TPM_OK;
}

/* Zooms a TpmImage.
 * Input: image, zfact
 * Output: zoomed image (according to TpmImageZoom)
 */
int Boa_ImageZoom(Obj context, Type *type_in_out, Obj *value_return)
{
  int zfact;
  Obj t;
  BOA_TYPECHECK_GET(T_Image);
  /* get args */
  TRY_RETURN(Boa_StringGetInt(context, "zfact", &zfact));
  TRY_RETURN(Boa_StringGet(context, "image", 0, &T_Image, &t));
  *value_return = Boa_Image_Create(TpmImageZoom(Boa_Image_Value(t), zfact));
  Boa_Release(t);
  return TPM_OK;
}

/* Computes the cooccurrence matrix of a TpmImage.
 * Input: image, dx, dy, levels
 * Output: cooccurrence image (stretched)
 */
int Boa_ImageCooc(Obj context, Type *type_in_out, Obj *value_return)
{
  int order, levels;
  Matrix cooc;
  Obj t;
  BOA_TYPECHECK_GET(T_Image);
  /* get args */
  if(Boa_StringGetInt(context, "order", &order) != TPM_OK) {
    order = 1;
  }
  if(Boa_StringGetInt(context, "levels", &levels) != TPM_OK) {
    levels = 64;
  }
  TRY_RETURN(Boa_StringGet(context, "image", 0, &T_Image, &t));
  cooc = MatrixCreate(levels, levels);
  VectorSet((Vector)cooc, 0);
#if 0
  CoocMatrixFromImage(Boa_Image_Value(t), 1, 0, cooc);
  CoocMatrixFromImage(Boa_Image_Value(t), 0, 1, cooc);
#endif
  CoocMatrixFromImage(Boa_Image_Value(t), 2, 2, cooc);
  if(order > 1) {
    CoocMatrixFromImage(Boa_Image_Value(t), 2, 0, cooc);
    CoocMatrixFromImage(Boa_Image_Value(t), 0, 2, cooc);
    CoocMatrixFromImage(Boa_Image_Value(t), 1, 1, cooc);
    if(order > 2) {
      CoocMatrixFromImage(Boa_Image_Value(t), 0, 3, cooc);
      CoocMatrixFromImage(Boa_Image_Value(t), 3, 0, cooc);
      CoocMatrixFromImage(Boa_Image_Value(t), 1, 2, cooc);
      CoocMatrixFromImage(Boa_Image_Value(t), 2, 1, cooc);
      if(order > 3) {
	CoocMatrixFromImage(Boa_Image_Value(t), 0, 4, cooc);
	CoocMatrixFromImage(Boa_Image_Value(t), 4, 0, cooc);
	CoocMatrixFromImage(Boa_Image_Value(t), 1, 3, cooc);
	CoocMatrixFromImage(Boa_Image_Value(t), 3, 1, cooc);
	CoocMatrixFromImage(Boa_Image_Value(t), 2, 2, cooc);
      }
    }
  }
  Boa_Release(t);
  VectorAddValue((Vector)cooc, 1);
  VectorApply((Vector)cooc, (MathFunc*)log);
  VectorRescale((Vector)cooc, 0, 255);
  *value_return = Boa_Image_Create(TpmImageFromMatrix(cooc));
  MatrixFree(cooc);
  return TPM_OK;
}
