#include <tpm/list.h>
#include <boa/boa.h>
#include <boa/vector.h>

/* Prototypes ****************************************************************/

int RankDistance(Obj context, Type *type_in_out, Obj *value_return);
int LinearCombination(Obj context, Type *type_in_out, Obj *value_return);
int MaxCombination(Obj context, Type *type_in_out, Obj *value_return);
int MinCombination(Obj context, Type *type_in_out, Obj *value_return);
int SymmetricMax(Obj context, Type *type_in_out, Obj *value_return);
int SymmetricMin(Obj context, Type *type_in_out, Obj *value_return);

/* Functions *****************************************************************/

/* Rank **********************************************************************/

typedef struct {
  int rank;
  double distance;
} RankRec;

static int RankCompare(RankRec *a, RankRec *b)
{
  return (a->distance > b->distance) ? 1 : -1;
}

static int CallMetric(Obj metric, 
		      Obj features, Obj test, Obj query, 
		      Obj *value_return)
{
  assert(Boa_StringSet(metric, "features", BOA_INSERT|BOA_REPLACE, features) 
	 == TPM_OK);
  assert(Boa_StringSet(metric, "test", BOA_INSERT|BOA_REPLACE, test)
	 == TPM_OK);
  if(Boa_ObjGet(metric, query, 0, NULL, value_return) != TPM_OK) {
    ErrorPrepend("metric failed");
    Boa_Release(query);
    return TPM_BOTTOM;
  }
  Boa_Release(query);
  return TPM_OK;
}

/* Produces a distance list for a set of members relative to a query.
 * Inputs: features (context), query (string), test (array of string), 
 *         metric (transforms (f,q,t) -> distance vector)
 * Output: vector of distances
 */
int RankDistance(Obj context, Type *type_in_out, Obj *value_return)
{
  Obj metric, features, tests, query, result;
  int ntests, i, rank;
  double cur_dist;
  Vector dists;
  RankRec *ranks;

  BOA_TYPECHECK_GET(T_Vector);
  TRY_RETURN(Boa_StringGet(context, "metric", 0, NULL, &metric));
  TRY_RETURN(Boa_StringGet(context, "features", 0, NULL, &features));
  if(Boa_StringGet(context, "query", 0, NULL, &query) != TPM_OK) {
    ErrorPrepend("no value for `query'");
    return TPM_BOTTOM;
  }
  if(Boa_StringGet(context, "test", 0, NULL, &tests) != TPM_OK) {
    ErrorPrepend("no value for `test'");
    return TPM_BOTTOM;
  }
  assert(Boa_StringGetInt(tests, "_length", &ntests) == TPM_OK);

  if(CallMetric(metric, features, tests, query, &result) != TPM_OK) {
    return TPM_BOTTOM;
  }
  Boa_Release(metric);
  dists = Boa_Vector_Value(result);

  /* Attach ranks */
  ranks = Allocate(ntests, RankRec);
  for(i=0;i<ntests;i++) {
    ranks[i].rank = i;
    ranks[i].distance = dists->d[i];
  }
  Boa_Release(result);

  /* Sort the ranks array to get rank ordering */
  qsort((void*)ranks, ntests, sizeof(RankRec), (CmpFunc*)RankCompare);

  /* Read ranks into output vector */
  dists = VectorCreate(ntests);
  rank = 0;
  cur_dist = -10e10;
  for(i=0;i<ntests;i++) {
    dists->d[ranks[i].rank] = rank;
    /* only increase rank if the distance has increased */
    if(ranks[i].distance > cur_dist) {
      rank++;
      cur_dist = ranks[i].distance;
    }
  }
  free(ranks);

  *value_return = Boa_Vector_Create(dists);
  return TPM_OK;
}

/* LinearCombination *********************************************************/

/* Produces a distance list for a set of members relative to a query.
 * Inputs: features (context), query (string), test (array of string), 
 *         metrics (list of transforms (f,q,t) -> distance vector)
 * Output: vector of distances
 */
int LinearCombination(Obj context, Type *type_in_out, Obj *value_return)
{
  Obj metrics, metric, features, tests, query, result;
  int ntests, i;
  Vector dists;

  BOA_TYPECHECK_GET(T_Vector);
  TRY_RETURN(Boa_StringGet(context, "metrics", 0, &T_List, &metrics));
  TRY_RETURN(Boa_StringGet(context, "features", 0, NULL, &features));
  if(Boa_StringGet(context, "query", 0, NULL, &query) != TPM_OK) {
    ErrorPrepend("no value for `query'");
    return TPM_BOTTOM;
  }
  if(Boa_StringGet(context, "test", 0, NULL, &tests) != TPM_OK) {
    ErrorPrepend("no value for `test'");
    return TPM_BOTTOM;
  }
  assert(Boa_StringGetInt(tests, "_length", &ntests) == TPM_OK);

  dists = VectorCreate(ntests);
  VectorSet(dists, 0);

  {ListIter(p, metric, Boa_List_Value(metrics)) {
    if(CallMetric(metric, 
		  Boa_Hold(features), Boa_Hold(tests), Boa_Hold(query), 
		  &result) != TPM_OK) {
      goto error;
    }
    for(i=0;i<ntests;i++) {
      dists->d[i] += Boa_Vector_Value(result)->d[i];
    }
    Boa_Release(result);
  }}

  Boa_Release(metrics);
  Boa_Release(features);
  Boa_Release(tests);
  Boa_Release(query);
  *value_return = Boa_Vector_Create(dists);
  return TPM_OK;

 error:
  VectorFree(dists);
  Boa_Release(metrics);
  Boa_Release(features);
  Boa_Release(tests);
  Boa_Release(query);
  return TPM_BOTTOM;
}

/* MaxCombination ************************************************************/

/* Produces a distance list for a set of members relative to a query.
 * Inputs: features (context), query (string), test (array of string), 
 *         metrics (list of transforms (f,q,t) -> distance vector)
 * Output: vector of distances
 */
int MaxCombination(Obj context, Type *type_in_out, Obj *value_return)
{
  Obj metrics, metric, features, tests, query, result;
  int ntests, i;
  Vector dists;

  BOA_TYPECHECK_GET(T_Vector);
  TRY_RETURN(Boa_StringGet(context, "metrics", 0, &T_List, &metrics));
  TRY_RETURN(Boa_StringGet(context, "features", 0, NULL, &features));
  if(Boa_StringGet(context, "query", 0, NULL, &query) != TPM_OK) {
    ErrorPrepend("no value for `query'");
    return TPM_BOTTOM;
  }
  if(Boa_StringGet(context, "test", 0, NULL, &tests) != TPM_OK) {
    ErrorPrepend("no value for `test'");
    return TPM_BOTTOM;
  }
  assert(Boa_StringGetInt(tests, "_length", &ntests) == TPM_OK);

  dists = VectorCreate(ntests);
  VectorSet(dists, 0);

  {ListIter(p, metric, Boa_List_Value(metrics)) {
    if(CallMetric(metric, 
		  Boa_Hold(features), Boa_Hold(tests), Boa_Hold(query), 
		  &result) != TPM_OK) {
      goto error;
    }
    for(i=0;i<ntests;i++) {
      if(Boa_Vector_Value(result)->d[i] > dists->d[i])
	dists->d[i] = Boa_Vector_Value(result)->d[i];
    }
    Boa_Release(result);
  }}

  Boa_Release(metrics);
  Boa_Release(features);
  Boa_Release(tests);
  Boa_Release(query);
  *value_return = Boa_Vector_Create(dists);
  return TPM_OK;

 error:
  VectorFree(dists);
  Boa_Release(metrics);
  Boa_Release(features);
  Boa_Release(tests);
  Boa_Release(query);
  return TPM_BOTTOM;
}

/* MinCombination ************************************************************/

/* Same as above except for one comparison and the initial value */
/* Produces a distance list for a set of members relative to a query.
 * Inputs: features (context), query (string), test (array of string), 
 *         metrics (list of transforms (f,q,t) -> distance vector)
 * Output: vector of distances
 */
int MinCombination(Obj context, Type *type_in_out, Obj *value_return)
{
  Obj metrics, metric, features, tests, query, result;
  int ntests, i;
  Vector dists;

  BOA_TYPECHECK_GET(T_Vector);
  TRY_RETURN(Boa_StringGet(context, "metrics", 0, &T_List, &metrics));
  TRY_RETURN(Boa_StringGet(context, "features", 0, NULL, &features));
  if(Boa_StringGet(context, "query", 0, NULL, &query) != TPM_OK) {
    ErrorPrepend("no value for `query'");
    return TPM_BOTTOM;
  }
  if(Boa_StringGet(context, "test", 0, NULL, &tests) != TPM_OK) {
    ErrorPrepend("no value for `test'");
    return TPM_BOTTOM;
  }
  assert(Boa_StringGetInt(tests, "_length", &ntests) == TPM_OK);

  dists = VectorCreate(ntests);
  VectorSet(dists, 1.0/0);

  {ListIter(p, metric, Boa_List_Value(metrics)) {
    if(CallMetric(metric, 
		  Boa_Hold(features), Boa_Hold(tests), Boa_Hold(query), 
		  &result) != TPM_OK) {
      goto error;
    }
    for(i=0;i<ntests;i++) {
      if(Boa_Vector_Value(result)->d[i] < dists->d[i])
	dists->d[i] = Boa_Vector_Value(result)->d[i];
    }
    Boa_Release(result);
  }}

  Boa_Release(metrics);
  Boa_Release(features);
  Boa_Release(tests);
  Boa_Release(query);
  *value_return = Boa_Vector_Create(dists);
  return TPM_OK;

 error:
  VectorFree(dists);
  Boa_Release(metrics);
  Boa_Release(features);
  Boa_Release(tests);
  Boa_Release(query);
  return TPM_BOTTOM;
}

/* SymmetricMax **************************************************************/

/* Produces a distance list for a set of members relative to a query.
 * Inputs: features (context), query (string), test (array of string), 
 *         metric (transforms (f,q,t) -> distance vector)
 * Output: vector of distances
 */
int SymmetricMax(Obj context, Type *type_in_out, Obj *value_return)
{
  Obj metric, features, tests, query, result;
  Obj query_test;
  int ntests, i, query_pos;
  Vector dists;
  double d;

  BOA_TYPECHECK_GET(T_Vector);
  TRY_RETURN(Boa_StringGet(context, "metric", 0, NULL, &metric));
  TRY_RETURN(Boa_StringGet(context, "features", 0, NULL, &features));
  if(Boa_StringGet(context, "query", 0, NULL, &query) != TPM_OK) {
    ErrorPrepend("no value for `query'");
    return TPM_BOTTOM;
  }
  if(Boa_StringGet(context, "test", 0, NULL, &tests) != TPM_OK) {
    ErrorPrepend("no value for `test'");
    return TPM_BOTTOM;
  }
  assert(Boa_StringGetInt(tests, "_length", &ntests) == TPM_OK);

  /* first get the distances from tests to query */
  if(CallMetric(metric, 
		Boa_Hold(features), Boa_Hold(tests), Boa_Hold(query), 
		&result) != TPM_OK) {
    goto error;
  }
  /* we're going to mutate the result */
  dists = Boa_Vector_Value(result);

  /* get the distances from query to tests */
#if 1
  query_test = Boa_Create(T_List);
  assert(Boa_IntSet(query_test, 0, BOA_INSERT, query) == TPM_OK);
  query = query_test;
  query_pos = 0;
#else
  /* must use this route if decorating a rank metric; slow! */
  /* find the query in tests */
  d = 1.0/0;
  for(i=0;i<ntests;i++) {
    if(dists->d[i] < d) { d = dists->d[i]; query_pos = i; }
  }
  Boa_Release(query);
  query = Boa_Hold(tests);
#endif
  for(i=0;i<ntests;i++) {
    Obj test, t;
    assert(Boa_IntGet(tests, i, 0, NULL, &test) == TPM_OK);
    if(CallMetric(metric, 
		  Boa_Hold(features), Boa_Hold(query), test,
		  &t) != TPM_OK) {
      Boa_Release(result);
      goto error;
    }
    d = Boa_Vector_Value(t)->d[query_pos];
    Boa_Release(t);
    if(d > dists->d[i]) dists->d[i] = d;
  }

  Boa_Release(metric);
  Boa_Release(features);
  Boa_Release(tests);
  Boa_Release(query);
  *value_return = result;
  return TPM_OK;

 error:
  Boa_Release(metric);
  Boa_Release(features);
  Boa_Release(tests);
  Boa_Release(query);
  return TPM_BOTTOM;
}

/* SymmetricMin **************************************************************/

/* Same as above except for one comparison */
/* Produces a distance list for a set of members relative to a query.
 * Inputs: features (context), query (string), test (array of string), 
 *         metric (transforms (f,q,t) -> distance vector)
 * Output: vector of distances
 */
int SymmetricMin(Obj context, Type *type_in_out, Obj *value_return)
{
  Obj metric, features, tests, query, result;
  Obj query_test;
  int ntests, i;
  Vector dists;

  BOA_TYPECHECK_GET(T_Vector);
  TRY_RETURN(Boa_StringGet(context, "metric", 0, NULL, &metric));
  TRY_RETURN(Boa_StringGet(context, "features", 0, NULL, &features));
  if(Boa_StringGet(context, "query", 0, NULL, &query) != TPM_OK) {
    ErrorPrepend("no value for `query'");
    return TPM_BOTTOM;
  }
  if(Boa_StringGet(context, "test", 0, NULL, &tests) != TPM_OK) {
    ErrorPrepend("no value for `test'");
    return TPM_BOTTOM;
  }
  assert(Boa_StringGetInt(tests, "_length", &ntests) == TPM_OK);

  /* first get the distances from tests to query */
  if(CallMetric(metric, 
		Boa_Hold(features), Boa_Hold(tests), Boa_Hold(query), 
		&result) != TPM_OK) {
    goto error;
  }
  /* we're going to mutate the result */
  dists = Boa_Vector_Value(result);

  /* get the distances from query to tests */
  query_test = Boa_Create(T_List);
  assert(Boa_IntSet(query_test, 0, BOA_INSERT, query) == TPM_OK);
  query = query_test;
  for(i=0;i<ntests;i++) {
    double d;
    Obj test, t;
    assert(Boa_IntGet(tests, i, 0, NULL, &test) == TPM_OK);
    if(CallMetric(metric, 
		  Boa_Hold(features), Boa_Hold(query), test,
		  &t) != TPM_OK) {
      Boa_Release(result);
      goto error;
    }
    d = Boa_Vector_Value(t)->d[0];
    Boa_Release(t);
    if(d < dists->d[i]) dists->d[i] = d;
  }

  Boa_Release(metric);
  Boa_Release(features);
  Boa_Release(tests);
  Boa_Release(query);
  *value_return = result;
  return TPM_OK;

 error:
  Boa_Release(metric);
  Boa_Release(features);
  Boa_Release(tests);
  Boa_Release(query);
  return TPM_BOTTOM;
}
