#include <tpm/list.h>
#include <boa/boa.h>
#include <boa/vector.h>

#define VERBOSE 0
#define BLOCK 1

typedef double DistFunc(Obj query, Obj test);
typedef double MeanICovDistFunc(Obj query_mean, Obj test_mean, Obj query_icov);
typedef double MeanCovDistFunc(Obj query_mean, Obj test_mean, 
			       Obj query_icov, Obj test_icov,
			       Obj query_cov, Obj test_cov);

/* Prototypes ****************************************************************/

int Intersection(Obj context, Type *type_in_out, Obj *value_return);
int Euclidean(Obj context, Type *type_in_out, Obj *value_return);
int Mahalanobis(Obj context, Type *type_in_out, Obj *value_return);
int Divergence(Obj context, Type *type_in_out, Obj *value_return);
int HierDistance(Obj context, Type *type_in_out, Obj *value_return);
int SpanTreeDistance(Obj context, Type *type_in_out, Obj *value_return);
int DistrDistance(Obj context, Type *type_in_out, Obj *value_return);
int MixtureDistance(Obj context, Type *type_in_out, Obj *value_return);

/* Functions *****************************************************************/

/* used by AttrDistance */
static Obj GetAttr(Obj attrs, char *name, char *attr)
{
  Obj result;
  if(Boa_StringGet(attrs, name, '/', NULL, &result)) {
    ErrorPrepend("%s has no attribute %s", name, attr);
    return NULL;
  }
  return result;
}

/* Produces a distance list for a set of members relative to a query.
 * Inputs: features (context), query (string), test (array of string)
 *         attribute (string)
 * Output: vector of distances
 */
static int AttrDistance(Obj context, DistFunc *proc, 
			Type *type_in_out, Obj *value_return)
{
  Obj attrs, tests, member, t, test;
  Obj query_attr, test_attr;
  char *query, *attr;
  Vector dists;
  int ntests, i, j;

  BOA_TYPECHECK_GET(T_Vector);
  TRY_RETURN(Boa_StringGet(context, "attribute", 0, &T_String, &t));
  attr = Boa_String_Value(t);
  Boa_Release(t);  /* this is safe since attr persists */
  TRY_RETURN(Boa_StringGet(context, "features", 0, NULL, &attrs));
  TRY_RETURN(Boa_StringGet(context, "query", 0, &T_String, &t));
  query = Boa_String_Value(t);
  Boa_Release(t);
  TRY_RETURN(Boa_StringGet(context, "test", 0, NULL, &tests));
  TRY_RETURN(Boa_StringGetInt(tests, "_length", &ntests));

  dists = VectorCreate(ntests);
 
  if(VERBOSE) fprintf(stderr, "attribute: %s\nquery: %s\n", attr, query);
  /* attrs/attr -> attrs */
  if(Boa_StringGet(attrs, attr, '/', NULL, &t) != TPM_OK) {
    fprintf(stderr, "database has no attribute %s\n", attr);
    goto error;
  }
  Boa_Release(attrs); attrs = t;
  /* get the query attribute */
  if(!(query_attr = GetAttr(attrs, query, attr))) goto error;

  /* try to get all test attributes */
  /* if this fails halfway through, there will be a lot of wasted work */
  if(BLOCK && (Boa_ObjGet(attrs, tests, '/', &T_List, &t) == TPM_OK)) {
    if(VERBOSE) fprintf(stderr, "block\n");
    i = 0;
    {ListIter(p, test_attr, Boa_List_Value(t)) {
      errno = 0;
      dists->d[i] = proc(query_attr, test_attr);
      if((errno == EDOM) || !finite(dists->d[i])) {
	assert(Boa_IntGet(tests, i, 0, NULL, &test) == TPM_OK);
	Boa_Print(stderr, test);
	Boa_Release(test);
	fprintf(stderr, " %g\n", dists->d[i]);
	dists->d[i] = 1e100;
      }
      i++;
    }}
    Boa_Release(t);
  }
  else {
    for(i=0;i<ntests;i++) {
      /* get the test attribute */
      assert(Boa_IntGet(tests, i, 0, &T_String, &test) == TPM_OK);
      test_attr = GetAttr(attrs, Boa_String_Value(test), attr);
      Boa_Release(test);
      if(!test_attr) {
	ErrorPrint(stderr);
	dists->d[i] = 1e100;
	continue;
      }
      errno = 0;
      dists->d[i] = proc(query_attr, test_attr);
      if((errno == EDOM) || !finite(dists->d[i])) {
	assert(Boa_IntGet(tests, i, 0, NULL, &test) == TPM_OK);
	Boa_Print(stderr, test);
	Boa_Release(test);
	fprintf(stderr, " %g\n", dists->d[i]);
	dists->d[i] = 1e100;
      }
      Boa_Release(test_attr);
    }
  }
  Boa_Release(query_attr);

  Boa_Release(tests);
  Boa_Release(attrs);
  *value_return = Boa_Vector_Create(dists);
  return TPM_OK;

 error:
  VectorFree(dists);
  Boa_Release(tests);
  Boa_Release(attrs);
  return TPM_ERROR;
}

/* Swain-style histogram intersection ****************************************/
static double VectorIntersect(Vector query, Vector test)
{
  double d = 0;
  int i;
  for(i=0;i<query->len;i++) {
    /* nominal penalty for having too much */
    if(query->d[i] < test->d[i]) d += 0.01;
    else d += query->d[i] - test->d[i];
  }
  return d;
}
/* query and test are arrays of int */
static double Intersect(Obj query, Obj test)
{
  assert(query->type == T_Vector);
  assert(test->type == T_Vector);
  assert(Boa_Vector_Value(query)->len == Boa_Vector_Value(test)->len);
  return VectorIntersect(Boa_Vector_Value(query), Boa_Vector_Value(test));
}
static double SymIntersect(Obj query, Obj test)
{
  double d1 = Intersect(query, test);
  double d2 = Intersect(test, query);
  if(d1 < d2) return d1;
  else return d2;
}
int Intersection(Obj context, Type *type_in_out, Obj *value_return)
{
  return AttrDistance(context, Intersect, type_in_out, value_return);
}

/* Euclidean distance ********************************************************/
/* query and test are vectors */
static double Euclid(Obj query, Obj test)
{
  assert(query->type == T_Vector);
  assert(test->type == T_Vector);
  assert(Boa_Vector_Value(query)->len == Boa_Vector_Value(test)->len);
  return VectorDistance(Boa_Vector_Value(query), Boa_Vector_Value(test));
}
int Euclidean(Obj context, Type *type_in_out, Obj *value_return)
{
  return AttrDistance(context, Euclid, type_in_out, value_return);
}

/* Mahalanobis distance ******************************************************/

/* Just like AttrDistance, but with mean, icov features */
/* Produces a distance list for a set of members relative to a query.
 * Inputs: features (context), query (string), test (array of string)
 *         mean (string), icov (string)
 * Output: vector of distances
 */
int MeanICovDistance(Obj context, MeanICovDistFunc *proc,
		     Type *type_in_out, Obj *value_return)
{
  Obj attrs, tests, member, t;
  Obj query_mean, query_icov, test_mean;
  char *query, *mean, *icov;
  Obj test;
  Vector dists;
  int ntests, i, j;

  BOA_TYPECHECK_GET(T_Vector);
  TRY_RETURN(Boa_StringGet(context, "mean", 0, &T_String, &t)); 
  mean = Boa_String_Value(t);
  Boa_Release(t);
  TRY_RETURN(Boa_StringGet(context, "icov", 0, &T_String, &t));
  icov = Boa_String_Value(t);
  Boa_Release(t);
  TRY_RETURN(Boa_StringGet(context, "features", 0, NULL, &attrs));
  TRY_RETURN(Boa_StringGet(context, "query", 0, &T_String, &t));
  query = Boa_String_Value(t);
  Boa_Release(t);
  TRY_RETURN(Boa_StringGet(context, "test", 0, NULL, &tests));
  assert(Boa_StringGetInt(tests, "_length", &ntests) == TPM_OK);

  dists = VectorCreate(ntests);

  if(VERBOSE) fprintf(stderr, "query: %s\n", query);
  /* get the query attributes */
  if(Boa_StringGet(attrs, icov, '/', NULL, &t) != TPM_OK) {
    fprintf(stderr, "database has no attribute %s\n", icov);
    goto error;
  }
  if(!(query_icov = GetAttr(t, query, icov))) goto error;
  Boa_Release(t);
  /* attrs/mean -> attrs */
  if(Boa_StringGet(attrs, mean, '/', NULL, &t) != TPM_OK) {
    fprintf(stderr, "database has no attribute %s\n", mean);
    goto error;
  }
  Boa_Release(attrs); attrs = t;
  if(!(query_mean = GetAttr(attrs, query, mean))) {
    Boa_Release(query_icov);
    goto error;
  }

  if(BLOCK && (Boa_ObjGet(attrs, tests, '/', &T_List, &t) == TPM_OK)) {
    if(VERBOSE) fprintf(stderr, "block\n");
    i = 0;
    {ListIter(p, test_mean, Boa_List_Value(t)) {
      errno = 0;
      dists->d[i] = proc(query_mean, test_mean, query_icov);
      if((errno == EDOM) || !finite(dists->d[i])) {
	assert(Boa_IntGet(tests, i, 0, NULL, &test) == TPM_OK);
	Boa_Print(stderr, test);
	Boa_Release(test);
	fprintf(stderr, " %g\n", dists->d[i]);
	dists->d[i] = 1e100;
      }
      i++;
    }}
    Boa_Release(t);
  }
  else {
    for(i=0;i<ntests;i++) {
      /* get the test attribute */
      assert(Boa_IntGet(tests, i, 0, &T_String, &test) == TPM_OK);
      test_mean = GetAttr(attrs, Boa_String_Value(test), mean);
      Boa_Release(test);
      if(!test_mean) {
	dists->d[i] = 1e100;
	continue;
      }
      errno = 0;
      dists->d[i] = proc(query_mean, test_mean, query_icov);
      if((errno == EDOM) || !finite(dists->d[i])) {
	assert(Boa_IntGet(tests, i, 0, NULL, &test) == TPM_OK);
	Boa_Print(stderr, test);
	Boa_Release(test);
	fprintf(stderr, " %g\n", dists->d[i]);
	dists->d[i] = 1e100;
      }
      Boa_Release(test_mean);
    }
  }
  Boa_Release(query_mean);
  Boa_Release(query_icov);

  Boa_Release(tests);
  Boa_Release(attrs);
  *value_return = Boa_Vector_Create(dists);
  return TPM_OK;

 error:
  VectorFree(dists);
  Boa_Release(tests);
  Boa_Release(attrs);
  return TPM_ERROR;
}

static double Mahal(Obj query_mean, Obj test_mean, Obj query_icov)
{
  assert(query_mean->type == T_Vector);
  assert(test_mean->type == T_Vector);
  assert(query_icov->type == T_Matrix);
  assert(Boa_Vector_Value(query_mean)->len == Boa_Vector_Value(test_mean)->len);
  assert(Boa_Vector_Value(query_mean)->len == Boa_Matrix_Value(query_icov)->rows);
  assert(Boa_Vector_Value(query_mean)->len == Boa_Matrix_Value(query_icov)->cols);
  return VectorMahalanobis(Boa_Vector_Value(query_mean), 
			   Boa_Vector_Value(test_mean), 
			   Boa_Matrix_Value(query_icov));
}

int Mahalanobis(Obj context, Type *type_in_out, Obj *value_return)
{
  return MeanICovDistance(context, Mahal, type_in_out, value_return);
}

static double MahalDiag(Obj query_mean, Obj test_mean, Obj query_icov)
{
  double d;
  Vector diag = MatrixDiagonal(Boa_Matrix_Value(query_icov));
  assert(query_mean->type == T_Vector);
  assert(test_mean->type == T_Vector);
  assert(query_icov->type == T_Matrix);
  assert(Boa_Vector_Value(query_mean)->len == Boa_Vector_Value(test_mean)->len);
  assert(Boa_Vector_Value(query_mean)->len == Boa_Matrix_Value(query_icov)->rows);
  assert(Boa_Vector_Value(query_mean)->len == Boa_Matrix_Value(query_icov)->cols);
  d = VectorWeightedDistance(Boa_Vector_Value(query_mean), 
			     Boa_Vector_Value(test_mean), 
			     diag);
  VectorFree(diag);
  return d;
}

int MahalanobisDiag(Obj context, Type *type_in_out, Obj *value_return)
{
  return MeanICovDistance(context, MahalDiag, type_in_out, value_return);
}

#ifndef NO_LAPACK
static double MahalEV(Obj query_mean, Obj test_mean, Obj query_icov)
{
  assert(query_mean->type == T_Vector);
  assert(test_mean->type == T_Vector);
  assert(query_icov->type == T_Matrix);
  assert(Boa_Vector_Value(query_mean)->len == Boa_Vector_Value(test_mean)->len);
  assert(Boa_Vector_Value(query_mean)->len == Boa_Matrix_Value(query_icov)->rows);
  assert(Boa_Vector_Value(query_mean)->len == Boa_Matrix_Value(query_icov)->cols);
  return VectorMahalanobisEv(Boa_Vector_Value(query_mean), 
			     Boa_Vector_Value(test_mean), 
			     Boa_Matrix_Value(query_icov));
}

int MahalanobisEV(Obj context, Type *type_in_out, Obj *value_return)
{
  return MeanICovDistance(context, MahalEV, type_in_out, value_return);
}
#endif

/* Divergence distance *******************************************************/

/* Just like AttrDistance, but with mean, icov, cov features */
/* Produces a distance list for a set of members relative to a query.
 * Inputs: features (context), query (string), test (array of string)
 *         mean (string), icov (string), cov (string)
 * Output: vector of distances
 */
int MeanCovDistance(Obj context, MeanCovDistFunc *proc,
		    Type *type_in_out, Obj *value_return)
{
  Obj attrs, tests, member, t;
  Obj attr_mean, attr_icov, attr_cov;
  Obj query_mean, query_icov, query_cov;
  Obj test_mean, test_icov, test_cov;
  Obj means, icovs, covs;
  char *query, *mean, *icov, *cov;
  Obj test;
  Vector dists;
  int ntests, i, j;

  BOA_TYPECHECK_GET(T_Vector);
  TRY_RETURN(Boa_StringGet(context, "mean", 0, &T_String, &t)); 
  mean = Boa_String_Value(t);
  Boa_Release(t);
  TRY_RETURN(Boa_StringGet(context, "icov", 0, &T_String, &t));
  icov = Boa_String_Value(t);
  Boa_Release(t);
  if(Boa_StringGet(context, "cov", 0, &T_String, &t) == TPM_OK) {
    cov = Boa_String_Value(t);
    Boa_Release(t);
  } 
  else cov = NULL;
  TRY_RETURN(Boa_StringGet(context, "features", 0, NULL, &attrs));
  TRY_RETURN(Boa_StringGet(context, "query", 0, &T_String, &t));
  query = Boa_String_Value(t);
  Boa_Release(t);
  TRY_RETURN(Boa_StringGet(context, "test", 0, NULL, &tests));
  TRY_RETURN(Boa_StringGetInt(tests, "_length", &ntests));

  dists = VectorCreate(ntests);

  if(Boa_StringGet(attrs, mean, '/', NULL, &attr_mean) != TPM_OK) {
    fprintf(stderr, "database has no attribute %s\n", mean);
    goto error;
  }
  if(Boa_StringGet(attrs, icov, '/', NULL, &attr_icov) != TPM_OK) {
    fprintf(stderr, "database has no attribute %s\n", icov);
    goto error;
  }
  if(cov) {
    if(Boa_StringGet(attrs, cov, '/', NULL, &attr_cov) != TPM_OK) {
      fprintf(stderr, "database has no attribute %s\n", cov);
      goto error;
    }
  }
  Boa_Release(attrs);

  if(VERBOSE) fprintf(stderr, "query: %s\n", query);
  /* get the query attributes */
  if(cov) {
    if(!(query_cov = GetAttr(attr_cov, query, cov))) goto error;
  }
  if(!(query_icov = GetAttr(attr_icov, query, icov))) {
    Boa_Release(query_cov);
    goto error;
  }
  if(!(query_mean = GetAttr(attr_mean, query, mean))) {
    Boa_Release(query_cov);
    Boa_Release(query_icov);
    goto error;
  }

  if(BLOCK && 
     (Boa_ObjGet(attr_mean, tests, '/', &T_List, &means) == TPM_OK) &&
     (Boa_ObjGet(attr_icov, tests, '/', &T_List, &icovs) == TPM_OK) &&
     (!cov || (Boa_ObjGet(attr_cov, tests, '/', &T_List, &covs) == TPM_OK))) {
    ListNode pm, pi, pc;
    if(VERBOSE) fprintf(stderr, "block\n");
    i = 0;
    if(cov) pc = ListFront(Boa_List_Value(covs));
    else pc = ListFront(Boa_List_Value(icovs));
    for(pm = ListFront(Boa_List_Value(means)), 
	pi = ListFront(Boa_List_Value(icovs));
	pm && pi && pc;
	pm = pm->next, pi = pi->next, pc = pc->next) {
      test_mean = (Obj)pm->data;
      test_icov = (Obj)pi->data;
      test_cov  = (Obj)pc->data;
      errno = 0;
      dists->d[i] = proc(query_mean, test_mean, query_icov, test_icov,
			 query_cov, test_cov);
      if((errno == EDOM) || 
	 (dists->d[i] != dists->d[i]) ||
	 !finite(dists->d[i])) {
	assert(Boa_IntGet(tests, i, 0, NULL, &test) == TPM_OK);
	Boa_Print(stderr, test);
	Boa_Release(test);
	fprintf(stderr, " %g\n", dists->d[i]);
	dists->d[i] = 1e100;
      }
      i++;
    }
    Boa_Release(means);
    Boa_Release(icovs);
    if(cov) Boa_Release(covs);
  }
  else {
    for(i=0;i<ntests;i++) {
      /* get the test attributes */
      assert(Boa_IntGet(tests, i, 0, &T_String, &test) == TPM_OK);
      test_mean = GetAttr(attr_mean, Boa_String_Value(test), mean);
      test_icov = GetAttr(attr_icov, Boa_String_Value(test), icov);
      if(cov) test_cov = GetAttr(attr_cov, Boa_String_Value(test), cov);
      Boa_Release(test);
      if(!test_mean || (cov && !test_icov) || !test_cov) {
	dists->d[i] = 1e100;
	continue;
      }
      errno = 0;
      dists->d[i] = proc(query_mean, test_mean, query_icov, test_icov, 
			 query_cov, test_cov);
      if((errno == EDOM) || 
	 (dists->d[i] != dists->d[i]) ||
	 !finite(dists->d[i])) {
	assert(Boa_IntGet(tests, i, 0, NULL, &test) == TPM_OK);
	Boa_Print(stderr, test);
	Boa_Release(test);
	fprintf(stderr, " %g\n", dists->d[i]);
	dists->d[i] = 1e100;
      }
      Boa_Release(test_mean);
      Boa_Release(test_icov);
      if(cov) Boa_Release(test_cov);
    }
  }
  Boa_Release(query_mean);
  Boa_Release(query_icov);
  if(cov) Boa_Release(query_cov);

  Boa_Release(tests);
  Boa_Release(attr_mean);
  Boa_Release(attr_icov);
  if(cov) Boa_Release(attr_cov);
  *value_return = Boa_Vector_Create(dists);
  return TPM_OK;

 error:
  VectorFree(dists);
  Boa_Release(tests);
  Boa_Release(attr_mean);
  Boa_Release(attr_icov);
  if(cov) Boa_Release(attr_cov);
  return TPM_ERROR;
}

static double Diver(Obj query_mean, Obj test_mean, 
		    Obj query_icov, Obj test_icov,
		    Obj query_cov, Obj test_cov)
{
  assert(query_mean->type == T_Vector);
  assert(test_mean->type == T_Vector);
  assert(query_icov->type == T_Matrix);
  assert(test_icov->type == T_Matrix);
  assert(query_cov->type == T_Matrix);
  assert(test_cov->type == T_Matrix);
  assert(Boa_Vector_Value(query_mean)->len == Boa_Vector_Value(test_mean)->len);
  assert(Boa_Vector_Value(query_mean)->len == Boa_Matrix_Value(query_icov)->rows);
  assert(Boa_Vector_Value(query_mean)->len == Boa_Matrix_Value(query_icov)->cols);
  assert(Boa_Vector_Value(query_mean)->len == Boa_Matrix_Value(query_cov)->rows);
  assert(Boa_Vector_Value(query_mean)->len == Boa_Matrix_Value(query_cov)->cols);
  return VectorDivergence(Boa_Vector_Value(query_mean), 
			  Boa_Vector_Value(test_mean), 
			  Boa_Matrix_Value(query_icov),
			  Boa_Matrix_Value(test_icov),
			  Boa_Matrix_Value(query_cov),
			  Boa_Matrix_Value(test_cov));
}

int Divergence(Obj context, Type *type_in_out, Obj *value_return)
{
  return MeanCovDistance(context, Diver, type_in_out, value_return);
}

#ifndef NO_LAPACK
static double DiverEV(Obj query_mean, Obj test_mean, 
		      Obj query_icov, Obj test_icov,
		      /* not used */
		      Obj query_cov, Obj test_cov)
{
  assert(query_mean->type == T_Vector);
  assert(test_mean->type == T_Vector);
  assert(query_icov->type == T_Matrix);
  assert(test_icov->type == T_Matrix);
  assert(Boa_Vector_Value(query_mean)->len == Boa_Vector_Value(test_mean)->len);
  assert(Boa_Vector_Value(query_mean)->len == Boa_Matrix_Value(query_icov)->rows);
  assert(Boa_Vector_Value(query_mean)->len == Boa_Matrix_Value(query_icov)->cols);
  return VectorDivergenceEv(Boa_Vector_Value(query_mean), 
			    Boa_Vector_Value(test_mean), 
			    Boa_Matrix_Value(query_icov),
			    Boa_Matrix_Value(test_icov));
}

int DivergenceEV(Obj context, Type *type_in_out, Obj *value_return)
{
  return MeanCovDistance(context, DiverEV, type_in_out, value_return);
}
#endif

/* Hierarchy *****************************************************************/

#include <tpm/hier.h>

#define RANDOM 0

/* Produces a distance list for a set of members relative to a query.
 * In this case, distance is cophenetic proximity as defined by Jain+Dubes.
 * Inputs: query (string), test (array of string), 
 *         hier (Hier), index (converts name to node number)
 * Output: vector of distances
 */
int HierDistance(Obj context, Type *type_in_out, Obj *value_return)
{
  Obj t, hierObj, tests, index;
  int ntests, i, j, left, right, dist, *d;
  Obj query, test;
  Vector dists;
  Hier hier;

  BOA_TYPECHECK_GET(T_Vector);
  if(Boa_StringGet(context, "hier", 0, NULL, &hierObj) != TPM_OK) {
    ErrorPrepend("no value for `hier'");
    return TPM_BOTTOM;
  }
  if(hierObj->type->type != T_Hier) {
    ErrorSet("wrong type for `hier'");
    return TPM_BOTTOM;
  }
  hier = Boa_Hier_Value(hierObj);
  if(Boa_StringGet(context, "index", 0, NULL, &index) != TPM_OK) {
    ErrorPrepend("no value for `index'");
    return TPM_BOTTOM;
  }
  if(Boa_StringGet(context, "query", 0, NULL, &query) != TPM_OK) {
    ErrorPrepend("no value for `query'");
    return TPM_BOTTOM;
  }
  if(Boa_StringGet(context, "test", 0, NULL, &tests) != TPM_OK) {
    ErrorPrepend("no value for `test'");
    return TPM_BOTTOM;
  }
  assert(Boa_StringGetInt(tests, "_length", &ntests) == TPM_OK);

  dists = VectorCreate(ntests);
  /* temporary storage */
  d = Allocate(hier->num_nodes, int);

  if(VERBOSE) {
    fprintf(stderr, "query: ");
    Boa_PrintNL(stderr, query);
  }

  /* find the node corresponding to the query */
  if(Boa_ObjGetInt(index, query, &i) != TPM_OK) {
    char *s = Boa_AsString(query);
    ErrorPrepend("No such member %s for query", s);
    free(s);
    goto error;
  }
  Boa_Release(query);

  /* if we didn't store the distance at each merge, we can approximate
   * by the height of the merge.
   */
  /* compute the height of all nodes */
  memset(d, 0, hier->num_nodes*sizeof(int));
  for(j=hier->num_nodes-1;j;j--) {
    int p = HierNodeAt(hier, j)->parent;
    if(d[p] < d[j]+1) d[p] = d[j]+1;
  }

  /* climb the tree and assign distances to nodes */
  left = right = i;
  for(dist=1;;dist++) {
    int limit;
    d[i] = 0;
    i = HierNodeAt(hier, i)->parent;
    if(i < 0) break;
    limit = HierRightLeaf(hier, i);
    for(j=i+1;j<=limit;j++) {
      if((j >= left) && (j <= right)) continue;
      d[j] = d[i];
    }
    left = i;
    right = limit;
  }
  /* check for nodes under diff't roots */
  for(j=0;j<hier->num_nodes;j++) {
    if((j >= left) || (j <= right)) continue;
    d[j] = INT_MAX;
  }

  /* look up all tests and assign distances */
  if(BLOCK && (Boa_ObjGet(index, tests, '/', &T_List, &t) == TPM_OK)) {
    if(VERBOSE) fprintf(stderr, "block\n");
    i = 0;
    {Obj pos; ListIter(p, pos, Boa_List_Value(t)) {
      dists->d[i] = (double)d[Boa_Int_Value(pos)];
#if RANDOM
      dists->d[i] += RandReal()/1000;
#endif
      i++;
    }}
  }
  else {
    for(i=0;i<ntests;i++) {
      assert(Boa_IntGet(tests, i, 0, NULL, &test) == TPM_OK);
      if(Boa_ObjGetInt(index, test, &j) != TPM_OK) {
	char *s = Boa_AsString(test);
	ErrorPrepend("No such member %s for test", s);
	free(s);
	Boa_Release(test);
	goto error;
      }
      Boa_Release(test);
      dists->d[i] = (double)d[j];
#if RANDOM
      dists->d[i] += RandReal()/1000;
#endif
    }
  }

  free(d);
  Boa_Release(hierObj);
  Boa_Release(index);
  Boa_Release(tests);
  *value_return = Boa_Vector_Create(dists);
  return TPM_OK;

 error:
  free(d);
  Boa_Release(hierObj);
  Boa_Release(index);
  Boa_Release(tests);
  VectorFree(dists);
  return TPM_BOTTOM;
}

/* SpanTree ******************************************************************/

#include <tpm/spanTree.h>

/* Returns the cost of i using the auxiliary array d. */
static float SpanTreeDistanceHelper(SpanTree tree, float *d, int i)
{
  if(d[i] == NAI) {
    d[i] = tree->node[i].edge + 
           SpanTreeDistanceHelper(tree, d, tree->node[i].parent);
  }
  return d[i];
}

/* Produces a distance list for a set of members relative to a query.
 * In this case, distance is path cost on a spanning tree.
 * Inputs: query (string), test (array of string), 
 *         tree (SpanTree), index (converts name to node number)
 * Output: vector of distances
 */
int SpanTreeDistance(Obj context, Type *type_in_out, Obj *value_return)
{
  Obj t, treeObj, tests, index;
  int ntests, i, j;
  char *query;
  Obj test;
  Vector dists;
  float *d, f;
  SpanTree tree;

  BOA_TYPECHECK_GET(T_Vector);
  if(Boa_StringGet(context, "tree", 0, NULL, &treeObj) != TPM_OK) {
    ErrorPrepend("no value for `tree'");
    return TPM_BOTTOM;
  }
  if(treeObj->type->type != T_SpanTree) {
    ErrorSet("wrong type for `tree'");
    return TPM_BOTTOM;
  }
  tree = Boa_SpanTree_Value(treeObj);
  if(Boa_StringGet(context, "index", 0, NULL, &index) != TPM_OK) {
    ErrorPrepend("no value for `index'");
    return TPM_BOTTOM;
  }
  if(Boa_StringGet(context, "query", 0, &T_String, &t) != TPM_OK) {
    ErrorPrepend("no value for `query'");
    return TPM_BOTTOM;
  }
  query = Boa_String_Value(t);
  Boa_Release(t);
  if(Boa_StringGet(context, "test", 0, NULL, &tests) != TPM_OK) {
    ErrorPrepend("no value for `test'");
    return TPM_BOTTOM;
  }
  assert(Boa_StringGetInt(tests, "_length", &ntests) == TPM_OK);

  dists = VectorCreate(ntests);
  /* temporary storage */
  d = Allocate(tree->num_nodes, float);

  if(VERBOSE) fprintf(stderr, "query: %s\n", query);

  /* find the node corresponding to the query */
  if(Boa_StringGetInt(index, query, &i) != TPM_OK) {
    ErrorPrepend("No such member `%s'", query);
    goto error;
  }

  /* d is a map which gives the distance from i to its ancestors */
  for(j=0;j<tree->num_nodes;j++) {
    d[j] = NAI;
  }
  /* compute costs going up */
  f = 0;
  j = i;
  while(j >= 0) {
    d[j] = f;
    f += tree->node[j].edge;
    j = tree->node[j].parent;
  }

  /* look up all tests and assign distances */
  if(BLOCK && (Boa_ObjGet(index, tests, '/', &T_List, &t) == TPM_OK)) {
    if(VERBOSE) fprintf(stderr, "block\n");
    i = 0;
    {Obj pos; ListIter(p, pos, Boa_List_Value(t)) {
      dists->d[i] = SpanTreeDistanceHelper(tree, d, Boa_Int_Value(pos));
      i++;
    }}
  }
  else {
    for(i=0;i<ntests;i++) {
      assert(Boa_IntGet(tests, i, 0, &T_String, &test) == TPM_OK);
      if(Boa_StringGetInt(index, Boa_String_Value(test), &j) != TPM_OK) {
	ErrorPrepend("No such member `%s'", Boa_String_Value(test));
	Boa_Release(test);
	goto error;
      }
      Boa_Release(test);
      dists->d[i] = SpanTreeDistanceHelper(tree, d, j);
    }
  }

  free(d);
  Boa_Release(treeObj);
  Boa_Release(index);
  Boa_Release(tests);
  *value_return = Boa_Vector_Create(dists);
  return TPM_OK;

 error:
  free(d);
  Boa_Release(treeObj);
  Boa_Release(index);
  Boa_Release(tests);
  VectorFree(dists);
  return TPM_BOTTOM;
}

/* Distribution **************************************************************/

static double RelativeEntropy(Vector p, Vector q, Matrix c)
{
  double d = 0;
  int i,j;
  assert(p->len == q->len);
  assert(p->len == c->rows);
  for(i=0;i<c->cols;i++) {
    double psum = 0, qsum = 0;
    for(j=0;j<c->rows;j++) {
      psum += c->d2[j][i] * p->d[j];
      qsum += c->d2[j][i] * q->d[j];
    }
    if(psum) {
      if(!qsum) d += 1e100;
      else d += psum * log(psum / qsum);
    }
  }
  if(VERBOSE) fprintf(stderr, "KL: %g\n", d);
  return d;
}

/* Produces a distance list for a set of members relative to a query.
 * In this case, distance is relative entropy between distributions,
 * each encoded as a mixture of prototype distributions.
 * Inputs: query (string), test (array of string), features (context),
 *         center (matrix [center][member]), membership (string)
 * Output: vector of distances
 */
int DistrDistance(Obj context, Type *type_in_out, Obj *value_return)
{
  Obj index, t, tests, attrs;
  Obj query_attr, test_attr;
  int ntests, i;
  char *query, *membership;
  Vector dists;
  Matrix center;

  BOA_TYPECHECK_GET(T_Vector);
  if(Boa_StringGet(context, "membership", 0, NULL, &t) != TPM_OK) {
    ErrorPrepend("no value for `membership'");
    return TPM_BOTTOM;
  }
  membership = Boa_String_Value(t);
  Boa_Release(t);
  if(Boa_StringGet(context, "center", 0, &T_Matrix, &t) != TPM_OK) {
    ErrorPrepend("no value for `center'");
    return TPM_BOTTOM;
  }
  center = Boa_Matrix_Value(t);
  Boa_Release(t);
  if(Boa_StringGet(context, "features", 0, NULL, &attrs) != TPM_OK) {
    ErrorPrepend("no value for `features'");
    return TPM_BOTTOM;
  }
  if(Boa_StringGet(context, "query", 0, &T_String, &t) != TPM_OK) {
    ErrorPrepend("no value for `query'");
    return TPM_BOTTOM;
  }
  query = Boa_String_Value(t);
  Boa_Release(t);
  if(Boa_StringGet(context, "test", 0, NULL, &tests) != TPM_OK) {
    ErrorPrepend("no value for `test'");
    return TPM_BOTTOM;
  }
  assert(Boa_StringGetInt(tests, "_length", &ntests) == TPM_OK);

  dists = VectorCreate(ntests);

  /* attrs/membership -> attrs */
  if(Boa_StringGet(attrs, membership, '/', NULL, &t) != TPM_OK) {
    ErrorPrepend("database has no attribute %s", membership);
    goto error;
  }
  Boa_Release(attrs); attrs = t;

  if(VERBOSE) fprintf(stderr, "query: %s\n", query);
  /* get the query attributes */
  if(!(query_attr = GetAttr(attrs, query, membership))) {
    goto error;
  }

  if(BLOCK && (Boa_ObjGet(attrs, tests, '/', &T_List, &t) == TPM_OK)) {
    if(VERBOSE) fprintf(stderr, "block\n");
    i = 0;
    {ListIter(p, test_attr, Boa_List_Value(t)) {
      dists->d[i] = RelativeEntropy(Boa_Vector_Value(query_attr), 
				    Boa_Vector_Value(test_attr), 
				    center);
      i++;
    }}
    Boa_Release(t);
  }
  else {
    for(i=0;i<ntests;i++) {
      Obj test;
      /* get the test attribute */
      assert(Boa_IntGet(tests, i, 0, &T_String, &test) == TPM_OK);
      test_attr = GetAttr(attrs, Boa_String_Value(test), membership);
      Boa_Release(test);
      if(!test_attr) {
	dists->d[i] = 1e100;
	continue;
      }
      dists->d[i] = RelativeEntropy(Boa_Vector_Value(query_attr), 
				    Boa_Vector_Value(test_attr), 
				    center);
      Boa_Release(test_attr);
    }
  }
  Boa_Release(query_attr);

  Boa_Release(tests);
  Boa_Release(attrs);
  *value_return = Boa_Vector_Create(dists);
  return TPM_OK;

 error:
  VectorFree(dists);
  Boa_Release(tests);
  Boa_Release(attrs);
  return TPM_ERROR;
}

/* Mixture *******************************************************************/

#define USE_COV 0

static double MixtureDivergence(Vector p, Vector q, Matrix mean, Matrix cov)
{
  double d = 0;
  int i,c;
  assert(p->len == q->len);
  assert(p->len == mean->rows);
  assert(mean->rows == cov->rows);
  for(i=0;i<mean->cols;i++) {
    double sum = 0, mean_sum = 0, cov_sum = 0;
    for(c=0;c<mean->rows;c++) {
      mean_sum += mean->d2[c][i] * (p->d[c] - q->d[c]);
#if USE_COV
      cov_sum += cov->d2[c][i] * (p->d[c] + q->d[c]);
      sum += p->d[c] + q->d[c];
#endif
    }
#if USE_COV
    if(cov_sum > 0) d += mean_sum * mean_sum / cov_sum * sum;
#else
    d += mean_sum * mean_sum;
#endif
  }
  if(VERBOSE) fprintf(stderr, "distance: %g\n", d);
  return d;
}

/* Produces a distance list for a set of members relative to a query.
 * In this case, distance is relative entropy between distributions,
 * each encoded as a mixture of prototype distributions.
 * Inputs: query (string), test (array of string), features (context),
 *         center (matrix [center][dim]), cov (matrix [center][dim*dim])
 *         membership (string)
 * Output: vector of distances
 */
int MixtureDistance(Obj context, Type *type_in_out, Obj *value_return)
{
  Obj index, t, tests, attrs;
  Obj query_attr, test_attr;
  int ntests, i;
  char *query, *membership;
  Vector dists;
  Matrix center, cov;

  BOA_TYPECHECK_GET(T_Vector);
  if(Boa_StringGet(context, "membership", 0, NULL, &t) != TPM_OK) {
    ErrorPrepend("no value for `membership'");
    return TPM_BOTTOM;
  }
  membership = Boa_String_Value(t);
  Boa_Release(t);
  if(Boa_StringGet(context, "center", 0, &T_Matrix, &t) != TPM_OK) {
    ErrorPrepend("no value for `center'");
    return TPM_BOTTOM;
  }
  center = Boa_Matrix_Value(t);
  Boa_Release(t);
  if(Boa_StringGet(context, "cov", 0, &T_Matrix, &t) != TPM_OK) {
    ErrorPrepend("no value for `cov'");
    return TPM_BOTTOM;
  }
  cov = Boa_Matrix_Value(t);
  Boa_Release(t);
  if(Boa_StringGet(context, "features", 0, NULL, &attrs) != TPM_OK) {
    ErrorPrepend("no value for `features'");
    return TPM_BOTTOM;
  }
  if(Boa_StringGet(context, "query", 0, &T_String, &t) != TPM_OK) {
    ErrorPrepend("no value for `query'");
    return TPM_BOTTOM;
  }
  query = Boa_String_Value(t);
  Boa_Release(t);
  if(Boa_StringGet(context, "test", 0, NULL, &tests) != TPM_OK) {
    ErrorPrepend("no value for `test'");
    return TPM_BOTTOM;
  }
  assert(Boa_StringGetInt(tests, "_length", &ntests) == TPM_OK);

  dists = VectorCreate(ntests);

  /* attrs/membership -> attrs */
  if(Boa_StringGet(attrs, membership, '/', NULL, &t) != TPM_OK) {
    ErrorPrepend("database has no attribute %s", membership);
    goto error;
  }
  Boa_Release(attrs); attrs = t;

  if(VERBOSE) fprintf(stderr, "query: %s\n", query);
  /* get the query attributes */
  if(!(query_attr = GetAttr(attrs, query, membership))) {
    goto error;
  }

  if(BLOCK && (Boa_ObjGet(attrs, tests, '/', &T_List, &t) == TPM_OK)) {
    if(VERBOSE) fprintf(stderr, "block\n");
    i = 0;
    {ListIter(p, test_attr, Boa_List_Value(t)) {
      dists->d[i] = MixtureDivergence(Boa_Vector_Value(query_attr), 
				      Boa_Vector_Value(test_attr), 
				      center, cov);
      i++;
    }}
    Boa_Release(t);
  }
  else {
    for(i=0;i<ntests;i++) {
      Obj test;
      /* get the test attribute */
      assert(Boa_IntGet(tests, i, 0, &T_String, &test) == TPM_OK);
      test_attr = GetAttr(attrs, Boa_String_Value(test), membership);
      Boa_Release(test);
      if(!test_attr) {
	dists->d[i] = 1e100;
	continue;
      }
      dists->d[i] = MixtureDivergence(Boa_Vector_Value(query_attr), 
				      Boa_Vector_Value(test_attr), 
				      center, cov);
      Boa_Release(test_attr);
    }
  }
  Boa_Release(query_attr);

  Boa_Release(tests);
  Boa_Release(attrs);
  *value_return = Boa_Vector_Create(dists);
  return TPM_OK;

 error:
  VectorFree(dists);
  Boa_Release(tests);
  Boa_Release(attrs);
  return TPM_ERROR;
}
