/* A Seg is represented in Tcl as a list of 8 integers.
 * A polygon is a list of <x,y> pairs
 * Commands:
 *   poly2seg coords		Convert polygon coords to a Seg
 *   seg2poly seg		Convert Seg into one polygon
 *   seg2polys seg		Convert Seg into multiple polygons
 *   readsegs obj arrayvar	Convert a Seg Hier into a Tcl array
 *   tcl_seg obj		Convert a Seg into a Tcl value
 *   showseg canvas seg color   Display a Seg on a canvas
 *   setdb name			Load a database
 *   member name		Returns the index of a member
 *
 *   within init segcon segs	begin within-image learning
 *   within get			get the learned Seg
 *   within pos|neg canvas x y	add an example at the canvas coords
 *   within clear               forget all examples
 *   within destroy		end within-image learning
 *
 *   across init across across_list within within_list	
 *                                              begin across-image learning
 *   across get mem-index label		        get the learned label
 *   across pos|neg mem-index label seg		add an example
 *   across destroy				end across-image learning
 *   across enable across flag                  enable/disable a tree
 */
#include <tcl.h>
#include "tcl_macros.h"
#include <tk.h> /* for Tk_Width */
#include "photobook.h"

extern Boa_DefineFunc Boa_SegMap_Define;

#define PACKINT(x, i) \
  ((unsigned)x[i]<<24) + ((unsigned)x[i+1]<<16) + \
  ((unsigned)x[i+2]<<8) + (unsigned)x[i+3]
#define SegString(str, seg) \
  sprintf(str, "%d %d %d %d %d %d %d %d",             \
	  PACKINT((seg).mask,0), PACKINT((seg).mask,4),   \
	  PACKINT((seg).mask,8), PACKINT((seg).mask,12),  \
	  PACKINT((seg).mask,16), PACKINT((seg).mask,20), \
	  PACKINT((seg).mask,24), PACKINT((seg).mask,28));
#define ReturnSeg(seg) SegString(interp->result, seg)

static Obj member_index = NULL;

/* Prototypes ****************************************************************/
int PhCmd(ClientData clientData, Tcl_Interp *interp,
	  int argc, char *argv[]);
int Phtcl_Init(Tcl_Interp *interp);

static int StringGets(Tcl_Interp *interp, char *str, List *result);
static int ObjGets(Tcl_Interp *interp, char *str, List *result);
static int GetSeg(Tcl_Interp *interp, char *str, Seg *seg);
static int MemberIndex(char *name);

/* Functions *****************************************************************/
int PhCmd(ClientData clientData, Tcl_Interp *interp,
	  int argc, char *argv[])
{
  Err(argc < 2, "missing directive: must be one of: poly2seg, seg2poly, setdb, member, within, across");
#if 0
  if(!strcmp(argv[1], "poly2seg")) {
    Tpm_Polygon *poly;
    Seg seg;
    XPoint point;
    int i,j;
    Err(argc != 3, "usage: poly2seg <coords>");
    poly = Tpm_PolygonCreate();
    /* add the values in <coords> to poly */
    Err(Tcl_SplitList(interp, argv[2], &argc, &argv) == TCL_ERROR,
	"bad coords list");
    MEM_ALLOC_NOTIFY(argv);
    Err(argc % 2, "bad poly");
    for(i=0;i<argc;i++) {
      GetInt(argv[i], j); point.x = j;
      i++;
      GetInt(argv[i], j); point.y = j;
      Tpm_PolygonAddPoint(poly, &point);
    }
    free(argv);
    seg = SegFromPolygon(poly, 512, 512);
    Tpm_PolygonFree(poly);
    i = seg.mask[8]<<24;
    ReturnSeg(seg);
  }
  else if(!strcmp(argv[1], "seg2poly")) {
    Tpm_Polygon *poly;
    Seg seg;
    XPoint point;
    int i;
    Err(argc != 3, "usage: seg2poly <segmask>");
    Err(GetSeg(interp, argv[2], &seg), "bad segmask");
    poly = SegToPolygon(&seg, 512, 512);
    for(i=0;i<poly->npoints;i++) {
      char str[10];
      sprintf(str, "%d", poly->points[i].x);
      Tcl_AppendElement(interp, str);
      sprintf(str, "%d", poly->points[i].y);
      Tcl_AppendElement(interp, str);
    }
    Tpm_PolygonFree(poly);
  }
  else if(!strcmp(argv[1], "seg2polys")) {
    Tpm_Polygon **polys, **pp;
    Seg seg;
    XPoint point;
    int i;
    Tcl_DString ds;
    Err(argc != 3, "usage: seg2polys <segmask>");
    Err(GetSeg(interp, argv[2], &seg), "bad segmask");
    polys = SegToPolygons(&seg, 512, 512);
    Tcl_DStringInit(&ds);
    for(pp = polys; *pp; pp++) {
      Tpm_Polygon *poly = *pp;
      Tcl_DStringStartSublist(&ds);
      for(i=0;i<poly->npoints;i++) {
	char str[10];
	sprintf(str, "%d", poly->points[i].x);
	Tcl_DStringAppendElement(&ds, str);
	sprintf(str, "%d", poly->points[i].y);
	Tcl_DStringAppendElement(&ds, str);
      }
      Tcl_DStringEndSublist(&ds);
    }
    Tpm_PolygonsFree(polys);
    Tcl_DStringResult(interp, &ds);
    Tcl_DStringFree(&ds);
  }
#endif
  if(!strcmp(argv[1], "seg")) {
    Err(argc < 3, "usage: seg show|size|fromTcl|toTcl|fromHier");
    if(!strcmp(argv[2], "fromHier")) {
      Obj obj;
      WithinNode *segments;
      int i;
      Err(argc != 5, "usage: seg fromHier hier arrayvar");
      GetPtr(argv[3], obj);
      segments = (WithinNode*)((Hier)obj->value)->node;
      for(i=0;i<Boa_Array_Length(obj->type);i++) {
	/* set arrayvar(i) to segments[i].seg */
	char name[10], value[10*8];
	sprintf(name, "%d", i);
	SegString(value, segments[i].seg);
	Tcl_SetVar2(interp, argv[4], name, value, 0);
      }
      ReturnInt(Boa_Array_Length(obj->type));
    }
    else if(!strcmp(argv[2], "toTcl")) {
      Obj segObj;
      Err(argc != 4, "usage: seg toTcl obj");
      GetPtr(argv[3], segObj);
      ReturnSeg(Boa_Segmask_Value(segObj));
      Boa_Release(segObj);
    }
    else if(!strcmp(argv[2], "fromTcl")) {
      Obj segObj;
      Seg seg;
      Err(argc != 4, "usage: seg fromTcl segmask");
      Err(GetSeg(interp, argv[3], &seg), "bad segmask");
      segObj = Boa_Create(T_Segmask);
      Boa_Segmask_Value(segObj) = seg;
      ReturnPtr(segObj);
    }
    else if(!strcmp(argv[2], "show")) {
      Seg seg;
      int x,y, xsize, ysize;
      char cmd[128];
      Err(argc != 6, "usage: seg show canvas segmask color");
      /* use the size of the canvas to determine scale */
      xsize = Tk_Width(Tk_NameToWindow(interp, argv[3], Tk_MainWindow(interp)));
      ysize = Tk_Height(Tk_NameToWindow(interp, argv[3], Tk_MainWindow(interp)));
      Err(GetSeg(interp, argv[4], &seg), "bad segmask");
      for(y=0;y<SEG_YCELLS;y++) {
	for(x=0;x<SEG_XCELLS;x++) {
	  if(!SegIsSet(&seg, x, y)) continue;
#if 1
	  sprintf(cmd, "%s create line %d %d %d %d -fill %s -width 2 -tags line",
		  argv[3], x*xsize/SEG_XCELLS, y*ysize/SEG_YCELLS, 
		  (x+1)*xsize/SEG_XCELLS, (y+1)*ysize/SEG_YCELLS, argv[5]);
	  Tcl_Eval(interp, cmd);
#if 1
	  sprintf(cmd, "%s create line %d %d %d %d -fill %s -width 2 -tags line",
		  argv[3], (x+1)*xsize/SEG_XCELLS, y*ysize/SEG_YCELLS, 
		  x*xsize/SEG_XCELLS, (y+1)*ysize/SEG_YCELLS, argv[5]);
	  Tcl_Eval(interp, cmd);
#endif
#else
	  sprintf(cmd, "%s create rect %d %d %d %d -outline %s -width 2 -tags line",
		  argv[3], (x+1)*xsize/SEG_XCELLS, y*ysize/SEG_YCELLS, 
		  x*xsize/SEG_XCELLS, (y+1)*ysize/SEG_YCELLS, argv[5]);
	  Tcl_Eval(interp, cmd);
#endif
	}
      }
    }
    else if(!strcmp(argv[2], "size")) {
      Seg seg;
      Err(argc != 4, "usage: seg size segmask");
      Err(GetSeg(interp, argv[3], &seg), "bad segmask");
      ReturnInt(SegSize(&seg));
    }
    else {
      Err(1, "unknown seg directive");
    }
  }
  else if(!strcmp(argv[1], "set_index")) {
    Err(argc < 2, "usage: set_index ?obj?");
    GetPtr(argv[2], member_index);
  }
  else if(!strcmp(argv[1], "member")) {
    Err(argc < 2, "usage: member name");
    sprintf(interp->result, "%d", MemberIndex(argv[2]));
  }
  else if(!strcmp(argv[1], "within")) {
    Err(argc < 3, "usage: within init|destroy|get|pos|neg");
    if(!strcmp(argv[2], "init")) {
      Obj within;
      List within_list;
      Err(argc != 5, "usage: within init within within_list");
      GetPtr(argv[3], within);
      if(StringGets(interp, argv[4], &within_list)) return TCL_ERROR;
      Ph_WithinInit(within, within_list);
      ListFree(within_list);
    }
    else if(!strcmp(argv[2], "destroy")) {
      Ph_WithinDestroy();
    }
    else if(!strcmp(argv[2], "save")) {
      Ph_WithinSave();
    }
    else if(!strcmp(argv[2], "clear")) {
      Ph_WithinClear();
    }
    else if(!strcmp(argv[2], "get")) {
      Seg seg = Ph_WithinGet();
      ReturnSeg(seg);
    }
    else if(!strcmp(argv[2], "pos") || !strcmp(argv[2], "neg")) {
      int x, y, xsize, ysize;
      Tk_Window win;
      Err(argc < 6, "usage: seg pos/neg canvas x y");
      /* use the size of the canvas to determine scale */
      GetWindow(argv[3], win);
      xsize = Tk_Width(win);
      ysize = Tk_Height(win);
      GetInt(argv[4], x);
      GetInt(argv[5], y);
      Ph_WithinEx(x*SEG_XCELLS/xsize, y*SEG_YCELLS/ysize, 
		  argv[2][0]=='p' ? 1 : -1);
    }
#if 0
    else if(!strcmp(argv[2], "bench")) {
      Seg seg;
      Err(argc < 4, "usage: within bench segmask");
      Err(GetSeg(interp, argv[3], &seg), "bad segmask");
      ReturnInt(Ph_WithinBench(&seg));
    }
#endif
    else Err(1, "unknown within directive");
  }
  else if(!strcmp(argv[1], "across")) {
    Err(argc < 3, "usage: across init|pos|neg|get|destroy");
    if(!strcmp(argv[2], "init")) {
      Obj within, across;
      List within_list, across_list;
      Err(argc != 7, "usage: across init across across_list within within_list");
      GetPtr(argv[3], across);
      if(StringGets(interp, argv[4], &across_list)) return TCL_ERROR;
      GetPtr(argv[5], within);
      if(StringGets(interp, argv[6], &within_list)) return TCL_ERROR;
      Ph_AcrossInit(member_index, 
		    across, across_list, within, within_list);
      ListFree(across_list);
      ListFree(within_list);
    }
    else if(!strcmp(argv[2], "destroy")) {
      Ph_AcrossDestroy();
    }
    else if(!strcmp(argv[2], "seg")) {
      int index, label;
      Seg seg;
      Err(argc < 5, "usage: across seg mem-index label");
      GetInt(argv[3], index);
      GetInt(argv[4], label);
      seg = Ph_AcrossSeg(index, label);
      ReturnSeg(seg);
    }
    else if(!strcmp(argv[2], "get")) {
      int index, label;
      List segments;
      AcrossData *data;
      Err(argc < 5, "usage: across get mem-index label");
      GetInt(argv[3], index);
      GetInt(argv[4], label);
      segments = Ph_AcrossGet(index, label);
      {ListIter(p, data, segments) {
	char str[256];
	sprintf(str, "%d %d", data->within, data->segment);
	Tcl_AppendElement(interp, str);
      }}
      ListFree(segments);
    }
    else if(!strcmp(argv[2], "pos") || !strcmp(argv[2], "neg")) {
      int index, label;
      Seg seg;
      Err(argc < 6, "usage: across pos/neg mem-index label seg");
      GetInt(argv[3], index);
      GetInt(argv[4], label);
      Err(GetSeg(interp, argv[5], &seg), "bad segmask");
      Ph_AcrossEx(index, label, &seg, argv[2][0]=='p' ? 1 : -1);
    }
    else if(!strcmp(argv[2], "enable")) {
      int index, flag;
      Err(argc != 5, "usage: across enable across flag");
      GetInt(argv[3], index);
      GetInt(argv[4], flag);
      Ph_AcrossEnable(index, flag);
    }
    else {
      Err(1, "unknown across directive");
    }
  }
  else {
    Err(1, "unknown directive");
  }
  return TCL_OK;
}

void PhShutdown(void)
{
  if(member_index) Boa_Release(member_index);
}

int Ph_Init(Tcl_Interp *interp)
{
  /* make sure that boaTcl is loaded */
  Tcl_Eval(interp, "catch boa");
  Boa_Hier_Pkg();
  Boa_Region_Pkg();
  Boa_Vector_Pkg();
  Boa_StringSet(Boa_Types, "segMap", BOA_INSERT, Boa_SegMap_Define());
  Tcl_CreateCommand(interp, "ph", PhCmd, NULL, 
		    (Tcl_CmdDeleteProc*)PhShutdown);
  return Tcl_PkgProvide(interp, "phTcl", "1");
}

static int StringGets(Tcl_Interp *interp, char *str, List *result)
{
  int i, argc;
  char **argv;

  if(Tcl_SplitList(interp, str, &argc, &argv) == TCL_ERROR) 
    return TCL_ERROR;
  MEM_ALLOC_NOTIFY(argv);
  *result = ListCreate(GenericFree);
  for(i=0;i<argc;i++) {
    ListAddRear(*result, strdup(argv[i]));
  }
  free(argv);
  return TCL_OK;
}

static int ObjGets(Tcl_Interp *interp, char *str, List *result)
{
  int i, argc;
  char **argv;

  if(Tcl_SplitList(interp, str, &argc, &argv) == TCL_ERROR) 
    return TCL_ERROR;
  MEM_ALLOC_NOTIFY(argv);
  *result = ListCreate((FreeFunc*)Boa_Free);
  for(i=0;i<argc;i++) {
    Obj obj;
    GetPtr(argv[i], obj);
    ListAddRear(*result, obj);
  }
  free(argv);
  return TCL_OK;
}

static int GetSeg(Tcl_Interp *interp, char *str, Seg *seg)
{
  int i, v, argc;
  char **argv;
  if((Tcl_SplitList(interp, str, &argc, &argv) == TCL_ERROR)
      || (argc != 8)) return TCL_ERROR;
  MEM_ALLOC_NOTIFY(argv);
  for(i=0;i<argc;i++) {
    GetInt(argv[i], v);
    seg->mask[i*4] = v >> 24;
    seg->mask[i*4+1] = (v >> 16) & 255;
    seg->mask[i*4+2] = (v >> 8) & 255;
    seg->mask[i*4+3] = v & 255;
  }
  free(argv);
  return TCL_OK;
}

static int MemberIndex(char *name)
{
  int i;
  if(member_index->type->type == T_Array) {
    for(i=0;i<Boa_Array_Length(member_index->type);i++)
      if(!strcmp(name, ((char**)member_index->value)[i])) return i;
    return NAI;
  }
  if(Boa_StringGetInt(member_index, name, &i) != TPM_OK) return NAI;
  return i;
}
