#include <boa/boa.h>
#include <boa/image.h>
#include <boa/vector.h>
#include <tpm/region.h>
#include "segmask.h"

/* Prototypes ****************************************************************/
int Boa_Histogram(Obj context, Type *type_in_out, Obj *value_return);
int Boa_VectorNormalize(Obj context, Type *type_in_out, Obj *value_return);
int Boa_MatrixInvert(Obj context, Type *type_in_out, Obj *value_return);
int Boa_MaskImage(Obj context, Type *type_in_out, Obj *value_return);
int Boa_AcrossImage(Obj context, Type *type_in_out, Obj *value_return);
int Boa_DCT(Obj context, Type *type_in_out, Obj *value_return);

/* Functions *****************************************************************/

/* Calls TpmImageHistogram on "image", returning the histogram as a vector.
 * "quant" specifies the number of distinguishable values in each color
 * channel (defaults to 256).
 * "image" can be an Image, Volume, Matrix, or Vector.
 */
int Boa_Histogram(Obj context, Type *type_in_out, Obj *value_return)
{
  Obj t;
  int i, quant;
  Vector result;

  BOA_TYPECHECK_GET(T_Vector);
  TRY_RETURN(Boa_StringGet(context, "image", 0, NULL, &t));

  if(t->type == T_Image) {
    int *bins, length;
    length = 256 * Boa_Image_Value(t)->channels;
    bins = TpmImageHistogram(Boa_Image_Value(t));
    MEM_ALLOC_NOTIFY(bins);
    Boa_Release(t);

    result = VectorCreate(length);
    for(i=0;i<length;i++) result->d[i] = (double)bins[i];
    free(bins);
  }
  else if(t->type == T_Volume) {
    int c, image_size;
    Volume v = Boa_Volume_Value(t);
    image_size = v->xdim * v->ydim;
    result = VectorCreate(256 * v->zdim);
    VectorSet(result, 0.0);
    for(c=0;c<v->zdim;c++) {
      for(i=0;i<image_size;i++) {
	unsigned char value = (unsigned char)v->d3[c][0][i];
	result->d[c*256 + value]++;
      }
    }
    Boa_Release(t);
  }
  else if((t->type == T_Matrix) || (t->type == T_Vector)) {
    Vector v = Boa_Vector_Value(t);
    result = VectorCreate(256);
    VectorSet(result, 0.0);
    for(i=0;i<v->len;i++) {
      unsigned char value = (unsigned char)v->d[i];
      result->d[value]++;
    }
    Boa_Release(t);
  }
  else {
    Boa_Release(t);
    ErrorString = "input to Boa_Histogram is not an image";
    return TPM_BOTTOM;
  }

  if(Boa_StringGetInt(context, "quant", &quant) == TPM_OK) {
    /* compress the histogram */
    Vector v = result;
    int factor = 256/quant, j;
    result = VectorCreate(v->len/factor);
    VectorSet(result, 0.0);
    for(i=0;i<result->len;i++) {
      for(j=0;j<factor;j++) {
	result->d[i] += v->d[i*factor+j];
      }
    }
    VectorFree(v);
  }
  *value_return = Boa_Vector_Create(result);
  return TPM_OK;
}

/* Normalize a vector so that it sums to 1.
 * Input: vector
 */
int Boa_VectorNormalize(Obj context, Type *type_in_out, Obj *value_return)
{
  Obj t;
  Vector v;
  BOA_TYPECHECK_GET(T_Vector);
  TRY_RETURN(Boa_StringGet(context, "vector", 0, NULL, &t));
  v = VectorCopy(Boa_Vector_Value(t));
  /* mutate the vector in-place */
  VectorScale(v, 1.0/VectorSum(v));
  *value_return = Boa_Vector_Create(v);
  return TPM_OK;
}

/* Invert a Matrix.
 * Input: matrix
 */
int Boa_MatrixInvert(Obj context, Type *type_in_out, Obj *value_return)
{
  Obj t;
  Matrix m;
  BOA_TYPECHECK_GET(T_Matrix);
  TRY_RETURN(Boa_StringGet(context, "matrix", 0, NULL, &t));
  m = MatrixCopy(Boa_Matrix_Value(t));
  MatrixInvert(m);
  *value_return = Boa_Matrix_Create(m);
  return TPM_OK;
}

/* Produce the subimage of image contained in r. 
 * The background color of each channel is given by "bg."
 */
static TpmImage MaskImage(TpmImage image, Tpm_Region *r, 
			  unsigned char bg)
{
  TpmImage result = TpmImageCreate(r->width, r->height, image->channels);
  int y,x,c;
  for(y=0;y<r->height;y++) {
    for(x=0;x<r->width;x++) {
      if(Tpm_RegionContains(r, r->x+x, r->y+y)) {
	for(c=0;c<image->channels;c++) 
	  result->d3[c][y][x] = image->d3[c][r->y+y][r->x+x];
      }
      else {
	for(c=0;c<image->channels;c++) 
	  result->d3[c][y][x] = bg;
      }
    }
  }
  return result;
}

/* Calls MaskImage to extract a mask.
 * Input:
 *   image
 *   segmask     (defaults to SegAll, i.e. no masking takes place)
 *   background  (defaults to black)
 */
int Boa_MaskImage(Obj context, Type *type_in_out, Obj *value_return)
{
  Obj imageObj, t;
  Seg seg;
  TpmImage image;
  Tpm_Region *region;
  int bg;
  BOA_TYPECHECK_GET(T_Image);
  /* get args */
  if(Boa_StringGet(context, "segmask", 0, &T_Segmask, &t) == TPM_OK) {
    seg = Boa_Segmask_Value(t);
    Boa_Release(t);
  }
  else seg = SegAll;
  if(Boa_StringGetInt(context, "background", &bg) != TPM_OK) {
    bg = 0;
  }
  TRY_RETURN(Boa_StringGet(context, "image", 0, &T_Image, &imageObj));
  image = Boa_Image_Value(imageObj);

  region = SegToRegion(&seg, image->width, image->height);
  image = MaskImage(image, region, bg);
  Tpm_RegionFree(region);
  Boa_Release(imageObj);
  *value_return = Boa_Image_Create(image);
  return TPM_OK;
}

/* Return an image corresponding to an across-hier node.
 * Inputs:
 *   across = across-hier
 *   i = which node
 *   within = maps node data into segmask
 *   index = maps integers into member names
 *   view = maps member names into images
 *   bg = background color for mask  (defaults to black)
 * Outputs:
 *   result = masked image  (uses Boa_MaskImage)
 */
int Boa_AcrossImage(Obj context, Type *type_in_out, Obj *value_return)
{
  Obj hierObj, t, within, segObj, indexObj, view, imageObj, bg, con;
  int i;
  struct AcrossData {
    int within,    /* which within-hier this node occurs in */
        member,    /* which member this node belongs to */
        segment;   /* which segment in the member's within-hier */
  };

  /* get the hier node */
  if(Boa_StringGet(context, "across", 0, NULL, &hierObj) != TPM_OK) {
    ErrorPrepend("no value for `across'");
    return TPM_BOTTOM;
  }
  if(Boa_StringGetInt(context, "i", &i) != TPM_OK) {
    ErrorPrepend("no value for `i'");
    return TPM_BOTTOM;
  }
  if(Boa_IntGet(hierObj, i, 0, NULL, &t) != TPM_OK) {
    ErrorPrepend("no index %d in the across-hier", i);
    return TPM_BOTTOM;
  }
  Boa_Release(hierObj);
  i = ((struct AcrossData *)t->value)->member;

  /* get the corresponding segmask */
  if(Boa_StringGet(context, "within", 0, NULL, &within) != TPM_OK) {
    ErrorPrepend("no value for `within'");
    return TPM_BOTTOM;
  }
  /* t_segmap's output must be split */
  if(Boa_ObjGet(within, t, '/', &T_Segmask, &segObj) != TPM_OK) {
    ErrorPrepend("no such segmask");
    return TPM_BOTTOM;
  }
  Boa_Release(within);
  Boa_Release(t);

  /* get the corresponding image */
  if(Boa_StringGet(context, "index", 0, NULL, &indexObj) != TPM_OK) {
    ErrorPrepend("no value for `index'");
    return TPM_BOTTOM;
  }
  if(Boa_IntGet(indexObj, i, 0, &T_String, &t) != TPM_OK) {
    ErrorPrepend("%d not found in index", i);
    return TPM_BOTTOM;
  }
  Boa_Release(indexObj);
  if(Boa_StringGet(context, "view", 0, NULL, &view) != TPM_OK) {
    ErrorPrepend("no value for `view'");
    return TPM_BOTTOM;
  }
  if(Boa_StringGet(view, Boa_String_Value(t), 0, &T_Image, &imageObj) != TPM_OK) {
    ErrorPrepend("no image for %s", Boa_String_Value(t));
    return TPM_BOTTOM;
  }
  Boa_Release(t);
  Boa_Release(view);

  /* get the masked image */
  con = Boa_Create(T_Dict);
  Boa_StringSet(con, "image", BOA_INSERT, imageObj);
  Boa_StringSet(con, "segmask", BOA_INSERT, segObj);
  if(Boa_StringGet(context, "background", 0, &T_Int, &bg) == TPM_OK) {
    Boa_StringSet(con, "background", BOA_INSERT, bg);
  }
  if(Boa_MaskImage(con, type_in_out, value_return) != TPM_OK) {
    ErrorPrepend("MaskImage failed");
    return TPM_BOTTOM;
  }
  Boa_Release(con);
  return TPM_OK;
}

/* Calls MatrixDCT on "image", returning parameters in a vector. 
 * "divide", a power of two, divides the number of params to keep 
 * (default is 1).
 * Destroys image.
 */
int Boa_DCT(Obj context, Type *type_in_out, Obj *value_return)
{
  Obj t;
  Vector result;
  Matrix m;
  int i, j, divide, keep;
  double *p;

  BOA_TYPECHECK_GET(T_Vector);
  TRY_RETURN(Boa_StringGet(context, "image", 0, NULL, &t));
  /* special case: Volume with zdim == 1 */
  if((t->type == T_Volume) && (Boa_Volume_Value(t)->zdim == 1)) {
    /* do nothing; treat it like a Matrix */
  }
  else {
    t = Boa_Matrix_Coerce(t);
    if(t->type == T_Volume) {
      /* zdim must be > 1; convert to grayscale */
      Matrix m = VolumeGray(Boa_Volume_Value(t));
      Boa_Release(t);
      t = Boa_Matrix_Create(m);
    }
    if(t->type != T_Matrix) { Boa_Release(t); return TPM_ERROR; }
  }
  m = Boa_Matrix_Value(t);
#if 0
  /* zero mean the image */
  VectorAddValue((Vector)m, -VectorSum((Vector)m)/m->len);
#endif
#if 1
  /* make shift-invariant by taking DFT magnitude */
  MatrixDFT(m, (Matrix)1);
#endif
  MatrixDCT(m);
  if(Boa_StringGetInt(context, "divide", &divide) != TPM_OK) {
    divide = 1;
  }
  divide /= 2;
  if(divide) {
    /* return the upper left triangle */
    keep = m->rows/divide;
    result = VectorCreate(keep*(keep+1)/2);
    p = result->d;
    for(i=0;i<keep;i++) {
      for(j=0;j<keep-i;j++) {
	*p++ = m->d2[i][j];
      }
    }
  }
  else {
    result = VectorCopy((Vector)m);
  }
  Boa_Release(t);
  /* square the dct coeffs so that we can average them */
  for(i=0;i<result->len;i++) result->d[i] *= result->d[i];
  *value_return = Boa_Vector_Create(result);
  return TPM_OK;
}

/* Returns a masked segment image.
 * Inputs:
 *   func = function which takes (image, segmask)
 *   view = maps member into image
 *   within = maps member/segment into segmask
 *   member = member name
 *   segment = segment name
 * Outputs:
 *   result = func(view(member), within(member/segment))
 */
int Boa_ViewWithin(Obj context, Type *type_in_out, Obj *value_return)
{
  Obj t, within, withinMember, segObj, view, imageObj, func, segment;
  char *member;
  int status;

  if(Boa_StringGet(context, "member", 0, &T_String, &t) != TPM_OK) {
    ErrorPrepend("no value for `member'");
    return TPM_BOTTOM;
  }
  member = Boa_String_Value(t);
  Boa_Release(t);

  if(Boa_StringGet(context, "func", 0, NULL, &func) != TPM_OK) {
    ErrorPrepend("no value for `func'");
    return TPM_BOTTOM;
  }

  /* get the image */
  if(Boa_StringGet(context, "view", 0, NULL, &view) != TPM_OK) {
    ErrorPrepend("no value for `view'");
    return TPM_BOTTOM;
  }
  if(Boa_StringGet(view, member, 0, &T_Image, &imageObj) != TPM_OK) {
    ErrorPrepend("no image for %s", member);
    return TPM_BOTTOM;
  }
  Boa_Release(view);

  /* curry with image */
  if(Boa_ObjGet(func, imageObj, 0, NULL, &t) != TPM_OK) {
    ErrorPrepend("func(image) failed");
    return TPM_BOTTOM;
  }
  Boa_Release(imageObj);
  Boa_Release(func);
  func = t;

  /* get the segmask */
  if(Boa_StringGet(context, "within", 0, NULL, &within) != TPM_OK) {
    ErrorPrepend("no value for `within'");
    return TPM_BOTTOM;
  }
  if(Boa_StringGet(within, member, 0, NULL, &withinMember) != TPM_OK) {
    ErrorPrepend("no within for `%s'", member);
    return TPM_BOTTOM;
  }
  Boa_Release(within);
  if(Boa_StringGet(context, "segment", 0, NULL, &segment) != TPM_OK) {
    ErrorPrepend("no value for `segment'");
    return TPM_BOTTOM;
  }
  if(Boa_ObjGet(withinMember, segment, 0, NULL, &segObj) != TPM_OK) {
    ErrorPrepend("no such segmask");
    return TPM_BOTTOM;
  }
  Boa_Release(segment);
  Boa_Release(withinMember);

  /* execute the function */
  status = Boa_ObjGet(func, segObj, 0, type_in_out, value_return);
  Boa_Release(segObj);
  Boa_Release(func);
  return status;
}
