#include <boa/boa.h>
#include <boa/vector.h>

/* Prototypes ****************************************************************/

int Boa_MRSAR(Obj context, Type *type_in_out, Obj *value_return);
Vector MRSAR(Matrix image);

/* Private */
static double sar_std(Matrix s, double *x);
static double MatrixCorrMN(Matrix a, int m, int n);
static void MatrixCorr(Matrix src, Matrix dest);
static void lse_neighbor_sum(Matrix win, int d, Matrix sum);

/* Functions *****************************************************************/

/* Calls MRSAR on "image", returning parameters in a vector.
 * MRSAR modifies image.
 */
int Boa_MRSAR(Obj context, Type *type_in_out, Obj *value_return)
{
  Obj t;
  Vector result;
  BOA_TYPECHECK_GET(T_Vector);
  TRY_RETURN(Boa_StringGet(context, "image", 0, NULL, &t));
  /* special case: Volume with zdim == 1 */
  if((t->type == T_Volume) && (Boa_Volume_Value(t)->zdim == 1)) {
    /* do nothing; treat it like a Matrix */
  }
  else {
    t = Boa_Matrix_Coerce(t);
    if(t->type == T_Volume) {
      /* zdim must be > 1; convert to grayscale */
      Matrix m = VolumeGray(Boa_Volume_Value(t));
      Boa_Release(t);
      t = Boa_Matrix_Create(m);
    }
    if(t->type != T_Matrix) { Boa_Release(t); return TPM_ERROR; }
  }
  result = MRSAR(Boa_Matrix_Value(t));
  Boa_Release(t);
  *value_return = Boa_Vector_Create(result);
  return TPM_OK;
}

/* Computes multiresolution SAR parameters over a single-channel image.
 * Modifies: image (subtracts mean from windows)
 * Output is concatentation of 4 AR params and std dev of error for each
 * resolution level. For example, for the three resolution levels 2,3,4, 
 * the vector will have 15 numbers, 5 for each level:
 * | AR 2 | dev 2 | AR 3 | dev 3 | AR 4 | dev 4 |
 */
Vector MRSAR(Matrix image)
{
#define MODEL_ORDER 2
#define LSE_SIZE (MODEL_ORDER*2)
#define NFEATURES (LSE_SIZE+1)
#define NLEVELS 3
  int levels[NLEVELS] = { 2, 3, 4 };

  Vector result;
  Matrix sum, A;
  double *B;
  int i, level;
  
  result = VectorCreate(NFEATURES * NLEVELS);
  sum = MatrixCreate(image->len, NFEATURES);
  A = MatrixCreate(LSE_SIZE, LSE_SIZE);
  B = Allocate(NFEATURES, double);

  /* debias */
  VectorAddValue((Vector)image, -VectorSum((Vector)image)/image->len);

  /* loop resolution levels */
  for(level = 0; level < NLEVELS; level++) {
    /* compute the features for this resolution */
    /* sum[i][LSE_SIZE] is the pixel value itself */
    lse_neighbor_sum(image, levels[level], sum);

    /* let A be the correlation matrix */
    MatrixCorr(sum, A);

    /* let B be the correlations with the pixel value */
    for(i=0; i<LSE_SIZE; i++) {
      B[i] = MatrixCorrMN(sum, i, LSE_SIZE);
    }

    /* estimate the parameters; place in B */
    MatrixSolve(A, B);
    /* let the last element be the std dev of the error */
    B[LSE_SIZE] = sar_std(sum, B);
    
    /* put into the feature vector */
    memcpy(&result->d[level*NFEATURES], B, NFEATURES*sizeof(double));
  }
  MatrixFree(sum);
  MatrixFree(A);
  free(B);
  return result;
}

/* Vector dot product. */
static double dot_product(double *a, double *b, int len)
{
  double r = 0.0;
  while(len--) r += *a++ * *b++;
  return r;
}

/* Compute the std dev of the error from the estimated parameters
 * to the true pixel values.
 */
static double sar_std(Matrix s, double *x)
{
  int i;
  double sum = 0.0, v;

  for(i=0;i<s->rows;i++) {
    v = dot_product(s->d2[i], x, LSE_SIZE) - s->d2[i][LSE_SIZE];
    sum += v*v;
  }
  v = sqrt(sum / s->rows);
  return v;
}

/* Computes the correlation between columns m and n in a */
static double MatrixCorrMN(Matrix a, int m, int n)
{
  int i;
  double r = 0.0;

  for(i=0;i<a->rows;i++) {
    r += a->d2[i][m] * a->d2[i][n];
  }
  return r / a->rows;
}

/* Computes a matrix of correlations of src, placing it in dest */
static void MatrixCorr(Matrix src, Matrix dest)
{
  int i,j;

  for(i=0;i<dest->rows;i++) {
    for(j=i;j<dest->cols;j++) {
      dest->d2[j][i] = dest->d2[i][j] = MatrixCorrMN(src, i, j);
    }
  }
}

/* Computes symmetric, noncausal AR features over all points in win */
static void lse_neighbor_sum(Matrix win, int d, Matrix sum)
{
  int i,j, i1,i2, j1,j2;
  double *p;

  p = sum->d;
  for(i=0;i<win->rows;i++) {
    i1 = (i - d + win->rows) % win->rows;
    i2 = (i + d) % win->rows;
    for(j=0;j<win->cols;j++) {
      j1 = (j - d + win->cols) % win->cols;
      j2 = (j + d) % win->cols;
      *p++ = win->d2[i1][j] + win->d2[i2][j];
      *p++ = win->d2[i][j1] + win->d2[i][j2];
      *p++ = win->d2[i1][j1] + win->d2[i2][j2];
      *p++ = win->d2[i1][j2] + win->d2[i2][j1];
      *p++ = win->d2[i][j];    /* last entry is true pixel value */
    }
  }
}
