#include <tpm/tpm.h>
#include <tpm/region.h>
#include "segmask.h"

/* Globals *******************************************************************/

Seg SegZero = { 
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0
  };
Seg SegAll = { 
  -1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1
  };

/* Prototypes ****************************************************************/

void SegSet(Seg *seg, int x, int y);
void SegClear(Seg *seg, int x, int y);
int SegIsSet(Seg *seg, int x, int y);
void SegUnion(Seg *result, Seg *seg);
void SegIntersect(Seg *result, Seg *seg);
void SegXor(Seg *result, Seg *seg);
int SegSize(Seg *seg);

Seg SegFromRegion(Tpm_Region *r, int xmax, int ymax);
Tpm_Region *SegToRegion(Seg *seg, int xmax, int ymax);

Seg SegFromPolygon(Tpm_Polygon *poly, int xmax, int ymax);
Tpm_Polygon *SegToPolygon(Seg *seg, int xmax, int ymax);
Tpm_Polygon **SegToPolygons(Seg *seg, int xmax, int ymax);

/* Functions *****************************************************************/

void SegSet(Seg *seg, int x, int y)
{
  if((x < 0) || (x >= SEG_XCELLS)) return;
  if((y < 0) || (y >= SEG_YCELLS)) return;
  seg->mask[(y*SEG_XCELLS + x)/8] |= 0x80 >> (x%8);
}

void SegClear(Seg *seg, int x, int y)
{
  if((x < 0) || (x >= SEG_XCELLS)) return;
  if((y < 0) || (y >= SEG_YCELLS)) return;
  seg->mask[(y*SEG_XCELLS + x)/8] &= ~(0x80 >> (x%8));
}

int SegIsSet(Seg *seg, int x, int y)
{
  if((x < 0) || (x >= SEG_XCELLS)) return 0;
  if((y < 0) || (y >= SEG_YCELLS)) return 0;
  return seg->mask[(y*SEG_XCELLS + x)/8] & (0x80 >> (x%8));
}

void SegUnion(Seg *result, Seg *seg)
{
  int i;
  for(i=0;i<SEG_SZ;i++) {
    result->mask[i] |= seg->mask[i];
  }
}

void SegIntersect(Seg *result, Seg *seg)
{
  int i;
  for(i=0;i<SEG_SZ;i++) {
    result->mask[i] &= seg->mask[i];
  }
}

void SegXor(Seg *result, Seg *seg)
{
  int i;
  for(i=0;i<SEG_SZ;i++) {
    result->mask[i] ^= seg->mask[i];
  }
}

/* Returns the number of cells in seg */
int SegSize(Seg *seg)
{
  int i, result = 0;
  for(i=0;i<SEG_SZ;i++) {
    result += NumBits(seg->mask[i]);
  }
  return result;
}

void SegPrint(Seg *seg)
{
  int i,j;
  for(i=0;i<SEG_YCELLS;i++) {
    for(j=0;j<SEG_XCELLS;j++) {
      printf("%d ", SegIsSet(seg, j, i) ? 1:0);
    }
    printf("\n");
  }
}

/* Conversions ***************************************************************/

Seg SegFromRegion(Tpm_Region *region, int xmax, int ymax)
{
  Seg seg = SegZero;
  int i,j, x,y;
  int cellx, celly;
  cellx = xmax/SEG_XCELLS;
  celly = ymax/SEG_YCELLS;
  /* convert the region bitmap into a mask */
  for(i=0;i<SEG_YCELLS;i++) {
    for(j=0;j<SEG_XCELLS;j++) {
      /* count number of points contained in the cell */
      int count = 0;
      for(y=0;y<celly;y++) {
	for(x=0;x<cellx;x++) {
	  if(Tpm_RegionContains(region, x+j*cellx, y+i*celly)) count++;
	}
      }
      /* threshold is 1/4 of the cell */
      if(count >= cellx*celly/4) SegSet(&seg, j, i);
    }
  }
  return seg;
}

Tpm_Region *SegToRegion(Seg *seg, int xmax, int ymax)
{
  Tpm_Region *region;
  int i,j;
  int cellx, celly;
  int minx = xmax, miny = ymax, maxx = -1, maxy = -1;

  /* determine the extent of the Seg */
  cellx = xmax / SEG_XCELLS;
  celly = ymax / SEG_YCELLS;
  for(i=0;i<SEG_YCELLS;i++) {
    for(j=0;j<SEG_XCELLS;j++) {
      if(SegIsSet(seg, j, i)) {
	int y = i*celly, x = j*cellx;
	if(y < miny) miny = y;
	y += celly-1;
	if(y > maxy) maxy = y;
	if(x < minx) minx = x;
	x += cellx-1;
	if(x > maxx) maxx = x;
      }
    }
  }
  region = Tpm_RegionCreate();
  /* is the region empty? */
  if(maxx < 0) return region;
  region->x = minx;
  region->y = miny;
  Tpm_RegionSetDims(region, maxx-minx+1, maxy-miny+1);

  /* convert the segment bitmap into a region */
  for(i=0;i<SEG_YCELLS;i++) {
    for(j=0;j<SEG_XCELLS;j++) {
      if(SegIsSet(seg, j, i)) {
	int y;
	for(y=0;y<celly;y++)
	  memset(&region->data[y+i*celly-region->y][j*cellx-region->x], 1, cellx);
      }
    }
  }
  return region;
}

Seg SegFromPolygon(Tpm_Polygon *poly, int xmax, int ymax)
{
  Seg seg;
  Tpm_Region *region = Tpm_RegionFromPolygon(poly);
  seg = SegFromRegion(region, xmax, ymax);
  Tpm_RegionFree(region);
  return seg;
}

Tpm_Polygon *SegToPolygon(Seg *seg, int xmax, int ymax)
{
  Tpm_Region *region = SegToRegion(seg, xmax, ymax);
  Tpm_Polygon *poly = Tpm_RegionPolygon(region);
  Tpm_RegionFree(region);
  return poly;
}

Tpm_Polygon **SegToPolygons(Seg *seg, int xmax, int ymax)
{
  Tpm_Region *region = SegToRegion(seg, xmax, ymax);
  Tpm_Polygon **polys = Tpm_RegionToPolygons(region);
  Tpm_RegionFree(region);
  return polys;
}
