/* Converts a numerical segment reference (within hier, member, segment #)
 * into a named reference "within/member/segment". 
 * Alternatively, converts a named reference into a numerical one.
 * attributes:
 *   within      Converts numbers into hier names, names into numbers
 *   member      Converts numbers into member names, and vice-versa
 */
#include "photobook.h"

typedef struct SegMap {
  Obj within, member;
} *SegMap;

/* Globals *******************************************************************/
Type T_SegMap = NULL;

/* Functions *****************************************************************/

static void Init(Obj self)
{
  SegMap map = (SegMap)self->value;
  /* dummy objects */
  map->within = Boa_Create(T_Merge);
  map->member = Boa_Create(T_Merge);
}

static void Final(Obj self)
{
  SegMap map = (SegMap)self->value;
  Boa_Release(map->within);
  Boa_Release(map->member);
}

static void Clean(Obj self, ObjFunc *func, Any data)
{
  SegMap map = (SegMap)self->value;
  func(map->within, data);
  func(map->member, data);
}

static int SPrint(Obj self, StrStream ss)
{
  SegMap map = (SegMap)self->value;
  StrStreamPuts(ss, "segMap(");
  Boa_SPrint(map->within, ss);
  StrStreamPutc(ss, ' ');
  Boa_SPrint(map->member, ss);
  return StrStreamPutc(ss, ')');
}

/* returns an array of 3 integers corresponding to within/member/segment */
static int StringGet(Obj self, char *name, int flags,
		     Type *type_in_out, Obj *value_return)
{
  SegMap map = (SegMap)self->value;
  char *p;
  int *v;
  Type t;

  t = Boa_Type_Create("(array 3 int)");
  BOA_TYPECHECK_GET(t);
  v = Allocate(3, int);
  /* split the name into three components: within/member/segment */
  name = strdup(name);
  p = strtok(name, "/");
  if(!p) goto error;
  /* convert within into a number */
  if(Boa_StringGetInt(map->within, p, &v[0]) != TPM_OK) {
    ErrorPrepend("no such within `%s'", p);
    goto error;
  }
  /* convert member name into a number */
  p = strtok(NULL, "/");
  if(!p) goto error;
  if(Boa_StringGetInt(map->member, p, &v[1]) != TPM_OK) {
    ErrorPrepend("no such member `%s'", p);
    goto error;
  }
  /* convert segment name into a number */
  p = strtok(NULL, "/");
  if(!p) goto error;
  v[2] = atoi(p);
  free(name);
  if(value_return) *value_return = Boa_CreateFromData(t, v);
  return TPM_OK;

 error:
  free(v);
  free(name);
  return TPM_BOTTOM;
}

static int StringSet(Obj self, char *name, int flags, Obj value)
{
  SegMap map = (SegMap)self->value;
  if(name[0] == '_') {
    if(!strcmp(name+1, "within")) {
      Boa_Release(map->within);
      map->within = value;
      return TPM_OK;
    }
    else if(!strcmp(name+1, "member")) {
      Boa_Release(map->member);
      map->member = value;
      return TPM_OK;
    }
  }
  return TPM_BOTTOM;
}

/* returns a string, within/member/segment, corresponding to 3 integers */
static int ObjGet(Obj self, Obj index, int flags,
		  Type *type_in_out, Obj *value_return)
{
  SegMap map = (SegMap)self->value;
  AcrossData *ind = index->value;
  Obj within, member;
  char *s;

  /* get the within name from map->within */
  if(Boa_IntGet(map->within, ind->within, 0, NULL, &within) != TPM_OK) {
    ErrorPrepend("no within %d", ind->within);
    return TPM_BOTTOM;
  }
  /* get the member name from map->member */
  if(Boa_IntGet(map->member, ind->member, 0, NULL, &member) != TPM_OK) {
    ErrorPrepend("no member %d", ind->member);
    Boa_Release(within);
    return TPM_BOTTOM;
  }
  s = Allocate(strlen(Boa_String_Value(within)) +
	       strlen(Boa_String_Value(member)) +
	       13, char);
  sprintf(s, "%s/%s/%d", 
	  Boa_String_Value(within), Boa_String_Value(member), ind->segment);
#if 0
  fprintf(stderr, "%d %d %d ", ind->within, ind->member, ind->segment);
  fprintf(stderr, "-> %s\n", s);
#endif
  if(value_return) *value_return = Boa_String_Create(s);
  Boa_Release(within);
  Boa_Release(member);
  return TPM_OK;
}

static struct Boa_Methods methods = {
  Init, Final, Clean,
  SPrint, Err_Copy, Generic_Compare, Generic_Hash,
  StringGet, StringSet,
  Err_IntGet, Err_IntSet,
  ObjGet, Generic_ObjSet
};

Obj Boa_SegMap_Define(void)
{
  T_SegMap = Boa_SimpleType_Create("segMap", &methods, 
				   sizeof(struct SegMap), sizeof(Obj));
  return T_SegMap;
}
