#include <boa/boa.h>
#include <boa/vector.h>
#include "segmask.h"

/* compute inverse covariance? only if you have lapack. */
#define INVERT 1

/* Prototypes ****************************************************************/

int MeanCovariance(Obj context, Type *type_in_out, Obj *value_return);
int WindowFeatures(Obj context, Type *type_in_out, Obj *value_return);
int WindowTest(Obj context, Type *type_in_out, Obj *value_return);

/* Functions *****************************************************************/

/* Produces a mean and covariance for a volume segment. 
 * Caller must free the results.
 */
static int Stats(Volume features, Seg *mask, 
		 Vector *mean_return, Matrix *cov_return)
{
  int i,j,k, x,y,m,n, nrows;
  int cellx, celly;
  Vector mean;
  Matrix cov;

  if(features->xdim % SEG_XCELLS) {
    fprintf(stderr, "warning: %d columns on the right edge will be ignored\n",
	    features->xdim % SEG_XCELLS);
  }
  if(features->ydim % SEG_YCELLS) {
    fprintf(stderr, "warning: %d rows on the bottom edge will be ignored\n",
	    features->ydim % SEG_YCELLS);
  }
  cellx = features->xdim / SEG_XCELLS;
  celly = features->ydim / SEG_YCELLS;
  if(!cellx || !celly) {
    ErrorSet("dimensions (%d by %d) must both be >= 16", 
		features->ydim, features->xdim);
    return TPM_ERROR;
  }

  mean = VectorCreate(features->zdim);
  VectorSet(mean, 0.0);
  cov = MatrixCreate(features->zdim, features->zdim);
  VectorSet((Vector)cov, 0.0);
  nrows = 0;
  for(i=0;i<SEG_YCELLS;i++) {
    for(j=0;j<SEG_XCELLS;j++) {
      if(!SegIsSet(mask, j, i)) continue;
      /* loop the pixels in the cell */
      for(y=0;y<celly;y++) {
	for(x=0;x<cellx;x++) {
	  int row = y+i*celly, col = x+j*cellx;
	  nrows++;
	  /* update mean */
	  for(k=0;k<features->zdim;k++) {
	    mean->d[k] += features->d3[k][row][col];
	  }
	  /* update correlation matrix (only upper right diag) */
	  for(m=0;m<features->zdim;m++) {
	    for(n=m;n<features->zdim;n++) {
	      cov->d2[m][n] += features->d3[m][row][col] * 
		               features->d3[n][row][col];
	    }
	  }
	}
      }
    }
  }
  fprintf(stderr, "nrows: %d\n", nrows);
  /* if nrows == 0, this will (correctly) produce NaNs */
  VectorScale(mean, 1.0/nrows);
  if(nrows >= mean->len) {
    VectorScale((Vector)cov, 1.0/nrows);
    /* turn the correlation matrix into a covariance */
    for(m=0;m<features->zdim;m++) {
      for(n=m;n<features->zdim;n++) {
	cov->d2[m][n] -= mean->d[m] * mean->d[n];
	/* fill in lower left diagonal */
	cov->d2[n][m] = cov->d2[m][n];
      }
    }
    /* unbias the estimator */
    VectorScale((Vector)cov, (double)nrows/(nrows-1));
    /* this mean, cov code has been checked with MatLab */
  }
  else {
    /* assume a diagonal covariance */
    VectorSet((Vector)cov, 0.0);
    for(m=0;m<features->zdim;m++) {
      cov->d2[m][m] = 1;
    }
  }
  *mean_return = mean;
  *cov_return = cov;
  return TPM_OK;
}

/* Produces a mean and covariance for a volume segment.
 * Inputs: volume, segmask
 * Outputs: mean (vector), cov (matrix)
 */
int MeanCovariance(Obj context, Type *type_in_out, Obj *value_return)
{
  Obj t;
  Volume features;
  Seg seg;
  Vector mean;
  Matrix cov, icov;

  /* get args */
  TRY_RETURN(Boa_StringGet(context, "volume", 0, &T_Volume, &t));
  features = Boa_Volume_Value(t);
  Boa_Release(t);
  if(Boa_StringGet(context, "segmask", 0, &T_Segmask, &t) == TPM_OK) {
    seg = Boa_Segmask_Value(t);
    Boa_Release(t);
  }
  else seg = SegAll;
  
  TRY_RETURN(Stats(features, &seg, &mean, &cov));

  BOA_TYPECHECK_GET(T_Dict);
  *value_return = Boa_Create(T_Dict);
  Boa_StringSet(*value_return, "mean", BOA_INSERT, Boa_Vector_Create(mean));
  if(INVERT) icov = MatrixCopy(cov);
  Boa_StringSet(*value_return, "cov", BOA_INSERT, Boa_Matrix_Create(cov));
  if(INVERT) {
    /* XXX this may be expensive */
    MatrixInvert(icov);
    Boa_StringSet(*value_return, "icov", BOA_INSERT, Boa_Matrix_Create(icov));
  }
  return TPM_OK;
}

/* Produces a feature Volume for multiple overlapping windows over an image
 * Volume.  Each window is centered at a pixel; this pixel moves across the
 * image in a raster scan, stepping by xinc and yinc.
 * Inputs:
 *   image
 *   xsize = width of windows
 *   xinc = distance between window centers (overlap is xsize-xinc)
 *   ysize = height of windows; defaults to xsize
 *   yinc = see xinc; defaults to xinc
 *   depth = feature vector length (extra components will be ignored)
 *   func = feature computation closure: (image)->(vector)
 * Output: feature volume
 */
int WindowFeatures(Obj context, Type *type_in_out, Obj *value_return)
{
  Obj imageObj, windowObj;
  Volume result;
  Volume image, window;
  int xsize, ysize, xinc, yinc, depth;
  Obj func;
  int y, x, i, j, k;
  int ylow, yhi, ylen, xlow, xhi, xlen;
  int xstart, xend, ystart, yend;
  Vector vector;

  BOA_TYPECHECK_GET(T_Volume);
  /* get args */
  TRY_RETURN(Boa_StringGetInt(context, "xsize", &xsize));
  TRY_RETURN(Boa_StringGetInt(context, "xinc", &xinc));
  if(Boa_StringGetInt(context, "ysize", &ysize) != TPM_OK) {
    ysize = xsize;
  }
  if(Boa_StringGetInt(context, "yinc", &yinc) != TPM_OK) {
    yinc = xinc;
  }
  TRY_RETURN(Boa_StringGetInt(context, "depth", &depth));
  TRY_RETURN(Boa_StringGet(context, "func", 0, NULL, &func));
  TRY_RETURN(Boa_StringGet(context, "image", 0, NULL, &imageObj));
  imageObj = Boa_Volume_Coerce(imageObj);
  if(imageObj->type != T_Volume) { Boa_Release(imageObj); ErrorString = "not a volume"; return TPM_ERROR; }
  image = Boa_Volume_Value(imageObj);

  xlen = image->xdim - xsize + 1;
  xlow = xlen/xinc+!!(xlen%xinc);
  if(xlow % 16) {
    xhi = xlow/16*16;
    ErrorSet("Cannot use xsize=%d, xinc=%d; Try xsize=%d or xinc=%d,%d,%d", 
		xsize, xinc, image->xdim-xhi*xinc+1, 
		xlen/15, xlen/31, xlen/47);
    return TPM_ERROR;
  }
  ylen = image->ydim - ysize + 1;
  ylow = ylen/yinc+!!(ylen%yinc);
  if(ylow % 16) {
    yhi = ylow/16*16;
    ErrorSet("Cannot use ysize=%d, yinc=%d; Try ysize=%d or yinc=%d,%d,%d", 
		ysize, yinc, image->ydim-yhi*yinc+1, 
		ylen/15, ylen/31, ylen/47);
    return TPM_ERROR;
  }

#if 1
  /* omit image border, so all windows are same size */
  ylow = (ysize-1)/2; yhi = image->ydim - 1 - ysize/2; 
  xlow = (xsize-1)/2; xhi = image->xdim - 1 - xsize/2; 
#else
  ylow = 0; yhi = image->ydim - 1;
  xlow = 0; xhi = image->xdim - 1;
#endif
  ylen = yhi - ylow + 1;
  xlen = xhi - xlow + 1;
  result = VolumeCreate(depth, ylen/yinc+!!(ylen % yinc), 
			       xlen/xinc+!!(xlen % xinc));
  window = VolumeCreate(image->zdim, ysize, xsize);
  windowObj = Boa_Volume_Create(window);
  /* loop across window centers */
  for(y = ylow; y <= yhi; y += yinc) {
    fprintf(stderr, "row: %d\n", y);
    for(x = xlow; x <= xhi; x += xinc) {
      Obj t;
      /* compute window extents */
      xstart = x - (xsize-1)/2; if(xstart < 0) xstart = 0;
      xend = x + xsize/2; if(xend >= image->xdim) xend = image->xdim-1;
      ystart = y - (ysize-1)/2; if(ystart < 0) ystart = 0;
      yend = y + ysize/2; if(yend >= image->ydim) yend = image->ydim-1;
      window->xdim = xend - xstart + 1;
      window->ydim = yend - ystart + 1;
      window->zydim = window->ydim * window->zdim;
      window->len = window->xdim * window->zydim;
      /* copy image data into the window */
      for(i = ystart; i <= yend; i++)
	for(j = xstart; j <= xend; j++)
	  for(k = 0; k < window->zdim; k++)
	    window->d3[k][i-ystart][j-xstart] = image->d3[k][i][j]; 
      /* compute feature vector */
      if(Boa_ObjGet(func, windowObj, 0, &T_Vector, &t) != TPM_OK) {
	ErrorPrepend("Call failed for (y=%d,x=%d)", y, x);
	ErrorPrint(stderr);
	continue;
      }
      vector = Boa_Vector_Value(t);
      if(vector->len < depth) {
	fprintf(stderr, "vector->len (%d) < depth (%d)\n", vector->len, depth);
	continue;
      }
      /* copy result into volume */
      for(i=0;i<depth;i++)
	result->d3[i][(y-ylow)/yinc][(x-xlow)/xinc] = vector->d[i];
      Boa_Release(t);
    }
  }
  Boa_Release(windowObj);
  Boa_Release(func);
  Boa_Release(imageObj);
  *value_return = Boa_Volume_Create(result);
  return TPM_OK;

 error:
  Boa_Release(func);
  Boa_Release(imageObj);
  return TPM_ERROR;
}

/* Test func for WindowFeatures.
 * Input: image
 * Output: vector
 */
int WindowTest(Obj context, Type *type_in_out, Obj *value_return)
{
  Volume image;
  Vector v;
  Obj t;

  BOA_TYPECHECK_GET(T_Vector);
  TRY_RETURN(Boa_StringGet(context, "image", 0, &T_Volume, &t));
  image = Boa_Volume_Value(t);
  Boa_Release(t);
  v = VectorCreate(3);
  v->d[0] = image->zdim;
  v->d[1] = image->ydim;
  v->d[2] = image->xdim;
  VectorPrint(stderr, v);
  *value_return = Boa_Vector_Create(v);
  return TPM_OK;
}
