/*
 * @(#)$Header: /amnt/kona/tangram/u/william/src/breakps/RCS/breakps.c,v 1.5 91/12/21 01:10:04 william Exp $
 *
 * Author: William Chia-Wei Cheng (william@cs.ucla.edu)
 * Modified by: Thomas Peter Minka (tpminka@mit.edu)
 *
 * Usage: breakps [-strip] [-k <pages_per_file>] [-o <out_prefix>] < tut1.ps
 *
 *      Breakps breaks up a PostScript file, such as 'tut1.ps' or 'ibuild.ps'
 *      in the InterViews /pub/papers directory, into files of size
 *      <pages_per_file> pages.  It generates output files <filename>.<number>,
 *      where <number> is automatically generated from 1.
 *      Default <pages_per_file> is 30.  Default <filename> is '/tmp/breakps'.
 *
 *      By default, breakps will not strip lines that start with %%.  If for
 *      some reason, this is not desirable, the -strip will cause the %%
 *      lines to not be printed.
 *
 * Here is how breakps works.
 *
 *      Breakps looks for the following strings for PostScript files
 *      generated by dvi2ps:
 *
 *              @start
 *              <...> @bop0
 *              <...> @bop1
 *
 *      If the PostScript files are generated by FrameMaker, the breakps
 *      looks for the following strings:
 *
 *              %%EndSetup
 *              %%Page: <...>
 *              %%EndPage: <...>
 *
 *      It accumulates character/font definition between @bop0 and @bop1
 *      (or %%EndPage and %%Page) and add it to every output file.  (This
 *      is kind of clumsy, but it seems to be a working hack.)
 *
 *      This program is custom made for PostScript files generated by
 *      the dvi2ps and FrameMaker used to generate the PostScript files
 *      in the InterViews distribution.  It would probably not work for
 *      PostScript files generated by any other programs.  It might not
 *      even work for PostScript files generated by previous or future
 *      versions of dvi2ps and FrameMaker.
 *      
 * To compile, simply do:
 *
 *      cc -o breakps -g breakps.c
 *
 * To run breakps, try:
 *
 *      breakps -k 5 -o out < tut1.ps
 *
 * The output files, out.1 and out.2, can be printed with lpr.
 * (Sometimes, breakps will generate an extra output file at the end.)
 *
 */

#include <stdio.h>

#ifndef TRUE
#define FALSE 0
#define TRUE 1
#endif

#define ACCUMULATED_BOP "/tmp/breakps.bop"
#define SINGLE_BOP "/tmp/breakps.bop-1"
#define SEARCH_BOP0 0
#define SEARCH_BOP1 1

#define DVI2PS 0
#define FRAMEMAKER 1

#define USE_DVI2PS 0

struct StrRec {
   char                 * line;
   struct StrRec        * next;
};

static int              stripComment = FALSE;
static int              fileType = DVI2PS;
static struct StrRec    * firstStr = NULL, * lastStr = NULL;

void Usage ()
{
   fprintf (stderr, "\nUsage:  breakps %s %s %s %s %s\n",
         "[-strip]", "[-k <pages_per_file>]", "[-p <start_page>]", "[-n <num_pages>]", "[-o <out_prefix>]");
   exit (-1);
}

void ReadHeader ()
{
   struct StrRec        * str_ptr;
   char                 inbuf[256], * s;
   int                  ok = FALSE, len;

   while (fgets (inbuf, 256, stdin) != NULL)
   {
      len = strlen (inbuf);
      if (!(stripComment && strncmp ("%%",inbuf,2) == 0))
      {
         s = (char *) calloc (len+2, sizeof (char));
         strcpy (s, inbuf);

         str_ptr = (struct StrRec *) calloc (1, sizeof (struct StrRec));
         str_ptr->next = NULL;
         str_ptr->line = s;

         if (lastStr == NULL)
            firstStr = str_ptr;
         else
            lastStr->next = str_ptr;
         lastStr = str_ptr;
      }
      if (len >= 7)
      {
         if (USE_DVI2PS && strncmp (inbuf, "@start", 6) == 0)
         {
            ok = TRUE;
            fileType = DVI2PS;
            printf ("Looks like a PostScript file generated by '%s'...\n",
                  "dvi2ps");
            break;
         }
         else if (strncmp (inbuf, "%%EndSetup", 10) == 0)
         {
            ok = TRUE;
            fileType = FRAMEMAKER;
            printf ("Looks like a PostScript file generated by '%s'...\n",
                  "FrameMaker");
            break;
         }
	 else if (!strncmp(inbuf, "%%EndProlog", 11)) {
	   ok = TRUE;
	   fileType = FRAMEMAKER;
            printf ("Looks like a PostScript file generated by '%s'...\n",
		    "unknown");
            break;
	 }
      }
   }
   if (!ok)
   {
      printf ("Abort!  Can not find \"@start\", \"%%EndSetup\", or \"%%EndProlog\".\n");
      exit (-1);
   }
}

void PrintBopStuff (fp)
   FILE *fp;
{
   FILE *acbopfp;
   char inbuf[256];

   if ((acbopfp = fopen (ACCUMULATED_BOP, "r")) == NULL)
   {
      printf ("Abort!  Can not open %s\n", ACCUMULATED_BOP); exit (-1);
   }

   while (fgets (inbuf, 256, acbopfp) != NULL)
      if (!(stripComment && strncmp ("%%",inbuf,2) == 0))
         fprintf (fp, "%s", inbuf);
   fclose (acbopfp);
}

main (argc, argv)
   int  argc;
   char *argv[];
{
   char inbuf[256], * next_page_str, * c_ptr, file_name[256], tmp_str[256];
   int  i, count, ok, len, search_mode, first_time = TRUE;
   int hold_over = FALSE, pre_page;
   FILE * fp, *bopfp;
   char *fileName = NULL;
   int              startPage = 0;
   int              numPages = -1;
   int              pagesPerFile = 30;
   int total_pages = 0;

   register struct StrRec       * str_ptr;

   for (--argc, ++argv; argc > 0; --argc, ++argv)
   {
      if (**argv == '-')
      {
         if (strcmp (*argv,"-k") == 0)
         {
	   --argc; ++argv;
            pagesPerFile = atoi (*argv);
            if (pagesPerFile <= 0) Usage ();
         }
         else if (strcmp (*argv,"-o") == 0)
            fileName = (--argc, *++argv);
         else if (strcmp (*argv,"-strip") == 0)
            stripComment = TRUE;
	 else if (strcmp (*argv,"-p") == 0)
	   startPage = (--argc, atoi(*++argv));
	 else if (strcmp (*argv,"-n") == 0)
	   numPages = (--argc, atoi(*++argv));
         else
            Usage ();
      }
      else
         Usage ();
   }

   if (!fileName) fileName = "/tmp/breakps";
   printf ("%1d pages per file ...\n", pagesPerFile);

   ReadHeader ();

   sprintf (inbuf, "rm -f %s", ACCUMULATED_BOP); /* accumulated bop stuff */
   system (inbuf);
   sprintf (inbuf, "rm -f %s", SINGLE_BOP); /* bop stuff for one file */
   system (inbuf);
   sprintf (inbuf, "touch %s", ACCUMULATED_BOP);
   system (inbuf);

   for (count = 1; ; count++)
   {
      strcpy (file_name, fileName);
      sprintf (file_name, "%s.%1d", fileName, count);

      if ((fp = fopen (file_name, "w")) == NULL)
      {
         printf ("Abort!  Can not open %s\n", file_name);
         exit (-1);
      }
      printf ("Generating %s ...\n", file_name);
      for (str_ptr = firstStr; str_ptr != NULL; str_ptr = str_ptr->next)
         fprintf (fp, "%s", str_ptr->line);

      if (fileType==DVI2PS && first_time)
      {
         ok = FALSE;
         while (fgets (inbuf, 256, stdin) != NULL)
         {
            len = strlen (inbuf);
            if (len >= 8 && strncmp (&inbuf[len-6], "@bop0", 5)==0)
            {
               inbuf[len-7] = '\0';
               next_page_str = (char *) calloc (len-5, sizeof (char));
               strcpy (next_page_str, inbuf);
               ok = TRUE;
               search_mode = SEARCH_BOP1;
               break;
            }
            else if (len >= 8 && strncmp (&inbuf[len-6], "@bop1", 5) == 0)
            {
               inbuf[len-1] = '\0';
               printf ("Abort!  '%s' found while looking for the first @bop0\n",
                     inbuf);
               exit (-1);
            }
            if (!(stripComment && strncmp ("%%",inbuf,2) == 0))
               fprintf (fp, "%s", inbuf);
         }
         if (!ok)
         {
            printf ("Abort!  Can not find @bop0\n"); exit (-1);
         }
         first_time = FALSE;
      }
      else if (fileType == FRAMEMAKER && first_time)
      {
         next_page_str = (char *) calloc (20, sizeof (char));
         strcpy (next_page_str, "(don't know yet)");
         first_time = FALSE;
      }

      if ((bopfp = fopen (SINGLE_BOP, "w")) == NULL)
      {
         printf ("Abort!  Can not open %s\n", SINGLE_BOP); exit (-1);
      }

      search_mode = SEARCH_BOP1;
      for (i = 0; i < pagesPerFile; i++, total_pages++)
      {
	if(total_pages == numPages) { ok = FALSE; break; }
         if (fileType == DVI2PS)
         {
            printf ("\tPhysical page %1d.  \tLogical page %s.\n",
                  pagesPerFile*(count-1)+i+1, next_page_str);
            fprintf (fp, "%s @bop0\n", next_page_str);
            cfree (next_page_str);
         }
         else if (fileType == FRAMEMAKER)
         {
            printf ("\tPhysical page %1d.  \tLogical page %s.\n",
                  pagesPerFile*(count-1)+i+1, next_page_str);
            cfree (next_page_str);
         }

         if (i == 0) PrintBopStuff (fp);

	 pre_page = TRUE;
         ok = FALSE;
         while (!ok && (hold_over || fgets (inbuf, 256, stdin) != NULL))
         {
	    hold_over = FALSE;
            len = strlen (inbuf);

            if (fileType == DVI2PS && len >= 8 &&
                  strncmp (&inbuf[len-6], "@bop0", 5) == 0)
            {
               ok = TRUE;
               inbuf[len-7] = '\0';
               next_page_str = (char *) calloc (len-5, sizeof (char));
               strcpy (next_page_str, inbuf);
               if (search_mode == SEARCH_BOP1)
                  printf ("Error: '%s @bop0' found %s '@bop1'.\n",
                        next_page_str, "while looking for");
               search_mode = SEARCH_BOP1;
               break;
            }
            else if (fileType == FRAMEMAKER && len >= 10 &&
		     !strncmp(inbuf, "%%EndPage:", 10))
            {
               ok = TRUE;
               inbuf[len-1] = '\0';
               strcpy (tmp_str, &inbuf[11]);
               next_page_str = (char *) calloc (len-9, sizeof (char));
               strcpy (next_page_str, tmp_str);
               if (search_mode == SEARCH_BOP1)
                  printf ("Error: '%%%%EndPage: %s' found %s '%%%%Page'.\n",
                        tmp_str, "while looking for");
               search_mode = SEARCH_BOP1;
               break;
            }
            else if (fileType == FRAMEMAKER && len >= 7 &&
		     !strncmp(inbuf, "%%Page:", 7))
            {
	      if(!pre_page) {
		ok = TRUE;
		strcpy (tmp_str, inbuf + 8);
		tmp_str[len-9] = '\0';
		next_page_str = (char *) calloc (len-6, sizeof (char));
		strcpy (next_page_str, tmp_str);
		search_mode = SEARCH_BOP1;
		hold_over = TRUE;
		break;
	      }
	      pre_page = FALSE;
            }

            if (!(stripComment && strncmp ("%%",inbuf,2) == 0))
               fprintf (fp, "%s", inbuf);

            switch (search_mode)
            {
               case SEARCH_BOP1:
                  if (fileType == DVI2PS && len >= 8 &&
                        strncmp (&inbuf[len-6], "@bop0", 5) == 0)
                     printf ("Error: '@bop0' found %s '@bop1'.\n",
                           "while looking for");
                  else if (fileType == FRAMEMAKER && len >= 10 &&
                        strncmp (inbuf, "%%EndPage:", 10) == 0)
                     printf ("Error: '%s' found while looking for '%s'.\n",
                           inbuf, "%%Page");
                  else if (fileType == DVI2PS && len >= 8 &&
                        strncmp (&inbuf[len-6], "@bop1", 5) == 0)
                     search_mode = SEARCH_BOP0;
                  else if (fileType == FRAMEMAKER && len >= 7 &&
                        strncmp (inbuf, "%%Page:", 7) == 0)
                     search_mode = SEARCH_BOP0;
                  else
                  {
                     if (!(stripComment && strncmp ("%%",inbuf,2) == 0))
                        fprintf (bopfp, "%s", inbuf);
                  }
                  break;
            }
         }
         if (!ok) break;
      }
      fclose (bopfp);
      sprintf (tmp_str, "cat %s >> %s", SINGLE_BOP, ACCUMULATED_BOP);
      system (tmp_str);
      if (ok)
      {  /* end of a single output file */
         if (fileType == DVI2PS)
            fprintf (fp, "@end\n");
         fclose (fp);
      }
      else
      {  /* end of everything */
         fclose (fp);
         break;
      }
   }
   return (0);
}
